////////////////////////////////////////////////////////////////////////////////////////////////////
// fastest_sort_w4.h
// Copyright(c) 2018 ozk
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
////////////////////////////////////////////////////////////////////////////////////////////////////
#pragma once

#include <cassert>
#include <iterator>
#include <utility>


namespace fsw4
{

	namespace detail
	{

		//---------+---------+---------+---------+---------+---------+---------+---------+---------+
		// FastestSortW4Impl<T, Compare>
		//---------+---------+---------+---------+---------+---------+---------+---------+---------+
		template<typename T, typename Compare>
		class FastestSortW4Impl
		{
		public:

			inline static void stable_sort2(T &v0, T &v1, const Compare &compare)
			{
				using std::swap;
				if (compare(v1, v0)) {
					swap(v0, v1); // ba
				}
			}

			static void stable_sort3(T &v0, T &v1, T &v2, const Compare &compare)
			{
				using std::swap;
				if (compare(v1, v0)) {
					if (compare(v2, v1)) {
						swap(v0, v2); // cba
					}
					else {
						if (compare(v2, v0)) {
							swap(v0, v1); swap(v1, v2); // bca
						}
						else {
							swap(v0, v1); // bac
						}
					}
				}
				else {
					if (compare(v2, v1)) {
						if (compare(v2, v0)) {
							swap(v0, v2); swap(v1, v2); // cab
						}
						else {
							swap(v1, v2); // acb
						}
					}
					else {
						// abc
					}
				}
			}

			static void stable_sort4(T &v0, T &v1, T &v2, T &v3, const Compare &compare)
			{
				using std::swap;
				if (compare(v1, v0)) {
					if (compare(v3, v2)) {
						if (compare(v2, v0)) {
							if (compare(v2, v1)) {
								swap(v0, v3); swap(v1, v2); // dcba
							}
							else {
								if (compare(v3, v1)) {
									swap(v0, v3); // dbca
								}
								else {
									swap(v0, v1); swap(v1, v3); // bdca
								}
							}
						}
						else {
							if (compare(v3, v0)) {
								if (compare(v3, v1)) {
									swap(v0, v3); swap(v2, v3); // dbac
								}
								else {
									swap(v0, v1); swap(v1, v3); swap(v2, v3); // bdac
								}
							}
							else {
								swap(v0, v1); swap(v2, v3); // badc
							}
						}
					}
					else {
						if (compare(v3, v0)) {
							if (compare(v2, v1)) {
								if (compare(v3, v1)) {
									swap(v0, v2); swap(v1, v3); swap(v2, v3); // cdba
								}
								else {
									swap(v0, v2); swap(v2, v3); // cbda
								}
							}
							else {
								swap(v0, v1); swap(v1, v2); swap(v2, v3); // bcda
							}
						}
						else {
							if (compare(v2, v0)) {
								if (compare(v2, v1)) {
									swap(v0, v2); // cbad
								}
								else {
									swap(v0, v1); swap(v1, v2); // bcad
								}
							}
							else {
								swap(v0, v1); // bacd
							}
						}
					}
				}
				else {
					if (compare(v3, v2)) {
						if (compare(v3, v0)) {
							if (compare(v2, v0)) {
								swap(v0, v3); swap(v1, v2); swap(v2, v3); // dcab
							}
							else {
								if (compare(v2, v1)) {
									swap(v0, v3); swap(v1, v3); // dacb
								}
								else {
									swap(v0, v3); swap(v1, v3); swap(v2, v3); // dabc
								}
							}
						}
						else {
							if (compare(v2, v1)) {
								swap(v1, v3); // adcb
							}
							else {
								if (compare(v3, v1)) {
									swap(v1, v3); swap(v2, v3); // adbc
								}
								else {
									swap(v2, v3); // abdc
								}
							}
						}
					}
					else {
						if (compare(v2, v0)) {
							if (compare(v3, v0)) {
								swap(v0, v2); swap(v1, v3); // cdab
							}
							else {
								if (compare(v3, v1)) {
									swap(v0, v2); swap(v1, v2); swap(v2, v3); // cadb
								}
								else {
									swap(v0, v2); swap(v1, v2); // cabd
								}
							}
						}
						else {
							if (compare(v2, v1)) {
								if (compare(v3, v1)) {
									swap(v1, v2); swap(v2, v3); // acdb
								}
								else {
									swap(v1, v2); // acbd
								}
							}
							else {
								// abcd
							}
						}
					}
				}
			}
		};

	} // detail

	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	// fastest_sort_w4<T, Compare>
	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	template<typename T, typename Compare = std::less<>>
	class fastest_sort_w4
	{
	public:

		static constexpr size_t max_size() noexcept
		{
			return 4;
		}

		template<typename Iterator>
		static void sort(const Iterator &first, const Iterator &last, const Compare &compare = Compare{})
		{
			using Impl = detail::FastestSortW4Impl<T, Compare>;
			Iterator i = first;
			if (i == last) {
				return;
			}
			T &v0 = *i; ++i;
			if (i == last) {
				return;
			}
			T &v1 = *i; ++i;
			if (i == last) {
				Impl::stable_sort2(v0, v1, compare);
				return;
			}
			T &v2 = *i; ++i;
			if (i == last) {
				Impl::stable_sort3(v0, v1, v2, compare);
				return;
			}
			T &v3 = *i; ++i;
			if (i == last) {
				Impl::stable_sort4(v0, v1, v2, v3, compare);
				return;
			}
			assert(false && "Too many input data. Please add more source code.");
		}
	};

	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	// sort2(v0, v1, compare)
	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	template<typename T, typename Compare = std::less<>>
	inline void sort2(T &v0, T &v1, const Compare &compare = Compare{})
	{
		using Impl = detail::FastestSortW4Impl<T, Compare>;
		Impl::stable_sort2(v0, v1, compare);
	}

	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	// sort3(v0, v1, v2, compare)
	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	template<typename T, typename Compare = std::less<>>
	inline void sort3(T &v0, T &v1, T &v2, const Compare &compare = Compare{})
	{
		using Impl = detail::FastestSortW4Impl<T, Compare>;
		Impl::stable_sort3(v0, v1, v2, compare);
	}

	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	// sort4(v0, v1, v2, v3, compare)
	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	template<typename T, typename Compare = std::less<>>
	inline void sort4(T &v0, T &v1, T &v2, T &v3, const Compare &compare = Compare{})
	{
		using Impl = detail::FastestSortW4Impl<T, Compare>;
		Impl::stable_sort4(v0, v1, v2, v3, compare);
	}

	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	// sort(first, last, compare)
	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	template<typename Iterator, typename Compare = std::less<>>
	inline void sort(const Iterator &first, const Iterator &last, const Compare &compare = Compare{})
	{
		using Value = typename std::iterator_traits<Iterator>::value_type;
		fastest_sort_w4<Value, Compare>::sort(first, last, compare);
	}

} // fsw4
