////////////////////////////////////////////////////////////////////////////////////////////////////
// fastest_sort_w8.h
// Copyright(c) 2018 ozk
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
////////////////////////////////////////////////////////////////////////////////////////////////////
#pragma once

#include <array>
#include <cassert>
#include <iterator>
#include <utility>


namespace fsw8
{

	namespace detail
	{

		//---------+---------+---------+---------+---------+---------+---------+---------+---------+
		// FastestSortW8Impl<T, Compare>
		//---------+---------+---------+---------+---------+---------+---------+---------+---------+
		template<typename T, typename Compare>
		class FastestSortW8Impl
		{
		public:

			inline static void stable_sort2(T &v0, T &v1, const Compare &compare)
			{
				using std::swap;
				if (compare(v1, v0)) {
					swap(v0, v1); // ba
				}
			}

			static void stable_sort3(T &v0, T &v1, T &v2, const Compare &compare)
			{
				using std::swap;
				if (compare(v1, v0)) {
					if (compare(v2, v1)) {
						swap(v0, v2); // cba
					}
					else {
						if (compare(v2, v0)) {
							swap(v0, v1); swap(v1, v2); // bca
						}
						else {
							swap(v0, v1); // bac
						}
					}
				}
				else {
					if (compare(v2, v1)) {
						if (compare(v2, v0)) {
							swap(v0, v2); swap(v1, v2); // cab
						}
						else {
							swap(v1, v2); // acb
						}
					}
					else {
						// abc
					}
				}
			}

			static void stable_sort4(T &v0, T &v1, T &v2, T &v3, const Compare &compare)
			{
				using std::swap;
				if (compare(v1, v0)) {
					if (compare(v3, v2)) {
						if (compare(v2, v0)) {
							if (compare(v2, v1)) {
								swap(v0, v3); swap(v1, v2); // dcba
							}
							else {
								if (compare(v3, v1)) {
									swap(v0, v3); // dbca
								}
								else {
									swap(v0, v1); swap(v1, v3); // bdca
								}
							}
						}
						else {
							if (compare(v3, v0)) {
								if (compare(v3, v1)) {
									swap(v0, v3); swap(v2, v3); // dbac
								}
								else {
									swap(v0, v1); swap(v1, v3); swap(v2, v3); // bdac
								}
							}
							else {
								swap(v0, v1); swap(v2, v3); // badc
							}
						}
					}
					else {
						if (compare(v3, v0)) {
							if (compare(v2, v1)) {
								if (compare(v3, v1)) {
									swap(v0, v2); swap(v1, v3); swap(v2, v3); // cdba
								}
								else {
									swap(v0, v2); swap(v2, v3); // cbda
								}
							}
							else {
								swap(v0, v1); swap(v1, v2); swap(v2, v3); // bcda
							}
						}
						else {
							if (compare(v2, v0)) {
								if (compare(v2, v1)) {
									swap(v0, v2); // cbad
								}
								else {
									swap(v0, v1); swap(v1, v2); // bcad
								}
							}
							else {
								swap(v0, v1); // bacd
							}
						}
					}
				}
				else {
					if (compare(v3, v2)) {
						if (compare(v3, v0)) {
							if (compare(v2, v0)) {
								swap(v0, v3); swap(v1, v2); swap(v2, v3); // dcab
							}
							else {
								if (compare(v2, v1)) {
									swap(v0, v3); swap(v1, v3); // dacb
								}
								else {
									swap(v0, v3); swap(v1, v3); swap(v2, v3); // dabc
								}
							}
						}
						else {
							if (compare(v2, v1)) {
								swap(v1, v3); // adcb
							}
							else {
								if (compare(v3, v1)) {
									swap(v1, v3); swap(v2, v3); // adbc
								}
								else {
									swap(v2, v3); // abdc
								}
							}
						}
					}
					else {
						if (compare(v2, v0)) {
							if (compare(v3, v0)) {
								swap(v0, v2); swap(v1, v3); // cdab
							}
							else {
								if (compare(v3, v1)) {
									swap(v0, v2); swap(v1, v2); swap(v2, v3); // cadb
								}
								else {
									swap(v0, v2); swap(v1, v2); // cabd
								}
							}
						}
						else {
							if (compare(v2, v1)) {
								if (compare(v3, v1)) {
									swap(v1, v2); swap(v2, v3); // acdb
								}
								else {
									swap(v1, v2); // acbd
								}
							}
							else {
								// abcd
							}
						}
					}
				}
			}

			static void stable_sort5(T &v0, T &v1, T &v2, T &v3, T &v4, const Compare &compare)
			{
				using std::swap;
				if (compare(v1, v0)) {
					if (compare(v3, v2)) {
						if (compare(v2, v0)) {
							if (compare(v4, v2)) {
								if (compare(v4, v3)) {
									if (compare(v3, v1)) {
										if (compare(v2, v1)) {
											swap(v0, v4); swap(v1, v3); // edcba
										}
										else {
											swap(v0, v4); swap(v1, v3); swap(v2, v3); // edbca
										}
									}
									else {
										if (compare(v4, v1)) {
											swap(v0, v4); swap(v2, v3); // ebdca
										}
										else {
											swap(v0, v1); swap(v1, v4); swap(v2, v3); // bedca
										}
									}
								}
								else {
									if (compare(v4, v1)) {
										if (compare(v2, v1)) {
											swap(v0, v3); swap(v1, v4); swap(v3, v4); // decba
										}
										else {
											swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); // debca
										}
									}
									else {
										if (compare(v3, v1)) {
											swap(v0, v3); swap(v2, v4); swap(v3, v4); // dbeca
										}
										else {
											swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); // bdeca
										}
									}
								}
							}
							else {
								if (compare(v4, v0)) {
									if (compare(v2, v1)) {
										if (compare(v4, v1)) {
											swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); // dceba
										}
										else {
											swap(v0, v3); swap(v1, v2); swap(v3, v4); // dcbea
										}
									}
									else {
										if (compare(v3, v1)) {
											swap(v0, v3); swap(v3, v4); // dbcea
										}
										else {
											swap(v0, v1); swap(v1, v3); swap(v3, v4); // bdcea
										}
									}
								}
								else {
									if (compare(v2, v1)) {
										swap(v0, v3); swap(v1, v2); // dcbae
									}
									else {
										if (compare(v3, v1)) {
											swap(v0, v3); // dbcae
										}
										else {
											swap(v0, v1); swap(v1, v3); // bdcae
										}
									}
								}
							}
						}
						else {
							if (compare(v4, v0)) {
								if (compare(v4, v1)) {
									if (compare(v3, v1)) {
										if (compare(v4, v3)) {
											swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); // edbac
										}
										else {
											swap(v0, v3); swap(v1, v4); swap(v2, v4); // debac
										}
									}
									else {
										if (compare(v3, v0)) {
											swap(v0, v4); swap(v2, v3); swap(v3, v4); // ebdac
										}
										else {
											swap(v0, v4); swap(v2, v4); // ebadc
										}
									}
								}
								else {
									if (compare(v4, v3)) {
										if (compare(v3, v0)) {
											swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); // bedac
										}
										else {
											swap(v0, v1); swap(v1, v4); swap(v2, v4); // beadc
										}
									}
									else {
										if (compare(v3, v1)) {
											swap(v0, v3); swap(v2, v4); // dbeac
										}
										else {
											swap(v0, v1); swap(v1, v3); swap(v2, v4); // bdeac
										}
									}
								}
							}
							else {
								if (compare(v3, v0)) {
									if (compare(v3, v1)) {
										if (compare(v4, v2)) {
											swap(v0, v3); swap(v2, v3); swap(v3, v4); // dbaec
										}
										else {
											swap(v0, v3); swap(v2, v3); // dbace
										}
									}
									else {
										if (compare(v4, v2)) {
											swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); // bdaec
										}
										else {
											swap(v0, v1); swap(v1, v3); swap(v2, v3); // bdace
										}
									}
								}
								else {
									if (compare(v4, v2)) {
										if (compare(v4, v3)) {
											swap(v0, v1); swap(v2, v4); // baedc
										}
										else {
											swap(v0, v1); swap(v2, v3); swap(v3, v4); // badec
										}
									}
									else {
										swap(v0, v1); swap(v2, v3); // badce
									}
								}
							}
						}
					}
					else {
						if (compare(v3, v0)) {
							if (compare(v4, v3)) {
								if (compare(v4, v2)) {
									if (compare(v2, v1)) {
										if (compare(v3, v1)) {
											swap(v0, v4); swap(v1, v2); swap(v2, v3); // ecdba
										}
										else {
											swap(v0, v4); swap(v1, v2); // ecbda
										}
									}
									else {
										if (compare(v4, v1)) {
											swap(v0, v4); // ebcda
										}
										else {
											swap(v0, v1); swap(v1, v4); // becda
										}
									}
								}
								else {
									if (compare(v4, v1)) {
										if (compare(v3, v1)) {
											swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); // cedba
										}
										else {
											swap(v0, v2); swap(v1, v4); swap(v2, v4); // cebda
										}
									}
									else {
										if (compare(v2, v1)) {
											swap(v0, v2); swap(v2, v4); // cbeda
										}
										else {
											swap(v0, v1); swap(v1, v2); swap(v2, v4); // bceda
										}
									}
								}
							}
							else {
								if (compare(v4, v0)) {
									if (compare(v3, v1)) {
										if (compare(v4, v1)) {
											swap(v0, v2); swap(v1, v3); swap(v2, v4); // cdeba
										}
										else {
											swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); // cdbea
										}
									}
									else {
										if (compare(v2, v1)) {
											swap(v0, v2); swap(v2, v3); swap(v3, v4); // cbdea
										}
										else {
											swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); // bcdea
										}
									}
								}
								else {
									if (compare(v2, v1)) {
										if (compare(v3, v1)) {
											swap(v0, v2); swap(v1, v3); swap(v2, v3); // cdbae
										}
										else {
											swap(v0, v2); swap(v2, v3); // cbdae
										}
									}
									else {
										swap(v0, v1); swap(v1, v2); swap(v2, v3); // bcdae
									}
								}
							}
						}
						else {
							if (compare(v4, v0)) {
								if (compare(v4, v1)) {
									if (compare(v2, v1)) {
										if (compare(v4, v2)) {
											swap(v0, v4); swap(v1, v2); swap(v3, v4); // ecbad
										}
										else {
											swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); // cebad
										}
									}
									else {
										if (compare(v2, v0)) {
											swap(v0, v4); swap(v3, v4); // ebcad
										}
										else {
											swap(v0, v4); swap(v2, v4); swap(v3, v4); // ebacd
										}
									}
								}
								else {
									if (compare(v4, v2)) {
										if (compare(v2, v0)) {
											swap(v0, v1); swap(v1, v4); swap(v3, v4); // becad
										}
										else {
											swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); // beacd
										}
									}
									else {
										if (compare(v2, v1)) {
											swap(v0, v2); swap(v2, v4); swap(v3, v4); // cbead
										}
										else {
											swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); // bcead
										}
									}
								}
							}
							else {
								if (compare(v2, v0)) {
									if (compare(v2, v1)) {
										if (compare(v4, v3)) {
											swap(v0, v2); swap(v3, v4); // cbaed
										}
										else {
											swap(v0, v2); // cbade
										}
									}
									else {
										if (compare(v4, v3)) {
											swap(v0, v1); swap(v1, v2); swap(v3, v4); // bcaed
										}
										else {
											swap(v0, v1); swap(v1, v2); // bcade
										}
									}
								}
								else {
									if (compare(v4, v2)) {
										swap(v0, v1); swap(v2, v4); swap(v3, v4); // baecd
									}
									else {
										if (compare(v4, v3)) {
											swap(v0, v1); swap(v3, v4); // baced
										}
										else {
											swap(v0, v1); // bacde
										}
									}
								}
							}
						}
					}
				}
				else {
					if (compare(v3, v2)) {
						if (compare(v3, v0)) {
							if (compare(v4, v0)) {
								if (compare(v2, v0)) {
									if (compare(v4, v2)) {
										if (compare(v4, v3)) {
											swap(v0, v4); swap(v1, v3); swap(v3, v4); // edcab
										}
										else {
											swap(v0, v3); swap(v1, v4); // decab
										}
									}
									else {
										swap(v0, v3); swap(v1, v2); swap(v2, v4); // dceab
									}
								}
								else {
									if (compare(v2, v1)) {
										if (compare(v4, v3)) {
											swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); // edacb
										}
										else {
											swap(v0, v3); swap(v1, v4); swap(v2, v3); // deacb
										}
									}
									else {
										if (compare(v4, v3)) {
											swap(v0, v4); swap(v1, v3); swap(v2, v4); // edabc
										}
										else {
											swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); // deabc
										}
									}
								}
							}
							else {
								if (compare(v4, v1)) {
									if (compare(v4, v2)) {
										if (compare(v2, v1)) {
											swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); // daecb
										}
										else {
											swap(v0, v3); swap(v1, v3); swap(v2, v4); // daebc
										}
									}
									else {
										if (compare(v2, v0)) {
											swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); // dcaeb
										}
										else {
											swap(v0, v3); swap(v1, v3); swap(v3, v4); // daceb
										}
									}
								}
								else {
									if (compare(v2, v1)) {
										if (compare(v2, v0)) {
											swap(v0, v3); swap(v1, v2); swap(v2, v3); // dcabe
										}
										else {
											swap(v0, v3); swap(v1, v3); // dacbe
										}
									}
									else {
										if (compare(v4, v2)) {
											swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); // dabec
										}
										else {
											swap(v0, v3); swap(v1, v3); swap(v2, v3); // dabce
										}
									}
								}
							}
						}
						else {
							if (compare(v4, v3)) {
								if (compare(v4, v0)) {
									if (compare(v2, v1)) {
										swap(v0, v4); swap(v1, v4); swap(v2, v3); // eadcb
									}
									else {
										if (compare(v3, v1)) {
											swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); // eadbc
										}
										else {
											swap(v0, v4); swap(v1, v4); swap(v2, v4); // eabdc
										}
									}
								}
								else {
									if (compare(v3, v1)) {
										if (compare(v2, v1)) {
											swap(v1, v4); swap(v2, v3); // aedcb
										}
										else {
											swap(v1, v4); swap(v2, v3); swap(v3, v4); // aedbc
										}
									}
									else {
										if (compare(v4, v1)) {
											swap(v1, v4); swap(v2, v4); // aebdc
										}
										else {
											swap(v2, v4); // abedc
										}
									}
								}
							}
							else {
								if (compare(v4, v2)) {
									if (compare(v4, v1)) {
										if (compare(v2, v1)) {
											swap(v1, v3); swap(v2, v4); swap(v3, v4); // adecb
										}
										else {
											swap(v1, v3); swap(v2, v4); // adebc
										}
									}
									else {
										if (compare(v3, v1)) {
											swap(v1, v3); swap(v2, v3); swap(v3, v4); // adbec
										}
										else {
											swap(v2, v3); swap(v3, v4); // abdec
										}
									}
								}
								else {
									if (compare(v2, v1)) {
										if (compare(v4, v1)) {
											swap(v1, v3); swap(v3, v4); // adceb
										}
										else {
											swap(v1, v3); // adcbe
										}
									}
									else {
										if (compare(v3, v1)) {
											swap(v1, v3); swap(v2, v3); // adbce
										}
										else {
											swap(v2, v3); // abdce
										}
									}
								}
							}
						}
					}
					else {
						if (compare(v2, v0)) {
							if (compare(v4, v0)) {
								if (compare(v3, v0)) {
									if (compare(v4, v2)) {
										swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); // ecdab
									}
									else {
										if (compare(v4, v3)) {
											swap(v0, v2); swap(v1, v4); swap(v2, v3); // cedab
										}
										else {
											swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); // cdeab
										}
									}
								}
								else {
									if (compare(v3, v1)) {
										if (compare(v4, v2)) {
											swap(v0, v4); swap(v1, v2); swap(v2, v4); // ecadb
										}
										else {
											swap(v0, v2); swap(v1, v4); // ceadb
										}
									}
									else {
										if (compare(v4, v2)) {
											swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); // ecabd
										}
										else {
											swap(v0, v2); swap(v1, v4); swap(v3, v4); // ceabd
										}
									}
								}
							}
							else {
								if (compare(v4, v1)) {
									if (compare(v4, v3)) {
										if (compare(v3, v1)) {
											swap(v0, v2); swap(v1, v2); swap(v2, v4); // caedb
										}
										else {
											swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); // caebd
										}
									}
									else {
										if (compare(v3, v0)) {
											swap(v0, v2); swap(v1, v3); swap(v3, v4); // cdaeb
										}
										else {
											swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); // cadeb
										}
									}
								}
								else {
									if (compare(v3, v1)) {
										if (compare(v3, v0)) {
											swap(v0, v2); swap(v1, v3); // cdabe
										}
										else {
											swap(v0, v2); swap(v1, v2); swap(v2, v3); // cadbe
										}
									}
									else {
										if (compare(v4, v3)) {
											swap(v0, v2); swap(v1, v2); swap(v3, v4); // cabed
										}
										else {
											swap(v0, v2); swap(v1, v2); // cabde
										}
									}
								}
							}
						}
						else {
							if (compare(v4, v2)) {
								if (compare(v4, v0)) {
									if (compare(v2, v1)) {
										if (compare(v3, v1)) {
											swap(v0, v4); swap(v1, v4); // eacdb
										}
										else {
											swap(v0, v4); swap(v1, v4); swap(v3, v4); // eacbd
										}
									}
									else {
										swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); // eabcd
									}
								}
								else {
									if (compare(v2, v1)) {
										if (compare(v3, v1)) {
											swap(v1, v4); // aecdb
										}
										else {
											swap(v1, v4); swap(v3, v4); // aecbd
										}
									}
									else {
										if (compare(v4, v1)) {
											swap(v1, v4); swap(v2, v4); swap(v3, v4); // aebcd
										}
										else {
											swap(v2, v4); swap(v3, v4); // abecd
										}
									}
								}
							}
							else {
								if (compare(v4, v3)) {
									if (compare(v4, v1)) {
										if (compare(v3, v1)) {
											swap(v1, v2); swap(v2, v4); // acedb
										}
										else {
											swap(v1, v2); swap(v2, v4); swap(v3, v4); // acebd
										}
									}
									else {
										if (compare(v2, v1)) {
											swap(v1, v2); swap(v3, v4); // acbed
										}
										else {
											swap(v3, v4); // abced
										}
									}
								}
								else {
									if (compare(v3, v1)) {
										if (compare(v4, v1)) {
											swap(v1, v2); swap(v2, v3); swap(v3, v4); // acdeb
										}
										else {
											swap(v1, v2); swap(v2, v3); // acdbe
										}
									}
									else {
										if (compare(v2, v1)) {
											swap(v1, v2); // acbde
										}
										else {
											// abcde
										}
									}
								}
							}
						}
					}
				}
			}

			static void stable_sort6(T &v0, T &v1, T &v2, T &v3, T &v4, T &v5, const Compare &compare)
			{
				using std::swap;
				if (compare(v1, v0)) {
					if (compare(v3, v2)) {
						if (compare(v2, v0)) {
							if (compare(v5, v4)) {
								if (compare(v5, v1)) {
									if (compare(v2, v1)) {
										if (compare(v4, v1)) {
											if (compare(v4, v2)) {
												if (compare(v4, v3)) {
													swap(v0, v5); swap(v1, v4); swap(v2, v3); // fedcba
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); // fdecba
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // dfecba
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v3, v4); // fdceba
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v5); // dfceba
													}
												}
												else {
													swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dcfeba
												}
											}
										}
										else {
											if (compare(v4, v0)) {
												if (compare(v5, v2)) {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); // fdcbea
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v3, v5); // dfcbea
													}
												}
												else {
													swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); // dcfbea
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v4, v5); // fdcbae
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v5); // dfcbae
													}
												}
												else {
													swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dcfbae
												}
											}
										}
									}
									else {
										if (compare(v4, v2)) {
											if (compare(v4, v1)) {
												if (compare(v4, v3)) {
													if (compare(v3, v1)) {
														swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); // fedbca
													}
													else {
														swap(v0, v5); swap(v1, v4); swap(v2, v4); // febdca
													}
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v2, v4); // fdebca
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); // dfebca
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); // fdbeca
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dfbeca
													}
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v5); swap(v2, v4); // fbedca
													}
													else {
														swap(v0, v5); swap(v2, v3); swap(v3, v4); // fbdeca
													}
												}
											}
										}
										else {
											if (compare(v4, v0)) {
												if (compare(v3, v1)) {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v2, v3); // fdbcea
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); // dfbcea
													}
												}
												else {
													swap(v0, v5); swap(v2, v3); // fbdcea
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v5); // fdbcae
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dfbcae
													}
												}
												else {
													swap(v0, v5); swap(v2, v3); swap(v4, v5); // fbdcae
												}
											}
										}
									}
								}
								else {
									if (compare(v4, v0)) {
										if (compare(v3, v1)) {
											if (compare(v5, v2)) {
												if (compare(v4, v2)) {
													swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dbfeca
												}
												else {
													swap(v0, v3); swap(v2, v5); swap(v3, v5); // dbfcea
												}
											}
											else {
												if (compare(v2, v1)) {
													swap(v0, v3); swap(v1, v2); swap(v3, v5); // dcbfea
												}
												else {
													swap(v0, v3); swap(v3, v5); // dbcfea
												}
											}
										}
										else {
											if (compare(v4, v2)) {
												if (compare(v4, v3)) {
													swap(v0, v1); swap(v1, v5); swap(v2, v4); // bfedca
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); // bfdeca
													}
													else {
														swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bdfeca
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v1, v5); swap(v2, v3); // bfdcea
													}
													else {
														swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); // bdfcea
													}
												}
												else {
													swap(v0, v1); swap(v1, v3); swap(v3, v5); // bdcfea
												}
											}
										}
									}
									else {
										if (compare(v3, v1)) {
											if (compare(v5, v0)) {
												if (compare(v2, v1)) {
													swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v5); // dcbfae
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dbfcae
													}
													else {
														swap(v0, v3); swap(v3, v5); swap(v4, v5); // dbcfae
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													swap(v0, v3); swap(v1, v2); swap(v4, v5); // dcbafe
												}
												else {
													swap(v0, v3); swap(v4, v5); // dbcafe
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v5, v3)) {
													swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v5); // bfdcae
												}
												else {
													swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bdfcae
												}
											}
											else {
												if (compare(v5, v0)) {
													swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v5); // bdcfae
												}
												else {
													swap(v0, v1); swap(v1, v3); swap(v4, v5); // bdcafe
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v4, v1)) {
									if (compare(v2, v1)) {
										if (compare(v5, v1)) {
											if (compare(v5, v2)) {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v5); // efdcba
													}
													else {
														swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // edfcba
													}
												}
												else {
													swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); // defcba
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v3)) {
														swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v5); // edcfba
													}
													else {
														swap(v0, v3); swap(v1, v4); swap(v3, v5); // decfba
													}
												}
												else {
													swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); // dcefba
												}
											}
										}
										else {
											if (compare(v5, v0)) {
												if (compare(v4, v2)) {
													if (compare(v4, v3)) {
														swap(v0, v4); swap(v1, v3); swap(v4, v5); // edcbfa
													}
													else {
														swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v5); // decbfa
													}
												}
												else {
													swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // dcebfa
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v3)) {
														swap(v0, v4); swap(v1, v3); // edcbaf
													}
													else {
														swap(v0, v3); swap(v1, v4); swap(v3, v4); // decbaf
													}
												}
												else {
													swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); // dcebaf
												}
											}
										}
									}
									else {
										if (compare(v5, v2)) {
											if (compare(v5, v1)) {
												if (compare(v5, v3)) {
													if (compare(v3, v1)) {
														swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // efdbca
													}
													else {
														swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v5); // efbdca
													}
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v5); // edfbca
													}
													else {
														swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // defbca
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v4, v3)) {
														swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // edbfca
													}
													else {
														swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); // debfca
													}
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v2, v5); swap(v4, v5); // ebfdca
													}
													else {
														swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // ebdfca
													}
												}
											}
										}
										else {
											if (compare(v5, v0)) {
												if (compare(v3, v1)) {
													if (compare(v4, v3)) {
														swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v5); // edbcfa
													}
													else {
														swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // debcfa
													}
												}
												else {
													swap(v0, v4); swap(v2, v3); swap(v4, v5); // ebdcfa
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v4, v3)) {
														swap(v0, v4); swap(v1, v3); swap(v2, v3); // edbcaf
													}
													else {
														swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); // debcaf
													}
												}
												else {
													swap(v0, v4); swap(v2, v3); // ebdcaf
												}
											}
										}
									}
								}
								else {
									if (compare(v5, v0)) {
										if (compare(v3, v1)) {
											if (compare(v4, v2)) {
												if (compare(v5, v2)) {
													swap(v0, v3); swap(v2, v4); swap(v3, v5); // dbefca
												}
												else {
													swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // dbecfa
												}
											}
											else {
												if (compare(v2, v1)) {
													swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v5); // dcbefa
												}
												else {
													swap(v0, v3); swap(v3, v4); swap(v4, v5); // dbcefa
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v5); // befdca
													}
													else {
														swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bedfca
													}
												}
												else {
													swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); // bdefca
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v3)) {
														swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v5); // bedcfa
													}
													else {
														swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // bdecfa
													}
												}
												else {
													swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v5); // bdcefa
												}
											}
										}
									}
									else {
										if (compare(v3, v1)) {
											if (compare(v4, v0)) {
												if (compare(v2, v1)) {
													swap(v0, v3); swap(v1, v2); swap(v3, v4); // dcbeaf
												}
												else {
													if (compare(v4, v2)) {
														swap(v0, v3); swap(v2, v4); swap(v3, v4); // dbecaf
													}
													else {
														swap(v0, v3); swap(v3, v4); // dbceaf
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													swap(v0, v3); swap(v1, v2); // dcbaef
												}
												else {
													swap(v0, v3); // dbcaef
												}
											}
										}
										else {
											if (compare(v4, v2)) {
												if (compare(v4, v3)) {
													swap(v0, v1); swap(v1, v4); swap(v2, v3); // bedcaf
												}
												else {
													swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); // bdecaf
												}
											}
											else {
												if (compare(v4, v0)) {
													swap(v0, v1); swap(v1, v3); swap(v3, v4); // bdceaf
												}
												else {
													swap(v0, v1); swap(v1, v3); // bdcaef
												}
											}
										}
									}
								}
							}
						}
						else {
							if (compare(v5, v4)) {
								if (compare(v5, v1)) {
									if (compare(v4, v0)) {
										if (compare(v3, v1)) {
											if (compare(v4, v1)) {
												if (compare(v4, v3)) {
													swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fedbac
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v5); // fdebac
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // dfebac
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fdbeac
												}
												else {
													swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); // dfbeac
												}
											}
										}
										else {
											if (compare(v3, v0)) {
												if (compare(v4, v1)) {
													swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v5); // febdac
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v5); swap(v2, v4); swap(v4, v5); // fbedac
													}
													else {
														swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fbdeac
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // febadc
												}
												else {
													swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // fbeadc
												}
											}
										}
									}
									else {
										if (compare(v3, v1)) {
											if (compare(v4, v2)) {
												if (compare(v5, v3)) {
													swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); // fdbaec
												}
												else {
													swap(v0, v3); swap(v1, v5); swap(v2, v5); // dfbaec
												}
											}
											else {
												if (compare(v5, v3)) {
													swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fdbace
												}
												else {
													swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v5); // dfbace
												}
											}
										}
										else {
											if (compare(v3, v0)) {
												if (compare(v4, v2)) {
													swap(v0, v5); swap(v2, v3); swap(v3, v5); // fbdaec
												}
												else {
													swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fbdace
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v3)) {
														swap(v0, v5); swap(v2, v5); swap(v3, v4); // fbaedc
													}
													else {
														swap(v0, v5); swap(v2, v5); // fbadec
													}
												}
												else {
													swap(v0, v5); swap(v2, v5); swap(v4, v5); // fbadce
												}
											}
										}
									}
								}
								else {
									if (compare(v5, v0)) {
										if (compare(v5, v3)) {
											if (compare(v3, v0)) {
												if (compare(v4, v0)) {
													if (compare(v4, v3)) {
														swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v5); // bfedac
													}
													else {
														swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bfdeac
													}
												}
												else {
													if (compare(v4, v2)) {
														swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); // bfdaec
													}
													else {
														swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bfdace
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v4, v0)) {
														swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // bfeadc
													}
													else {
														swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); // bfaedc
													}
												}
												else {
													if (compare(v4, v2)) {
														swap(v0, v1); swap(v1, v5); swap(v2, v5); // bfadec
													}
													else {
														swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v5); // bfadce
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v4, v0)) {
													swap(v0, v3); swap(v2, v5); swap(v3, v4); // dbfeac
												}
												else {
													if (compare(v4, v2)) {
														swap(v0, v3); swap(v2, v5); // dbfaec
													}
													else {
														swap(v0, v3); swap(v2, v5); swap(v4, v5); // dbface
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); // bdfeac
												}
												else {
													if (compare(v4, v2)) {
														swap(v0, v1); swap(v1, v3); swap(v2, v5); // bdfaec
													}
													else {
														swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v5); // bdface
													}
												}
											}
										}
									}
									else {
										if (compare(v3, v0)) {
											if (compare(v3, v1)) {
												if (compare(v4, v2)) {
													swap(v0, v3); swap(v2, v3); swap(v3, v5); // dbafec
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dbafce
													}
													else {
														swap(v0, v3); swap(v2, v3); swap(v4, v5); // dbacfe
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); // bdafec
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bdafce
													}
													else {
														swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v5); // bdacfe
													}
												}
											}
										}
										else {
											if (compare(v4, v2)) {
												if (compare(v4, v3)) {
													swap(v0, v1); swap(v2, v5); swap(v3, v4); // bafedc
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v2, v5); // bafdec
													}
													else {
														swap(v0, v1); swap(v2, v3); swap(v3, v5); // badfec
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v2, v5); swap(v4, v5); // bafdce
													}
													else {
														swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v5); // badfce
													}
												}
												else {
													swap(v0, v1); swap(v2, v3); swap(v4, v5); // badcfe
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v4, v1)) {
									if (compare(v5, v0)) {
										if (compare(v3, v1)) {
											if (compare(v5, v1)) {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); // efdbac
													}
													else {
														swap(v0, v4); swap(v1, v3); swap(v2, v5); // edfbac
													}
												}
												else {
													swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); // defbac
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); // edbfac
												}
												else {
													swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // debfac
												}
											}
										}
										else {
											if (compare(v3, v0)) {
												if (compare(v5, v1)) {
													swap(v0, v4); swap(v1, v5); swap(v2, v5); // efbdac
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v2, v5); // ebfdac
													}
													else {
														swap(v0, v4); swap(v2, v3); swap(v3, v5); // ebdfac
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); // efbadc
												}
												else {
													swap(v0, v4); swap(v2, v5); swap(v3, v4); // ebfadc
												}
											}
										}
									}
									else {
										if (compare(v3, v1)) {
											if (compare(v5, v2)) {
												if (compare(v4, v3)) {
													swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // edbafc
												}
												else {
													swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v5); // debafc
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); // edbacf
												}
												else {
													swap(v0, v3); swap(v1, v4); swap(v2, v4); // debacf
												}
											}
										}
										else {
											if (compare(v3, v0)) {
												if (compare(v5, v2)) {
													swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // ebdafc
												}
												else {
													swap(v0, v4); swap(v2, v3); swap(v3, v4); // ebdacf
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // ebafdc
													}
													else {
														swap(v0, v4); swap(v2, v4); swap(v4, v5); // ebadfc
													}
												}
												else {
													swap(v0, v4); swap(v2, v4); // ebadcf
												}
											}
										}
									}
								}
								else {
									if (compare(v4, v0)) {
										if (compare(v4, v3)) {
											if (compare(v3, v0)) {
												if (compare(v5, v0)) {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v1, v4); swap(v2, v5); // befdac
													}
													else {
														swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); // bedfac
													}
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bedafc
													}
													else {
														swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); // bedacf
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v5, v0)) {
														swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); // befadc
													}
													else {
														swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // beafdc
													}
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v5); // beadfc
													}
													else {
														swap(v0, v1); swap(v1, v4); swap(v2, v4); // beadcf
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v5, v0)) {
													swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // dbefac
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v3); swap(v2, v4); swap(v4, v5); // dbeafc
													}
													else {
														swap(v0, v3); swap(v2, v4); // dbeacf
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // bdefac
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v5); // bdeafc
													}
													else {
														swap(v0, v1); swap(v1, v3); swap(v2, v4); // bdeacf
													}
												}
											}
										}
									}
									else {
										if (compare(v3, v0)) {
											if (compare(v3, v1)) {
												if (compare(v4, v2)) {
													if (compare(v5, v2)) {
														swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dbaefc
													}
													else {
														swap(v0, v3); swap(v2, v3); swap(v3, v4); // dbaecf
													}
												}
												else {
													swap(v0, v3); swap(v2, v3); // dbacef
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v5, v2)) {
														swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bdaefc
													}
													else {
														swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); // bdaecf
													}
												}
												else {
													swap(v0, v1); swap(v1, v3); swap(v2, v3); // bdacef
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v5); // baefdc
													}
													else {
														swap(v0, v1); swap(v2, v4); swap(v4, v5); // baedfc
													}
												}
												else {
													swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v5); // badefc
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v3)) {
														swap(v0, v1); swap(v2, v4); // baedcf
													}
													else {
														swap(v0, v1); swap(v2, v3); swap(v3, v4); // badecf
													}
												}
												else {
													swap(v0, v1); swap(v2, v3); // badcef
												}
											}
										}
									}
								}
							}
						}
					}
					else {
						if (compare(v3, v0)) {
							if (compare(v5, v4)) {
								if (compare(v5, v1)) {
									if (compare(v3, v1)) {
										if (compare(v4, v1)) {
											if (compare(v5, v2)) {
												if (compare(v4, v2)) {
													swap(v0, v5); swap(v1, v4); // fecdba
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v5); swap(v1, v2); swap(v2, v4); // fcedba
													}
													else {
														swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); // fcdeba
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v5); // cfedba
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cfdeba
													}
													else {
														swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); // cdfeba
													}
												}
											}
										}
										else {
											if (compare(v4, v0)) {
												if (compare(v5, v2)) {
													swap(v0, v5); swap(v1, v2); swap(v2, v3); // fcdbea
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); // cfdbea
													}
													else {
														swap(v0, v2); swap(v1, v3); swap(v2, v5); // cdfbea
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v5); // fcdbae
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cfdbae
													}
													else {
														swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v5); // cdfbae
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v3)) {
											if (compare(v4, v1)) {
												if (compare(v4, v2)) {
													if (compare(v2, v1)) {
														swap(v0, v5); swap(v1, v4); swap(v3, v4); // fecbda
													}
													else {
														swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); // febcda
													}
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); // fcebda
													}
													else {
														swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // cfebda
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v5, v2)) {
														swap(v0, v5); swap(v1, v2); swap(v3, v4); // fcbeda
													}
													else {
														swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); // cfbeda
													}
												}
												else {
													if (compare(v4, v2)) {
														swap(v0, v5); swap(v2, v4); swap(v3, v4); // fbecda
													}
													else {
														swap(v0, v5); swap(v3, v4); // fbceda
													}
												}
											}
										}
										else {
											if (compare(v4, v0)) {
												if (compare(v2, v1)) {
													if (compare(v5, v2)) {
														swap(v0, v5); swap(v1, v2); // fcbdea
													}
													else {
														swap(v0, v2); swap(v1, v5); swap(v2, v5); // cfbdea
													}
												}
												else {
													swap(v0, v5); // fbcdea
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v5, v2)) {
														swap(v0, v5); swap(v1, v2); swap(v4, v5); // fcbdae
													}
													else {
														swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v5); // cfbdae
													}
												}
												else {
													swap(v0, v5); swap(v4, v5); // fbcdae
												}
											}
										}
									}
								}
								else {
									if (compare(v4, v0)) {
										if (compare(v2, v1)) {
											if (compare(v5, v3)) {
												if (compare(v4, v3)) {
													swap(v0, v2); swap(v2, v5); swap(v3, v4); // cbfeda
												}
												else {
													swap(v0, v2); swap(v2, v5); // cbfdea
												}
											}
											else {
												if (compare(v3, v1)) {
													swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); // cdbfea
												}
												else {
													swap(v0, v2); swap(v2, v3); swap(v3, v5); // cbdfea
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v4, v2)) {
													swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); // bfecda
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v1); swap(v1, v5); swap(v3, v4); // bfceda
													}
													else {
														swap(v0, v1); swap(v1, v5); // bfcdea
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); // bcfeda
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v1, v2); swap(v2, v5); // bcfdea
													}
													else {
														swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); // bcdfea
													}
												}
											}
										}
									}
									else {
										if (compare(v2, v1)) {
											if (compare(v5, v0)) {
												if (compare(v3, v1)) {
													swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cdbfae
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v2, v5); swap(v4, v5); // cbfdae
													}
													else {
														swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cbdfae
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v5); // cdbafe
												}
												else {
													swap(v0, v2); swap(v2, v3); swap(v4, v5); // cbdafe
												}
											}
										}
										else {
											if (compare(v5, v3)) {
												if (compare(v5, v2)) {
													swap(v0, v1); swap(v1, v5); swap(v4, v5); // bfcdae
												}
												else {
													swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v5); // bcfdae
												}
											}
											else {
												if (compare(v5, v0)) {
													swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bcdfae
												}
												else {
													swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v5); // bcdafe
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v4, v1)) {
									if (compare(v3, v1)) {
										if (compare(v5, v1)) {
											if (compare(v4, v2)) {
												if (compare(v5, v2)) {
													swap(v0, v4); swap(v1, v5); swap(v4, v5); // efcdba
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v5); // ecfdba
													}
													else {
														swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // ecdfba
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v1, v4); swap(v2, v5); // cefdba
													}
													else {
														swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); // cedfba
													}
												}
												else {
													swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // cdefba
												}
											}
										}
										else {
											if (compare(v5, v0)) {
												if (compare(v4, v2)) {
													swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v5); // ecdbfa
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cedbfa
													}
													else {
														swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v5); // cdebfa
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													swap(v0, v4); swap(v1, v2); swap(v2, v3); // ecdbaf
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); // cedbaf
													}
													else {
														swap(v0, v2); swap(v1, v3); swap(v2, v4); // cdebaf
													}
												}
											}
										}
									}
									else {
										if (compare(v5, v3)) {
											if (compare(v5, v1)) {
												if (compare(v5, v2)) {
													if (compare(v2, v1)) {
														swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v5); // efcbda
													}
													else {
														swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // efbcda
													}
												}
												else {
													if (compare(v4, v2)) {
														swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // ecfbda
													}
													else {
														swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); // cefbda
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v4, v2)) {
														swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v5); // ecbfda
													}
													else {
														swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // cebfda
													}
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // ebfcda
													}
													else {
														swap(v0, v4); swap(v3, v5); swap(v4, v5); // ebcfda
													}
												}
											}
										}
										else {
											if (compare(v5, v0)) {
												if (compare(v2, v1)) {
													if (compare(v4, v2)) {
														swap(v0, v4); swap(v1, v2); swap(v4, v5); // ecbdfa
													}
													else {
														swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v5); // cebdfa
													}
												}
												else {
													swap(v0, v4); swap(v4, v5); // ebcdfa
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v4, v2)) {
														swap(v0, v4); swap(v1, v2); // ecbdaf
													}
													else {
														swap(v0, v2); swap(v1, v4); swap(v2, v4); // cebdaf
													}
												}
												else {
													swap(v0, v4); // ebcdaf
												}
											}
										}
									}
								}
								else {
									if (compare(v5, v0)) {
										if (compare(v2, v1)) {
											if (compare(v4, v3)) {
												if (compare(v5, v3)) {
													swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // cbefda
												}
												else {
													swap(v0, v2); swap(v2, v4); swap(v4, v5); // cbedfa
												}
											}
											else {
												if (compare(v3, v1)) {
													swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cdbefa
												}
												else {
													swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cbdefa
												}
											}
										}
										else {
											if (compare(v4, v2)) {
												if (compare(v5, v2)) {
													swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // befcda
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v5); // becfda
													}
													else {
														swap(v0, v1); swap(v1, v4); swap(v4, v5); // becdfa
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // bcefda
													}
													else {
														swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v5); // bcedfa
													}
												}
												else {
													swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bcdefa
												}
											}
										}
									}
									else {
										if (compare(v2, v1)) {
											if (compare(v4, v0)) {
												if (compare(v3, v1)) {
													swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); // cdbeaf
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v2); swap(v2, v4); // cbedaf
													}
													else {
														swap(v0, v2); swap(v2, v3); swap(v3, v4); // cbdeaf
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													swap(v0, v2); swap(v1, v3); swap(v2, v3); // cdbaef
												}
												else {
													swap(v0, v2); swap(v2, v3); // cbdaef
												}
											}
										}
										else {
											if (compare(v4, v3)) {
												if (compare(v4, v2)) {
													swap(v0, v1); swap(v1, v4); // becdaf
												}
												else {
													swap(v0, v1); swap(v1, v2); swap(v2, v4); // bcedaf
												}
											}
											else {
												if (compare(v4, v0)) {
													swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); // bcdeaf
												}
												else {
													swap(v0, v1); swap(v1, v2); swap(v2, v3); // bcdaef
												}
											}
										}
									}
								}
							}
						}
						else {
							if (compare(v5, v4)) {
								if (compare(v5, v1)) {
									if (compare(v4, v0)) {
										if (compare(v2, v1)) {
											if (compare(v4, v1)) {
												if (compare(v4, v2)) {
													swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v5); // fecbad
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // fcebad
													}
													else {
														swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); // cfebad
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v5); // fcbead
												}
												else {
													swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cfbead
												}
											}
										}
										else {
											if (compare(v2, v0)) {
												if (compare(v4, v1)) {
													swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // febcad
												}
												else {
													if (compare(v4, v2)) {
														swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // fbecad
													}
													else {
														swap(v0, v5); swap(v3, v4); swap(v4, v5); // fbcead
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); // febacd
												}
												else {
													swap(v0, v5); swap(v2, v4); swap(v3, v5); // fbeacd
												}
											}
										}
									}
									else {
										if (compare(v2, v1)) {
											if (compare(v5, v2)) {
												if (compare(v4, v3)) {
													swap(v0, v5); swap(v1, v2); swap(v3, v5); // fcbaed
												}
												else {
													swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v5); // fcbade
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); // cfbaed
												}
												else {
													swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cfbade
												}
											}
										}
										else {
											if (compare(v2, v0)) {
												if (compare(v4, v3)) {
													swap(v0, v5); swap(v3, v5); // fbcaed
												}
												else {
													swap(v0, v5); swap(v3, v5); swap(v4, v5); // fbcade
												}
											}
											else {
												if (compare(v4, v2)) {
													swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fbaecd
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v5); swap(v2, v5); swap(v3, v5); // fbaced
													}
													else {
														swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fbacde
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v5, v0)) {
										if (compare(v5, v2)) {
											if (compare(v2, v0)) {
												if (compare(v4, v0)) {
													if (compare(v4, v2)) {
														swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // bfecad
													}
													else {
														swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v5); // bfcead
													}
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v1); swap(v1, v5); swap(v3, v5); // bfcaed
													}
													else {
														swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v5); // bfcade
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v0)) {
														swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); // bfeacd
													}
													else {
														swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bfaecd
													}
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); // bfaced
													}
													else {
														swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bfacde
													}
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v4, v0)) {
													swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cbfead
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v2); swap(v2, v5); swap(v3, v5); // cbfaed
													}
													else {
														swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cbfade
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bcfead
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); // bcfaed
													}
													else {
														swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bcfade
													}
												}
											}
										}
									}
									else {
										if (compare(v2, v0)) {
											if (compare(v2, v1)) {
												if (compare(v4, v3)) {
													swap(v0, v2); swap(v3, v5); // cbafed
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v3, v5); swap(v4, v5); // cbafde
													}
													else {
														swap(v0, v2); swap(v4, v5); // cbadfe
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v0, v1); swap(v1, v2); swap(v3, v5); // bcafed
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v5); // bcafde
													}
													else {
														swap(v0, v1); swap(v1, v2); swap(v4, v5); // bcadfe
													}
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v4, v2)) {
													swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bafecd
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v1); swap(v2, v5); swap(v3, v5); // bafced
													}
													else {
														swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bafcde
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v0, v1); swap(v3, v5); // bacfed
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v3, v5); swap(v4, v5); // bacfde
													}
													else {
														swap(v0, v1); swap(v4, v5); // bacdfe
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v4, v1)) {
									if (compare(v5, v0)) {
										if (compare(v2, v1)) {
											if (compare(v5, v1)) {
												if (compare(v4, v2)) {
													if (compare(v5, v2)) {
														swap(v0, v4); swap(v1, v5); swap(v3, v5); // efcbad
													}
													else {
														swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); // ecfbad
													}
												}
												else {
													swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cefbad
												}
											}
											else {
												if (compare(v4, v2)) {
													swap(v0, v4); swap(v1, v2); swap(v3, v5); // ecbfad
												}
												else {
													swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); // cebfad
												}
											}
										}
										else {
											if (compare(v2, v0)) {
												if (compare(v5, v1)) {
													swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); // efbcad
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v4); swap(v2, v5); swap(v3, v5); // ebfcad
													}
													else {
														swap(v0, v4); swap(v3, v5); // ebcfad
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // efbacd
												}
												else {
													swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // ebfacd
												}
											}
										}
									}
									else {
										if (compare(v2, v1)) {
											if (compare(v4, v2)) {
												if (compare(v5, v3)) {
													swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v5); // ecbafd
												}
												else {
													swap(v0, v4); swap(v1, v2); swap(v3, v4); // ecbadf
												}
											}
											else {
												if (compare(v5, v3)) {
													swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cebafd
												}
												else {
													swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); // cebadf
												}
											}
										}
										else {
											if (compare(v2, v0)) {
												if (compare(v5, v3)) {
													swap(v0, v4); swap(v3, v4); swap(v4, v5); // ebcafd
												}
												else {
													swap(v0, v4); swap(v3, v4); // ebcadf
												}
											}
											else {
												if (compare(v5, v2)) {
													swap(v0, v4); swap(v2, v4); swap(v3, v5); // ebafcd
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // ebacfd
													}
													else {
														swap(v0, v4); swap(v2, v4); swap(v3, v4); // ebacdf
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v4, v0)) {
										if (compare(v4, v2)) {
											if (compare(v2, v0)) {
												if (compare(v5, v0)) {
													if (compare(v5, v2)) {
														swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); // befcad
													}
													else {
														swap(v0, v1); swap(v1, v4); swap(v3, v5); // becfad
													}
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v5); // becafd
													}
													else {
														swap(v0, v1); swap(v1, v4); swap(v3, v4); // becadf
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v0)) {
														swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // befacd
													}
													else {
														swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); // beafcd
													}
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // beacfd
													}
													else {
														swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); // beacdf
													}
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v5, v0)) {
													swap(v0, v2); swap(v2, v4); swap(v3, v5); // cbefad
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cbeafd
													}
													else {
														swap(v0, v2); swap(v2, v4); swap(v3, v4); // cbeadf
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); // bcefad
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // bceafd
													}
													else {
														swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); // bceadf
													}
												}
											}
										}
									}
									else {
										if (compare(v2, v0)) {
											if (compare(v2, v1)) {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v3, v4); swap(v4, v5); // cbaefd
													}
													else {
														swap(v0, v2); swap(v3, v4); // cbaedf
													}
												}
												else {
													swap(v0, v2); // cbadef
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v5); // bcaefd
													}
													else {
														swap(v0, v1); swap(v1, v2); swap(v3, v4); // bcaedf
													}
												}
												else {
													swap(v0, v1); swap(v1, v2); // bcadef
												}
											}
										}
										else {
											if (compare(v4, v2)) {
												if (compare(v5, v2)) {
													swap(v0, v1); swap(v2, v4); swap(v3, v5); // baefcd
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v5); // baecfd
													}
													else {
														swap(v0, v1); swap(v2, v4); swap(v3, v4); // baecdf
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v0, v1); swap(v3, v4); swap(v4, v5); // bacefd
													}
													else {
														swap(v0, v1); swap(v3, v4); // bacedf
													}
												}
												else {
													swap(v0, v1); // bacdef
												}
											}
										}
									}
								}
							}
						}
					}
				}
				else {
					if (compare(v3, v2)) {
						if (compare(v3, v0)) {
							if (compare(v5, v4)) {
								if (compare(v4, v1)) {
									if (compare(v4, v0)) {
										if (compare(v2, v0)) {
											if (compare(v4, v2)) {
												if (compare(v4, v3)) {
													swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v5); // fedcab
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // fdecab
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); // dfecab
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v5); // fdceab
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v3, v4); // dfceab
													}
												}
												else {
													swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); // dcfeab
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v4, v3)) {
													swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fedacb
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); // fdeacb
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v2, v4); // dfeacb
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); // fedabc
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // fdeabc
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v5); // dfeabc
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v2)) {
											if (compare(v2, v1)) {
												if (compare(v5, v0)) {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fdaecb
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); // dfaecb
													}
												}
												else {
													swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dafecb
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); // fdaebc
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dfaebc
													}
												}
												else {
													swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); // dafebc
												}
											}
										}
										else {
											if (compare(v2, v0)) {
												if (compare(v5, v2)) {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v3, v5); // fdcaeb
													}
													else {
														swap(v0, v3); swap(v1, v5); // dfcaeb
													}
												}
												else {
													if (compare(v5, v0)) {
														swap(v0, v3); swap(v1, v2); swap(v2, v5); // dcfaeb
													}
													else {
														swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); // dcafeb
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); // fdaceb
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v2, v3); // dfaceb
													}
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); // dafceb
													}
													else {
														swap(v0, v3); swap(v1, v3); swap(v3, v5); // dacfeb
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v5, v0)) {
										if (compare(v2, v1)) {
											if (compare(v2, v0)) {
												if (compare(v5, v2)) {
													if (compare(v5, v3)) {
														swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v5); // fdcabe
													}
													else {
														swap(v0, v3); swap(v1, v5); swap(v4, v5); // dfcabe
													}
												}
												else {
													swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v5); // dcfabe
												}
											}
											else {
												if (compare(v5, v3)) {
													swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fdacbe
												}
												else {
													swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v5); // dfacbe
												}
											}
										}
										else {
											if (compare(v4, v2)) {
												if (compare(v5, v3)) {
													swap(v0, v5); swap(v1, v3); swap(v2, v5); // fdabec
												}
												else {
													swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); // dfabec
												}
											}
											else {
												if (compare(v5, v3)) {
													swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v5); // fdabce
												}
												else {
													swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dfabce
												}
											}
										}
									}
									else {
										if (compare(v2, v1)) {
											if (compare(v2, v0)) {
												if (compare(v5, v1)) {
													swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dcafbe
												}
												else {
													swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v5); // dcabfe
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v5, v2)) {
														swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dafcbe
													}
													else {
														swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v5); // dacfbe
													}
												}
												else {
													swap(v0, v3); swap(v1, v3); swap(v4, v5); // dacbfe
												}
											}
										}
										else {
											if (compare(v5, v1)) {
												if (compare(v4, v2)) {
													swap(v0, v3); swap(v1, v3); swap(v2, v5); // dafbec
												}
												else {
													swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v5); // dafbce
												}
											}
											else {
												if (compare(v4, v2)) {
													swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); // dabfec
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dabfce
													}
													else {
														swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v5); // dabcfe
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v5, v1)) {
									if (compare(v5, v0)) {
										if (compare(v2, v0)) {
											if (compare(v5, v2)) {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v1, v5); swap(v2, v3); // efdcab
													}
													else {
														swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); // edfcab
													}
												}
												else {
													swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // defcab
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v3)) {
														swap(v0, v4); swap(v1, v3); swap(v3, v5); // edcfab
													}
													else {
														swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v5); // decfab
													}
												}
												else {
													swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // dcefab
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); // efdacb
													}
													else {
														swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // edfacb
													}
												}
												else {
													swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v5); // defacb
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // efdabc
													}
													else {
														swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); // edfabc
													}
												}
												else {
													swap(v0, v3); swap(v1, v4); swap(v2, v5); // defabc
												}
											}
										}
									}
									else {
										if (compare(v5, v2)) {
											if (compare(v2, v1)) {
												if (compare(v4, v0)) {
													if (compare(v4, v3)) {
														swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); // edafcb
													}
													else {
														swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // deafcb
													}
												}
												else {
													swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); // daefcb
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v4, v3)) {
														swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // edafbc
													}
													else {
														swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); // deafbc
													}
												}
												else {
													swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // daefbc
												}
											}
										}
										else {
											if (compare(v2, v0)) {
												if (compare(v4, v2)) {
													if (compare(v4, v3)) {
														swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v5); // edcafb
													}
													else {
														swap(v0, v3); swap(v1, v4); swap(v4, v5); // decafb
													}
												}
												else {
													if (compare(v4, v0)) {
														swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v5); // dceafb
													}
													else {
														swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dcaefb
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v4, v3)) {
														swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // edacfb
													}
													else {
														swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v5); // deacfb
													}
												}
												else {
													if (compare(v4, v2)) {
														swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // daecfb
													}
													else {
														swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v5); // dacefb
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v4, v0)) {
										if (compare(v2, v1)) {
											if (compare(v2, v0)) {
												if (compare(v4, v2)) {
													if (compare(v4, v3)) {
														swap(v0, v4); swap(v1, v3); swap(v3, v4); // edcabf
													}
													else {
														swap(v0, v3); swap(v1, v4); // decabf
													}
												}
												else {
													swap(v0, v3); swap(v1, v2); swap(v2, v4); // dceabf
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); // edacbf
												}
												else {
													swap(v0, v3); swap(v1, v4); swap(v2, v3); // deacbf
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v4, v3)) {
													swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v5); // edabfc
												}
												else {
													swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // deabfc
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v0, v4); swap(v1, v3); swap(v2, v4); // edabcf
												}
												else {
													swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); // deabcf
												}
											}
										}
									}
									else {
										if (compare(v2, v1)) {
											if (compare(v2, v0)) {
												if (compare(v4, v1)) {
													swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); // dcaebf
												}
												else {
													swap(v0, v3); swap(v1, v2); swap(v2, v3); // dcabef
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v4, v2)) {
														swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); // daecbf
													}
													else {
														swap(v0, v3); swap(v1, v3); swap(v3, v4); // dacebf
													}
												}
												else {
													swap(v0, v3); swap(v1, v3); // dacbef
												}
											}
										}
										else {
											if (compare(v4, v1)) {
												if (compare(v5, v2)) {
													swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v5); // daebfc
												}
												else {
													swap(v0, v3); swap(v1, v3); swap(v2, v4); // daebcf
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v5, v2)) {
														swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dabefc
													}
													else {
														swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); // dabecf
													}
												}
												else {
													swap(v0, v3); swap(v1, v3); swap(v2, v3); // dabcef
												}
											}
										}
									}
								}
							}
						}
						else {
							if (compare(v5, v4)) {
								if (compare(v4, v1)) {
									if (compare(v5, v0)) {
										if (compare(v2, v1)) {
											if (compare(v4, v3)) {
												if (compare(v4, v0)) {
													swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v5); // feadcb
												}
												else {
													swap(v0, v5); swap(v1, v5); swap(v2, v4); // faedcb
												}
											}
											else {
												if (compare(v4, v2)) {
													swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); // fadecb
												}
												else {
													swap(v0, v5); swap(v1, v5); swap(v2, v3); // fadceb
												}
											}
										}
										else {
											if (compare(v4, v0)) {
												if (compare(v3, v1)) {
													swap(v0, v5); swap(v1, v4); swap(v2, v5); // feadbc
												}
												else {
													swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); // feabdc
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v4, v3)) {
														swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v5); // faedbc
													}
													else {
														swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fadebc
													}
												}
												else {
													swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // faebdc
												}
											}
										}
									}
									else {
										if (compare(v2, v1)) {
											if (compare(v4, v2)) {
												if (compare(v4, v3)) {
													swap(v1, v5); swap(v2, v4); // afedcb
												}
												else {
													if (compare(v5, v3)) {
														swap(v1, v5); swap(v2, v3); swap(v3, v4); // afdecb
													}
													else {
														swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // adfecb
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v3)) {
														swap(v1, v5); swap(v2, v3); // afdceb
													}
													else {
														swap(v1, v3); swap(v2, v5); swap(v3, v5); // adfceb
													}
												}
												else {
													swap(v1, v3); swap(v3, v5); // adcfeb
												}
											}
										}
										else {
											if (compare(v4, v3)) {
												if (compare(v3, v1)) {
													swap(v1, v5); swap(v2, v4); swap(v4, v5); // afedbc
												}
												else {
													swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // afebdc
												}
											}
											else {
												if (compare(v5, v3)) {
													swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // afdebc
												}
												else {
													swap(v1, v3); swap(v2, v5); swap(v3, v4); // adfebc
												}
											}
										}
									}
								}
								else {
									if (compare(v3, v1)) {
										if (compare(v5, v3)) {
											if (compare(v5, v0)) {
												if (compare(v2, v1)) {
													swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v5); // fadcbe
												}
												else {
													if (compare(v4, v2)) {
														swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); // fadbec
													}
													else {
														swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fadbce
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													swap(v1, v5); swap(v2, v3); swap(v4, v5); // afdcbe
												}
												else {
													if (compare(v4, v2)) {
														swap(v1, v5); swap(v2, v3); swap(v3, v5); // afdbec
													}
													else {
														swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // afdbce
													}
												}
											}
										}
										else {
											if (compare(v5, v1)) {
												if (compare(v2, v1)) {
													if (compare(v5, v2)) {
														swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // adfcbe
													}
													else {
														swap(v1, v3); swap(v3, v5); swap(v4, v5); // adcfbe
													}
												}
												else {
													if (compare(v4, v2)) {
														swap(v1, v3); swap(v2, v5); // adfbec
													}
													else {
														swap(v1, v3); swap(v2, v5); swap(v4, v5); // adfbce
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v4, v2)) {
														swap(v1, v3); swap(v2, v3); swap(v3, v5); // adbfec
													}
													else {
														swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // adbfce
													}
												}
												else {
													if (compare(v2, v1)) {
														swap(v1, v3); swap(v4, v5); // adcbfe
													}
													else {
														swap(v1, v3); swap(v2, v3); swap(v4, v5); // adbcfe
													}
												}
											}
										}
									}
									else {
										if (compare(v5, v1)) {
											if (compare(v5, v0)) {
												if (compare(v4, v2)) {
													if (compare(v4, v3)) {
														swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); // fabedc
													}
													else {
														swap(v0, v5); swap(v1, v5); swap(v2, v5); // fabdec
													}
												}
												else {
													swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v5); // fabdce
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v3)) {
														swap(v1, v5); swap(v2, v5); swap(v3, v4); // afbedc
													}
													else {
														swap(v1, v5); swap(v2, v5); // afbdec
													}
												}
												else {
													swap(v1, v5); swap(v2, v5); swap(v4, v5); // afbdce
												}
											}
										}
										else {
											if (compare(v4, v2)) {
												if (compare(v4, v3)) {
													swap(v2, v5); swap(v3, v4); // abfedc
												}
												else {
													if (compare(v5, v3)) {
														swap(v2, v5); // abfdec
													}
													else {
														swap(v2, v3); swap(v3, v5); // abdfec
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v3)) {
														swap(v2, v5); swap(v4, v5); // abfdce
													}
													else {
														swap(v2, v3); swap(v3, v5); swap(v4, v5); // abdfce
													}
												}
												else {
													swap(v2, v3); swap(v4, v5); // abdcfe
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v5, v1)) {
									if (compare(v4, v0)) {
										if (compare(v2, v1)) {
											if (compare(v5, v3)) {
												if (compare(v5, v0)) {
													swap(v0, v4); swap(v1, v5); swap(v2, v4); // efadcb
												}
												else {
													swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v5); // eafdcb
												}
											}
											else {
												if (compare(v5, v2)) {
													swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // eadfcb
												}
												else {
													swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v5); // eadcfb
												}
											}
										}
										else {
											if (compare(v5, v0)) {
												if (compare(v3, v1)) {
													swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v5); // efadbc
												}
												else {
													swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // efabdc
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v1, v4); swap(v2, v5); // eafdbc
													}
													else {
														swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); // eadfbc
													}
												}
												else {
													swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); // eafbdc
												}
											}
										}
									}
									else {
										if (compare(v2, v1)) {
											if (compare(v5, v2)) {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v1, v4); swap(v2, v5); swap(v4, v5); // aefdcb
													}
													else {
														swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // aedfcb
													}
												}
												else {
													swap(v1, v3); swap(v2, v4); swap(v3, v5); // adefcb
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v3)) {
														swap(v1, v4); swap(v2, v3); swap(v4, v5); // aedcfb
													}
													else {
														swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // adecfb
													}
												}
												else {
													swap(v1, v3); swap(v3, v4); swap(v4, v5); // adcefb
												}
											}
										}
										else {
											if (compare(v5, v3)) {
												if (compare(v3, v1)) {
													swap(v1, v4); swap(v2, v5); // aefdbc
												}
												else {
													swap(v1, v4); swap(v2, v5); swap(v3, v4); // aefbdc
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v1, v4); swap(v2, v3); swap(v3, v5); // aedfbc
												}
												else {
													swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // adefbc
												}
											}
										}
									}
								}
								else {
									if (compare(v3, v1)) {
										if (compare(v4, v3)) {
											if (compare(v4, v0)) {
												if (compare(v2, v1)) {
													swap(v0, v4); swap(v1, v4); swap(v2, v3); // eadcbf
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // eadbfc
													}
													else {
														swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); // eadbcf
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													swap(v1, v4); swap(v2, v3); // aedcbf
												}
												else {
													if (compare(v5, v2)) {
														swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // aedbfc
													}
													else {
														swap(v1, v4); swap(v2, v3); swap(v3, v4); // aedbcf
													}
												}
											}
										}
										else {
											if (compare(v4, v1)) {
												if (compare(v2, v1)) {
													if (compare(v4, v2)) {
														swap(v1, v3); swap(v2, v4); swap(v3, v4); // adecbf
													}
													else {
														swap(v1, v3); swap(v3, v4); // adcebf
													}
												}
												else {
													if (compare(v5, v2)) {
														swap(v1, v3); swap(v2, v4); swap(v4, v5); // adebfc
													}
													else {
														swap(v1, v3); swap(v2, v4); // adebcf
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v5, v2)) {
														swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // adbefc
													}
													else {
														swap(v1, v3); swap(v2, v3); swap(v3, v4); // adbecf
													}
												}
												else {
													if (compare(v2, v1)) {
														swap(v1, v3); // adcbef
													}
													else {
														swap(v1, v3); swap(v2, v3); // adbcef
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v1)) {
											if (compare(v4, v0)) {
												if (compare(v5, v2)) {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // eabfdc
													}
													else {
														swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v5); // eabdfc
													}
												}
												else {
													swap(v0, v4); swap(v1, v4); swap(v2, v4); // eabdcf
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v3)) {
														swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // aebfdc
													}
													else {
														swap(v1, v4); swap(v2, v4); swap(v4, v5); // aebdfc
													}
												}
												else {
													swap(v1, v4); swap(v2, v4); // aebdcf
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v2, v4); swap(v3, v5); swap(v4, v5); // abefdc
													}
													else {
														swap(v2, v4); swap(v4, v5); // abedfc
													}
												}
												else {
													swap(v2, v3); swap(v3, v4); swap(v4, v5); // abdefc
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v3)) {
														swap(v2, v4); // abedcf
													}
													else {
														swap(v2, v3); swap(v3, v4); // abdecf
													}
												}
												else {
													swap(v2, v3); // abdcef
												}
											}
										}
									}
								}
							}
						}
					}
					else {
						if (compare(v2, v0)) {
							if (compare(v5, v4)) {
								if (compare(v4, v1)) {
									if (compare(v4, v0)) {
										if (compare(v3, v0)) {
											if (compare(v5, v2)) {
												if (compare(v4, v2)) {
													swap(v0, v5); swap(v1, v4); swap(v4, v5); // fecdab
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v5); // fcedab
													}
													else {
														swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fcdeab
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v0, v2); swap(v1, v5); swap(v2, v4); // cfedab
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); // cfdeab
													}
													else {
														swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cdfeab
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v4, v2)) {
													swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v5); // fecadb
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // fceadb
													}
													else {
														swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); // cfeadb
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													swap(v0, v5); swap(v1, v4); swap(v3, v5); // fecabd
												}
												else {
													if (compare(v5, v2)) {
														swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); // fceabd
													}
													else {
														swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cfeabd
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v3)) {
											if (compare(v3, v1)) {
												if (compare(v5, v0)) {
													if (compare(v5, v2)) {
														swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); // fcaedb
													}
													else {
														swap(v0, v2); swap(v1, v5); swap(v3, v4); // cfaedb
													}
												}
												else {
													swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); // cafedb
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v5, v2)) {
														swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fcaebd
													}
													else {
														swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v5); // cfaebd
													}
												}
												else {
													swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cafebd
												}
											}
										}
										else {
											if (compare(v3, v0)) {
												if (compare(v5, v3)) {
													if (compare(v5, v2)) {
														swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); // fcdaeb
													}
													else {
														swap(v0, v2); swap(v1, v5); swap(v2, v3); // cfdaeb
													}
												}
												else {
													if (compare(v5, v0)) {
														swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); // cdfaeb
													}
													else {
														swap(v0, v2); swap(v1, v3); swap(v3, v5); // cdafeb
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v5, v2)) {
														swap(v0, v5); swap(v1, v2); swap(v2, v5); // fcadeb
													}
													else {
														swap(v0, v2); swap(v1, v5); // cfadeb
													}
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v1, v2); swap(v2, v5); // cafdeb
													}
													else {
														swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); // cadfeb
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v5, v0)) {
										if (compare(v3, v1)) {
											if (compare(v3, v0)) {
												if (compare(v5, v2)) {
													swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fcdabe
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v5); // cfdabe
													}
													else {
														swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cdfabe
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v5); // fcadbe
												}
												else {
													swap(v0, v2); swap(v1, v5); swap(v4, v5); // cfadbe
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v4, v3)) {
													swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); // fcabed
												}
												else {
													swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fcabde
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v0, v2); swap(v1, v5); swap(v3, v5); // cfabed
												}
												else {
													swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v5); // cfabde
												}
											}
										}
									}
									else {
										if (compare(v3, v1)) {
											if (compare(v3, v0)) {
												if (compare(v5, v1)) {
													swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v5); // cdafbe
												}
												else {
													swap(v0, v2); swap(v1, v3); swap(v4, v5); // cdabfe
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v5); // cafdbe
													}
													else {
														swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cadfbe
													}
												}
												else {
													swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v5); // cadbfe
												}
											}
										}
										else {
											if (compare(v5, v1)) {
												if (compare(v4, v3)) {
													swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); // cafbed
												}
												else {
													swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cafbde
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v0, v2); swap(v1, v2); swap(v3, v5); // cabfed
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v5); // cabfde
													}
													else {
														swap(v0, v2); swap(v1, v2); swap(v4, v5); // cabdfe
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v5, v1)) {
									if (compare(v5, v0)) {
										if (compare(v3, v0)) {
											if (compare(v4, v2)) {
												if (compare(v5, v2)) {
													swap(v0, v4); swap(v1, v5); // efcdab
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v1, v2); swap(v2, v5); // ecfdab
													}
													else {
														swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); // ecdfab
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v5); // cefdab
													}
													else {
														swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cedfab
													}
												}
												else {
													swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); // cdefab
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v4, v2)) {
													if (compare(v5, v2)) {
														swap(v0, v4); swap(v1, v5); swap(v3, v4); // efcadb
													}
													else {
														swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); // ecfadb
													}
												}
												else {
													swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cefadb
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v5, v2)) {
														swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v5); // efcabd
													}
													else {
														swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // ecfabd
													}
												}
												else {
													swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); // cefabd
												}
											}
										}
									}
									else {
										if (compare(v5, v3)) {
											if (compare(v3, v1)) {
												if (compare(v4, v0)) {
													if (compare(v4, v2)) {
														swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // ecafdb
													}
													else {
														swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v5); // ceafdb
													}
												}
												else {
													swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // caefdb
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v4, v2)) {
														swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); // ecafbd
													}
													else {
														swap(v0, v2); swap(v1, v4); swap(v3, v5); // ceafbd
													}
												}
												else {
													swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); // caefbd
												}
											}
										}
										else {
											if (compare(v3, v0)) {
												if (compare(v4, v3)) {
													if (compare(v4, v2)) {
														swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // ecdafb
													}
													else {
														swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v5); // cedafb
													}
												}
												else {
													if (compare(v4, v0)) {
														swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cdeafb
													}
													else {
														swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v5); // cdaefb
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v4, v2)) {
														swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v5); // ecadfb
													}
													else {
														swap(v0, v2); swap(v1, v4); swap(v4, v5); // ceadfb
													}
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v5); // caedfb
													}
													else {
														swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cadefb
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v4, v0)) {
										if (compare(v3, v1)) {
											if (compare(v3, v0)) {
												if (compare(v4, v2)) {
													swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); // ecdabf
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v2); swap(v1, v4); swap(v2, v3); // cedabf
													}
													else {
														swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); // cdeabf
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													swap(v0, v4); swap(v1, v2); swap(v2, v4); // ecadbf
												}
												else {
													swap(v0, v2); swap(v1, v4); // ceadbf
												}
											}
										}
										else {
											if (compare(v4, v2)) {
												if (compare(v5, v3)) {
													swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // ecabfd
												}
												else {
													swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); // ecabdf
												}
											}
											else {
												if (compare(v5, v3)) {
													swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v5); // ceabfd
												}
												else {
													swap(v0, v2); swap(v1, v4); swap(v3, v4); // ceabdf
												}
											}
										}
									}
									else {
										if (compare(v3, v1)) {
											if (compare(v3, v0)) {
												if (compare(v4, v1)) {
													swap(v0, v2); swap(v1, v3); swap(v3, v4); // cdaebf
												}
												else {
													swap(v0, v2); swap(v1, v3); // cdabef
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v4, v3)) {
														swap(v0, v2); swap(v1, v2); swap(v2, v4); // caedbf
													}
													else {
														swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); // cadebf
													}
												}
												else {
													swap(v0, v2); swap(v1, v2); swap(v2, v3); // cadbef
												}
											}
										}
										else {
											if (compare(v4, v1)) {
												if (compare(v5, v3)) {
													swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // caebfd
												}
												else {
													swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); // caebdf
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v5); // cabefd
													}
													else {
														swap(v0, v2); swap(v1, v2); swap(v3, v4); // cabedf
													}
												}
												else {
													swap(v0, v2); swap(v1, v2); // cabdef
												}
											}
										}
									}
								}
							}
						}
						else {
							if (compare(v5, v4)) {
								if (compare(v4, v1)) {
									if (compare(v5, v0)) {
										if (compare(v3, v1)) {
											if (compare(v4, v2)) {
												if (compare(v4, v0)) {
													swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // feacdb
												}
												else {
													swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); // faecdb
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v0, v5); swap(v1, v5); swap(v3, v4); // facedb
												}
												else {
													swap(v0, v5); swap(v1, v5); // facdeb
												}
											}
										}
										else {
											if (compare(v4, v0)) {
												if (compare(v2, v1)) {
													swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); // feacbd
												}
												else {
													swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // feabcd
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v4, v2)) {
														swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // faecbd
													}
													else {
														swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v5); // facebd
													}
												}
												else {
													swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); // faebcd
												}
											}
										}
									}
									else {
										if (compare(v3, v1)) {
											if (compare(v5, v2)) {
												if (compare(v4, v2)) {
													swap(v1, v5); swap(v2, v4); swap(v3, v4); // afecdb
												}
												else {
													if (compare(v4, v3)) {
														swap(v1, v5); swap(v3, v4); // afcedb
													}
													else {
														swap(v1, v5); // afcdeb
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v1, v2); swap(v2, v5); swap(v3, v4); // acfedb
												}
												else {
													if (compare(v5, v3)) {
														swap(v1, v2); swap(v2, v5); // acfdeb
													}
													else {
														swap(v1, v2); swap(v2, v3); swap(v3, v5); // acdfeb
													}
												}
											}
										}
										else {
											if (compare(v4, v2)) {
												if (compare(v2, v1)) {
													swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // afecbd
												}
												else {
													swap(v1, v5); swap(v2, v4); swap(v3, v5); // afebcd
												}
											}
											else {
												if (compare(v5, v2)) {
													swap(v1, v5); swap(v3, v4); swap(v4, v5); // afcebd
												}
												else {
													swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // acfebd
												}
											}
										}
									}
								}
								else {
									if (compare(v2, v1)) {
										if (compare(v5, v2)) {
											if (compare(v5, v0)) {
												if (compare(v3, v1)) {
													swap(v0, v5); swap(v1, v5); swap(v4, v5); // facdbe
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v5); swap(v1, v5); swap(v3, v5); // facbed
													}
													else {
														swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v5); // facbde
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													swap(v1, v5); swap(v4, v5); // afcdbe
												}
												else {
													if (compare(v4, v3)) {
														swap(v1, v5); swap(v3, v5); // afcbed
													}
													else {
														swap(v1, v5); swap(v3, v5); swap(v4, v5); // afcbde
													}
												}
											}
										}
										else {
											if (compare(v5, v1)) {
												if (compare(v3, v1)) {
													if (compare(v5, v3)) {
														swap(v1, v2); swap(v2, v5); swap(v4, v5); // acfdbe
													}
													else {
														swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // acdfbe
													}
												}
												else {
													if (compare(v4, v3)) {
														swap(v1, v2); swap(v2, v5); swap(v3, v5); // acfbed
													}
													else {
														swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // acfbde
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v4, v3)) {
														swap(v1, v2); swap(v3, v5); // acbfed
													}
													else {
														swap(v1, v2); swap(v3, v5); swap(v4, v5); // acbfde
													}
												}
												else {
													if (compare(v3, v1)) {
														swap(v1, v2); swap(v2, v3); swap(v4, v5); // acdbfe
													}
													else {
														swap(v1, v2); swap(v4, v5); // acbdfe
													}
												}
											}
										}
									}
									else {
										if (compare(v5, v1)) {
											if (compare(v5, v0)) {
												if (compare(v4, v2)) {
													swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fabecd
												}
												else {
													if (compare(v4, v3)) {
														swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); // fabced
													}
													else {
														swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fabcde
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // afbecd
												}
												else {
													if (compare(v4, v3)) {
														swap(v1, v5); swap(v2, v5); swap(v3, v5); // afbced
													}
													else {
														swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // afbcde
													}
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v4, v2)) {
													swap(v2, v5); swap(v3, v4); swap(v4, v5); // abfecd
												}
												else {
													if (compare(v4, v3)) {
														swap(v2, v5); swap(v3, v5); // abfced
													}
													else {
														swap(v2, v5); swap(v3, v5); swap(v4, v5); // abfcde
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													swap(v3, v5); // abcfed
												}
												else {
													if (compare(v5, v3)) {
														swap(v3, v5); swap(v4, v5); // abcfde
													}
													else {
														swap(v4, v5); // abcdfe
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v5, v1)) {
									if (compare(v4, v0)) {
										if (compare(v3, v1)) {
											if (compare(v5, v2)) {
												if (compare(v5, v0)) {
													swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); // efacdb
												}
												else {
													swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // eafcdb
												}
											}
											else {
												if (compare(v5, v3)) {
													swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v5); // eacfdb
												}
												else {
													swap(v0, v4); swap(v1, v4); swap(v4, v5); // eacdfb
												}
											}
										}
										else {
											if (compare(v5, v0)) {
												if (compare(v2, v1)) {
													swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // efacbd
												}
												else {
													swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); // efabcd
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v5, v2)) {
														swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); // eafcbd
													}
													else {
														swap(v0, v4); swap(v1, v4); swap(v3, v5); // eacfbd
													}
												}
												else {
													swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // eafbcd
												}
											}
										}
									}
									else {
										if (compare(v3, v1)) {
											if (compare(v4, v2)) {
												if (compare(v5, v2)) {
													swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // aefcdb
												}
												else {
													if (compare(v5, v3)) {
														swap(v1, v4); swap(v3, v5); swap(v4, v5); // aecfdb
													}
													else {
														swap(v1, v4); swap(v4, v5); // aecdfb
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // acefdb
													}
													else {
														swap(v1, v2); swap(v2, v4); swap(v4, v5); // acedfb
													}
												}
												else {
													swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // acdefb
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v2, v1)) {
													swap(v1, v4); swap(v2, v5); swap(v3, v5); // aefcbd
												}
												else {
													swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // aefbcd
												}
											}
											else {
												if (compare(v4, v2)) {
													swap(v1, v4); swap(v3, v5); // aecfbd
												}
												else {
													swap(v1, v2); swap(v2, v4); swap(v3, v5); // acefbd
												}
											}
										}
									}
								}
								else {
									if (compare(v2, v1)) {
										if (compare(v4, v2)) {
											if (compare(v4, v0)) {
												if (compare(v3, v1)) {
													swap(v0, v4); swap(v1, v4); // eacdbf
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v5); // eacbfd
													}
													else {
														swap(v0, v4); swap(v1, v4); swap(v3, v4); // eacbdf
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													swap(v1, v4); // aecdbf
												}
												else {
													if (compare(v5, v3)) {
														swap(v1, v4); swap(v3, v4); swap(v4, v5); // aecbfd
													}
													else {
														swap(v1, v4); swap(v3, v4); // aecbdf
													}
												}
											}
										}
										else {
											if (compare(v4, v1)) {
												if (compare(v3, v1)) {
													if (compare(v4, v3)) {
														swap(v1, v2); swap(v2, v4); // acedbf
													}
													else {
														swap(v1, v2); swap(v2, v3); swap(v3, v4); // acdebf
													}
												}
												else {
													if (compare(v5, v3)) {
														swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // acebfd
													}
													else {
														swap(v1, v2); swap(v2, v4); swap(v3, v4); // acebdf
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v1, v2); swap(v3, v4); swap(v4, v5); // acbefd
													}
													else {
														swap(v1, v2); swap(v3, v4); // acbedf
													}
												}
												else {
													if (compare(v3, v1)) {
														swap(v1, v2); swap(v2, v3); // acdbef
													}
													else {
														swap(v1, v2); // acbdef
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v1)) {
											if (compare(v4, v0)) {
												if (compare(v5, v2)) {
													swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); // eabfcd
												}
												else {
													if (compare(v5, v3)) {
														swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // eabcfd
													}
													else {
														swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); // eabcdf
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													swap(v1, v4); swap(v2, v4); swap(v3, v5); // aebfcd
												}
												else {
													if (compare(v5, v3)) {
														swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // aebcfd
													}
													else {
														swap(v1, v4); swap(v2, v4); swap(v3, v4); // aebcdf
													}
												}
											}
										}
										else {
											if (compare(v4, v2)) {
												if (compare(v5, v2)) {
													swap(v2, v4); swap(v3, v5); // abefcd
												}
												else {
													if (compare(v5, v3)) {
														swap(v2, v4); swap(v3, v4); swap(v4, v5); // abecfd
													}
													else {
														swap(v2, v4); swap(v3, v4); // abecdf
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v5, v3)) {
														swap(v3, v4); swap(v4, v5); // abcefd
													}
													else {
														swap(v3, v4); // abcedf
													}
												}
												else {
													// abcdef
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}

			static void stable_sort7(T &v0, T &v1, T &v2, T &v3, T &v4, T &v5, T &v6, const Compare &compare)
			{
				using std::swap;
				if (compare(v1, v0)) {
					if (compare(v3, v2)) {
						if (compare(v2, v0)) {
							if (compare(v5, v4)) {
								if (compare(v6, v2)) {
									if (compare(v6, v3)) {
										if (compare(v3, v1)) {
											if (compare(v2, v1)) {
												if (compare(v4, v1)) {
													if (compare(v4, v2)) {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); // gfedcba
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v5, v6); // fgedcba
																}
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // fegdcba
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); // gfdecba
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // fgdecba
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // gdfecba
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v5)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v3); // gfdceba
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v5, v6); // fgdceba
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); // gdfceba
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v3, v5); // gdcfeba
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v5); // gfdcbea
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // fgdcbea
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // gdfcbea
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v5); // gdcfbea
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v4, v5); // gdcbfea
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // gfdcbae
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v6); // fgdcbae
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gdfcbae
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gdcfbae
																}
																else {
																	swap(v0, v6); swap(v1, v3); swap(v4, v5); swap(v5, v6); // gdcbfae
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v4, v6); // gdcbafe
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v1)) {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v5); // gfedbca
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // fgedbca
																}
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v5, v6); // fegdbca
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // gfdebca
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fgdebca
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); // gdfebca
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v5)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); // gfdbeca
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // fgdbeca
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); // gdfbeca
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); // gdbfeca
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // gfdbcea
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fgdbcea
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v5); // gdfbcea
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // gdbfcea
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v5); // gdbcfea
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gfdbcae
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // fgdbcae
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // gdfbcae
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gdbfcae
																}
																else {
																	swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // gdbcfae
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v6); // gdbcafe
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v1)) {
												if (compare(v5, v1)) {
													if (compare(v4, v3)) {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // gfebdca
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fgebdca
																}
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // fegbdca
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); // gfbedca
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // fgbedca
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v5); // gfbdeca
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v5); // gfbdcea
																}
															}
															else {
																swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // gfbdcae
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v5, v6); // fgbdeca
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // fgbdcea
																}
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v6); // fgbdcae
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v2, v5); swap(v3, v4); // gbfedca
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v2, v5); // gbfdeca
																}
																else {
																	swap(v0, v6); swap(v2, v3); swap(v3, v5); // gbdfeca
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v2, v5); swap(v4, v5); // gbfdcea
																}
																else {
																	swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // gbdfcea
																}
															}
															else {
																swap(v0, v6); swap(v2, v3); swap(v4, v5); // gbdcfea
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // gbfdcae
															}
															else {
																swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gbdfcae
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // gbdcfae
															}
															else {
																swap(v0, v6); swap(v2, v3); swap(v4, v6); // gbdcafe
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v5, v1)) {
														if (compare(v6, v4)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // fbgedca
															}
															else {
																swap(v0, v5); swap(v2, v6); swap(v5, v6); // fbgdeca
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // febgdca
															}
															else {
																swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fbegdca
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); // bgfedca
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // bfgedca
																}
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bfegdca
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); // bgfdeca
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v5, v6); // bfgdeca
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); // bgdfeca
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // fbgdcea
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v5); // bgfdcea
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // bfgdcea
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bgdfcea
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v5); // bgdcfea
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v6); swap(v4, v6); // fbgdcae
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // bgfdcae
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v6); // bfgdcae
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bgdfcae
																}
																else {
																	swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // bgdcfae
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v6); // bgdcafe
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v1)) {
											if (compare(v2, v1)) {
												if (compare(v4, v1)) {
													if (compare(v4, v2)) {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fedgcba
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fdgecba
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // fdegcba
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dgfecba
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // dfgecba
																}
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); // dfegcba
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // dgfceba
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v5, v6); // dgcfeba
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // fdgceba
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); // dfgceba
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fdgcbea
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dgfcbea
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // dfgcbea
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dgcfbea
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v5); // dgcbfea
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // fdgcbae
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // dgfcbae
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dfgcbae
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v6); // dgcfbae
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dgcbfae
																}
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v6); // dgcbafe
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // fedgbca
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // fdgebca
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fdegbca
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // dgfebca
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dfgebca
																}
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // dfegbca
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); // dgfbeca
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // dgbfeca
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v5, v6); // fdgbeca
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // dfgbeca
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // fdgbcea
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // dgfbcea
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dfgbcea
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dgbfcea
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // dgbcfea
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v6); // fdgbcae
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dgfbcae
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // dfgbcae
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // dgbfcae
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dgbcfae
																}
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // dgbcafe
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v5, v1)) {
													if (compare(v5, v3)) {
														if (compare(v6, v4)) {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // fdbgeca
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fdbgcea
																}
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // fdbgcae
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fedbgca
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // fdebgca
																}
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fdbegca
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dfebgca
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); // dfbgeca
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // dfbegca
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // dfbgcea
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dfbgcae
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // dbgfeca
																}
																else {
																	swap(v0, v3); swap(v2, v5); swap(v3, v6); // dbfgeca
																}
															}
															else {
																swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // dbfegca
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dbgfcea
																}
																else {
																	swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // dbfgcea
																}
															}
															else {
																swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // dbgcfea
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v5)) {
																swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // dbgfcae
															}
															else {
																swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dbfgcae
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dbgcfae
															}
															else {
																swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // dbgcafe
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // febdgca
															}
															else {
																swap(v0, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // fbedgca
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // fbdgeca
															}
															else {
																swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fbdegca
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // bfedgca
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // bfdgeca
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bfdegca
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // bdgfeca
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); // bdfgeca
																}
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // bdfegca
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fbdgcea
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bfdgcea
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bdgfcea
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // bdfgcea
																}
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // bdgcfea
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // bdgfcae
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bdgcfae
																}
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // bdgcafe
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // fbdgcae
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // bfdgcae
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bdfgcae
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v4)) {
										if (compare(v2, v1)) {
											if (compare(v6, v1)) {
												if (compare(v4, v0)) {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v5, v6); // fdcgeba
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v3, v6); // dfcgeba
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // dcgfeba
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); // dcfgeba
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fdcgbea
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v5); // dfcgbea
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dcgfbea
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // dcfgbea
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // dcgbfea
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v0)) {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // dcgfbae
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dcgbfae
															}
														}
														else {
															swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // dcgbafe
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v6); // fdcgbae
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dfcgbae
															}
														}
														else {
															swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dcfgbae
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v4, v6); swap(v5, v6); // fdcbgea
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dfcbgea
															}
														}
														else {
															swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dcfbgea
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v4, v6); // fdcbgae
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v6); // dfcbgae
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // dcfbgae
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v4, v5); swap(v5, v6); // fdcbage
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dfcbage
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dcfbage
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v5); // dcbgfea
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dcbgfae
															}
														}
														else {
															if (compare(v6, v0)) {
																swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v6); // dcbgafe
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v4, v6); // dcbagfe
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dcbfgea
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v6); // dcbfgae
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dcbfage
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v4, v5); swap(v5, v6); // dcbafge
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v5, v2)) {
													if (compare(v4, v0)) {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // fdbcgea
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dfbcgea
															}
														}
														else {
															swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dbfcgea
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v6); // fdbcgae
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // dfbcgae
																}
															}
															else {
																swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // dbfcgae
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // fdbcage
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dfbcage
																}
															}
															else {
																swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dbfcage
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v3); swap(v3, v6); swap(v4, v5); // dbcgfea
															}
															else {
																swap(v0, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dbcgfae
															}
														}
														else {
															if (compare(v6, v0)) {
																swap(v0, v3); swap(v3, v6); swap(v4, v6); // dbcgafe
															}
															else {
																swap(v0, v3); swap(v4, v6); // dbcagfe
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dbcfgea
															}
															else {
																swap(v0, v3); swap(v3, v5); swap(v4, v6); // dbcfgae
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dbcfage
															}
															else {
																swap(v0, v3); swap(v4, v5); swap(v5, v6); // dbcafge
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v4, v0)) {
														if (compare(v5, v1)) {
															swap(v0, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // fbdcgea
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // bfdcgea
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bdfcgea
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v3); swap(v4, v6); // fbdcgae
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v6); // bfdcgae
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // bdfcgae
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // fbdcage
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // bfdcage
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bdfcage
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v5); // bdcgfea
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bdcgfae
															}
														}
														else {
															if (compare(v6, v0)) {
																swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v6); // bdcgafe
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v4, v6); // bdcagfe
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bdcfgea
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v6); // bdcfgae
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bdcfage
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v4, v5); swap(v5, v6); // bdcafge
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v5, v3)) {
											if (compare(v6, v0)) {
												if (compare(v4, v1)) {
													if (compare(v4, v3)) {
														if (compare(v2, v1)) {
															if (compare(v6, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // fedcgba
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v5, v6); // fedcbga
															}
														}
														else {
															if (compare(v3, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // fedbcga
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v5, v6); // febdcga
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fdecgba
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fdcegba
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // fdecbga
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v5, v6); // fdcebga
																}
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v5, v6); // fdebcga
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v5, v1)) {
															if (compare(v3, v1)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // fdbecga
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v2, v4); swap(v5, v6); // fbedcga
																}
																else {
																	swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // fbdecga
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v5, v6); // bfedcga
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // bfdecga
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v2, v1)) {
																swap(v0, v5); swap(v1, v3); swap(v5, v6); // fdcbega
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v5, v6); // fdbcega
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v3); swap(v5, v6); // fbdcega
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v5, v6); // bfdcega
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v2, v1)) {
														if (compare(v4, v1)) {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v4); swap(v2, v3); // fedcbag
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); // fdecbag
																}
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v3, v4); // fdcebag
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v3); // fdcbeag
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v4, v5); // fdcbaeg
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); // fedbcag
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v4); // fdebcag
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); // fdbecag
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v3); // fdbceag
																}
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v5); // fdbcaeg
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); // febdcag
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v2, v4); // fbedcag
																}
																else {
																	swap(v0, v5); swap(v2, v3); swap(v3, v4); // fbdecag
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v2, v3); // fbdceag
															}
															else {
																swap(v0, v5); swap(v2, v3); swap(v4, v5); // fbdcaeg
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); // bfedcag
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); // bfdecag
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v3); // bfdceag
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v5); // bfdcaeg
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v1)) {
												if (compare(v6, v0)) {
													if (compare(v4, v1)) {
														if (compare(v6, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // dfecgba
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v6); // dfcegba
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // dcfegba
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dfecbga
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // dfebcga
																}
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dfcebga
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dcfebga
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v5, v6); // dfcbega
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // dcfbega
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dfbecga
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // dfbcega
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v4, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // dfecbag
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v5); // dfcebag
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dcfebag
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v3); swap(v1, v5); swap(v3, v5); // dfcbeag
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); // dcfbeag
																}
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v5); // dfcbaeg
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dcfbaeg
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); // dfebcag
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dfbecag
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); // dfbceag
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dfbcaeg
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v3, v1)) {
														if (compare(v6, v0)) {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dbfecga
															}
															else {
																swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // dbfcega
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dbfecag
																}
																else {
																	swap(v0, v3); swap(v2, v5); swap(v3, v5); // dbfceag
																}
															}
															else {
																swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dbfcaeg
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bdfecga
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // bdfcega
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bdfecag
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); // bdfceag
																}
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bdfcaeg
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v0)) {
															if (compare(v2, v1)) {
																swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v5, v6); // dcbfega
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v3, v5); swap(v5, v6); // dbcfega
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v5, v6); // bdcfega
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																swap(v0, v3); swap(v1, v2); swap(v3, v5); // dcbfeag
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v3, v5); // dbcfeag
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v3, v5); // bdcfeag
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v2, v1)) {
																swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v5); // dcbfaeg
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v3, v5); swap(v4, v5); // dbcfaeg
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v5); // bdcfaeg
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																swap(v0, v3); swap(v1, v2); swap(v4, v5); // dcbafeg
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v4, v5); // dbcafeg
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v4, v5); // bdcafeg
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v6, v2)) {
									if (compare(v6, v3)) {
										if (compare(v3, v1)) {
											if (compare(v2, v1)) {
												if (compare(v5, v1)) {
													if (compare(v5, v2)) {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v5); // gefdcba
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // egfdcba
																}
																else {
																	swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v6); // efgdcba
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // gedfcba
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // egdfcba
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); // gdefcba
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v5); // gedcfba
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // egdcfba
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // gdecfba
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v5); // gdcefba
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v3); // gedcbfa
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v6); // egdcbfa
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); // gdecbfa
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v6); swap(v1, v3); swap(v3, v4); // gdcebfa
															}
															else {
																swap(v0, v6); swap(v1, v3); // gdcbefa
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v5, v6); // gedcbaf
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // egdcbaf
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // gdecbaf
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v1)) {
																	swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v5, v6); // gdcebaf
																}
																else {
																	swap(v0, v6); swap(v1, v3); swap(v5, v6); // gdcbeaf
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v4, v6); swap(v5, v6); // gdcbaef
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v1)) {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); // gefdbca
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v6); // egfdbca
																}
																else {
																	swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // efgdbca
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); // gedfbca
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // egdfbca
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // gdefbca
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // gedbfca
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // egdbfca
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v5); // gdebfca
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // gdbefca
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v1)) {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); // gedbcfa
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // egdbcfa
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); // gdebcfa
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); // gdbecfa
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); // gdbcefa
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // gedbcaf
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // egdbcaf
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v5, v6); // gdebcaf
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // gdbecaf
																}
																else {
																	swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v5, v6); // gdbceaf
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // gdbcaef
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v1)) {
												if (compare(v4, v1)) {
													if (compare(v5, v3)) {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); // gefbdca
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // egfbdca
																}
																else {
																	swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // efgbdca
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // gebfdca
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // egbfdca
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v5); // gebdfca
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v2, v4); // gebdcfa
																}
															}
															else {
																swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v5, v6); // gebdcaf
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // egbdfca
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v6); // egbdcfa
																}
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // egbdcaf
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // gbefdca
																}
																else {
																	swap(v0, v6); swap(v2, v4); swap(v4, v5); // gbedfca
																}
															}
															else {
																swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // gbdefca
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v2, v4); // gbedcfa
																}
																else {
																	swap(v0, v6); swap(v2, v3); swap(v3, v4); // gbdecfa
																}
															}
															else {
																swap(v0, v6); swap(v2, v3); // gbdcefa
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v2, v4); swap(v5, v6); // gbedcaf
															}
															else {
																swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // gbdecaf
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v2, v3); swap(v5, v6); // gbdceaf
															}
															else {
																swap(v0, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // gbdcaef
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v4, v1)) {
														if (compare(v6, v5)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ebgfdca
															}
															else {
																swap(v0, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // ebgdfca
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // efbgdca
															}
															else {
																swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // ebfgdca
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // bgefdca
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // begfdca
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // befgdca
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v5); // bgedfca
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // begdfca
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bgdefca
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v6); swap(v4, v6); // ebgdcfa
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); // bgedcfa
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v6); // begdcfa
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); // bgdecfa
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); // bgdcefa
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // ebgdcaf
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v5, v6); // bgedcaf
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // begdcaf
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // bgdecaf
																}
																else {
																	swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v5, v6); // bgdceaf
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // bgdcaef
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v1)) {
											if (compare(v2, v1)) {
												if (compare(v5, v1)) {
													if (compare(v5, v2)) {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // efdgcba
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // edgfcba
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // edfgcba
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // dgefcba
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // degfcba
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // defgcba
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dgecfba
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dgcefba
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // edgcfba
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // degcfba
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // edgcbfa
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // dgecbfa
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); // degcbfa
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v6); // dgcebfa
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v3, v6); // dgcbefa
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // edgcbaf
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dgecbaf
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // degcbaf
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v1)) {
																	swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dgcebaf
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v5, v6); // dgcbeaf
																}
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dgcbaef
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // efdgbca
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // edgfbca
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // edfgbca
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dgefbca
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // degfbca
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); // defgbca
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // dgebfca
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dgbefca
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // edgbfca
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // degbfca
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v6); // edgbcfa
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); // dgebcfa
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // degbcfa
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // dgbecfa
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); // dgbcefa
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // edgbcaf
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // dgebcaf
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // degbcaf
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dgbecaf
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // dgbceaf
																}
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dgbcaef
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v4, v1)) {
													if (compare(v4, v3)) {
														if (compare(v6, v5)) {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // edbgfca
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // edbgcfa
																}
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // edbgcaf
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // efdbgca
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v6); // edfbgca
																}
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // edbfgca
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // defbgca
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // debgfca
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // debfgca
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); // debgcfa
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // debgcaf
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v5, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dbgefca
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // dbegfca
																}
																else {
																	swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dbefgca
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // dbgecfa
																}
																else {
																	swap(v0, v3); swap(v2, v4); swap(v3, v6); // dbegcfa
																}
															}
															else {
																swap(v0, v3); swap(v2, v6); swap(v3, v6); // dbgcefa
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dbgecaf
															}
															else {
																swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // dbegcaf
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // dbgceaf
															}
															else {
																swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dbgcaef
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v6); // efbdgca
															}
															else {
																swap(v0, v4); swap(v2, v5); swap(v4, v6); // ebfdgca
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ebdgfca
															}
															else {
																swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // ebdfgca
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v6); // befdgca
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bedgfca
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // bedfgca
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bdgefca
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // bdegfca
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bdefgca
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // ebdgcfa
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // bedgcfa
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // bdgecfa
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); // bdegcfa
																}
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); // bdgcefa
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bdgecaf
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // bdgceaf
																}
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bdgcaef
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ebdgcaf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bedgcaf
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // bdegcaf
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v5)) {
										if (compare(v2, v1)) {
											if (compare(v6, v1)) {
												if (compare(v5, v0)) {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // edcgfba
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v5); // decgfba
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dcgefba
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // dcegfba
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v6); // edcgbfa
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v3, v6); // decgbfa
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // dcgebfa
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); // dcegbfa
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); // dcgbefa
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v4, v0)) {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dcgebaf
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // dcgbeaf
															}
														}
														else {
															swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dcgbaef
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // edcgbaf
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v5, v6); // decgbaf
															}
														}
														else {
															swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // dcegbaf
														}
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v4, v6); // edcbgfa
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v6); // decbgfa
															}
														}
														else {
															swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // dcebgfa
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v4); swap(v1, v3); swap(v4, v6); swap(v5, v6); // edcbgaf
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // decbgaf
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dcebgaf
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v4); swap(v1, v3); swap(v5, v6); // edcbagf
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v5, v6); // decbagf
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // dcebagf
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v4, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v3); swap(v1, v2); swap(v3, v6); // dcbgefa
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v5, v6); // dcbgeaf
															}
														}
														else {
															if (compare(v6, v0)) {
																swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dcbgaef
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v4, v6); swap(v5, v6); // dcbagef
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v6); // dcbegfa
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dcbegaf
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v5, v6); // dcbeagf
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v5, v6); // dcbaegf
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v4, v2)) {
													if (compare(v5, v0)) {
														if (compare(v4, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v6); // edbcgfa
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // debcgfa
															}
														}
														else {
															swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // dbecgfa
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // edbcgaf
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // debcgaf
																}
															}
															else {
																swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dbecgaf
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v5, v6); // edbcagf
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // debcagf
																}
															}
															else {
																swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // dbecagf
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v4, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v3); swap(v3, v6); // dbcgefa
															}
															else {
																swap(v0, v3); swap(v3, v6); swap(v5, v6); // dbcgeaf
															}
														}
														else {
															if (compare(v6, v0)) {
																swap(v0, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dbcgaef
															}
															else {
																swap(v0, v3); swap(v4, v6); swap(v5, v6); // dbcagef
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v3); swap(v3, v4); swap(v4, v6); // dbcegfa
															}
															else {
																swap(v0, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dbcegaf
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v3); swap(v3, v4); swap(v5, v6); // dbceagf
															}
															else {
																swap(v0, v3); swap(v5, v6); // dbcaegf
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v5, v0)) {
														if (compare(v4, v1)) {
															swap(v0, v4); swap(v2, v3); swap(v4, v6); // ebdcgfa
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v6); // bedcgfa
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // bdecgfa
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // ebdcgaf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // bedcgaf
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bdecgaf
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v3); swap(v5, v6); // ebdcagf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v5, v6); // bedcagf
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // bdecagf
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v4, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v3); swap(v3, v6); // bdcgefa
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v5, v6); // bdcgeaf
															}
														}
														else {
															if (compare(v6, v0)) {
																swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bdcgaef
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v4, v6); swap(v5, v6); // bdcagef
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v6); // bdcegfa
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bdcegaf
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v5, v6); // bdceagf
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v5, v6); // bdcaegf
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v3)) {
											if (compare(v6, v0)) {
												if (compare(v5, v1)) {
													if (compare(v5, v3)) {
														if (compare(v2, v1)) {
															if (compare(v6, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v6); // efdcgba
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // efdcbga
															}
														}
														else {
															if (compare(v3, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // efdbcga
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // efbdcga
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // edfcgba
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v6); // edcfgba
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // edfcbga
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // edcfbga
																}
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // edfbcga
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v1)) {
															if (compare(v3, v1)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // edbfcga
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // ebfdcga
																}
																else {
																	swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ebdfcga
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // befdcga
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bedfcga
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v2, v1)) {
																swap(v0, v4); swap(v1, v3); swap(v4, v5); swap(v5, v6); // edcbfga
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // edbcfga
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // ebdcfga
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // bedcfga
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v2, v1)) {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v5); // efdcbag
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // edfcbag
																}
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v5); // edcfbag
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v3); swap(v4, v5); // edcbfag
															}
															else {
																swap(v0, v4); swap(v1, v3); // edcbafg
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // efdbcag
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v5); // edfbcag
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // edbfcag
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v5); // edbcfag
																}
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v3); // edbcafg
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v5); // efbdcag
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v2, v5); swap(v4, v5); // ebfdcag
																}
																else {
																	swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // ebdfcag
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v2, v3); swap(v4, v5); // ebdcfag
															}
															else {
																swap(v0, v4); swap(v2, v3); // ebdcafg
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v5); // befdcag
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bedfcag
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v5); // bedcfag
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v3); // bedcafg
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v1)) {
												if (compare(v6, v0)) {
													if (compare(v5, v1)) {
														if (compare(v6, v1)) {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // defcgba
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // decfgba
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dcefgba
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // defcbga
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // defbcga
																}
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v5, v6); // decfbga
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // dcefbga
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // decbfga
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dcebfga
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // debfcga
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // debcfga
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v5, v1)) {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); // defcbag
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v3, v5); // decfbag
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); // dcefbag
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v5); // decbfag
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // dcebfag
																}
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v3); swap(v1, v4); swap(v3, v4); // decbafg
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); // dcebafg
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // defbcag
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); // debfcag
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // debcfag
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); // debcafg
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v3, v1)) {
														if (compare(v6, v0)) {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // dbefcga
															}
															else {
																swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dbecfga
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v3); swap(v2, v4); swap(v3, v5); // dbefcag
																}
																else {
																	swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // dbecfag
																}
															}
															else {
																swap(v0, v3); swap(v2, v4); swap(v3, v4); // dbecafg
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // bdefcga
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bdecfga
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); // bdefcag
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // bdecfag
																}
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); // bdecafg
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v6, v0)) {
															if (compare(v2, v1)) {
																swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dcbefga
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dbcefga
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bdcefga
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v5); // dcbefag
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v3, v4); swap(v4, v5); // dbcefag
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v5); // bdcefag
																}
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v2, v1)) {
																swap(v0, v3); swap(v1, v2); swap(v3, v4); // dcbeafg
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v3, v4); // dbceafg
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v3, v4); // bdceafg
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																swap(v0, v3); swap(v1, v2); // dcbaefg
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v3); // dbcaefg
																}
																else {
																	swap(v0, v1); swap(v1, v3); // bdcaefg
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
						else {
							if (compare(v5, v4)) {
								if (compare(v6, v0)) {
									if (compare(v6, v1)) {
										if (compare(v3, v1)) {
											if (compare(v6, v3)) {
												if (compare(v4, v0)) {
													if (compare(v4, v1)) {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // gfedbac
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v6); // fgedbac
																}
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); // fegdbac
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gfdebac
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // fgdebac
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // gdfebac
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v5)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // gfdbeac
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); // fgdbeac
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v5, v6); // gdfbeac
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // gdbfeac
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gfdbaec
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // fgdbaec
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // gdfbaec
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // gfdbace
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fgdbace
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v6); // gdfbace
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gdbfaec
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // gdbface
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // gdbafec
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gdbafce
																}
																else {
																	swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // gdbacfe
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); // fedgbac
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); // fdgebac
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // fdegbac
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dgfebac
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // dfgebac
																}
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dfegbac
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // dgfbeac
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); // dgbfeac
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); // fdgbeac
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); // dfgbeac
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v5); // fdgbaec
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dgfbaec
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // dfgbaec
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // fdgbace
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // dgfbace
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dfgbace
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // dgbfaec
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dgbface
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v5); // dgbafec
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // dgbafce
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v6); // dgbacfe
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v0)) {
												if (compare(v4, v0)) {
													if (compare(v4, v3)) {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gfebdac
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // fgebdac
																}
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); // fegbdac
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // gfbedac
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); // fgbedac
																}
															}
															else {
																swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // gbfedac
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v5)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v5, v6); // gfbdeac
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v6); // fgbdeac
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v2, v5); swap(v5, v6); // gbfdeac
															}
															else {
																swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // gbdfeac
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v5, v1)) {
															if (compare(v6, v5)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // gfbdaec
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v5); // fgbdaec
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // gbfdaec
																}
																else {
																	swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gbdfaec
																}
															}
															else {
																swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // gbdafec
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v6); // gfbdace
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // fgbdace
																}
															}
															else {
																swap(v0, v6); swap(v2, v5); swap(v4, v6); // gbfdace
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // gbdface
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gbdafce
																}
																else {
																	swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // gbdacfe
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v4, v0)) {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // gfebadc
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fgebadc
																}
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // fegbadc
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gfbeadc
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // fgbeadc
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // gfbaedc
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gfbadec
																}
															}
															else {
																swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // gfbadce
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); // fgbaedc
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // fgbadec
																}
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fgbadce
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gbfeadc
															}
															else {
																swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // gbfaedc
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gbfadec
															}
															else {
																swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // gbfadce
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v2, v6); swap(v3, v5); // gbafedc
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // gbafdec
																}
																else {
																	swap(v0, v6); swap(v2, v6); swap(v4, v5); // gbadfec
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gbafdce
																}
																else {
																	swap(v0, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // gbadfce
																}
															}
															else {
																swap(v0, v6); swap(v2, v6); swap(v4, v6); // gbadcfe
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v3)) {
											if (compare(v3, v0)) {
												if (compare(v4, v0)) {
													if (compare(v5, v1)) {
														if (compare(v6, v4)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v2, v6); swap(v3, v4); // fbgedac
															}
															else {
																swap(v0, v5); swap(v2, v6); // fbgdeac
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // febgdac
															}
															else {
																swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // fbegdac
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // bgfedac
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); // bfgedac
																}
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // bfegdac
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v5, v6); // bgfdeac
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); // bfgdeac
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // bgdfeac
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v6); swap(v4, v5); // fbgdaec
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // bgfdaec
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v5); // bfgdaec
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bgdfaec
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // bgdafec
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // fbgdace
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v6); // bgfdace
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // bfgdace
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // bgdface
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bgdafce
																}
																else {
																	swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // bgdacfe
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v4, v0)) {
														if (compare(v5, v1)) {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // febgadc
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // fbgeadc
																}
																else {
																	swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fbegadc
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bgfeadc
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // bfgeadc
																}
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bfegadc
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // bgfaedc
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); // bgafedc
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v6); swap(v3, v5); // fbgaedc
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); // bfgaedc
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // fbgadec
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bgfadec
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // bfgadec
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fbgadce
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // bgfadce
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bfgadce
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // bgafdec
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v5); // bgadfec
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bgafdce
																}
																else {
																	swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // bgadfce
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v6); // bgadcfe
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v4, v0)) {
													if (compare(v5, v3)) {
														if (compare(v4, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // fedbgac
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v6); // fdebgac
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); // fdbgeac
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // fdbegac
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // dfebgac
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // dfbgeac
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dfbegac
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v2, v6); swap(v3, v5); // dbgfeac
																}
																else {
																	swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // dbfgeac
																}
															}
															else {
																swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dbfegac
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // fdbgaec
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dfbgaec
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // dbgfaec
																}
																else {
																	swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dbfgaec
																}
															}
															else {
																swap(v0, v3); swap(v2, v6); swap(v4, v5); // dbgafec
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v0)) {
																swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dbgface
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // dbgafce
																}
																else {
																	swap(v0, v3); swap(v2, v6); swap(v4, v6); // dbgacfe
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fdbgace
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // dfbgace
																}
															}
															else {
																swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // dbfgace
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v6); // febdgac
															}
															else {
																swap(v0, v5); swap(v2, v4); swap(v4, v6); // fbedgac
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v5); swap(v2, v3); swap(v3, v6); // fbdgeac
															}
															else {
																swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // fbdegac
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v6); // bfedgac
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); // bfdgeac
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // bfdegac
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); // bdgfeac
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // bdfgeac
																}
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bdfegac
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // fbdgaec
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // bfdgaec
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // bdgfaec
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bdfgaec
																}
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v5); // bdgafec
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bdgface
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // bdgafce
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v6); // bdgacfe
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fbdgace
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bfdgace
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // bdfgace
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v4)) {
										if (compare(v3, v0)) {
											if (compare(v3, v1)) {
												if (compare(v5, v0)) {
													if (compare(v5, v1)) {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fdbagec
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // fdbagce
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fdbacge
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // dfbagec
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v6); // dfbagce
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // dfbacge
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															swap(v0, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // dbfagec
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v3); swap(v2, v5); swap(v4, v6); // dbfagce
															}
															else {
																swap(v0, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // dbfacge
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // dbagfec
															}
															else {
																swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dbagfce
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // dbagcfe
															}
															else {
																swap(v0, v3); swap(v2, v3); swap(v4, v6); // dbacgfe
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dbafgec
															}
															else {
																swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // dbafgce
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dbafcge
															}
															else {
																swap(v0, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // dbacfge
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fbdagec
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // fbdagce
															}
															else {
																swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fbdacge
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bfdagec
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // bfdagce
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bfdacge
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // bdfagec
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v6); // bdfagce
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // bdfacge
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // bdagfec
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bdagfce
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // bdagcfe
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v6); // bdacgfe
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bdafgec
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // bdafgce
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bdafcge
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // bdacfge
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v3)) {
												if (compare(v5, v0)) {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // fbagedc
															}
															else {
																swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fbagdec
															}
														}
														else {
															swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // fbagdce
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // bfagedc
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bfagdec
															}
														}
														else {
															swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // bfagdce
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v4, v3)) {
															if (compare(v6, v5)) {
																swap(v0, v1); swap(v2, v6); swap(v3, v5); // bagfedc
															}
															else {
																swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v5, v6); // bafgedc
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v5); // bagfdec
																}
																else {
																	swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bafgdec
																}
															}
															else {
																swap(v0, v1); swap(v2, v6); swap(v4, v5); // bagdfec
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v5)) {
																swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bagfdce
															}
															else {
																swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v6); // bafgdce
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v2, v6); swap(v4, v5); swap(v5, v6); // bagdfce
															}
															else {
																swap(v0, v1); swap(v2, v6); swap(v4, v6); // bagdcfe
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v5, v0)) {
														if (compare(v5, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // fbadgec
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v5); swap(v2, v5); swap(v4, v6); // fbadgce
																}
																else {
																	swap(v0, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // fbadcge
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // bfadgec
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v6); // bfadgce
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // bfadcge
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															swap(v0, v1); swap(v2, v5); swap(v4, v6); swap(v5, v6); // bafdgec
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v1); swap(v2, v5); swap(v4, v6); // bafdgce
															}
															else {
																swap(v0, v1); swap(v2, v5); swap(v4, v5); swap(v5, v6); // bafdcge
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v5); // badgfec
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // badgfce
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v6); // badgcfe
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v4, v6); // badcgfe
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // badfgec
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v6); // badfgce
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // badfcge
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v4, v5); swap(v5, v6); // badcfge
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v5, v1)) {
											if (compare(v6, v2)) {
												if (compare(v4, v3)) {
													if (compare(v4, v1)) {
														if (compare(v3, v0)) {
															if (compare(v3, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fedbagc
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // febdagc
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // febagdc
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // febadgc
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v3, v0)) {
																swap(v0, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // fbedagc
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // fbeagdc
																}
																else {
																	swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fbeadgc
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fbaegdc
															}
															else {
																swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // fbaedgc
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v4, v1)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // fdebagc
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dfebagc
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fdbeagc
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // dfbeagc
																}
															}
															else {
																swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fbdeagc
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // fdbaegc
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v5, v6); // dfbaegc
															}
														}
														else {
															if (compare(v3, v0)) {
																swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // fbdaegc
															}
															else {
																swap(v0, v5); swap(v2, v5); swap(v5, v6); // fbadegc
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v3, v1)) {
														if (compare(v4, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fedbacg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v5); // fdebacg
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // dfebacg
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fdbeacg
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); // dfbeacg
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v5); // febdacg
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v2, v4); swap(v4, v5); // fbedacg
																}
																else {
																	swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fbdeacg
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // febadcg
															}
															else {
																swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // fbeadcg
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); // fdbaecg
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v5); // dfbaecg
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fdbaceg
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v5); // dfbaceg
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v2, v3); swap(v3, v5); // fbdaecg
															}
															else {
																swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fbdaceg
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v2, v5); swap(v3, v4); // fbaedcg
																}
																else {
																	swap(v0, v5); swap(v2, v5); // fbadecg
																}
															}
															else {
																swap(v0, v5); swap(v2, v5); swap(v4, v5); // fbadceg
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v3)) {
												if (compare(v6, v2)) {
													if (compare(v4, v0)) {
														if (compare(v3, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // bfedagc
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bfdeagc
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // bfeagdc
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bfeadgc
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bfaegdc
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // bfaedgc
																}
															}
															else {
																if (compare(v3, v0)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // bfdaegc
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v5, v6); // bfadegc
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bafegdc
																}
																else {
																	swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v5, v6); // bafedgc
																}
															}
															else {
																swap(v0, v1); swap(v2, v5); swap(v5, v6); // bafdegc
															}
														}
													}
												}
												else {
													if (compare(v3, v0)) {
														if (compare(v4, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v5); // bfedacg
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bfdeacg
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); // bfdaecg
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bfdaceg
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																if (compare(v4, v0)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // bfeadcg
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); // bfaedcg
																}
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v5); // bfadecg
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v5); // bfadceg
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v1); swap(v2, v5); swap(v3, v4); // bafedcg
																}
																else {
																	swap(v0, v1); swap(v2, v5); // bafdecg
																}
															}
															else {
																swap(v0, v1); swap(v2, v5); swap(v4, v5); // bafdceg
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v3, v1)) {
														if (compare(v4, v0)) {
															if (compare(v6, v2)) {
																swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // dbfeagc
															}
															else {
																swap(v0, v3); swap(v2, v5); swap(v3, v4); // dbfeacg
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v3); swap(v2, v5); swap(v5, v6); // dbfaegc
																}
																else {
																	swap(v0, v3); swap(v2, v5); // dbfaecg
																}
															}
															else {
																swap(v0, v3); swap(v2, v5); swap(v4, v5); // dbfaceg
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v6, v2)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // bdfeagc
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); // bdfeacg
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v5, v6); // bdfaegc
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v5); // bdfaecg
																}
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v5); // bdfaceg
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v6, v2)) {
															if (compare(v3, v0)) {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // dbafegc
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // bdafegc
																}
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v5, v6); // badfegc
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v2, v3); swap(v3, v5); // dbafecg
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); // bdafecg
																}
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v3, v5); // badfecg
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v3, v0)) {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dbafceg
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bdafceg
																}
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v5); // badfceg
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v2, v3); swap(v4, v5); // dbacfeg
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v5); // bdacfeg
																}
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v4, v5); // badcfeg
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v6, v0)) {
									if (compare(v6, v1)) {
										if (compare(v3, v1)) {
											if (compare(v6, v3)) {
												if (compare(v5, v0)) {
													if (compare(v5, v1)) {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v5, v6); // gefdbac
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // egfdbac
																}
																else {
																	swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v5); // efgdbac
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // gedfbac
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // egdfbac
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gdefbac
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gedbfac
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // egdbfac
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // gdebfac
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gdbefac
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // gedbafc
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); // egdbafc
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v6); // gdebafc
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gedbacf
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // egdbacf
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // gdebacf
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // gdbeafc
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gdbeacf
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); // gdbaefc
																}
																else {
																	swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // gdbaecf
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gdbacef
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // efdgbac
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // edgfbac
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // edfgbac
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // dgefbac
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); // degfbac
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // defgbac
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dgebfac
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // dgbefac
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v5); // edgbfac
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // degbfac
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); // edgbafc
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // dgebafc
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); // degbafc
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v5, v6); // edgbacf
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dgebacf
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // degbacf
																}
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); // dgbeafc
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // dgbeacf
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v6); // dgbaefc
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v5, v6); // dgbaecf
																}
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // dgbacef
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v0)) {
												if (compare(v5, v0)) {
													if (compare(v5, v3)) {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // gefbdac
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // egfbdac
																}
																else {
																	swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // efgbdac
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gebfdac
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // egbfdac
																}
															}
															else {
																swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gbefdac
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // gebdfac
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v5); // egbdfac
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // gbedfac
															}
															else {
																swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gbdefac
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v6); // gebdafc
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v6); // egbdafc
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v2, v4); swap(v4, v6); // gbedafc
																}
																else {
																	swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // gbdeafc
																}
															}
															else {
																swap(v0, v6); swap(v2, v3); swap(v3, v6); // gbdaefc
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // gebdacf
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v5, v6); // egbdacf
																}
															}
															else {
																swap(v0, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // gbedacf
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gbdeacf
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // gbdaecf
																}
																else {
																	swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gbdacef
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v5, v0)) {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gefbadc
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // egfbadc
																}
																else {
																	swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); // efgbadc
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // gebfadc
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); // egbfadc
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gebafdc
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // gebadfc
																}
															}
															else {
																swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gebadcf
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // egbafdc
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); // egbadfc
																}
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // egbadcf
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // gbefadc
															}
															else {
																swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gbeafdc
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // gbeadfc
															}
															else {
																swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gbeadcf
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // gbaefdc
																}
																else {
																	swap(v0, v6); swap(v2, v6); swap(v3, v4); // gbaedfc
																}
															}
															else {
																swap(v0, v6); swap(v2, v6); // gbadefc
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // gbaedcf
																}
																else {
																	swap(v0, v6); swap(v2, v6); swap(v5, v6); // gbadecf
																}
															}
															else {
																swap(v0, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // gbadcef
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v3)) {
											if (compare(v3, v0)) {
												if (compare(v5, v0)) {
													if (compare(v4, v1)) {
														if (compare(v6, v5)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // ebgfdac
															}
															else {
																swap(v0, v4); swap(v2, v6); swap(v4, v5); // ebgdfac
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // efbgdac
															}
															else {
																swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ebfgdac
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bgefdac
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // begfdac
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // befgdac
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // bgedfac
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v5); // begdfac
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bgdefac
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v6); // ebgdafc
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v6); // bgedafc
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); // begdafc
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // bgdeafc
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); // bgdaefc
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v6); swap(v5, v6); // ebgdacf
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // bgedacf
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v5, v6); // begdacf
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bgdeacf
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // bgdaecf
																}
																else {
																	swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bgdacef
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v5, v0)) {
														if (compare(v4, v1)) {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // efbgadc
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v2, v6); swap(v3, v5); // ebgfadc
																}
																else {
																	swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // ebfgadc
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // bgefadc
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); // begfadc
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // befgadc
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bgeafdc
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // bgaefdc
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // ebgafdc
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // begafdc
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v6); swap(v3, v4); // ebgadfc
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // bgeadfc
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); // begadfc
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // ebgadcf
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bgeadcf
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // begadcf
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); // bgaedfc
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); // bgadefc
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // bgaedcf
																}
																else {
																	swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v5, v6); // bgadecf
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // bgadcef
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v5, v0)) {
													if (compare(v4, v3)) {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // efdbgac
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // edfbgac
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // edbgfac
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // edbfgac
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // defbgac
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // debgfac
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // debfgac
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // dbgefac
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dbegfac
																}
																else {
																	swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // dbefgac
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); // edbgafc
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // debgafc
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v2, v6); swap(v3, v4); // dbgeafc
																}
																else {
																	swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // dbegafc
																}
															}
															else {
																swap(v0, v3); swap(v2, v6); // dbgaefc
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v0)) {
																swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // dbgeacf
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v3); swap(v2, v6); swap(v5, v6); // dbgaecf
																}
																else {
																	swap(v0, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // dbgacef
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // edbgacf
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // debgacf
																}
															}
															else {
																swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dbegacf
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // efbdgac
															}
															else {
																swap(v0, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // ebfdgac
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // ebdgfac
															}
															else {
																swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ebdfgac
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // befdgac
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // bedgfac
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bedfgac
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // bdgefac
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bdegfac
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // bdefgac
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v3); swap(v3, v6); // ebdgafc
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); // bedgafc
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); // bdgeafc
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // bdegafc
																}
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v6); // bdgaefc
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // bdgeacf
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v5, v6); // bdgaecf
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // bdgacef
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // ebdgacf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // bedgacf
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bdegacf
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v5)) {
										if (compare(v3, v0)) {
											if (compare(v3, v1)) {
												if (compare(v4, v0)) {
													if (compare(v4, v1)) {
														if (compare(v4, v3)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // edbagfc
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // edbagcf
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // edbacgf
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v6); // debagfc
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // debagcf
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v5, v6); // debacgf
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															swap(v0, v3); swap(v2, v4); swap(v4, v6); // dbeagfc
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // dbeagcf
															}
															else {
																swap(v0, v3); swap(v2, v4); swap(v5, v6); // dbeacgf
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v2, v3); swap(v3, v6); // dbagefc
															}
															else {
																swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // dbagecf
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dbagcef
															}
															else {
																swap(v0, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // dbacgef
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // dbaegfc
															}
															else {
																swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dbaegcf
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // dbaecgf
															}
															else {
																swap(v0, v3); swap(v2, v3); swap(v5, v6); // dbacegf
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // ebdagfc
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ebdagcf
															}
															else {
																swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // ebdacgf
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // bedagfc
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bedagcf
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // bedacgf
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v6); // bdeagfc
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // bdeagcf
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v5, v6); // bdeacgf
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); // bdagefc
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // bdagecf
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bdagcef
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // bdacgef
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // bdaegfc
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bdaegcf
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // bdaecgf
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v5, v6); // bdacegf
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v3)) {
												if (compare(v4, v0)) {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ebagfdc
															}
															else {
																swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // ebagdfc
															}
														}
														else {
															swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ebagdcf
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // beagfdc
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // beagdfc
															}
														}
														else {
															swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // beagdcf
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v5); // bagefdc
															}
															else {
																swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // baegfdc
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v2, v6); swap(v3, v4); // bagedfc
																}
																else {
																	swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v6); // baegdfc
																}
															}
															else {
																swap(v0, v1); swap(v2, v6); // bagdefc
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v5, v6); // bagedcf
															}
															else {
																swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // baegdcf
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v2, v6); swap(v5, v6); // bagdecf
															}
															else {
																swap(v0, v1); swap(v2, v6); swap(v4, v6); swap(v5, v6); // bagdcef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v4, v0)) {
														if (compare(v4, v1)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v2, v4); swap(v4, v6); // ebadgfc
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // ebadgcf
																}
																else {
																	swap(v0, v4); swap(v2, v4); swap(v5, v6); // ebadcgf
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v6); // beadgfc
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // beadgcf
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v5, v6); // beadcgf
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															swap(v0, v1); swap(v2, v4); swap(v4, v6); // baedgfc
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v1); swap(v2, v4); swap(v4, v6); swap(v5, v6); // baedgcf
															}
															else {
																swap(v0, v1); swap(v2, v4); swap(v5, v6); // baedcgf
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v2, v3); swap(v3, v6); // badgefc
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v5, v6); // badgecf
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // badgcef
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v4, v6); swap(v5, v6); // badcgef
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v6); // badegfc
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // badegcf
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v5, v6); // badecgf
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v5, v6); // badcegf
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v1)) {
											if (compare(v6, v2)) {
												if (compare(v5, v3)) {
													if (compare(v5, v1)) {
														if (compare(v3, v0)) {
															if (compare(v3, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // efdbagc
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v5, v6); // efbdagc
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // efbagdc
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // efbadgc
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v3, v0)) {
																swap(v0, v4); swap(v2, v5); swap(v5, v6); // ebfdagc
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ebfagdc
																}
																else {
																	swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // ebfadgc
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // ebafgdc
															}
															else {
																swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ebafdgc
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v5, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v5, v6); // edfbagc
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // defbagc
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v3)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // edbfagc
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // debfagc
																}
															}
															else {
																swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // ebdfagc
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // edbafgc
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // debafgc
															}
														}
														else {
															if (compare(v3, v0)) {
																swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ebdafgc
															}
															else {
																swap(v0, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // ebadfgc
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v3, v1)) {
														if (compare(v5, v1)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); // efdbacg
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); // edfbacg
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); // defbacg
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); // edbfacg
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // debfacg
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); // efbdacg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v2, v5); // ebfdacg
																}
																else {
																	swap(v0, v4); swap(v2, v3); swap(v3, v5); // ebdfacg
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); // efbadcg
															}
															else {
																swap(v0, v4); swap(v2, v5); swap(v3, v4); // ebfadcg
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // edbafcg
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v5); // debafcg
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); // edbacfg
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v4); // debacfg
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // ebdafcg
															}
															else {
																swap(v0, v4); swap(v2, v3); swap(v3, v4); // ebdacfg
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // ebafdcg
																}
																else {
																	swap(v0, v4); swap(v2, v4); swap(v4, v5); // ebadfcg
																}
															}
															else {
																swap(v0, v4); swap(v2, v4); // ebadcfg
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v3)) {
												if (compare(v6, v2)) {
													if (compare(v5, v0)) {
														if (compare(v3, v0)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v5, v6); // befdagc
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // bedfagc
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // befagdc
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // befadgc
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // beafgdc
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // beafdgc
																}
															}
															else {
																if (compare(v3, v0)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bedafgc
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // beadfgc
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v6); // baefgdc
																}
																else {
																	swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // baefdgc
																}
															}
															else {
																swap(v0, v1); swap(v2, v4); swap(v4, v5); swap(v5, v6); // baedfgc
															}
														}
													}
												}
												else {
													if (compare(v3, v0)) {
														if (compare(v5, v0)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); // befdacg
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); // bedfacg
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bedafcg
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); // bedacfg
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v5, v3)) {
																if (compare(v5, v0)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); // befadcg
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // beafdcg
																}
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v5); // beadfcg
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v4); // beadcfg
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v5); // baefdcg
																}
																else {
																	swap(v0, v1); swap(v2, v4); swap(v4, v5); // baedfcg
																}
															}
															else {
																swap(v0, v1); swap(v2, v4); // baedcfg
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v3, v1)) {
														if (compare(v5, v0)) {
															if (compare(v6, v2)) {
																swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dbefagc
															}
															else {
																swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // dbefacg
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // dbeafgc
																}
																else {
																	swap(v0, v3); swap(v2, v4); swap(v4, v5); // dbeafcg
																}
															}
															else {
																swap(v0, v3); swap(v2, v4); // dbeacfg
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v6, v2)) {
																swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bdefagc
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // bdefacg
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // bdeafgc
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v5); // bdeafcg
																}
															}
															else {
																swap(v0, v1); swap(v1, v3); swap(v2, v4); // bdeacfg
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v6, v2)) {
															if (compare(v3, v0)) {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dbaefgc
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bdaefgc
																}
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // badefgc
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dbaefcg
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bdaefcg
																}
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v5); // badefcg
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v3, v0)) {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v2, v3); swap(v3, v4); // dbaecfg
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); // bdaecfg
																}
															}
															else {
																swap(v0, v1); swap(v2, v3); swap(v3, v4); // badecfg
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v3, v1)) {
																	swap(v0, v3); swap(v2, v3); // dbacefg
																}
																else {
																	swap(v0, v1); swap(v1, v3); swap(v2, v3); // bdacefg
																}
															}
															else {
																swap(v0, v1); swap(v2, v3); // badcefg
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
					else {
						if (compare(v3, v0)) {
							if (compare(v5, v4)) {
								if (compare(v6, v3)) {
									if (compare(v6, v2)) {
										if (compare(v2, v1)) {
											if (compare(v3, v1)) {
												if (compare(v4, v1)) {
													if (compare(v4, v3)) {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); // gfecdba
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // fgecdba
																}
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fegcdba
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v3, v4); // gfcedba
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v5, v6); // fgcedba
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); // gcfedba
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v5)) {
																swap(v0, v6); swap(v1, v5); // gfcdeba
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v5, v6); // fgcdeba
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v5); // gcfdeba
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); // gcdfeba
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v6, v5)) {
																swap(v0, v6); swap(v1, v5); swap(v4, v5); // gfcdbea
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v4, v6); swap(v5, v6); // fgcdbea
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v5); // gcfdbea
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // gcdfbea
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v5); // gcdbfea
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v4, v5); swap(v5, v6); // gfcdbae
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v4, v6); // fgcdbae
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // gcfdbae
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gcdfbae
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // gcdbfae
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v6); // gcdbafe
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v4, v1)) {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // gfecbda
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fgecbda
																}
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // fegcbda
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v5); // gfcebda
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fgcebda
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v5); swap(v3, v5); // gfcbeda
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v5); // gfcbdea
																}
															}
															else {
																swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gfcbdae
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v5, v6); // fgcbeda
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fgcbdea
																}
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v6); // fgcbdae
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v1)) {
															if (compare(v4, v1)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // gcfebda
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); // gcfbeda
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // gcfbdea
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v2); swap(v3, v5); // gcbfeda
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v5); // gcbfdea
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v4, v5); // gcbdfea
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gcfbdae
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gcbfdae
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v2); swap(v4, v5); swap(v5, v6); // gcbdfae
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v4, v6); // gcbdafe
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v1)) {
												if (compare(v5, v1)) {
													if (compare(v4, v2)) {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); // gfebcda
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // fgebcda
																}
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fegbcda
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // gfbecda
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fgbecda
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); // gfbceda
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // gfbcdea
																}
															}
															else {
																swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gfbcdae
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // fgbceda
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fgbcdea
																}
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // fgbcdae
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // gbfecda
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v2, v5); swap(v3, v5); // gbfceda
																}
																else {
																	swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // gbfcdea
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v3, v5); // gbcfeda
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v3, v5); swap(v4, v5); // gbcfdea
																}
																else {
																	swap(v0, v6); swap(v4, v5); // gbcdfea
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gbfcdae
															}
															else {
																swap(v0, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gbcfdae
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v4, v5); swap(v5, v6); // gbcdfae
															}
															else {
																swap(v0, v6); swap(v4, v6); // gbcdafe
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v5, v1)) {
														if (compare(v6, v4)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fbgecda
															}
															else {
																swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // fbgceda
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // febgcda
															}
															else {
																swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // fbegcda
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bgfecda
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bfgecda
																}
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // bfegcda
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); // bgfceda
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // bfgceda
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v3, v5); // bgcfeda
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fbgcdea
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bgfcdea
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bfgcdea
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v5); // bgcfdea
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v4, v5); // bgcdfea
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // fbgcdae
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bgfcdae
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // bfgcdae
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bgcfdae
																}
																else {
																	swap(v0, v1); swap(v1, v6); swap(v4, v5); swap(v5, v6); // bgcdfae
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v4, v6); // bgcdafe
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v1)) {
											if (compare(v3, v1)) {
												if (compare(v4, v1)) {
													if (compare(v5, v2)) {
														if (compare(v6, v4)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // fcgedba
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v5, v6); // fcgdeba
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fecgdba
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fcegdba
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // cgfedba
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); // cfgedba
																}
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // cfegdba
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v5, v6); // cgfdeba
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v6); // cfgdeba
																}
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // cgdfeba
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // fcgdbea
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // cgfdbea
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v5); // cfgdbea
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cgdfbea
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // cgdbfea
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v6); // fcgdbae
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v6); // cgfdbae
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // cfgdbae
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // cgdfbae
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cgdbfae
																}
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // cgdbafe
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v5, v6); // fecgbda
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fcgebda
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // fcegbda
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cgfebda
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // cfgebda
																}
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cfegbda
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // cgfbeda
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); // cgbfeda
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // fcgbeda
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); // cfgbeda
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fcgbdea
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cgfbdea
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // cfgbdea
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // cgbfdea
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v5); // cgbdfea
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // fcgbdae
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // cgfbdae
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cfgbdae
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cgbfdae
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // cgbdfae
																}
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v6); // cgbdafe
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v5, v1)) {
													if (compare(v5, v2)) {
														if (compare(v6, v4)) {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v5, v6); // fcbgeda
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fcbgdea
																}
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v6); // fcbgdae
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fecbgda
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fcebgda
																}
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fcbegda
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // cfebgda
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // cfbgeda
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cfbegda
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cfbgdea
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // cfbgdae
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v2, v6); swap(v3, v5); // cbgfeda
																}
																else {
																	swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // cbfgeda
																}
															}
															else {
																swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cbfegda
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // cbgfdea
																}
																else {
																	swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cbfgdea
																}
															}
															else {
																swap(v0, v2); swap(v2, v6); swap(v4, v5); // cbgdfea
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v5)) {
																swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cbgfdae
															}
															else {
																swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // cbfgdae
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // cbgdfae
															}
															else {
																swap(v0, v2); swap(v2, v6); swap(v4, v6); // cbgdafe
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v1)) {
														if (compare(v6, v4)) {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v3, v6); swap(v5, v6); // fbcgeda
																}
																else {
																	swap(v0, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fbcgdea
																}
															}
															else {
																swap(v0, v5); swap(v3, v6); swap(v4, v6); // fbcgdae
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // febcgda
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fbecgda
																}
																else {
																	swap(v0, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fbcegda
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bfecgda
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v5, v6); // bfcgeda
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bfcegda
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bfcgdea
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v6); // bfcgdae
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); // bcgfeda
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // bcfgeda
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bcfegda
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // bcgfdea
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bcfgdea
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v5); // bcgdfea
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v5)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bcgfdae
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // bcfgdae
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // bcgdfae
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v6); // bcgdafe
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v4)) {
										if (compare(v3, v1)) {
											if (compare(v6, v1)) {
												if (compare(v4, v0)) {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // fcdgeba
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); // cfdgeba
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); // cdgfeba
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // cdfgeba
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fcdgbea
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // cfdgbea
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // cdgfbea
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cdfgbea
																}
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v5); // cdgbfea
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v0)) {
															if (compare(v5, v1)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cdgfbae
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // cdgbfae
															}
														}
														else {
															swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v6); // cdgbafe
														}
													}
													else {
														if (compare(v5, v2)) {
															swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // fcdgbae
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cfdgbae
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // cdfgbae
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // fcdbgea
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cfdbgea
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // cdfbgea
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v6); // fcdbgae
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // cfdbgae
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v6); // cdfbgae
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // fcdbage
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cfdbage
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // cdfbage
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // cdbgfea
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cdbgfae
															}
														}
														else {
															if (compare(v6, v0)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // cdbgafe
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v6); // cdbagfe
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cdbfgea
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // cdbfgae
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cdbfage
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // cdbafge
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v5, v3)) {
													if (compare(v4, v0)) {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v4, v6); swap(v5, v6); // fcbdgea
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // cfbdgea
															}
														}
														else {
															swap(v0, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // cbfdgea
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	swap(v0, v5); swap(v1, v2); swap(v4, v6); // fcbdgae
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v6); // cfbdgae
																}
															}
															else {
																swap(v0, v2); swap(v2, v5); swap(v4, v6); // cbfdgae
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	swap(v0, v5); swap(v1, v2); swap(v4, v5); swap(v5, v6); // fcbdage
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // cfbdage
																}
															}
															else {
																swap(v0, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // cbfdage
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // cbdgfea
															}
															else {
																swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cbdgfae
															}
														}
														else {
															if (compare(v6, v0)) {
																swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // cbdgafe
															}
															else {
																swap(v0, v2); swap(v2, v3); swap(v4, v6); // cbdagfe
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cbdfgea
															}
															else {
																swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // cbdfgae
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cbdfage
															}
															else {
																swap(v0, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // cbdafge
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v4, v0)) {
														if (compare(v5, v1)) {
															swap(v0, v5); swap(v4, v6); swap(v5, v6); // fbcdgea
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v1, v5); swap(v4, v6); swap(v5, v6); // bfcdgea
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // bcfdgea
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v4, v6); // fbcdgae
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v1); swap(v1, v5); swap(v4, v6); // bfcdgae
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v6); // bcfdgae
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v4, v5); swap(v5, v6); // fbcdage
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v1); swap(v1, v5); swap(v4, v5); swap(v5, v6); // bfcdage
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // bcfdage
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // bcdgfea
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bcdgfae
															}
														}
														else {
															if (compare(v6, v0)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // bcdgafe
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v6); // bcdagfe
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bcdfgea
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // bcdfgae
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bcdfage
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // bcdafge
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v5, v2)) {
											if (compare(v6, v0)) {
												if (compare(v4, v1)) {
													if (compare(v4, v2)) {
														if (compare(v3, v1)) {
															if (compare(v6, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v4, v6); swap(v5, v6); // fecdgba
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v5, v6); // fecdbga
															}
														}
														else {
															if (compare(v2, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v5, v6); // fecbdga
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // febcdga
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // fcedgba
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fcdegba
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v5, v6); // fcedbga
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // fcdebga
																}
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // fcebdga
															}
														}
													}
												}
												else {
													if (compare(v4, v3)) {
														if (compare(v5, v1)) {
															if (compare(v2, v1)) {
																swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v5, v6); // fcbedga
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // fbecdga
																}
																else {
																	swap(v0, v5); swap(v3, v4); swap(v5, v6); // fbcedga
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // bfecdga
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v5, v6); // bfcedga
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v3, v1)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v5, v6); // fcdbega
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v5, v6); // fcbdega
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v5, v6); // fbcdega
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v5, v6); // bfcdega
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v3, v1)) {
														if (compare(v4, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v4); // fecdbag
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v4); // fcedbag
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); // fcdebag
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v3); // fcdbeag
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v5); // fcdbaeg
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v3, v4); // fecbdag
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); // fcebdag
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v2); swap(v3, v4); // fcbedag
																}
																else {
																	swap(v0, v5); swap(v1, v2); // fcbdeag
																}
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v4, v5); // fcbdaeg
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); // febcdag
															}
															else {
																swap(v0, v5); swap(v2, v4); swap(v3, v4); // fbecdag
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v3, v4); // fbcedag
																}
																else {
																	swap(v0, v5); // fbcdeag
																}
															}
															else {
																swap(v0, v5); swap(v4, v5); // fbcdaeg
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); // bfecdag
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v3, v4); // bfcedag
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v5); // bfcdeag
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v4, v5); // bfcdaeg
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v1)) {
												if (compare(v6, v0)) {
													if (compare(v4, v1)) {
														if (compare(v6, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v6); // cfedgba
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // cfdegba
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cdfegba
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v3, v1)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // cfedbga
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cfebdga
																}
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cfdebga
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // cdfebga
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // cfdbega
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v5, v6); // cdfbega
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // cfbedga
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v5, v6); // cfbdega
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v4, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v5); // cfedbag
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cfdebag
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); // cdfebag
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); // cfdbeag
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v5); // cdfbeag
																}
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cfdbaeg
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v5); // cdfbaeg
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // cfebdag
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); // cfbedag
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v5); // cfbdeag
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v5); // cfbdaeg
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v2, v1)) {
														if (compare(v6, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // cbfedga
															}
															else {
																swap(v0, v2); swap(v2, v5); swap(v5, v6); // cbfdega
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v2); swap(v2, v5); swap(v3, v4); // cbfedag
																}
																else {
																	swap(v0, v2); swap(v2, v5); // cbfdeag
																}
															}
															else {
																swap(v0, v2); swap(v2, v5); swap(v4, v5); // cbfdaeg
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // bcfedga
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v5, v6); // bcfdega
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); // bcfedag
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v5); // bcfdeag
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v5); // bcfdaeg
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v0)) {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // cdbfega
																}
																else {
																	swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // cbdfega
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // bcdfega
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); // cdbfeag
																}
																else {
																	swap(v0, v2); swap(v2, v3); swap(v3, v5); // cbdfeag
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); // bcdfeag
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cdbfaeg
																}
																else {
																	swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cbdfaeg
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bcdfaeg
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v5); // cdbafeg
																}
																else {
																	swap(v0, v2); swap(v2, v3); swap(v4, v5); // cbdafeg
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v5); // bcdafeg
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v6, v3)) {
									if (compare(v6, v2)) {
										if (compare(v2, v1)) {
											if (compare(v3, v1)) {
												if (compare(v5, v1)) {
													if (compare(v5, v3)) {
														if (compare(v5, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // gefcdba
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // egfcdba
																}
																else {
																	swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // efgcdba
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v5); // gecfdba
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // egcfdba
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // gcefdba
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v4, v5); // gecdfba
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v4, v5); swap(v5, v6); // egcdfba
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v5); // gcedfba
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // gcdefba
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); // gecdbfa
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v4, v6); // egcdbfa
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v4); // gcedbfa
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); // gcdebfa
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v3); // gcdbefa
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v5, v6); // gecdbaf
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v4, v6); swap(v5, v6); // egcdbaf
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v5, v6); // gcedbaf
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v1)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // gcdebaf
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v5, v6); // gcdbeaf
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // gcdbaef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v5, v1)) {
														if (compare(v5, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); // gefcbda
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // egfcbda
																}
																else {
																	swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // efgcbda
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v3, v5); // gecfbda
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v6); // egcfbda
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v5); // gecbfda
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v3, v4); // gecbdfa
																}
															}
															else {
																swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v5, v6); // gecbdaf
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // egcbfda
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v6); // egcbdfa
																}
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // egcbdaf
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v1)) {
															if (compare(v5, v1)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); // gcefbda
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // gcebfda
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); // gcebdfa
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v5); // gcbefda
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v3, v4); // gcbedfa
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); // gcbdefa
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // gcebdaf
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v5, v6); // gcbedaf
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v2); swap(v5, v6); // gcbdeaf
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v4, v6); swap(v5, v6); // gcbdaef
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v1)) {
												if (compare(v4, v1)) {
													if (compare(v5, v2)) {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // gefbcda
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // egfbcda
																}
																else {
																	swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // efgbcda
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); // gebfcda
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // egbfcda
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // gebcfda
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); // gebcdfa
																}
															}
															else {
																swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // gebcdaf
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // egbcfda
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // egbcdfa
																}
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // egbcdaf
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v2, v4); swap(v3, v5); // gbefcda
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // gbecfda
																}
																else {
																	swap(v0, v6); swap(v2, v4); swap(v3, v4); // gbecdfa
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v3, v4); swap(v4, v5); // gbcefda
																}
																else {
																	swap(v0, v6); swap(v3, v4); // gbcedfa
																}
															}
															else {
																swap(v0, v6); // gbcdefa
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // gbecdaf
															}
															else {
																swap(v0, v6); swap(v3, v4); swap(v5, v6); // gbcedaf
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v5, v6); // gbcdeaf
															}
															else {
																swap(v0, v6); swap(v4, v6); swap(v5, v6); // gbcdaef
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v4, v1)) {
														if (compare(v6, v5)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // ebgfcda
															}
															else {
																swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ebgcfda
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // efbgcda
															}
															else {
																swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ebfgcda
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); // bgefcda
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // begfcda
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // befgcda
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // bgecfda
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // begcfda
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v5); // bgcefda
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // ebgcdfa
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); // bgecdfa
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // begcdfa
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v6); swap(v3, v4); // bgcedfa
															}
															else {
																swap(v0, v1); swap(v1, v6); // bgcdefa
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ebgcdaf
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // bgecdaf
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // begcdaf
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v5, v6); // bgcedaf
																}
																else {
																	swap(v0, v1); swap(v1, v6); swap(v5, v6); // bgcdeaf
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v4, v6); swap(v5, v6); // bgcdaef
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v1)) {
											if (compare(v3, v1)) {
												if (compare(v5, v1)) {
													if (compare(v4, v2)) {
														if (compare(v6, v5)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ecgfdba
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // ecgdfba
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v6); // efcgdba
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // ecfgdba
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cgefdba
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // cegfdba
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cefgdba
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // cgedfba
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v5); // cegdfba
																}
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cgdefba
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v6); // ecgdbfa
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v6); // cgedbfa
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v6); // cegdbfa
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // cgdebfa
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); // cgdbefa
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // ecgdbaf
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // cgedbaf
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v5, v6); // cegdbaf
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v1)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cgdebaf
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // cgdbeaf
																}
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cgdbaef
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // efcgbda
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // ecgfbda
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ecfgbda
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // cgefbda
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); // cegfbda
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // cefgbda
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cgebfda
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // cgbefda
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ecgbfda
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // cegbfda
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // ecgbdfa
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // cgebdfa
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); // cegbdfa
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); // cgbedfa
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); // cgbdefa
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ecgbdaf
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cgebdaf
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // cegbdaf
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // cgbedaf
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v5, v6); // cgbdeaf
																}
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // cgbdaef
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v4, v1)) {
													if (compare(v4, v2)) {
														if (compare(v6, v5)) {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ecbgfda
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v6); // ecbgdfa
																}
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ecbgdaf
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v6); // efcbgda
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // ecfbgda
																}
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v6); // ecbfgda
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cefbgda
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cebgfda
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // cebfgda
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // cebgdfa
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cebgdaf
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // cbgefda
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cbegfda
																}
																else {
																	swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // cbefgda
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v2, v6); swap(v3, v4); // cbgedfa
																}
																else {
																	swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // cbegdfa
																}
															}
															else {
																swap(v0, v2); swap(v2, v6); // cbgdefa
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // cbgedaf
															}
															else {
																swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cbegdaf
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v2, v6); swap(v5, v6); // cbgdeaf
															}
															else {
																swap(v0, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // cbgdaef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v1)) {
														if (compare(v6, v5)) {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ebcgfda
																}
																else {
																	swap(v0, v4); swap(v3, v6); swap(v4, v6); // ebcgdfa
																}
															}
															else {
																swap(v0, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ebcgdaf
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // efbcgda
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // ebfcgda
																}
																else {
																	swap(v0, v4); swap(v3, v5); swap(v4, v6); // ebcfgda
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // befcgda
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // becgfda
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v6); // becfgda
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v6); // becgdfa
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // becgdaf
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // bcgefda
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bcegfda
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // bcefgda
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); // bcgedfa
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // bcegdfa
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v6); // bcgdefa
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // bcgedaf
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bcegdaf
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v5, v6); // bcgdeaf
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // bcgdaef
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v5)) {
										if (compare(v3, v1)) {
											if (compare(v6, v1)) {
												if (compare(v5, v0)) {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ecdgfba
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // cedgfba
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // cdgefba
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cdegfba
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // ecdgbfa
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); // cedgbfa
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); // cdgebfa
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // cdegbfa
																}
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v6); // cdgbefa
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v4, v0)) {
															if (compare(v4, v1)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // cdgebaf
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v5, v6); // cdgbeaf
															}
														}
														else {
															swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // cdgbaef
														}
													}
													else {
														if (compare(v4, v2)) {
															swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ecdgbaf
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // cedgbaf
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cdegbaf
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v6); // ecdbgfa
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // cedbgfa
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v6); // cdebgfa
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // ecdbgaf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cedbgaf
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // cdebgaf
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v5, v6); // ecdbagf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // cedbagf
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v5, v6); // cdebagf
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v4, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); // cdbgefa
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // cdbgeaf
															}
														}
														else {
															if (compare(v6, v0)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cdbgaef
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // cdbagef
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // cdbegfa
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cdbegaf
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // cdbeagf
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v5, v6); // cdbaegf
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v4, v3)) {
													if (compare(v5, v0)) {
														if (compare(v4, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v4, v6); // ecbdgfa
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v6); // cebdgfa
															}
														}
														else {
															swap(v0, v2); swap(v2, v4); swap(v4, v6); // cbedgfa
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	swap(v0, v4); swap(v1, v2); swap(v4, v6); swap(v5, v6); // ecbdgaf
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // cebdgaf
																}
															}
															else {
																swap(v0, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // cbedgaf
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	swap(v0, v4); swap(v1, v2); swap(v5, v6); // ecbdagf
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v5, v6); // cebdagf
																}
															}
															else {
																swap(v0, v2); swap(v2, v4); swap(v5, v6); // cbedagf
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v4, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v2, v3); swap(v3, v6); // cbdgefa
															}
															else {
																swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // cbdgeaf
															}
														}
														else {
															if (compare(v6, v0)) {
																swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cbdgaef
															}
															else {
																swap(v0, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // cbdagef
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // cbdegfa
															}
															else {
																swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cbdegaf
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // cbdeagf
															}
															else {
																swap(v0, v2); swap(v2, v3); swap(v5, v6); // cbdaegf
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v5, v0)) {
														if (compare(v4, v1)) {
															swap(v0, v4); swap(v4, v6); // ebcdgfa
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v1, v4); swap(v4, v6); // becdgfa
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v6); // bcedgfa
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v4, v6); swap(v5, v6); // ebcdgaf
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v1); swap(v1, v4); swap(v4, v6); swap(v5, v6); // becdgaf
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // bcedgaf
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v5, v6); // ebcdagf
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v1); swap(v1, v4); swap(v5, v6); // becdagf
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v5, v6); // bcedagf
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v4, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); // bcdgefa
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // bcdgeaf
															}
														}
														else {
															if (compare(v6, v0)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bcdgaef
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // bcdagef
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // bcdegfa
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bcdegaf
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // bcdeagf
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v5, v6); // bcdaegf
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v2)) {
											if (compare(v6, v0)) {
												if (compare(v5, v1)) {
													if (compare(v5, v2)) {
														if (compare(v3, v1)) {
															if (compare(v6, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v4, v6); // efcdgba
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v4, v5); swap(v5, v6); // efcdbga
															}
														}
														else {
															if (compare(v2, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // efcbdga
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // efbcdga
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v6); // ecfdgba
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // ecdfgba
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // ecfdbga
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ecdfbga
																}
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ecfbdga
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v4, v1)) {
															if (compare(v2, v1)) {
																swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ecbfdga
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ebfcdga
																}
																else {
																	swap(v0, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ebcfdga
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // befcdga
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // becfdga
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v3, v1)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // ecdbfga
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v4, v5); swap(v5, v6); // ecbdfga
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v4, v5); swap(v5, v6); // ebcdfga
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v4, v5); swap(v5, v6); // becdfga
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v3, v1)) {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v5); swap(v4, v5); // efcdbag
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v5); // ecfdbag
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // ecdfbag
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v5); // ecdbfag
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v3); // ecdbafg
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v5); // efcbdag
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // ecfbdag
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v5); // ecbfdag
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v4, v5); // ecbdfag
																}
															}
															else {
																swap(v0, v4); swap(v1, v2); // ecbdafg
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // efbcdag
															}
															else {
																swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // ebfcdag
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v3, v5); swap(v4, v5); // ebcfdag
																}
																else {
																	swap(v0, v4); swap(v4, v5); // ebcdfag
																}
															}
															else {
																swap(v0, v4); // ebcdafg
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // befcdag
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v5); // becfdag
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v4); swap(v4, v5); // becdfag
															}
															else {
																swap(v0, v1); swap(v1, v4); // becdafg
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v1)) {
												if (compare(v6, v0)) {
													if (compare(v5, v1)) {
														if (compare(v6, v1)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // cefdgba
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cedfgba
																}
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // cdefgba
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v5, v6); // cefdbga
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // cefbdga
																}
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // cedfbga
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cdefbga
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cedbfga
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // cdebfga
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cebfdga
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // cebdfga
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v5, v1)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v5); // cefdbag
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); // cedfbag
																}
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // cdefbag
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cedbfag
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v5); // cdebfag
																}
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); // cedbafg
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v4); // cdebafg
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); // cefbdag
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // cebfdag
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v5); // cebdfag
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v4); // cebdafg
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v2, v1)) {
														if (compare(v6, v0)) {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cbefdga
															}
															else {
																swap(v0, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // cbedfga
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // cbefdag
																}
																else {
																	swap(v0, v2); swap(v2, v4); swap(v4, v5); // cbedfag
																}
															}
															else {
																swap(v0, v2); swap(v2, v4); // cbedafg
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bcefdga
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // bcedfga
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // bcefdag
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v5); // bcedfag
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v4); // bcedafg
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v6, v0)) {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cdbefga
																}
																else {
																	swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cbdefga
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bcdefga
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cdbefag
																}
																else {
																	swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cbdefag
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bcdefag
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); // cdbeafg
																}
																else {
																	swap(v0, v2); swap(v2, v3); swap(v3, v4); // cbdeafg
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); // bcdeafg
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v0, v2); swap(v1, v3); swap(v2, v3); // cdbaefg
																}
																else {
																	swap(v0, v2); swap(v2, v3); // cbdaefg
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v3); // bcdaefg
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
						else {
							if (compare(v5, v4)) {
								if (compare(v6, v0)) {
									if (compare(v6, v1)) {
										if (compare(v2, v1)) {
											if (compare(v6, v2)) {
												if (compare(v4, v0)) {
													if (compare(v4, v1)) {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gfecbad
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // fgecbad
																}
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); // fegcbad
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gfcebad
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v6); // fgcebad
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gcfebad
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v5)) {
																swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v5, v6); // gfcbead
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v3, v6); // fgcbead
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // gcfbead
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v5, v6); // gcbfead
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gfcbaed
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v5); // fgcbaed
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gcfbaed
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v6); // gfcbade
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fgcbade
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // gcfbade
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gcbfaed
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v6); // gcbfade
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v5); // gcbafed
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gcbafde
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v6); // gcbadfe
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v3, v6); // fecgbad
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // fcgebad
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); // fcegbad
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // cgfebad
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cfgebad
																}
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // cfegbad
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); // cgfbead
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // cgbfead
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); // fcgbead
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // cfgbead
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // fcgbaed
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // cgfbaed
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cfgbaed
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fcgbade
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cgfbade
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // cfgbade
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cgbfaed
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // cgbfade
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // cgbafed
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cgbafde
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // cgbadfe
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v0)) {
												if (compare(v4, v0)) {
													if (compare(v4, v2)) {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // gfebcad
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); // fgebcad
																}
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // fegbcad
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gfbecad
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // fgbecad
																}
															}
															else {
																swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gbfecad
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v5)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // gfbcead
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); // fgbcead
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // gbfcead
															}
															else {
																swap(v0, v6); swap(v3, v5); swap(v5, v6); // gbcfead
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v3)) {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gfbcaed
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // fgbcaed
																}
															}
															else {
																swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gbfcaed
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // gfbcade
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fgbcade
																}
															}
															else {
																swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // gbfcade
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gbcfaed
															}
															else {
																swap(v0, v6); swap(v3, v5); swap(v4, v6); // gbcfade
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v3, v6); swap(v4, v5); // gbcafed
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gbcafde
																}
																else {
																	swap(v0, v6); swap(v3, v6); swap(v4, v6); // gbcadfe
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v4, v0)) {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gfebacd
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // fgebacd
																}
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fegbacd
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // gfbeacd
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fgbeacd
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); // gfbaecd
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // gfbaced
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gfbacde
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // fgbaecd
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fgbaced
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // fgbacde
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // gbfeacd
															}
															else {
																swap(v0, v6); swap(v2, v5); swap(v3, v6); // gbfaecd
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // gbfaced
															}
															else {
																swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gbfacde
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // gbafecd
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gbafced
																}
																else {
																	swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // gbafcde
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // gbacfed
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gbacfde
																}
																else {
																	swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // gbacdfe
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v2)) {
											if (compare(v2, v0)) {
												if (compare(v4, v0)) {
													if (compare(v5, v1)) {
														if (compare(v6, v4)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // fbgecad
															}
															else {
																swap(v0, v5); swap(v2, v6); swap(v3, v6); // fbgcead
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); // febgcad
															}
															else {
																swap(v0, v5); swap(v2, v4); swap(v3, v6); // fbegcad
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bgfecad
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // bfgecad
																}
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); // bfegcad
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // bgfcead
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); // bfgcead
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v5, v6); // bgcfead
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // fbgcaed
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bgfcaed
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // bfgcaed
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fbgcade
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // bgfcade
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bfgcade
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bgcfaed
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v6); // bgcfade
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v5); // bgcafed
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bgcafde
																}
																else {
																	swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v6); // bgcadfe
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v0)) {
														if (compare(v5, v1)) {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // febgacd
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fbgeacd
																}
																else {
																	swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // fbegacd
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // bgfeacd
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bfgeacd
																}
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // bfegacd
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); // bgfaecd
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // bgafecd
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // fbgaecd
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // bfgaecd
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fbgaced
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // bgfaced
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bfgaced
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // fbgacde
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bgfacde
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // bfgacde
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bgafced
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // bgafcde
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // bgacfed
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bgacfde
																}
																else {
																	swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // bgacdfe
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v4, v0)) {
													if (compare(v5, v2)) {
														if (compare(v4, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v6); // fecbgad
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // fcebgad
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v5); swap(v1, v2); swap(v3, v6); // fcbgead
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v6); // fcbegad
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v4, v1)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cfebgad
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); // cfbgead
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // cfbegad
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // cbgfead
																}
																else {
																	swap(v0, v2); swap(v2, v5); swap(v3, v6); // cbfgead
																}
															}
															else {
																swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // cbfegad
															}
														}
													}
												}
												else {
													if (compare(v4, v3)) {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v5); // fcbgaed
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // cfbgaed
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cbgfaed
																}
																else {
																	swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // cbfgaed
																}
															}
															else {
																swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // cbgafed
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // cbgfade
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cbgafde
																}
																else {
																	swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // cbgadfe
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fcbgade
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cfbgade
																}
															}
															else {
																swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cbfgade
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // febcgad
															}
															else {
																swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // fbecgad
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v5); swap(v3, v6); // fbcgead
															}
															else {
																swap(v0, v5); swap(v3, v4); swap(v4, v6); // fbcegad
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // bfecgad
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v5); swap(v3, v6); // bfcgead
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v6); // bfcegad
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // bcgfead
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); // bcfgead
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // bcfegad
															}
														}
													}
												}
												else {
													if (compare(v4, v3)) {
														if (compare(v5, v2)) {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v3, v6); swap(v4, v5); // fbcgaed
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v5); // bfcgaed
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bcgfaed
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // bcfgaed
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // bcgafed
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v0)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // bcgfade
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bcgafde
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // bcgadfe
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fbcgade
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bfcgade
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bcfgade
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v4)) {
										if (compare(v2, v0)) {
											if (compare(v2, v1)) {
												if (compare(v5, v0)) {
													if (compare(v5, v1)) {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fcbaged
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v6); // fcbagde
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fcbadge
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cfbaged
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // cfbagde
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cfbadge
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cbfaged
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // cbfagde
															}
															else {
																swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cbfadge
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v3, v6); swap(v4, v5); // cbagfed
															}
															else {
																swap(v0, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cbagfde
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v2); swap(v3, v6); swap(v4, v6); // cbagdfe
															}
															else {
																swap(v0, v2); swap(v4, v6); // cbadgfe
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cbafged
															}
															else {
																swap(v0, v2); swap(v3, v5); swap(v4, v6); // cbafgde
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cbafdge
															}
															else {
																swap(v0, v2); swap(v4, v5); swap(v5, v6); // cbadfge
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															swap(v0, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fbcaged
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v5); swap(v3, v5); swap(v4, v6); // fbcagde
															}
															else {
																swap(v0, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fbcadge
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bfcaged
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v6); // bfcagde
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bfcadge
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bcfaged
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // bcfagde
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bcfadge
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v5); // bcagfed
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bcagfde
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v6); // bcagdfe
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v4, v6); // bcadgfe
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bcafged
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v6); // bcafgde
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bcafdge
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v4, v5); swap(v5, v6); // bcadfge
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v2)) {
												if (compare(v5, v0)) {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															swap(v0, v5); swap(v2, v5); swap(v3, v6); // fbagecd
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // fbagced
															}
															else {
																swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fbagcde
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); // bfagecd
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // bfagced
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bfagcde
															}
														}
													}
												}
												else {
													if (compare(v4, v3)) {
														if (compare(v4, v2)) {
															if (compare(v6, v5)) {
																swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v5, v6); // bagfecd
															}
															else {
																swap(v0, v1); swap(v2, v5); swap(v3, v6); // bafgecd
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bagfced
																}
																else {
																	swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v5); // bafgced
																}
															}
															else {
																swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v5); // bagcfed
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v5)) {
																swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v6); // bagfcde
															}
															else {
																swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bafgcde
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bagcfde
															}
															else {
																swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v6); // bagcdfe
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v0)) {
														if (compare(v5, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fbacged
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // fbacgde
																}
																else {
																	swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fbacdge
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bfacged
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // bfacgde
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bfacdge
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bafcged
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v6); // bafcgde
															}
															else {
																swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bafcdge
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v3, v6); swap(v4, v5); // bacgfed
															}
															else {
																swap(v0, v1); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bacgfde
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v1); swap(v3, v6); swap(v4, v6); // bacgdfe
															}
															else {
																swap(v0, v1); swap(v4, v6); // bacdgfe
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bacfged
															}
															else {
																swap(v0, v1); swap(v3, v5); swap(v4, v6); // bacfgde
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bacfdge
															}
															else {
																swap(v0, v1); swap(v4, v5); swap(v5, v6); // bacdfge
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v5, v1)) {
											if (compare(v6, v3)) {
												if (compare(v4, v2)) {
													if (compare(v4, v1)) {
														if (compare(v2, v0)) {
															if (compare(v2, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fecbagd
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // febcagd
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // febagcd
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // febacgd
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v2, v0)) {
																swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fbecagd
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fbeagcd
																}
																else {
																	swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // fbeacgd
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // fbaegcd
															}
															else {
																swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fbaecgd
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v4, v1)) {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fcebagd
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // cfebagd
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fcbeagd
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cfbeagd
																}
															}
															else {
																swap(v0, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fbceagd
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v5, v6); // fcbaegd
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // cfbaegd
															}
														}
														else {
															if (compare(v2, v0)) {
																swap(v0, v5); swap(v3, v5); swap(v5, v6); // fbcaegd
															}
															else {
																swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // fbacegd
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v2, v1)) {
														if (compare(v4, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v5); // fecbadg
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // fcebadg
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); // cfebadg
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v5); // fcbeadg
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cfbeadg
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // febcadg
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // fbecadg
																}
																else {
																	swap(v0, v5); swap(v3, v4); swap(v4, v5); // fbceadg
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); // febacdg
															}
															else {
																swap(v0, v5); swap(v2, v4); swap(v3, v5); // fbeacdg
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v2); swap(v3, v5); // fcbaedg
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v5); // fcbadeg
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); // cfbaedg
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cfbadeg
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v3, v5); // fbcaedg
															}
															else {
																swap(v0, v5); swap(v3, v5); swap(v4, v5); // fbcadeg
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fbaecdg
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v2, v5); swap(v3, v5); // fbacedg
																}
																else {
																	swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fbacdeg
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v4, v2)) {
													if (compare(v4, v0)) {
														if (compare(v2, v0)) {
															if (compare(v6, v3)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bfecagd
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // bfecadg
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bfeagcd
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // bfeacgd
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); // bfeacdg
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v6, v2)) {
																swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // bfaegcd
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bfaecgd
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bfaecdg
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v6); // bafegcd
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bafecgd
																}
																else {
																	swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bafecdg
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v0)) {
														if (compare(v4, v0)) {
															if (compare(v6, v3)) {
																swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bfceagd
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v5); // bfceadg
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v5, v6); // bfcaegd
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v3, v5); // bfcaedg
																}
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v5); // bfcadeg
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // bfacegd
																}
																else {
																	swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); // bfacedg
																}
															}
															else {
																swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bfacdeg
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v5, v6); // bafcegd
																}
																else {
																	swap(v0, v1); swap(v2, v5); swap(v3, v5); // bafcedg
																}
															}
															else {
																swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bafcdeg
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v2, v1)) {
														if (compare(v4, v0)) {
															if (compare(v6, v3)) {
																swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cbfeagd
															}
															else {
																swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cbfeadg
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // cbfaegd
																}
																else {
																	swap(v0, v2); swap(v2, v5); swap(v3, v5); // cbfaedg
																}
															}
															else {
																swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cbfadeg
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v6, v3)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bcfeagd
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bcfeadg
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // bcfaegd
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); // bcfaedg
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bcfadeg
															}
														}
													}
												}
												else {
													if (compare(v4, v3)) {
														if (compare(v6, v3)) {
															if (compare(v2, v0)) {
																if (compare(v2, v1)) {
																	swap(v0, v2); swap(v3, v5); swap(v5, v6); // cbafegd
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v5, v6); // bcafegd
																}
															}
															else {
																swap(v0, v1); swap(v3, v5); swap(v5, v6); // bacfegd
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v2, v1)) {
																	swap(v0, v2); swap(v3, v5); // cbafedg
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v3, v5); // bcafedg
																}
															}
															else {
																swap(v0, v1); swap(v3, v5); // bacfedg
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v2, v0)) {
																if (compare(v2, v1)) {
																	swap(v0, v2); swap(v3, v5); swap(v4, v5); // cbafdeg
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v5); // bcafdeg
																}
															}
															else {
																swap(v0, v1); swap(v3, v5); swap(v4, v5); // bacfdeg
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v2, v1)) {
																	swap(v0, v2); swap(v4, v5); // cbadfeg
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v4, v5); // bcadfeg
																}
															}
															else {
																swap(v0, v1); swap(v4, v5); // bacdfeg
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v6, v0)) {
									if (compare(v6, v1)) {
										if (compare(v2, v1)) {
											if (compare(v6, v2)) {
												if (compare(v5, v0)) {
													if (compare(v5, v1)) {
														if (compare(v5, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // gefcbad
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // egfcbad
																}
																else {
																	swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // efgcbad
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v5, v6); // gecfbad
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // egcfbad
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // gcefbad
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gecbfad
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v5); // egcbfad
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gcebfad
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gcbefad
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v6); // gecbafd
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v3, v6); // egcbafd
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // gcebafd
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gecbadf
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v5, v6); // egcbadf
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gcebadf
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v6); // gcbeafd
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gcbeadf
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v3, v6); // gcbaefd
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v5, v6); // gcbaedf
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gcbadef
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v5); // efcgbad
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ecgfbad
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // ecfgbad
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cgefbad
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // cegfbad
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); // cefgbad
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // cgebfad
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cgbefad
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // ecgbfad
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cegbfad
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); // ecgbafd
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); // cgebafd
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // cegbafd
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // ecgbadf
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // cgebadf
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cegbadf
																}
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // cgbeafd
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cgbeadf
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); // cgbaefd
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // cgbaedf
																}
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cgbadef
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v0)) {
												if (compare(v5, v0)) {
													if (compare(v5, v2)) {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gefbcad
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // egfbcad
																}
																else {
																	swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // efgbcad
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // gebfcad
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // egbfcad
																}
															}
															else {
																swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // gbefcad
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gebcfad
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // egbcfad
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gbecfad
															}
															else {
																swap(v0, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gbcefad
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v5, v3)) {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // gebcafd
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); // egbcafd
																}
															}
															else {
																swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // gbecafd
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gebcadf
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // egbcadf
																}
															}
															else {
																swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gbecadf
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v3, v4); swap(v4, v6); // gbceafd
															}
															else {
																swap(v0, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gbceadf
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v3, v6); // gbcaefd
																}
																else {
																	swap(v0, v6); swap(v3, v6); swap(v5, v6); // gbcaedf
																}
															}
															else {
																swap(v0, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gbcadef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v5, v0)) {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // gefbacd
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); // egfbacd
																}
																else {
																	swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // efgbacd
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gebfacd
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // egbfacd
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // gebafcd
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); // gebacfd
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // gebacdf
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // egbafcd
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // egbacfd
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // egbacdf
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gbefacd
															}
															else {
																swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // gbeafcd
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v2, v4); swap(v3, v6); // gbeacfd
															}
															else {
																swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // gbeacdf
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gbaefcd
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // gbaecfd
																}
																else {
																	swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gbaecdf
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v2, v6); swap(v3, v6); // gbacefd
																}
																else {
																	swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // gbacedf
																}
															}
															else {
																swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gbacdef
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v2)) {
											if (compare(v2, v0)) {
												if (compare(v5, v0)) {
													if (compare(v4, v1)) {
														if (compare(v6, v5)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ebgfcad
															}
															else {
																swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // ebgcfad
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // efbgcad
															}
															else {
																swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // ebfgcad
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // bgefcad
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // begfcad
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // befgcad
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bgecfad
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // begcfad
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bgcefad
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v5, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v6); swap(v3, v6); // ebgcafd
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // bgecafd
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); // begcafd
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // ebgcadf
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bgecadf
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // begcadf
																}
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v6); // bgceafd
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bgceadf
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v1); swap(v1, v6); swap(v3, v6); // bgcaefd
																}
																else {
																	swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v5, v6); // bgcaedf
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bgcadef
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v0)) {
														if (compare(v4, v1)) {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); // efbgacd
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // ebgfacd
																}
																else {
																	swap(v0, v4); swap(v2, v5); swap(v3, v6); // ebfgacd
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bgefacd
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // begfacd
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); // befgacd
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // bgeafcd
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bgaefcd
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ebgafcd
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // begafcd
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // ebgacfd
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); // bgeacfd
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // begacfd
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ebgacdf
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // bgeacdf
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // begacdf
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // bgaecfd
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bgaecdf
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); // bgacefd
																}
																else {
																	swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // bgacedf
																}
															}
															else {
																swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bgacdef
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v5, v0)) {
													if (compare(v4, v2)) {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // efcbgad
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ecfbgad
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v5); // ecbgfad
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ecbfgad
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v5, v1)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // cefbgad
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // cebgfad
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cebfgad
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cbgefad
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // cbegfad
																}
																else {
																	swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cbefgad
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v4, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v3, v6); // ecbgafd
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); // cebgafd
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // cbgeafd
																}
																else {
																	swap(v0, v2); swap(v2, v4); swap(v3, v6); // cbegafd
																}
															}
															else {
																swap(v0, v2); swap(v2, v6); swap(v3, v6); // cbgaefd
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cbgeadf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // cbgaedf
																}
																else {
																	swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cbgadef
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v5, v6); // ecbgadf
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // cebgadf
																}
															}
															else {
																swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // cbegadf
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // efbcgad
															}
															else {
																swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ebfcgad
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v4); swap(v3, v6); swap(v4, v5); // ebcgfad
															}
															else {
																swap(v0, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ebcfgad
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // befcgad
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v5); // becgfad
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // becfgad
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bcgefad
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // bcegfad
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bcefgad
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v4, v2)) {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v3, v6); // ebcgafd
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v3, v6); // becgafd
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // bcgeafd
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); // bcegafd
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); // bcgaefd
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v0)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bcgeadf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // bcgaedf
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bcgadef
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v4); swap(v3, v6); swap(v5, v6); // ebcgadf
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v5, v6); // becgadf
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // bcegadf
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v5)) {
										if (compare(v2, v0)) {
											if (compare(v2, v1)) {
												if (compare(v4, v0)) {
													if (compare(v4, v1)) {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v6); // ecbagfd
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ecbagdf
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v5, v6); // ecbadgf
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // cebagfd
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cebagdf
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // cebadgf
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // cbeagfd
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cbeagdf
															}
															else {
																swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // cbeadgf
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v3, v6); // cbagefd
															}
															else {
																swap(v0, v2); swap(v3, v6); swap(v5, v6); // cbagedf
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cbagdef
															}
															else {
																swap(v0, v2); swap(v4, v6); swap(v5, v6); // cbadgef
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v3, v4); swap(v4, v6); // cbaegfd
															}
															else {
																swap(v0, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cbaegdf
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v3, v4); swap(v5, v6); // cbaedgf
															}
															else {
																swap(v0, v2); swap(v5, v6); // cbadegf
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															swap(v0, v4); swap(v3, v4); swap(v4, v6); // ebcagfd
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ebcagdf
															}
															else {
																swap(v0, v4); swap(v3, v4); swap(v5, v6); // ebcadgf
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v6); // becagfd
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // becagdf
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v5, v6); // becadgf
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // bceagfd
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bceagdf
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // bceadgf
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v2); swap(v3, v6); // bcagefd
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v5, v6); // bcagedf
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bcagdef
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v4, v6); swap(v5, v6); // bcadgef
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v6); // bcaegfd
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bcaegdf
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v5, v6); // bcaedgf
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v5, v6); // bcadegf
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v2)) {
												if (compare(v4, v0)) {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // ebagfcd
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v2, v4); swap(v3, v6); // ebagcfd
															}
															else {
																swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // ebagcdf
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // beagfcd
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); // beagcfd
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // beagcdf
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v5, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bagefcd
															}
															else {
																swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v5); // baegfcd
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v6); // bagecfd
																}
																else {
																	swap(v0, v1); swap(v2, v4); swap(v3, v6); // baegcfd
																}
															}
															else {
																swap(v0, v1); swap(v2, v6); swap(v3, v6); // bagcefd
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bagecdf
															}
															else {
																swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v5, v6); // baegcdf
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v5, v6); // bagcedf
															}
															else {
																swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bagcdef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v0)) {
														if (compare(v4, v1)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // ebacgfd
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ebacgdf
																}
																else {
																	swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // ebacdgf
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // beacgfd
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // beacgdf
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // beacdgf
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v6); // baecgfd
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // baecgdf
															}
															else {
																swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v5, v6); // baecdgf
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v3, v6); // bacgefd
															}
															else {
																swap(v0, v1); swap(v3, v6); swap(v5, v6); // bacgedf
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v1); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bacgdef
															}
															else {
																swap(v0, v1); swap(v4, v6); swap(v5, v6); // bacdgef
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v1); swap(v3, v4); swap(v4, v6); // bacegfd
															}
															else {
																swap(v0, v1); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bacegdf
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v1); swap(v3, v4); swap(v5, v6); // bacedgf
															}
															else {
																swap(v0, v1); swap(v5, v6); // bacdegf
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v1)) {
											if (compare(v6, v3)) {
												if (compare(v5, v2)) {
													if (compare(v5, v1)) {
														if (compare(v2, v0)) {
															if (compare(v2, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v5, v6); // efcbagd
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // efbcagd
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // efbagcd
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // efbacgd
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v2, v0)) {
																swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // ebfcagd
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // ebfagcd
																}
																else {
																	swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ebfacgd
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ebafgcd
															}
															else {
																swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // ebafcgd
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v5, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // ecfbagd
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cefbagd
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v5, v6); // ecbfagd
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // cebfagd
																}
															}
															else {
																swap(v0, v4); swap(v3, v5); swap(v5, v6); // ebcfagd
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ecbafgd
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cebafgd
															}
														}
														else {
															if (compare(v2, v0)) {
																swap(v0, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ebcafgd
															}
															else {
																swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ebacfgd
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v2, v1)) {
														if (compare(v5, v1)) {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v5); swap(v3, v5); // efcbadg
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); // ecfbadg
																}
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cefbadg
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v3, v5); // ecbfadg
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); // cebfadg
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); // efbcadg
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v2, v5); swap(v3, v5); // ebfcadg
																}
																else {
																	swap(v0, v4); swap(v3, v5); // ebcfadg
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // efbacdg
															}
															else {
																swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // ebfacdg
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v5); // ecbafdg
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v3, v4); // ecbadfg
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cebafdg
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); // cebadfg
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v3, v4); swap(v4, v5); // ebcafdg
															}
															else {
																swap(v0, v4); swap(v3, v4); // ebcadfg
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v2, v4); swap(v3, v5); // ebafcdg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // ebacfdg
																}
																else {
																	swap(v0, v4); swap(v2, v4); swap(v3, v4); // ebacdfg
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v2)) {
												if (compare(v5, v2)) {
													if (compare(v5, v0)) {
														if (compare(v2, v0)) {
															if (compare(v6, v3)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // befcagd
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); // befcadg
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // befagcd
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // befacgd
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // befacdg
																}
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v6, v2)) {
																swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // beafgcd
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // beafcgd
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); // beafcdg
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // baefgcd
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v5, v6); // baefcgd
																}
																else {
																	swap(v0, v1); swap(v2, v4); swap(v3, v5); // baefcdg
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v0)) {
														if (compare(v5, v0)) {
															if (compare(v6, v3)) {
																swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v5, v6); // becfagd
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v3, v5); // becfadg
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // becafgd
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v5); // becafdg
																}
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v3, v4); // becadfg
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // beacfgd
																}
																else {
																	swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // beacfdg
																}
															}
															else {
																swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); // beacdfg
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // baecfgd
																}
																else {
																	swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v5); // baecfdg
																}
															}
															else {
																swap(v0, v1); swap(v2, v4); swap(v3, v4); // baecdfg
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v2, v1)) {
														if (compare(v5, v0)) {
															if (compare(v6, v3)) {
																swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // cbefagd
															}
															else {
																swap(v0, v2); swap(v2, v4); swap(v3, v5); // cbefadg
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cbeafgd
																}
																else {
																	swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cbeafdg
																}
															}
															else {
																swap(v0, v2); swap(v2, v4); swap(v3, v4); // cbeadfg
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v6, v3)) {
																swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // bcefagd
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); // bcefadg
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bceafgd
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // bceafdg
																}
															}
															else {
																swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); // bceadfg
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v6, v3)) {
															if (compare(v2, v0)) {
																if (compare(v2, v1)) {
																	swap(v0, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cbaefgd
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bcaefgd
																}
															}
															else {
																swap(v0, v1); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bacefgd
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v2, v1)) {
																	swap(v0, v2); swap(v3, v4); swap(v4, v5); // cbaefdg
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v5); // bcaefdg
																}
															}
															else {
																swap(v0, v1); swap(v3, v4); swap(v4, v5); // bacefdg
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v2, v0)) {
																if (compare(v2, v1)) {
																	swap(v0, v2); swap(v3, v4); // cbaedfg
																}
																else {
																	swap(v0, v1); swap(v1, v2); swap(v3, v4); // bcaedfg
																}
															}
															else {
																swap(v0, v1); swap(v3, v4); // bacedfg
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v2, v1)) {
																	swap(v0, v2); // cbadefg
																}
																else {
																	swap(v0, v1); swap(v1, v2); // bcadefg
																}
															}
															else {
																swap(v0, v1); // bacdefg
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
				else {
					if (compare(v3, v2)) {
						if (compare(v3, v0)) {
							if (compare(v5, v4)) {
								if (compare(v6, v0)) {
									if (compare(v6, v5)) {
										if (compare(v4, v1)) {
											if (compare(v4, v2)) {
												if (compare(v4, v0)) {
													if (compare(v5, v3)) {
														if (compare(v4, v3)) {
															if (compare(v2, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v5, v6); // gfedcab
															}
															else {
																if (compare(v2, v1)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // gfedacb
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v6); // gfedabc
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // gfdecab
															}
															else {
																if (compare(v2, v1)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gfdeacb
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // gfdeabc
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v2, v0)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gdfecab
															}
															else {
																if (compare(v2, v1)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // gdfeacb
																}
																else {
																	swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gdfeabc
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dgfecab
															}
															else {
																if (compare(v2, v1)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); // dgfeacb
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // dgfeabc
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v5, v0)) {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // gfdaecb
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); // gdfaecb
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); // dgfaecb
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // gdafecb
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); // dgafecb
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); // gfdaebc
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // gdfaebc
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v5, v6); // dgfaebc
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); // gdafebc
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // dgafebc
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v2, v0)) {
														if (compare(v4, v0)) {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v5, v6); // gfdceab
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // gdfceab
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); // dgfceab
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // gfdcaeb
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gdfcaeb
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dgfcaeb
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gfdaceb
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // gdfaceb
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v5); // dgfaceb
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gdafceb
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dgafceb
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v0)) {
															if (compare(v6, v2)) {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v5, v6); // gdcfeab
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v3, v5); // dgcfeab
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); // dcgfeab
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gdcfaeb
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v5); // dgcfaeb
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // dcgfaeb
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v6, v2)) {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v5); // gdcafeb
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v4, v5); // dgcafeb
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v5); // dcgafeb
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // gdacfeb
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v5); // dgacfeb
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v3)) {
												if (compare(v5, v0)) {
													if (compare(v2, v1)) {
														if (compare(v2, v0)) {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v6); // gfdcabe
																}
																else {
																	swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // gdfcabe
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v6); // gdcfabe
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // gfdacbe
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gdfacbe
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // gfdabec
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gdfabec
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gfdabce
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // gdfabce
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v2, v0)) {
															if (compare(v5, v1)) {
																swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gdcafbe
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v6); // gdcabfe
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // gdafcbe
																}
																else {
																	swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gdacfbe
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // gdacbfe
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // gdafbec
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gdafbce
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v5); // gdabfec
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // gdabfce
																}
																else {
																	swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v6); // gdabcfe
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v0)) {
														if (compare(v2, v1)) {
															if (compare(v2, v0)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dgfcabe
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // dgfacbe
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // dgfabec
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v6); // dgfabce
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v2, v1)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dgafcbe
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dgafbec
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // dgafbce
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // dgabfec
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dgabfce
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v5, v0)) {
															if (compare(v6, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dgcfabe
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dcgfabe
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v6, v2)) {
																	swap(v0, v3); swap(v1, v6); swap(v4, v5); swap(v5, v6); // dgcafbe
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // dcgafbe
																}
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // dgacfbe
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v6, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v4, v6); // dgcabfe
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v6); // dcgabfe
															}
														}
														else {
															if (compare(v2, v1)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v6); // dgacbfe
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // dgabcfe
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v2, v0)) {
											if (compare(v6, v2)) {
												if (compare(v4, v2)) {
													if (compare(v6, v4)) {
														if (compare(v6, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v4); // fgedcab
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); // fgdecab
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // fdgecab
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dfgecab
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v6); // fegdcab
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // fedgcab
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); // fdegcab
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // dfegcab
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v6, v3)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v3); // fgdceab
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); // fdgceab
															}
														}
														else {
															swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // dfgceab
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v5); // fgdcaeb
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // fdgcaeb
																}
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dfgcaeb
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // fgdcabe
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fdgcabe
																}
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // dfgcabe
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v5, v3)) {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v6); // fedcgab
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // fdecgab
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v5); swap(v1, v3); swap(v3, v6); // fdcgeab
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v6); // fdcegab
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dfecgab
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v5, v6); // dfcgeab
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dfcegab
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // dcfgeab
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dcfegab
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v5); // fdcgaeb
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dfcgaeb
															}
														}
														else {
															swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dcfgaeb
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fdcgabe
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v6); // dfcgabe
															}
														}
														else {
															swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // dcfgabe
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v4, v0)) {
													if (compare(v6, v4)) {
														if (compare(v6, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v5); // fgedacb
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fgdeacb
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fdgeacb
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // dfgeacb
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // fegdacb
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fedgacb
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // fdegacb
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dfegacb
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); // fgdaecb
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // fdgaecb
																}
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v5, v6); // dfgaecb
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fgdaceb
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fdgaceb
																}
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // dfgaceb
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v3)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fgdacbe
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // fdgacbe
															}
														}
														else {
															swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v6); // dfgacbe
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v6, v4)) {
														if (compare(v6, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // fgedabc
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fgdeabc
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // fdgeabc
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); // dfgeabc
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v5); // fegdabc
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // fedgabc
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fdegabc
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // dfegabc
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v6, v3)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // fgdaebc
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); // fdgaebc
															}
														}
														else {
															swap(v0, v3); swap(v1, v5); swap(v2, v6); // dfgaebc
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fgdabec
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // fdgabec
																}
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v5); // dfgabec
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // fgdabce
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fdgabce
																}
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // dfgabce
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v1)) {
										if (compare(v6, v2)) {
											if (compare(v2, v1)) {
												if (compare(v5, v0)) {
													if (compare(v5, v3)) {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); // fdagecb
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // fdagceb
																}
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fdagcbe
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // fedagcb
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fdeagcb
																}
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // fdaegcb
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v0)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // dfeagcb
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // dfagecb
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dfaegcb
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dfagceb
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // dfagcbe
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // dagfecb
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); // dafgecb
																}
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // dafegcb
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dagfceb
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // dafgceb
																}
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // dagcfeb
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v5)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // dagfcbe
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dafgcbe
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dagcfbe
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // dagcbfe
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v5, v3)) {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // fdagebc
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fdagbec
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // fdagbce
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fedagbc
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // fdeagbc
																}
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fdaegbc
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v0)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v6); // dfeagbc
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); // dfagebc
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // dfaegbc
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // dfagbec
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dfagbce
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); // dagfebc
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // dafgebc
																}
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dafegbc
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // dagfbec
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dafgbec
																}
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v5); // dagbfec
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v5)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dagfbce
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // dafgbce
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // dagbfce
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v6); // dagbcfe
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v0)) {
												if (compare(v5, v2)) {
													if (compare(v4, v0)) {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // fedcagb
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fdecagb
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // dfecagb
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fdceagb
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v5, v6); // dfceagb
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fdcageb
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v5, v6); // fdcaegb
																}
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v6); // fdcagbe
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v5); swap(v4, v6); swap(v5, v6); // dfcageb
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v5, v6); // dfcaegb
																}
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v4, v6); // dfcagbe
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // dcfageb
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v6); // dcfagbe
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // dcfeagb
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v5, v6); // dcfaegb
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // dcagfeb
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dcafgeb
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // dcafegb
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dcagfbe
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // dcafgbe
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // dcagbfe
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v4, v2)) {
														if (compare(v4, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fedacgb
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // fdeacgb
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v5, v6); // dfeacgb
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fdaecgb
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // dfaecgb
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fdacgeb
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // fdacegb
																}
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // fdacgbe
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // dfacgeb
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v5, v6); // dfacegb
																}
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v6); // dfacgbe
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dafecgb
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dafcgeb
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // dafcegb
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v5); // dacgfeb
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dacfgeb
																}
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v5, v6); // dacfegb
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dacgfbe
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v6); // dacgbfe
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // dafcgbe
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v6); // dacfgbe
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v2, v1)) {
											if (compare(v2, v0)) {
												if (compare(v4, v1)) {
													if (compare(v4, v0)) {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v5); // fedcabg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // fdecabg
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); // dfecabg
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v5); // fdceabg
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v3, v4); // dfceabg
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); // dcfeabg
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v3, v5); // fdcaebg
															}
															else {
																swap(v0, v3); swap(v1, v5); // dfcaebg
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v5); // dcfaebg
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); // dcafebg
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v6, v4)) {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fdcabge
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v4, v5); swap(v5, v6); // dfcabge
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // dcfabge
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v5); // fdcabeg
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v4, v5); // dfcabeg
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v5); // dcfabeg
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dcafbge
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dcafbeg
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v6); // dcabgfe
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // dcabfge
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v5); // dcabfeg
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v4, v2)) {
														if (compare(v4, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fedacbg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); // fdeacbg
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v4); // dfeacbg
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fdaecbg
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); // dfaecbg
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); // fdacebg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fdacbge
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fdacbeg
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v3); // dfacebg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // dfacbge
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v5); // dfacbeg
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v1)) {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dafecbg
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); // dafcebg
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dafcbge
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dafcbeg
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v3); swap(v3, v5); // dacfebg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dacfbge
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v5); // dacfbeg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v4, v6); // dacbgfe
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v4, v5); swap(v5, v6); // dacbfge
																}
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v4, v5); // dacbfeg
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v2)) {
												if (compare(v5, v0)) {
													if (compare(v4, v1)) {
														if (compare(v4, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // fedabgc
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fdeabgc
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // dfeabgc
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // fdaebgc
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dfaebgc
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // fdabgec
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v5, v6); // fdabegc
																}
															}
															else {
																swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v6); // fdabgce
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dfabgec
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // dfabegc
																}
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // dfabgce
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v6, v4)) {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // dafbgec
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v6); // dafbgce
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // dafebgc
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v5, v6); // dafbegc
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // dabgfec
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dabfgec
																}
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // dabfegc
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dabgfce
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // dabfgce
																}
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // dabgcfe
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v4, v1)) {
														if (compare(v4, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); // fedabcg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // fdeabcg
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v5); // dfeabcg
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); // fdaebcg
															}
															else {
																swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dfaebcg
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v3); swap(v2, v5); // fdabecg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // fdabcge
																}
																else {
																	swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v5); // fdabceg
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); // dfabecg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dfabcge
																}
																else {
																	swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dfabceg
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); // dafebcg
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); // dafbecg
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // dafbcge
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v5); // dafbceg
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); // dabfecg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dabfcge
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dabfceg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v6); // dabcgfe
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // dabcfge
																}
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v5); // dabcfeg
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v6, v0)) {
									if (compare(v6, v4)) {
										if (compare(v5, v1)) {
											if (compare(v5, v2)) {
												if (compare(v5, v0)) {
													if (compare(v4, v3)) {
														if (compare(v5, v3)) {
															if (compare(v2, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // gefdcab
															}
															else {
																if (compare(v2, v1)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v6); // gefdacb
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // gefdabc
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gedfcab
															}
															else {
																if (compare(v2, v1)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // gedfacb
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gedfabc
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v2, v0)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // gdefcab
															}
															else {
																if (compare(v2, v1)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gdefacb
																}
																else {
																	swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // gdefabc
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); // dgefcab
															}
															else {
																if (compare(v2, v1)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // dgefacb
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dgefabc
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v4, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gedafcb
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // gdeafcb
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v5); // dgeafcb
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gdaefcb
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dgaefcb
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // gedafbc
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gdeafbc
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // dgeafbc
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // gdaefbc
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dgaefbc
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v2, v0)) {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // gedcfab
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gdecfab
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // dgecfab
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v6); // gedcafb
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // gdecafb
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); // dgecafb
																}
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // gedacfb
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); // gdeacfb
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v4); // dgeacfb
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // gdaecfb
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); // dgaecfb
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v0)) {
															if (compare(v6, v2)) {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gdcefab
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v5); // dgcefab
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // dcgefab
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v6); // gdceafb
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v3, v4); // dgceafb
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); // dcgeafb
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v6, v2)) {
																if (compare(v6, v3)) {
																	swap(v0, v6); swap(v1, v3); swap(v3, v6); // gdcaefb
																}
																else {
																	swap(v0, v3); swap(v1, v6); // dgcaefb
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); // dcgaefb
															}
														}
														else {
															if (compare(v6, v3)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); // gdacefb
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); // dgacefb
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v3)) {
												if (compare(v4, v0)) {
													if (compare(v2, v1)) {
														if (compare(v2, v0)) {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // gedcabf
																}
																else {
																	swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gdecabf
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gdceabf
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gedacbf
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // gdeacbf
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); // gedabfc
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // gdeabfc
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // gedabcf
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gdeabcf
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v2, v0)) {
															if (compare(v4, v1)) {
																swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v5, v6); // gdcaebf
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gdcabef
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gdaecbf
																}
																else {
																	swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // gdacebf
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gdacbef
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); // gdaebfc
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // gdaebcf
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v3); swap(v2, v6); // gdabefc
																}
																else {
																	swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v5, v6); // gdabecf
																}
															}
															else {
																swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // gdabcef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v4, v0)) {
														if (compare(v2, v1)) {
															if (compare(v2, v0)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // dgecabf
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v5, v6); // dgeacbf
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v6); // dgeabfc
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // dgeabcf
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v2, v1)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // dgaecbf
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // dgaebfc
																}
																else {
																	swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dgaebcf
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); // dgabefc
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // dgabecf
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v4, v0)) {
															if (compare(v6, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v5, v6); // dgceabf
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // dcgeabf
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v6, v2)) {
																	swap(v0, v3); swap(v1, v6); swap(v5, v6); // dgcaebf
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v5, v6); // dcgaebf
																}
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v5, v6); // dgacebf
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v6, v2)) {
																swap(v0, v3); swap(v1, v6); swap(v4, v6); swap(v5, v6); // dgcabef
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // dcgabef
															}
														}
														else {
															if (compare(v2, v1)) {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // dgacbef
															}
															else {
																swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dgabcef
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v2, v0)) {
											if (compare(v6, v2)) {
												if (compare(v5, v2)) {
													if (compare(v6, v5)) {
														if (compare(v6, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v5); // egfdcab
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // egdfcab
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // edgfcab
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // degfcab
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v3)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // efgdcab
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // efdgcab
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // edfgcab
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // defgcab
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v6, v3)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v5); // egdcfab
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // edgcfab
															}
														}
														else {
															swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // degcfab
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); // egdcafb
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); // edgcafb
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // degcafb
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v5, v6); // egdcabf
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // edgcabf
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // degcabf
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v4, v3)) {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // efdcgab
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // edfcgab
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v5); // edcgfab
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // edcfgab
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // defcgab
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // decgfab
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v6); // decfgab
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dcegfab
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // dcefgab
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v3, v6); // edcgafb
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v6); // decgafb
															}
														}
														else {
															swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // dcegafb
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v5, v6); // edcgabf
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // decgabf
															}
														}
														else {
															swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dcegabf
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v5, v0)) {
													if (compare(v6, v5)) {
														if (compare(v6, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v5); // egfdacb
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); // egdfacb
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // edgfacb
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // degfacb
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v3)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v5, v6); // efgdacb
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // efdgacb
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); // edfgacb
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // defgacb
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // egdafcb
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // edgafcb
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // degafcb
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); // egdacfb
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // edgacfb
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v6); // degacfb
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v3)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // egdacbf
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // edgacbf
															}
														}
														else {
															swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // degacbf
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v6, v5)) {
														if (compare(v6, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v5, v6); // egfdabc
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // egdfabc
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); // edgfabc
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // degfabc
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v3)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v6); // efgdabc
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); // efdgabc
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // edfgabc
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // defgabc
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v6, v3)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // egdafbc
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // edgafbc
															}
														}
														else {
															swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v5); // degafbc
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // egdabfc
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); // edgabfc
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v6); // degabfc
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // egdabcf
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // edgabcf
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v5, v6); // degabcf
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v1)) {
										if (compare(v6, v2)) {
											if (compare(v2, v1)) {
												if (compare(v4, v0)) {
													if (compare(v4, v3)) {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // edagfcb
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); // edagcfb
																}
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // edagcbf
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // efdagcb
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // edfagcb
																}
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // edafgcb
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v6); // defagcb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // deagfcb
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // deafgcb
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // deagcfb
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // deagcbf
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v5, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dagefcb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // daegfcb
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // daefgcb
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // dagecfb
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); // daegcfb
																}
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); // dagcefb
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dagecbf
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // daegcbf
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // dagcebf
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dagcbef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v4, v3)) {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // edagfbc
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // edagbfc
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // edagbcf
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // efdagbc
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // edfagbc
																}
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // edafgbc
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // defagbc
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // deagfbc
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // deafgbc
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); // deagbfc
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // deagbcf
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // dagefbc
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // daegfbc
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // daefgbc
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); // dagebfc
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // daegbfc
																}
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); // dagbefc
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // dagebcf
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // daegbcf
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v5, v6); // dagbecf
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // dagbcef
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v0)) {
												if (compare(v4, v2)) {
													if (compare(v5, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v5, v6); // efdcagb
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // edfcagb
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // defcagb
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v5, v6); // edcfagb
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // decfagb
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v6); // edcagfb
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // edcafgb
																}
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // edcagbf
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v4); swap(v4, v6); // decagfb
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v4, v5); swap(v5, v6); // decafgb
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v4, v6); swap(v5, v6); // decagbf
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v6); // dceagfb
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // dceagbf
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dcefagb
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // dceafgb
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); // dcagefb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // dcaegfb
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dcaefgb
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // dcagebf
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dcaegbf
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dcagbef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v5, v2)) {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // efdacgb
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // edfacgb
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // defacgb
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // edafcgb
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // deafcgb
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // edacgfb
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // edacfgb
																}
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // edacgbf
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v6); // deacgfb
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // deacfgb
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // deacgbf
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // daefcgb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // daecgfb
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // daecfgb
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v3); swap(v3, v6); // dacgefb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v6); // dacegfb
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dacefgb
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v5, v6); // dacgebf
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dacgbef
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // daecgbf
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dacegbf
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v2, v1)) {
											if (compare(v2, v0)) {
												if (compare(v5, v1)) {
													if (compare(v5, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v5); swap(v2, v3); // efdcabg
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); // edfcabg
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // defcabg
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v4); swap(v1, v3); swap(v3, v5); // edcfabg
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v5); // decfabg
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // dcefabg
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v5); // edcafbg
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v4, v5); // decafbg
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v5); // dceafbg
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dcaefbg
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v5)) {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v5, v6); // edcabgf
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v5, v6); // decabgf
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v5, v6); // dceabgf
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v4); swap(v1, v3); swap(v3, v4); // edcabfg
																}
																else {
																	swap(v0, v3); swap(v1, v4); // decabfg
																}
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v4); // dceabfg
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v5)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // dcaebgf
															}
															else {
																swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); // dcaebfg
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // dcabgef
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v5, v6); // dcabegf
																}
																else {
																	swap(v0, v3); swap(v1, v2); swap(v2, v3); // dcabefg
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v5, v2)) {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); // efdacbg
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // edfacbg
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v5); // defacbg
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); // edafcbg
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // deafcbg
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // edacfbg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // edacbgf
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); // edacbfg
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v5); // deacfbg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v5, v6); // deacbgf
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v3); // deacbfg
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); // daefcbg
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // daecfbg
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // daecbgf
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); // daecbfg
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v5); // dacefbg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v5, v6); // dacebgf
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v3, v4); // dacebfg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v3); swap(v4, v6); swap(v5, v6); // dacbgef
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v5, v6); // dacbegf
																}
																else {
																	swap(v0, v3); swap(v1, v3); // dacbefg
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v2)) {
												if (compare(v4, v0)) {
													if (compare(v5, v1)) {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // efdabgc
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // edfabgc
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v5, v6); // defabgc
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // edafbgc
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // deafbgc
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v6); // edabgfc
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // edabfgc
																}
															}
															else {
																swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // edabgcf
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // deabgfc
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // deabfgc
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // deabgcf
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v6, v5)) {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v6); // daebgfc
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // daebgcf
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // daefbgc
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // daebfgc
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); // dabgefc
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // dabegfc
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dabefgc
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // dabgecf
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dabegcf
																}
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dabgcef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v5, v1)) {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // efdabcg
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); // edfabcg
																}
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v5); // defabcg
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // edafbcg
															}
															else {
																swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); // deafbcg
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v5); // edabfcg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v5, v6); // edabcgf
																}
																else {
																	swap(v0, v4); swap(v1, v3); swap(v2, v4); // edabcfg
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // deabfcg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // deabcgf
																}
																else {
																	swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); // deabcfg
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v5, v1)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // daefbcg
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v5); // daebfcg
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v5, v6); // daebcgf
															}
															else {
																swap(v0, v3); swap(v1, v3); swap(v2, v4); // daebcfg
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dabefcg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // dabecgf
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); // dabecfg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // dabcgef
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v5, v6); // dabcegf
																}
																else {
																	swap(v0, v3); swap(v1, v3); swap(v2, v3); // dabcefg
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
						else {
							if (compare(v5, v4)) {
								if (compare(v6, v3)) {
									if (compare(v6, v5)) {
										if (compare(v4, v2)) {
											if (compare(v4, v1)) {
												if (compare(v4, v3)) {
													if (compare(v5, v0)) {
														if (compare(v4, v0)) {
															if (compare(v2, v1)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gfeadcb
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // gfeadbc
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gfeabdc
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // gfaedcb
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); // gfaedbc
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // gfaebdc
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v2, v1)) {
																swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); // gafedcb
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // gafedbc
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gafebdc
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																swap(v1, v6); swap(v2, v5); swap(v3, v4); // agfedcb
															}
															else {
																if (compare(v3, v1)) {
																	swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // agfedbc
																}
																else {
																	swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // agfebdc
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v5, v6); // gfadecb
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); // gafdecb
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); // gadfecb
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v1, v6); swap(v2, v5); // agfdecb
															}
															else {
																swap(v1, v6); swap(v2, v3); swap(v3, v5); // agdfecb
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); // gfadebc
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v5, v6); // gafdebc
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // gadfebc
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v1, v6); swap(v2, v5); swap(v5, v6); // agfdebc
															}
															else {
																swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // agdfebc
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v0)) {
													if (compare(v3, v1)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v5); // gfadbec
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // gafdbec
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gadfbec
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // gadbfec
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); // gfabedc
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // gfabdec
																}
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // gafbedc
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gafbdec
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); // gabfedc
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // gabfdec
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v5); // gabdfec
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v3, v1)) {
															if (compare(v5, v3)) {
																swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // agfdbec
															}
															else {
																swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // agdfbec
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // agfbedc
															}
															else {
																swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // agfbdec
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v3)) {
																	swap(v1, v6); swap(v2, v6); swap(v3, v5); // agbfedc
																}
																else {
																	swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // agbfdec
																}
															}
															else {
																if (compare(v3, v1)) {
																	swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // agdbfec
																}
																else {
																	swap(v1, v6); swap(v2, v6); swap(v4, v5); // agbdfec
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v2, v6); swap(v3, v5); // abgfedc
															}
															else {
																if (compare(v5, v3)) {
																	swap(v2, v6); swap(v3, v5); swap(v4, v5); // abgfdec
																}
																else {
																	swap(v2, v6); swap(v4, v5); // abgdfec
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v0)) {
												if (compare(v2, v1)) {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // gfadceb
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v5); // gafdceb
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // gadfceb
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v5); // gadcfeb
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v6); // gfadcbe
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // gafdcbe
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gadfcbe
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // gadcfbe
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v6); // gadcbfe
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // gfadbce
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v6); // gafdbce
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // gadfbce
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gadbfce
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // gadbcfe
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gfabdce
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // gafbdce
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gabfdce
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // gabdfce
																}
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v6); // gabdcfe
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v5, v3)) {
														if (compare(v2, v1)) {
															if (compare(v4, v1)) {
																swap(v1, v6); swap(v2, v5); swap(v4, v5); // agfdceb
															}
															else {
																swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // agfdcbe
															}
														}
														else {
															if (compare(v3, v1)) {
																swap(v1, v6); swap(v2, v5); swap(v4, v6); // agfdbce
															}
															else {
																swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // agfbdce
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v5, v2)) {
																swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // agdfceb
															}
															else {
																swap(v1, v6); swap(v2, v3); swap(v4, v5); // agdcfeb
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // agdfcbe
																}
																else {
																	swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // agdcfbe
																}
															}
															else {
																swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // agdfbce
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v6, v1)) {
															if (compare(v3, v1)) {
																swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // agdbfce
															}
															else {
																if (compare(v5, v3)) {
																	swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // agbfdce
																}
																else {
																	swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // agbdfce
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // abgfdce
															}
															else {
																swap(v2, v6); swap(v4, v5); swap(v5, v6); // abgdfce
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v2, v1)) {
																swap(v1, v6); swap(v2, v3); swap(v4, v6); // agdcbfe
															}
															else {
																swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // agdbcfe
															}
														}
														else {
															if (compare(v6, v1)) {
																swap(v1, v6); swap(v2, v6); swap(v4, v6); // agbdcfe
															}
															else {
																swap(v2, v6); swap(v4, v6); // abgdcfe
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v3, v1)) {
											if (compare(v2, v1)) {
												if (compare(v4, v3)) {
													if (compare(v6, v4)) {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // fgeadcb
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); // fgaedcb
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // fagedcb
															}
															else {
																swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // afgedcb
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v6, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fegadcb
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // feagdcb
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // faegdcb
															}
															else {
																swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // afegdcb
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v6, v0)) {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v5); // fgadecb
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v5); // fgadceb
																}
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // fgadcbe
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v5, v6); // fagdecb
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // fagdceb
																}
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v6); // fagdcbe
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v2)) {
																swap(v1, v5); swap(v2, v6); swap(v5, v6); // afgdecb
															}
															else {
																swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // afgdceb
															}
														}
														else {
															swap(v1, v5); swap(v2, v6); swap(v4, v6); // afgdcbe
														}
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v6, v4)) {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fgeadbc
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // fgaedbc
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); // fagedbc
															}
															else {
																swap(v1, v5); swap(v2, v6); swap(v3, v4); // afgedbc
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v6, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // fegadbc
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // feagdbc
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // faegdbc
															}
															else {
																swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // afegdbc
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v6, v0)) {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v5, v6); // fgadebc
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // fgadbec
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v6); // fgadbce
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v6); // fagdebc
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v5); // fagdbec
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // fagdbce
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															swap(v1, v5); swap(v2, v6); // afgdebc
														}
														else {
															if (compare(v4, v2)) {
																swap(v1, v5); swap(v2, v6); swap(v4, v5); // afgdbec
															}
															else {
																swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // afgdbce
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v1)) {
												if (compare(v4, v1)) {
													if (compare(v6, v4)) {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // fgeabdc
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fgaebdc
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // fagebdc
															}
															else {
																swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // afgebdc
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v6, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); // fegabdc
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // feagbdc
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // faegbdc
															}
															else {
																swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // afegbdc
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v6, v0)) {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // fgabedc
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fgabdec
																}
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // fgabdce
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); // fagbedc
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // fagbdec
																}
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fagbdce
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v1, v5); swap(v2, v6); swap(v3, v5); // afgbedc
															}
															else {
																swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // afgbdec
															}
														}
														else {
															swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // afgbdce
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v6, v4)) {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // fabgedc
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fabgdec
															}
														}
														else {
															swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // fabgdce
														}
													}
													else {
														if (compare(v4, v0)) {
															swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // feabgdc
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // faebgdc
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fabegdc
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // afebgdc
															}
															else {
																if (compare(v6, v4)) {
																	swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // afbgedc
																}
																else {
																	swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // afbegdc
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // afbgdec
															}
															else {
																swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // afbgdce
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																swap(v2, v5); swap(v3, v6); swap(v5, v6); // abfgedc
															}
															else {
																swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // abfegdc
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // abfgdec
															}
															else {
																swap(v2, v5); swap(v3, v6); swap(v4, v6); // abfgdce
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v2)) {
										if (compare(v6, v1)) {
											if (compare(v2, v1)) {
												if (compare(v4, v2)) {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v6); // feadgcb
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // faedgcb
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // fadgecb
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fadegcb
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // afedgcb
															}
															else {
																if (compare(v6, v4)) {
																	swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // afdgecb
																}
																else {
																	swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // afdegcb
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // adgfecb
																}
																else {
																	swap(v1, v3); swap(v2, v5); swap(v3, v6); // adfgecb
																}
															}
															else {
																swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // adfegcb
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fadgceb
															}
															else {
																swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // afdgceb
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // adgfceb
																}
																else {
																	swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // adfgceb
																}
															}
															else {
																swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // adgcfeb
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // adgfcbe
																}
																else {
																	swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // adgcfbe
																}
															}
															else {
																swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // adgcbfe
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // fadgcbe
															}
															else {
																if (compare(v5, v3)) {
																	swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // afdgcbe
																}
																else {
																	swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // adfgcbe
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // feadgbc
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v6); // faedgbc
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); // fadgebc
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // fadegbc
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v1, v5); swap(v2, v4); swap(v4, v6); // afedgbc
															}
															else {
																if (compare(v6, v4)) {
																	swap(v1, v5); swap(v2, v3); swap(v3, v6); // afdgebc
																}
																else {
																	swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // afdegbc
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v2, v6); swap(v3, v5); // adgfebc
																}
																else {
																	swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // adfgebc
																}
															}
															else {
																swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // adfegbc
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // fadgbec
															}
															else {
																swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // afdgbec
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // adgfbec
																}
																else {
																	swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // adfgbec
																}
															}
															else {
																swap(v1, v3); swap(v2, v6); swap(v4, v5); // adgbfec
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // adgfbce
															}
															else {
																if (compare(v5, v2)) {
																	swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // adgbfce
																}
																else {
																	swap(v1, v3); swap(v2, v6); swap(v4, v6); // adgbcfe
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fadgbce
															}
															else {
																if (compare(v5, v3)) {
																	swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // afdgbce
																}
																else {
																	swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // adfgbce
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v5, v3)) {
													if (compare(v5, v0)) {
														if (compare(v4, v1)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v5, v6); // feadbgc
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // faedbgc
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fadebgc
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fadbgec
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // fadbegc
																}
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // fadbgce
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v3)) {
																swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // afedbgc
															}
															else {
																swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // afdebgc
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // afdbgec
																}
																else {
																	swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // afdbegc
																}
															}
															else {
																swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // afdbgce
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v6, v4)) {
															if (compare(v4, v2)) {
																swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // adfbgec
															}
															else {
																swap(v1, v3); swap(v2, v5); swap(v4, v6); // adfbgce
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // adfebgc
															}
															else {
																swap(v1, v3); swap(v2, v5); swap(v5, v6); // adfbegc
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // adbgfec
																}
																else {
																	swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // adbfgec
																}
															}
															else {
																swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // adbfegc
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // adbgfce
																}
																else {
																	swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // adbfgce
																}
															}
															else {
																swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // adbgcfe
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // feabdgc
															}
															else {
																if (compare(v4, v1)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // faebdgc
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // fabedgc
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // fabdgec
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v5, v6); // fabdegc
																}
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v6); // fabdgce
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // afebdgc
															}
															else {
																swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // afbedgc
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // afbdgec
																}
																else {
																	swap(v1, v5); swap(v2, v5); swap(v5, v6); // afbdegc
																}
															}
															else {
																swap(v1, v5); swap(v2, v5); swap(v4, v6); // afbdgce
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v2, v5); swap(v3, v4); swap(v5, v6); // abfedgc
															}
															else {
																if (compare(v6, v4)) {
																	swap(v2, v5); swap(v4, v6); swap(v5, v6); // abfdgec
																}
																else {
																	swap(v2, v5); swap(v5, v6); // abfdegc
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v2, v3); swap(v3, v6); swap(v4, v5); // abdgfec
																}
																else {
																	swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // abdfgec
																}
															}
															else {
																swap(v2, v3); swap(v3, v5); swap(v5, v6); // abdfegc
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v2)) {
																swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // abdgfce
															}
															else {
																swap(v2, v3); swap(v3, v6); swap(v4, v6); // abdgcfe
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v2, v5); swap(v4, v6); // abfdgce
															}
															else {
																swap(v2, v3); swap(v3, v5); swap(v4, v6); // abdfgce
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v2, v1)) {
											if (compare(v6, v1)) {
												if (compare(v5, v3)) {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // feadcgb
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v5, v6); // faedcgb
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // fadecgb
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // fadcgeb
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v5, v6); // fadcegb
																}
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v6); // fadcgbe
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v1, v5); swap(v2, v4); swap(v5, v6); // afedcgb
															}
															else {
																swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // afdecgb
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // afdcgeb
																}
																else {
																	swap(v1, v5); swap(v2, v3); swap(v5, v6); // afdcegb
																}
															}
															else {
																swap(v1, v5); swap(v2, v3); swap(v4, v6); // afdcgbe
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // adfecgb
															}
															else {
																if (compare(v6, v4)) {
																	swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // adfcgeb
																}
																else {
																	swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // adfcegb
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v3, v6); swap(v4, v5); // adcgfeb
																}
																else {
																	swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // adcfgeb
																}
															}
															else {
																swap(v1, v3); swap(v3, v5); swap(v5, v6); // adcfegb
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // adcgfbe
															}
															else {
																swap(v1, v3); swap(v3, v6); swap(v4, v6); // adcgbfe
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // adfcgbe
															}
															else {
																swap(v1, v3); swap(v3, v5); swap(v4, v6); // adcfgbe
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v5); // feadcbg
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v4); // faedcbg
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); // fadecbg
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); // fadcebg
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v1, v5); swap(v2, v4); // afedcbg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v1, v5); swap(v2, v3); swap(v3, v4); // afdecbg
																}
																else {
																	swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // adfecbg
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v1, v5); swap(v2, v3); // afdcebg
																}
																else {
																	swap(v1, v3); swap(v2, v5); swap(v3, v5); // adfcebg
																}
															}
															else {
																swap(v1, v3); swap(v3, v5); // adcfebg
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v6, v4)) {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // fadcbge
															}
															else {
																if (compare(v5, v3)) {
																	swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // afdcbge
																}
																else {
																	swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // adfcbge
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v5); // fadcbeg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v1, v5); swap(v2, v3); swap(v4, v5); // afdcbeg
																}
																else {
																	swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // adfcbeg
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // adcfbge
															}
															else {
																swap(v1, v3); swap(v3, v5); swap(v4, v5); // adcfbeg
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v4, v6); // adcbgfe
																}
																else {
																	swap(v1, v3); swap(v4, v5); swap(v5, v6); // adcbfge
																}
															}
															else {
																swap(v1, v3); swap(v4, v5); // adcbfeg
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v4, v2)) {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); // feadbcg
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v5); // faedbcg
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fadebcg
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); // fadbecg
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v3)) {
																swap(v1, v5); swap(v2, v4); swap(v4, v5); // afedbcg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // afdebcg
																}
																else {
																	swap(v1, v3); swap(v2, v5); swap(v3, v4); // adfebcg
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	swap(v1, v5); swap(v2, v3); swap(v3, v5); // afdbecg
																}
																else {
																	swap(v1, v3); swap(v2, v5); // adfbecg
																}
															}
															else {
																swap(v1, v3); swap(v2, v3); swap(v3, v5); // adbfecg
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v6, v4)) {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fadbcge
															}
															else {
																if (compare(v5, v3)) {
																	swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // afdbcge
																}
																else {
																	swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // adfbcge
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fadbceg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // afdbceg
																}
																else {
																	swap(v1, v3); swap(v2, v5); swap(v4, v5); // adfbceg
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v4)) {
																swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // adbfcge
															}
															else {
																swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // adbfceg
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v2, v3); swap(v4, v6); // adbcgfe
																}
																else {
																	swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // adbcfge
																}
															}
															else {
																swap(v1, v3); swap(v2, v3); swap(v4, v5); // adbcfeg
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); // feabdcg
															}
															else {
																if (compare(v4, v1)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // faebdcg
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); // fabedcg
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v5); // fabdecg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // fabdcge
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v5); // fabdceg
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v1)) {
																swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // afebdcg
															}
															else {
																if (compare(v4, v3)) {
																	swap(v1, v5); swap(v2, v5); swap(v3, v4); // afbedcg
																}
																else {
																	swap(v1, v5); swap(v2, v5); // afbdecg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // afbdcge
															}
															else {
																swap(v1, v5); swap(v2, v5); swap(v4, v5); // afbdceg
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v4, v2)) {
															if (compare(v4, v3)) {
																swap(v2, v5); swap(v3, v4); // abfedcg
															}
															else {
																swap(v2, v5); // abfdecg
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v2, v5); swap(v4, v5); swap(v5, v6); // abfdcge
															}
															else {
																swap(v2, v5); swap(v4, v5); // abfdceg
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v2, v3); swap(v3, v5); // abdfecg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // abdfcge
																}
																else {
																	swap(v2, v3); swap(v3, v5); swap(v4, v5); // abdfceg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v2, v3); swap(v4, v6); // abdcgfe
																}
																else {
																	swap(v2, v3); swap(v4, v5); swap(v5, v6); // abdcfge
																}
															}
															else {
																swap(v2, v3); swap(v4, v5); // abdcfeg
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v6, v3)) {
									if (compare(v6, v4)) {
										if (compare(v5, v2)) {
											if (compare(v5, v1)) {
												if (compare(v5, v3)) {
													if (compare(v4, v0)) {
														if (compare(v5, v0)) {
															if (compare(v2, v1)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // gefadcb
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gefadbc
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // gefabdc
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // geafdcb
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // geafdbc
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); // geafbdc
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v2, v1)) {
																swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // gaefdcb
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gaefdbc
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // gaefbdc
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // agefdcb
															}
															else {
																if (compare(v3, v1)) {
																	swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // agefdbc
																}
																else {
																	swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // agefbdc
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // geadfcb
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v5); // gaedfcb
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // gadefcb
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v1, v6); swap(v2, v4); swap(v4, v5); // agedfcb
															}
															else {
																swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // agdefcb
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v5); // geadfbc
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // gaedfbc
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gadefbc
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // agedfbc
															}
															else {
																swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // agdefbc
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v0)) {
													if (compare(v3, v1)) {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); // geadbfc
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v6); // gaedbfc
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // gadebfc
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); // gadbefc
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v0)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // geabfdc
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); // geabdfc
																}
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gaebfdc
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // gaebdfc
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // gabefdc
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); // gabedfc
																}
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v6); // gabdefc
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v3, v1)) {
															if (compare(v4, v3)) {
																swap(v1, v6); swap(v2, v4); swap(v4, v6); // agedbfc
															}
															else {
																swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // agdebfc
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // agebfdc
															}
															else {
																swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // agebdfc
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // agbefdc
																}
																else {
																	swap(v1, v6); swap(v2, v6); swap(v3, v4); // agbedfc
																}
															}
															else {
																if (compare(v3, v1)) {
																	swap(v1, v6); swap(v2, v3); swap(v3, v6); // agdbefc
																}
																else {
																	swap(v1, v6); swap(v2, v6); // agbdefc
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v2, v6); swap(v3, v4); swap(v4, v5); // abgefdc
																}
																else {
																	swap(v2, v6); swap(v3, v4); // abgedfc
																}
															}
															else {
																swap(v2, v6); // abgdefc
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v0)) {
												if (compare(v2, v1)) {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v6); // geadcfb
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v4); // gaedcfb
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); // gadecfb
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); // gadcefb
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // geadcbf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v5, v6); // gaedcbf
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // gadecbf
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v5, v6); // gadcebf
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // gadcbef
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v4, v1)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v5, v6); // geadbcf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // gaedbcf
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gadebcf
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // gadbecf
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gadbcef
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // geabdcf
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gaebdcf
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // gabedcf
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v5, v6); // gabdecf
																}
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // gabdcef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v4, v3)) {
														if (compare(v2, v1)) {
															if (compare(v5, v1)) {
																swap(v1, v6); swap(v2, v4); // agedcfb
															}
															else {
																swap(v1, v6); swap(v2, v4); swap(v5, v6); // agedcbf
															}
														}
														else {
															if (compare(v3, v1)) {
																swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // agedbcf
															}
															else {
																swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // agebdcf
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v4, v2)) {
																swap(v1, v6); swap(v2, v3); swap(v3, v4); // agdecfb
															}
															else {
																swap(v1, v6); swap(v2, v3); // agdcefb
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // agdecbf
																}
																else {
																	swap(v1, v6); swap(v2, v3); swap(v5, v6); // agdcebf
																}
															}
															else {
																swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // agdebcf
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v6, v1)) {
															if (compare(v3, v1)) {
																swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // agdbecf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // agbedcf
																}
																else {
																	swap(v1, v6); swap(v2, v6); swap(v5, v6); // agbdecf
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v2, v6); swap(v3, v4); swap(v5, v6); // abgedcf
															}
															else {
																swap(v2, v6); swap(v5, v6); // abgdecf
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v2, v1)) {
																swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // agdcbef
															}
															else {
																swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // agdbcef
															}
														}
														else {
															if (compare(v6, v1)) {
																swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // agbdcef
															}
															else {
																swap(v2, v6); swap(v4, v6); swap(v5, v6); // abgdcef
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v3, v1)) {
											if (compare(v2, v1)) {
												if (compare(v5, v3)) {
													if (compare(v6, v5)) {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); // egfadcb
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // egafdcb
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // eagfdcb
															}
															else {
																swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // aegfdcb
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v6, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // efgadcb
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // efagdcb
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // eafgdcb
															}
															else {
																swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // aefgdcb
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v0)) {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v5); // egadfcb
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v4); // egadcfb
																}
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v5, v6); // egadcbf
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // eagdfcb
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v6); // eagdcfb
																}
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // eagdcbf
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // aegdfcb
															}
															else {
																swap(v1, v4); swap(v2, v6); swap(v4, v6); // aegdcfb
															}
														}
														else {
															swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // aegdcbf
														}
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v6, v5)) {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // egfadbc
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // egafdbc
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // eagfdbc
															}
															else {
																swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // aegfdbc
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v6, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); // efgadbc
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // efagdbc
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // eafgdbc
															}
															else {
																swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // aefgdbc
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v0)) {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // egadfbc
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v6); // egadbfc
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // egadbcf
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v5); // eagdfbc
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v6); // eagdbfc
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v5, v6); // eagdbcf
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															swap(v1, v4); swap(v2, v6); swap(v4, v5); // aegdfbc
														}
														else {
															if (compare(v5, v2)) {
																swap(v1, v4); swap(v2, v6); // aegdbfc
															}
															else {
																swap(v1, v4); swap(v2, v6); swap(v5, v6); // aegdbcf
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v1)) {
												if (compare(v5, v1)) {
													if (compare(v6, v5)) {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // egfabdc
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // egafbdc
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); // eagfbdc
															}
															else {
																swap(v1, v4); swap(v2, v6); swap(v3, v5); // aegfbdc
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v6, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // efgabdc
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // efagbdc
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // eafgbdc
															}
															else {
																swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // aefgbdc
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v0)) {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // egabfdc
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // egabdfc
																}
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // egabdcf
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // eagbfdc
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); // eagbdfc
																}
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // eagbdcf
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // aegbfdc
															}
															else {
																swap(v1, v4); swap(v2, v6); swap(v3, v4); // aegbdfc
															}
														}
														else {
															swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // aegbdcf
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v6, v5)) {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // eabgfdc
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // eabgdfc
															}
														}
														else {
															swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // eabgdcf
														}
													}
													else {
														if (compare(v5, v0)) {
															swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // efabgdc
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // eafbgdc
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // eabfgdc
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // aefbgdc
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // aebgfdc
																}
																else {
																	swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // aebfgdc
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // aebgdfc
															}
															else {
																swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // aebgdcf
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v5)) {
																swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // abegfdc
															}
															else {
																swap(v2, v4); swap(v3, v5); swap(v4, v6); // abefgdc
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v2, v4); swap(v3, v6); swap(v4, v6); // abegdfc
															}
															else {
																swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // abegdcf
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v2)) {
										if (compare(v6, v1)) {
											if (compare(v2, v1)) {
												if (compare(v5, v2)) {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // efadgcb
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v6); // eafdgcb
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // eadgfcb
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // eadfgcb
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v1, v4); swap(v2, v5); swap(v4, v6); // aefdgcb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // aedgfcb
																}
																else {
																	swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // aedfgcb
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // adgefcb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // adegfcb
																}
																else {
																	swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // adefgcb
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // eadgcfb
															}
															else {
																swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // aedgcfb
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // adgecfb
																}
																else {
																	swap(v1, v3); swap(v2, v4); swap(v3, v6); // adegcfb
																}
															}
															else {
																swap(v1, v3); swap(v2, v6); swap(v3, v6); // adgcefb
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // adgecbf
																}
																else {
																	swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // adgcebf
																}
															}
															else {
																swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // adgcbef
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // eadgcbf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // aedgcbf
																}
																else {
																	swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // adegcbf
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v6); // efadgbc
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // eafdgbc
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // eadgfbc
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // eadfgbc
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // aefdgbc
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // aedgfbc
																}
																else {
																	swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // aedfgbc
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // adgefbc
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // adegfbc
																}
																else {
																	swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // adefgbc
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); // eadgbfc
															}
															else {
																swap(v1, v4); swap(v2, v3); swap(v3, v6); // aedgbfc
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v1, v3); swap(v2, v6); swap(v3, v4); // adgebfc
																}
																else {
																	swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // adegbfc
																}
															}
															else {
																swap(v1, v3); swap(v2, v6); // adgbefc
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // adgebcf
															}
															else {
																if (compare(v4, v2)) {
																	swap(v1, v3); swap(v2, v6); swap(v5, v6); // adgbecf
																}
																else {
																	swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // adgbcef
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // eadgbcf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // aedgbcf
																}
																else {
																	swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // adegbcf
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v4, v3)) {
													if (compare(v4, v0)) {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // efadbgc
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v5, v6); // eafdbgc
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // eadfbgc
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // eadbgfc
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // eadbfgc
																}
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // eadbgcf
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																swap(v1, v4); swap(v2, v5); swap(v5, v6); // aefdbgc
															}
															else {
																swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // aedfbgc
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // aedbgfc
																}
																else {
																	swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // aedbfgc
																}
															}
															else {
																swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // aedbgcf
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v6, v5)) {
															if (compare(v5, v2)) {
																swap(v1, v3); swap(v2, v4); swap(v4, v6); // adebgfc
															}
															else {
																swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // adebgcf
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // adefbgc
															}
															else {
																swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // adebfgc
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v4)) {
																swap(v1, v3); swap(v2, v3); swap(v3, v6); // adbgefc
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // adbegfc
																}
																else {
																	swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // adbefgc
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // adbgecf
																}
																else {
																	swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // adbegcf
																}
															}
															else {
																swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // adbgcef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // efabdgc
															}
															else {
																if (compare(v5, v1)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // eafbdgc
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // eabfdgc
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v6); // eabdgfc
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // eabdfgc
																}
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // eabdgcf
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // aefbdgc
															}
															else {
																swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // aebfdgc
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	swap(v1, v4); swap(v2, v4); swap(v4, v6); // aebdgfc
																}
																else {
																	swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // aebdfgc
																}
															}
															else {
																swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // aebdgcf
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // abefdgc
															}
															else {
																if (compare(v6, v5)) {
																	swap(v2, v4); swap(v4, v6); // abedgfc
																}
																else {
																	swap(v2, v4); swap(v4, v5); swap(v5, v6); // abedfgc
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v2, v3); swap(v3, v6); // abdgefc
															}
															else {
																if (compare(v6, v5)) {
																	swap(v2, v3); swap(v3, v4); swap(v4, v6); // abdegfc
																}
																else {
																	swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // abdefgc
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v2)) {
																swap(v2, v3); swap(v3, v6); swap(v5, v6); // abdgecf
															}
															else {
																swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // abdgcef
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v2, v4); swap(v4, v6); swap(v5, v6); // abedgcf
															}
															else {
																swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // abdegcf
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v2, v1)) {
											if (compare(v6, v1)) {
												if (compare(v4, v3)) {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v5, v6); // efadcgb
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // eafdcgb
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // eadfcgb
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v6); // eadcgfb
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // eadcfgb
																}
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // eadcgbf
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // aefdcgb
															}
															else {
																swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // aedfcgb
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v1, v4); swap(v2, v3); swap(v4, v6); // aedcgfb
																}
																else {
																	swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // aedcfgb
																}
															}
															else {
																swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // aedcgbf
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // adefcgb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // adecgfb
																}
																else {
																	swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // adecfgb
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v1, v3); swap(v3, v6); // adcgefb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v3, v4); swap(v4, v6); // adcegfb
																}
																else {
																	swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // adcefgb
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																swap(v1, v3); swap(v3, v6); swap(v5, v6); // adcgebf
															}
															else {
																swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // adcgbef
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // adecgbf
															}
															else {
																swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // adcegbf
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); // efadcbg
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v5); // eafdcbg
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // eadfcbg
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v5); // eadcfbg
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v1, v4); swap(v2, v5); swap(v4, v5); // aefdcbg
																}
																else {
																	swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // aedfcbg
																}
															}
															else {
																swap(v1, v3); swap(v2, v4); swap(v3, v5); // adefcbg
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	swap(v1, v4); swap(v2, v3); swap(v4, v5); // aedcfbg
																}
																else {
																	swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // adecfbg
																}
															}
															else {
																swap(v1, v3); swap(v3, v4); swap(v4, v5); // adcefbg
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v6, v5)) {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v5, v6); // eadcbgf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v1, v4); swap(v2, v3); swap(v5, v6); // aedcbgf
																}
																else {
																	swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // adecbgf
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); // eadcbfg
															}
															else {
																if (compare(v4, v3)) {
																	swap(v1, v4); swap(v2, v3); // aedcbfg
																}
																else {
																	swap(v1, v3); swap(v2, v4); swap(v3, v4); // adecbfg
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v5)) {
																swap(v1, v3); swap(v3, v4); swap(v5, v6); // adcebgf
															}
															else {
																swap(v1, v3); swap(v3, v4); // adcebfg
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v1, v3); swap(v4, v6); swap(v5, v6); // adcbgef
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v5, v6); // adcbegf
																}
																else {
																	swap(v1, v3); // adcbefg
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v5, v2)) {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v5); // efadbcg
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v5); // eafdbcg
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); // eadfbcg
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // eadbfcg
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v1, v4); swap(v2, v5); // aefdbcg
																}
																else {
																	swap(v1, v4); swap(v2, v3); swap(v3, v5); // aedfbcg
																}
															}
															else {
																swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // adefbcg
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // aedbfcg
																}
																else {
																	swap(v1, v3); swap(v2, v4); swap(v4, v5); // adebfcg
																}
															}
															else {
																swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // adbefcg
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v6, v5)) {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // eadbcgf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // aedbcgf
																}
																else {
																	swap(v1, v3); swap(v2, v4); swap(v5, v6); // adebcgf
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); // eadbcfg
															}
															else {
																if (compare(v4, v3)) {
																	swap(v1, v4); swap(v2, v3); swap(v3, v4); // aedbcfg
																}
																else {
																	swap(v1, v3); swap(v2, v4); // adebcfg
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v5)) {
																swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // adbecgf
															}
															else {
																swap(v1, v3); swap(v2, v3); swap(v3, v4); // adbecfg
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // adbcgef
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v3); swap(v2, v3); swap(v5, v6); // adbcegf
																}
																else {
																	swap(v1, v3); swap(v2, v3); // adbcefg
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // efabdcg
															}
															else {
																if (compare(v5, v1)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); // eafbdcg
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // eabfdcg
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v5); // eabdfcg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v5, v6); // eabdcgf
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v4); // eabdcfg
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v1)) {
																swap(v1, v4); swap(v2, v5); swap(v3, v4); // aefbdcg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // aebfdcg
																}
																else {
																	swap(v1, v4); swap(v2, v4); swap(v4, v5); // aebdfcg
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v1, v4); swap(v2, v4); swap(v5, v6); // aebdcgf
															}
															else {
																swap(v1, v4); swap(v2, v4); // aebdcfg
															}
														}
													}
												}
												else {
													if (compare(v4, v3)) {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																swap(v2, v4); swap(v3, v5); swap(v4, v5); // abefdcg
															}
															else {
																swap(v2, v4); swap(v4, v5); // abedfcg
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v2, v4); swap(v5, v6); // abedcgf
															}
															else {
																swap(v2, v4); // abedcfg
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v2, v3); swap(v3, v4); swap(v4, v5); // abdefcg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v2, v3); swap(v3, v4); swap(v5, v6); // abdecgf
																}
																else {
																	swap(v2, v3); swap(v3, v4); // abdecfg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v2, v3); swap(v4, v6); swap(v5, v6); // abdcgef
															}
															else {
																if (compare(v6, v5)) {
																	swap(v2, v3); swap(v5, v6); // abdcegf
																}
																else {
																	swap(v2, v3); // abdcefg
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
					else {
						if (compare(v2, v0)) {
							if (compare(v5, v4)) {
								if (compare(v6, v0)) {
									if (compare(v6, v5)) {
										if (compare(v4, v1)) {
											if (compare(v4, v3)) {
												if (compare(v4, v0)) {
													if (compare(v5, v2)) {
														if (compare(v4, v2)) {
															if (compare(v3, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // gfecdab
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gfecadb
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // gfecabd
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v5, v6); // gfcedab
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gfceadb
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v6); // gfceabd
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v3, v0)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // gcfedab
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gcfeadb
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // gcfeabd
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); // cgfedab
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cgfeadb
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cgfeabd
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v5, v0)) {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v5, v6); // gfcaedb
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // gcfaedb
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); // cgfaedb
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); // gcafedb
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v3, v5); // cgafedb
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v1, v5); swap(v3, v6); // gfcaebd
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); // gcfaebd
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // cgfaebd
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // gcafebd
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v5, v6); // cgafebd
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v2)) {
													if (compare(v3, v0)) {
														if (compare(v4, v0)) {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v1, v5); swap(v5, v6); // gfcdeab
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v5, v6); // gcfdeab
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // gcdfeab
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v5); swap(v4, v6); swap(v5, v6); // gfcdaeb
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // gcfdaeb
																}
															}
															else {
																if (compare(v5, v0)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gcdfaeb
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // gcdafeb
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gfcadeb
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gcfadeb
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // gcafdeb
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v5); // gcadfeb
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v0)) {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v5); // cgfdeab
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); // cgdfeab
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // cdgfeab
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v3, v0)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v5); // cgfdaeb
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cgfadeb
																}
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cgdfaeb
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cdgfaeb
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v6, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v5); // cgdafeb
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // cdgafeb
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v5); // cgafdeb
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v4, v5); // cgadfeb
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v2)) {
												if (compare(v5, v0)) {
													if (compare(v3, v1)) {
														if (compare(v3, v0)) {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v1, v5); swap(v4, v6); // gfcdabe
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v6); // gcfdabe
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // gcdfabe
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v6); // gfcadbe
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // gcfadbe
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v5); // gfcabed
															}
															else {
																swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gfcabde
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // gcfabed
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gcfabde
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v3, v0)) {
															if (compare(v5, v1)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gcdafbe
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // gcdabfe
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gcafdbe
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // gcadfbe
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v6); // gcadbfe
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gcafbed
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // gcafbde
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // gcabfed
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gcabfde
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // gcabdfe
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v5, v0)) {
														if (compare(v3, v1)) {
															if (compare(v3, v0)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // cgfdabe
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cgfadbe
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cgfabed
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // cgfabde
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v3, v1)) {
																swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cgafdbe
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cgafbed
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v6); // cgafbde
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v5); // cgabfed
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cgabfde
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v5, v0)) {
															if (compare(v6, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cgdfabe
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // cdgfabe
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v6, v3)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // cgdafbe
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cdgafbe
																}
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v4, v5); swap(v5, v6); // cgadfbe
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v6, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v6); // cgdabfe
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // cdgabfe
															}
														}
														else {
															if (compare(v3, v1)) {
																swap(v0, v2); swap(v1, v6); swap(v4, v6); // cgadbfe
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v6); // cgabdfe
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v3, v0)) {
											if (compare(v6, v3)) {
												if (compare(v4, v3)) {
													if (compare(v6, v4)) {
														if (compare(v6, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); // fgecdab
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v3, v4); // fgcedab
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); // fcgedab
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // cfgedab
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // fegcdab
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v6); // fecgdab
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // fcegdab
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cfegdab
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v6, v2)) {
																swap(v0, v5); swap(v1, v6); // fgcdeab
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); // fcgdeab
															}
														}
														else {
															swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v5, v6); // cfgdeab
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v5); swap(v1, v6); swap(v4, v5); // fgcdaeb
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v5); // fcgdaeb
																}
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // cfgdaeb
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v5); swap(v1, v6); swap(v4, v5); swap(v5, v6); // fgcdabe
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // fcgdabe
																}
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v6); // cfgdabe
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v5, v2)) {
														if (compare(v4, v3)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v4, v6); // fecdgab
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v6); // fcedgab
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); // fcdgeab
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // fcdegab
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // cfedgab
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // cfdgeab
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cfdegab
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); // cdfgeab
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // cdfegab
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // fcdgaeb
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cfdgaeb
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // cdfgaeb
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fcdgabe
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // cfdgabe
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cdfgabe
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v4, v0)) {
													if (compare(v6, v4)) {
														if (compare(v6, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // fgecadb
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v5); // fgceadb
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // fcgeadb
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cfgeadb
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fegcadb
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fecgadb
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fcegadb
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // cfegadb
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v4, v3)) {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v5); swap(v1, v6); swap(v3, v5); // fgcaedb
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); // fcgaedb
																}
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // cfgaedb
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v5); // fgcadeb
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // fcgadeb
																}
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cfgadeb
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v2)) {
																swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fgcadbe
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fcgadbe
															}
														}
														else {
															swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // cfgadbe
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v6, v4)) {
														if (compare(v6, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fgecabd
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fgceabd
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fcgeabd
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // cfgeabd
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // fegcabd
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v5); // fecgabd
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // fcegabd
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); // cfegabd
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v6, v2)) {
																swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v5, v6); // fgcaebd
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // fcgaebd
															}
														}
														else {
															swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); // cfgaebd
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fgcabed
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fcgabed
																}
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // cfgabed
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v6); // fgcabde
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // fcgabde
																}
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cfgabde
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v1)) {
										if (compare(v6, v3)) {
											if (compare(v3, v1)) {
												if (compare(v5, v0)) {
													if (compare(v5, v2)) {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // fcagedb
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fcagdeb
																}
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // fcagdbe
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v6); // fecagdb
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // fceagdb
																}
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fcaegdb
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cfeagdb
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v5, v6); // cfagedb
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cfaegdb
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cfagdeb
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v6); // cfagdbe
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); // cagfedb
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // cafgedb
																}
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cafegdb
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // cagfdeb
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cafgdeb
																}
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v5); // cagdfeb
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v5)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cagfdbe
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // cafgdbe
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // cagdfbe
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v6); // cagdbfe
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v5, v2)) {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); // fcagebd
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // fcagbed
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fcagbde
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fecagbd
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fceagbd
																}
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // fcaegbd
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // cfeagbd
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v5); swap(v3, v6); // cfagebd
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v6); // cfaegbd
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v5); // cfagbed
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cfagbde
															}
														}
													}
												}
												else {
													if (compare(v4, v3)) {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // cagfebd
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); // cafgebd
																}
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // cafegbd
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cagfbed
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // cafgbed
																}
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // cagbfed
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v5)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // cagfbde
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cafgbde
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cagbfde
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // cagbdfe
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v0)) {
												if (compare(v5, v3)) {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v4, v5); swap(v5, v6); // fecdagb
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // fcedagb
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fcdeagb
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v5, v6); // cfedagb
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // cfdeagb
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fcdageb
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // fcdaegb
																}
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // fcdagbe
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // cfdageb
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v5, v6); // cfdaegb
																}
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v6); // cfdagbe
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cdfageb
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // cdfagbe
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cdfeagb
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // cdfaegb
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v5); // cdagfeb
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cdafgeb
																}
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v5, v6); // cdafegb
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cdagfbe
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v6); // cdafgbe
																}
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v6); // cdagbfe
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v5, v2)) {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fecadgb
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fceadgb
																}
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // fcaedgb
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // fcadgeb
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v5, v6); // fcadegb
																}
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v6); // fcadgbe
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // cfeadgb
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v5, v6); // cfaedgb
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v5); swap(v4, v6); swap(v5, v6); // cfadgeb
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v5, v6); // cfadegb
																}
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v4, v6); // cfadgbe
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // cafedgb
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // cafdgeb
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v5, v6); // cafdegb
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // cadgfeb
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cadfgeb
																}
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // cadfegb
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cadgfbe
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // cadgbfe
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v6); // cafdgbe
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // cadfgbe
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v3, v1)) {
											if (compare(v3, v0)) {
												if (compare(v4, v1)) {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v4, v5); // fecdabg
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v5); // fcedabg
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fcdeabg
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); // cfedabg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); // cfdeabg
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cdfeabg
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); // fcdaebg
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v2, v3); // cfdaebg
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); // cdfaebg
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v3, v5); // cdafebg
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v6, v4)) {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fcdabge
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // cfdabge
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cdfabge
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fcdabeg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v5); // cfdabeg
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cdfabeg
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cdafbge
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v5); // cdafbeg
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v3); swap(v4, v6); // cdabgfe
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v4, v5); swap(v5, v6); // cdabfge
																}
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v4, v5); // cdabfeg
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v5, v2)) {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v5); // fecadbg
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // fceadbg
																}
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); // fcaedbg
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v5); // fcadebg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // fcadbge
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v5); // fcadbeg
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); // cfeadbg
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v2); swap(v1, v5); swap(v3, v4); // cfaedbg
																}
																else {
																	swap(v0, v2); swap(v1, v5); // cfadebg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v5); swap(v4, v5); swap(v5, v6); // cfadbge
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v4, v5); // cfadbeg
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v4, v1)) {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); // cafedbg
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); // cafdebg
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // cafdbge
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v5); // cafdbeg
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v4, v1)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); // cadfebg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cadfbge
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cadfbeg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v6); // cadbgfe
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // cadbfge
																}
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v5); // cadbfeg
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v3)) {
												if (compare(v5, v0)) {
													if (compare(v4, v1)) {
														if (compare(v4, v0)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v5, v6); // fecabgd
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // fceabgd
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cfeabgd
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fcaebgd
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cfaebgd
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fcabged
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // fcabegd
																}
															}
															else {
																swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // fcabgde
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cfabged
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v5, v6); // cfabegd
																}
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v6); // cfabgde
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v6, v4)) {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cafbged
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // cafbgde
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cafebgd
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // cafbegd
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v5); // cabgfed
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cabfged
																}
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v5, v6); // cabfegd
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cabgfde
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v6); // cabfgde
																}
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v6); // cabgdfe
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v4, v1)) {
														if (compare(v4, v0)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v4); swap(v3, v5); // fecabdg
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); // fceabdg
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cfeabdg
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fcaebdg
															}
															else {
																swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v5); // cfaebdg
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); // fcabedg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fcabdge
																}
																else {
																	swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fcabdeg
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v5); swap(v3, v5); // cfabedg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cfabdge
																}
																else {
																	swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v5); // cfabdeg
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cafebdg
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); // cafbedg
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cafbdge
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cafbdeg
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v2); swap(v3, v5); // cabfedg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cabfdge
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v5); // cabfdeg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v4, v6); // cabdgfe
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v4, v5); swap(v5, v6); // cabdfge
																}
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v4, v5); // cabdfeg
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v6, v0)) {
									if (compare(v6, v4)) {
										if (compare(v5, v1)) {
											if (compare(v5, v3)) {
												if (compare(v5, v0)) {
													if (compare(v4, v2)) {
														if (compare(v5, v2)) {
															if (compare(v3, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gefcdab
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // gefcadb
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gefcabd
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gecfdab
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v6); // gecfadb
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gecfabd
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v3, v0)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gcefdab
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // gcefadb
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gcefabd
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // cgefdab
															}
															else {
																if (compare(v3, v1)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); // cgefadb
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // cgefabd
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v4, v0)) {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gecafdb
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gceafdb
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cgeafdb
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // gcaefdb
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v5); // cgaefdb
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v5); // gecafbd
															}
															else {
																if (compare(v6, v2)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // gceafbd
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cgeafbd
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gcaefbd
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cgaefbd
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v2)) {
													if (compare(v3, v0)) {
														if (compare(v5, v0)) {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v4); swap(v4, v5); swap(v5, v6); // gecdfab
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // gcedfab
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gcdefab
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v4, v2)) {
																	swap(v0, v6); swap(v1, v4); swap(v4, v6); // gecdafb
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v6); // gcedafb
																}
															}
															else {
																if (compare(v4, v0)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // gcdeafb
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); // gcdaefb
																}
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v6); // gecadfb
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // gceadfb
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); // gcaedfb
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); // gcadefb
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v5); // cgedfab
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cgdefab
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cdgefab
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v3, v0)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v4); // cgedafb
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); // cgeadfb
																}
															}
															else {
																if (compare(v6, v3)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); // cgdeafb
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // cdgeafb
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v6, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v3); // cgdaefb
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); // cdgaefb
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v3, v4); // cgaedfb
															}
															else {
																swap(v0, v2); swap(v1, v6); // cgadefb
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v2)) {
												if (compare(v4, v0)) {
													if (compare(v3, v1)) {
														if (compare(v3, v0)) {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v4); swap(v4, v6); swap(v5, v6); // gecdabf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // gcedabf
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gcdeabf
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gecadbf
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gceadbf
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v4); swap(v3, v6); // gecabfd
															}
															else {
																swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v5, v6); // gecabdf
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); // gceabfd
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // gceabdf
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v3, v0)) {
															if (compare(v4, v1)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // gcdaebf
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gcdabef
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // gcaedbf
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v5, v6); // gcadebf
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // gcadbef
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // gcaebfd
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gcaebdf
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); // gcabefd
																}
																else {
																	swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // gcabedf
																}
															}
															else {
																swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gcabdef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v4, v0)) {
														if (compare(v3, v1)) {
															if (compare(v3, v0)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v5, v6); // cgedabf
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // cgeadbf
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // cgeabfd
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cgeabdf
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v3, v1)) {
																swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v5, v6); // cgaedbf
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v6); // cgaebfd
																}
																else {
																	swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cgaebdf
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v3, v6); // cgabefd
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v5, v6); // cgabedf
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v4, v0)) {
															if (compare(v6, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // cgdeabf
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cdgeabf
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v6, v3)) {
																	swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v5, v6); // cgdaebf
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // cdgaebf
																}
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v5, v6); // cgadebf
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v6, v3)) {
																swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // cgdabef
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cdgabef
															}
														}
														else {
															if (compare(v3, v1)) {
																swap(v0, v2); swap(v1, v6); swap(v4, v6); swap(v5, v6); // cgadbef
															}
															else {
																swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cgabdef
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v3, v0)) {
											if (compare(v6, v3)) {
												if (compare(v5, v3)) {
													if (compare(v6, v5)) {
														if (compare(v6, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // egfcdab
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v5); // egcfdab
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // ecgfdab
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cegfdab
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v2)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // efgcdab
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // efcgdab
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ecfgdab
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // cefgdab
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v6, v2)) {
																swap(v0, v4); swap(v1, v6); swap(v4, v5); // egcdfab
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v5); // ecgdfab
															}
														}
														else {
															swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // cegdfab
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v4); swap(v1, v6); // egcdafb
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v6); // ecgdafb
																}
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v6); // cegdafb
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v4); swap(v1, v6); swap(v5, v6); // egcdabf
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v5, v6); // ecgdabf
																}
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // cegdabf
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v4, v2)) {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v5); swap(v4, v6); swap(v5, v6); // efcdgab
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // ecfdgab
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // ecdgfab
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ecdfgab
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v6); // cefdgab
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cedgfab
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // cedfgab
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // cdegfab
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cdefgab
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); // ecdgafb
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // cedgafb
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); // cdegafb
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // ecdgabf
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cedgabf
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // cdegabf
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v1)) {
												if (compare(v5, v0)) {
													if (compare(v6, v5)) {
														if (compare(v6, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); // egfcadb
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v3, v5); // egcfadb
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); // ecgfadb
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // cegfadb
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v2)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // efgcadb
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v5, v6); // efcgadb
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // ecfgadb
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cefgadb
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v5); // egcafdb
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // ecgafdb
																}
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cegafdb
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v4); swap(v1, v6); swap(v3, v4); // egcadfb
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); // ecgadfb
																}
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // cegadfb
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v2)) {
																swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v5, v6); // egcadbf
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // ecgadbf
															}
														}
														else {
															swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cegadbf
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v6, v5)) {
														if (compare(v6, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // egfcabd
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v5, v6); // egcfabd
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // ecgfabd
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cegfabd
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v2)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); // efgcabd
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v3, v6); // efcgabd
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); // ecfgabd
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // cefgabd
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v6, v2)) {
																swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // egcafbd
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ecgafbd
															}
														}
														else {
															swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // cegafbd
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v6); // egcabfd
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // ecgabfd
																}
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); // cegabfd
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // egcabdf
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ecgabdf
																}
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // cegabdf
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v1)) {
										if (compare(v6, v3)) {
											if (compare(v3, v1)) {
												if (compare(v4, v0)) {
													if (compare(v4, v2)) {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ecagfdb
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // ecagdfb
																}
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ecagdbf
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // efcagdb
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ecfagdb
																}
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // ecafgdb
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // cefagdb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ceagfdb
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v6); // ceafgdb
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v6); // ceagdfb
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ceagdbf
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // cagefdb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // caegfdb
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // caefgdb
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); // cagedfb
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // caegdfb
																}
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); // cagdefb
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // cagedbf
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // caegdbf
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v5, v6); // cagdebf
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // cagdbef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v4, v2)) {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // ecagfbd
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); // ecagbfd
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // ecagbdf
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v6); // efcagbd
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // ecfagbd
																}
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ecafgbd
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cefagbd
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v5); // ceagfbd
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ceafgbd
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v4); swap(v3, v6); // ceagbfd
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v5, v6); // ceagbdf
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cagefbd
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // caegfbd
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // caefgbd
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // cagebfd
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); // caegbfd
																}
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); // cagbefd
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cagebdf
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // caegbdf
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // cagbedf
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cagbdef
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v3, v0)) {
												if (compare(v4, v3)) {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v5); swap(v5, v6); // efcdagb
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v5, v6); // ecfdagb
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // ecdfagb
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // cefdagb
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cedfagb
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // ecdagfb
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ecdafgb
																}
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ecdagbf
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v6); // cedagfb
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // cedafgb
																}
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // cedagbf
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // cdeagfb
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cdeagbf
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // cdefagb
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cdeafgb
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v3); swap(v3, v6); // cdagefb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v6); // cdaegfb
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cdaefgb
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v5, v6); // cdagebf
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cdaegbf
																}
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cdagbef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v4, v2)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v5, v6); // efcadgb
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // ecfadgb
																}
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ecafdgb
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v6); // ecadgfb
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // ecadfgb
																}
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // ecadgbf
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cefadgb
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ceafdgb
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v4); swap(v4, v6); // ceadgfb
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v4, v5); swap(v5, v6); // ceadfgb
																}
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v4, v6); swap(v5, v6); // ceadgbf
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // caefdgb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v6); // caedgfb
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // caedfgb
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); // cadgefb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // cadegfb
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cadefgb
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // cadgebf
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cadgbef
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // caedgbf
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cadegbf
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v3, v1)) {
											if (compare(v3, v0)) {
												if (compare(v5, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v5); // efcdabg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v2); swap(v2, v5); // ecfdabg
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); // ecdfabg
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v5); // cefdabg
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cedfabg
																}
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); // cdefabg
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // ecdafbg
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v5); // cedafbg
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cdeafbg
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v5); // cdaefbg
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v5)) {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // ecdabgf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v5, v6); // cedabgf
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // cdeabgf
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); // ecdabfg
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v2); swap(v1, v4); swap(v2, v3); // cedabfg
																}
																else {
																	swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); // cdeabfg
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v5)) {
																swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v5, v6); // cdaebgf
															}
															else {
																swap(v0, v2); swap(v1, v3); swap(v3, v4); // cdaebfg
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v3); swap(v4, v6); swap(v5, v6); // cdabgef
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v3); swap(v5, v6); // cdabegf
																}
																else {
																	swap(v0, v2); swap(v1, v3); // cdabefg
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v4, v2)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v5); swap(v3, v4); // efcadbg
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); // ecfadbg
																}
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // ecafdbg
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v5); // ecadfbg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v5, v6); // ecadbgf
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v4); // ecadbfg
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cefadbg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v5); // ceafdbg
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v4, v5); // ceadfbg
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v2); swap(v1, v4); swap(v5, v6); // ceadbgf
															}
															else {
																swap(v0, v2); swap(v1, v4); // ceadbfg
															}
														}
													}
												}
												else {
													if (compare(v4, v3)) {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // caefdbg
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v5); // caedfbg
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v5, v6); // caedbgf
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); // caedbfg
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v5, v1)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cadefbg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // cadebgf
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); // cadebfg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // cadbgef
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v5, v6); // cadbegf
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v2, v3); // cadbefg
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v3)) {
												if (compare(v4, v0)) {
													if (compare(v5, v1)) {
														if (compare(v5, v0)) {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // efcabgd
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ecfabgd
																}
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // cefabgd
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // ecafbgd
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v5, v6); // ceafbgd
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // ecabgfd
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ecabfgd
																}
															}
															else {
																swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ecabgdf
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v6); // ceabgfd
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ceabfgd
																}
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ceabgdf
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v6, v5)) {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // caebgfd
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // caebgdf
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // caefbgd
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // caebfgd
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v2); swap(v3, v6); // cabgefd
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v6); // cabegfd
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cabefgd
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v5, v6); // cabgedf
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cabegdf
																}
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cabgdef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v5, v1)) {
														if (compare(v5, v0)) {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v5); // efcabdg
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // ecfabdg
																}
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); // cefabdg
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); // ecafbdg
															}
															else {
																swap(v0, v2); swap(v1, v4); swap(v3, v5); // ceafbdg
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // ecabfdg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // ecabdgf
																}
																else {
																	swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); // ecabdfg
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v5); // ceabfdg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v5, v6); // ceabdgf
																}
																else {
																	swap(v0, v2); swap(v1, v4); swap(v3, v4); // ceabdfg
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); // caefbdg
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // caebfdg
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // caebdgf
															}
															else {
																swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); // caebdfg
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v5); // cabefdg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v5, v6); // cabedgf
																}
																else {
																	swap(v0, v2); swap(v1, v2); swap(v3, v4); // cabedfg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v2); swap(v1, v2); swap(v4, v6); swap(v5, v6); // cabdgef
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v2); swap(v1, v2); swap(v5, v6); // cabdegf
																}
																else {
																	swap(v0, v2); swap(v1, v2); // cabdefg
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
						else {
							if (compare(v5, v4)) {
								if (compare(v6, v2)) {
									if (compare(v6, v5)) {
										if (compare(v4, v3)) {
											if (compare(v4, v1)) {
												if (compare(v4, v2)) {
													if (compare(v5, v0)) {
														if (compare(v4, v0)) {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // gfeacdb
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); // gfeacbd
																}
															}
															else {
																swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // gfeabcd
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gfaecdb
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // gfaecbd
																}
															}
															else {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gfaebcd
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // gafecdb
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gafecbd
																}
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // gafebcd
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // agfecdb
																}
																else {
																	swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // agfecbd
																}
															}
															else {
																swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // agfebcd
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // gfacedb
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); // gafcedb
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v3, v5); // gacfedb
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v1, v6); swap(v2, v5); swap(v3, v5); // agfcedb
															}
															else {
																swap(v1, v6); swap(v3, v5); // agcfedb
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); // gfacebd
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // gafcebd
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v5, v6); // gacfebd
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // agfcebd
															}
															else {
																swap(v1, v6); swap(v3, v5); swap(v5, v6); // agcfebd
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v0)) {
													if (compare(v2, v1)) {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // gfacbed
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gafcbed
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gacfbed
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v5); // gacbfed
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																if (compare(v4, v2)) {
																	swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // gfabecd
																}
																else {
																	swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gfabced
																}
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); // gafbecd
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // gafbced
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // gabfecd
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gabfced
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // gabcfed
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v2, v1)) {
															if (compare(v5, v2)) {
																swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // agfcbed
															}
															else {
																swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // agcfbed
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v1, v6); swap(v2, v5); swap(v3, v6); // agfbecd
															}
															else {
																swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // agfbced
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v2)) {
																	swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // agbfecd
																}
																else {
																	swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // agbfced
																}
															}
															else {
																if (compare(v2, v1)) {
																	swap(v1, v6); swap(v3, v6); swap(v4, v5); // agcbfed
																}
																else {
																	swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // agbcfed
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v2, v6); swap(v3, v5); swap(v5, v6); // abgfecd
															}
															else {
																if (compare(v5, v2)) {
																	swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // abgfced
																}
																else {
																	swap(v2, v6); swap(v3, v6); swap(v4, v5); // abgcfed
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v0)) {
												if (compare(v3, v1)) {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gfacdeb
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // gafcdeb
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v5); // gacfdeb
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v4, v5); // gacdfeb
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // gfacdbe
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gafcdbe
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gacfdbe
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v4, v5); swap(v5, v6); // gacdfbe
																}
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v4, v6); // gacdbfe
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gfacbde
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // gafcbde
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v6); // gacfbde
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gacbfde
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v6); // gacbdfe
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // gfabcde
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gafbcde
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // gabfcde
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gabcfde
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // gabcdfe
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v5, v2)) {
														if (compare(v3, v1)) {
															if (compare(v4, v1)) {
																swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // agfcdeb
															}
															else {
																swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // agfcdbe
															}
														}
														else {
															if (compare(v2, v1)) {
																swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // agfcbde
															}
															else {
																swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // agfbcde
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v5, v3)) {
																swap(v1, v6); swap(v3, v5); swap(v4, v5); // agcfdeb
															}
															else {
																swap(v1, v6); swap(v4, v5); // agcdfeb
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v5, v3)) {
																	swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // agcfdbe
																}
																else {
																	swap(v1, v6); swap(v4, v5); swap(v5, v6); // agcdfbe
																}
															}
															else {
																swap(v1, v6); swap(v3, v5); swap(v4, v6); // agcfbde
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v6, v1)) {
															if (compare(v2, v1)) {
																swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // agcbfde
															}
															else {
																if (compare(v5, v2)) {
																	swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // agbfcde
																}
																else {
																	swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // agbcfde
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v2, v6); swap(v3, v5); swap(v4, v6); // abgfcde
															}
															else {
																swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // abgcfde
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v3, v1)) {
																swap(v1, v6); swap(v4, v6); // agcdbfe
															}
															else {
																swap(v1, v6); swap(v3, v6); swap(v4, v6); // agcbdfe
															}
														}
														else {
															if (compare(v6, v1)) {
																swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // agbcdfe
															}
															else {
																swap(v2, v6); swap(v3, v6); swap(v4, v6); // abgcdfe
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v2, v1)) {
											if (compare(v3, v1)) {
												if (compare(v4, v2)) {
													if (compare(v6, v4)) {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); // fgeacdb
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fgaecdb
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fagecdb
															}
															else {
																swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // afgecdb
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v6, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // fegacdb
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // feagcdb
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // faegcdb
															}
															else {
																swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // afegcdb
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v6, v0)) {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); // fgacedb
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fgacdeb
																}
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fgacdbe
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // fagcedb
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fagcdeb
																}
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // fagcdbe
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v3)) {
																swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // afgcedb
															}
															else {
																swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // afgcdeb
															}
														}
														else {
															swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // afgcdbe
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v6, v4)) {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // fgeacbd
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fgaecbd
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // fagecbd
															}
															else {
																swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // afgecbd
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v6, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fegacbd
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // feagcbd
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); // faegcbd
															}
															else {
																swap(v1, v5); swap(v2, v4); swap(v3, v6); // afegcbd
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v6, v0)) {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // fgacebd
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fgacbed
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // fgacbde
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); // fagcebd
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // fagcbed
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fagcbde
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															swap(v1, v5); swap(v2, v6); swap(v3, v6); // afgcebd
														}
														else {
															if (compare(v4, v3)) {
																swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // afgcbed
															}
															else {
																swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // afgcbde
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v1)) {
												if (compare(v4, v1)) {
													if (compare(v6, v4)) {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fgeabcd
															}
															else {
																swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // fgaebcd
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fagebcd
															}
															else {
																swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // afgebcd
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v6, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // fegabcd
															}
															else {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); // feagbcd
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // faegbcd
															}
															else {
																swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // afegbcd
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v6, v0)) {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); // fgabecd
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // fgabced
																}
																else {
																	swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fgabcde
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // fagbecd
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fagbced
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // fagbcde
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // afgbecd
														}
														else {
															if (compare(v4, v3)) {
																swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // afgbced
															}
															else {
																swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // afgbcde
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v0)) {
													if (compare(v6, v4)) {
														if (compare(v4, v2)) {
															swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); // fabgecd
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // fabgced
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fabgcde
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // feabgcd
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // faebgcd
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // fabegcd
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v4, v1)) {
																swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // afebgcd
															}
															else {
																if (compare(v6, v4)) {
																	swap(v1, v5); swap(v2, v5); swap(v3, v6); // afbgecd
																}
																else {
																	swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // afbegcd
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // afbgced
															}
															else {
																swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // afbgcde
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																swap(v2, v5); swap(v3, v6); // abfgecd
															}
															else {
																swap(v2, v5); swap(v3, v4); swap(v4, v6); // abfegcd
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v2, v5); swap(v3, v6); swap(v4, v5); // abfgced
															}
															else {
																swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // abfgcde
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v3)) {
										if (compare(v6, v1)) {
											if (compare(v3, v1)) {
												if (compare(v5, v2)) {
													if (compare(v5, v0)) {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v5, v6); // facgedb
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // facgdeb
																}
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v6); // facgdbe
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // feacgdb
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // faecgdb
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // facegdb
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v2)) {
																swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // afecgdb
															}
															else {
																if (compare(v6, v4)) {
																	swap(v1, v5); swap(v3, v6); swap(v5, v6); // afcgedb
																}
																else {
																	swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // afcegdb
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // afcgdeb
															}
															else {
																swap(v1, v5); swap(v3, v6); swap(v4, v6); // afcgdbe
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v2, v6); swap(v3, v5); // acgfedb
																}
																else {
																	swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // acfgedb
																}
															}
															else {
																swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // acfegdb
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // acgfdeb
																}
																else {
																	swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // acfgdeb
																}
															}
															else {
																swap(v1, v2); swap(v2, v6); swap(v4, v5); // acgdfeb
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v5)) {
																swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // acgfdbe
															}
															else {
																swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // acfgdbe
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // acgdfbe
															}
															else {
																swap(v1, v2); swap(v2, v6); swap(v4, v6); // acgdbfe
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // feacgbd
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // faecgbd
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v0, v5); swap(v1, v5); swap(v3, v6); // facgebd
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v6); // facegbd
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // afecgbd
															}
															else {
																if (compare(v6, v4)) {
																	swap(v1, v5); swap(v3, v6); // afcgebd
																}
																else {
																	swap(v1, v5); swap(v3, v4); swap(v4, v6); // afcegbd
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // acgfebd
																}
																else {
																	swap(v1, v2); swap(v2, v5); swap(v3, v6); // acfgebd
																}
															}
															else {
																swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // acfegbd
															}
														}
													}
												}
												else {
													if (compare(v4, v3)) {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v5); // facgbed
															}
															else {
																swap(v1, v5); swap(v3, v6); swap(v4, v5); // afcgbed
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // acgfbed
																}
																else {
																	swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // acfgbed
																}
															}
															else {
																swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // acgbfed
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // acgfbde
															}
															else {
																if (compare(v5, v3)) {
																	swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // acgbfde
																}
																else {
																	swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // acgbdfe
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // facgbde
															}
															else {
																if (compare(v5, v2)) {
																	swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // afcgbde
																}
																else {
																	swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // acfgbde
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v5, v2)) {
													if (compare(v5, v0)) {
														if (compare(v4, v1)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // feacbgd
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // faecbgd
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // facebgd
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // facbged
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v5, v6); // facbegd
																}
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v6); // facbgde
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v2)) {
																swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // afecbgd
															}
															else {
																swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // afcebgd
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // afcbged
																}
																else {
																	swap(v1, v5); swap(v3, v5); swap(v5, v6); // afcbegd
																}
															}
															else {
																swap(v1, v5); swap(v3, v5); swap(v4, v6); // afcbgde
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v6, v4)) {
															if (compare(v4, v3)) {
																swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // acfbged
															}
															else {
																swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // acfbgde
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // acfebgd
															}
															else {
																swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // acfbegd
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v3, v6); swap(v4, v5); // acbgfed
																}
																else {
																	swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // acbfged
																}
															}
															else {
																swap(v1, v2); swap(v3, v5); swap(v5, v6); // acbfegd
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // acbgfde
																}
																else {
																	swap(v1, v2); swap(v3, v5); swap(v4, v6); // acbfgde
																}
															}
															else {
																swap(v1, v2); swap(v3, v6); swap(v4, v6); // acbgdfe
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // feabcgd
															}
															else {
																if (compare(v4, v1)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // faebcgd
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fabecgd
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fabcged
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // fabcegd
																}
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // fabcgde
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v1)) {
																swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // afebcgd
															}
															else {
																swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // afbecgd
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // afbcged
																}
																else {
																	swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // afbcegd
																}
															}
															else {
																swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // afbcgde
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v6, v4)) {
															if (compare(v4, v3)) {
																swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // abfcged
															}
															else {
																swap(v2, v5); swap(v3, v5); swap(v4, v6); // abfcgde
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // abfecgd
															}
															else {
																swap(v2, v5); swap(v3, v5); swap(v5, v6); // abfcegd
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v3, v6); swap(v4, v5); // abcgfed
																}
																else {
																	swap(v3, v5); swap(v4, v6); swap(v5, v6); // abcfged
																}
															}
															else {
																swap(v3, v5); swap(v5, v6); // abcfegd
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v3, v6); swap(v4, v5); swap(v5, v6); // abcgfde
																}
																else {
																	swap(v3, v5); swap(v4, v6); // abcfgde
																}
															}
															else {
																swap(v3, v6); swap(v4, v6); // abcgdfe
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v3, v1)) {
											if (compare(v6, v1)) {
												if (compare(v5, v2)) {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // feacdgb
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // faecdgb
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v5, v6); // facedgb
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v5); swap(v4, v6); swap(v5, v6); // facdgeb
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v5, v6); // facdegb
																}
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v4, v6); // facdgbe
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v4, v2)) {
																swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // afecdgb
															}
															else {
																swap(v1, v5); swap(v3, v4); swap(v5, v6); // afcedgb
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v1, v5); swap(v4, v6); swap(v5, v6); // afcdgeb
																}
																else {
																	swap(v1, v5); swap(v5, v6); // afcdegb
																}
															}
															else {
																swap(v1, v5); swap(v4, v6); // afcdgbe
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // acfedgb
															}
															else {
																if (compare(v6, v4)) {
																	swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // acfdgeb
																}
																else {
																	swap(v1, v2); swap(v2, v5); swap(v5, v6); // acfdegb
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // acdgfeb
																}
																else {
																	swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // acdfgeb
																}
															}
															else {
																swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // acdfegb
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // acdgfbe
															}
															else {
																swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // acdgbfe
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v1, v2); swap(v2, v5); swap(v4, v6); // acfdgbe
															}
															else {
																swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // acdfgbe
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // feacdbg
															}
															else {
																swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); // faecdbg
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v5); swap(v3, v4); // facedbg
															}
															else {
																swap(v0, v5); swap(v1, v5); // facdebg
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																swap(v1, v5); swap(v2, v4); swap(v3, v4); // afecdbg
															}
															else {
																if (compare(v4, v3)) {
																	swap(v1, v5); swap(v3, v4); // afcedbg
																}
																else {
																	swap(v1, v5); // afcdebg
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v1, v2); swap(v2, v5); swap(v3, v4); // acfedbg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v1, v2); swap(v2, v5); // acfdebg
																}
																else {
																	swap(v1, v2); swap(v2, v3); swap(v3, v5); // acdfebg
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v6, v4)) {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v4, v5); swap(v5, v6); // facdbge
															}
															else {
																if (compare(v5, v2)) {
																	swap(v1, v5); swap(v4, v5); swap(v5, v6); // afcdbge
																}
																else {
																	swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // acfdbge
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v5); swap(v1, v5); swap(v4, v5); // facdbeg
															}
															else {
																if (compare(v5, v2)) {
																	swap(v1, v5); swap(v4, v5); // afcdbeg
																}
																else {
																	swap(v1, v2); swap(v2, v5); swap(v4, v5); // acfdbeg
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // acdfbge
															}
															else {
																swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // acdfbeg
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v2, v3); swap(v4, v6); // acdbgfe
																}
																else {
																	swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // acdbfge
																}
															}
															else {
																swap(v1, v2); swap(v2, v3); swap(v4, v5); // acdbfeg
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v5, v2)) {
													if (compare(v5, v0)) {
														if (compare(v4, v1)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); // feacbdg
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // faecbdg
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v5); // facebdg
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v5); swap(v3, v5); // facbedg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // facbdge
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v5); // facbdeg
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v2)) {
																swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // afecbdg
															}
															else {
																swap(v1, v5); swap(v3, v4); swap(v4, v5); // afcebdg
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v1, v5); swap(v3, v5); // afcbedg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // afcbdge
																}
																else {
																	swap(v1, v5); swap(v3, v5); swap(v4, v5); // afcbdeg
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v4, v1)) {
																swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // acfebdg
															}
															else {
																swap(v1, v2); swap(v2, v5); swap(v3, v5); // acfbedg
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // acfbdge
															}
															else {
																swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // acfbdeg
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v1, v2); swap(v3, v5); // acbfedg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // acbfdge
																}
																else {
																	swap(v1, v2); swap(v3, v5); swap(v4, v5); // acbfdeg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v4, v6); // acbdgfe
																}
																else {
																	swap(v1, v2); swap(v4, v5); swap(v5, v6); // acbdfge
																}
															}
															else {
																swap(v1, v2); swap(v4, v5); // acbdfeg
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v4, v0)) {
																swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // feabcdg
															}
															else {
																if (compare(v4, v1)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); // faebcdg
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fabecdg
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); // fabcedg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fabcdge
																}
																else {
																	swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fabcdeg
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v1)) {
																swap(v1, v5); swap(v2, v4); swap(v3, v5); // afebcdg
															}
															else {
																swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // afbecdg
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v1, v5); swap(v2, v5); swap(v3, v5); // afbcedg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // afbcdge
																}
																else {
																	swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // afbcdeg
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v3)) {
															if (compare(v4, v2)) {
																swap(v2, v5); swap(v3, v4); swap(v4, v5); // abfecdg
															}
															else {
																swap(v2, v5); swap(v3, v5); // abfcedg
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // abfcdge
															}
															else {
																swap(v2, v5); swap(v3, v5); swap(v4, v5); // abfcdeg
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																swap(v3, v5); // abcfedg
															}
															else {
																if (compare(v6, v4)) {
																	swap(v3, v5); swap(v4, v5); swap(v5, v6); // abcfdge
																}
																else {
																	swap(v3, v5); swap(v4, v5); // abcfdeg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	swap(v4, v6); // abcdgfe
																}
																else {
																	swap(v4, v5); swap(v5, v6); // abcdfge
																}
															}
															else {
																swap(v4, v5); // abcdfeg
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v6, v2)) {
									if (compare(v6, v4)) {
										if (compare(v5, v3)) {
											if (compare(v5, v1)) {
												if (compare(v5, v2)) {
													if (compare(v4, v0)) {
														if (compare(v5, v0)) {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gefacdb
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // gefacbd
																}
															}
															else {
																swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); // gefabcd
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // geafcdb
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // geafcbd
																}
															}
															else {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // geafbcd
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); // gaefcdb
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // gaefcbd
																}
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gaefbcd
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	swap(v1, v6); swap(v2, v4); swap(v3, v5); // agefcdb
																}
																else {
																	swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // agefcbd
																}
															}
															else {
																swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // agefbcd
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // geacfdb
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // gaecfdb
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v5); // gacefdb
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // agecfdb
															}
															else {
																swap(v1, v6); swap(v3, v4); swap(v4, v5); // agcefdb
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // geacfbd
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gaecfbd
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gacefbd
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // agecfbd
															}
															else {
																swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // agcefbd
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v0)) {
													if (compare(v2, v1)) {
														if (compare(v4, v2)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); // geacbfd
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // gaecbfd
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v6); // gacebfd
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v3, v6); // gacbefd
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v0)) {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // geabfcd
																}
																else {
																	swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // geabcfd
																}
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // gaebfcd
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); // gaebcfd
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gabefcd
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // gabecfd
																}
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); // gabcefd
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v2, v1)) {
															if (compare(v4, v2)) {
																swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // agecbfd
															}
															else {
																swap(v1, v6); swap(v3, v4); swap(v4, v6); // agcebfd
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // agebfcd
															}
															else {
																swap(v1, v6); swap(v2, v4); swap(v3, v6); // agebcfd
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // agbefcd
																}
																else {
																	swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // agbecfd
																}
															}
															else {
																if (compare(v2, v1)) {
																	swap(v1, v6); swap(v3, v6); // agcbefd
																}
																else {
																	swap(v1, v6); swap(v2, v6); swap(v3, v6); // agbcefd
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // abgefcd
																}
																else {
																	swap(v2, v6); swap(v3, v4); swap(v4, v6); // abgecfd
																}
															}
															else {
																swap(v2, v6); swap(v3, v6); // abgcefd
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v0)) {
												if (compare(v3, v1)) {
													if (compare(v5, v1)) {
														if (compare(v4, v2)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // geacdfb
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); // gaecdfb
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v6); swap(v3, v4); // gacedfb
															}
															else {
																swap(v0, v6); swap(v1, v6); // gacdefb
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // geacdbf
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // gaecdbf
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v5, v6); // gacedbf
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v5, v6); // gacdebf
																}
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v4, v6); swap(v5, v6); // gacdbef
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v4, v1)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // geacbdf
															}
															else {
																if (compare(v4, v2)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gaecbdf
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gacebdf
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v5, v6); // gacbedf
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gacbdef
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v4, v0)) {
																swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // geabcdf
															}
															else {
																swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // gaebcdf
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gabecdf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // gabcedf
																}
																else {
																	swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gabcdef
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v4, v2)) {
														if (compare(v3, v1)) {
															if (compare(v5, v1)) {
																swap(v1, v6); swap(v2, v4); swap(v3, v4); // agecdfb
															}
															else {
																swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // agecdbf
															}
														}
														else {
															if (compare(v2, v1)) {
																swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // agecbdf
															}
															else {
																swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // agebcdf
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v4, v3)) {
																swap(v1, v6); swap(v3, v4); // agcedfb
															}
															else {
																swap(v1, v6); // agcdefb
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v3)) {
																	swap(v1, v6); swap(v3, v4); swap(v5, v6); // agcedbf
																}
																else {
																	swap(v1, v6); swap(v5, v6); // agcdebf
																}
															}
															else {
																swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // agcebdf
															}
														}
													}
												}
												else {
													if (compare(v4, v3)) {
														if (compare(v6, v1)) {
															if (compare(v2, v1)) {
																swap(v1, v6); swap(v3, v6); swap(v5, v6); // agcbedf
															}
															else {
																if (compare(v4, v2)) {
																	swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // agbecdf
																}
																else {
																	swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // agbcedf
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // abgecdf
															}
															else {
																swap(v2, v6); swap(v3, v6); swap(v5, v6); // abgcedf
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v3, v1)) {
																swap(v1, v6); swap(v4, v6); swap(v5, v6); // agcdbef
															}
															else {
																swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // agcbdef
															}
														}
														else {
															if (compare(v6, v1)) {
																swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // agbcdef
															}
															else {
																swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // abgcdef
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v2, v1)) {
											if (compare(v3, v1)) {
												if (compare(v5, v2)) {
													if (compare(v6, v5)) {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // egfacdb
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); // egafcdb
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // eagfcdb
															}
															else {
																swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // aegfcdb
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v6, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // efgacdb
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // efagcdb
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // eafgcdb
															}
															else {
																swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // aefgcdb
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v0)) {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // egacfdb
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); // egacdfb
																}
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // egacdbf
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // eagcfdb
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // eagcdfb
																}
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // eagcdbf
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // aegcfdb
															}
															else {
																swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // aegcdfb
															}
														}
														else {
															swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // aegcdbf
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v6, v5)) {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // egfacbd
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // egafcbd
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // eagfcbd
															}
															else {
																swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // aegfcbd
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v6, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // efgacbd
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); // efagcbd
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // eafgcbd
															}
															else {
																swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // aefgcbd
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v0)) {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // egacfbd
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // egacbfd
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // egacbdf
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // eagcfbd
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); // eagcbfd
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // eagcbdf
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // aegcfbd
														}
														else {
															if (compare(v5, v3)) {
																swap(v1, v4); swap(v2, v6); swap(v3, v6); // aegcbfd
															}
															else {
																swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // aegcbdf
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v1)) {
												if (compare(v5, v1)) {
													if (compare(v6, v5)) {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // egfabcd
															}
															else {
																swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // egafbcd
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // eagfbcd
															}
															else {
																swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // aegfbcd
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v6, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // efgabcd
															}
															else {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // efagbcd
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); // eafgbcd
															}
															else {
																swap(v1, v4); swap(v2, v5); swap(v3, v6); // aefgbcd
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v0)) {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // egabfcd
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); // egabcfd
																}
																else {
																	swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // egabcdf
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // eagbfcd
															}
															else {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // eagbcfd
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // eagbcdf
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // aegbfcd
														}
														else {
															if (compare(v5, v3)) {
																swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // aegbcfd
															}
															else {
																swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // aegbcdf
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v0)) {
													if (compare(v6, v5)) {
														if (compare(v5, v2)) {
															swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // eabgfcd
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); // eabgcfd
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // eabgcdf
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // efabgcd
														}
														else {
															if (compare(v5, v1)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // eafbgcd
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // eabfgcd
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v2)) {
															if (compare(v5, v1)) {
																swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // aefbgcd
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // aebgfcd
																}
																else {
																	swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // aebfgcd
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v1, v4); swap(v2, v4); swap(v3, v6); // aebgcfd
															}
															else {
																swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // aebgcdf
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v5)) {
																swap(v2, v4); swap(v3, v6); swap(v4, v5); // abegfcd
															}
															else {
																swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // abefgcd
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v2, v4); swap(v3, v6); // abegcfd
															}
															else {
																swap(v2, v4); swap(v3, v6); swap(v5, v6); // abegcdf
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v3)) {
										if (compare(v6, v1)) {
											if (compare(v3, v1)) {
												if (compare(v4, v2)) {
													if (compare(v4, v0)) {
														if (compare(v6, v5)) {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // eacgfdb
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v6); // eacgdfb
																}
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // eacgdbf
															}
														}
														else {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // efacgdb
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // eafcgdb
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v6); // eacfgdb
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // aefcgdb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // aecgfdb
																}
																else {
																	swap(v1, v4); swap(v3, v5); swap(v4, v6); // aecfgdb
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v1, v4); swap(v3, v6); swap(v4, v6); // aecgdfb
															}
															else {
																swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // aecgdbf
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // acgefdb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // acegfdb
																}
																else {
																	swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // acefgdb
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v1, v2); swap(v2, v6); swap(v3, v4); // acgedfb
																}
																else {
																	swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // acegdfb
																}
															}
															else {
																swap(v1, v2); swap(v2, v6); // acgdefb
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v4)) {
																swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // acgedbf
															}
															else {
																swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // acegdbf
															}
														}
														else {
															if (compare(v4, v1)) {
																swap(v1, v2); swap(v2, v6); swap(v5, v6); // acgdebf
															}
															else {
																swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // acgdbef
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // efacgbd
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // eafcgbd
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v5); // eacgfbd
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // eacfgbd
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // aefcgbd
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v4); swap(v3, v6); swap(v4, v5); // aecgfbd
																}
																else {
																	swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // aecfgbd
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // acgefbd
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // acegfbd
																}
																else {
																	swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // acefgbd
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v4, v2)) {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v3, v6); // eacgbfd
															}
															else {
																swap(v1, v4); swap(v3, v6); // aecgbfd
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v4)) {
																	swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // acgebfd
																}
																else {
																	swap(v1, v2); swap(v2, v4); swap(v3, v6); // acegbfd
																}
															}
															else {
																swap(v1, v2); swap(v2, v6); swap(v3, v6); // acgbefd
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // acgebdf
															}
															else {
																if (compare(v4, v3)) {
																	swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // acgbedf
																}
																else {
																	swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // acgbdef
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v5, v6); // eacgbdf
															}
															else {
																if (compare(v4, v2)) {
																	swap(v1, v4); swap(v3, v6); swap(v5, v6); // aecgbdf
																}
																else {
																	swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // acegbdf
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v4, v2)) {
													if (compare(v4, v0)) {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // efacbgd
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // eafcbgd
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v5, v6); // eacfbgd
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v6); // eacbgfd
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // eacbfgd
																}
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // eacbgdf
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // aefcbgd
															}
															else {
																swap(v1, v4); swap(v3, v5); swap(v5, v6); // aecfbgd
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v1, v4); swap(v3, v4); swap(v4, v6); // aecbgfd
																}
																else {
																	swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // aecbfgd
																}
															}
															else {
																swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // aecbgdf
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v6, v5)) {
															if (compare(v5, v3)) {
																swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // acebgfd
															}
															else {
																swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // acebgdf
															}
														}
														else {
															if (compare(v5, v1)) {
																swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // acefbgd
															}
															else {
																swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // acebfgd
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																swap(v1, v2); swap(v3, v6); // acbgefd
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v3, v4); swap(v4, v6); // acbegfd
																}
																else {
																	swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // acbefgd
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v1, v2); swap(v3, v6); swap(v5, v6); // acbgedf
																}
																else {
																	swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // acbegdf
																}
															}
															else {
																swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // acbgdef
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // efabcgd
															}
															else {
																if (compare(v5, v1)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // eafbcgd
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // eabfcgd
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // eabcgfd
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // eabcfgd
																}
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // eabcgdf
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v1)) {
																swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // aefbcgd
															}
															else {
																swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // aebfcgd
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // aebcgfd
																}
																else {
																	swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // aebcfgd
																}
															}
															else {
																swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // aebcgdf
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v6, v5)) {
															if (compare(v5, v3)) {
																swap(v2, v4); swap(v3, v4); swap(v4, v6); // abecgfd
															}
															else {
																swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // abecgdf
															}
														}
														else {
															if (compare(v5, v2)) {
																swap(v2, v4); swap(v3, v5); swap(v5, v6); // abefcgd
															}
															else {
																swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // abecfgd
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																swap(v3, v6); // abcgefd
															}
															else {
																if (compare(v6, v5)) {
																	swap(v3, v4); swap(v4, v6); // abcegfd
																}
																else {
																	swap(v3, v4); swap(v4, v5); swap(v5, v6); // abcefgd
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	swap(v3, v6); swap(v5, v6); // abcgedf
																}
																else {
																	swap(v3, v4); swap(v4, v6); swap(v5, v6); // abcegdf
																}
															}
															else {
																swap(v3, v6); swap(v4, v6); swap(v5, v6); // abcgdef
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v3, v1)) {
											if (compare(v6, v1)) {
												if (compare(v4, v2)) {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // efacdgb
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // eafcdgb
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // eacfdgb
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v4); swap(v4, v6); // eacdgfb
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v4, v5); swap(v5, v6); // eacdfgb
																}
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v4, v6); swap(v5, v6); // eacdgbf
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // aefcdgb
															}
															else {
																swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // aecfdgb
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	swap(v1, v4); swap(v4, v6); // aecdgfb
																}
																else {
																	swap(v1, v4); swap(v4, v5); swap(v5, v6); // aecdfgb
																}
															}
															else {
																swap(v1, v4); swap(v4, v6); swap(v5, v6); // aecdgbf
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // acefdgb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v2, v4); swap(v4, v6); // acedgfb
																}
																else {
																	swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // acedfgb
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v1, v2); swap(v2, v3); swap(v3, v6); // acdgefb
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // acdegfb
																}
																else {
																	swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // acdefgb
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v1)) {
																swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // acdgebf
															}
															else {
																swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // acdgbef
															}
														}
														else {
															if (compare(v4, v3)) {
																swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // acedgbf
															}
															else {
																swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // acdegbf
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); // efacdbg
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // eafcdbg
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v5); // eacfdbg
															}
															else {
																swap(v0, v4); swap(v1, v4); swap(v4, v5); // eacdfbg
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // aefcdbg
															}
															else {
																if (compare(v5, v3)) {
																	swap(v1, v4); swap(v3, v5); swap(v4, v5); // aecfdbg
																}
																else {
																	swap(v1, v4); swap(v4, v5); // aecdfbg
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // acefdbg
																}
																else {
																	swap(v1, v2); swap(v2, v4); swap(v4, v5); // acedfbg
																}
															}
															else {
																swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // acdefbg
															}
														}
													}
												}
												else {
													if (compare(v4, v3)) {
														if (compare(v6, v5)) {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); swap(v5, v6); // eacdbgf
															}
															else {
																if (compare(v4, v2)) {
																	swap(v1, v4); swap(v5, v6); // aecdbgf
																}
																else {
																	swap(v1, v2); swap(v2, v4); swap(v5, v6); // acedbgf
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																swap(v0, v4); swap(v1, v4); // eacdbfg
															}
															else {
																if (compare(v4, v2)) {
																	swap(v1, v4); // aecdbfg
																}
																else {
																	swap(v1, v2); swap(v2, v4); // acedbfg
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v5)) {
																swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // acdebgf
															}
															else {
																swap(v1, v2); swap(v2, v3); swap(v3, v4); // acdebfg
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // acdbgef
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v2, v3); swap(v5, v6); // acdbegf
																}
																else {
																	swap(v1, v2); swap(v2, v3); // acdbefg
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v2, v1)) {
												if (compare(v4, v2)) {
													if (compare(v4, v0)) {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // efacbdg
															}
															else {
																if (compare(v5, v2)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); // eafcbdg
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v3, v5); // eacfbdg
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v5); // eacbfdg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v5, v6); // eacbdgf
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v3, v4); // eacbdfg
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																swap(v1, v4); swap(v2, v5); swap(v3, v5); // aefcbdg
															}
															else {
																swap(v1, v4); swap(v3, v5); // aecfbdg
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v1, v4); swap(v3, v4); swap(v4, v5); // aecbfdg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v4); swap(v3, v4); swap(v5, v6); // aecbdgf
																}
																else {
																	swap(v1, v4); swap(v3, v4); // aecbdfg
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v5, v3)) {
															if (compare(v5, v1)) {
																swap(v1, v2); swap(v2, v4); swap(v3, v5); // acefbdg
															}
															else {
																swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // acebfdg
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // acebdgf
															}
															else {
																swap(v1, v2); swap(v2, v4); swap(v3, v4); // acebdfg
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v1, v2); swap(v3, v4); swap(v4, v5); // acbefdg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v3, v4); swap(v5, v6); // acbedgf
																}
																else {
																	swap(v1, v2); swap(v3, v4); // acbedfg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v1, v2); swap(v4, v6); swap(v5, v6); // acbdgef
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v2); swap(v5, v6); // acbdegf
																}
																else {
																	swap(v1, v2); // acbdefg
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v1)) {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); // efabcdg
															}
															else {
																if (compare(v5, v1)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // eafbcdg
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); // eabfcdg
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // eabcfdg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // eabcdgf
																}
																else {
																	swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); // eabcdfg
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v1)) {
																swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // aefbcdg
															}
															else {
																swap(v1, v4); swap(v2, v4); swap(v3, v5); // aebfcdg
															}
														}
														else {
															if (compare(v5, v3)) {
																swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // aebcfdg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // aebcdgf
																}
																else {
																	swap(v1, v4); swap(v2, v4); swap(v3, v4); // aebcdfg
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																swap(v2, v4); swap(v3, v5); // abefcdg
															}
															else {
																swap(v2, v4); swap(v3, v4); swap(v4, v5); // abecfdg
															}
														}
														else {
															if (compare(v6, v5)) {
																swap(v2, v4); swap(v3, v4); swap(v5, v6); // abecdgf
															}
															else {
																swap(v2, v4); swap(v3, v4); // abecdfg
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v3)) {
																swap(v3, v4); swap(v4, v5); // abcefdg
															}
															else {
																if (compare(v6, v5)) {
																	swap(v3, v4); swap(v5, v6); // abcedgf
																}
																else {
																	swap(v3, v4); // abcedfg
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																swap(v4, v6); swap(v5, v6); // abcdgef
															}
															else {
																if (compare(v6, v5)) {
																	swap(v5, v6); // abcdegf
																}
																else {
																	// abcdefg
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}

			static void stable_sort8(T &v0, T &v1, T &v2, T &v3, T &v4, T &v5, T &v6, T &v7, const Compare &compare)
			{
				using std::swap;
				if (compare(v1, v0)) {
					if (compare(v3, v2)) {
						if (compare(v2, v0)) {
							if (compare(v5, v4)) {
								if (compare(v7, v6)) {
									if (compare(v6, v4)) {
										if (compare(v4, v0)) {
											if (compare(v6, v1)) {
												if (compare(v2, v1)) {
													if (compare(v7, v3)) {
														if (compare(v5, v3)) {
															if (compare(v6, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); // hgfedcba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // hfgedcba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fhgedcba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); // hgfdecba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v5, v6); // hfgdecba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v5, v6); swap(v6, v7); // fhgdecba
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v5); // hgfdceba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // hfgdceba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhgdceba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // hgfdcbea
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v6); // hfgdcbea
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); // fhgdcbea
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // hfdgecba
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hfdgceba
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // hfdcgeba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // hfdgcbea
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v6); // hfdcgbea
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fhdgecba
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhdgceba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhdcgeba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fhdgcbea
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); // fhdcgbea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); // hgdfecba
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // hdgfecba
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); // hdfgecba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // hgdfceba
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hdgfceba
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // hdfgceba
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hdfcgeba
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v5); // hgdcfeba
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // hdgcfeba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v5); // hdcgfeba
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hdcfgeba
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hgdfcbea
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // hdgfcbea
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hdfgcbea
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // hdfcgbea
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // hgdcfbea
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hdgcfbea
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hdcgfbea
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v6); // hdcfgbea
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v6); // hgdcbfea
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // hdgcbfea
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v6); // hdcgbfea
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dhgfecba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v6, v7); // dhfgecba
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); // fdhgecba
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dfhgecba
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhgfceba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dhgcfeba
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fdhgceba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dhfgceba
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfhgceba
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdhcgeba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhfcgeba
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dfhcgeba
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdchgeba
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dfchgeba
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dhcgfeba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhcfgeba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dchgfeba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dchfgeba
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dcfhgeba
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dhgfcbea
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhfgcbea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdhgcbea
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dfhgcbea
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fdhcgbea
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dhfcgbea
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); // dfhcgbea
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fdchgbea
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v6); // dfchgbea
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhgcfbea
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhcgfbea
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dhcfgbea
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dhgcbfea
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dhcgbfea
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dchgfbea
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dchgbfea
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dchfgbea
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); // dcfhgbea
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // hgfedbca
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v5, v6); // hgfdebca
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // hgfdbeca
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v6); // hgfdbcea
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hgfebdca
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // hgfbedca
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hgfbdeca
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // hgfbdcea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v3)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // hgdfebca
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); // hdgfebca
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); // dhgfebca
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hgdfbeca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // hdgfbeca
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dhgfbeca
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // hgdfbcea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hdgfbcea
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhgfbcea
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // hgdbfeca
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v5); // hdgbfeca
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); // dhgbfeca
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hgdbfcea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // hdgbfcea
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dhgbfcea
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // hgdbcfea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v6); // hdgbcfea
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); // dhgbcfea
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); // hgbfedca
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // hgbfdeca
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v5); // hgbdfeca
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hgbfdcea
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // hgbdfcea
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v6); // hgbdcfea
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v6, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); // hfgedbca
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); // hfgdebca
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v5); // hfgdbeca
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // hfgdbcea
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // hfgebdca
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); // hfgbedca
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // hfgbdeca
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hfgbdcea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v3)) {
																		swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); // hfdgebca
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // hdfgebca
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dhfgebca
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // hfdgbeca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hdfgbeca
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhfgbeca
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hfdgbcea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // hdfgbcea
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dhfgbcea
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); // fhgedbca
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v5, v7); // fhgdebca
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); // fhgdbeca
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhgdbcea
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fhgebdca
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); // fhgbedca
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fhgbdeca
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhgbdcea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); // fhdgebca
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fdhgebca
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v6, v7); // dfhgebca
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fhdgbeca
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdhgbeca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dfhgbeca
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhdgbcea
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fdhgbcea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfhgbcea
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v1)) {
													if (compare(v5, v1)) {
														if (compare(v7, v5)) {
															if (compare(v6, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hfdbgeca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // hdfbgeca
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dhfbgeca
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // hfdbgcea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v6); // hdfbgcea
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); // dhfbgcea
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // hfbgedca
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hfbgdeca
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // hfbdgeca
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // hfbgdcea
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v6); // hfbdgcea
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // hfdcbgea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hdfcbgea
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhfcbgea
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hdcfbgea
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhcfbgea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dchfbgea
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hfdbcgea
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // hfbdcgea
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // hdfbcgea
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dhfbcgea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhdbgeca
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdhbgeca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfhbgeca
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fhdbgcea
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); // fdhbgcea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dfhbgcea
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fhbgedca
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhbgdeca
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhbdgeca
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fhbgdcea
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); // fhbdgcea
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhdcbgea
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdhcbgea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dfhcbgea
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdchbgea
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dfchbgea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dcfhbgea
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhdbcgea
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhbdcgea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdhbcgea
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfhbcgea
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v3, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // hdbgfeca
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hdbgfcea
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // hdbgcfea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dhbgfeca
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhbgfcea
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dhbgcfea
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v4, v6); // hdcbgfea
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v6); // dhcbgfea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); // dchbgfea
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v6); // hdbcgfea
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); // dhbcgfea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); // hbgfedca
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); // hbgfdeca
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v4, v5); // hbgdfeca
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); // hbdgfeca
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hbgfdcea
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); // hbgdfcea
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hbdgfcea
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v4, v6); // hbgdcfea
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); // hbdgcfea
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v4, v6); // hbdcgfea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hdbfgeca
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // hdbfgcea
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hdbfcgea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhbfgeca
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dhbfgcea
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhbfcgea
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v4, v5); swap(v5, v6); // hdcbfgea
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dhcbfgea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dchbfgea
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // hdbcfgea
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dhbcfgea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); // hbfgedca
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hbfgdeca
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); // hbfdgeca
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hbdfgeca
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); // hbfgdcea
																			}
																			else {
																				swap(v0, v7); swap(v2, v5); swap(v4, v6); // hbfdgcea
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); // hbdfgcea
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); // hbfdcgea
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hbdfcgea
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); // hbdcfgea
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v6, v5)) {
															if (compare(v5, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); // bhgfedca
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); // bhgfdeca
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v5); // bhgdfeca
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dbhgfeca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); // bhdgfeca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bdhgfeca
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bhgfdcea
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); // bhgdfcea
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbhgfcea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bhdgfcea
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdhgfcea
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dbhgcfea
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); // dbhcgfea
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v6); // dcbhgfea
																		}
																		else {
																			swap(v0, v3); swap(v3, v7); swap(v4, v6); // dbchgfea
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v6); // bhgdcfea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); // bhdgcfea
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bdhgcfea
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v6); // bhdcgfea
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); // bdhcgfea
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v6); // bdchgfea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); // bhfgedca
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bhfgdeca
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); // bhfdgeca
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbhfgeca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bhdfgeca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdhfgeca
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); // bhfgdcea
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v6); // bhfdgcea
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dbhfgcea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); // bhdfgcea
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bdhfgcea
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbhfcgea
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dbhcfgea
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dcbhfgea
																		}
																		else {
																			swap(v0, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dbchfgea
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); // bhfdcgea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bhdfcgea
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdhfcgea
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); // bhdcfgea
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bdhcfgea
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bdchfgea
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v3, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdbhgeca
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fdbhgcea
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdbhcgea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dfbhgeca
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); // dfbhgcea
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dfbhcgea
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdcbhgea
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfcbhgea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcfbhgea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdbchgea
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfbchgea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fbhgedca
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbhgdeca
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbhdgeca
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbdhgeca
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fbhgdcea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); // fbhdgcea
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fbdhgcea
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbhdcgea
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbdhcgea
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbdchgea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bfhgedca
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfhgdeca
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfhdgeca
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dbfhgeca
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfdhgeca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bdfhgeca
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bfhgdcea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); // bfhdgcea
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bfdhgcea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); // dbfhgcea
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); // bdfhgcea
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dbfhcgea
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbfchgea
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcbfhgea
																		}
																		else {
																			swap(v0, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbcfhgea
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfhdcgea
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfdhcgea
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bdfhcgea
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfdchgea
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdfchgea
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdcfhgea
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v5, v3)) {
													if (compare(v3, v1)) {
														if (compare(v2, v1)) {
															if (compare(v6, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgfdcbae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v6, v7); // hfgdcbae
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhgdcbae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hfdgcbae
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhdgcbae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fdhgcbae
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v6, v7); // hfdcgbae
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhdcgbae
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdhcgbae
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdchgbae
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfdcbgae
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); // fhdcbgae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fdhcbgae
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fdchbgae
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v4, v7); swap(v5, v6); // fdcbhgae
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); // hfdcbage
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v7); // fhdcbage
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); // fdhcbage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v7); // fdchbage
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v4, v7); // fdcbhage
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v4, v5); swap(v5, v7); // fdcbahge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v6, v7); // hgfdbcae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfgdbcae
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); // fhgdbcae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfdgbcae
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fhdgbcae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdhgbcae
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hfdbgcae
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhdbgcae
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdhbgcae
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdbhgcae
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfdbcgae
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fhdbcgae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); // fdhbcgae
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fdbhcgae
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); // fdbchgae
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hfdbcage
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); // fhdbcage
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v7); // fdhbcage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); // fdbhcage
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v7); // fdbchage
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); // fdbcahge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v7, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hgfbdcae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfgbdcae
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fhgbdcae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hfbgdcae
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhbgdcae
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v6, v7); // hfbdgcae
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfbdcgae
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); // hfbdcage
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhbdgcae
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); // fhbdcgae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v7); // fhbdcage
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbhgdcae
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbhdgcae
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbdhgcae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); // fbhdcgae
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fbdhcgae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); // fbdchgae
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v4, v7); // fbhdcage
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); // fbdhcage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v2, v3); swap(v4, v7); // fbdchage
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); // fbdcahge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgbfdcae
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbgfdcae
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hbfgdcae
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v4, v6); swap(v6, v7); // hbfdgcae
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhgfdcae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bhfgdcae
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfhgdcae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v6, v7); // bhfdgcae
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfhdgcae
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfdhgcae
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbfdcgae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhfdcgae
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); // bfhdcgae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bfdhcgae
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); // bfdchgae
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); // hbfdcage
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); // bhfdcage
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v7); // bfhdcage
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); // bfdhcage
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v7); // bfdchage
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); // bfdcahge
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v2, v1)) {
															if (compare(v6, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgdfcbae
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hdgfcbae
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdfgcbae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhgfcbae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dhfgcbae
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); // dfhgcbae
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hdfcgbae
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhfcgbae
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dfhcgbae
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dfchgbae
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdfcbgae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dhfcbgae
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfhcbgae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfchbgae
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dfcbhgae
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hdfcbage
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); // dhfcbage
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dfhcbage
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dfchbage
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v7); // dfcbhage
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dfcbahge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hgdfbcae
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdgfbcae
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hdfgbcae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dhgfbcae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); // dhfgbcae
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dfhgbcae
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v6, v7); // hdfbgcae
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dhfbgcae
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfhbgcae
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dfbhgcae
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdfbcgae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhfbcgae
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dfhbcgae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfbhcgae
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dfbchgae
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); // hdfbcage
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dhfbcage
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); // dfhbcage
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dfbhcage
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); // dfbchage
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dfbcahge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v7, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgdbfcae
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdgbfcae
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdbgfcae
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hdbfgcae
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdbfcgae
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hdbfcage
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhgbfcae
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dhbgfcae
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhbfgcae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dhbfcgae
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); // dhbfcage
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dbhgfcae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbhfgcae
																			}
																			else {
																				swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dbfhgcae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dbhfcgae
																			}
																			else {
																				swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbfhcgae
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dbfchgae
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); // dbhfcage
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dbfhcage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); // dbfchage
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dbfcahge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgbdfcae
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbgdfcae
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbdgfcae
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hbdfgcae
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhgdfcae
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhdgfcae
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bhdfgcae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bdhgfcae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdhfgcae
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bdfhgcae
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbdfcgae
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhdfcgae
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bdhfcgae
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdfhcgae
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bdfchgae
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hbdfcage
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bhdfcage
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); // bdhfcage
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bdfhcage
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); // bdfchage
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bdfcahge
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v5)) {
													if (compare(v7, v3)) {
														if (compare(v5, v0)) {
															if (compare(v6, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgdcfbae
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v6, v7); // hgdcbfae
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hgdbcfae
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v6, v7); // hgbdcfae
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdgcfbae
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdcgfbae
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hdgcbfae
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hdcgbfae
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v6, v7); // hdgbcfae
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hdbgcfae
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v4, v6); swap(v6, v7); // hbgdcfae
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hbdgcfae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v6, v7); // bhgdcfae
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bhdgcfae
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v4, v6); swap(v6, v7); // hdcbgfae
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v6, v7); // hdbcgfae
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v3); swap(v4, v6); swap(v6, v7); // hbdcgfae
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v6, v7); // bhdcgfae
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgdcbafe
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdgcbafe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdcgbafe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdcbgafe
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v4, v7); swap(v5, v6); // hdcbagfe
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgdbcafe
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdgbcafe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdbgcafe
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdbcgafe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); // hdbcagfe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgbdcafe
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbgdcafe
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbdgcafe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbdcgafe
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); // hbdcagfe
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhgdcafe
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhdgcafe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhdcgafe
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); // bhdcagfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dhgcfbae
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dhcgfbae
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dchgfbae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); // dhgcbfae
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dhgbcfae
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v7); // dhcgbfae
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); // dchgbfae
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dhcbgfae
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dchbgfae
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); // dhbgcfae
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dhbcgfae
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dhgcbafe
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dhcgbafe
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dchgbafe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhcbgafe
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dchbgafe
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dhcbagfe
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dchbagfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhgbcafe
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dhbgcafe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhbcgafe
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dhbcagfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); // dbhgcfae
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dbhcgfae
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dbhgcafe
																			}
																			else {
																				swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbhcgafe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dbhcagfe
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); // bdhgcfae
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bdhcgfae
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bdhgcafe
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdhcgafe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bdhcagfe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dcbhgfae
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dbchgfae
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bdchgfae
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcbhgafe
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbchgafe
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdchgafe
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dcbhagfe
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dbchagfe
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bdchagfe
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v4, v7); swap(v5, v6); // dcbahgfe
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v4, v7); swap(v5, v6); // dbcahgfe
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v4, v7); swap(v5, v6); // bdcahgfe
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v5, v1)) {
															if (compare(v6, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hdcfgbae
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhcfgbae
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dchfgbae
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dcfhgbae
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdcfbgae
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dhcfbgae
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dchfbgae
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcfhbgae
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dcfbhgae
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hdcfbage
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v7); // dhcfbage
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); // dchfbage
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dcfhbage
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); // dcfbhage
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dcfbahge
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdcbfgae
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhcbfgae
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dchbfgae
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v4, v5); swap(v5, v7); // hdcbfage
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dhcbfage
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dchbfage
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v4, v7); // hdcbafge
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v7); // dhcbafge
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); // dchbafge
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcbhfgae
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dcbhfage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v7); // dcbhafge
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v4, v7); // dcbahfge
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dcbfhgae
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v7); // dcbfhage
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dcbfahge
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v4, v5); swap(v5, v7); // dcbafhge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v7, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdbcfgae
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhbcfgae
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbhcfgae
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); // hdbcfage
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dhbcfage
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dbhcfage
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v7); // hdbcafge
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); // dhbcafge
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); // dbhcafge
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbchfgae
																		}
																		else {
																			swap(v0, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dbchfage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v3, v7); swap(v4, v7); // dbchafge
																		}
																		else {
																			swap(v0, v3); swap(v4, v7); // dbcahfge
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dbcfhgae
																		}
																		else {
																			swap(v0, v3); swap(v3, v5); swap(v4, v7); // dbcfhage
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dbcfahge
																		}
																		else {
																			swap(v0, v3); swap(v4, v5); swap(v5, v7); // dbcafhge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbdcfgae
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhdcfgae
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdhcfgae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); // hbdcfage
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); // bhdcfage
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bdhcfage
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		swap(v0, v7); swap(v2, v3); swap(v4, v7); // hbdcafge
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v7); // bhdcafge
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); // bdhcafge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdchfgae
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bdchfage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v7); // bdchafge
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v4, v7); // bdcahfge
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bdcfhgae
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v7); // bdcfhage
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bdcfahge
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v4, v5); swap(v5, v7); // bdcafhge
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v0)) {
											if (compare(v4, v1)) {
												if (compare(v2, v1)) {
													if (compare(v5, v3)) {
														if (compare(v7, v3)) {
															if (compare(v4, v3)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hfegdcba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhegdcba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fehgdcba
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // hfedgcba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhedgcba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); // fehdgcba
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v5, v6); // hfedcgba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v5, v6); swap(v6, v7); // fhedcgba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fehdcgba
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); // hfedcbga
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v5, v7); // fhedcbga
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); // fehdcbga
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hfdegcba
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // hfdecgba
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v5, v6); // hfdcegba
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); // hfdecbga
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); // hfdcebga
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhdegcba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fhdecgba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v5, v6); swap(v6, v7); // fhdcegba
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); // fhdecbga
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v5, v7); // fhdcebga
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fedhgcba
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fdhegcba
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdehgcba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdhecgba
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fdhcegba
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fedhcgba
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fdehcgba
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fedchgba
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdechgba
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fdchegba
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdcehgba
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fedhcbga
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fdhecbga
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); // fdehcbga
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); // fedchbga
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fdechbga
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v5, v7); // fedcbhga
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); // fdecbhga
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); // fdhcebga
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v5, v7); // fdchebga
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fdcehbga
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v5, v7); // fdcebhga
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // hdfegcba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dhfegcba
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfhegcba
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dfehgcba
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hdfecgba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhfecgba
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dfhecgba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); // dfehcgba
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dfechgba
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // hdfcegba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dhfcegba
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); // dfhcegba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v5, v6); // dfchegba
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dfcehgba
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v5, v6); // hdcfegba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dhcfegba
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dchfegba
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); // dcfhegba
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dcfehgba
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // hdfecbga
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dhfecbga
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); // dfhecbga
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); // dfehcbga
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); // hdfcebga
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); // dhfcebga
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); // dfhcebga
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); // hdcfebga
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v5, v7); // dhcfebga
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); // dfechbga
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v7); // dfchebga
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v7); // dfcehbga
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dfecbhga
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dfcebhga
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); // dchfebga
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); // dcfhebga
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); // dcfehbga
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dcfebhga
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v4)) {
														if (compare(v7, v5)) {
															if (compare(v4, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // hfegdbca
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v6); // hfedgbca
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // hfedbgca
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v5); // hfedbcga
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hfegbdca
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // hfebgdca
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hfebdgca
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // hfebdcga
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v5, v3)) {
																		swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // hfdegbca
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hdfegbca
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhfegbca
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hfdebgca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // hdfebgca
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dhfebgca
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // hfdebcga
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); // hdfebcga
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); // dhfebcga
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fhegdbca
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); // fhedgbca
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhedbgca
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); // fhedbcga
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhegbdca
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fhebgdca
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhebdgca
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fhebdcga
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fhdegbca
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdhegbca
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dfhegbca
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhdebgca
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fdhebgca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfhebgca
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fhdebcga
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); // fdhebcga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dfhebcga
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v7, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fehgdbca
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fehdgbca
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v5, v6); swap(v6, v7); // fehdbgca
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v5, v7); // fehdbcga
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fehgbdca
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fehbgdca
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fehbdgca
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); // fehbdcga
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v3)) {
																		swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fedhgbca
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fdehgbca
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); // dfehgbca
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fedhbgca
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdehbgca
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dfehbgca
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); // fedhbcga
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fdehbcga
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); // dfehbcga
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fedbhgca
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdebhgca
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfebhgca
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fedbhcga
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); // fdebhcga
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dfebhcga
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); // fedbchga
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v5, v7); // fdebchga
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); // dfebchga
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v3)) {
																		swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // febhgdca
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // febhdgca
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // febdhgca
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // febhdcga
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); // febdhcga
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v5, v7); // febdchga
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v7, v1)) {
														if (compare(v7, v5)) {
															if (compare(v4, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // hfdbegca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v5, v6); // hdfbegca
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); // dhfbegca
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); // hfdbecga
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); // hdfbecga
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); // dhfbecga
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hfbegdca
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // hfbedgca
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v5, v6); // hfbdegca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); // hfbedcga
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); // hfbdecga
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v5); // hfdcbega
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // hdfcbega
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dhfcbega
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v5); // hdcfbega
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dhcfbega
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dchfbega
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // hfdbcega
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v5); // hfbdcega
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v5); // hdfbcega
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); // dhfbcega
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fhdbegca
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v5, v6); swap(v6, v7); // fdhbegca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dfhbegca
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); // fhdbecga
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v5, v7); // fdhbecga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); // dfhbecga
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhbegdca
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fhbedgca
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v5, v6); swap(v6, v7); // fhbdegca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); // fhbedcga
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v5, v7); // fhbdecga
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); // fhdcbega
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fdhcbega
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); // dfhcbega
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fdchbega
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v5); // dfchbega
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); // dcfhbega
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fhdbcega
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); // fhbdcega
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); // fdhbcega
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dfhbcega
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v4)) {
															if (compare(v3, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fdbhegca
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); // fdbhecga
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fdbhcega
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); // dfbhegca
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); // dfbhecga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); // dfbhcega
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v4, v7); swap(v5, v7); // fdcbhega
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dfcbhega
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dcfbhega
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); // fdbchega
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dfbchega
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbhegdca
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fbhedgca
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v5, v6); swap(v6, v7); // fbhdegca
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fbdhegca
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); // fbhedcga
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v5, v7); // fbhdecga
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); // fbdhecga
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); // fbhdcega
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fbdhcega
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); // fbdchega
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdbehgca
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fdbehcga
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); // fdbechga
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dfbehgca
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); // dfbehcga
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dfbechga
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v5, v7); // fdcbehga
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v5, v7); // dfcbehga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); // dcfbehga
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v5, v7); // fdbcehga
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); // dfbcehga
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fbehgdca
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbehdgca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbedhgca
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbdehgca
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fbehdcga
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); // fbedhcga
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fbdehcga
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v5, v7); // fbedchga
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); // fbdechga
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v5, v7); // fbdcehga
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v7, v1)) {
															if (compare(v3, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // hdbfegca
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); // hdbfecga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // hdbfcega
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dhbfegca
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); // dhbfecga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dhbfcega
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v4, v5); // hdcbfega
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v5); // dhcbfega
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); // dchbfega
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v5); // hdbcfega
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); // dhbcfega
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hbfegdca
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); // hbfedgca
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v5, v6); // hbfdegca
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); // hbdfegca
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v4); // hbfedcga
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v5); // hbfdecga
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v5); // hbdfecga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v5); swap(v4, v5); // hbfdcega
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); // hbdfcega
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v4, v5); // hbdcfega
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bhfegdca
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); // bhfedgca
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v5, v6); // bhfdegca
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dbhfegca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); // bhdfegca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bdhfegca
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); // bhfedcga
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); // bhfdecga
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); // dbhfecga
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); // bhdfecga
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); // bdhfecga
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dbhfcega
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); // dbhcfega
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v5); // dcbhfega
																		}
																		else {
																			swap(v0, v3); swap(v3, v7); swap(v4, v5); // dbchfega
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v5); // bhfdcega
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bhdfcega
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bdhfcega
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v5); // bhdcfega
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); // bdhcfega
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v5); // bdchfega
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v4)) {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfhegdca
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bfhedgca
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v5, v6); swap(v6, v7); // bfhdegca
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); // dbfhegca
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bfdhegca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); // bdfhegca
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); // bfhedcga
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v5, v7); // bfhdecga
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); // bfdhecga
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); // dbfhecga
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); // bdfhecga
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); // dbfhcega
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dbfchega
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dcbfhega
																		}
																		else {
																			swap(v0, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dbcfhega
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); // bfhdcega
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bfdhcega
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); // bdfhcega
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); // bfdchega
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bdfchega
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bdcfhega
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bfehgdca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfehdgca
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfedhgca
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dbfehgca
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfdehgca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bdfehgca
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bfehdcga
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); // bfedhcga
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); // dbfehcga
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bfdehcga
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); // bdfehcga
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dbfechga
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); // dbfcehga
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v5, v7); // dcbfehga
																		}
																		else {
																			swap(v0, v3); swap(v3, v5); swap(v5, v7); // dbcfehga
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v5, v7); // bfedchga
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); // bfdechga
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bdfechga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v5, v7); // bfdcehga
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); // bdfcehga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v5, v7); // bdcfehga
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v2)) {
												if (compare(v7, v3)) {
													if (compare(v3, v1)) {
														if (compare(v2, v1)) {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v6, v7); // hfedcbag
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v5, v7); swap(v6, v7); // fhedcbag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fehdcbag
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v6, v7); // hfdecbag
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fhdecbag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hdfecbag
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v6, v7); // hfdcebag
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v5, v7); swap(v6, v7); // fhdcebag
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v6, v7); // hdfcebag
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v6, v7); // hdcfebag
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v6, v7); // hfdcbeag
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhdcbeag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hdfcbeag
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hdcfbeag
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v4, v5); swap(v6, v7); // hdcbfeag
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfdcbaeg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v6, v7); // fhdcbaeg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdfcbaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v1)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdcfbaeg
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdcbfaeg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v4, v7); swap(v6, v7); // hdcbafeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v6, v7); // hfedbcag
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhedbcag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v5, v7); swap(v6, v7); // fehdbcag
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hfdebcag
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhdebcag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v6, v7); // hdfebcag
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v6, v7); // hfdbecag
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fhdbecag
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v6, v7); // hdfbecag
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v6, v7); // hdbfecag
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hfdbceag
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhdbceag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v6, v7); // hdfbceag
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hdbfceag
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v6, v7); // hdbcfeag
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfdbcaeg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fhdbcaeg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdfbcaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdbfcaeg
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdbcfaeg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v6, v7); // hdbcafeg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hfebdcag
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhebdcag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fehbdcag
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v6, v7); // hfbedcag
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fhbedcag
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v6, v7); // hfbdecag
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v6, v7); // hfbdceag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfbdcaeg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v5, v7); swap(v6, v7); // fhbdecag
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhbdceag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v6, v7); // fhbdcaeg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v6, v7); // hbfedcag
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v5); swap(v6, v7); // hbfdecag
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v6, v7); // hbdfecag
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v5); swap(v4, v5); swap(v6, v7); // hbfdceag
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hbdfceag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v4, v5); swap(v6, v7); // hbdcfeag
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbfdcaeg
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbdfcaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbdcfaeg
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v4, v7); swap(v6, v7); // hbdcafeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v1)) {
																	if (compare(v7, v4)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fbhedcag
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v5, v7); swap(v6, v7); // fbhdecag
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // febhdcag
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbehdcag
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v6, v7); // bhfedcag
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bfhedcag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfehdcag
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v6, v7); // bhfdecag
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v5, v7); swap(v6, v7); // bfhdecag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v6, v7); // bhdfecag
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbhdceag
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v6, v7); // bhfdceag
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfhdceag
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bhdfceag
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v6, v7); // bhdcfeag
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v7); swap(v4, v7); swap(v6, v7); // fbhdcaeg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhfdcaeg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v6, v7); // bfhdcaeg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhdfcaeg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhdcfaeg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v6, v7); // bhdcafeg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v2, v1)) {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fedhcbag
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdhecbag
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fdehcbag
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhfecbag
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dfhecbag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v6, v7); // dfehcbag
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dhfcebag
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dhcfebag
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fdhcebag
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v6, v7); // dfhcebag
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdhcbeag
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhfcbeag
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dfhcbeag
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhcfbeag
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dhcbfeag
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fdhcbaeg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dhfcbaeg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfhcbaeg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v1)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dhcfbaeg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhcbfaeg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dhcbafeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fedhbcag
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fdhebcag
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdehbcag
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dhfebcag
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfhebcag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dfehbcag
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v6, v7); // dhfbecag
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dhbfecag
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v5, v7); swap(v6, v7); // fdhbecag
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dfhbecag
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdhbceag
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dhfbceag
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfhbceag
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhbfceag
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dhbcfeag
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v6, v7); // fdhbcaeg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhfbcaeg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dfhbcaeg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dhbfcaeg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhbcfaeg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dhbcafeg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v4)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fdbhecag
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdbhceag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fdbhcaeg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fedbhcag
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdebhcag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdbehcag
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfebhcag
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v6, v7); // dfbhecag
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dfbehcag
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dfbhceag
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfbhcaeg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dbhfecag
																			}
																			else {
																				swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v6, v7); // dbfhecag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dbfehcag
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbhfceag
																			}
																			else {
																				swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dbfhceag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dbhcfeag
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dbhfcaeg
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbfhcaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbhcfaeg
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dbhcafeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // febdhcag
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbedhcag
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fbdhecag
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbdehcag
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfedhcag
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bfdhecag
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfdehcag
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bdhfecag
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v6, v7); // bdfhecag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bdfehcag
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbdhceag
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfdhceag
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdhfceag
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bdfhceag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bdhcfeag
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bdhfcaeg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdhcfaeg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bdhcafeg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fbdhcaeg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bfdhcaeg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdfhcaeg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v4)) {
													if (compare(v2, v1)) {
														if (compare(v7, v1)) {
															if (compare(v4, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fdchebag
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v6, v7); // dfchebag
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dchfebag
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v6, v7); // dcfhebag
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdchbeag
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dfchbeag
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dchfbeag
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dcfhbeag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dchbfeag
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dchfbaeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dchbfaeg
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dchbafeg
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fdchbaeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfchbaeg
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcfhbaeg
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdcbheag
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfcbheag
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcfbheag
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v4, v7); swap(v6, v7); // fdcbhaeg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dfcbhaeg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dcfbhaeg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdcbaheg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfcbaheg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcfbaheg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dcbhfeag
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcbhfaeg
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dcbhafeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v4, v7); swap(v6, v7); // dcbahfeg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcbfheag
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dcbfhaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcbfaheg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcbafheg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdbcheag
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfbcheag
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbfcheag
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v6, v7); // fdbchaeg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dfbchaeg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dbfchaeg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdbcaheg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfbcaheg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbfcaheg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dbchfeag
																		}
																		else {
																			swap(v0, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbchfaeg
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dbchafeg
																		}
																		else {
																			swap(v0, v3); swap(v4, v7); swap(v6, v7); // dbcahfeg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbcfheag
																		}
																		else {
																			swap(v0, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dbcfhaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbcfaheg
																		}
																		else {
																			swap(v0, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbcafheg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		swap(v0, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbdcheag
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfdcheag
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdfcheag
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v3); swap(v4, v7); swap(v6, v7); // fbdchaeg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v6, v7); // bfdchaeg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bdfchaeg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbdcaheg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfdcaheg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdfcaheg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bdchfeag
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdchfaeg
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bdchafeg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v4, v7); swap(v6, v7); // bdcahfeg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdcfheag
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bdcfhaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdcfaheg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdcafheg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v7, v0)) {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fedchbag
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v5, v7); swap(v6, v7); // fedcbhag
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fedbchag
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v5, v7); swap(v6, v7); // febdchag
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdechbag
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdcehbag
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fdecbhag
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fdcebhag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v5, v7); swap(v6, v7); // fdebchag
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fdbechag
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v5, v7); swap(v6, v7); // fbedchag
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fbdechag
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v5, v7); swap(v6, v7); // bfedchag
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bfdechag
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v5, v7); swap(v6, v7); // fdcbehag
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v5, v7); swap(v6, v7); // fdbcehag
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v3); swap(v5, v7); swap(v6, v7); // fbdcehag
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v5, v7); swap(v6, v7); // bfdcehag
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v6, v7); // fedcbahg
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v6, v7); // fdecbahg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v6, v7); // fdcebahg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v1, v3); swap(v6, v7); // fdcbeahg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v4, v5); swap(v6, v7); // fdcbaehg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v6, v7); // fedbcahg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v6, v7); // fdebcahg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v6, v7); // fdbecahg
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v6, v7); // fdbceahg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v6, v7); // fdbcaehg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v6, v7); // febdcahg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v6, v7); // fbedcahg
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v6, v7); // fbdecahg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v2, v3); swap(v6, v7); // fbdceahg
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v4, v5); swap(v6, v7); // fbdcaehg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v6, v7); // bfedcahg
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v6, v7); // bfdecahg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v6, v7); // bfdceahg
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v6, v7); // bfdcaehg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v7, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dfechbag
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dfcehbag
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dcfehbag
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfecbhag
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dfebchag
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfcebhag
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcfebhag
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dfcbehag
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dcfbehag
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfbechag
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dfbcehag
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dfecbahg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dfcebahg
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dcfebahg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v6, v7); // dfcbeahg
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v6, v7); // dcfbeahg
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dfcbaehg
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dcfbaehg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v6, v7); // dfebcahg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dfbecahg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v6, v7); // dfbceahg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dfbcaehg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbfechag
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dbfcehag
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dbfecahg
																			}
																			else {
																				swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v6, v7); // dbfceahg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dbfcaehg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdfechag
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bdfcehag
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bdfecahg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v6, v7); // bdfceahg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bdfcaehg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dcbfehag
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dbcfehag
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bdcfehag
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v6, v7); // dcbfeahg
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v3, v5); swap(v6, v7); // dbcfeahg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v6, v7); // bdcfeahg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dcbfaehg
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dbcfaehg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bdcfaehg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v4, v5); swap(v6, v7); // dcbafehg
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v4, v5); swap(v6, v7); // dbcafehg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v4, v5); swap(v6, v7); // bdcafehg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v7, v4)) {
										if (compare(v4, v0)) {
											if (compare(v7, v1)) {
												if (compare(v2, v1)) {
													if (compare(v6, v3)) {
														if (compare(v5, v3)) {
															if (compare(v7, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v6, v7); // ghfedcba
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gfhedcba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); // fghedcba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v6, v7); // ghfdecba
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v5, v7); swap(v6, v7); // gfhdecba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v5, v7); // fghdecba
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v6, v7); // ghfdceba
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfhdceba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); // fghdceba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghfdcbea
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v6, v7); // gfhdcbea
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fghdcbea
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gfdhecba
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfdhceba
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfdcheba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gfdhcbea
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v6, v7); // gfdchbea
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); // fgdhecba
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fgdhceba
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); // fgdcheba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgdhcbea
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgdchbea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v6, v7); // ghdfecba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gdhfecba
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v6, v7); // gdfhecba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ghdfceba
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdhfceba
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gdfhceba
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdfcheba
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v6, v7); // ghdcfeba
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gdhcfeba
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gdchfeba
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdcfheba
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghdfcbea
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gdhfcbea
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdfhcbea
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gdfchbea
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghdcfbea
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdhcfbea
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdchfbea
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gdcfhbea
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v6, v7); // ghdcbfea
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gdhcbfea
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gdchbfea
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v7, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); // dghfecba
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); // dgfhecba
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fdghecba
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); // dfghecba
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dghfceba
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); // dgfhceba
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); // dghcfeba
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdghceba
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dfghceba
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fdgcheba
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dgfcheba
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfgcheba
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fdcgheba
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfcgheba
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v5); // dgchfeba
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dgcfheba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); // dcghfeba
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dcgfheba
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcfgheba
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dghfcbea
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dgfhcbea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fdghcbea
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); // dfghcbea
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdgchbea
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgfchbea
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dfgchbea
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdcghbea
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dfcghbea
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dghcfbea
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dgchfbea
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgcfhbea
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); // dghcbfea
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v6); // dgchbfea
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dcghfbea
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); // dcghbfea
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcgfhbea
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dcfghbea
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ghfedbca
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v5, v7); swap(v6, v7); // ghfdebca
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghfdbeca
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v6, v7); // ghfdbcea
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghfebdca
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ghfbedca
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghfbdeca
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ghfbdcea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ghdfebca
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v6, v7); // gdhfebca
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dghfebca
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghdfbeca
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gdhfbeca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dghfbeca
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ghdfbcea
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdhfbcea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dghfbcea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ghdbfeca
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v6, v7); // gdhbfeca
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dghbfeca
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghdbfcea
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdhbfcea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dghbfcea
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ghdbcfea
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v6, v7); // gdhbcfea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dghbcfea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v6, v7); // ghbfedca
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ghbfdeca
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v6, v7); // ghbdfeca
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghbfdcea
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghbdfcea
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v6, v7); // ghbdcfea
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v7, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v6, v7); // gfhedbca
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v6, v7); // gfhdebca
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v6, v7); // gfhdbeca
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfhdbcea
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gfhebdca
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v6, v7); // gfhbedca
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gfhbdeca
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfhbdcea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v3)) {
																		swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v6, v7); // gfdhebca
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gdfhebca
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); // dgfhebca
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gfdhbeca
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdfhbeca
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dgfhbeca
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfdhbcea
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gdfhbcea
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); // dgfhbcea
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fghedbca
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v5, v6); swap(v6, v7); // fghdebca
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fghdbeca
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); // fghdbcea
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fghebdca
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fghbedca
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fghbdeca
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fghbdcea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fgdhebca
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); // fdghebca
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); // dfghebca
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgdhbeca
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fdghbeca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); // dfghbeca
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fgdhbcea
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdghbcea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dfghbcea
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v1)) {
													if (compare(v5, v1)) {
														if (compare(v6, v5)) {
															if (compare(v7, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfdbheca
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdfbheca
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgfbheca
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gfdbhcea
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v6, v7); // gdfbhcea
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dgfbhcea
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gfbhedca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfbhdeca
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfbdheca
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gfbhdcea
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v6, v7); // gfbdhcea
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfdcbhea
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdfcbhea
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dgfcbhea
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdcfbhea
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dgcfbhea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dcgfbhea
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfdbchea
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfbdchea
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdfbchea
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgfbchea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fgdbheca
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); // fdgbheca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dfgbheca
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgdbhcea
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdgbhcea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfgbhcea
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); // fgbhedca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fgbhdeca
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); // fgbdheca
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgbhdcea
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgbdhcea
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); // fgdcbhea
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fdgcbhea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfgcbhea
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fdcgbhea
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfcgbhea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcfgbhea
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fgdbchea
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); // fgbdchea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); // fdgbchea
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dfgbchea
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v3, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gdbhfeca
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdbhfcea
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gdbhcfea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); // dgbhfeca
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dgbhfcea
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); // dgbhcfea
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v4, v7); swap(v6, v7); // gdcbhfea
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dgcbhfea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dcgbhfea
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v6, v7); // gdbchfea
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dgbchfea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v6, v7); // gbhfedca
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gbhfdeca
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v4, v5); swap(v6, v7); // gbhdfeca
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gbdhfeca
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbhfdcea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbhdfcea
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbdhfcea
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v4, v7); swap(v6, v7); // gbhdcfea
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gbdhcfea
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v4, v7); swap(v6, v7); // gbdchfea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdbfheca
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gdbfhcea
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdbfchea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dgbfheca
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgbfhcea
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dgbfchea
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdcbfhea
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgcbfhea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcgbfhea
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdbcfhea
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgbcfhea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gbfhedca
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbfhdeca
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbfdheca
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbdfheca
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gbfhdcea
																			}
																			else {
																				swap(v0, v6); swap(v2, v5); swap(v4, v7); swap(v6, v7); // gbfdhcea
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gbdfhcea
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbfdchea
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbdfchea
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbdcfhea
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v7, v5)) {
															if (compare(v5, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v6, v7); // bghfedca
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bghfdeca
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v6, v7); // bghdfeca
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); // dbghfeca
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bgdhfeca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); // bdghfeca
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bghfdcea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bghdfcea
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgdhfcea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dbghfcea
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bdghfcea
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); // dbghcfea
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dbgchfea
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dcbghfea
																		}
																		else {
																			swap(v0, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dbcghfea
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v6, v7); // bghdcfea
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bgdhcfea
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); // bdghcfea
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v6, v7); // bgdchfea
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bdgchfea
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bdcghfea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bgfhedca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgfhdeca
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgfdheca
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dbgfheca
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgdfheca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bdgfheca
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bgfhdcea
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v6, v7); // bgfdhcea
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbgfhcea
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bgdfhcea
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdgfhcea
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dbgfchea
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbgcfhea
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcbgfhea
																		}
																		else {
																			swap(v0, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbcgfhea
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgfdchea
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgdfchea
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bdgfchea
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgdcfhea
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdgcfhea
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdcgfhea
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v3, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fdbgheca
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdbghcea
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fdbgchea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfbgheca
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dfbghcea
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfbgchea
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v4, v6); swap(v5, v7); // fdcbghea
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dfcbghea
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dcfbghea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); // fdbcghea
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dfbcghea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); // fbghedca
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fbghdeca
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); // fbgdheca
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fbdgheca
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbghdcea
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbgdhcea
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbdghcea
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); // fbgdchea
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fbdgchea
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); // fbdcghea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); // bfghedca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bfghdeca
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); // bfgdheca
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbfgheca
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bfdgheca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdfgheca
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfghdcea
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfgdhcea
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfdghcea
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dbfghcea
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bdfghcea
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbfgchea
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dbfcghea
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dcbfghea
																		}
																		else {
																			swap(v0, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dbcfghea
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); // bfgdchea
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bfdgchea
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdfgchea
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); // bfdcghea
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bdfcghea
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bdcfghea
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v5, v3)) {
													if (compare(v3, v1)) {
														if (compare(v2, v1)) {
															if (compare(v7, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); // ghfdcbae
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v7); // gfhdcbae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); // fghdcbae
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); // gfdhcbae
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fgdhcbae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdghcbae
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v7); // gfdchbae
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); // fgdchbae
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fdgchbae
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fdcghbae
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); // gfdcbhae
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgdcbhae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdgcbhae
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdcgbhae
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdcbghae
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfdcbahe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v6, v7); // fgdcbahe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fdgcbahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fdcgbahe
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v4, v6); swap(v6, v7); // fdcbgahe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdcbaghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v7); // ghfdbcae
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); // gfhdbcae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fghdbcae
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gfdhbcae
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgdhbcae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fdghbcae
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); // gfdbhcae
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fgdbhcae
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); // fdgbhcae
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fdbghcae
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gfdbchae
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgdbchae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdgbchae
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdbgchae
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdbcghae
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfdbcahe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fgdbcahe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v6, v7); // fdgbcahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fdbgcahe
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v6, v7); // fdbcgahe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdbcaghe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); // ghfbdcae
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gfhbdcae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fghbdcae
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); // gfbhdcae
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fgbhdcae
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v7); // gfbdhcae
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); // gfbdchae
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfbdcahe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); // fgbdhcae
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgbdchae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v6, v7); // fgbdcahe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fbghdcae
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); // fbgdhcae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fbdghcae
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbgdchae
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbdgchae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbdcghae
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v6); swap(v4, v6); swap(v6, v7); // fbgdcahe
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fbdgcahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v2, v3); swap(v4, v6); swap(v6, v7); // fbdcgahe
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbdcaghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ghbfdcae
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gbhfdcae
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); // gbfhdcae
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v4, v7); // gbfdhcae
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bghfdcae
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); // bgfhdcae
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bfghdcae
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v7); // bgfdhcae
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); // bfgdhcae
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bfdghcae
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); // gbfdchae
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); // bgfdchae
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfgdchae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfdgchae
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfdcghae
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbfdcahe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgfdcahe
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v6, v7); // bfgdcahe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bfdgcahe
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v6, v7); // bfdcgahe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfdcaghe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v2, v1)) {
															if (compare(v7, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ghdfcbae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); // gdhfcbae
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gdfhcbae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dghfcbae
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgfhcbae
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dfghcbae
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); // gdfchbae
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dgfchbae
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); // dfgchbae
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v7); // dfcghbae
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gdfcbhae
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgfcbhae
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dfgcbhae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dfcgbhae
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfcbghae
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdfcbahe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dgfcbahe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfgcbahe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfcgbahe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dfcbgahe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfcbaghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); // ghdfbcae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gdhfbcae
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); // gdfhbcae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dghfbcae
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dgfhbcae
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfghbcae
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v7); // gdfbhcae
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); // dgfbhcae
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dfgbhcae
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); // dfbghcae
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); // gdfbchae
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dgfbchae
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfgbchae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dfbgchae
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfbcghae
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdfbcahe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgfbcahe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dfgbcahe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfbgcahe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dfbcgahe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfbcaghe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v6, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ghdbfcae
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); // gdhbfcae
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gdbhfcae
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); // gdbfhcae
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gdbfchae
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdbfcahe
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dghbfcae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgbhfcae
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dgbfhcae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgbfchae
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dgbfcahe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbghfcae
																			}
																			else {
																				swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dbgfhcae
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); // dbfghcae
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbgfchae
																			}
																			else {
																				swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dbfgchae
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbfcghae
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dbgfcahe
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbfgcahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dbfcgahe
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbfcaghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); // ghbdfcae
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); // gbhdfcae
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gbdhfcae
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); // gbdfhcae
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); // bghdfcae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bgdhfcae
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); // bgdfhcae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdghfcae
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bdgfhcae
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); // bdfghcae
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gbdfchae
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bgdfchae
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdgfchae
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bdfgchae
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdfcghae
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbdfcahe
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgdfcahe
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bdgfcahe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdfgcahe
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bdfcgahe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdfcaghe
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v5)) {
													if (compare(v6, v3)) {
														if (compare(v5, v0)) {
															if (compare(v7, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); // ghdcfbae
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v7); // ghdcbfae
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); // ghdbcfae
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v7); // ghbdcfae
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gdhcfbae
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gdchfbae
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); // gdhcbfae
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v7); // gdchbfae
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v7); // gdhbcfae
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); // gdbhcfae
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v4, v7); // gbhdcfae
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); // gbdhcfae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v7); // bghdcfae
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); // bgdhcfae
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v4, v7); // gdcbhfae
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v7); // gdbchfae
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v3); swap(v4, v7); // gbdchfae
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v7); // bgdchfae
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); // ghdcbafe
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gdhcbafe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gdchbafe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v3); swap(v4, v7); swap(v5, v6); // gdcbhafe
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdcbahfe
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ghdbcafe
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); // gdhbcafe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gdbhcafe
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); // gdbchafe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdbcahfe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); // ghbdcafe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); // gbhdcafe
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gbdhcafe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); // gbdchafe
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbdcahfe
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); // bghdcafe
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bgdhcafe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); // bgdchafe
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgdcahfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dghcfbae
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgchfbae
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcghfbae
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dghcbfae
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); // dghbcfae
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dgchbfae
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dcghbfae
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v7); // dgcbhfae
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); // dcgbhfae
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dgbhcfae
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); // dgbchfae
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dghcbafe
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgchbafe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcghbafe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dgcbhafe
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dcgbhafe
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgcbahfe
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcgbahfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dghbcafe
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgbhcafe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dgbchafe
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgbcahfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dbghcfae
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); // dbgchfae
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbghcafe
																			}
																			else {
																				swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dbgchafe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbgcahfe
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bdghcfae
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); // bdgchfae
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdghcafe
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bdgchafe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdgcahfe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v7); // dcbghfae
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v3, v6); swap(v4, v7); // dbcghfae
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v7); // bdcghfae
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dcbghafe
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dbcghafe
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bdcghafe
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcbgahfe
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbcgahfe
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdcgahfe
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcbaghfe
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbcaghfe
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdcaghfe
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v5, v1)) {
															if (compare(v7, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v7); // gdcfhbae
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dgcfhbae
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dcgfhbae
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); // dcfghbae
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gdcfbhae
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgcfbhae
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcgfbhae
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dcfgbhae
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcfbghae
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdcfbahe
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dgcfbahe
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dcgfbahe
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcfgbahe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dcfbgahe
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcfbaghe
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v4, v5); swap(v5, v7); // gdcbfhae
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dgcbfhae
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dcgbfhae
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdcbfahe
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgcbfahe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcgbfahe
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v4, v6); swap(v6, v7); // gdcbafhe
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dgcbafhe
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dcgbafhe
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dcbgfhae
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcbgfahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dcbgafhe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v4, v6); swap(v6, v7); // dcbagfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcbfghae
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dcbfgahe
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcbfaghe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcbafghe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v6, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); // gdbcfhae
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dgbcfhae
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dbgcfhae
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdbcfahe
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgbcfahe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbgcfahe
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v6, v7); // gdbcafhe
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dgbcafhe
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dbgcafhe
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dbcgfhae
																		}
																		else {
																			swap(v0, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbcgfahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dbcgafhe
																		}
																		else {
																			swap(v0, v3); swap(v4, v6); swap(v6, v7); // dbcagfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbcfghae
																		}
																		else {
																			swap(v0, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dbcfgahe
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbcfaghe
																		}
																		else {
																			swap(v0, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbcafghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); // gbdcfhae
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); // bgdcfhae
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bdgcfhae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbdcfahe
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgdcfahe
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdgcfahe
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		swap(v0, v6); swap(v2, v3); swap(v4, v6); swap(v6, v7); // gbdcafhe
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v6, v7); // bgdcafhe
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bdgcafhe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bdcgfhae
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdcgfahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bdcgafhe
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v4, v6); swap(v6, v7); // bdcagfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdcfghae
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bdcfgahe
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdcfaghe
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdcafghe
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v7, v0)) {
											if (compare(v4, v1)) {
												if (compare(v2, v1)) {
													if (compare(v5, v3)) {
														if (compare(v6, v3)) {
															if (compare(v4, v3)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfehdcba
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fgehdcba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // feghdcba
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfedhcba
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); // fgedhcba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fegdhcba
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v5, v7); swap(v6, v7); // gfedchba
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v5, v7); // fgedchba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); // fegdchba
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v6, v7); // gfedcbha
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v5, v6); swap(v6, v7); // fgedcbha
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fegdcbha
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfdehcba
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gfdechba
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v5, v7); swap(v6, v7); // gfdcehba
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v6, v7); // gfdecbha
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v6, v7); // gfdcebha
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fgdehcba
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); // fgdechba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v5, v7); // fgdcehba
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fgdecbha
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v5, v6); swap(v6, v7); // fgdcebha
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fedghcba
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdgehcba
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fdeghcba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fdgechba
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); // fdgcehba
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fedgchba
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); // fdegchba
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); // fedcghba
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fdecghba
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v5, v7); // fdcgehba
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fdceghba
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fedgcbha
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdgecbha
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fdegcbha
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fedcgbha
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdecgbha
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v5, v6); swap(v6, v7); // fedcbgha
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fdecbgha
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fdgcebha
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fdcgebha
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdcegbha
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fdcebgha
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v4, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gdfehcba
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); // dgfehcba
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dfgehcba
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfeghcba
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdfechba
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dgfechba
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfgechba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dfegchba
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfecghba
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gdfcehba
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); // dgfcehba
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dfgcehba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dfcgehba
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfceghba
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gdcfehba
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v5, v7); // dgcfehba
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); // dcgfehba
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dcfgehba
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcfeghba
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gdfecbha
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgfecbha
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dfgecbha
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v6, v7); // dfegcbha
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v6, v7); // gdfcebha
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dgfcebha
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v6, v7); // dfgcebha
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v6, v7); // gdcfebha
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dgcfebha
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dfecgbha
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v6, v7); // dfcgebha
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dfcegbha
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfecbgha
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfcebgha
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dcgfebha
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v6, v7); // dcfgebha
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dcfegbha
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcfebgha
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v6, v5)) {
															if (compare(v4, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gfehdbca
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v6, v7); // gfedhbca
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfedbhca
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v6, v7); // gfedbcha
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfehbdca
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gfebhdca
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfebdhca
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gfebdcha
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v5, v3)) {
																		swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gfdehbca
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdfehbca
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dgfehbca
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfdebhca
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gdfebhca
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgfebhca
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gfdebcha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v6, v7); // gdfebcha
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dgfebcha
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgehdbca
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgedhbca
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); // fgedbhca
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgedbcha
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fgehbdca
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgebhdca
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fgebdhca
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgebdcha
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgdehbca
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fdgehbca
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); // dfgehbca
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fgdebhca
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); // fdgebhca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dfgebhca
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgdebcha
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fdgebcha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfgebcha
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v6, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // feghdbca
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); // fegdhbca
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v5, v7); // fegdbhca
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v5, v6); swap(v6, v7); // fegdbcha
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); // feghbdca
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fegbhdca
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); // fegbdhca
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fegbdcha
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v3)) {
																		swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fedghbca
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdeghbca
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dfeghbca
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); // fedgbhca
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fdegbhca
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfegbhca
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fedgbcha
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdegbcha
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dfegbcha
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fedbghca
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); // fdebghca
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dfebghca
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fedbgcha
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdebgcha
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfebgcha
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fedbcgha
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v5, v6); swap(v6, v7); // fdebcgha
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dfebcgha
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // febghdca
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // febgdhca
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); // febdghca
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // febgdcha
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // febdgcha
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v5, v6); swap(v6, v7); // febdcgha
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v6, v1)) {
														if (compare(v6, v5)) {
															if (compare(v4, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gfdbehca
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v5, v7); swap(v6, v7); // gdfbehca
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dgfbehca
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v6, v7); // gfdbecha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v6, v7); // gdfbecha
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v6, v7); // dgfbecha
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfbehdca
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gfbedhca
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v5, v7); swap(v6, v7); // gfbdehca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v6, v7); // gfbedcha
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v6, v7); // gfbdecha
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v6, v7); // gfdcbeha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gdfcbeha
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgfcbeha
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gdcfbeha
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgcfbeha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcgfbeha
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gfdbceha
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v6, v7); // gfbdceha
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v6, v7); // gdfbceha
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dgfbceha
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); // fgdbehca
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v5, v7); // fdgbehca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); // dfgbehca
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fgdbecha
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v5, v6); swap(v6, v7); // fdgbecha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dfgbecha
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fgbehdca
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); // fgbedhca
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v5, v7); // fgbdehca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fgbedcha
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v5, v6); swap(v6, v7); // fgbdecha
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgdcbeha
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdgcbeha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dfgcbeha
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdcgbeha
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dfcgbeha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dcfgbeha
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgdbceha
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgbdceha
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdgbceha
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfgbceha
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v3, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); // fdbgehca
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fdbgecha
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdbgceha
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dfbgehca
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v6, v7); // dfbgecha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dfbgceha
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdcbgeha
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfcbgeha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcfbgeha
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdbcgeha
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfbcgeha
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fbgehdca
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); // fbgedhca
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v6); swap(v5, v7); // fbgdehca
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); // fbdgehca
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fbgedcha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v5, v6); swap(v6, v7); // fbgdecha
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fbdgecha
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbgdceha
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbdgceha
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbdcgeha
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fdbeghca
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdbegcha
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fdbecgha
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfbeghca
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dfbegcha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfbecgha
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v5, v6); swap(v6, v7); // fdcbegha
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dfcbegha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dcfbegha
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v5, v6); swap(v6, v7); // fdbcegha
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dfbcegha
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbeghdca
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fbegdhca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); // fbedghca
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fbdeghca
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbegdcha
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbedgcha
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbdegcha
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v5, v6); swap(v6, v7); // fbedcgha
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fbdecgha
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v5, v6); swap(v6, v7); // fbdcegha
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v6, v1)) {
															if (compare(v3, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gdbfehca
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v6, v7); // gdbfecha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gdbfceha
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); // dgbfehca
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dgbfecha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgbfceha
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v4, v5); swap(v6, v7); // gdcbfeha
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dgcbfeha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dcgbfeha
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v6, v7); // gdbcfeha
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dgbcfeha
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbfehdca
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gbfedhca
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v5, v7); swap(v6, v7); // gbfdehca
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gbdfehca
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v6, v7); // gbfedcha
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v5); swap(v6, v7); // gbfdecha
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v6, v7); // gbdfecha
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v5); swap(v4, v5); swap(v6, v7); // gbfdceha
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gbdfceha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v4, v5); swap(v6, v7); // gbdcfeha
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgfehdca
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bgfedhca
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v5, v7); swap(v6, v7); // bgfdehca
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); // dbgfehca
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bgdfehca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); // bdgfehca
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v6, v7); // bgfedcha
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v6, v7); // bgfdecha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dbgfecha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v6, v7); // bgdfecha
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bdgfecha
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbgfceha
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dbgcfeha
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dcbgfeha
																		}
																		else {
																			swap(v0, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dbcgfeha
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v6, v7); // bgfdceha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bgdfceha
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdgfceha
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v6, v7); // bgdcfeha
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bdgcfeha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bdcgfeha
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v4, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bfgehdca
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); // bfgedhca
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v5, v7); // bfgdehca
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dbfgehca
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); // bfdgehca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bdfgehca
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bfgedcha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v5, v6); swap(v6, v7); // bfgdecha
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bfdgecha
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v6, v7); // dbfgecha
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v6, v7); // bdfgecha
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dbfgceha
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbfcgeha
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcbfgeha
																		}
																		else {
																			swap(v0, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbcfgeha
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfgdceha
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfdgceha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bdfgceha
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfdcgeha
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdfcgeha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdcfgeha
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfeghdca
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bfegdhca
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); // bfedghca
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbfeghca
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bfdeghca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdfeghca
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfegdcha
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfedgcha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dbfegcha
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfdegcha
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bdfegcha
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbfecgha
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dbfcegha
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dcbfegha
																		}
																		else {
																			swap(v0, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dbcfegha
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v5, v6); swap(v6, v7); // bfedcgha
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bfdecgha
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdfecgha
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v5, v6); swap(v6, v7); // bfdcegha
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bdfcegha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bdcfegha
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v2)) {
												if (compare(v6, v3)) {
													if (compare(v3, v1)) {
														if (compare(v2, v1)) {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); // gfedcbah
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v5, v6); // fgedcbah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // fegdcbah
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); // gfdecbah
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // fgdecbah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // gdfecbah
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); // gfdcebah
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v5, v6); // fgdcebah
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); // gdfcebah
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); // gdcfebah
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v5); // gfdcbeah
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // fgdcbeah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // gdfcbeah
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v5); // gdcfbeah
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v4, v5); // gdcbfeah
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // gfdcbaeh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v6); // fgdcbaeh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gdfcbaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v1)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gdcfbaeh
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v4, v5); swap(v5, v6); // gdcbfaeh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v4, v6); // gdcbafeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v5); // gfedbcah
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // fgedbcah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v5, v6); // fegdbcah
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // gfdebcah
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fgdebcah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); // gdfebcah
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); // gfdbecah
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // fgdbecah
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); // gdfbecah
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); // gdbfecah
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // gfdbceah
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fgdbceah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v5); // gdfbceah
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // gdbfceah
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v5); // gdbcfeah
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gfdbcaeh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // fgdbcaeh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // gdfbcaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gdbfcaeh
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // gdbcfaeh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v6); // gdbcafeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // gfebdcah
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fgebdcah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // fegbdcah
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); // gfbedcah
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // fgbedcah
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); // gfbdecah
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v5); // gfbdceah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // gfbdcaeh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v5, v6); // fgbdecah
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // fgbdceah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v6); // fgbdcaeh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v4); // gbfedcah
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v5); // gbfdecah
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v5); // gbdfecah
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v5); swap(v4, v5); // gbfdceah
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // gbdfceah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v4, v5); // gbdcfeah
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // gbfdcaeh
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gbdfcaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // gbdcfaeh
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v4, v6); // gbdcafeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v4)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // fbgedcah
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v5, v6); // fbgdecah
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // febgdcah
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fbegdcah
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); // bgfedcah
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // bfgedcah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bfegdcah
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); // bgfdecah
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v5, v6); // bfgdecah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); // bgdfecah
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // fbgdceah
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v5); // bgfdceah
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // bfgdceah
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bgdfceah
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v5); // bgdcfeah
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v6); swap(v4, v6); // fbgdcaeh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // bgfdcaeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v6); // bfgdcaeh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bgdfcaeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // bgdcfaeh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v6); // bgdcafeh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v2, v1)) {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fedgcbah
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fdgecbah
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // fdegcbah
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dgfecbah
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // dfgecbah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); // dfegcbah
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // dgfcebah
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v5, v6); // dgcfebah
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // fdgcebah
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); // dfgcebah
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fdgcbeah
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dgfcbeah
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // dfgcbeah
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dgcfbeah
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v5); // dgcbfeah
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // fdgcbaeh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // dgfcbaeh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dfgcbaeh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v1)) {
																				swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v6); // dgcfbaeh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dgcbfaeh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v6); // dgcbafeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // fedgbcah
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // fdgebcah
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fdegbcah
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // dgfebcah
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dfgebcah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // dfegbcah
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); // dgfbecah
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // dgbfecah
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v5, v6); // fdgbecah
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // dfgbecah
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // fdgbceah
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // dgfbceah
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dfgbceah
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dgbfceah
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // dgbcfeah
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v6); // fdgbcaeh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dgfbcaeh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // dfgbcaeh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // dgbfcaeh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dgbcfaeh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // dgbcafeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v4)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // fdbgecah
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fdbgceah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // fdbgcaeh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fedbgcah
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // fdebgcah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fdbegcah
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dfebgcah
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); // dfbgecah
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // dfbegcah
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // dfbgceah
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dfbgcaeh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // dbgfecah
																			}
																			else {
																				swap(v0, v3); swap(v2, v5); swap(v3, v6); // dbfgecah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // dbfegcah
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dbgfceah
																			}
																			else {
																				swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // dbfgceah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // dbgcfeah
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // dbgfcaeh
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dbfgcaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dbgcfaeh
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // dbgcafeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // febdgcah
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // fbedgcah
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // fbdgecah
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fbdegcah
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // bfedgcah
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // bfdgecah
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bfdegcah
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // bdgfecah
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); // bdfgecah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // bdfegcah
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fbdgceah
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bfdgceah
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bdgfceah
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // bdfgceah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // bdgcfeah
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // bdgfcaeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bdgcfaeh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // bdgcafeh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // fbdgcaeh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // bfdgcaeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bdfgcaeh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v4)) {
													if (compare(v2, v1)) {
														if (compare(v6, v1)) {
															if (compare(v4, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v5, v6); // fdcgebah
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v6); // dfcgebah
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // dcgfebah
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); // dcfgebah
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fdcgbeah
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v5); // dfcgbeah
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dcgfbeah
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // dcfgbeah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // dcgbfeah
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // dcgfbaeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dcgbfaeh
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // dcgbafeh
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v6); // fdcgbaeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dfcgbaeh
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dcfgbaeh
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v4, v6); swap(v5, v6); // fdcbgeah
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dfcbgeah
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dcfbgeah
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v4, v6); // fdcbgaeh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v6); // dfcbgaeh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // dcfbgaeh
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v4, v5); swap(v5, v6); // fdcbageh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dfcbageh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dcfbageh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v5); // dcbgfeah
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dcbgfaeh
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v6); // dcbgafeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v4, v6); // dcbagfeh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dcbfgeah
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v6); // dcbfgaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dcbfageh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v4, v5); swap(v5, v6); // dcbafgeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // fdbcgeah
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dfbcgeah
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dbfcgeah
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v6); // fdbcgaeh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // dfbcgaeh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // dbfcgaeh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // fdbcageh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dfbcageh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dbfcageh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v3, v6); swap(v4, v5); // dbcgfeah
																		}
																		else {
																			swap(v0, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dbcgfaeh
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v3, v6); swap(v4, v6); // dbcgafeh
																		}
																		else {
																			swap(v0, v3); swap(v4, v6); // dbcagfeh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dbcfgeah
																		}
																		else {
																			swap(v0, v3); swap(v3, v5); swap(v4, v6); // dbcfgaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dbcfageh
																		}
																		else {
																			swap(v0, v3); swap(v4, v5); swap(v5, v6); // dbcafgeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		swap(v0, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // fbdcgeah
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // bfdcgeah
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bdfcgeah
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v3); swap(v4, v6); // fbdcgaeh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v6); // bfdcgaeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // bdfcgaeh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // fbdcageh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // bfdcageh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bdfcageh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v5); // bdcgfeah
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bdcgfaeh
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v6); // bdcgafeh
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v4, v6); // bdcagfeh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bdcfgeah
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v6); // bdcfgaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bdcfageh
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v4, v5); swap(v5, v6); // bdcafgeh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v6, v0)) {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // fedcgbah
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v5, v6); // fedcbgah
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // fedbcgah
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v5, v6); // febdcgah
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fdecgbah
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fdcegbah
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // fdecbgah
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v5, v6); // fdcebgah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v5, v6); // fdebcgah
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // fdbecgah
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v5, v6); // fbedcgah
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // fbdecgah
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v5, v6); // bfedcgah
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // bfdecgah
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v5, v6); // fdcbegah
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v5, v6); // fdbcegah
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v3); swap(v5, v6); // fbdcegah
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v5, v6); // bfdcegah
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); // fedcbagh
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); // fdecbagh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); // fdcebagh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v1, v3); // fdcbeagh
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v4, v5); // fdcbaegh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); // fedbcagh
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); // fdebcagh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); // fdbecagh
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); // fdbceagh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v4, v5); // fdbcaegh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); // febdcagh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v4); // fbedcagh
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v4); // fbdecagh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v2, v3); // fbdceagh
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v4, v5); // fbdcaegh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); // bfedcagh
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); // bfdecagh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); // bfdceagh
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v4, v5); // bfdcaegh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // dfecgbah
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v6); // dfcegbah
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // dcfegbah
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dfecbgah
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // dfebcgah
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dfcebgah
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dcfebgah
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v5, v6); // dfcbegah
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // dcfbegah
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dfbecgah
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // dfbcegah
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // dfecbagh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v5); // dfcebagh
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dcfebagh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); // dfcbeagh
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); // dcfbeagh
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v5); swap(v4, v5); // dfcbaegh
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dcfbaegh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); // dfebcagh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dfbecagh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); // dfbceagh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dfbcaegh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dbfecgah
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // dbfcegah
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dbfecagh
																			}
																			else {
																				swap(v0, v3); swap(v2, v5); swap(v3, v5); // dbfceagh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dbfcaegh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bdfecgah
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // bdfcegah
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bdfecagh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); // bdfceagh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bdfcaegh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v5, v6); // dcbfegah
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v3, v5); swap(v5, v6); // dbcfegah
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v5, v6); // bdcfegah
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); // dcbfeagh
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v3, v5); // dbcfeagh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v5); // bdcfeagh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v5); swap(v4, v5); // dcbfaegh
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v3, v5); swap(v4, v5); // dbcfaegh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v5); swap(v4, v5); // bdcfaegh
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v4, v5); // dcbafegh
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v4, v5); // dbcafegh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v4, v5); // bdcafegh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v7, v6)) {
									if (compare(v7, v4)) {
										if (compare(v4, v1)) {
											if (compare(v6, v1)) {
												if (compare(v5, v1)) {
													if (compare(v5, v2)) {
														if (compare(v2, v1)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // hgefdcba
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hegfdcba
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // hefgdcba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v6); // hefdgcba
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // hefdcgba
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v5); // hgedfcba
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // hegdfcba
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hedgfcba
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // hedfgcba
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hedfcgba
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // hgdefcba
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hdgefcba
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // hdegfcba
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hdefgcba
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // hdefcgba
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhgefcba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dhegfcba
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhefgcba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dhefcgba
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v6, v4)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hgefdbca
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // hgefbdca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // hgedfbca
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hgdefbca
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // hegfdbca
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hefgdbca
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v5); // hegdfbca
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); // hegfbdca
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // hefgbdca
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // hedgfbca
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // hdgefbca
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hdegfbca
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dhgefbca
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhegfbca
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // hefdgbca
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hedfgbca
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // hdefgbca
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dhefgbca
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v2)) {
																if (compare(v6, v4)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); // hgedcfba
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); // hgdecfba
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // hdgecfba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dhgecfba
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v6); // hegdcfba
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // hedgcfba
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); // hdegcfba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v6, v7); // dhegcfba
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v6); // hedcgfba
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // hdecgfba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dhecgfba
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // hedcfgba
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hdecfgba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhecfgba
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v3)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v3); // hgdcefba
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); // hdgcefba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v6, v7); // dhgcefba
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v6); // hdcgefba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v6, v7); // dhcgefba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v6, v7); // dchgefba
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v6); // hdcegfba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dhcegfba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dchegfba
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hdcefgba
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhcefgba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dchefgba
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v2, v1)) {
															if (compare(v5, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v5, v6); // hgedcbfa
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // hgdecbfa
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hdgecbfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhgecbfa
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v5, v6); // hgdcebfa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // hdgcebfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dhgcebfa
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v5, v6); // hdcgebfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dhcgebfa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dchgebfa
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v5, v6); swap(v6, v7); // hgedcbaf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hgdecbaf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdgecbaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dhgecbaf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v5, v6); swap(v6, v7); // hgdcebaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hdgcebaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); // dhgcebaf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hdcgebaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v5, v7); // dhcgebaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); // dchgebaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v6); // hgedbfca
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // hgedbcfa
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgedbcaf
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hgebfdca
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // hgebdfca
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hgebdcfa
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgebdcaf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // hgdebfca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); // hdgebfca
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); // dhgebfca
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hgdebcfa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // hdgebcfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dhgebcfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgdebcaf
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hdgebcaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhgebcaf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v5, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // hegdcbfa
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hedgcbfa
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // hdegcbfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dhegcbfa
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // hedcgbfa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hdecgbfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhecgbfa
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hdcegbfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhcegbfa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dchegbfa
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hegdcbaf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hedgcbaf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hdegcbaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); // dhegcbaf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hedcgbaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdecgbaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dhecgbaf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdcegbaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dhcegbaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dchegbaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); // hegdbfca
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v5, v6); // hegdbcfa
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v5, v6); swap(v6, v7); // hegdbcaf
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // hegbfdca
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); // hegbdfca
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // hegbdcfa
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hegbdcaf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); // hedgbfca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // hdegbfca
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dhegbfca
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // hedgbcfa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hdegbcfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhegbcfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hedgbcaf
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdegbcaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dhegbcaf
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v6, v0)) {
														if (compare(v7, v3)) {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v5); // hefdcbga
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // hedfcbga
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); // hdefcbga
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v5); // hedcfbga
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // hdecfbga
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v5); // hdcefbga
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v5, v6); // hedcbgfa
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // hdecbgfa
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v5, v6); // hdcebgfa
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); // hedcbfga
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); // hdecbfga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v4); // hdcebfga
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v5, v6); swap(v6, v7); // hedcbgaf
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hdecbgaf
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hdcebgaf
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dhecbgfa
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dhcebgfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dchebgfa
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhecbgaf
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhcebgaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dchebgaf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); // dhefcbga
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dhecfbga
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dhcefbga
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dchefbga
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); // dhecbfga
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v7); // dhcebfga
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); // dchebfga
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v5, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v6, v7); // hefdcbag
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hedfcbag
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v6, v7); // hedcfbag
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v6, v7); // hedcbfag
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v6, v7); // hdefcbag
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hdecfbag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hdcefbag
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v6, v7); // hdecbfag
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v6, v7); // hdcebfag
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v5, v7); // hedcbagf
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); // hdecbagf
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v5, v7); // hdcebagf
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v5, v7); swap(v6, v7); // hedcbafg
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hdecbafg
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hdcebafg
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dhefcbag
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhecfbag
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhcefbag
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dchefbag
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dhecbfag
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dhcebfag
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dchebfag
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dhecbagf
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dhcebagf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dchebagf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhecbafg
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhcebafg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dchebafg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v6, v2)) {
															if (compare(v3, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v5, v6); // hefdbgca
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // hedfbgca
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hdefbgca
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhefbgca
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // hedbgfca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v6); // hdebgfca
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); // dhebgfca
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hedbfgca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // hdebfgca
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dhebfgca
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hebgfdca
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // hebgdfca
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hefbgdca
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // hebfgdca
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // hefbdgca
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hebfdgca
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v6); // hebdgfca
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // hebdfgca
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); // hefdbcga
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); // hedfbcga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); // hefbdcga
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // hdefbcga
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dhefbcga
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // hedbfcga
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v5); // hdebfcga
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); // dhebfcga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // hebfdcga
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v5); // hebdfcga
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v6, v7); // hefdbcag
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v6, v7); // hedfbcag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v6, v7); // hefbdcag
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hdefbcag
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhefbcag
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hedbfcag
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v6, v7); // hdebfcag
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dhebfcag
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hebfdcag
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v6, v7); // hebdfcag
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hedbgcfa
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // hedbcgfa
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); // hedbcfga
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v6, v7); // hedbcfag
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hebgdcfa
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // hebdgcfa
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v5, v6); // hebdcgfa
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); // hebdcfga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v6, v7); // hebdcfag
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // hdebgcfa
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v5, v6); // hdebcgfa
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); // hdebcfga
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v6, v7); // hdebcfag
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dhebgcfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); // dhebcgfa
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); // dhebcfga
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v6, v7); // dhebcfag
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hedbgcaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdebgcaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhebgcaf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hebgdcaf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hebdgcaf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hedbcgaf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v5, v6); swap(v6, v7); // hebdcgaf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v5, v6); swap(v6, v7); // hdebcgaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dhebcgaf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); // hedbcagf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hedbcafg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v5, v7); // hebdcagf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v5, v7); swap(v6, v7); // hebdcafg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v5, v7); // hdebcagf
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v5, v7); swap(v6, v7); // hdebcafg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); // dhebcagf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dhebcafg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v0)) {
												if (compare(v6, v2)) {
													if (compare(v5, v2)) {
														if (compare(v7, v1)) {
															if (compare(v6, v4)) {
																if (compare(v6, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); // hgdbefca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); // hdgbefca
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); // dhgbefca
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // hgbefdca
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); // hgbedfca
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); // hgbdefca
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); // hbgefdca
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); // hbgedfca
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); // hbgdefca
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); // hdbgefca
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v6, v7); // dhbgefca
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v6); // hbdgefca
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hbegfdca
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); // hbefgdca
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hbefdgca
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); // hbegdfca
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v4); swap(v4, v6); // hbedgfca
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); // hbedfgca
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // hdbegfca
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dhbegfca
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); // hbdegfca
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hdbefgca
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhbefgca
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hbdefgca
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); // bhgefdca
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bhegfdca
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); // bhefgdca
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bhefdgca
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); // bhgedfca
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); // bhegdfca
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v6); // bhedgfca
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); // bhedfgca
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); // bhgdefca
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); // bhdgefca
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); // bhdegfca
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bhdefgca
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v6, v7); // dbhgefca
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dbhegfca
																			}
																			else {
																				swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbhefgca
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v6, v7); // bdhgefca
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bdhegfca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdhefgca
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v6, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // hgdcbefa
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hdgcbefa
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhgcbefa
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // hgdbecfa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v5, v6); // hdgbecfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); // dhgbecfa
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hgdbcefa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // hdgbcefa
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dhgbcefa
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // hdbgecfa
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hdbegcfa
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hdbgcefa
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dhbgecfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhbegcfa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhbgcefa
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dbhgecfa
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbhegcfa
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbhgcefa
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // hgbedcfa
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); // hbgedcfa
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hbegdcfa
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); // hbedgcfa
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v5, v6); // hgbdecfa
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v5, v6); // hbgdecfa
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); // hbdgecfa
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hbdegcfa
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // hgbdcefa
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); // hbgdcefa
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hbdgcefa
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); // bhgedcfa
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bhegdcfa
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v5, v6); // bhgdecfa
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); // bhgdcefa
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); // bhdgecfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bhdgcefa
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); // bhedgcfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bhdegcfa
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bdhgecfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdhegcfa
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdhgcefa
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v4, v2)) {
															if (compare(v3, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // hdbefcga
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // hdbecgfa
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); // hdbecfga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hdbefcag
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v6, v7); // hdbecfag
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dhbefcga
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dhbecgfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); // dhbecfga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhbefcag
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dhbecfag
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); // hbefdcga
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v4, v5); // hbedfcga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); // hbdefcga
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hbefdcag
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v4, v5); swap(v6, v7); // hbedfcag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hbdefcag
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v4); swap(v5, v6); // hbedcgfa
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); // hbedcfga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v6, v7); // hbedcfag
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); // hbdecgfa
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v4); // hbdecfga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v6, v7); // hbdecfag
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hdcgbefa
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v3); swap(v4, v6); swap(v5, v6); // hdcbgefa
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v5, v6); // hdcbegfa
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // hdbcgefa
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v5, v6); // hdbcegfa
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); // hbdcgefa
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v5, v6); // hbdcegfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v3); // hdcbefga
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); // hdbcefga
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); // hbdcefga
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v6, v7); // hdcbefag
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v6, v7); // hdbcefag
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v6, v7); // hbdcefag
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v2)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhcgbefa
																			}
																			else {
																				if (compare(v6, v4)) {
																					swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dhcbgefa
																				}
																				else {
																					swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v5, v6); // dhcbegfa
																				}
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dhbcgefa
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); // dhbcegfa
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dchgbefa
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dchbgefa
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); // dchbegfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); // dhcbefga
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); // dchbefga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); // dhbcefga
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v6, v7); // dhcbefag
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v6, v7); // dchbefag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v6, v7); // dhbcefag
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); // bhefdcga
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v5); // bhedfcga
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dbhefcga
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bhdefcga
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bdhefcga
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bhefdcag
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v6, v7); // bhedfcag
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbhefcag
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bhdefcag
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdhefcag
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v5, v6); // bhedcgfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); // bhedcfga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v6, v7); // bhedcfag
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); // bhdecgfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); // bhdecfga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v6, v7); // bhdecfag
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dbhecgfa
																			}
																			else {
																				swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); // dbhecfga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dbhecfag
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bdhecgfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); // bdhecfga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bdhecfag
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v4)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dcbhgefa
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dbhcgefa
																			}
																			else {
																				swap(v0, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dbchgefa
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v5, v6); // dcbhegfa
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); // dbhcegfa
																			}
																			else {
																				swap(v0, v3); swap(v3, v7); swap(v5, v6); // dbchegfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); // dcbhefga
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v7); // dbhcefga
																			}
																			else {
																				swap(v0, v3); swap(v3, v7); // dbchefga
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v6, v7); // dcbhefag
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v6, v7); // dbhcefag
																			}
																			else {
																				swap(v0, v3); swap(v3, v7); swap(v6, v7); // dbchefag
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v6, v4)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); // bhdcgefa
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bdhcgefa
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bdchgefa
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v5, v6); // bhdcegfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); // bdhcegfa
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v5, v6); // bdchegfa
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); // bhdcefga
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); // bdhcefga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); // bdchefga
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v6, v7); // bhdcefag
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v6, v7); // bdhcefag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v6, v7); // bdchefag
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v4)) {
													if (compare(v4, v0)) {
														if (compare(v3, v1)) {
															if (compare(v6, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgdcbeaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdgcbeaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dhgcbeaf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdcgbeaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dhcgbeaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dchgbeaf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hgdbecaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v5, v6); swap(v6, v7); // hdgbecaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dhgbecaf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgdbceaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdgbceaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhgbceaf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hdbgecaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); // dhbgecaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); // dbhgecaf
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdbgceaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dhbgceaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dbhgceaf
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdcbgeaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhcbgeaf
																			}
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dchbgeaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcbhgeaf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdbcgeaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhbcgeaf
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbhcgeaf
																			}
																			else {
																				swap(v0, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbchgeaf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hgbedcaf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v5, v6); swap(v6, v7); // hgbdecaf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hbgedcaf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v5, v6); swap(v6, v7); // hbgdecaf
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hbdgecaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgbdceaf
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbgdceaf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbdgceaf
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbdcgeaf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bhgedcaf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v5, v6); swap(v6, v7); // bhgdecaf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bhdgecaf
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); // bdhgecaf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhgdceaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhdgceaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bdhgceaf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhdcgeaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdhcgeaf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdchgeaf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v2, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); // hgdcbaef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hdgcbaef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhgcbaef
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hdcgbaef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhcgbaef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dchgbaef
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v4, v6); swap(v5, v7); // hdcbgaef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dhcbgaef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dchbgaef
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdcbagef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhcbagef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dchbagef
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hgdbcaef
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); // hgbdcaef
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); // hdgbcaef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dhgbcaef
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hdbgcaef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhbgcaef
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); // hbgdcaef
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hbdgcaef
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); // hdbcgaef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dhbcgaef
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); // hbdcgaef
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdbcagef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhbcagef
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbdcagef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbhgcaef
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dbhcgaef
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dcbhgaef
																		}
																		else {
																			swap(v0, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dbchgaef
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); // bhgdcaef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bhdgcaef
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdhgcaef
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); // bhdcgaef
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bdhcgaef
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bdchgaef
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcbhagef
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbhcagef
																			}
																			else {
																				swap(v0, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbchagef
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcbahgef
																		}
																		else {
																			swap(v0, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbcahgef
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhdcagef
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdhcagef
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdchagef
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdcahgef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v7, v3)) {
															if (compare(v3, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v3); swap(v5, v6); swap(v6, v7); // hdcbegaf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v5, v7); // hdcbeagf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v5, v7); swap(v6, v7); // hdcbeafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v3); swap(v4, v7); swap(v5, v7); // hdcbaegf
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdcbaefg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdbegcaf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hdbecgaf
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); // hdbecagf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hdbecafg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v0)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v5, v6); swap(v6, v7); // hdbcegaf
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v5, v7); // hdbceagf
																				}
																				else {
																					swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v5, v7); swap(v6, v7); // hdbceafg
																				}
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); // hdbcaegf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdbcaefg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbegdcaf
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbedgcaf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbdegcaf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v2, v4); swap(v5, v6); swap(v6, v7); // hbedcgaf
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hbdecgaf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v5, v6); swap(v6, v7); // hbdcegaf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v4); swap(v5, v7); // hbedcagf
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v5, v7); swap(v6, v7); // hbedcafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); // hbdecagf
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hbdecafg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v3); swap(v5, v7); // hbdceagf
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v5, v7); swap(v6, v7); // hbdceafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); // hbdcaegf
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbdcaefg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dhcbegaf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v5, v7); // dhcbeagf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dhcbeafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dhcbaegf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhcbaefg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dchbegaf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); // dchbeagf
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dchbeafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dchbaegf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dchbaefg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dhbegcaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhbecgaf
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dhbecagf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhbecafg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dhbcegaf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); // dhbceagf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dhbceafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dhbcaegf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhbcaefg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v4, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dbhegcaf
																			}
																			else {
																				swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbhecgaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dbhcegaf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dbhecagf
																			}
																			else {
																				swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbhecafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); // dbhceagf
																			}
																			else {
																				swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dbhceafg
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dcbhegaf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v5, v7); // dcbheagf
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dcbheafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dbchegaf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v3, v7); swap(v5, v7); // dbcheagf
																			}
																			else {
																				swap(v0, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dbcheafg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dcbhaegf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dbhcaegf
																			}
																			else {
																				swap(v0, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dbchaegf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v4, v7); swap(v5, v7); // dcbahegf
																		}
																		else {
																			swap(v0, v3); swap(v4, v7); swap(v5, v7); // dbcahegf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcbhaefg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbhcaefg
																			}
																			else {
																				swap(v0, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbchaefg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcbahefg
																		}
																		else {
																			swap(v0, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbcahefg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhegdcaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhedgcaf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhdegcaf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v5, v6); swap(v6, v7); // bhedcgaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bhdecgaf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v5, v6); swap(v6, v7); // bhdcegaf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v5, v7); // bhedcagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v5, v7); swap(v6, v7); // bhedcafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); // bhdecagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bhdecafg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v5, v7); // bhdceagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v5, v7); swap(v6, v7); // bhdceafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); // bhdcaegf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhdcaefg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bdhegcaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdhecgaf
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bdhecagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdhecafg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v0)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bdhcegaf
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); // bdhceagf
																				}
																				else {
																					swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bdhceafg
																				}
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bdhcaegf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdhcaefg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bdchegaf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v5, v7); // bdcheagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bdcheafg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bdchaegf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdchaefg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v4, v7); swap(v5, v7); // bdcahegf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdcahefg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v7, v1)) {
											if (compare(v5, v1)) {
												if (compare(v6, v1)) {
													if (compare(v6, v2)) {
														if (compare(v2, v1)) {
															if (compare(v7, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehgfdcba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ehfgdcba
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efhgdcba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v6, v7); // ehfdgcba
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efhdgcba
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehgdfcba
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehdgfcba
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ehdfgcba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v6, v7); // ehgdcfba
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ehdgcfba
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // edhgfcba
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); // edhgcfba
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efdhgcba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edhfgcba
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // edfhgcba
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dehgfcba
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dehgcfba
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dehfgcba
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); // defhgcba
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ehgfdbca
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehfgdbca
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehfdgbca
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); // ehgfbdca
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ehfgbdca
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // efhgdbca
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efhgbdca
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v6, v7); // efhdgbca
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // efdhgbca
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); // ehgdfbca
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ehdgfbca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edhgfbca
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dehgfbca
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehdfgbca
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // edhfgbca
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edfhgbca
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dehfgbca
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // defhgbca
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v5, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehfdcgba
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehdfcgba
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // edhfcgba
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dehfcgba
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); // efhdcgba
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // efdhcgba
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edfhcgba
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // defhcgba
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v6, v7); // ehdcgfba
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // edhcgfba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); // dehcgfba
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehdcfgba
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edhcfgba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dehcfgba
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // edchgfba
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v6); // dechgfba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); // dcehgfba
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edchfgba
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dechfgba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dcehfgba
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); // efdchgba
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // edfchgba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // defchgba
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // edcfhgba
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // decfhgba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcefhgba
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v2, v1)) {
															if (compare(v6, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); // ehfdcbga
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ehdfcbga
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); // edhfcbga
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dehfcbga
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); // ehdcfbga
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // edhcfbga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); // dehcfbga
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // edchfbga
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v5); // dechfbga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); // dcehfbga
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehfdcbag
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehdfcbag
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // edhfcbag
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dehfcbag
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehdcfbag
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edhcfbag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dehcfbag
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edchfbag
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dechfbag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dcehfbag
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); // ehfdbgca
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v7); // ehfdbcga
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v6, v7); // ehfdbcag
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehfbgdca
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ehfbdgca
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); // ehfbdcga
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ehfbdcag
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ehdfbgca
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edhfbgca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dehfbgca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); // ehdfbcga
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // edhfbcga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); // dehfbcga
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ehdfbcag
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edhfbcag
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dehfbcag
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v6, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v7); // efhdcbga
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); // efdhcbga
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // edfhcbga
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); // defhcbga
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v7); // efdchbga
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); // edfchbga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // defchbga
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v7); // edcfhbga
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // decfhbga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dcefhbga
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v6, v7); // efhdcbag
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // efdhcbag
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edfhcbag
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // defhcbag
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v6, v7); // efdchbag
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // edfchbag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // defchbag
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // edcfhbag
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // decfhbag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcefhbag
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efhdbgca
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); // efhdbcga
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efhdbcag
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // efhbgdca
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efhbdgca
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // efhbdcga
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efhbdcag
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efdhbgca
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // edfhbgca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); // defhbgca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // efdhbcga
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); // edfhbcga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); // defhbcga
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efdhbcag
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // edfhbcag
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v6, v7); // defhbcag
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v6, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v3)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehgdcbfa
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehdgcbfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // edhgcbfa
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehdcgbfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edhcgbfa
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edchgbfa
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); // ehdcbgfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // edhcbgfa
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // edchbgfa
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v7); // ehdcbfga
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); // edhcbfga
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v7); // edchbfga
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v6, v7); // ehdcbfag
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // edhcbfag
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // edchbfag
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dehgcbfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dehcgbfa
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dechgbfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dcehgbfa
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); // dehcbgfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v5, v6); // dechbgfa
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); // dcehbgfa
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); // dehcbfga
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); // dechbfga
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); // dcehbfga
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v6, v7); // dehcbfag
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v6, v7); // dechbfag
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v6, v7); // dcehbfag
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); // ehgdcbaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ehdgcbaf
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); // ehdcgbaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehdcbgaf
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edhgcbaf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // edhcgbaf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // edchgbaf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edhcbgaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edchbgaf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); // ehdcbagf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // edhcbagf
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // edchbagf
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehdcbafg
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edhcbafg
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edchbafg
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dehgcbaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dehcgbaf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dechgbaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcehgbaf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dehcbgaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dechbgaf
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dcehbgaf
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); // dehcbagf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v5, v7); // dechbagf
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); // dcehbagf
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dehcbafg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dechbafg
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dcehbafg
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v2)) {
														if (compare(v5, v2)) {
															if (compare(v3, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v7); // ehgdbfca
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); // ehdgbfca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // edhgbfca
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v6, v7); // dehgbfca
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ehdbgfca
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v6, v7); // edhbgfca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dehbgfca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehdbfgca
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edhbfgca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dehbfgca
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ehgbfdca
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehbgfdca
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ehbfgdca
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehbfdgca
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); // ehgbdfca
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ehbgdfca
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v6, v7); // ehbdgfca
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehbdfgca
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); // ehgdbcfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ehgbdcfa
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ehdgbcfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edhgbcfa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dehgbcfa
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehdbgcfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edhbgcfa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dehbgcfa
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehbgdcfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehbdgcfa
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehgdbcaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehgbdcaf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehdgbcaf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // edhgbcaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); // dehgbcaf
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ehdbgcaf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); // edhbgcaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dehbgcaf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ehbgdcaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); // ehbdgcaf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v7, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ehdbcgfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehdbcgaf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ehdbfcga
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); // ehdbcfga
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ehbfdcga
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); // ehbdfcga
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); // ehbdcgfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v7); // ehbdcfga
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehbdcgaf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); // edhbcgfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edhbcgaf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); // edhbfcga
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v7); // edhbcfga
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dehbcgfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dehbcgaf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dehbfcga
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); // dehbcfga
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehdbfcag
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edhbfcag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dehbfcag
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehbfdcag
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehbdfcag
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ehdbcfag
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v6, v7); // ehbdcfag
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v6, v7); // edhbcfag
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dehbcfag
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ehdbcagf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehdbcafg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); // ehbdcagf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehbdcafg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); // edhbcagf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edhbcafg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dehbcagf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dehbcafg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v0)) {
												if (compare(v5, v2)) {
													if (compare(v6, v2)) {
														if (compare(v4, v1)) {
															if (compare(v7, v5)) {
																if (compare(v7, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebhgfdca
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ebhfgdca
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebhfdgca
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ebhgdfca
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v4, v6); swap(v6, v7); // ebhdgfca
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebhdfgca
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // edbhgfca
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); // debhgfca
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ebdhgfca
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edbhfgca
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // debhfgca
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebdhfgca
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // efdbhgca
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); // edfbhgca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // defbhgca
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efbhgdca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // efbhdgca
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); // efbdhgca
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebfhgdca
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ebfhdgca
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); // ebfdhgca
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // edbfhgca
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // debfhgca
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ebdfhgca
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v6, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // behgfdca
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // behgdfca
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // behfgdca
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // befhgdca
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // behfdgca
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // befhdgca
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v6, v7); // behdgfca
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // behdfgca
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bedhgfca
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bedhfgca
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); // befdhgca
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bedfhgca
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); // dbehgfca
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dbehfgca
																			}
																			else {
																				swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbefhgca
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); // bdehgfca
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bdehfgca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdefhgca
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v5, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); // efdcbhga
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // edfcbhga
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); // defcbhga
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); // efdbhcga
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v7); // edfbhcga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); // defbhcga
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // efdbchga
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); // edfbchga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // defbchga
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // edbhfcga
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); // edbfhcga
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // edbfchga
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); // debhfcga
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // debfhcga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); // debfchga
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); // dbehfcga
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dbefhcga
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); // dbefchga
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); // efbhdcga
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v7); // efbdhcga
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); // efbdchga
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ebhfdcga
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); // ebfhdcga
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v2, v5); swap(v4, v7); // ebfdhcga
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); // ebfdchga
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); // ebhdfcga
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ebdhfcga
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); // ebdfhcga
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ebdfchga
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // behfdcga
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); // befhdcga
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v7); // befdhcga
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); // befdchga
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); // behdfcga
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bedhfcga
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); // bedfhcga
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bedfchga
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); // bdehfcga
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bdefhcga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); // bdefchga
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v7, v2)) {
															if (compare(v3, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edbhgcfa
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // edbhcgfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); // edbhcfga
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // edbhgcaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edbhcgaf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // debhgcfa
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); // debhcgfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); // debhcfga
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // debhgcaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // debhcgaf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebhgdcfa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebhdgcfa
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebdhgcfa
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ebhgdcaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); // ebhdgcaf
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ebdhgcaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); // ebhdcgfa
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v4, v7); // ebhdcfga
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebhdcgaf
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ebdhcgfa
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); // ebdhcfga
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebdhcgaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v4, v7); swap(v5, v6); // edcbhgfa
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); // edbchgfa
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); // ebdchgfa
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edcbhgaf
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edbchgaf
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebdchgaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // edcfbhga
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v4, v7); // edcbhfga
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v4, v5); swap(v5, v7); // edcbfhga
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v7); // edbchfga
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); // edbcfhga
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v4, v7); // ebdchfga
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); // ebdcfhga
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // decbhgfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dcebhgfa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // debchgfa
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // decbhgaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcebhgaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // debchgaf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v1)) {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v5, v7); // decfbhga
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v7); // decbhfga
																				}
																				else {
																					swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // decbfhga
																				}
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); // debchfga
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // debcfhga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); // dcefbhga
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); // dcebhfga
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dcebfhga
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v6, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // behgdcfa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // behdgcfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bedhgcfa
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dbehgcfa
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bdehgcfa
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // behgdcaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); // behdgcaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bedhgcaf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbehgcaf
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdehgcaf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); // behdcgfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v7); // behdcfga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // behdcgaf
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bedhcgfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); // bedhcfga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bedhcgaf
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); // dbehcgfa
																			}
																			else {
																				swap(v0, v3); swap(v2, v4); swap(v3, v7); // dbehcfga
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dbehcgaf
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); // bdehcgfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); // bdehcfga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bdehcgaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dcbehgfa
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dbechgfa
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dbcehgfa
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcbehgaf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbechgaf
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbcehgaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v7); // dcbehfga
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); // dbechfga
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v4, v7); // dbcehfga
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dcbefhga
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dbecfhga
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dbcefhga
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); // bedchgfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bdechgfa
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bdcehgfa
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bedchgaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdechgaf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdcehgaf
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v7); // bedchfga
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); // bdechfga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v7); // bdcehfga
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); // bedcfhga
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bdecfhga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bdcefhga
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v5)) {
													if (compare(v4, v1)) {
														if (compare(v4, v3)) {
															if (compare(v3, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v3); swap(v4, v7); swap(v6, v7); // edcbhfag
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v4, v7); swap(v5, v7); // edcbhagf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edcbhafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v5, v7); // edcbahgf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v5, v7); swap(v6, v7); // edcbahfg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edbhfcag
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // edbhcfag
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // edbhcagf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edbhcafg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v6, v7); // edbchfag
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); // edbchagf
																				}
																				else {
																					swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edbchafg
																				}
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v5, v7); // edbcahgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v5, v7); swap(v6, v7); // edbcahfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebhfdcag
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebhdfcag
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebdhfcag
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v7); swap(v4, v7); swap(v6, v7); // ebhdcfag
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ebdhcfag
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v4, v7); swap(v6, v7); // ebdchfag
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); // ebhdcagf
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebhdcafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ebdhcagf
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebdhcafg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); // ebdchagf
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebdchafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v5, v7); // ebdcahgf
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v5, v7); swap(v6, v7); // ebdcahfg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // decbhfag
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // decbhagf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // decbhafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v5, v7); // decbahgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // decbahfg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dcebhfag
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dcebhagf
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcebhafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); // dcebahgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dcebahfg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // debhfcag
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v6, v7); // debhcfag
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); // debhcagf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // debhcafg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // debchfag
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // debchagf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // debchafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); // debcahgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // debcahfg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v7, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dbehfcag
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v6, v7); // dbehcfag
																			}
																			else {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dbechfag
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); // dbehcagf
																			}
																			else {
																				swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dbehcafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dbechagf
																			}
																			else {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbechafg
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dcbehfag
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dcbehagf
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcbehafg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dbcehfag
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dbcehagf
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbcehafg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v5, v7); // dcbeahgf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); // dbecahgf
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v5, v7); // dbceahgf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v5, v7); // dcbaehgf
																		}
																		else {
																			swap(v0, v3); swap(v5, v7); // dbcaehgf
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dcbeahfg
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dbecahfg
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dbceahfg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v5, v7); swap(v6, v7); // dcbaehfg
																		}
																		else {
																			swap(v0, v3); swap(v5, v7); swap(v6, v7); // dbcaehfg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // behfdcag
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // behdfcag
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bedhfcag
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v6, v7); // behdcfag
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bedhcfag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v6, v7); // bedchfag
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); // behdcagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // behdcafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bedhcagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bedhcafg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); // bedchagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bedchafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v5, v7); // bedcahgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v5, v7); swap(v6, v7); // bedcahfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bdehfcag
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v6, v7); // bdehcfag
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); // bdehcagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bdehcafg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bdechfag
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bdechagf
																				}
																				else {
																					swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdechafg
																				}
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); // bdecahgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bdecahfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bdcehfag
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bdcehagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdcehafg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v5, v7); // bdceahgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bdceahfg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v5, v7); // bdcaehgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v5, v7); swap(v6, v7); // bdcaehfg
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v0)) {
														if (compare(v3, v1)) {
															if (compare(v5, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efdcbhag
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edfcbhag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // defcbhag
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edcfbhag
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // decfbhag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dcefbhag
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // efdbhcag
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v6, v7); // edfbhcag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // defbhcag
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efdbchag
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edfbchag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // defbchag
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // edbfhcag
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // debfhcag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbefhcag
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edbfchag
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // debfchag
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dbefchag
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edcbfhag
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // decbfhag
																			}
																		}
																		else {
																			if (compare(v4, v1)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcebfhag
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcbefhag
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edbcfhag
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // debcfhag
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbecfhag
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbcefhag
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // efbhdcag
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v6, v7); // efbdhcag
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ebfhdcag
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v4, v7); swap(v6, v7); // ebfdhcag
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ebdfhcag
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efbdchag
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebfdchag
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebdfchag
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebdcfhag
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // befhdcag
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v6, v7); // befdhcag
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bedfhcag
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdefhcag
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // befdchag
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bedfchag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bdefchag
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bedcfhag
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdecfhag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdcefhag
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v2, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v6, v7); // efdcbahg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // edfcbahg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v6, v7); // defcbahg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // edcfbahg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v6, v7); // decfbahg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v6, v7); // dcefbahg
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v4, v5); swap(v6, v7); // edcbfahg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // decbfahg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dcebfahg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v6, v7); // edcbafhg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v6, v7); // decbafhg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v6, v7); // dcebafhg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // efdbcahg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v6, v7); // efbdcahg
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v6, v7); // edfbcahg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // defbcahg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // edbfcahg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v6, v7); // debfcahg
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v5); swap(v4, v5); swap(v6, v7); // ebfdcahg
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ebdfcahg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v6, v7); // edbcfahg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // debcfahg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v4, v5); swap(v6, v7); // ebdcfahg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v6, v7); // edbcafhg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v6, v7); // debcafhg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v6, v7); // ebdcafhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v6, v7); // dbefcahg
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dbecfahg
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dcbefahg
																		}
																		else {
																			swap(v0, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dbcefahg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v6, v7); // befdcahg
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bedfcahg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v6, v7); // bdefcahg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v6, v7); // bedcfahg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bdecfahg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bdcefahg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v6, v7); // dcbeafhg
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v6, v7); // dbecafhg
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v6, v7); // dbceafhg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v6, v7); // dcbaefhg
																		}
																		else {
																			swap(v0, v3); swap(v6, v7); // dbcaefhg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v6, v7); // bedcafhg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v6, v7); // bdecafhg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v6, v7); // bdceafhg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v6, v7); // bdcaefhg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v4)) {
										if (compare(v4, v1)) {
											if (compare(v7, v1)) {
												if (compare(v5, v1)) {
													if (compare(v5, v2)) {
														if (compare(v2, v1)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ghefdcba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gehfdcba
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gefhdcba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v6, v7); // gefdhcba
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gefdchba
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v6, v7); // ghedfcba
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gehdfcba
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gedhfcba
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gedfhcba
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gedfchba
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ghdefcba
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdhefcba
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gdehfcba
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdefhcba
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gdefchba
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dghefcba
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); // dgehfcba
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dgefhcba
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); // dgefchba
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v7, v4)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghefdbca
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ghefbdca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghedfbca
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghdefbca
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gehfdbca
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gefhdbca
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v6, v7); // gehdfbca
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v6, v7); // gehfbdca
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gefhbdca
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gedhfbca
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gdhefbca
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdehfbca
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dghefbca
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dgehfbca
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gefdhbca
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gedfhbca
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gdefhbca
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgefhbca
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v7, v2)) {
																if (compare(v7, v4)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v6, v7); // ghedcfba
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v6, v7); // ghdecfba
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gdhecfba
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); // dghecfba
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v6, v7); // gehdcfba
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gedhcfba
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v6, v7); // gdehcfba
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); // dgehcfba
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v6, v7); // gedchfba
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gdechfba
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); // dgechfba
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gedcfhba
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdecfhba
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dgecfhba
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v4)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v6, v7); // ghdcefba
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v6, v7); // gdhcefba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); // dghcefba
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v6, v7); // gdchefba
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v7); // dgchefba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); // dcghefba
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gdcehfba
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v7); // dgcehfba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); // dcgehfba
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdcefhba
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dgcefhba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dcgefhba
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v4)) {
														if (compare(v2, v1)) {
															if (compare(v5, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v5, v7); swap(v6, v7); // ghedcbfa
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ghdecbfa
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdhecbfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dghecbfa
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v5, v7); swap(v6, v7); // ghdcebfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gdhcebfa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); // dghcebfa
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gdchebfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v5, v6); // dgchebfa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); // dcghebfa
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v5, v7); // ghedcbaf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); // ghdecbaf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gdhecbaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dghecbaf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v5, v7); // ghdcebaf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); // gdhcebaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dghcebaf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v5, v7); // gdchebaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dgchebaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dcghebaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v6, v7); // ghedbfca
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghedbcfa
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); // ghedbcaf
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghebfdca
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ghebdfca
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghebdcfa
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ghebdcaf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ghdebfca
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v6, v7); // gdhebfca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dghebfca
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghdebcfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gdhebcfa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dghebcfa
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ghdebcaf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); // gdhebcaf
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dghebcaf
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v5, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gehdcbfa
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gedhcbfa
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gdehcbfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); // dgehcbfa
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gedchbfa
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdechbfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dgechbfa
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdcehbfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dgcehbfa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dcgehbfa
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); // gehdcbaf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gedhcbaf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); // gdehcbaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dgehcbaf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); // gedchbaf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gdechbaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgechbaf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gdcehbaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgcehbaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcgehbaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v6, v7); // gehdbfca
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v5, v7); swap(v6, v7); // gehdbcfa
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v5, v7); // gehdbcaf
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gehbfdca
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v6, v7); // gehbdfca
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gehbdcfa
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); // gehbdcaf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v6, v7); // gedhbfca
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gdehbfca
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); // dgehbfca
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gedhbcfa
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdehbcfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dgehbcfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); // gedhbcaf
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gdehbcaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgehbcaf
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v7, v0)) {
														if (compare(v6, v3)) {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v6, v7); // gefdcbha
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gedfcbha
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v6, v7); // gdefcbha
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v6, v7); // gedcfbha
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gdecfbha
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gdcefbha
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v5, v7); swap(v6, v7); // gedcbhfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gdecbhfa
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gdcebhfa
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v6, v7); // gedcbfha
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v6, v7); // gdecbfha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v6, v7); // gdcebfha
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v5, v7); // gedcbhaf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); // gdecbhaf
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v5, v7); // gdcebhaf
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgecbhfa
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgcebhfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcgebhfa
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dgecbhaf
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dgcebhaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dcgebhaf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dgefcbha
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgecfbha
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgcefbha
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcgefbha
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dgecbfha
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dgcebfha
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dcgebfha
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v5, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v5); // gefdcbah
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // gedfcbah
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v5); // gedcfbah
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); // gedcbfah
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); // gdefcbah
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // gdecfbah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v5); // gdcefbah
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); // gdecbfah
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v4); // gdcebfah
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v5, v6); swap(v6, v7); // gedcbahf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gdecbahf
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gdcebahf
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v5, v6); // gedcbafh
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // gdecbafh
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v5, v6); // gdcebafh
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // dgefcbah
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dgecfbah
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dgcefbah
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dcgefbah
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // dgecbfah
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v6); // dgcebfah
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // dcgebfah
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgecbahf
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgcebahf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcgebahf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dgecbafh
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dgcebafh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dcgebafh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v7, v2)) {
															if (compare(v3, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v5, v7); swap(v6, v7); // gefdbhca
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gedfbhca
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdefbhca
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dgefbhca
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gedbhfca
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v6, v7); // gdebhfca
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dgebhfca
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gedbfhca
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdebfhca
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgebfhca
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gebhfdca
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gebhdfca
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gefbhdca
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gebfhdca
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gefbdhca
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gebfdhca
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v6, v7); // gebdhfca
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gebdfhca
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v6, v7); // gefdbcha
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v6, v7); // gedfbcha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v6, v7); // gefbdcha
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gdefbcha
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgefbcha
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gedbfcha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v6, v7); // gdebfcha
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dgebfcha
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gebfdcha
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v6, v7); // gebdfcha
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); // gefdbcah
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); // gedfbcah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); // gefbdcah
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // gdefbcah
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dgefbcah
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // gedbfcah
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v5); // gdebfcah
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // dgebfcah
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // gebfdcah
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v5); // gebdfcah
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gedbhcfa
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gedbchfa
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v6, v7); // gedbcfha
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); // gedbcfah
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gebhdcfa
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gebdhcfa
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v5, v7); swap(v6, v7); // gebdchfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v6, v7); // gebdcfha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); // gebdcfah
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdebhcfa
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v5, v7); swap(v6, v7); // gdebchfa
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v6, v7); // gdebcfha
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); // gdebcfah
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgebhcfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dgebchfa
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v6, v7); // dgebcfha
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); // dgebcfah
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gedbhcaf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); // gdebhcaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dgebhcaf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gebhdcaf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); // gebdhcaf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); // gedbchaf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v5, v7); // gebdchaf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v5, v7); // gdebchaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); // dgebchaf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gedbcahf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // gedbcafh
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v5, v6); swap(v6, v7); // gebdcahf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v5, v6); // gebdcafh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v5, v6); swap(v6, v7); // gdebcahf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v5, v6); // gdebcafh
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dgebcahf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // dgebcafh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v0)) {
												if (compare(v7, v2)) {
													if (compare(v5, v2)) {
														if (compare(v6, v1)) {
															if (compare(v7, v4)) {
																if (compare(v7, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v6, v7); // ghdbefca
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v6, v7); // gdhbefca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v6, v7); // dghbefca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ghbefdca
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v6, v7); // ghbedfca
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v6, v7); // ghbdefca
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gbhefdca
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v6, v7); // gbhedfca
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v6, v7); // gbhdefca
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v6, v7); // gdbhefca
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); // dgbhefca
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v6, v7); // gbdhefca
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbehfdca
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gbefhdca
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbefdhca
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gbehdfca
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v4); swap(v4, v7); swap(v6, v7); // gbedhfca
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbedfhca
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gdbehfca
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); // dgbehfca
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gbdehfca
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdbefhca
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dgbefhca
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbdefhca
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bghefdca
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgehfdca
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bgefhdca
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgefdhca
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v6, v7); // bghedfca
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bgehdfca
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v6, v7); // bgedhfca
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgedfhca
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v6, v7); // bghdefca
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v6, v7); // bgdhefca
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bgdehfca
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgdefhca
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); // dbghefca
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); // dbgehfca
																			}
																			else {
																				swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dbgefhca
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); // bdghefca
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); // bdgehfca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bdgefhca
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v7, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghdcbefa
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdhcbefa
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dghcbefa
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ghdbecfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v5, v7); swap(v6, v7); // gdhbecfa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dghbecfa
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghdbcefa
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdhbcefa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dghbcefa
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gdbhecfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdbehcfa
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdbhcefa
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); // dgbhecfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dgbehcfa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dgbhcefa
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); // dbghecfa
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dbgehcfa
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dbghcefa
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ghbedcfa
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gbhedcfa
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbehdcfa
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbedhcfa
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v5, v7); swap(v6, v7); // ghbdecfa
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v5, v7); swap(v6, v7); // gbhdecfa
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gbdhecfa
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbdehcfa
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghbdcefa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbhdcefa
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbdhcefa
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bghedcfa
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgehdcfa
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v5, v7); swap(v6, v7); // bghdecfa
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bghdcefa
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bgdhecfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgdhcefa
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgedhcfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgdehcfa
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); // bdghecfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bdgehcfa
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bdghcefa
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v4, v2)) {
															if (compare(v3, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gdbefcha
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gdbechfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v6, v7); // gdbecfha
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // gdbefcah
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); // gdbecfah
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgbefcha
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgbechfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dgbecfha
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dgbefcah
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // dgbecfah
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gbefdcha
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v4, v5); swap(v6, v7); // gbedfcha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gbdefcha
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // gbefdcah
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v4, v5); // gbedfcah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // gbdefcah
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v4); swap(v5, v7); swap(v6, v7); // gbedchfa
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v6, v7); // gbedcfha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); // gbedcfah
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gbdechfa
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v6, v7); // gbdecfha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v4); // gbdecfah
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdchbefa
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdcbhefa
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v5, v7); swap(v6, v7); // gdcbehfa
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdbchefa
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v5, v7); swap(v6, v7); // gdbcehfa
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbdchefa
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v5, v7); swap(v6, v7); // gbdcehfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v6, v7); // gdcbefha
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v6, v7); // gdbcefha
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v6, v7); // gbdcefha
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v3); // gdcbefah
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); // gdbcefah
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); // gbdcefah
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v2)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dgchbefa
																			}
																			else {
																				if (compare(v7, v4)) {
																					swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgcbhefa
																				}
																				else {
																					swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dgcbehfa
																				}
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgbchefa
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dgbcehfa
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dcghbefa
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcgbhefa
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dcgbehfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v6, v7); // dgcbefha
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v6, v7); // dcgbefha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v6, v7); // dgbcefha
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); // dgcbefah
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); // dcgbefah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); // dgbcefah
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bgefdcha
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v6, v7); // bgedfcha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbgefcha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bgdefcha
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdgefcha
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // bgefdcah
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v5); // bgedfcah
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dbgefcah
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bgdefcah
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bdgefcah
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v5, v7); swap(v6, v7); // bgedchfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v6, v7); // bgedcfha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); // bgedcfah
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bgdechfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v6, v7); // bgdecfha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); // bgdecfah
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbgechfa
																			}
																			else {
																				swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dbgecfha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // dbgecfah
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdgechfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bdgecfha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // bdgecfah
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v4)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcbghefa
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbgchefa
																			}
																			else {
																				swap(v0, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbcghefa
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dcbgehfa
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dbgcehfa
																			}
																			else {
																				swap(v0, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dbcgehfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v6, v7); // dcbgefha
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v6, v7); // dbgcefha
																			}
																			else {
																				swap(v0, v3); swap(v3, v6); swap(v6, v7); // dbcgefha
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); // dcbgefah
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v6); // dbgcefah
																			}
																			else {
																				swap(v0, v3); swap(v3, v6); // dbcgefah
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v7, v4)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgdchefa
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdgchefa
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdcghefa
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v5, v7); swap(v6, v7); // bgdcehfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bdgcehfa
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bdcgehfa
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v6, v7); // bgdcefha
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v6, v7); // bdgcefha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v6, v7); // bdcgefha
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); // bgdcefah
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); // bdgcefah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); // bdcgefah
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v4)) {
													if (compare(v4, v0)) {
														if (compare(v3, v1)) {
															if (compare(v7, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); // ghdcbeaf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gdhcbeaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dghcbeaf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gdchbeaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgchbeaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcghbeaf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); // ghdbecaf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v5, v7); // gdhbecaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); // dghbecaf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ghdbceaf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); // gdhbceaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dghbceaf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); // gdbhecaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dgbhecaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dbghecaf
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gdbhceaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgbhceaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbghceaf
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v4, v7); swap(v5, v7); // gdcbheaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dgcbheaf
																			}
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dcgbheaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dcbgheaf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); // gdbcheaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dgbcheaf
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dbgcheaf
																			}
																			else {
																				swap(v0, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dbcgheaf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); // ghbedcaf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v5, v7); // ghbdecaf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); // gbhedcaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v5, v7); // gbhdecaf
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); // gbdhecaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); // ghbdceaf
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); // gbhdceaf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gbdhceaf
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); // gbdcheaf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); // bghedcaf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v5, v7); // bghdecaf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); // bgdhecaf
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bdghecaf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); // bghdceaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bgdhceaf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdghceaf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); // bgdcheaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bdgcheaf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bdcgheaf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v2, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghdcbaef
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdhcbaef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dghcbaef
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdchbaef
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dgchbaef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dcghbaef
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdcbhaef
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgcbhaef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcgbhaef
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v4, v6); swap(v5, v7); // gdcbahef
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dgcbahef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dcgbahef
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghdbcaef
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghbdcaef
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdhbcaef
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dghbcaef
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdbhcaef
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dgbhcaef
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbhdcaef
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbdhcaef
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdbchaef
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgbchaef
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbdchaef
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); // gdbcahef
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dgbcahef
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); // gbdcahef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dbghcaef
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbgchaef
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcbghaef
																		}
																		else {
																			swap(v0, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbcghaef
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bghdcaef
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgdhcaef
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bdghcaef
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgdchaef
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdgchaef
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdcghaef
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dcbgahef
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dbgcahef
																			}
																			else {
																				swap(v0, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dbcgahef
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v4, v6); swap(v5, v7); // dcbaghef
																		}
																		else {
																			swap(v0, v3); swap(v4, v6); swap(v5, v7); // dbcaghef
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); // bgdcahef
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bdgcahef
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bdcgahef
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v4, v6); swap(v5, v7); // bdcaghef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v6, v3)) {
															if (compare(v3, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v3); swap(v5, v7); // gdcbehaf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v5, v6); swap(v6, v7); // gdcbeahf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v5, v6); // gdcbeafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdcbaehf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v4, v6); swap(v5, v6); // gdcbaefh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gdbehcaf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); // gdbechaf
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gdbecahf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // gdbecafh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v0)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v5, v7); // gdbcehaf
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v5, v6); swap(v6, v7); // gdbceahf
																				}
																				else {
																					swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v5, v6); // gdbceafh
																				}
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdbcaehf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // gdbcaefh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gbehdcaf
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); // gbedhcaf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gbdehcaf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v2, v4); swap(v5, v7); // gbedchaf
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); // gbdechaf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v5, v7); // gbdcehaf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v4); swap(v5, v6); swap(v6, v7); // gbedcahf
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v5, v6); // gbedcafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gbdecahf
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // gbdecafh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v3); swap(v5, v6); swap(v6, v7); // gbdceahf
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v5, v6); // gbdceafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbdcaehf
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // gbdcaefh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v5, v7); // dgcbehaf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dgcbeahf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v5, v6); // dgcbeafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgcbaehf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dgcbaefh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); // dcgbehaf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dcgbeahf
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // dcgbeafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcgbaehf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dcgbaefh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgbehcaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dgbechaf
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgbecahf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dgbecafh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); // dgbcehaf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dgbceahf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // dgbceafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgbcaehf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dgbcaefh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v4, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbgehcaf
																			}
																			else {
																				swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dbgechaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); // dbgcehaf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbgecahf
																			}
																			else {
																				swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dbgecafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dbgceahf
																			}
																			else {
																				swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // dbgceafh
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v5, v7); // dcbgehaf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dcbgeahf
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v5, v6); // dcbgeafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v3, v6); swap(v5, v7); // dbcgehaf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dbcgeahf
																			}
																			else {
																				swap(v0, v3); swap(v3, v6); swap(v5, v6); // dbcgeafh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcbgaehf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbgcaehf
																			}
																			else {
																				swap(v0, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbcgaehf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcbagehf
																		}
																		else {
																			swap(v0, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbcagehf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dcbgaefh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dbgcaefh
																			}
																			else {
																				swap(v0, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dbcgaefh
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v4, v6); swap(v5, v6); // dcbagefh
																		}
																		else {
																			swap(v0, v3); swap(v4, v6); swap(v5, v6); // dbcagefh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bgehdcaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); // bgedhcaf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bgdehcaf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v5, v7); // bgedchaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); // bgdechaf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v5, v7); // bgdcehaf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v5, v6); swap(v6, v7); // bgedcahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v5, v6); // bgedcafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bgdecahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // bgdecafh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v5, v6); swap(v6, v7); // bgdceahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v5, v6); // bgdceafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgdcaehf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // bgdcaefh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdgehcaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bdgechaf
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdgecahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bdgecafh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); // bdgcehaf
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bdgceahf
																				}
																				else {
																					swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // bdgceafh
																				}
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdgcaehf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bdgcaefh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v5, v7); // bdcgehaf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bdcgeahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v5, v6); // bdcgeafh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdcgaehf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bdcgaefh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdcagehf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v4, v6); swap(v5, v6); // bdcagefh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v1)) {
											if (compare(v5, v1)) {
												if (compare(v7, v1)) {
													if (compare(v7, v2)) {
														if (compare(v2, v1)) {
															if (compare(v6, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // eghfdcba
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); // egfhdcba
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // efghdcba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v7); // egfdhcba
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); // efgdhcba
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); // eghdfcba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // egdhfcba
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); // egdfhcba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v7); // eghdcfba
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); // egdhcfba
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edghfcba
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // edgfhcba
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // edghcfba
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // efdghcba
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); // edfghcba
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // deghfcba
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); // deghcfba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // degfhcba
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // defghcba
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eghfdbca
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // egfhdbca
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); // egfdhbca
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // eghfbdca
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egfhbdca
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); // efghdbca
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // efghbdca
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v7); // efgdhbca
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); // efdghbca
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eghdfbca
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egdhfbca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // edghfbca
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); // deghfbca
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // egdfhbca
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); // edgfhbca
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // edfghbca
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // degfhbca
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // defghbca
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); // egfdchba
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // egdfchba
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edgfchba
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // degfchba
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efgdchba
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efdgchba
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // edfgchba
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // defgchba
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v7); // egdchfba
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); // edgchfba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // degchfba
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); // egdcfhba
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // edgcfhba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // degcfhba
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v7); // edcghfba
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // decghfba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dceghfba
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // edcgfhba
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // decgfhba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcegfhba
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efdcghba
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edfcghba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // defcghba
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edcfghba
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // decfghba
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dcefghba
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v2, v1)) {
															if (compare(v7, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egfdcbha
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egdfcbha
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // edgfcbha
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // degfcbha
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egdcfbha
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edgcfbha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // degcfbha
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edcgfbha
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // decgfbha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dcegfbha
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // egfdcbah
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // egdfcbah
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // edgfcbah
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // degfcbah
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // egdcfbah
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // edgcfbah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // degcfbah
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // edcgfbah
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v5); // decgfbah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // dcegfbah
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egfdbhca
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v6, v7); // egfdbcha
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v6); // egfdbcah
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // egfbhdca
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egfbdhca
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // egfbdcha
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // egfbdcah
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egdfbhca
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // edgfbhca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); // degfbhca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // egdfbcha
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edgfbcha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // degfbcha
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // egdfbcah
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // edgfbcah
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // degfbcah
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v7, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v6, v7); // efgdcbha
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // efdgcbha
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edfgcbha
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // defgcbha
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v6, v7); // efdcgbha
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // edfcgbha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // defcgbha
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // edcfgbha
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // decfgbha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcefgbha
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v6); // efgdcbah
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // efdgcbah
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // edfgcbah
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // defgcbah
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v6); // efdcgbah
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // edfcgbah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // defcgbah
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v6); // edcfgbah
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // decfgbah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dcefgbah
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); // efgdbhca
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efgdbcha
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // efgdbcah
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); // efgbhdca
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // efgbdhca
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efgbdcha
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // efgbdcah
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // efdgbhca
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edfgbhca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // defgbhca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efdgbcha
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // edfgbcha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v6, v7); // defgbcha
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // efdgbcah
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // edfgbcah
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); // defgbcah
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v7, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); // eghdcbfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // egdhcbfa
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edghcbfa
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); // egdchbfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // edgchbfa
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // edcghbfa
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egdcbhfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edgcbhfa
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edcgbhfa
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v6, v7); // egdcbfha
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // edgcbfha
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // edcgbfha
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v6); // egdcbfah
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // edgcbfah
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v6); // edcgbfah
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // deghcbfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // degchbfa
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // decghbfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dceghbfa
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // degcbhfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // decgbhfa
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dcegbhfa
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v6, v7); // degcbfha
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v6, v7); // decgbfha
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v6, v7); // dcegbfha
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); // degcbfah
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); // decgbfah
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); // dcegbfah
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v7, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eghdcbaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egdhcbaf
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egdchbaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); // egdcbhaf
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // edghcbaf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edgchbaf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edcghbaf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // edgcbhaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // edcgbhaf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egdcbahf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edgcbahf
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edcgbahf
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // egdcbafh
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // edgcbafh
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // edcgbafh
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // deghcbaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // degchbaf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // decghbaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dceghbaf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); // degcbhaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v5, v7); // decgbhaf
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); // dcegbhaf
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // degcbahf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // decgbahf
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dcegbahf
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // degcbafh
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v5, v6); // decgbafh
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // dcegbafh
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v2)) {
														if (compare(v5, v2)) {
															if (compare(v3, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v6, v7); // eghdbfca
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // egdhbfca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); // edghbfca
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); // deghbfca
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); // egdbhfca
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v7); // edgbhfca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); // degbhfca
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // egdbfhca
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); // edgbfhca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // degbfhca
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eghbfdca
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // egbhfdca
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); // egbfhdca
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // egbfdhca
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // eghbdfca
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); // egbhdfca
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v7); // egbdhfca
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); // egbdfhca
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eghdbcfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egdhbcfa
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eghbdcfa
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // edghbcfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); // deghbcfa
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // egdbhcfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); // edgbhcfa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // degbhcfa
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // egbhdcfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); // egbdhcfa
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); // eghdbcaf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // egdhbcaf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // eghbdcaf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edghbcaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // deghbcaf
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egdbhcaf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edgbhcaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // degbhcaf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egbhdcaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egbdhcaf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v0)) {
															if (compare(v6, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egdbchfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // egdbchaf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egdbfcha
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // egdbcfha
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egbfdcha
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egbdfcha
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egbdchfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v6, v7); // egbdcfha
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); // egbdchaf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edgbchfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); // edgbchaf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edgbfcha
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v6, v7); // edgbcfha
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // degbchfa
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // degbchaf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // degbfcha
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // degbcfha
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // egdbfcah
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // edgbfcah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // degbfcah
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // egbfdcah
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // egbdfcah
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // egdbcfah
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v6); // egbdcfah
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v6); // edgbcfah
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // degbcfah
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egdbcahf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // egdbcafh
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egbdcahf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // egbdcafh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edgbcahf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // edgbcafh
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // degbcahf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // degbcafh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v0)) {
												if (compare(v5, v2)) {
													if (compare(v7, v2)) {
														if (compare(v4, v1)) {
															if (compare(v6, v5)) {
																if (compare(v6, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ebghfdca
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); // ebgfhdca
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ebgfdhca
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); // ebghdfca
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v4, v7); // ebgdhfca
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); // ebgdfhca
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); // edbghfca
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // debghfca
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); // ebdghfca
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // edbgfhca
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // debgfhca
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ebdgfhca
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efdbghca
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edfbghca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // defbghca
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // efbghdca
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efbgdhca
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efbdghca
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ebfghdca
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebfgdhca
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebfdghca
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edbfghca
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // debfghca
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebdfghca
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v7, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // beghfdca
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); // beghdfca
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); // begfhdca
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // befghdca
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // begfdhca
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // befgdhca
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v7); // begdhfca
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); // begdfhca
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); // bedghfca
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bedgfhca
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // befdghca
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bedfghca
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dbeghfca
																			}
																			else {
																				swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbegfhca
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dbefghca
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bdeghfca
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdegfhca
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bdefghca
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v5, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efdcbgha
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edfcbgha
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // defcbgha
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // efdbgcha
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v6, v7); // edfbgcha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // defbgcha
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efdbcgha
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edfbcgha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // defbcgha
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edbgfcha
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // edbfgcha
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edbfcgha
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // debgfcha
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // debfgcha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // debfcgha
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dbegfcha
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbefgcha
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dbefcgha
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // efbgdcha
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v6, v7); // efbdgcha
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efbdcgha
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebgfdcha
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ebfgdcha
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v2, v5); swap(v4, v6); swap(v6, v7); // ebfdgcha
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebfdcgha
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebgdfcha
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebdgfcha
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ebdfgcha
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebdfcgha
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // begfdcha
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // befgdcha
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v6, v7); // befdgcha
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // befdcgha
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // begdfcha
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bedgfcha
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bedfgcha
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bedfcgha
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bdegfcha
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdefgcha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bdefcgha
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v6, v2)) {
															if (compare(v3, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // edbghcfa
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edbgchfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // edbgcfha
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edbghcaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // edbgchaf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // debghcfa
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // debgchfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v6, v7); // debgcfha
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // debghcaf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); // debgchaf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ebghdcfa
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); // ebgdhcfa
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ebdghcfa
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebghdcaf
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebgdhcaf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebdghcaf
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebgdchfa
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v4, v6); swap(v6, v7); // ebgdcfha
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); // ebgdchaf
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebdgchfa
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ebdgcfha
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ebdgchaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edcbghfa
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edbcghfa
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebdcghfa
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v4, v6); swap(v5, v7); // edcbghaf
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); // edbcghaf
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); // ebdcghaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edcfbgha
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v4, v6); swap(v6, v7); // edcbgfha
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edcbfgha
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v6, v7); // edbcgfha
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edbcfgha
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v4, v6); swap(v6, v7); // ebdcgfha
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebdcfgha
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // decbghfa
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcebghfa
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // debcghfa
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // decbghaf
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dcebghaf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // debcghaf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v1)) {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // decfbgha
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // decbgfha
																				}
																				else {
																					swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // decbfgha
																				}
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // debcgfha
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // debcfgha
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dcefbgha
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dcebgfha
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcebfgha
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v7, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // beghdcfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); // begdhcfa
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bedghcfa
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbeghcfa
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdeghcfa
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // beghdcaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // begdhcaf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bedghcaf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dbeghcaf
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bdeghcaf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // begdchfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v6, v7); // begdcfha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); // begdchaf
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bedgchfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bedgcfha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bedgchaf
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dbegchfa
																			}
																			else {
																				swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v6, v7); // dbegcfha
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); // dbegchaf
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bdegchfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v6, v7); // bdegcfha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); // bdegchaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcbeghfa
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbecghfa
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbceghfa
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dcbeghaf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dbecghaf
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dbceghaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dcbegfha
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dbecgfha
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dbcegfha
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcbefgha
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbecfgha
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbcefgha
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bedcghfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdecghfa
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdceghfa
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); // bedcghaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bdecghaf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bdceghaf
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v6, v7); // bedcgfha
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bdecgfha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bdcegfha
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bedcfgha
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdecfgha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdcefgha
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v5)) {
													if (compare(v4, v1)) {
														if (compare(v4, v3)) {
															if (compare(v3, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v3); swap(v4, v6); // edcbgfah
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edcbgahf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v4, v6); swap(v5, v6); // edcbgafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v5, v6); swap(v6, v7); // edcbaghf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v5, v6); // edcbagfh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // edbgfcah
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // edbgcfah
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edbgcahf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // edbgcafh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v6); // edbcgfah
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edbcgahf
																				}
																				else {
																					swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // edbcgafh
																				}
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v5, v6); swap(v6, v7); // edbcaghf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v5, v6); // edbcagfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ebgfdcah
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // ebgdfcah
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ebdgfcah
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v2, v6); swap(v4, v6); // ebgdcfah
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // ebdgcfah
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v4, v6); // ebdcgfah
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebgdcahf
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // ebgdcafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebdgcahf
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ebdgcafh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebdcgahf
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // ebdcgafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v5, v6); swap(v6, v7); // ebdcaghf
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v5, v6); // ebdcagfh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v6); // decbgfah
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // decbgahf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // decbgafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // decbaghf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v5, v6); // decbagfh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // dcebgfah
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcebgahf
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dcebgafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dcebaghf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // dcebagfh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // debgfcah
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); // debgcfah
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // debgcahf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // debgcafh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // debcgfah
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // debcgahf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // debcgafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // debcaghf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // debcagfh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v6, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // dbegfcah
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v6); // dbegcfah
																			}
																			else {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // dbecgfah
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dbegcahf
																			}
																			else {
																				swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // dbegcafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbecgahf
																			}
																			else {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dbecgafh
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v6); // dcbegfah
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcbegahf
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dcbegafh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v3, v4); swap(v4, v6); // dbcegfah
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbcegahf
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dbcegafh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dcbeaghf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dbecaghf
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dbceaghf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v5, v6); swap(v6, v7); // dcbaeghf
																		}
																		else {
																			swap(v0, v3); swap(v5, v6); swap(v6, v7); // dbcaeghf
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v5, v6); // dcbeagfh
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // dbecagfh
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v5, v6); // dbceagfh
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v5, v6); // dcbaegfh
																		}
																		else {
																			swap(v0, v3); swap(v5, v6); // dbcaegfh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // begfdcah
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // begdfcah
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bedgfcah
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v6); // begdcfah
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // bedgcfah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v6); // bedcgfah
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // begdcahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // begdcafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bedgcahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bedgcafh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bedcgahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // bedcgafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v5, v6); swap(v6, v7); // bedcaghf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v5, v6); // bedcagfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // bdegfcah
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); // bdegcfah
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bdegcahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // bdegcafh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // bdecgfah
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdecgahf
																				}
																				else {
																					swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bdecgafh
																				}
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bdecaghf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // bdecagfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v6); // bdcegfah
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdcegahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bdcegafh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bdceaghf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v5, v6); // bdceagfh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v5, v6); swap(v6, v7); // bdcaeghf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v5, v6); // bdcaegfh
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v0)) {
														if (compare(v3, v1)) {
															if (compare(v5, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // efdcbgah
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // edfcbgah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // defcbgah
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // edcfbgah
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v5, v6); // decfbgah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // dcefbgah
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // efdbgcah
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v6); // edfbgcah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // defbgcah
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // efdbcgah
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // edfbcgah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // defbcgah
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // edbfgcah
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // debfgcah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dbefgcah
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // edbfcgah
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // debfcgah
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // dbefcgah
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v4, v5); swap(v5, v6); // edcbfgah
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // decbfgah
																			}
																		}
																		else {
																			if (compare(v4, v1)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dcebfgah
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dcbefgah
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // edbcfgah
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // debcfgah
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dbecfgah
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dbcefgah
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // efbgdcah
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v6); // efbdgcah
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // ebfgdcah
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v4, v6); // ebfdgcah
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // ebdfgcah
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // efbdcgah
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // ebfdcgah
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ebdfcgah
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // ebdcfgah
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // befgdcah
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v6); // befdgcah
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // bedfgcah
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bdefgcah
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // befdcgah
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bedfcgah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // bdefcgah
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // bedcfgah
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bdecfgah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bdcefgah
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v2, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v5); // efdcbagh
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // edfcbagh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); // defcbagh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v5); // edcfbagh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); // decfbagh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); // dcefbagh
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v4, v5); // edcbfagh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v4); swap(v4, v5); // decbfagh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // dcebfagh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); // edcbafgh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v4); // decbafgh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v4); // dcebafgh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // efdbcagh
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v5); // efbdcagh
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v5); // edfbcagh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // defbcagh
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // edbfcagh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); // debfcagh
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v5); swap(v4, v5); // ebfdcagh
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // ebdfcagh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v4, v5); // edbcfagh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // debcfagh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v4, v5); // ebdcfagh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); // edbcafgh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v4); // debcafgh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); // ebdcafgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); // dbefcagh
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // dbecfagh
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); swap(v4, v5); // dcbefagh
																		}
																		else {
																			swap(v0, v3); swap(v3, v4); swap(v4, v5); // dbcefagh
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v5); // befdcagh
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bedfcagh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); // bdefcagh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v4, v5); // bedcfagh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // bdecfagh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); swap(v4, v5); // bdcefagh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v3, v4); // dcbeafgh
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v4); // dbecafgh
																			}
																			else {
																				swap(v0, v3); swap(v3, v4); // dbceafgh
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v2); // dcbaefgh
																		}
																		else {
																			swap(v0, v3); // dbcaefgh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); // bedcafgh
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v4); // bdecafgh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v3, v4); // bdceafgh
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); // bdcaefgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
						else {
							if (compare(v5, v4)) {
								if (compare(v7, v6)) {
									if (compare(v6, v4)) {
										if (compare(v6, v0)) {
											if (compare(v6, v3)) {
												if (compare(v6, v5)) {
													if (compare(v6, v1)) {
														if (compare(v5, v1)) {
															if (compare(v4, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hgfedbac
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v5, v6); swap(v6, v7); // hgfdebac
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hgdfebac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgfdbeac
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgdfbeac
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgfebdac
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hgfbedac
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgfbdeac
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hgfebadc
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); // hgfbeadc
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); // hgfdbaec
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hgdfbaec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgfdbace
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgdfbace
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hgfbdaec
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgfbdace
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hgfbaedc
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgfbadec
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hgfbadce
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v6, v7); // hgbfedac
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hgbfdeac
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgbfdaec
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hgbfdace
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v4, v0)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hgbfeadc
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgbfaedc
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hgbfadec
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); // hgbfadce
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hgdbfeac
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgdbfaec
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hgdbface
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v6, v7); // hgbdfeac
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgbdfaec
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); // hgbdface
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hgdbafec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); // hgdbafce
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hgdbacfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v6, v7); // hgbdafec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v7); // hgbdafce
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); // hgbdacfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hgbafedc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgbafdec
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hgbadfec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hgbafdce
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); // hgbadfce
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hgbadcfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v7, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v6, v7); // hbgfedac
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hbgfdeac
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v4, v5); swap(v6, v7); // hbgdfeac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbgfdaec
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbgdfaec
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v4, v7); swap(v6, v7); // hbgdafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hbgfdace
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); // hbgdface
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v2, v6); swap(v4, v7); // hbgdafce
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); // hbgdacfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v6, v7); // bhgfedac
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bhgfdeac
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v6, v7); // bhgdfeac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhgfdaec
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhgdfaec
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v6, v7); // bhgdafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bhgfdace
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); // bhgdface
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v7); // bhgdafce
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); // bhgdacfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hbgfeadc
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbgfaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hbgfadec
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); // hbgfadce
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hbgafedc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbgafdec
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hbgadfec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hbgafdce
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); // hbgadfce
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hbgadcfe
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bhgfeadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhgfaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bhgfadec
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); // bhgfadce
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bhgafedc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhgafdec
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bhgadfec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bhgafdce
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); // bhgadfce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bhgadcfe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v3, v0)) {
															if (compare(v4, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v6, v7); // hfgedbac
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v6, v7); // hfgdebac
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v6, v7); // hfgdbeac
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hfgebdac
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v6, v7); // hfgbedac
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hfgbdeac
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hfbgedac
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hbfgedac
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bhfgedac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfbgdeac
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbfgdeac
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhfgdeac
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfgdbaec
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfgbdaec
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hfbgdaec
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hbfgdaec
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bhfgdaec
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); // hfgdbace
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hfgbdace
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfbgdace
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbfgdace
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhfgdace
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfgebadc
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hfgbeadc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfgbaedc
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hfgbadec
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); // hfgbadce
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); // hfbgeadc
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hfbgaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfbgadec
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hfbgadce
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); // hbfgeadc
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hbfgaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbfgadec
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hbfgadce
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); // bhfgeadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bhfgaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhfgadec
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bhfgadce
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v4, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fhgedbac
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v5, v7); swap(v6, v7); // fhgdebac
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhgdbeac
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhgebdac
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fhgbedac
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhgbdeac
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); // fhbgedac
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); // fbhgedac
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); // bfhgedac
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fhbgdeac
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fbhgdeac
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bfhgdeac
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v6, v7); // fhgdbaec
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fhgbdaec
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); // fhbgdaec
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); // fbhgdaec
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); // bfhgdaec
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v7); // fhgdbace
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); // fhgbdace
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fhbgdace
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fbhgdace
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bfhgdace
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fhgebadc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v6, v7); // fhgbeadc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fhgbaedc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhgbadec
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fhgbadce
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); // fhbgeadc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); // fhbgaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fhbgadec
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhbgadce
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v6); // fbhgeadc
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); // fbhgaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fbhgadec
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbhgadce
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); // bfhgeadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); // bfhgaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bfhgadec
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfhgadce
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v3)) {
													if (compare(v3, v1)) {
														if (compare(v6, v1)) {
															if (compare(v6, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v6, v7); // hdgfebac
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hdgfbeac
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdgfbaec
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hdgfbace
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v6, v7); // hdgbfeac
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdgbfaec
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); // hdgbface
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v6, v7); // hdgbafec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v7); // hdgbafce
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); // hdgbacfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v6, v7); // hfdgebac
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fhdgebac
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hdfgebac
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hfdgbeac
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhdgbeac
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdfgbeac
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfdgbaec
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fhdgbaec
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hdfgbaec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hfdgbace
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); // fhdgbace
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdfgbace
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfdbgeac
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fhdbgeac
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdfbgeac
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hfdbgaec
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); // fhdbgaec
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); // hdfbgaec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfdbgace
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fhdbgace
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdfbgace
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hdbgfeac
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdbfgeac
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdbgfaec
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hdbfgaec
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hdbgafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hdbgface
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdbfgace
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); // hdbgafce
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hdbgacfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfbdgeac
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); // fhbdgeac
																		}
																	}
																	else {
																		swap(v0, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbfdgeac
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); // hfbdgaec
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v6); // fhbdgaec
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); // hbfdgaec
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfbdgace
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v6, v7); // fhbdgace
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbfdgace
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hbdgfeac
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbdfgeac
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbdgfaec
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hbdfgaec
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hbdgafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hbdgface
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbdfgace
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); // hbdgafce
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hbdgacfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		swap(v0, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); // fbhdgeac
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhfdgeac
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); // bfhdgeac
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v7); swap(v4, v6); // fbhdgaec
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); // bhfdgaec
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v6); // bfhdgaec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v7); swap(v4, v6); swap(v6, v7); // fbhdgace
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhfdgace
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v6, v7); // bfhdgace
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bhdgfeac
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhdfgeac
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhdgfaec
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bhdfgaec
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bhdgafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bhdgface
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhdfgace
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); // bhdgafce
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bhdgacfe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v6, v1)) {
															if (compare(v6, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dhgfebac
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhgfbeac
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dhgfbaec
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); // dhgfbace
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dhgbfeac
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhgbfaec
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dhgbface
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dhgbafec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); // dhgbafce
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dhgbacfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); // fdhgebac
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); // dhfgebac
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); // dfhgebac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fdhgbeac
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dhfgbeac
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); // dfhgbeac
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); // fdhgbaec
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhfgbaec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dfhgbaec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fdhgbace
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dhfgbace
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfhgbace
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); // fdhbgeac
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhfbgeac
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dfhbgeac
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v6); // fdhbgaec
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dhfbgaec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); // dfhbgaec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v6, v7); // fdhbgace
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhfbgace
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dfhbgace
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); // dhbgfeac
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dhbfgeac
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dhbgfaec
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); // dhbfgaec
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); // dhbgafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhbgface
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dhbfgace
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dhbgafce
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhbgacfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v7, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); // dbhgfeac
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dbhfgeac
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dbhgfaec
																			}
																			else {
																				swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); // dbhfgaec
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); // dbhgafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbhgface
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dbhfgace
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dbhgafce
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbhgacfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fdbhgeac
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfbhgeac
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbfhgeac
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); // fdbhgaec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dfbhgaec
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dbfhgaec
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fdbhgace
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfbhgace
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbfhgace
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); // bdhgfeac
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bdhfgeac
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bdhgfaec
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); // bdhfgaec
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); // bdhgafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdhgface
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bdhfgace
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bdhgafce
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdhgacfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fbdhgeac
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bfdhgeac
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdfhgeac
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); // fbdhgaec
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); // bfdhgaec
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bdfhgaec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fbdhgace
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bfdhgace
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdfhgace
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v0)) {
												if (compare(v5, v1)) {
													if (compare(v7, v1)) {
														if (compare(v3, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfdbagec
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fhdbagec
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); // fdhbagec
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hfdbagce
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhdbagce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdhbagce
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); // hfdbacge
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fhdbacge
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); // fdhbacge
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdfbagec
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhfbagec
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dfhbagec
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); // hdfbagce
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dhfbagce
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfhbagce
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v7); // hdfbacge
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); // dhfbacge
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dfhbacge
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfbagedc
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hfbagdec
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfbdagec
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfbadgec
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); // hfbdagce
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v7); // hfbdacge
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfbagdce
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hfbadgce
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); // hfbadcge
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fhbagedc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); // fhbagdec
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); // fhbdagec
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fhbadgec
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhbdagce
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); // fhbdacge
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fhbagdce
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhbadgce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fhbadcge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v0)) {
															if (compare(v7, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fbhagedc
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); // fbhagdec
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); // fbhdagec
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fbhadgec
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbhdagce
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); // fbhdacge
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fbhagdce
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbhadgce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fbhadcge
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fdbhagec
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdbhagce
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fdbhacge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfbhagec
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dfbhagce
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); // dfbhacge
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fbdhagec
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbdhagce
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fbdhacge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdbahgec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfbahgec
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbdahgec
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fdbahgce
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); // dfbahgce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fbdahgce
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); // fdbahcge
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v7); // dfbahcge
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); // fbdahcge
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fdbachge
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); // dfbachge
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fbdachge
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbahgedc
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fbahgdec
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbahdgec
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbadhgec
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbahgdce
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fbahdgce
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); // fbadhgce
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); // fbahdcge
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v4, v7); // fbadhcge
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); // fbadchge
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v7, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbfagedc
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hbfagdec
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbfdagec
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbfadgec
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); // hbfdagce
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v4, v7); // hbfdacge
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbfagdce
																			}
																			else {
																				swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hbfadgce
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); // hbfadcge
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdbfagec
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hdbfagce
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); // hdbfacge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dhbfagec
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhbfagce
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dhbfacge
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbdfagec
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hbdfagce
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); // hbdfacge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhfagedc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bhfagdec
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhfdagec
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhfadgec
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); // bhfdagce
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v7); // bhfdacge
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhfagdce
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bhfadgce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); // bhfadcge
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dbhfagec
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbhfagce
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dbhfacge
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhdfagec
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bhdfagce
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); // bhdfacge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bdhfagec
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdhfagce
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bdhfacge
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v0)) {
															if (compare(v7, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bfhagedc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); // bfhagdec
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); // bfhdagec
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bfhadgec
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfhdagce
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); // bfhdacge
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bfhagdce
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfhadgce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bfhadcge
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbfhagec
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dbfhagce
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); // dbfhacge
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bfdhagec
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfdhagce
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bfdhacge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdfhagec
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bdfhagce
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); // bdfhacge
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbfahgec
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfdahgec
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdfahgec
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); // dbfahgce
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bfdahgce
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); // bdfahgce
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v4, v7); // dbfahcge
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); // bfdahcge
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v7); // bdfahcge
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); // dbfachge
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bfdachge
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); // bdfachge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfahgedc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bfahgdec
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfahdgec
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfadhgec
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfahgdce
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bfahdgce
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); // bfadhgce
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); // bfahdcge
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v7); // bfadhcge
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); // bfadchge
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v0)) {
													if (compare(v3, v1)) {
														if (compare(v6, v2)) {
															if (compare(v7, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); // hdbagfec
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hdbafgec
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hdbagfce
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdbafgce
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdbagcfe
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v6); // dhbagfec
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); // dhbafgec
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v6, v7); // dhbagfce
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhbafgce
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhbagcfe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v2, v7); swap(v4, v6); // dbhagfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v2, v7); swap(v4, v6); swap(v6, v7); // dbhagfce
																			}
																			else {
																				swap(v0, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbhagcfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); // dbahgfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dbahgfce
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbahgcfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); // dbhafgec
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dbahfgec
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbafhgec
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbhafgce
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbahfgce
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dbafhgce
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hdbafcge
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hdbacgfe
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); // hdbacfge
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); // dhbafcge
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); // dhbacgfe
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v7); // dhbacfge
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		swap(v0, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); // dbhafcge
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); // dbhacgfe
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v4, v7); // dbhacfge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dbahfcge
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); // dbafhcge
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dbahcgfe
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); // dbahcfge
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); // dbachgfe
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v4, v7); // dbachfge
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dbafchge
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); // dbacfhge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v7, v3)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); // hbdagfec
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hbdafgec
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hbdagfce
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbdafgce
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbdagcfe
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); // bhdagfec
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bhdafgec
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bhdagfce
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhdafgce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhdagcfe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v6); // bdhagfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v6, v7); // bdhagfce
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdhagcfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); // bdahgfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bdahgfce
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdahgcfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); // bdhafgec
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bdahfgec
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdafhgec
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdhafgce
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdahfgce
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bdafhgce
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v5, v2)) {
																		swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hbdafcge
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hbdacgfe
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); // hbdacfge
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bhdafcge
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bhdacgfe
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); // bhdacfge
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); // bdhafcge
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); // bdhacgfe
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v7); // bdhacfge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bdahfcge
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); // bdafhcge
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bdahcgfe
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); // bdahcfge
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); // bdachgfe
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v7); // bdachfge
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bdafchge
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); // bdacfhge
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v0)) {
														if (compare(v7, v1)) {
															if (compare(v6, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); // hbagfedc
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hbagfdec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); // hbagdfec
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v4, v6); // hbadgfec
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbagfdce
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hbagdfce
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v4, v6); swap(v6, v7); // hbadgfce
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbagdcfe
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbadgcfe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); // hbadcgfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hbafgedc
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); // hbafgdec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hbafdgec
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); // hbadfgec
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hbafgdce
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbafdgce
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbadfgce
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hbafdcge
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); // hbadfcge
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v4, v7); // hbadcfge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); // bhagfedc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bhagfdec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); // bhagdfec
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v6); // bhadgfec
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhagfdce
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bhagdfce
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v6, v7); // bhadgfce
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhagdcfe
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhadgcfe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); // bhadcgfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bhafgedc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); // bhafgdec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bhafdgec
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); // bhadfgec
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bhafgdce
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhafdgce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhadfgce
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bhafdcge
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); // bhadfcge
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v7); // bhadcfge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v5); // bahgfedc
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bahfgedc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bahgfdec
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v6); // bahfgdec
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bahfdgec
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bafhgedc
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bafhgdec
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bafhdgec
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bafdhgec
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bahgfdce
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bahfgdce
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bafhgdce
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bahfdgce
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bafhdgce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v4, v7); swap(v5, v6); // bafdhgce
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bahfdcge
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v7); // bafhdcge
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v5); swap(v4, v7); // bafdhcge
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v4, v5); swap(v5, v7); // bafdchge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v6); // bahgdfec
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v4, v6); // bahdgfec
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v6); // badhgfec
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bahgdfce
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v4, v6); swap(v6, v7); // bahdgfce
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // badhgfce
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bahgdcfe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bahdgcfe
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // badhgcfe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v4, v7); swap(v5, v6); // bahdcgfe
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // badhcgfe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v4, v7); swap(v5, v6); // badchgfe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v4, v5); swap(v5, v6); // bahdfgec
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // badhfgec
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // badfhgec
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bahdfgce
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // badhfgce
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // badfhgce
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v4, v5); swap(v5, v7); // bahdfcge
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // badhfcge
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v7); // badfhcge
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // badfchge
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v4, v7); // bahdcfge
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v7); // badhcfge
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v4, v7); // badchfge
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v4, v5); swap(v5, v7); // badcfhge
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v0)) {
											if (compare(v4, v3)) {
												if (compare(v7, v4)) {
													if (compare(v7, v5)) {
														if (compare(v3, v0)) {
															if (compare(v6, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hfegdbac
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v6, v7); // hfedgbac
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfedbgac
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfegbdac
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hfebgdac
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfebdgac
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfbegdac
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbfegdac
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhfegdac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hfbedgac
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hbfedgac
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bhfedgac
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); // hfedbagc
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hfebdagc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); // hfbedagc
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); // hbfedagc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); // bhfedagc
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfedbacg
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfebdacg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hfbedacg
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hbfedacg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bhfedacg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hfegbadc
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfebgadc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hfebagdc
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); // hfebadgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hfebadcg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hfbegadc
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfbeagdc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hfbeadgc
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfbeadcg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hbfegadc
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbfeagdc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hbfeadgc
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbfeadcg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bhfegadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhfeagdc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bhfeadgc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhfeadcg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v6, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhegdbac
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhedgbac
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); // fhedbgac
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fhegbdac
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhebgdac
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fhebdgac
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fhbegdac
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fbhegdac
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bfhegdac
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); // fhbedgac
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); // fbhedgac
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); // bfhedgac
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v7); // fhedbagc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); // fhebdagc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); // fhbedagc
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v4); // fbhedagc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); // bfhedagc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v6, v7); // fhedbacg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fhebdacg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v6, v7); // fhbedacg
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v6, v7); // fbhedacg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v6, v7); // bfhedacg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); // fhegbadc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fhebgadc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhebagdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fhebadgc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhebadcg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); // fhbegadc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fhbeagdc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); // fhbeadgc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fhbeadcg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); // fbhegadc
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fbheagdc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); // fbheadgc
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fbheadcg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); // bfhegadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bfheagdc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); // bfheadgc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bfheadcg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v7, v1)) {
															if (compare(v6, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fehgdbac
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); // fehdgbac
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fedhgbac
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v5, v6); // fehdbgac
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); // fedhbgac
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); // fehgbdac
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fehbgdac
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); // fehbdgac
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); // fehgbadc
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); // fehbgadc
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); // fehdbagc
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); // fedhbagc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v6, v7); // fehdbacg
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v6, v7); // fedhbacg
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); // fehbdagc
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v6, v7); // fehbdacg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fehbagdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); // fehbadgc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fehbadcg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v3, v0)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // febhgdac
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // febhdgac
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); // febhdagc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // febhdacg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v0)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // febhgadc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // febhagdc
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // febhadgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // febhadcg
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fedbhgac
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); // fedbhagc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fedbhacg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); // febdhgac
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v7); // febdhagc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v6, v7); // febdhacg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fedbahgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fedbahcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fedbachg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); // febdahgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // febdahcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v6, v7); // febdachg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // febahgdc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); // febahdgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // febadhgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // febahdcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // febadhcg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // febadchg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v5, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbehgdac
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fbehdgac
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); // fbedhgac
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); // fbehdagc
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v4, v7); // fbedhagc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); // fbedahgc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fbehdacg
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v4, v7); swap(v6, v7); // fbedhacg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbedahcg
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v4, v5); swap(v6, v7); // fbedachg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfehgdac
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bfehdgac
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); // bfedhgac
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); // bfehdagc
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v7); // bfedhagc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); // bfedahgc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bfehdacg
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v6, v7); // bfedhacg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfedahcg
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v6, v7); // bfedachg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fbehgadc
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbehagdc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fbehadgc
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbehadcg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fbeahgdc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); // fbeahdgc
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fbeadhgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fbeahdcg
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbeadhcg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fbeadchg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bfehgadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfehagdc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bfehadgc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfehadcg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bfeahgdc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); // bfeahdgc
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bfeadhgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bfeahdcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfeadhcg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bfeadchg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v3, v1)) {
														if (compare(v4, v1)) {
															if (compare(v7, v4)) {
																if (compare(v6, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hfdegbac
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhdegbac
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fdhegbac
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfdebgac
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fhdebgac
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); // fdhebgac
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hfdebagc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); // fhdebagc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); // fdhebagc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfdebacg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fhdebacg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v6, v7); // fdhebacg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdehgbac
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fdehbgac
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); // fdehbagc
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fdehbacg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); // fdebhgac
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v7); // fdebhagc
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v6, v7); // fdebhacg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); // fdebahgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdebahcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v6, v7); // fdebachg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hfdbegac
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); // fhdbegac
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v5, v6); // fdhbegac
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); // hfdbeagc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); // fhdbeagc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); // fdhbeagc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hfdbeacg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v6, v7); // fhdbeacg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v6, v7); // fdhbeacg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); // fdbhegac
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fdbehgac
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); // fdbheagc
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); // fdbehagc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fdbeahgc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v6, v7); // fdbheacg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fdbehacg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdbeahcg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fdbeachg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v7, v3)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v5, v6); swap(v6, v7); // hfbdegac
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v5, v6); // fhbdegac
																		}
																	}
																	else {
																		swap(v0, v5); swap(v2, v7); swap(v5, v6); // fbhdegac
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v5, v7); // hfbdeagc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); // fhbdeagc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); // fbhdeagc
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v5, v7); swap(v6, v7); // hfbdeacg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v6, v7); // fhbdeacg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v6, v7); // fbhdeacg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); // fbdhegac
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fbdehgac
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v2, v3); swap(v3, v7); // fbdheagc
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); // fbdehagc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fbdeahgc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v6, v7); // fbdheacg
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fbdehacg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbdeahcg
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fbdeachg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		swap(v0, v7); swap(v2, v5); swap(v5, v6); swap(v6, v7); // hbfdegac
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v5, v6); swap(v6, v7); // bhfdegac
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v5, v6); // bfhdegac
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v5); swap(v5, v7); // hbfdeagc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v5, v7); // bhfdeagc
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); // bfhdeagc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v5); swap(v5, v7); swap(v6, v7); // hbfdeacg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v5, v7); swap(v6, v7); // bhfdeacg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v6, v7); // bfhdeacg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); // bfdhegac
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bfdehgac
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); // bfdheagc
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); // bfdehagc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bfdeahgc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v6, v7); // bfdheacg
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bfdehacg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfdeahcg
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bfdeachg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v1)) {
															if (compare(v7, v4)) {
																if (compare(v6, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdfegbac
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dhfegbac
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); // dfhegbac
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hdfebgac
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhfebgac
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dfhebgac
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); // hdfebagc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dhfebagc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); // dfhebagc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hdfebacg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhfebacg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dfhebacg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dfehgbac
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfehbgac
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dfehbagc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfehbacg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dfebhgac
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); // dfebhagc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dfebhacg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dfebahgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfebahcg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dfebachg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v3)) {
																		swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v5, v6); swap(v6, v7); // hdfbegac
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dhfbegac
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); // dfhbegac
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v5, v7); // hdfbeagc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); // dhfbeagc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); // dfhbeagc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v5, v7); swap(v6, v7); // hdfbeacg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dhfbeacg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v6, v7); // dfhbeacg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dfbhegac
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfbehgac
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); // dfbheagc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dfbehagc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); // dfbeahgc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dfbheacg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfbehacg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dfbeahcg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v6, v7); // dfbeachg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v7, v5)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hdbfegac
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); // dhbfegac
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); // dbhfegac
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); // hdbfeagc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); // dhbfeagc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); // dbhfeagc
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hdbfeacg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v6, v7); // dhbfeacg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v6, v7); // dbhfeacg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dbfhegac
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbfehgac
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); // dbfheagc
																			}
																			else {
																				swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dbfehagc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); // dbfeahgc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dbfheacg
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbfehacg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dbfeahcg
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v6, v7); // dbfeachg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hbdfegac
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bhdfegac
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); // bdhfegac
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); // hbdfeagc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); // bhdfeagc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); // bdhfeagc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hbdfeacg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bhdfeacg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v6, v7); // bdhfeacg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bdfhegac
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdfehgac
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); // bdfheagc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bdfehagc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); // bdfeahgc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bdfheacg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdfehacg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bdfeahcg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v6, v7); // bdfeachg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v0)) {
												if (compare(v7, v1)) {
													if (compare(v5, v1)) {
														if (compare(v3, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hfdbaegc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); // fhdbaegc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v5); // fdhbaegc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfdbaecg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fhdbaecg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v6, v7); // fdhbaecg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hfdbaceg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhdbaceg
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdhbaceg
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); // hdfbaegc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dhfbaegc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); // dfhbaegc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdfbaecg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhfbaecg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dfhbaecg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v6, v7); // hdfbaceg
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dhfbaceg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfhbaceg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hfbaegdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); // hfbaedgc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); // hfbdaegc
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hfbadegc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfbdaecg
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v6, v7); // hfbdaceg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hfbaedcg
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfbadecg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hfbadceg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); // fhbaegdc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); // fhbaedgc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v5); // fhbdaegc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); // fhbadegc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v6, v7); // fhbdaecg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhbdaceg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v6, v7); // fhbaedcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fhbadecg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhbadceg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v5, v3)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hbfaegdc
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); // hbfaedgc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); // hbfdaegc
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hbfadegc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbfdaecg
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v4, v7); swap(v6, v7); // hbfdaceg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hbfaedcg
																			}
																			else {
																				swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbfadecg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hbfadceg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hdbfaegc
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdbfaecg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hdbfaceg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); // dhbfaegc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dhbfaecg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhbfaceg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hbdfaegc
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbdfaecg
																		}
																	}
																	else {
																		swap(v0, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hbdfaceg
																	}
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); // hdbafegc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v5); // dhbafegc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); // hbdafegc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hdbafecg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v6, v7); // dhbafecg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hbdafecg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdbafceg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhbafceg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbdafceg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hdbacfeg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v6, v7); // dhbacfeg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hbdacfeg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); // hbafegdc
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v5); // hbafedgc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); // hbafdegc
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v4, v5); // hbadfegc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v6, v7); // hbafedcg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hbafdecg
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v4, v5); swap(v6, v7); // hbadfecg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbafdceg
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbadfceg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v4, v7); swap(v6, v7); // hbadcfeg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v5, v0)) {
															if (compare(v5, v3)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bhfaegdc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); // bhfaedgc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); // bhfdaegc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bhfadegc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhfdaecg
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v6, v7); // bhfdaceg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bhfaedcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhfadecg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bhfadceg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); // dbhfaegc
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dbhfaecg
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbhfaceg
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bhdfaegc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhdfaecg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bhdfaceg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); // bdhfaegc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bdhfaecg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdhfaceg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v4, v5); // dbhafegc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); // bhdafegc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v5); // bdhafegc
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v4, v5); swap(v6, v7); // dbhafecg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bhdafecg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v6, v7); // bdhafecg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbhafceg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhdafceg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdhafceg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v4, v7); swap(v6, v7); // dbhacfeg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bhdacfeg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v6, v7); // bdhacfeg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); // bhafegdc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); // bhafedgc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); // bhafdegc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v5); // bhadfegc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v6, v7); // bhafedcg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bhafdecg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v6, v7); // bhadfecg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhafdceg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhadfceg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v6, v7); // bhadcfeg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v7, v3)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); // fbhaegdc
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); // fbhaedgc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v5); swap(v2, v7); swap(v4, v5); // fbhdaegc
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); // fbhadegc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v2, v7); swap(v4, v5); swap(v6, v7); // fbhdaecg
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbhdaceg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v6, v7); // fbhaedcg
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fbhadecg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbhadceg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); // fdbhaegc
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fdbhaecg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdbhaceg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dfbhaegc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfbhaecg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dfbhaceg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); // fbdhaegc
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fbdhaecg
																		}
																	}
																	else {
																		swap(v0, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbdhaceg
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); // bfhaegdc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); // bfhaedgc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v5); // bfhdaegc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); // bfhadegc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v6, v7); // bfhdaecg
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfhdaceg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v6, v7); // bfhaedcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bfhadecg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfhadceg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dbfhaegc
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbfhaecg
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dbfhaceg
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); // bfdhaegc
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bfdhaecg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfdhaceg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bdfhaegc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdfhaecg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bdfhaceg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v0)) {
													if (compare(v3, v1)) {
														if (compare(v4, v2)) {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fdbahegc
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); // fdbaehgc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdbahecg
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fdbaehcg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v6, v7); // fdbaechg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); // dfbahegc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v5, v7); // dfbaehgc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfbahecg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v5, v7); swap(v6, v7); // dfbaehcg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v6, v7); // dfbaechg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v4)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); // dbfahegc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); // dbahfegc
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dbafhegc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbfahecg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dbahfecg
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbafhecg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v5, v7); // dbfaehgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v5); swap(v5, v7); swap(v6, v7); // dbfaehcg
																			}
																			else {
																				swap(v0, v3); swap(v2, v5); swap(v6, v7); // dbfaechg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); // dbafehgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dbafehcg
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v6, v7); // dbafechg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fdbahceg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdbacheg
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fdbacehg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v6, v7); // dfbahceg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfbacheg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v6, v7); // dfbacehg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		swap(v0, v3); swap(v2, v5); swap(v4, v7); swap(v6, v7); // dbfahceg
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbfacheg
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v4, v5); swap(v6, v7); // dbfacehg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbahfceg
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dbafhceg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbafcheg
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dbafcehg
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dbahcfeg
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v4, v7); swap(v6, v7); // dbachfeg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbacfheg
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v4, v5); swap(v6, v7); // dbacfehg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fbdahegc
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); // fbdaehgc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbdahecg
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fbdaehcg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v6, v7); // fbdaechg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bfdahegc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); // bfdaehgc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfdahecg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bfdaehcg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v6, v7); // bfdaechg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v4)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); // bdfahegc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); // bdahfegc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bdafhegc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdfahecg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bdahfecg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdafhecg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v5, v7); // bdfaehgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v5, v7); swap(v6, v7); // bdfaehcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v6, v7); // bdfaechg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); // bdafehgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bdafehcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v6, v7); // bdafechg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v7, v2)) {
																		swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fbdahceg
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbdacheg
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fbdacehg
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bfdahceg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfdacheg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bfdacehg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v6, v7); // bdfahceg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdfacheg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v6, v7); // bdfacehg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdahfceg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bdafhceg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdafcheg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bdafcehg
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bdahcfeg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v6, v7); // bdachfeg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdacfheg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v6, v7); // bdacfehg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v5, v1)) {
															if (compare(v7, v4)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fbahegdc
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); // fbahedgc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fbahdegc
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); // fbadhegc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fbahedcg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbahdecg
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbadhecg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fbahdceg
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v4, v7); swap(v6, v7); // fbadhceg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbadcheg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbaehgdc
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fbaehdgc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); // fbaedhgc
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v5, v7); // fbadehgc
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbaehdcg
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fbaedhcg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v5, v7); swap(v6, v7); // fbadehcg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v6, v7); // fbaedchg
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v6, v7); // fbadechg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v4, v5); swap(v6, v7); // fbadcehg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v4)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bfahegdc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); // bfahedgc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bfahdegc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); // bfadhegc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bfahedcg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfahdecg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfadhecg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bfahdceg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v6, v7); // bfadhceg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfadcheg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfaehgdc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bfaehdgc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); // bfaedhgc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v5, v7); // bfadehgc
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfaehdcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bfaedhcg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v5, v7); swap(v6, v7); // bfadehcg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v6, v7); // bfaedchg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v6, v7); // bfadechg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v6, v7); // bfadcehg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																if (compare(v7, v4)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v5, v6); // bahfegdc
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); // bahfedgc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v6, v7); // bahfedcg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bafhegdc
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v5, v7); // bafhedgc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bafhedcg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bafehgdc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bafehdgc
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v5, v7); // bafedhgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bafehdcg
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bafedhcg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v6, v7); // bafedchg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v5); // bahfdegc
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bahfdecg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bahfdceg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bafhdegc
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bafhdecg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bafhdceg
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v5); swap(v4, v7); swap(v5, v7); // bafdhegc
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bafdhecg
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v5); swap(v4, v7); swap(v6, v7); // bafdhceg
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bafdcheg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v5); swap(v5, v7); // bafdehgc
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v5, v7); swap(v6, v7); // bafdehcg
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v2, v5); swap(v6, v7); // bafdechg
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v4, v5); swap(v6, v7); // bafdcehg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v4, v5); // bahdfegc
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v5); // badhfegc
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // badfhegc
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v5, v7); // badfehgc
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v4, v5); swap(v6, v7); // bahdfecg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // badhfecg
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // badfhecg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // badfehcg
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v6, v7); // badfechg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bahdfceg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // badhfceg
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // badfhceg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // badfcheg
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // badfcehg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v4, v7); swap(v6, v7); // bahdcfeg
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // badhcfeg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v4, v7); swap(v6, v7); // badchfeg
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // badcfheg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v4, v5); swap(v6, v7); // badcfehg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v7, v4)) {
										if (compare(v7, v0)) {
											if (compare(v7, v3)) {
												if (compare(v7, v5)) {
													if (compare(v7, v1)) {
														if (compare(v5, v1)) {
															if (compare(v4, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); // ghfedbac
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v5, v7); // ghfdebac
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); // ghdfebac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); // ghfdbeac
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ghdfbeac
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ghfebdac
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); // ghfbedac
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ghfbdeac
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghfebadc
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ghfbeadc
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghfdbaec
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghdfbaec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); // ghfdbace
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ghdfbace
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghfbdaec
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ghfbdace
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghfbaedc
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ghfbadec
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghfbadce
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); // ghbfedac
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); // ghbfdeac
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ghbfdaec
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghbfdace
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v4, v0)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); // ghbfeadc
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ghbfaedc
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); // ghbfadec
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ghbfadce
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); // ghdbfeac
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ghdbfaec
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghdbface
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v5); // ghbdfeac
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); // ghbdfaec
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghbdface
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); // ghdbafec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ghdbafce
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghdbacfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v6); // ghbdafec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v6, v7); // ghbdafce
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghbdacfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); // ghbafedc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ghbafdec
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); // ghbadfec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghbafdce
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ghbadfce
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghbadcfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v6, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); // gbhfedac
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); // gbhfdeac
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v4, v5); // gbhdfeac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gbhfdaec
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); // gbhdfaec
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v4, v6); // gbhdafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbhfdace
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbhdface
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v2, v7); swap(v4, v6); swap(v6, v7); // gbhdafce
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbhdacfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); // bghfedac
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); // bghfdeac
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v5); // bghdfeac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bghfdaec
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); // bghdfaec
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v6); // bghdafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bghfdace
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bghdface
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v6, v7); // bghdafce
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bghdacfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); // gbhfeadc
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gbhfaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); // gbhfadec
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gbhfadce
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); // gbhafedc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gbhafdec
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); // gbhadfec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbhafdce
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gbhadfce
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbhadcfe
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); // bghfeadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bghfaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); // bghfadec
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bghfadce
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); // bghafedc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bghafdec
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); // bghadfec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bghafdce
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bghadfce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bghadcfe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v3, v0)) {
															if (compare(v4, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); // gfhedbac
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); // gfhdebac
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v5); // gfhdbeac
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); // gfhebdac
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); // gfhbedac
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); // gfhbdeac
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); // gfbhedac
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); // gbfhedac
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); // bgfhedac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gfbhdeac
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gbfhdeac
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bgfhdeac
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); // gfhdbaec
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gfhbdaec
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfbhdaec
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbfhdaec
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgfhdaec
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfhdbace
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfhbdace
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gfbhdace
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gbfhdace
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bgfhdace
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gfhebadc
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); // gfhbeadc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gfhbaedc
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); // gfhbadec
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gfhbadce
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gfbheadc
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfbhaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gfbhadec
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfbhadce
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gbfheadc
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbfhaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gbfhadec
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbfhadce
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bgfheadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgfhaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bgfhadec
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgfhadce
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v4, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); // fghedbac
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v5, v6); // fghdebac
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); // fghdbeac
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fghebdac
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); // fghbedac
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fghbdeac
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fgbhedac
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fbghedac
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bfghedac
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgbhdeac
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbghdeac
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfghdeac
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v6); // fghdbaec
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); // fghbdaec
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fgbhdaec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fbghdaec
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bfghdaec
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v6, v7); // fghdbace
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fghbdace
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); // fgbhdace
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); // fbghdace
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); // bfghdace
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); // fghebadc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); // fghbeadc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); // fghbaedc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fghbadec
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fghbadce
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v6, v7); // fgbheadc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fgbhaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgbhadec
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fgbhadce
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v6, v7); // fbgheadc
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fbghaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbghadec
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fbghadce
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v6, v7); // bfgheadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bfghaedc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfghadec
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bfghadce
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v3)) {
													if (compare(v3, v1)) {
														if (compare(v7, v1)) {
															if (compare(v7, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); // gdhfebac
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); // gdhfbeac
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gdhfbaec
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdhfbace
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v5); // gdhbfeac
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); // gdhbfaec
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdhbface
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v6); // gdhbafec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v6, v7); // gdhbafce
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdhbacfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); // gfdhebac
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); // fgdhebac
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); // gdfhebac
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); // gfdhbeac
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fgdhbeac
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gdfhbeac
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gfdhbaec
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); // fgdhbaec
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdfhbaec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfdhbace
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fgdhbace
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gdfhbace
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gfdbheac
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgdbheac
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); // gdfbheac
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfdbhaec
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fgdbhaec
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdfbhaec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gfdbhace
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); // fgdbhace
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); // gdfbhace
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); // gdbhfeac
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gdbfheac
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gdbhfaec
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdbfhaec
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); // gdbhafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdbhface
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gdbfhace
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gdbhafce
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdbhacfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); // gfbdheac
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgbdheac
																		}
																	}
																	else {
																		swap(v0, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); // gbfdheac
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfbdhaec
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v6, v7); // fgbdhaec
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbfdhaec
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); // gfbdhace
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v7); // fgbdhace
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); // gbfdhace
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); // gbdhfeac
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gbdfheac
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gbdhfaec
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbdfhaec
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); // gbdhafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbdhface
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gbdfhace
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gbdhafce
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbdhacfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		swap(v0, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbgdheac
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); // bgfdheac
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfgdheac
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v6); swap(v4, v7); swap(v6, v7); // fbgdhaec
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgfdhaec
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v6, v7); // bfgdhaec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v6); swap(v4, v7); // fbgdhace
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); // bgfdhace
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v7); // bfgdhace
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); // bgdhfeac
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bgdfheac
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bgdhfaec
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgdfhaec
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); // bgdhafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgdhface
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bgdfhace
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bgdhafce
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgdhacfe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v7, v1)) {
															if (compare(v7, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); // dghfebac
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dghfbeac
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); // dghfbaec
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dghfbace
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); // dghbfeac
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dghbfaec
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dghbface
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); // dghbafec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dghbafce
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dghbacfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fdghebac
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dgfhebac
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v6, v7); // dfghebac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdghbeac
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgfhbeac
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dfghbeac
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fdghbaec
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dgfhbaec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfghbaec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); // fdghbace
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgfhbace
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dfghbace
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdgbheac
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dgfbheac
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfgbheac
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v6, v7); // fdgbhaec
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgfbhaec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dfgbhaec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v7); // fdgbhace
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dgfbhace
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); // dfgbhace
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dgbhfeac
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgbfheac
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgbhfaec
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dgbfhaec
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dgbhafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dgbhface
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); // dgbfhace
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); // dgbhafce
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dgbhacfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v6, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dbghfeac
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbgfheac
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbghfaec
																			}
																			else {
																				swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dbgfhaec
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dbghafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dbghface
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); // dbgfhace
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); // dbghafce
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dbghacfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdbgheac
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dfbgheac
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dbfgheac
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fdbghaec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfbghaec
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbfghaec
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); // fdbghace
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dfbghace
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dbfghace
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bdghfeac
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdgfheac
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdghfaec
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bdgfhaec
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bdghafec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bdghface
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); // bdgfhace
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); // bdghafce
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bdghacfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbdgheac
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfdgheac
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bdfgheac
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fbdghaec
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bfdghaec
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdfghaec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); // fbdghace
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); // bfdghace
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bdfghace
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v0)) {
												if (compare(v5, v1)) {
													if (compare(v6, v1)) {
														if (compare(v3, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gfdbahec
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgdbahec
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdgbahec
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfdbahce
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fgdbahce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); // fdgbahce
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gfdbache
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgdbache
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdgbache
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); // gdfbahec
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dgfbahec
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfgbahec
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdfbahce
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgfbahce
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dfgbahce
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v6, v7); // gdfbache
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dgfbache
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfgbache
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gfbahedc
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfbahdec
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); // gfbdahec
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gfbadhec
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfbdahce
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v6, v7); // gfbdache
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gfbahdce
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfbadhce
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gfbadche
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgbahedc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fgbahdec
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgbdahec
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgbadhec
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); // fgbdahce
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgbdache
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); // fgbahdce
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fgbadhce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgbadche
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v6, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbgahedc
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fbgahdec
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbgdahec
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbgadhec
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); // fbgdahce
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbgdache
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); // fbgahdce
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fbgadhce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbgadche
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdbgahec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fdbgahce
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdbgache
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dfbgahec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfbgahce
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dfbgache
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbdgahec
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fbdgahce
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbdgache
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fdbaghec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); // dfbaghec
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fbdaghec
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdbaghce
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfbaghce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbdaghce
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fdbagche
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v6, v7); // dfbagche
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fbdagche
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdbacghe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfbacghe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbdacghe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fbaghedc
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbaghdec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fbagdhec
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); // fbadghec
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fbaghdce
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbagdhce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbadghce
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fbagdche
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v4, v6); swap(v6, v7); // fbadgche
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbadcghe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v6, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gbfahedc
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbfahdec
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); // gbfdahec
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gbfadhec
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbfdahce
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v4, v6); swap(v6, v7); // gbfdache
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gbfahdce
																			}
																			else {
																				swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbfadhce
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gbfadche
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gdbfahec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdbfahce
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gdbfache
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgbfahec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dgbfahce
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgbfache
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gbdfahec
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbdfahce
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gbdfache
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bgfahedc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgfahdec
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); // bgfdahec
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bgfadhec
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgfdahce
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v6, v7); // bgfdache
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bgfahdce
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgfadhce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bgfadche
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbgfahec
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dbgfahce
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbgfache
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bgdfahec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgdfahce
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bgdfache
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdgfahec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bdgfahce
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdgfache
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v6, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfgahedc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bfgahdec
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfgdahec
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfgadhec
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); // bfgdahce
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfgdache
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); // bfgahdce
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bfgadhce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfgadche
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dbfgahec
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbfgahce
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dbfgache
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfdgahec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bfdgahce
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfdgache
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bdfgahec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdfgahce
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bdfgache
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); // dbfaghec
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bfdaghec
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); // bdfaghec
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbfaghce
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfdaghce
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdfaghce
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v4, v6); swap(v6, v7); // dbfagche
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bfdagche
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v6, v7); // bdfagche
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbfacghe
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfdacghe
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdfacghe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bfaghedc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfaghdec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bfagdhec
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); // bfadghec
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bfaghdce
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfagdhce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfadghce
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bfagdche
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v6, v7); // bfadgche
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfadcghe
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v0)) {
													if (compare(v3, v1)) {
														if (compare(v7, v2)) {
															if (compare(v6, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gdbahfec
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdbafhec
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); // gdbahfce
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gdbafhce
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gdbahcfe
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v6, v7); // dgbahfec
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgbafhec
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v7); // dgbahfce
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); // dgbafhce
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); // dgbahcfe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v2, v6); swap(v4, v7); swap(v6, v7); // dbgahfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v2, v6); swap(v4, v7); // dbgahfce
																			}
																			else {
																				swap(v0, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); // dbgahcfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dbaghfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); // dbaghfce
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dbaghcfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbgafhec
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbagfhec
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dbafghec
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); // dbgafhce
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dbagfhce
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbafghce
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdbafche
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdbachfe
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gdbacfhe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgbafche
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgbachfe
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v6, v7); // dgbacfhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		swap(v0, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbgafche
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbgachfe
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v4, v6); swap(v6, v7); // dbgacfhe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbagfche
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dbafgche
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbagchfe
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dbagcfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbacghfe
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v4, v6); swap(v6, v7); // dbacgfhe
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbafcghe
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbacfghe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v6, v3)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gbdahfec
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbdafhec
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); // gbdahfce
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gbdafhce
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gbdahcfe
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bgdahfec
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgdafhec
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); // bgdahfce
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bgdafhce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bgdahcfe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v6, v7); // bdgahfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v7); // bdgahfce
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); // bdgahcfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bdaghfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); // bdaghfce
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bdaghcfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdgafhec
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdagfhec
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bdafghec
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); // bdgafhce
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bdagfhce
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdafghce
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v5, v2)) {
																		swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbdafche
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbdachfe
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gbdacfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgdafche
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgdachfe
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bgdacfhe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdgafche
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdgachfe
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v6, v7); // bdgacfhe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdagfche
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bdafgche
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdagchfe
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bdagcfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdacghfe
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v6, v7); // bdacgfhe
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdafcghe
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdacfghe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v0)) {
														if (compare(v6, v1)) {
															if (compare(v7, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gbahfedc
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbahfdec
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gbahdfec
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v4, v7); swap(v6, v7); // gbadhfec
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gbahfdce
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); // gbahdfce
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v4, v7); // gbadhfce
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gbahdcfe
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); // gbadhcfe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbadchfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbafhedc
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gbafhdec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbafdhec
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbadfhec
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); // gbafhdce
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gbafdhce
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); // gbadfhce
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbafdche
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbadfche
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v4, v6); swap(v6, v7); // gbadcfhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bgahfedc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgahfdec
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bgahdfec
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v6, v7); // bgadhfec
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bgahfdce
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); // bgahdfce
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v7); // bgadhfce
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bgahdcfe
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); // bgadhcfe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgadchfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgafhedc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bgafhdec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgafdhec
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgadfhec
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); // bgafhdce
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bgafdhce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); // bgadfhce
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgafdche
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgadfche
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v6, v7); // bgadcfhe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // baghfedc
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bagfhedc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // baghfdec
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bagfhdec
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bagfdhec
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bafghedc
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bafghdec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bafgdhec
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v4, v6); swap(v5, v7); // bafdghec
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // baghfdce
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v7); // bagfhdce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bafghdce
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bagfdhce
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bafgdhce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bafdghce
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bagfdche
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bafgdche
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v5); swap(v4, v6); swap(v6, v7); // bafdgche
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bafdcghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // baghdfec
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v4, v7); swap(v6, v7); // bagdhfec
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // badghfec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v7); // baghdfce
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v4, v7); // bagdhfce
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v7); // badghfce
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // baghdcfe
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v4, v7); swap(v5, v6); // bagdhcfe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // badghcfe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bagdchfe
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // badgchfe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // badcghfe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bagdfhec
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // badgfhec
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // badfghec
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v6); swap(v4, v5); swap(v5, v7); // bagdfhce
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // badgfhce
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // badfghce
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bagdfche
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // badgfche
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // badfgche
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // badfcghe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v4, v6); swap(v6, v7); // bagdcfhe
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // badgcfhe
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v4, v6); swap(v6, v7); // badcgfhe
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // badcfghe
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v0)) {
											if (compare(v4, v3)) {
												if (compare(v6, v4)) {
													if (compare(v6, v5)) {
														if (compare(v3, v0)) {
															if (compare(v7, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); // gfehdbac
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v7); // gfedhbac
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); // gfedbhac
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gfehbdac
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); // gfebhdac
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gfebdhac
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gfbehdac
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gbfehdac
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bgfehdac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); // gfbedhac
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); // gbfedhac
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); // bgfedhac
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfedbahc
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfebdahc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gfbedahc
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gbfedahc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bgfedahc
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // gfedbach
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gfebdach
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // gfbedach
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // gbfedach
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // bgfedach
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfehbadc
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gfebhadc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfebahdc
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gfebadhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // gfebadch
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfbehadc
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gfbeahdc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfbeadhc
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gfbeadch
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbfehadc
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gbfeahdc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbfeadhc
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gbfeadch
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgfehadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bgfeahdc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgfeadhc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bgfeadch
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v7, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fgehdbac
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); // fgedhbac
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgedbhac
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgehbdac
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fgebhdac
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgebdhac
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgbehdac
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbgehdac
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfgehdac
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fgbedhac
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fbgedhac
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bfgedhac
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v6, v7); // fgedbahc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fgebdahc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v6, v7); // fgbedahc
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v6, v7); // fbgedahc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v6, v7); // bfgedahc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v6); // fgedbach
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // fgebdach
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); // fgbedach
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v4); // fbgedach
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); // bfgedach
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fgehbadc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); // fgebhadc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fgebahdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgebadhc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fgebadch
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fgbehadc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgbeahdc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fgbeadhc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // fgbeadch
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fbgehadc
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbgeahdc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fbgeadhc
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // fbgeadch
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bfgehadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfgeahdc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bfgeadhc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // bfgeadch
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v6, v1)) {
															if (compare(v7, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // feghdbac
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fegdhbac
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fedghbac
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v5, v7); swap(v6, v7); // fegdbhac
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fedgbhac
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // feghbdac
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fegbhdac
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fegbdhac
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v6, v7); // feghbadc
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fegbhadc
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v6, v7); // fegdbahc
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v6, v7); // fedgbahc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); // fegdbach
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); // fedgbach
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v6, v7); // fegbdahc
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); // fegbdach
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fegbahdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fegbadhc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // fegbadch
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v3, v0)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // febghdac
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // febgdhac
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // febgdahc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // febgdach
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v0)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); // febghadc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // febgahdc
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // febgadhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // febgadch
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fedbghac
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fedbgahc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // fedbgach
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // febdghac
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v6, v7); // febdgahc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v6); // febdgach
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fedbaghc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fedbagch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fedbacgh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // febdaghc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // febdagch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v4, v5); // febdacgh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // febaghdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // febagdhc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // febadghc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // febagdch
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // febadgch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // febadcgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v5, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fbeghdac
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbegdhac
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbedghac
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fbegdahc
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v4, v6); swap(v6, v7); // fbedgahc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbedaghc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // fbegdach
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v4, v6); // fbedgach
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // fbedagch
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v4, v5); // fbedacgh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bfeghdac
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfegdhac
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfedghac
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bfegdahc
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v6, v7); // bfedgahc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfedaghc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // bfegdach
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v6); // bfedgach
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // bfedagch
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v4, v5); // bfedacgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); // fbeghadc
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fbegahdc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbegadhc
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fbegadch
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbeaghdc
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fbeagdhc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbeadghc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // fbeagdch
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fbeadgch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // fbeadcgh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); // bfeghadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bfegahdc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfegadhc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bfegadch
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfeaghdc
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bfeagdhc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfeadghc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // bfeagdch
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bfeadgch
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // bfeadcgh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v3, v1)) {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																if (compare(v7, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); // gfdehbac
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fgdehbac
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdgehbac
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gfdebhac
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgdebhac
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fdgebhac
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfdebahc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fgdebahc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v6, v7); // fdgebahc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gfdebach
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // fgdebach
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); // fdgebach
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fdeghbac
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdegbhac
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fdegbahc
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // fdegbach
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdebghac
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v6, v7); // fdebgahc
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v6); // fdebgach
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdebaghc
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // fdebagch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v4, v5); // fdebacgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); // gfdbehac
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fgdbehac
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v5, v7); swap(v6, v7); // fdgbehac
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gfdbeahc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v6, v7); // fgdbeahc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v6, v7); // fdgbeahc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // gfdbeach
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); // fgdbeach
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); // fdgbeach
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fdbgehac
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdbeghac
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v6, v7); // fdbgeahc
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fdbegahc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdbeaghc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); // fdbgeach
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // fdbegach
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fdbeagch
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fdbeacgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v3)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v5, v7); // gfbdehac
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v5, v7); swap(v6, v7); // fgbdehac
																		}
																	}
																	else {
																		swap(v0, v5); swap(v2, v6); swap(v5, v7); swap(v6, v7); // fbgdehac
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v5, v6); swap(v6, v7); // gfbdeahc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v6, v7); // fgbdeahc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v6, v7); // fbgdeahc
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v5, v6); // gfbdeach
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); // fgbdeach
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); // fbgdeach
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fbdgehac
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbdeghac
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v6, v7); // fbdgeahc
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fbdegahc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbdeaghc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); // fbdgeach
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // fbdegach
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fbdeagch
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fbdeacgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		swap(v0, v6); swap(v2, v5); swap(v5, v7); // gbfdehac
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v5, v7); // bgfdehac
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v5, v7); swap(v6, v7); // bfgdehac
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v5); swap(v5, v6); swap(v6, v7); // gbfdeahc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v5, v6); swap(v6, v7); // bgfdeahc
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v6, v7); // bfgdeahc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v5); swap(v5, v6); // gbfdeach
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v5, v6); // bgfdeach
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); // bfgdeach
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bfdgehac
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfdeghac
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v6, v7); // bfdgeahc
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bfdegahc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfdeaghc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); // bfdgeach
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // bfdegach
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bfdeagch
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bfdeacgh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																if (compare(v7, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gdfehbac
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgfehbac
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dfgehbac
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); // gdfebhac
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dgfebhac
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfgebhac
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gdfebahc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgfebahc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dfgebahc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // gdfebach
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dgfebach
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // dfgebach
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); // dfeghbac
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dfegbhac
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfegbahc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dfegbach
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfebghac
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dfebgahc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // dfebgach
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfebaghc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dfebagch
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // dfebacgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v3)) {
																		swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v5, v7); // gdfbehac
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); // dgfbehac
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dfgbehac
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v5, v6); swap(v6, v7); // gdfbeahc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dgfbeahc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v6, v7); // dfgbeahc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v5, v6); // gdfbeach
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // dgfbeach
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); // dfgbeach
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); // dfbgehac
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dfbeghac
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dfbgeahc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfbegahc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dfbeaghc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // dfbgeach
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dfbegach
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // dfbeagch
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v4); // dfbeacgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v6, v5)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); // gdbfehac
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dgbfehac
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dbgfehac
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gdbfeahc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v6, v7); // dgbfeahc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v6, v7); // dbgfeahc
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // gdbfeach
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); // dgbfeach
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); // dbgfeach
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); // dbfgehac
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dbfeghac
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dbfgeahc
																			}
																			else {
																				swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbfegahc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dbfeaghc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // dbfgeach
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dbfegach
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // dbfeagch
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v4); // dbfeacgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); // gbdfehac
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); // bgdfehac
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bdgfehac
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gbdfeahc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bgdfeahc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v6, v7); // bdgfeahc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // gbdfeach
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // bgdfeach
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); // bdgfeach
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); // bdfgehac
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bdfeghac
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bdfgeahc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdfegahc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bdfeaghc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // bdfgeach
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bdfegach
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // bdfeagch
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v4); // bdfeacgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v0)) {
												if (compare(v6, v1)) {
													if (compare(v5, v1)) {
														if (compare(v3, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfdbaehc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fgdbaehc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v6, v7); // fdgbaehc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gfdbaech
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // fgdbaech
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v5); // fdgbaech
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // gfdbaceh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fgdbaceh
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // fdgbaceh
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdfbaehc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgfbaehc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dfgbaehc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // gdfbaech
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dgfbaech
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // dfgbaech
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v4, v6); // gdfbaceh
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // dgfbaceh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dfgbaceh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); // gfbaehdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gfbaedhc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfbdaehc
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfbadehc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // gfbdaech
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v4, v6); // gfbdaceh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // gfbaedch
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gfbadech
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // gfbadceh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fgbaehdc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v6, v7); // fgbaedhc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v6, v7); // fgbdaehc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fgbadehc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v5); // fgbdaech
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // fgbdaceh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); // fgbaedch
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // fgbadech
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fgbadceh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v5, v3)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); // gbfaehdc
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gbfaedhc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbfdaehc
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbfadehc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // gbfdaech
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v4, v6); // gbfdaceh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // gbfaedch
																			}
																			else {
																				swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gbfadech
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // gbfadceh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdbfaehc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gdbfaech
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // gdbfaceh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dgbfaehc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // dgbfaech
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dgbfaceh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbdfaehc
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gbdfaech
																		}
																	}
																	else {
																		swap(v0, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // gbdfaceh
																	}
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gdbafehc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v6, v7); // dgbafehc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gbdafehc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // gdbafech
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v5); // dgbafech
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // gbdafech
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gdbafceh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // dgbafceh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gbdafceh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // gdbacfeh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v6); // dgbacfeh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // gbdacfeh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gbafehdc
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v6, v7); // gbafedhc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gbafdehc
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v4, v5); swap(v6, v7); // gbadfehc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v5); // gbafedch
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // gbafdech
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v4, v5); // gbadfech
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gbafdceh
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // gbadfceh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v4, v6); // gbadcfeh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v0)) {
															if (compare(v5, v3)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); // bgfaehdc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bgfaedhc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgfdaehc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgfadehc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // bgfdaech
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v4, v6); // bgfdaceh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // bgfaedch
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bgfadech
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // bgfadceh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dbgfaehc
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // dbgfaech
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dbgfaceh
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgdfaehc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bgdfaech
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // bgdfaceh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bdgfaehc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // bdgfaech
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bdgfaceh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v4, v5); swap(v6, v7); // dbgafehc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bgdafehc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v6, v7); // bdgafehc
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v4, v5); // dbgafech
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // bgdafech
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v5); // bdgafech
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // dbgafceh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bgdafceh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // bdgafceh
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v4, v6); // dbgacfeh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // bgdacfeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v6); // bdgacfeh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bgafehdc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v6, v7); // bgafedhc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bgafdehc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v6, v7); // bgadfehc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); // bgafedch
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // bgafdech
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v5); // bgadfech
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bgafdceh
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // bgadfceh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v6); // bgadcfeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v3)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fbgaehdc
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v6, v7); // fbgaedhc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v5); swap(v2, v6); swap(v4, v5); swap(v6, v7); // fbgdaehc
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fbgadehc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v2, v6); swap(v4, v5); // fbgdaech
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // fbgdaceh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v3, v5); // fbgaedch
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // fbgadech
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fbgadceh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fdbgaehc
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // fdbgaech
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fdbgaceh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfbgaehc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dfbgaech
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // dfbgaceh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fbdgaehc
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // fbdgaech
																		}
																	}
																	else {
																		swap(v0, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fbdgaceh
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bfgaehdc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v6, v7); // bfgaedhc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v6, v7); // bfgdaehc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bfgadehc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v5); // bfgdaech
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // bfgdaceh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); // bfgaedch
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // bfgadech
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bfgadceh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbfgaehc
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dbfgaech
																		}
																	}
																	else {
																		swap(v0, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // dbfgaceh
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bfdgaehc
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // bfdgaech
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bfdgaceh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdfgaehc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bdfgaech
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // bdfgaceh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v0)) {
													if (compare(v3, v1)) {
														if (compare(v4, v2)) {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdbagehc
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fdbaeghc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fdbagech
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // fdbaegch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); // fdbaecgh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfbagehc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v5, v6); swap(v6, v7); // dfbaeghc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // dfbagech
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v5, v6); // dfbaegch
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); // dfbaecgh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v4)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbfagehc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dbagfehc
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbafgehc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // dbfagech
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // dbagfech
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dbafgech
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v5); swap(v5, v6); swap(v6, v7); // dbfaeghc
																			}
																			else {
																				swap(v0, v3); swap(v2, v5); swap(v5, v6); // dbfaegch
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); // dbfaecgh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dbafeghc
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // dbafegch
																			}
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v3, v5); // dbafecgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // fdbagceh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fdbacgeh
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fdbacegh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v6); // dfbagceh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // dfbacgeh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v5); swap(v4, v5); // dfbacegh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		swap(v0, v3); swap(v2, v5); swap(v4, v6); // dbfagceh
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // dbfacgeh
																		}
																		else {
																			swap(v0, v3); swap(v2, v5); swap(v4, v5); // dbfacegh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dbagfceh
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // dbafgceh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dbafcgeh
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dbafcegh
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // dbagcfeh
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v4, v6); // dbacgfeh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // dbacfgeh
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v4, v5); // dbacfegh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																if (compare(v5, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbdagehc
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fbdaeghc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fbdagech
																			}
																			else {
																				swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // fbdaegch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); // fbdaecgh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfdagehc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bfdaeghc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bfdagech
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // bfdaegch
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); // bfdaecgh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v4)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdfagehc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bdagfehc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdafgehc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // bdfagech
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // bdagfech
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bdafgech
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v5, v6); swap(v6, v7); // bdfaeghc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v5, v6); // bdfaegch
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); // bdfaecgh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bdafeghc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // bdafegch
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); // bdafecgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v2)) {
																		swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // fbdagceh
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fbdacgeh
																		}
																		else {
																			swap(v0, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fbdacegh
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // bfdagceh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bfdacgeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bfdacegh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v6); // bdfagceh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // bdfacgeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v5); swap(v4, v5); // bdfacegh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bdagfceh
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // bdafgceh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bdafcgeh
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bdafcegh
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // bdagcfeh
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v6); // bdacgfeh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // bdacfgeh
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v5); // bdacfegh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); // fbagehdc
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fbagedhc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbagdehc
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbadgehc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // fbagedch
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fbagdech
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // fbadgech
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // fbagdceh
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v4, v6); // fbadgceh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // fbadcgeh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fbaeghdc
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbaegdhc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fbaedghc
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v5, v6); swap(v6, v7); // fbadeghc
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fbaegdch
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // fbaedgch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v5, v6); // fbadegch
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v4); // fbaedcgh
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); // fbadecgh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v4, v5); // fbadcegh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); // bfagehdc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bfagedhc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfagdehc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfadgehc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // bfagedch
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bfagdech
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // bfadgech
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // bfagdceh
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v6); // bfadgceh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // bfadcgeh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bfaeghdc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfaegdhc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bfaedghc
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v5, v6); swap(v6, v7); // bfadeghc
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bfaegdch
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // bfaedgch
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v5, v6); // bfadegch
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); // bfaedcgh
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); // bfadecgh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v4, v5); // bfadcegh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bagfehdc
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v6, v7); // bagfedhc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v3, v5); // bagfedch
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v5, v7); // bafgehdc
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bafgedhc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v5, v6); // bafgedch
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bafeghdc
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bafegdhc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bafedghc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bafegdch
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v5, v6); // bafedgch
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v4); // bafedcgh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bagfdehc
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v5); // bagfdech
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bagfdceh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bafgdehc
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bafgdech
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v6); // bafgdceh
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bafdgehc
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v4, v6); swap(v5, v6); // bafdgech
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v5); swap(v4, v6); // bafdgceh
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v4, v5); swap(v5, v6); // bafdcgeh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v5); swap(v5, v6); swap(v6, v7); // bafdeghc
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v5, v6); // bafdegch
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v2, v5); // bafdecgh
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v4, v5); // bafdcegh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v6); swap(v4, v5); swap(v6, v7); // bagdfehc
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // badgfehc
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // badfgehc
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // badfeghc
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v6); swap(v4, v5); // bagdfech
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v5); // badgfech
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // badfgech
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v5, v6); // badfegch
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v5); // badfecgh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v6); swap(v4, v5); swap(v5, v6); // bagdfceh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // badgfceh
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v6); // badfgceh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // badfcgeh
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v5); swap(v4, v5); // badfcegh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v6); swap(v4, v6); // bagdcfeh
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v6); // badgcfeh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v4, v6); // badcgfeh
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v4, v5); swap(v5, v6); // badcfgeh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v4, v5); // badcfegh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v7, v6)) {
									if (compare(v7, v4)) {
										if (compare(v6, v0)) {
											if (compare(v6, v3)) {
												if (compare(v6, v1)) {
													if (compare(v5, v0)) {
														if (compare(v5, v1)) {
															if (compare(v3, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v4)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgefdbac
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hegfdbac
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hefgdbac
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgedfbac
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v6, v7); // hegdfbac
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgdefbac
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v6, v4)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hgefbdac
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v6, v7); // hegfbdac
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hefgbdac
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgefbadc
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hegfbadc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); // hefgbadc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v3, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgebfdac
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hegbfdac
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hgbefdac
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hgebfadc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hegbfadc
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgbefadc
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v6, v7); // hgedbfac
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v6, v7); // hegdbfac
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hgdebfac
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v6, v7); // hgdbefac
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hgebdfac
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v6, v7); // hegbdfac
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v6, v7); // hgbedfac
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v6, v7); // hgbdefac
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v3, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgedbafc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v5, v7); swap(v6, v7); // hegdbafc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgdebafc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hgdbeafc
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgdbaefc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgebdafc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hegbdafc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hgbedafc
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v5, v7); swap(v6, v7); // hgbdeafc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgbdaefc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); // hgebafdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hegbafdc
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hgebadfc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hegbadfc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hgbeafdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgbeadfc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v6, v7); // hgbaefdc
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hgbaedfc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgbadefc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v3, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); // hgedbacf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v5, v7); // hegdbacf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hgebdacf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); // hegbdacf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); // hgbedacf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgebadcf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hegbadcf
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hgbeadcf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); // hgbaedcf
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hgdebacf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); // hgdbeacf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hgdbaecf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgdbacef
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v5, v7); // hgbdeacf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); // hgbdaecf
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgbdacef
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hgbadecf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgbadcef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v0)) {
														if (compare(v5, v0)) {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hebgfdac
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hbgefdac
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbegfdac
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bhgefdac
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhegfdac
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hefbgdac
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hebfgdac
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hbefgdac
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bhefgdac
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v4)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v6, v7); // hbgedfac
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v6, v7); // hbgdefac
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hebgdfac
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hbegdfac
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v6, v7); // bhgedfac
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bhegdfac
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v6, v7); // bhgdefac
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hebgdafc
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hbgedafc
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbegdafc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bhgedafc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhegdafc
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hebgdacf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); // hbgedacf
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hbegdacf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); // bhgedacf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bhegdacf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v2, v6); swap(v5, v7); swap(v6, v7); // hbgdeafc
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v5, v7); // hbgdeacf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbgdaefc
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); // hbgdaecf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbgdacef
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v5, v7); swap(v6, v7); // bhgdeafc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v5, v7); // bhgdeacf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhgdaefc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); // bhgdaecf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhgdacef
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hebgfadc
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbgefadc
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hbegfadc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhgefadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bhegfadc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hefbgadc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hebfgadc
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbefgadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhefgadc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v4)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hbgeafdc
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v6, v7); // hbgaefdc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); // hebgafdc
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); // hbegafdc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bhgeafdc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); // bhegafdc
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v6, v7); // bhgaefdc
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hebgadfc
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbgeadfc
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hbegadfc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhgeadfc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bhegadfc
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hebgadcf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hbgeadcf
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbegadcf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bhgeadcf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhegadcf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hbgaedfc
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbgadefc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); // hbgaedcf
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hbgadecf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbgadcef
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bhgaedfc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhgadefc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); // bhgaedcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bhgadecf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhgadcef
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v4)) {
													if (compare(v7, v3)) {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hdgefbac
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v6, v7); // hdgebfac
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v6, v7); // hdgbefac
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v6, v7); // hdbgefac
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v6, v7); // hbdgefac
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v6, v7); // bhdgefac
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hdgebafc
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v5, v7); swap(v6, v7); // hdgbeafc
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hdbgeafc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hbdgeafc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bhdgeafc
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdgbaefc
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdbgaefc
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbdgaefc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhdgaefc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); // hdgebacf
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v5, v7); // hdgbeacf
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); // hdbgeacf
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); // hdgbaecf
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdgbacef
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hdbgaecf
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdbgacef
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v0)) {
																		swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); // hbdgeacf
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hbdgaecf
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbdgacef
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); // bhdgeacf
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bhdgaecf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhdgacef
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhgefbac
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dhgebfac
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v6, v7); // dhgbefac
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); // dhbgefac
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); // dbhgefac
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); // bdhgefac
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhgebafc
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dhgbeafc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); // dhbgeafc
																			}
																			else {
																				swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); // dbhgeafc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); // bdhgeafc
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhgbaefc
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dhbgaefc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dbhgaefc
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bdhgaefc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dhgebacf
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); // dhgbeacf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dhbgeacf
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dhgbaecf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhgbacef
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhbgaecf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dhbgacef
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v0)) {
																		swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dbhgeacf
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbhgaecf
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dbhgacef
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bdhgeacf
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdhgaecf
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bdhgacef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v4, v3)) {
															if (compare(v6, v5)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hedgfbac
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v6, v7); // hedgbfac
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hedbgfac
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v6, v7); // hebdgfac
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hedgbafc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hedbgafc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hebdgafc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); // hedgbacf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hedbgacf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); // hebdgacf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hefdgbac
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v5, v6); swap(v6, v7); // hefdbgac
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hefbdgac
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hebfdgac
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hedfgbac
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hedfbgac
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hedbfgac
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hebdfgac
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdegfbac
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hdefgbac
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hdegbfac
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdegbafc
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hdegbacf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dhegfbac
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhefgbac
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); // dhegbfac
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dhegbafc
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhegbacf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdefbgac
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v6, v7); // hdebgfac
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdebfgac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdebgafc
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); // hdebgacf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dhefbgac
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dhebgfac
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhebfgac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhebgafc
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dhebgacf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hdbegfac
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdbefgac
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); // dhbegfac
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dhbefgac
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbefdgac
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v4); swap(v4, v6); swap(v6, v7); // hbedgfac
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbedfgac
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hbdegfac
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbdefgac
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdbegafc
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dhbegafc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hdbegacf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhbegacf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbedgafc
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbdegafc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); // hbedgacf
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hbdegacf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhefdgac
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v6, v7); // bhedgfac
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhedfgac
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bhdegfac
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhdefgac
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); // dbhegfac
																		}
																		else {
																			swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dbhefgac
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); // bdhegfac
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bdhefgac
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhedgafc
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhdegafc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dbhegafc
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bdhegafc
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); // bhedgacf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bhdegacf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbhegacf
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdhegacf
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v1)) {
												if (compare(v4, v3)) {
													if (compare(v4, v1)) {
														if (compare(v3, v0)) {
															if (compare(v5, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v5, v7); // hefdbagc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); // hedfbagc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hedbfagc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v5, v7); swap(v6, v7); // hefdbacg
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hedfbacg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hedbfacg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); // hefbdagc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hebfdagc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); // hebdfagc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hefbdacg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hebfdacg
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hebdfacg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hedbagfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); // hedbafgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hedbafcg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hedbagcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hedbacgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hedbacfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); // hebdagfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v7); // hebdafgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v6, v7); // hebdafcg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hebdagcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); // hebdacgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hebdacfg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v5, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hefbagdc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hefbadgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hefbadcg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hebfagdc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); // hebfadgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hebfadcg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hebagfdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hebafgdc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hebafdgc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hebafdcg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hebagdfc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hebadgfc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); // hebadfgc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hebadfcg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hebagdcf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hebadgcf
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hebadcgf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hebadcfg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v4, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hbeagfdc
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbeagdfc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); // hbedagfc
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hbeadgfc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbedagcf
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); // hbedacgf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hbeagdcf
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbeadgcf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hbeadcgf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); // hbagefdc
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); // hbaegfdc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); // hbagedfc
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hbaegdfc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); // hbaedgfc
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hbagedcf
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbaegdcf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hbaedgcf
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); // hbaedcgf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v3, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hbefdagc
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); // hbedfagc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hbefagdc
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); // hbefadgc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbeafgdc
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hbeafdgc
																			}
																		}
																		else {
																			if (compare(v3, v0)) {
																				swap(v0, v7); swap(v2, v4); swap(v4, v7); // hbedafgc
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); // hbeadfgc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hbaefgdc
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); // hbaefdgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v4); // hbaedfgc
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbefdacg
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbedfacg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v2, v4); swap(v4, v7); swap(v6, v7); // hbedafcg
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbedacfg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			if (compare(v5, v0)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hbefadcg
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbeafdcg
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hbeadfcg
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbeadcfg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hbaefdcg
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v6, v7); // hbaedfcg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hbaedcfg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v0)) {
															if (compare(v5, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hdefbagc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); // dhefbagc
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); // hdebfagc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dhebfagc
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hdbefagc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); // dhbefagc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hbdefagc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdefbacg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dhefbacg
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdebfacg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhebfacg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdbefacg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dhbefacg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbdefacg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); // hdebagfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v7); // hdebafgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v6, v7); // hdebafcg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dhebagfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); // dhebafgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dhebafcg
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v6, v2)) {
																				if (compare(v6, v5)) {
																					swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hdbeagfc
																				}
																				else {
																					swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); // hdbeafgc
																				}
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hdbeafcg
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				if (compare(v6, v5)) {
																					swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); // dhbeagfc
																				}
																				else {
																					swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); // dhbeafgc
																				}
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v6, v7); // dhbeafcg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hbdeagfc
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); // hbdeafgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hbdeafcg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hdbagefc
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); // hdbaegfc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); // hdbaefgc
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v6, v7); // hdbaefcg
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); // dhbagefc
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v5, v6); // dhbaegfc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); // dhbaefgc
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v6, v7); // dhbaefcg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hbdagefc
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); // hbdaegfc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); // hbdaefgc
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v6, v7); // hbdaefcg
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hbagdefc
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); // hbadgefc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v7); swap(v5, v6); // hbadegfc
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); // hbadefgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v6, v7); // hbadefcg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v7, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdebagcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); // hdebacgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdebacfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdbeagcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hdbeacgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdbeacfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdbagecf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hdbaegcf
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); // hdbaecgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hdbaecfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hdbagcef
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdbacgef
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hdbacegf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdbacefg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhebagcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dhebacgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhebacfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dhbeagcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); // dhbeacgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dhbeacfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhbagecf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v5, v6); swap(v6, v7); // dhbaegcf
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v5, v7); // dhbaecgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v5, v7); swap(v6, v7); // dhbaecfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); // dhbagcef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhbacgef
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); // dhbacegf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhbacefg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbdeagcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hbdeacgf
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbdeacfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbdagecf
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hbdaegcf
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); // hbdaecgf
																			}
																			else {
																				swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hbdaecfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hbdagcef
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbdacgef
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hbdacegf
																		}
																		else {
																			swap(v0, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbdacefg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbagdecf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbadgecf
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v5, v6); swap(v6, v7); // hbadegcf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v2, v7); swap(v5, v7); // hbadecgf
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v5, v7); swap(v6, v7); // hbadecfg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hbagdcef
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); // hbadgcef
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbadcgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); // hbadcegf
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbadcefg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v0)) {
													if (compare(v6, v2)) {
														if (compare(v5, v2)) {
															if (compare(v6, v5)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); // dbheagfc
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); // dbhagefc
																			}
																			else {
																				swap(v0, v3); swap(v2, v7); swap(v5, v6); // dbhaegfc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dbahgefc
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); // dbahegfc
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); // bhedagfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bhdeagfc
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bhdagefc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); // bhdaegfc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v4, v0)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); // bdheagfc
																			}
																			else {
																				if (compare(v6, v4)) {
																					swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); // bdhagefc
																				}
																				else {
																					swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v5, v6); // bdhaegfc
																				}
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bdahgefc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); // bdahegfc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bhefdagc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); // bhedfagc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bhdefagc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v7); // bhedafgc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); // bhdeafgc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); // bhdaefgc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); // dbhefagc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); // bdhefagc
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v4); // dbheafgc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); // bdheafgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v7); // dbhaefgc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); // bdhaefgc
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v7); // dbahefgc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); // bdahefgc
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbhagecf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhdagecf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdhagecf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbahgecf
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdahgecf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); // dbhagcef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bhdagcef
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); // bdhagcef
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dbahgcef
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bdahgcef
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhedagcf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhdeagcf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dbheagcf
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bdheagcf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v5, v6); swap(v6, v7); // dbhaegcf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bhdaegcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v5, v6); swap(v6, v7); // bdhaegcf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dbahegcf
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bdahegcf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v7, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v3)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); // bhedacgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bhdeacgf
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); // dbheacgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); // bdheacgf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v5, v7); // dbhaecgf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); // bhdaecgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v5, v7); // bdhaecgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbhacgef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhdacgef
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdhacgef
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); // dbhacegf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bhdacegf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); // bdhacegf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); // dbahecgf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbahcgef
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dbahcegf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbachgef
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); // dbachegf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); // bdahecgf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdahcgef
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bdahcegf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdachgef
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); // bdachegf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v7, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhefdacg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhedfacg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhdefacg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v6, v7); // bhedafcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bhdeafcg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v6, v7); // bhdaefcg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dbhefacg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bdhefacg
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v6, v7); // dbheafcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v6, v7); // bdheafcg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v7); swap(v6, v7); // dbhaefcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v6, v7); // bdhaefcg
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v6, v7); // dbahefcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v6, v7); // bdahefcg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v3)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhedacfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhdeacfg
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dbheacfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bdheacfg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v7); swap(v5, v7); swap(v6, v7); // dbhaecfg
																			}
																			else {
																				if (compare(v7, v3)) {
																					swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bhdaecfg
																				}
																				else {
																					swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v5, v7); swap(v6, v7); // bdhaecfg
																				}
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dbahecfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bdahecfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbhacefg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhdacefg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdhacefg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbahcefg
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbachefg
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdahcefg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdachefg
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v0)) {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bhefagdc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bheagfdc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bheafgdc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); // bhagefdc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); // bhaegfdc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bhaefgdc
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); // bhefadgc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bheafdgc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); // bhaefdgc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bhefadcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bheafdcg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bhaefdcg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); // bhagedfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bhagdefc
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bheagdfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bhaegdfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v4, v0)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bheadgfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); // bhaedgfc
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); // bhadgefc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v5, v6); // bhadegfc
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); // bheadfgc
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); // bhaedfgc
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); // bhadefgc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bheadfcg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v6, v7); // bhaedfcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v6, v7); // bhadefcg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bheagdcf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bheadgcf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bhagedcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhaegdcf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bhaedgcf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhagdecf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhadgecf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v5, v6); swap(v6, v7); // bhadegcf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bhagdcef
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); // bhadgcef
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bheadcgf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bheadcfg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); // bhaedcgf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bhaedcfg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v5, v7); // bhadecgf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v5, v7); swap(v6, v7); // bhadecfg
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhadcgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); // bhadcegf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhadcefg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); // bahgefdc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v6); // bahegfdc
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bahefgdc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v5); // bahefdgc
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bahefdcg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v5, v6); // bahgedfc
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bahegdfc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v5, v6); // bahedgfc
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v4); // bahedfgc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v6, v7); // bahedfcg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bahgdefc
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v4, v6); swap(v5, v6); // bahdgefc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v7); swap(v5, v6); // bahdegfc
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); // bahdefgc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v6, v7); // bahdefcg
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // badhgefc
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v5, v6); // badhegfc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v3); swap(v3, v7); // badhefgc
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v6, v7); // badhefcg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bahgedcf
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bahgdecf
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bahdgecf
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // badhgecf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bahegdcf
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bahedgcf
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v5, v6); swap(v6, v7); // bahdegcf
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // badhegcf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v5, v7); // bahedcgf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v5, v7); // bahdecgf
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v5, v7); // badhecgf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bahedcfg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v5, v7); swap(v6, v7); // bahdecfg
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // badhecfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v4)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bahgdcef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v4, v6); swap(v5, v7); // bahdgcef
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // badhgcef
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bahdcgef
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // badhcgef
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // badchgef
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v4, v7); swap(v5, v7); // bahdcegf
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // badhcegf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v4, v7); swap(v5, v7); // badchegf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bahdcefg
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // badhcefg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // badchefg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v5, v0)) {
											if (compare(v5, v3)) {
												if (compare(v5, v1)) {
													if (compare(v6, v0)) {
														if (compare(v6, v1)) {
															if (compare(v3, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v6, v5)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehgfdbac
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ehfgdbac
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); // efhgdbac
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); // ehfdgbac
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v6); // efhdgbac
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); // efdhgbac
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v6, v5)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ehgfbdac
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehfgbdac
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // efhgbdac
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehgfbadc
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ehfgbadc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); // efhgbadc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v3, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ehfbgdac
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); // efhbgdac
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // efbhgdac
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehfbgadc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // efhbgadc
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efbhgadc
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehfdbgac
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); // efhdbgac
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // efdhbgac
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efdbhgac
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehfbdgac
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // efhbdgac
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efbhdgac
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efbdhgac
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v3, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); // ehfdbagc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v5); // efhdbagc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); // efdhbagc
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // efdbhagc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); // efdbahgc
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ehfbdagc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); // efhbdagc
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // efbhdagc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); // efbdhagc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v5, v7); // efbdahgc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ehfbagdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); // efhbagdc
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); // ehfbadgc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); // efhbadgc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // efbhagdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); // efbhadgc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efbahgdc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // efbahdgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); // efbadhgc
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v3, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehfdbacg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v6, v7); // efhdbacg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehfbdacg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // efhbdacg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efbhdacg
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ehfbadcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v6, v7); // efhbadcg
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // efbhadcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efbahdcg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // efdhbacg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efdbhacg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // efdbahcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v6, v7); // efdbachg
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efbdhacg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v5, v7); swap(v6, v7); // efbdahcg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v6, v7); // efbdachg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // efbadhcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v6, v7); // efbadchg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v0)) {
														if (compare(v6, v0)) {
															if (compare(v6, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehgbfdac
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ehbgfdac
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ebhgfdac
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // behgfdac
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); // ehbfgdac
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); // ebhfgdac
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ebfhgdac
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); // behfgdac
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // befhgdac
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ehbfdgac
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ebhfdgac
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebfhdgac
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebfdhgac
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // behfdgac
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // befhdgac
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // befdhgac
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); // ehbfdagc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); // ebhfdagc
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ebfhdagc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); // behfdagc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // befhdagc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ehbfdacg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ebhfdacg
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebfhdacg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // behfdacg
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // befhdacg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); // ebfdhagc
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebfdhacg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v5); swap(v5, v7); // ebfdahgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v2, v5); swap(v5, v7); swap(v6, v7); // ebfdahcg
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v6, v7); // ebfdachg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); // befdhagc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // befdhacg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v5, v7); // befdahgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v5, v7); swap(v6, v7); // befdahcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v6, v7); // befdachg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v6, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ehgbfadc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); // ehbgfadc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); // ebhgfadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); // behgfadc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ehbfgadc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ebhfgadc
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebfhgadc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // behfgadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // befhgadc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); // ehbfagdc
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); // ebhfagdc
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ebfhagdc
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebfahgdc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); // behfagdc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // befhagdc
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // befahgdc
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); // ehbfadgc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); // ebhfadgc
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); // ebfhadgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); // behfadgc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); // befhadgc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v6, v7); // ehbfadcg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v6, v7); // ebhfadcg
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ebfhadcg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v6, v7); // behfadcg
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // befhadcg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ebfahdgc
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); // ebfadhgc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebfahdcg
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ebfadhcg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v6, v7); // ebfadchg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // befahdgc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); // befadhgc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // befahdcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // befadhcg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v6, v7); // befadchg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v5)) {
													if (compare(v7, v1)) {
														if (compare(v7, v3)) {
															if (compare(v6, v5)) {
																if (compare(v6, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehgdfbac
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v6, v7); // ehgdbfac
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ehgbdfac
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); // ehbgdfac
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehdgfbac
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ehdgbfac
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); // ehdbgfac
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v6); // ehbdgfac
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ehdfgbac
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehdfbgac
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ehdbfgac
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); // ehbdfgac
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v1)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ehdfbagc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); // ehdbfagc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v5); // ehbdfagc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v1)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehdfbacg
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ehdbfacg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v6, v7); // ehbdfacg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // edhgfbac
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); // edhfgbac
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // edhfbgac
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); // edhfbagc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // edhfbacg
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); // dehgfbac
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dehfgbac
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); // dehfbgac
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); // dehfbagc
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v6, v7); // dehfbacg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); // edhgbfac
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v6); // edhbgfac
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); // edhbfgac
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v5); // edhbfagc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v6, v7); // edhbfacg
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); // dehgbfac
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); // dehbgfac
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dehbfgac
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); // dehbfagc
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dehbfacg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); // edbhgfac
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // edbhfgac
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // debhgfac
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // debhfgac
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); // ebhgdfac
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v4, v6); // ebhdgfac
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); // ebhdfgac
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); // ebdhgfac
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ebdhfgac
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); // edbhfagc
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // debhfagc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // edbhfacg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // debhfacg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v4, v5); // ebhdfagc
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); // ebdhfagc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v4, v5); swap(v6, v7); // ebhdfacg
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ebdhfacg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); // behgdfac
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v6); // behdgfac
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); // behdfgac
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); // bedhgfac
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bedhfgac
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dbehgfac
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbehfgac
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bdehgfac
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdehfgac
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v5); // behdfagc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); // bedhfagc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dbehfagc
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bdehfagc
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v6, v7); // behdfacg
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bedhfacg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbehfacg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdehfacg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v3)) {
														if (compare(v6, v2)) {
															if (compare(v6, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v1)) {
																		swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // edfhgbac
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edfhbgac
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edfbhgac
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edbfhgac
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebdfhgac
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bedfhgac
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // edfhbagc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); // edfbhagc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // edbfhagc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ebdfhagc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bedfhagc
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v5, v7); // edfbahgc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); // edbfahgc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); // ebdfahgc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); // bedfahgc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edfhbacg
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edfbhacg
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edbfhacg
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v5, v7); swap(v6, v7); // edfbahcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v6, v7); // edfbachg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // edbfahcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v6, v7); // edbfachg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v0)) {
																		swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebdfhacg
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ebdfahcg
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v6, v7); // ebdfachg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bedfhacg
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bedfahcg
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v6, v7); // bedfachg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v6, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v1)) {
																		swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // defhgbac
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // defhbgac
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // defbhgac
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // debfhgac
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dbefhgac
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bdefhgac
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); // defhbagc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // defbhagc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); // debfhagc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); // dbefhagc
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); // bdefhagc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); // defbahgc
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // debfahgc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dbefahgc
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bdefahgc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // defhbacg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // defbhacg
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // debfhacg
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // defbahcg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v6, v7); // defbachg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // debfahcg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // debfachg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v0)) {
																		swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dbefhacg
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbefahcg
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dbefachg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bdefhacg
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdefahcg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bdefachg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v1)) {
												if (compare(v7, v3)) {
													if (compare(v7, v1)) {
														if (compare(v3, v0)) {
															if (compare(v6, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehgdbafc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehdgbafc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ehdbgafc
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); // ehgdbacf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ehdgbacf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehdbgacf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehgbdafc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ehbgdafc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); // ehbdgafc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ehgbdacf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehbgdacf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehbdgacf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); // ehdbagfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); // ehdbafgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v6, v7); // ehdbafcg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ehdbagcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); // ehdbacgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ehdbacfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v5, v6); // ehbdagfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); // ehbdafgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v6, v7); // ehbdafcg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v5, v6); swap(v6, v7); // ehbdagcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v5, v7); // ehbdacgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v5, v7); swap(v6, v7); // ehbdacfg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v6, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v6, v7); // ehgbafdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ehgbadfc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); // ehgbadcf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); // ehbgafdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); // ehbgadfc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ehbgadcf
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); // ehbagfdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ehbafgdc
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ehbagdfc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehbagdcf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); // ehbadgfc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ehbadgcf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); // ehbafdgc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); // ehbadfgc
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ehbafdcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v6, v7); // ehbadfcg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); // ehbadcgf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ehbadcfg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v5, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v3, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ebhgdafc
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); // ebhdgafc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); // ebhgafdc
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); // ebhgadfc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); // ebhagfdc
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ebhagdfc
																			}
																		}
																		else {
																			if (compare(v3, v0)) {
																				swap(v0, v4); swap(v2, v7); swap(v5, v6); // ebhdagfc
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); // ebhadgfc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ebahgfdc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebahgdfc
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ebahdgfc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebhgdacf
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebhdgacf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v7); swap(v5, v6); swap(v6, v7); // ebhdagcf
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v5, v7); // ebhdacgf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v0)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ebhgadcf
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebhagdcf
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ebhadgcf
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); // ebhadcgf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ebahgdcf
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebahdgcf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ebahdcgf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ebhafgdc
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); // ebhafdgc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v4); swap(v2, v7); // ebhdafgc
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v4); // ebhadfgc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v7); swap(v6, v7); // ebhdafcg
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v5, v7); swap(v6, v7); // ebhdacfg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ebhafdcg
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v6, v7); // ebhadfcg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ebhadcfg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebahfgdc
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ebafhgdc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ebahfdgc
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); // ebafhdgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); // ebahdfgc
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebahfdcg
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ebafhdcg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ebahdfcg
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebahdcfg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v7, v0)) {
															if (compare(v7, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // edhgbafc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); // edhbgafc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v5, v6); // edhbagfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); // edhbafgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v6, v7); // edhbafcg
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); // dehgbafc
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dehbgafc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); // dehbagfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); // dehbafgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v6, v7); // dehbafcg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // edbhgafc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // debhgafc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ebdhgafc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); // edbhagfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // debhagfc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); // ebdhagfc
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); // edbhafgc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); // debhafgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); // ebdhafgc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v6, v7); // edbhafcg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // debhafcg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v6, v7); // ebdhafcg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // edbahgfc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); // edbahfgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // edbafhgc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); // debahgfc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v7); // debahfgc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); // debafhgc
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ebdahgfc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); // ebdahfgc
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ebdafhgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); // ebadhgfc
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v4, v7); // ebadhfgc
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ebafdhgc
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); // ebadfhgc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // edbahfcg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edbafhcg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // edbafchg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v6, v7); // debahfcg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // debafhcg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v6, v7); // debafchg
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ebdahfcg
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebdafhcg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ebdafchg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebafdhcg
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v4); swap(v2, v4); swap(v4, v7); swap(v6, v7); // ebadhfcg
																				}
																				else {
																					swap(v0, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebadfhcg
																				}
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ebafdchg
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v4, v5); swap(v6, v7); // ebadfchg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v0)) {
															if (compare(v7, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edhgbacf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edhbgacf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v5, v6); swap(v6, v7); // edhbagcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v5, v7); // edhbacgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v5, v7); swap(v6, v7); // edhbacfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dehgbacf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dehbgacf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dehbagcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); // dehbacgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dehbacfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edbhgacf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // debhgacf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebdhgacf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // edbhagcf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // debhagcf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ebdhagcf
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); // edbhacgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // debhacgf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); // ebdhacgf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // edbhacfg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // debhacfg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ebdhacfg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edbahgcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // edbahcgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edbahcfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // debahgcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); // debahcgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // debahcfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); // edbachgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // edbachfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v6, v7); // edbacfhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v5, v7); // debachgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v5, v7); swap(v6, v7); // debachfg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v6, v7); // debacfhg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebdahgcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ebdahcgf
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebdahcfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); // ebdachgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ebdachfg
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v6, v7); // ebdacfhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebadhgcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); // ebadhcgf
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebadhcfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v2, v4); swap(v5, v7); // ebadchgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v5, v7); swap(v6, v7); // ebadchfg
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v6, v7); // ebadcfhg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v0)) {
													if (compare(v5, v2)) {
														if (compare(v6, v2)) {
															if (compare(v6, v5)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // behgdafc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); // behdgafc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bedhgafc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v5, v6); // behdagfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); // bedhagfc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bedahgfc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbehgafc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdehgafc
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dbehagfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bdehagfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); // dbeahgfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); // bdeahgfc
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dbaehgfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bdaehgfc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); // dbehafgc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v4); swap(v4, v7); // dbeahfgc
																			}
																			else {
																				swap(v0, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); // dbeafhgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); // dbaehfgc
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dbaefhgc
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); // behdafgc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); // bedhafgc
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); // bedahfgc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bedafhgc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); // bdehafgc
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v7); // bdeahfgc
																				}
																				else {
																					swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); // bdeafhgc
																				}
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); // bdaehfgc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bdaefhgc
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v7, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v6, v7); // behdafcg
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v6, v7); // bedhafcg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dbehafcg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bdehafcg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v4, v7); swap(v6, v7); // dbeahfcg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bedahfcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v6, v7); // bdeahfcg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dbaehfcg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bdaehfcg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbeafhcg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bedafhcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdeafhcg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbaefhcg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdaefhcg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v4, v5); swap(v6, v7); // dbeafchg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bedafchg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v6, v7); // bdeafchg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dbaefchg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bdaefchg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v6, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // behgdacf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // behdgacf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bedhgacf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v5, v6); swap(v6, v7); // behdagcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bedhagcf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bedahgcf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dbehgacf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bdehgacf
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbehagcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdehagcf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbeahgcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdeahgcf
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbaehgcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdaehgcf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v5, v7); // behdacgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); // bedhacgf
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dbehacgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bdehacgf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); // dbeahcgf
																			}
																			else {
																				if (compare(v4, v3)) {
																					swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bedahcgf
																				}
																				else {
																					swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); // bdeahcgf
																				}
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dbaehcgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bdaehcgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v5, v7); // dbeachgf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); // bedachgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v5, v7); // bdeachgf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); // dbaechgf
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v5, v7); // dbacehgf
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); // bdaechgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v5, v7); // bdacehgf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v5, v7); swap(v6, v7); // behdacfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bedhacfg
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbehacfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdehacfg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbeahcfg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bedahcfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdeahcfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v5, v7); swap(v6, v7); // dbeachfg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bedachfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v5, v7); swap(v6, v7); // bdeachfg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v6, v7); // dbeacfhg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v6, v7); // bedacfhg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v6, v7); // bdeacfhg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbaehcfg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dbaechfg
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v6, v7); // dbaecfhg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v2, v3); swap(v5, v7); swap(v6, v7); // dbacehfg
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v6, v7); // dbacefhg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdaehcfg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bdaechfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v6, v7); // bdaecfhg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v5, v7); swap(v6, v7); // bdacehfg
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v6, v7); // bdacefhg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v2)) {
															if (compare(v6, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); // behgafdc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); // behagfdc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // behafgdc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // beahgfdc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // beahfgdc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // beafhgdc
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); // behgadfc
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // behagdfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // beahgdfc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // behgadcf
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // behagdcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // beahgdcf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); // behadgfc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // beahdgfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); // beadhgfc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // behadgcf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // beahdgcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // beadhgcf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); // behafdgc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // beahfdgc
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); // beafhdgc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // beafdhgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); // behadfgc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); // beahdfgc
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v7); // beadhfgc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); // beadfhgc
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // behafdcg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // beahfdcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // beafhdcg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // beafdhcg
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // beafdchg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v6, v7); // behadfcg
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // beahdfcg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v6, v7); // beadhfcg
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // beadfhcg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v6, v7); // beadfchg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); // behadcgf
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // behadcfg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // beahdcgf
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // beahdcfg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); // beadhcgf
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // beadhcfg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v5, v7); // beadchgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v5, v7); swap(v6, v7); // beadchfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v6, v7); // beadcfhg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v7, v5)) {
																if (compare(v7, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // baehgfdc
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // baehfgdc
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // baehfdgc
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // baehfdcg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // baehgdfc
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // baehdgfc
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v7); // baehdfgc
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // baehdfcg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v4); swap(v4, v7); swap(v5, v6); // baedhgfc
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v4, v7); // baedhfgc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v4, v7); swap(v6, v7); // baedhfcg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // badehgfc
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v7); // badehfgc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // badehfcg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // baefhgdc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v7); // baefhdgc
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // baefdhgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // baefhdcg
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // baefdhcg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // baefdchg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v4); swap(v4, v5); swap(v5, v7); // baedfhgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // baedfhcg
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v4, v5); swap(v6, v7); // baedfchg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // badefhgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // badefhcg
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // badefchg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // baehgdcf
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // baehdgcf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // baedhgcf
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // badehgcf
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // baehdcgf
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v4, v7); swap(v5, v7); // baedhcgf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // badehcgf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // baehdcfg
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // baedhcfg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // badehcfg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v2, v4); swap(v5, v7); // baedchgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v4); swap(v5, v7); swap(v6, v7); // baedchfg
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v6, v7); // baedcfhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v5, v7); // badechgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // badechfg
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v6, v7); // badecfhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		swap(v0, v1); swap(v2, v3); swap(v5, v7); // badcehgf
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v2, v3); swap(v5, v7); swap(v6, v7); // badcehfg
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v6, v7); // badcefhg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v4)) {
										if (compare(v7, v0)) {
											if (compare(v7, v3)) {
												if (compare(v7, v1)) {
													if (compare(v5, v0)) {
														if (compare(v5, v1)) {
															if (compare(v3, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v4)) {
																		swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ghefdbac
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); // gehfdbac
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gefhdbac
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); // ghedfbac
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v5); // gehdfbac
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ghdefbac
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v7, v4)) {
																		swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); // ghefbdac
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); // gehfbdac
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); // gefhbdac
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ghefbadc
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); // gehfbadc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gefhbadc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v3, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ghebfdac
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); // gehbfdac
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); // ghbefdac
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghebfadc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gehbfadc
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ghbefadc
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v7); // ghedbfac
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); // gehdbfac
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); // ghdebfac
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); // ghdbefac
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); // ghebdfac
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); // gehbdfac
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); // ghbedfac
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); // ghbdefac
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v3, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); // ghedbafc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v5, v6); // gehdbafc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ghdebafc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); // ghdbeafc
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ghdbaefc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ghebdafc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); // gehbdafc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); // ghbedafc
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v5, v6); // ghbdeafc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); // ghbdaefc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ghebafdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); // gehbafdc
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghebadfc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gehbadfc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); // ghbeafdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ghbeadfc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); // ghbaefdc
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); // ghbaedfc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ghbadefc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v3, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghedbacf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v5, v6); swap(v6, v7); // gehdbacf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghebdacf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gehbdacf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ghbedacf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ghebadcf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gehbadcf
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghbeadcf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ghbaedcf
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghdebacf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ghdbeacf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghdbaecf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ghdbacef
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v5, v6); swap(v6, v7); // ghbdeacf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghbdaecf
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); // ghbdacef
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghbadecf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ghbadcef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v0)) {
														if (compare(v5, v0)) {
															if (compare(v5, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gebhfdac
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); // gbhefdac
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gbehfdac
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); // bghefdac
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bgehfdac
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gefbhdac
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); // gebfhdac
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); // gbefhdac
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); // bgefhdac
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v4)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v4); // gbhedfac
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); // gbhdefac
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); // gebhdfac
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); // gbehdfac
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); // bghedfac
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); // bgehdfac
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v6); swap(v2, v7); // bghdefac
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gebhdafc
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); // gbhedafc
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gbehdafc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); // bghedafc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bgehdafc
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gebhdacf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gbhedacf
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbehdacf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bghedacf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgehdacf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v2, v7); swap(v5, v6); // gbhdeafc
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v5, v6); swap(v6, v7); // gbhdeacf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); // gbhdaefc
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbhdaecf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); // gbhdacef
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v5, v6); // bghdeafc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v5, v6); swap(v6, v7); // bghdeacf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); // bghdaefc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bghdaecf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); // bghdacef
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gebhfadc
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gbhefadc
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbehfadc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bghefadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgehfadc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gefbhadc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gebfhadc
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gbefhadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bgefhadc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v4)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); // gbheafdc
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v6); // gbhaefdc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gebhafdc
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gbehafdc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); // bgheafdc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bgehafdc
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); // bghaefdc
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gebhadfc
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gbheadfc
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbehadfc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bgheadfc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgehadfc
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gebhadcf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbheadcf
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gbehadcf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgheadcf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bgehadcf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); // gbhaedfc
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gbhadefc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gbhaedcf
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbhadecf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gbhadcef
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); // bghaedfc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bghadefc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bghaedcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bghadecf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bghadcef
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v4)) {
													if (compare(v6, v3)) {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); // gdhefbac
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); // gdhebfac
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v3); swap(v2, v7); // gdhbefac
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); // gdbhefac
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v7); // gbdhefac
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); // bgdhefac
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); // gdhebafc
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v5, v6); // gdhbeafc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); // gdbheafc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); // gbdheafc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); // bgdheafc
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); // gdhbaefc
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gdbhaefc
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gbdhaefc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bgdhaefc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gdhebacf
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v5, v6); swap(v6, v7); // gdhbeacf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gdbheacf
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdhbaecf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); // gdhbacef
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdbhaecf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gdbhacef
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v0)) {
																		swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gbdheacf
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbdhaecf
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gbdhacef
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bgdheacf
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgdhaecf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bgdhacef
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dghefbac
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); // dghebfac
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); // dghbefac
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v6, v7); // dgbhefac
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v6, v7); // dbghefac
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v6, v7); // bdghefac
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dghebafc
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); // dghbeafc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dgbheafc
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dbgheafc
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bdgheafc
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dghbaefc
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgbhaefc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbghaefc
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdghaefc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dghebacf
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dghbeacf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); // dgbheacf
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dghbaecf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dghbacef
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dgbhaecf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgbhacef
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v0)) {
																		swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); // dbgheacf
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dbghaecf
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbghacef
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); // bdgheacf
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bdghaecf
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdghacef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v4, v3)) {
															if (compare(v7, v5)) {
																if (compare(v5, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); // gedhfbac
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); // gedhbfac
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); // gedbhfac
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v7); // gebdhfac
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); // gedhbafc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gedbhafc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); // gebdhafc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gedhbacf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gedbhacf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gebdhacf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); // gefdhbac
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v5, v7); // gefdbhac
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); // gefbdhac
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gebfdhac
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gedfhbac
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); // gedfbhac
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gedbfhac
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); // gebdfhac
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gdehfbac
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); // gdefhbac
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); // gdehbfac
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gdehbafc
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdehbacf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgehfbac
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dgefhbac
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dgehbfac
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgehbafc
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dgehbacf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gdefbhac
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v7); // gdebhfac
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); // gdebfhac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); // gdebhafc
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdebhacf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgefbhac
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); // dgebhfac
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dgebfhac
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dgebhafc
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgebhacf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); // gdbehfac
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gdbefhac
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dgbehfac
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgbefhac
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gbefdhac
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v4); swap(v4, v7); // gbedhfac
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); // gbedfhac
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); // gbdehfac
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gbdefhac
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gdbehafc
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgbehafc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdbehacf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dgbehacf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); // gbedhafc
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gbdehafc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbedhacf
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbdehacf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bgefdhac
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v7); // bgedhfac
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); // bgedfhac
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); // bgdehfac
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bgdefhac
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dbgehfac
																		}
																		else {
																			swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dbgefhac
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bdgehfac
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bdgefhac
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); // bgedhafc
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bgdehafc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbgehafc
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdgehafc
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgedhacf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgdehacf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dbgehacf
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bdgehacf
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v1)) {
												if (compare(v4, v3)) {
													if (compare(v4, v1)) {
														if (compare(v3, v0)) {
															if (compare(v5, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v5, v6); swap(v6, v7); // gefdbahc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gedfbahc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gedbfahc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v5, v6); // gefdbach
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // gedfbach
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gedbfach
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gefbdahc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gebfdahc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gebdfahc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // gefbdach
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gebfdach
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // gebdfach
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gedbahfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gedbafhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // gedbafch
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gedbahcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gedbachf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gedbacfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gebdahfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v6, v7); // gebdafhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v6); // gebdafch
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); // gebdahcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gebdachf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // gebdacfh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v5, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gefbahdc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gefbadhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gefbadch
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gebfahdc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gebfadhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // gebfadch
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); // gebahfdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gebafhdc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gebafdhc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gebafdch
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gebahdfc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gebadhfc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gebadfhc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // gebadfch
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gebahdcf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gebadhcf
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gebadchf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gebadcfh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v4, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); // gbeahfdc
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gbeahdfc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbedahfc
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbeadhfc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); // gbedahcf
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbedachf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbeahdcf
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gbeadhcf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbeadchf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v6, v7); // gbahefdc
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gbaehfdc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gbahedfc
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbaehdfc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gbaedhfc
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); // gbahedcf
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gbaehdcf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); // gbaedhcf
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gbaedchf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v3, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbefdahc
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbedfahc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbefahdc
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gbefadhc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gbeafhdc
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbeafdhc
																			}
																		}
																		else {
																			if (compare(v3, v0)) {
																				swap(v0, v6); swap(v2, v4); swap(v4, v6); swap(v6, v7); // gbedafhc
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gbeadfhc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbaefhdc
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gbaefdhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v6, v7); // gbaedfhc
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gbefdach
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // gbedfach
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v2, v4); swap(v4, v6); // gbedafch
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // gbedacfh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			if (compare(v5, v0)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // gbefadch
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gbeafdch
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // gbeadfch
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gbeadcfh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // gbaefdch
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v4); // gbaedfch
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // gbaedcfh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v0)) {
															if (compare(v5, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdefbahc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dgefbahc
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdebfahc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgebfahc
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdbefahc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dgbefahc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbdefahc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gdefbach
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // dgefbach
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // gdebfach
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dgebfach
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gdbefach
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // dgbefach
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gbdefach
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdebahfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v6, v7); // gdebafhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v6); // gdebafch
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgebahfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dgebafhc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // dgebafch
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v2)) {
																				if (compare(v7, v5)) {
																					swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdbeahfc
																				}
																				else {
																					swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gdbeafhc
																				}
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // gdbeafch
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				if (compare(v7, v5)) {
																					swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dgbeahfc
																				}
																				else {
																					swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v6, v7); // dgbeafhc
																				}
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); // dgbeafch
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbdeahfc
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gbdeafhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // gbdeafch
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdbahefc
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gdbaehfc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v6, v7); // gdbaefhc
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); // gdbaefch
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgbahefc
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v5, v7); swap(v6, v7); // dgbaehfc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v6, v7); // dgbaefhc
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); // dgbaefch
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbdahefc
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gbdaehfc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v6, v7); // gbdaefhc
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); // gbdaefch
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbahdefc
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbadhefc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v6); swap(v5, v7); swap(v6, v7); // gbadehfc
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v6, v7); // gbadefhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); // gbadefch
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v6, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); // gdebahcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdebachf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // gdebacfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gdbeahcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdbeachf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gdbeacfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gdbahecf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); // gdbaehcf
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gdbaechf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // gdbaecfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdbahcef
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gdbachef
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdbacehf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gdbacefh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dgebahcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgebachf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dgebacfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); // dgbeahcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dgbeachf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // dgbeacfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); // dgbahecf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v5, v7); // dgbaehcf
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v5, v6); swap(v6, v7); // dgbaechf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v5, v6); // dgbaecfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgbahcef
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); // dgbachef
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgbacehf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // dgbacefh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gbdeahcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbdeachf
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gbdeacfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gbdahecf
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); // gbdaehcf
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gbdaechf
																			}
																			else {
																				swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // gbdaecfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbdahcef
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gbdachef
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbdacehf
																		}
																		else {
																			swap(v0, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gbdacefh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gbahdecf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); // gbadhecf
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v5, v7); // gbadehcf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v2, v6); swap(v5, v6); swap(v6, v7); // gbadechf
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v5, v6); // gbadecfh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbahdcef
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbadhcef
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); // gbadchef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbadcehf
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // gbadcefh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v0)) {
													if (compare(v7, v2)) {
														if (compare(v5, v2)) {
															if (compare(v7, v5)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dbgeahfc
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbgahefc
																			}
																			else {
																				swap(v0, v3); swap(v2, v6); swap(v5, v7); swap(v6, v7); // dbgaehfc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dbaghefc
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dbagehfc
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgedahfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgdeahfc
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgdahefc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bgdaehfc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v4, v0)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bdgeahfc
																			}
																			else {
																				if (compare(v7, v4)) {
																					swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdgahefc
																				}
																				else {
																					swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v5, v7); swap(v6, v7); // bdgaehfc
																				}
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bdaghefc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bdagehfc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgefdahc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgedfahc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgdefahc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v6, v7); // bgedafhc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bgdeafhc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v6, v7); // bgdaefhc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dbgefahc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bdgefahc
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v6, v7); // dbgeafhc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v6, v7); // bdgeafhc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v6); swap(v6, v7); // dbgaefhc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v6, v7); // bdgaefhc
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v6, v7); // dbagefhc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v6, v7); // bdagefhc
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v4)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); // dbgahecf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bgdahecf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); // bdgahecf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dbaghecf
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bdaghecf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbgahcef
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgdahcef
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdgahcef
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbaghcef
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdaghcef
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); // bgedahcf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bgdeahcf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); // dbgeahcf
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); // bdgeahcf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v5, v7); // dbgaehcf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); // bgdaehcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v5, v7); // bdgaehcf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); // dbagehcf
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); // bdagehcf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v6, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgedachf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgdeachf
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dbgeachf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bdgeachf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v5, v6); swap(v6, v7); // dbgaechf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bgdaechf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v5, v6); swap(v6, v7); // bdgaechf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); // dbgachef
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bgdachef
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); // bdgachef
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbgacehf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgdacehf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdgacehf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dbagechf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dbagchef
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbagcehf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); // dbacghef
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbacgehf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bdagechf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bdagchef
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdagcehf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); // bdacghef
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdacgehf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bgefdach
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // bgedfach
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bgdefach
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v6); // bgedafch
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // bgdeafch
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); // bgdaefch
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // dbgefach
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // bdgefach
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v4); // dbgeafch
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); // bdgeafch
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v6); // dbgaefch
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); // bdgaefch
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v6); // dbagefch
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); // bdagefch
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // bgedacfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bgdeacfh
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // dbgeacfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // bdgeacfh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v6); swap(v5, v6); // dbgaecfh
																			}
																			else {
																				if (compare(v6, v3)) {
																					swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // bgdaecfh
																				}
																				else {
																					swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v5, v6); // bdgaecfh
																				}
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // dbagecfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // bdagecfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // dbgacefh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bgdacefh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // bdgacefh
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dbagcefh
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // dbacgefh
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bdagcefh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // bdacgefh
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v0)) {
														if (compare(v5, v2)) {
															if (compare(v5, v3)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgefahdc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); // bgeahfdc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bgeafhdc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v6, v7); // bgahefdc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bgaehfdc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgaefhdc
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bgefadhc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgeafdhc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bgaefdhc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // bgefadch
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bgeafdch
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // bgaefdch
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bgahedfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgahdefc
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bgeahdfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgaehdfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v4, v0)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgeadhfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bgaedhfc
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgadhefc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v5, v7); swap(v6, v7); // bgadehfc
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bgeadfhc
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v6, v7); // bgaedfhc
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v6, v7); // bgadefhc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // bgeadfch
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); // bgaedfch
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); // bgadefch
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgeahdcf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bgeadhcf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); // bgahedcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bgaehdcf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); // bgaedhcf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bgahdecf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); // bgadhecf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v5, v7); // bgadehcf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgahdcef
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgadhcef
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgeadchf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bgeadcfh
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bgaedchf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // bgaedcfh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v5, v6); swap(v6, v7); // bgadechf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v5, v6); // bgadecfh
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); // bgadchef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgadcehf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // bgadcefh
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v6, v7); // baghefdc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bagehfdc
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bagefhdc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bagefdhc
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v5); // bagefdch
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // baghedfc
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bagehdfc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bagedhfc
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v6, v7); // bagedfhc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v3, v4); // bagedfch
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // baghdefc
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bagdhefc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v6); swap(v5, v7); swap(v6, v7); // bagdehfc
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v6, v7); // bagdefhc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); // bagdefch
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // badghefc
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // badgehfc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v6, v7); // badgefhc
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v6); // badgefch
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v5, v7); // baghedcf
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // baghdecf
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v4, v7); swap(v5, v7); // bagdhecf
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // badghecf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bagehdcf
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v5, v7); // bagedhcf
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v5, v7); // bagdehcf
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v5, v7); // badgehcf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bagedchf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v5, v6); swap(v6, v7); // bagdechf
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // badgechf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v5, v6); // bagedcfh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v5, v6); // bagdecfh
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v5, v6); // badgecfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v4)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // baghdcef
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bagdhcef
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // badghcef
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v4, v6); swap(v5, v7); // bagdchef
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // badgchef
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v4, v6); swap(v5, v7); // badcghef
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bagdcehf
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // badgcehf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // badcgehf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v4, v6); swap(v5, v6); // bagdcefh
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // badgcefh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v4, v6); swap(v5, v6); // badcgefh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v5, v0)) {
											if (compare(v5, v3)) {
												if (compare(v5, v1)) {
													if (compare(v7, v0)) {
														if (compare(v7, v1)) {
															if (compare(v3, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // eghfdbac
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egfhdbac
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // efghdbac
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egfdhbac
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v6, v7); // efgdhbac
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // efdghbac
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); // eghfbdac
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // egfhbdac
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efghbdac
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // eghfbadc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egfhbadc
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // efghbadc
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v3, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egfbhdac
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // efgbhdac
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efbghdac
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // egfbhadc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efgbhadc
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // efbghadc
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); // egfdbhac
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efgdbhac
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efdgbhac
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // efdbghac
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // egfbdhac
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efgbdhac
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // efbgdhac
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); // efbdghac
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v3, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egfdbahc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v6, v7); // efgdbahc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // efdgbahc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efdbgahc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // efdbaghc
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egfbdahc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // efgbdahc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efbgdahc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efbdgahc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v5, v6); swap(v6, v7); // efbdaghc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); // egfbahdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // efgbahdc
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // egfbadhc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v6, v7); // efgbadhc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); // efbgahdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // efbgadhc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // efbaghdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efbagdhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // efbadghc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v3, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // egfdbach
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v5); // efgdbach
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // egfbdach
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // efgbdach
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // efbgdach
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // egfbadch
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); // efgbadch
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // efbgadch
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // efbagdch
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // efdgbach
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // efdbgach
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // efdbagch
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v5); // efdbacgh
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // efbdgach
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v5, v6); // efbdagch
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); // efbdacgh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // efbadgch
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); // efbadcgh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v0)) {
														if (compare(v7, v0)) {
															if (compare(v7, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // eghbfdac
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egbhfdac
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebghfdac
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // beghfdac
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // egbfhdac
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ebgfhdac
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebfghdac
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // begfhdac
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // befghdac
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egbfdhac
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebgfdhac
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ebfgdhac
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); // ebfdghac
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // begfdhac
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // befgdhac
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); // befdghac
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // egbfdahc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ebgfdahc
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebfgdahc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // begfdahc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // befgdahc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // egbfdach
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // ebgfdach
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ebfgdach
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // begfdach
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // befgdach
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebfdgahc
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // ebfdgach
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v2, v5); swap(v5, v6); swap(v6, v7); // ebfdaghc
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v5, v6); // ebfdagch
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); // ebfdacgh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // befdgahc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // befdgach
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v5, v6); swap(v6, v7); // befdaghc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v5, v6); // befdagch
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); // befdacgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v7, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); // eghbfadc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // egbhfadc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ebghfadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // beghfadc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egbfhadc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebgfhadc
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ebfghadc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // begfhadc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // befghadc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // egbfahdc
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ebgfahdc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); // ebfgahdc
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ebfaghdc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // begfahdc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); // befgahdc
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // befaghdc
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v6, v7); // egbfadhc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v6, v7); // ebgfadhc
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ebfgadhc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v6, v7); // begfadhc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // befgadhc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); // egbfadch
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); // ebgfadch
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // ebfgadch
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); // begfadch
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // befgadch
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebfagdhc
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ebfadghc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ebfagdch
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // ebfadgch
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); // ebfadcgh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // befagdhc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // befadghc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // befagdch
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // befadgch
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); // befadcgh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v5)) {
													if (compare(v6, v1)) {
														if (compare(v6, v3)) {
															if (compare(v7, v5)) {
																if (compare(v7, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); // eghdfbac
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v6); // eghdbfac
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); // eghbdfac
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // egbhdfac
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // egdhfbac
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); // egdhbfac
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // egdbhfac
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v6, v7); // egbdhfac
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egdfhbac
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // egdfbhac
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egdbfhac
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egbdfhac
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v1)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egdfbahc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // egdbfahc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v6, v7); // egbdfahc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v1)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // egdfbach
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // egdbfach
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v5); // egbdfach
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edghfbac
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // edgfhbac
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edgfbhac
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // edgfbahc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // edgfbach
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // deghfbac
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // degfhbac
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // degfbhac
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v6, v7); // degfbahc
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); // degfbach
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // edghbfac
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v6, v7); // edgbhfac
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edgbfhac
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v6, v7); // edgbfahc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v5); // edgbfach
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v6, v7); // deghbfac
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // degbhfac
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // degbfhac
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // degbfahc
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // degbfach
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v7, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // edbghfac
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edbgfhac
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); // debghfac
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // debgfhac
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ebghdfac
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v4, v7); swap(v6, v7); // ebgdhfac
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebgdfhac
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ebdghfac
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebdgfhac
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // edbgfahc
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // debgfahc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // edbgfach
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // debgfach
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v4, v5); swap(v6, v7); // ebgdfahc
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ebdgfahc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v4, v5); // ebgdfach
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // ebdgfach
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // beghdfac
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v6, v7); // begdhfac
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // begdfhac
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bedghfac
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bedgfhac
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); // dbeghfac
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dbegfhac
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); // bdeghfac
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bdegfhac
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v6, v7); // begdfahc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bedgfahc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbegfahc
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdegfahc
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v5); // begdfach
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // bedgfach
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dbegfach
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bdegfach
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v3)) {
														if (compare(v7, v2)) {
															if (compare(v7, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edfghbac
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // edfgbhac
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); // edfbghac
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // edbfghac
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ebdfghac
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bedfghac
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edfgbahc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edfbgahc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edbfgahc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebdfgahc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bedfgahc
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v5, v6); swap(v6, v7); // edfbaghc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // edbfaghc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ebdfaghc
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bedfaghc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // edfgbach
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // edfbgach
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // edbfgach
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v5, v6); // edfbagch
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); // edfbacgh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // edbfagch
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v5); // edbfacgh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v0)) {
																		swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ebdfgach
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // ebdfagch
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v5); // ebdfacgh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bedfgach
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // bedfagch
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v5); // bedfacgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v7, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // defghbac
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); // defgbhac
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // defbghac
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // debfghac
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbefghac
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdefghac
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // defgbahc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // defbgahc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // debfgahc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dbefgahc
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bdefgahc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // defbaghc
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // debfaghc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbefaghc
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdefaghc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // defgbach
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // defbgach
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // debfgach
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // defbagch
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v4); // defbacgh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // debfagch
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // debfacgh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v0)) {
																		swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // dbefgach
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dbefagch
																		}
																		else {
																			swap(v0, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // dbefacgh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // bdefgach
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bdefagch
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // bdefacgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v1)) {
												if (compare(v6, v3)) {
													if (compare(v6, v1)) {
														if (compare(v3, v0)) {
															if (compare(v7, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); // eghdbafc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // egdhbafc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egdbhafc
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eghdbacf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egdhbacf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // egdbhacf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // eghbdafc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egbhdafc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egbdhafc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eghbdacf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // egbhdacf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); // egbdhacf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // egdbahfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v6, v7); // egdbafhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); // egdbafch
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); // egdbahcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // egdbachf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // egdbacfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v5, v7); swap(v6, v7); // egbdahfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v6, v7); // egbdafhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); // egbdafch
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v5, v7); // egbdahcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v5, v6); swap(v6, v7); // egbdachf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v5, v6); // egbdacfh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v7, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); // eghbafdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); // eghbadfc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // eghbadcf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v6, v7); // egbhafdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // egbhadfc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); // egbhadcf
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // egbahfdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egbafhdc
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egbahdfc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // egbahdcf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // egbadhfc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); // egbadhcf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // egbafdhc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v6, v7); // egbadfhc
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // egbafdch
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); // egbadfch
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // egbadchf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // egbadcfh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v5, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v3, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebghdafc
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebgdhafc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v6, v7); // ebghafdc
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ebghadfc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ebgahfdc
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebgahdfc
																			}
																		}
																		else {
																			if (compare(v3, v0)) {
																				swap(v0, v4); swap(v2, v6); swap(v5, v7); swap(v6, v7); // ebgdahfc
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ebgadhfc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); // ebaghfdc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ebaghdfc
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebagdhfc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ebghdacf
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); // ebgdhacf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v2, v6); swap(v5, v7); // ebgdahcf
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v5, v6); swap(v6, v7); // ebgdachf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v0)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); // ebghadcf
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ebgahdcf
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); // ebgadhcf
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ebgadchf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebaghdcf
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ebagdhcf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebagdchf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebgafhdc
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ebgafdhc
																		}
																	}
																	else {
																		if (compare(v3, v0)) {
																			swap(v0, v4); swap(v2, v6); swap(v6, v7); // ebgdafhc
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v6, v7); // ebgadfhc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v6); // ebgdafch
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v5, v6); // ebgdacfh
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // ebgafdch
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v4); // ebgadfch
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // ebgadcfh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ebagfhdc
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebafghdc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebagfdhc
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ebafgdhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ebagdfhc
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ebagfdch
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // ebafgdch
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // ebagdfch
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ebagdcfh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v6, v0)) {
															if (compare(v6, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edghbafc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edgbhafc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v5, v7); swap(v6, v7); // edgbahfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v6, v7); // edgbafhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); // edgbafch
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // deghbafc
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // degbhafc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // degbahfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v6, v7); // degbafhc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); // degbafch
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edbghafc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // debghafc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebdghafc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // edbgahfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // debgahfc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ebdgahfc
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v6, v7); // edbgafhc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // debgafhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v6, v7); // ebdgafhc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); // edbgafch
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // debgafch
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); // ebdgafch
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edbaghfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // edbagfhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edbafghc
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // debaghfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v6, v7); // debagfhc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // debafghc
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebdaghfc
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ebdagfhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebdafghc
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebadghfc
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v4, v6); swap(v6, v7); // ebadgfhc
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebafdghc
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebadfghc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // edbagfch
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // edbafgch
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // edbafcgh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v6); // debagfch
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // debafgch
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v5); // debafcgh
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // ebdagfch
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ebdafgch
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // ebdafcgh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ebafdgch
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v4); swap(v2, v4); swap(v4, v6); // ebadgfch
																				}
																				else {
																					swap(v0, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // ebadfgch
																				}
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // ebafdcgh
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v4, v5); // ebadfcgh
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v6, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // edghbacf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); // edgbhacf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v5, v7); // edgbahcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v5, v6); swap(v6, v7); // edgbachf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v5, v6); // edgbacfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); // deghbacf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // degbhacf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); // degbahcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // degbachf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // degbacfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // edbghacf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // debghacf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ebdghacf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); // edbgahcf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // debgahcf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); // ebdgahcf
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // edbgachf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // debgachf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ebdgachf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // edbgacfh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // debgacfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // ebdgacfh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // edbaghcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edbagchf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // edbagcfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); // debaghcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // debagchf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // debagcfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // edbacghf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // edbacgfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v3); swap(v3, v4); // edbacfgh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v5, v6); swap(v6, v7); // debacghf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v4); swap(v5, v6); // debacgfh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v4); // debacfgh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ebdaghcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebdagchf
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ebdagcfh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ebdacghf
																			}
																			else {
																				swap(v0, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // ebdacgfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v3); swap(v3, v4); // ebdacfgh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); // ebadghcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebadgchf
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // ebadgcfh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v5, v6); swap(v6, v7); // ebadcghf
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v5, v6); // ebadcgfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); // ebadcfgh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v0)) {
													if (compare(v5, v2)) {
														if (compare(v7, v2)) {
															if (compare(v7, v5)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // beghdafc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // begdhafc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bedghafc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v5, v7); swap(v6, v7); // begdahfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bedgahfc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bedaghfc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dbeghafc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bdeghafc
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbegahfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdegahfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbeaghfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdeaghfc
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dbaeghfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bdaeghfc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dbegafhc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v4); swap(v4, v6); swap(v6, v7); // dbeagfhc
																			}
																			else {
																				swap(v0, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbeafghc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dbaegfhc
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dbaefghc
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v6, v7); // begdafhc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v6, v7); // bedgafhc
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bedagfhc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bedafghc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v0)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bdegafhc
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v6, v7); // bdeagfhc
																				}
																				else {
																					swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdeafghc
																				}
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bdaegfhc
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bdaefghc
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v6, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); // begdafch
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); // bedgafch
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // dbegafch
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // bdegafch
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v4, v6); // dbeagfch
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // bedagfch
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v6); // bdeagfch
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // dbaegfch
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // bdaegfch
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // dbeafgch
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bedafgch
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // bdeafgch
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dbaefgch
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bdaefgch
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v4, v5); // dbeafcgh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bedafcgh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v5); // bdeafcgh
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dbaefcgh
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bdaefcgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v7, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // beghdacf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); // begdhacf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bedghacf
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v5, v7); // begdahcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); // bedgahcf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bedaghcf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dbeghacf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bdeghacf
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dbegahcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bdegahcf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); // dbeaghcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); // bdeaghcf
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dbaeghcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bdaeghcf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v5, v6); swap(v6, v7); // begdachf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bedgachf
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbegachf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdegachf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbeagchf
																			}
																			else {
																				if (compare(v4, v3)) {
																					swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bedagchf
																				}
																				else {
																					swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdeagchf
																				}
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dbaegchf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bdaegchf
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v5, v6); swap(v6, v7); // dbeacghf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bedacghf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v5, v6); swap(v6, v7); // bdeacghf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dbaecghf
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v5, v6); swap(v6, v7); // dbaceghf
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bdaecghf
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v5, v6); swap(v6, v7); // bdaceghf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v5, v6); // begdacfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // bedgacfh
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dbegacfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bdegacfh
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // dbeagcfh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bedagcfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // bdeagcfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); swap(v5, v6); // dbeacgfh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // bedacgfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); swap(v5, v6); // bdeacgfh
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v3); swap(v2, v4); // dbeacfgh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v3); swap(v3, v4); // bedacfgh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v4); // bdeacfgh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dbaegcfh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // dbaecgfh
																			}
																			else {
																				swap(v0, v3); swap(v2, v3); swap(v3, v4); // dbaecfgh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v2, v3); swap(v5, v6); // dbacegfh
																		}
																		else {
																			swap(v0, v3); swap(v2, v3); // dbacefgh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bdaegcfh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // bdaecgfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v3, v4); // bdaecfgh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); swap(v5, v6); // bdacegfh
																		}
																		else {
																			swap(v0, v1); swap(v1, v3); swap(v2, v3); // bdacefgh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v7, v2)) {
															if (compare(v7, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v6, v7); // beghafdc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // begahfdc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // begafhdc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); // beaghfdc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // beagfhdc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // beafghdc
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // beghadfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // begahdfc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // beaghdfc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); // beghadcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // begahdcf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // beaghdcf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // begadhfc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // beagdhfc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // beadghfc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); // begadhcf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // beagdhcf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); // beadghcf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v6, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // begafdhc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // beagfdhc
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // beafgdhc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // beafdghc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v6, v7); // begadfhc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // beagdfhc
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v6, v7); // beadgfhc
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // beadfghc
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // begafdch
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // beagfdch
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // beafgdch
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // beafdgch
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // beafdcgh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); // begadfch
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // beagdfch
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v6); // beadgfch
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // beadfgch
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v5); // beadfcgh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // begadchf
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // begadcfh
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // beagdchf
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // beagdcfh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // beadgchf
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // beadgcfh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v5, v6); swap(v6, v7); // beadcghf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v5, v6); // beadcgfh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); // beadcfgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v5)) {
																if (compare(v6, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v7); // baeghfdc
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // baegfhdc
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // baegfdhc
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // baegfdch
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // baeghdfc
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // baegdhfc
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // baegdfhc
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v6); // baegdfch
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // baedghfc
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v4, v6); swap(v6, v7); // baedgfhc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v4, v6); // baedgfch
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // badeghfc
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // badegfhc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v6); // badegfch
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // baefghdc
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // baefgdhc
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // baefdghc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v6); // baefgdch
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // baefdgch
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v5); // baefdcgh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // baedfghc
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v4, v5); swap(v5, v6); // baedfgch
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v4, v5); // baedfcgh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // badefghc
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // badefgch
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v5); // badefcgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // baeghdcf
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // baegdhcf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v2, v4); swap(v4, v6); swap(v5, v7); // baedghcf
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // badeghcf
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // baegdchf
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // baedgchf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // badegchf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // baegdcfh
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v4, v6); swap(v5, v6); // baedgcfh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // badegcfh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v4); swap(v5, v6); swap(v6, v7); // baedcghf
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v5, v6); // baedcgfh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); // baedcfgh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // badecghf
																			}
																			else {
																				swap(v0, v1); swap(v2, v3); swap(v3, v4); swap(v5, v6); // badecgfh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v3, v4); // badecfgh
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v2, v3); swap(v5, v6); swap(v6, v7); // badceghf
																		}
																		else {
																			swap(v0, v1); swap(v2, v3); swap(v5, v6); // badcegfh
																		}
																	}
																	else {
																		swap(v0, v1); swap(v2, v3); // badcefgh
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
					else {
						if (compare(v3, v0)) {
							if (compare(v5, v4)) {
								if (compare(v7, v6)) {
									if (compare(v6, v4)) {
										if (compare(v4, v0)) {
											if (compare(v6, v1)) {
												if (compare(v3, v1)) {
													if (compare(v7, v2)) {
														if (compare(v5, v2)) {
															if (compare(v6, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // hgfecdba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hfgecdba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhgecdba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); // hgfcedba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // hfgcedba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fhgcedba
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // hgfcdeba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hfgcdeba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhgcdeba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hgfcdbea
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // hfgcdbea
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fhgcdbea
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v5, v6); // hfcgedba
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hfcgdeba
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v4, v6); swap(v5, v6); // hfcdgeba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v6); // hfcgdbea
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v4, v6); // hfcdgbea
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fhcgedba
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhcgdeba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhcdgeba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fhcgdbea
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v4, v6); swap(v5, v7); // fhcdgbea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); // hgcfedba
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); // hcgfedba
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // hcfgedba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v5); // hgcfdeba
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // hcgfdeba
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hcfgdeba
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // hcfdgeba
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v4, v5); // hgcdfeba
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v5); // hcgdfeba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // hcdgfeba
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hcdfgeba
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hgcfdbea
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hcgfdbea
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // hcfgdbea
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v6); // hcfdgbea
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v4, v5); swap(v5, v6); // hgcdfbea
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // hcgdfbea
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hcdgfbea
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // hcdfgbea
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v4, v6); // hgcdbfea
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v6); // hcgdbfea
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // hcdgbfea
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v3)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v6, v7); // chgfedba
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // chfgedba
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fchgedba
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); // cfhgedba
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // chgfdeba
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v6, v7); // chgdfeba
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fchgdeba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chfgdeba
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cfhgdeba
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fchdgeba
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcdhgeba
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chfdgeba
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); // cfhdgeba
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cfdhgeba
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // chdgfeba
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chdfgeba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); // cdhgfeba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cdhfgeba
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdfhgeba
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chgfdbea
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // chfgdbea
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fchgdbea
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); // cfhgdbea
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); // fchdgbea
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fcdhgbea
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); // chfdgbea
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v6); // cfhdgbea
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); // cfdhgbea
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chgdfbea
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chdgfbea
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // chdfgbea
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v6, v7); // chgdbfea
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // chdgbfea
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cdhgfbea
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); // cdhgbfea
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); // cdhfgbea
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cdfhgbea
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hgfecbda
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // hgfcebda
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hgfcbeda
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // hgfcbdea
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // hgfebcda
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); // hgfbecda
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // hgfbceda
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hgfbcdea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v2)) {
																		swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v5, v6); // hgcfebda
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // hcgfebda
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // chgfebda
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hgcfbeda
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hcgfbeda
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chgfbeda
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v6); // hgcfbdea
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // hcgfbdea
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // chgfbdea
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v5); // hgcbfeda
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hgcbfdea
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v6); // hgcbdfea
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // hcgbfeda
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hcgbfdea
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // hcgbdfea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // chgbfeda
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chgbfdea
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // chgbdfea
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // hgbfecda
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hgbfceda
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // hgbfcdea
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // hgbcfeda
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hgbcfdea
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // hgbcdfea
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // hfgecbda
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); // hfgcebda
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // hfgcbeda
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hfgcbdea
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hfgebcda
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // hfgbecda
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hfgbceda
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // hfgbcdea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v2)) {
																		swap(v0, v7); swap(v1, v5); swap(v3, v6); // hfcgebda
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); // hcfgebda
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); // chfgebda
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v5); // hfcgbeda
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // hcfgbeda
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // chfgbeda
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hfcgbdea
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hcfgbdea
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chfgbdea
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fhgecbda
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); // fhgcebda
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fhgcbeda
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhgcbdea
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhgebcda
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fhgbecda
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhgbceda
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fhgbcdea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v5, v7); // fhcgebda
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); // fchgebda
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); // cfhgebda
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fhcgbeda
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fchgbeda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); // cfhgbeda
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhcgbdea
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fchgbdea
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cfhgbdea
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v1)) {
													if (compare(v5, v1)) {
														if (compare(v7, v5)) {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hfcbgeda
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v6); // hfcbgdea
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v4, v5); swap(v5, v6); // hfcdbgea
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hfcbdgea
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); // hfbgecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // hfbgceda
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hfbgcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hfbcgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // hfbcgdea
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hfbcdgea
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // hcfdbgea
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hcdfbgea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chfdbgea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chdfbgea
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cdhfbgea
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hcfbgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // hcfbgdea
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hcfbdgea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chfbgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // chfbgdea
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chfbdgea
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhcbgeda
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fhcbgdea
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhcdbgea
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhcbdgea
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); // fhbgecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fhbgceda
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhbgcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhbcgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fhbcgdea
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhbcdgea
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fchdbgea
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcdhbgea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); // cfhdbgea
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cfdhbgea
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdfhbgea
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fchbgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fchbgdea
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fchbdgea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cfhbgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); // cfhbgdea
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cfhbdgea
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v2, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v5); // hcbgfeda
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hcbgfdea
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v6); // hcdbgfea
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v6); // hcbgdfea
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v4, v6); // hcbdgfea
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); // chbgfeda
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // chbgfdea
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); // chbgdfea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); // chdbgfea
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v6); // cdhbgfea
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v6); // chbdgfea
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); // hbgfecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hbgfceda
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); // hbgfcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); // hbgcfeda
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hbgcfdea
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); // hbgcdfea
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v3, v6); swap(v4, v5); // hbcgfeda
																		}
																		else {
																			swap(v0, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hbcgfdea
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v3, v6); swap(v4, v6); // hbcgdfea
																		}
																		else {
																			swap(v0, v7); swap(v4, v6); // hbcdgfea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hcbfgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v6); // hcbfgdea
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hcbfdgea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // hcdbfgea
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v4, v5); swap(v5, v6); // hcbdfgea
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // chbfgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); // chbfgdea
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // chbfdgea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // chdbfgea
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); // cdhbfgea
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); // chbdfgea
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); // hbfgecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); // hbfgceda
																			}
																			else {
																				swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hbfgcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hbfcgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); // hbfcgdea
																			}
																			else {
																				swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hbfcdgea
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hbcfgeda
																		}
																		else {
																			swap(v0, v7); swap(v3, v5); swap(v4, v6); // hbcfgdea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hbcfdgea
																		}
																		else {
																			swap(v0, v7); swap(v4, v5); swap(v5, v6); // hbcdfgea
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v6, v5)) {
															if (compare(v7, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); // bhgfecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bhgfceda
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); // bhgfcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); // bhgcfeda
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bhgcfdea
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); // bhgcdfea
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v5); // bhcgfeda
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bhcgfdea
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v6); // bhcgdfea
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v4, v6); // bhcdgfea
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); // cbhgfeda
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cbhgfdea
																			}
																			else {
																				swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); // cbhgdfea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); // cdbhgfea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v4, v6); // cbhdgfea
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); // cbdhgfea
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); // bchgfeda
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bchgfdea
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); // bchgdfea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v6); // bchdgfea
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); // bcdhgfea
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); // bhfgecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); // bhfgceda
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bhfgcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bhfcgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); // bhfcgdea
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bhfcdgea
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bhcfgeda
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v6); // bhcfgdea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bhcfdgea
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v4, v5); swap(v5, v6); // bhcdfgea
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cbhfgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); // cbhfgdea
																			}
																			else {
																				swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cbhfdgea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cdbhfgea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); // cbhdfgea
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cbdhfgea
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bchfgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); // bchfgdea
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bchfdgea
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); // bchdfgea
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bcdhfgea
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcbhgeda
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fcbhgdea
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcdbhgea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcbhdgea
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcbdhgea
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfbhgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cfbhgdea
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfbhdgea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfdbhgea
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdfbhgea
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfbdhgea
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); // fbhgecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fbhgceda
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbhgcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbhcgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fbhcgdea
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbhcdgea
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbchgeda
																		}
																		else {
																			swap(v0, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fbchgdea
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbchdgea
																		}
																		else {
																			swap(v0, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbcdhgea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); // bfhgecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bfhgceda
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfhgcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfhcgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bfhcgdea
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfhcdgea
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfchgeda
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bfchgdea
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfchdgea
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfcdhgea
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbfhgeda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cbfhgdea
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbfhdgea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdbfhgea
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbfdhgea
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbdfhgea
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcfhgeda
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bcfhgdea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcfhdgea
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcfdhgea
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcdfhgea
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v3)) {
												if (compare(v5, v2)) {
													if (compare(v2, v1)) {
														if (compare(v3, v1)) {
															if (compare(v6, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgfcdbae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hfgcdbae
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhgcdbae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hfcgdbae
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhcgdbae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fchgdbae
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v4, v6); swap(v6, v7); // hfcdgbae
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhcdgbae
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fchdgbae
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcdhgbae
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfcdbgae
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v4, v7); swap(v5, v6); // fhcdbgae
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); // fchdbgae
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fcdhbgae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); // fcdbhgae
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v4, v5); swap(v5, v7); // hfcdbage
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v4, v7); // fhcdbage
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v7); // fchdbage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); // fcdhbage
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v7); // fcdbhage
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); // fcdbahge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hgfcbdae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfgcbdae
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fhgcbdae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfcgbdae
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fhcgbdae
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hfcbgdae
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfcbdgae
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hfcbdage
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhcbgdae
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fhcbdgae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v7); // fhcbdage
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fchgbdae
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fchbgdae
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fchbdgae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcbhgdae
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fcbhdgae
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v4, v7); swap(v5, v6); // fcbdhgae
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); // fchbdage
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v7); // fcbhdage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v1, v2); swap(v4, v7); // fcbdhage
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v4, v5); swap(v5, v7); // fcbdahge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v7, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgfbcdae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hfgbcdae
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhgbcdae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfbgcdae
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fhbgcdae
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hfbcgdae
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfbcdgae
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hfbcdage
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhbcgdae
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fhbcdgae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); // fhbcdage
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fbhgcdae
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbhcgdae
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fbhcdgae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbchgdae
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fbchdgae
																			}
																			else {
																				swap(v0, v5); swap(v4, v7); swap(v5, v6); // fbcdhgae
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); // fbhcdage
																		}
																		else {
																			swap(v0, v5); swap(v3, v7); swap(v4, v7); // fbchdage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v4, v7); // fbcdhage
																		}
																		else {
																			swap(v0, v5); swap(v4, v5); swap(v5, v7); // fbcdahge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hgbfcdae
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hbgfcdae
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbfgcdae
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hbfcgdae
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bhgfcdae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhfgcdae
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bfhgcdae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bhfcgdae
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfhcgdae
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfchgdae
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbfcdgae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhfcdgae
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bfhcdgae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bfchdgae
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v4, v7); swap(v5, v6); // bfcdhgae
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hbfcdage
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bhfcdage
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); // bfhcdage
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v7); // bfchdage
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v4, v7); // bfcdhage
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v4, v5); swap(v5, v7); // bfcdahge
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v3, v1)) {
															if (compare(v6, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgcfdbae
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcgfdbae
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hcfgdbae
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v6, v7); // hcfdgbae
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // chgfdbae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chfgdbae
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cfhgdbae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chfdgbae
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v6, v7); // cfhdgbae
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cfdhgbae
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcfdbgae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); // chfdbgae
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfhdbgae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfdhbgae
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cfdbhgae
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); // hcfdbage
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v7); // chfdbage
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); // cfhdbage
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cfdhbage
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); // cfdbhage
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cfdbahge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hgcfbdae
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hcgfbdae
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcfgbdae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); // chgfbdae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // chfgbdae
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfhgbdae
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hcfbgdae
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chfbgdae
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cfhbgdae
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfbhgdae
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcfbdgae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // chfbdgae
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfhbdgae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cfbhdgae
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); // cfbdhgae
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hcfbdage
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); // chfbdage
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cfhbdage
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); // cfbhdage
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v7); // cfbdhage
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); // cfbdahge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v7, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgcbfdae
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcgbfdae
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcbgfdae
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hcbfgdae
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcbfdgae
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hcbfdage
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // chgbfdae
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chbgfdae
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // chbfgdae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chbfdgae
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // chbfdage
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbhgfdae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cbhfgdae
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbfhgdae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbhfdgae
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cbfhdgae
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); // cbfdhgae
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cbhfdage
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); // cbfhdage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v2, v5); swap(v4, v7); // cbfdhage
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); // cbfdahge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgbcfdae
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbgcfdae
																			}
																			else {
																				swap(v0, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbcgfdae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hbcfgdae
																			}
																			else {
																				swap(v0, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbcfdgae
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hbcfdage
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhgcfdae
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhcgfdae
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bhcfgdae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhcfdgae
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bhcfdage
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bchgfdae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bchfgdae
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcfhgdae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bchfdgae
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bcfhdgae
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); // bcfdhgae
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bchfdage
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); // bcfhdage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v7); // bcfdhage
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); // bcfdahge
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v5)) {
													if (compare(v7, v2)) {
														if (compare(v5, v0)) {
															if (compare(v6, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgcdfbae
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v4, v6); swap(v6, v7); // hgcdbfae
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hgcbdfae
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hgbcdfae
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcgdfbae
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcdgfbae
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v6, v7); // hcgdbfae
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hcdgbfae
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hcgbdfae
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hcbgdfae
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hbgcdfae
																			}
																			else {
																				swap(v0, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hbcgdfae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bhgcdfae
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bhcgdfae
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v6, v7); // hcdbgfae
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v4, v6); swap(v6, v7); // hcbdgfae
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v4, v6); swap(v6, v7); // hbcdgfae
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v4, v6); swap(v6, v7); // bhcdgfae
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgcdbafe
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcgdbafe
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcdgbafe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcdbgafe
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); // hcdbagfe
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgcbdafe
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcgbdafe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcbgdafe
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcbdgafe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v4, v7); swap(v5, v6); // hcbdagfe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgbcdafe
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbgcdafe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbcgdafe
																			}
																			else {
																				swap(v0, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbcdgafe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v4, v7); swap(v5, v6); // hbcdagfe
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhgcdafe
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhcgdafe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhcdgafe
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v4, v7); swap(v5, v6); // bhcdagfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); // chgdfbae
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // chdgfbae
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdhgfbae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v7); // chgdbfae
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); // chgbdfae
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); // chdgbfae
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cdhgbfae
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // chdbgfae
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v6, v7); // cdhbgfae
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // chbgdfae
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v6, v7); // chbdgfae
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); // chgdbafe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // chdgbafe
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdhgbafe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chdbgafe
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdhbgafe
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // chdbagfe
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); // cdhbagfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // chgbdafe
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chbgdafe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chbdgafe
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); // chbdagfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cbhgdfae
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v4, v6); swap(v6, v7); // cbhdgfae
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbhgdafe
																			}
																			else {
																				swap(v0, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbhdgafe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); // cbhdagfe
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bchgdfae
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v6, v7); // bchdgfae
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bchgdafe
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bchdgafe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); // bchdagfe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cdbhgfae
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cbdhgfae
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bcdhgfae
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdbhgafe
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbdhgafe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcdhgafe
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cdbhagfe
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cbdhagfe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bcdhagfe
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); // cdbahgfe
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); // cbdahgfe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); // bcdahgfe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v5, v1)) {
															if (compare(v6, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hcdfgbae
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chdfgbae
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cdhfgbae
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdfhgbae
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcdfbgae
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // chdfbgae
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdhfbgae
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cdfhbgae
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); // cdfbhgae
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v7, v2)) {
																		swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hcdfbage
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); // chdfbage
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cdhfbage
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); // cdfhbage
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v7); // cdfbhage
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); // cdfbahge
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcdbfgae
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chdbfgae
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdhbfgae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); // hcdbfage
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // chdbfage
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); // cdhbfage
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v7); // hcdbafge
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); // chdbafge
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v7); // cdhbafge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdbhfgae
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cdbhfage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); // cdbhafge
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v7); // cdbahfge
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cdbfhgae
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); // cdbfhage
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cdbfahge
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); // cdbafhge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v7, v3)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcbdfgae
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chbdfgae
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbhdfgae
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v4, v5); swap(v5, v7); // hcbdfage
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); // chbdfage
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); // cbhdfage
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v4, v7); // hcbdafge
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v7); // chbdafge
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v7); swap(v4, v7); // cbhdafge
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbdhfgae
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cbdhfage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); // cbdhafge
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v4, v7); // cbdahfge
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cbdfhgae
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); // cbdfhage
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cbdfahge
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); // cbdafhge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbcdfgae
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhcdfgae
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bchdfgae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v4, v5); swap(v5, v7); // hbcdfage
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v4, v5); swap(v5, v7); // bhcdfage
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); // bchdfage
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		swap(v0, v7); swap(v4, v7); // hbcdafge
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v4, v7); // bhcdafge
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v7); // bchdafge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcdhfgae
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bcdhfage
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); // bcdhafge
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v7); // bcdahfge
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bcdfhgae
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); // bcdfhage
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bcdfahge
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); // bcdafhge
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v0)) {
											if (compare(v4, v1)) {
												if (compare(v3, v1)) {
													if (compare(v5, v2)) {
														if (compare(v7, v2)) {
															if (compare(v4, v2)) {
																if (compare(v6, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // hfegcdba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fhegcdba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fehgcdba
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hfecgdba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhecgdba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fehcgdba
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // hfecdgba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fhecdgba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fehcdgba
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); // hfecdbga
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); // fhecdbga
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fehcdbga
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hfcegdba
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v5, v6); // hfcedgba
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v5, v6); // hfcdegba
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v4); // hfcedbga
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); // hfcdebga
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhcegdba
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fhcedgba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v5, v6); swap(v6, v7); // fhcdegba
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v5, v7); // fhcedbga
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v5, v7); // fhcdebga
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fechgdba
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fchegdba
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fcehgdba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fchedgba
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v5, v6); swap(v6, v7); // fchdegba
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fechdgba
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcehdgba
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fecdhgba
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcedhgba
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fcdhegba
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcdehgba
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fechdbga
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v4, v7); swap(v5, v7); // fecdhbga
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v5, v7); // fecdbhga
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); // fchedbga
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fcehdbga
																			}
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); // fcedhbga
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v5, v7); // fcedbhga
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v5, v7); // fchdebga
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); // fcdhebga
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fcdehbga
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); // fcdebhga
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hcfegdba
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chfegdba
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cfhegdba
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfehgdba
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // hcfedgba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // chfedgba
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); // cfhedgba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cfehdgba
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); // cfedhgba
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v5, v6); // hcfdegba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v5, v6); swap(v6, v7); // chfdegba
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v5, v6); // cfhdegba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); // cfdhegba
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cfdehgba
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // hcdfegba
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // chdfegba
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); // cdhfegba
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cdfhegba
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdfehgba
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); // hcfedbga
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); // chfedbga
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); // cfhedbga
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); // cfehdbga
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); // hcfdebga
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); // hcdfebga
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v5, v7); // chfdebga
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); // cfhdebga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); // chdfebga
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v7); // cfedhbga
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); // cfdhebga
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); // cfdehbga
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); // cfedbhga
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cfdebhga
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); // cdhfebga
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); // cdfhebga
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cdfehbga
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); // cdfebhga
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v4)) {
														if (compare(v7, v5)) {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); // hfegcbda
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // hfecgbda
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hfecbgda
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // hfecbdga
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // hfegbcda
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hfebgcda
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // hfebcgda
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); // hfebcdga
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v5, v2)) {
																		swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v6); // hfcegbda
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // hcfegbda
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // chfegbda
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hfcebgda
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hcfebgda
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chfebgda
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v5); // hfcebdga
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // hcfebdga
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // chfebdga
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); // fhegcbda
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fhecgbda
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhecbgda
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fhecbdga
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fhegbcda
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhebgcda
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fhebcgda
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); // fhebcdga
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fhcegbda
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fchegbda
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); // cfhegbda
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhcebgda
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fchebgda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cfhebgda
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fhcebdga
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fchebdga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); // cfhebdga
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fehgcbda
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fehcgbda
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fehcbgda
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); // fehcbdga
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); // fehgbcda
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fehbgcda
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fehbcgda
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fehbcdga
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fechgbda
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcehgbda
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cfehgbda
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fechbgda
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fcehbgda
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfehbgda
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v5, v7); // fechbdga
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); // fcehbdga
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cfehbdga
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fecbhgda
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fecbhdga
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v5, v7); // fecbdhga
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcebhgda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fcebhdga
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); // fcebdhga
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cfebhgda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); // cfebhdga
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cfebdhga
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // febhgcda
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // febhcgda
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); // febhcdga
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // febchgda
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // febchdga
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); // febcdhga
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v7, v1)) {
														if (compare(v7, v5)) {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v5, v6); // hfcbegda
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v3, v5); // hfcbedga
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v4, v5); // hfcdbega
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v5); // hfcbdega
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // hfbegcda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hfbecgda
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // hfbecdga
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // hfbcegda
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); // hfbcedga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // hfbcdega
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v5); // hcfdbega
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // hcdfbega
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); // chfdbega
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // chdfbega
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); // cdhfbega
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // hcfbegda
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); // hcfbedga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // hcfbdega
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // chfbegda
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); // chfbedga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // chfbdega
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fhcbegda
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v5, v7); // fhcbedga
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v4, v7); swap(v5, v7); // fhcdbega
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fhcbdega
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fhbegcda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhbecgda
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fhbecdga
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fhbcegda
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); // fhbcedga
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fhbcdega
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); // fchdbega
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fcdhbega
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v5); // cfhdbega
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); // cfdhbega
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cdfhbega
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fchbegda
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); // fchbedga
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fchbdega
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); // cfhbegda
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); // cfhbedga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); // cfhbdega
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v4)) {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fcbhegda
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v5, v7); // fcbhedga
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); // fcdbhega
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fcbhdega
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v4, v7); swap(v5, v7); // fcbdhega
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cfbhegda
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); // cfbhedga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cfbhdega
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cfdbhega
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); // cdfbhega
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); // cfbdhega
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fbhegcda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbhecgda
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fbhecdga
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fbhcegda
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); // fbhcedga
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fbhcdega
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fbchegda
																		}
																		else {
																			swap(v0, v5); swap(v3, v7); swap(v5, v7); // fbchedga
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fbchdega
																		}
																		else {
																			swap(v0, v5); swap(v4, v7); swap(v5, v7); // fbcdhega
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcbehgda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fcbehdga
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v5, v7); // fcbedhga
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v5, v7); // fcdbehga
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v5, v7); // fcbdehga
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfbehgda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cfbehdga
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); // cfbedhga
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); // cfdbehga
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v5, v7); // cdfbehga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v5, v7); // cfbdehga
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbehgcda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fbehcgda
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); // fbehcdga
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbechgda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fbechdga
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); // fbecdhga
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbcehgda
																		}
																		else {
																			swap(v0, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fbcehdga
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v3, v4); swap(v5, v7); // fbcedhga
																		}
																		else {
																			swap(v0, v5); swap(v5, v7); // fbcdehga
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v7, v1)) {
															if (compare(v2, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v5, v6); // hcbfegda
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v5); // hcbfedga
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v5); // hcdbfega
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v5); // hcbfdega
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v4, v5); // hcbdfega
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); // chbfegda
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); // chbfedga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); // chbfdega
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); // chdbfega
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v5); // cdhbfega
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v5); // chbdfega
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); // hbfegcda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hbfecgda
																			}
																			else {
																				swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); // hbfecdga
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); // hbfcegda
																			}
																			else {
																				swap(v0, v7); swap(v2, v5); swap(v3, v5); // hbfcedga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); // hbfcdega
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v3, v5); swap(v5, v6); // hbcfegda
																		}
																		else {
																			swap(v0, v7); swap(v3, v5); // hbcfedga
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v3, v5); swap(v4, v5); // hbcfdega
																		}
																		else {
																			swap(v0, v7); swap(v4, v5); // hbcdfega
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); // bhfegcda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bhfecgda
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bhfecdga
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); // bhfcegda
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); // bhfcedga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bhfcdega
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v5, v6); // bhcfegda
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); // bhcfedga
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v5); // bhcfdega
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v4, v5); // bhcdfega
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); // cbhfegda
																			}
																			else {
																				swap(v0, v2); swap(v2, v7); swap(v3, v5); // cbhfedga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); // cbhfdega
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); // cdbhfega
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v4, v5); // cbhdfega
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); // cbdhfega
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); // bchfegda
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); // bchfedga
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); // bchfdega
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v5); // bchdfega
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); // bcdhfega
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v4)) {
															if (compare(v5, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bfhegcda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfhecgda
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bfhecdga
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bfhcegda
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); // bfhcedga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bfhcdega
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bfchegda
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v5, v7); // bfchedga
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bfchdega
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v4, v7); swap(v5, v7); // bfcdhega
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cbfhegda
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); // cbfhedga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cbfhdega
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cdbfhega
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); // cbfdhega
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cbdfhega
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bcfhegda
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); // bcfhedga
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bcfhdega
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); // bcfdhega
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bcdfhega
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfehgcda
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bfehcgda
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); // bfehcdga
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfechgda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bfechdga
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); // bfecdhga
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfcehgda
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bfcehdga
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v5, v7); // bfcedhga
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v5, v7); // bfcdehga
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbfehgda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cbfehdga
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); // cbfedhga
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); // cdbfehga
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v5, v7); // cbfdehga
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); // cbdfehga
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcfehgda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bcfehdga
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); // bcfedhga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v5, v7); // bcfdehga
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); // bcdfehga
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v3)) {
												if (compare(v7, v2)) {
													if (compare(v2, v1)) {
														if (compare(v3, v1)) {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v6, v7); // hfecdbag
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fhecdbag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fehcdbag
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v6, v7); // hfcedbag
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fhcedbag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v6, v7); // hcfedbag
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v6, v7); // hfcdebag
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v5, v7); swap(v6, v7); // fhcdebag
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v6, v7); // hcfdebag
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v6, v7); // hcdfebag
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v4, v5); swap(v6, v7); // hfcdbeag
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhcdbeag
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v6, v7); // hcfdbeag
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hcdfbeag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v6, v7); // hcdbfeag
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfcdbaeg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v4, v7); swap(v6, v7); // fhcdbaeg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcfdbaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v1)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcdfbaeg
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcdbfaeg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v6, v7); // hcdbafeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hfecbdag
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhecbdag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fehcbdag
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hfcebdag
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhcebdag
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v6, v7); // hfcbedag
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hfcbdeag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfcbdaeg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fhcbedag
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhcbdeag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fhcbdaeg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hcfebdag
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v6, v7); // hcfbedag
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hcfbdeag
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v6, v7); // hcbfedag
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hcbfdeag
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v4, v5); swap(v6, v7); // hcbdfeag
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcfbdaeg
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcbfdaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v2); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcbdfaeg
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v4, v7); swap(v6, v7); // hcbdafeg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v6, v7); // hfebcdag
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fhebcdag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fehbcdag
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hfbecdag
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhbecdag
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v6, v7); // hfbcedag
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hfbcdeag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfbcdaeg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fhbcedag
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhbcdeag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fhbcdaeg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hbfecdag
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v6, v7); // hbfcedag
																			}
																			else {
																				swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hbfcdeag
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v3, v5); swap(v6, v7); // hbcfedag
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hbcfdeag
																			}
																			else {
																				swap(v0, v7); swap(v4, v5); swap(v6, v7); // hbcdfeag
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbfcdaeg
																		}
																		else {
																			swap(v0, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbcfdaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbcdfaeg
																		}
																		else {
																			swap(v0, v7); swap(v4, v7); swap(v6, v7); // hbcdafeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v1)) {
																	if (compare(v7, v4)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbhecdag
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fbhcedag
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // febhcdag
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fbehcdag
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bhfecdag
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfhecdag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bfehcdag
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v6, v7); // bhfcedag
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bfhcedag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v6, v7); // bhcfedag
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbhcdeag
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bhfcdeag
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfhcdeag
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bhcfdeag
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v4, v5); swap(v6, v7); // bhcdfeag
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fbhcdaeg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhfcdaeg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bfhcdaeg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhcfdaeg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhcdfaeg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v4, v7); swap(v6, v7); // bhcdafeg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v3, v1)) {
															if (compare(v4, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v4)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fchedbag
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v5, v7); swap(v6, v7); // fchdebag
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fechdbag
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcehdbag
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // chfedbag
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v6, v7); // cfhedbag
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cfehdbag
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v5, v7); swap(v6, v7); // chfdebag
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v6, v7); // cfhdebag
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // chdfebag
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fchdbeag
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chfdbeag
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v6, v7); // cfhdbeag
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chdfbeag
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // chdbfeag
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v6, v7); // fchdbaeg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v6, v7); // chfdbaeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfhdbaeg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v1)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // chdfbaeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chdbfaeg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // chdbafeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v4, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fechbdag
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fchebdag
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fcehbdag
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chfebdag
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cfhebdag
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfehbdag
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // chfbedag
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v6, v7); // chbfedag
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fchbedag
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v6, v7); // cfhbedag
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fchbdeag
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chfbdeag
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cfhbdeag
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // chbfdeag
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v6, v7); // chbdfeag
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fchbdaeg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // chfbdaeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfhbdaeg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chbfdaeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chbdfaeg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v6, v7); // chbdafeg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v4)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fcbhedag
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcbhdeag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fcbhdaeg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fecbhdag
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcebhdag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcbehdag
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cfebhdag
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cfbhedag
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfbehdag
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfbhdeag
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cfbhdaeg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v6, v7); // cbhfedag
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cbfhedag
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbfehdag
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cbhfdeag
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbfhdeag
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v4, v5); swap(v6, v7); // cbhdfeag
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbhfdaeg
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cbfhdaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbhdfaeg
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v4, v7); swap(v6, v7); // cbhdafeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v1)) {
																	if (compare(v7, v4)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fbchedag
																			}
																			else {
																				swap(v0, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbchdeag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fbchdaeg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // febchdag
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbechdag
																			}
																			else {
																				swap(v0, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbcehdag
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfechdag
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bfchedag
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfcehdag
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfchdeag
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bfchdaeg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v6, v7); // bchfedag
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bcfhedag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcfehdag
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bchfdeag
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcfhdeag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v6, v7); // bchdfeag
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bchfdaeg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bcfhdaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bchdfaeg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v6, v7); // bchdafeg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v4)) {
													if (compare(v3, v1)) {
														if (compare(v7, v1)) {
															if (compare(v4, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fcdhebag
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v6, v7); // cfdhebag
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v6, v7); // cdhfebag
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cdfhebag
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcdhbeag
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cfdhbeag
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cdhfbeag
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdfhbeag
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v6, v7); // cdhbfeag
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdhfbaeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdhbfaeg
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v6, v7); // cdhbafeg
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fcdhbaeg
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfdhbaeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cdfhbaeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcdbheag
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfdbheag
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdfbheag
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v6, v7); // fcdbhaeg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cfdbhaeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v6, v7); // cdfbhaeg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcdbaheg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfdbaheg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdfbaheg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cdbhfeag
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdbhfaeg
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cdbhafeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v6, v7); // cdbahfeg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdbfheag
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cdbfhaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdbfaheg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdbafheg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcbdheag
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfbdheag
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbfdheag
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v4, v7); swap(v6, v7); // fcbdhaeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v6, v7); // cfbdhaeg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v4, v7); swap(v6, v7); // cbfdhaeg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcbdaheg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfbdaheg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbfdaheg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cbdhfeag
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbdhfaeg
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cbdhafeg
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v4, v7); swap(v6, v7); // cbdahfeg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbdfheag
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cbdfhaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbdfaheg
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbdafheg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		swap(v0, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbcdheag
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfcdheag
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcfdheag
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v4, v7); swap(v6, v7); // fbcdhaeg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v4, v7); swap(v6, v7); // bfcdhaeg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v6, v7); // bcfdhaeg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbcdaheg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfcdaheg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcfdaheg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bcdhfeag
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcdhfaeg
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bcdhafeg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v6, v7); // bcdahfeg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcdfheag
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bcdfhaeg
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcdfaheg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcdafheg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v7, v0)) {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fecdhbag
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v5, v7); swap(v6, v7); // fecdbhag
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fecbdhag
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // febcdhag
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcedhbag
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcdehbag
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v5, v7); swap(v6, v7); // fcedbhag
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fcdebhag
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fcebdhag
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fcbedhag
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fbecdhag
																			}
																			else {
																				swap(v0, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fbcedhag
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bfecdhag
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bfcedhag
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v5, v7); swap(v6, v7); // fcdbehag
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v5, v7); swap(v6, v7); // fcbdehag
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v5, v7); swap(v6, v7); // fbcdehag
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v5, v7); swap(v6, v7); // bfcdehag
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v6, v7); // fecdbahg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v6, v7); // fcedbahg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v6, v7); // fcdebahg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v6, v7); // fcdbeahg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v6, v7); // fcdbaehg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v6, v7); // fecbdahg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v6, v7); // fcebdahg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v6, v7); // fcbedahg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v6, v7); // fcbdeahg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v4, v5); swap(v6, v7); // fcbdaehg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v6, v7); // febcdahg
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v6, v7); // fbecdahg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v3, v4); swap(v6, v7); // fbcedahg
																			}
																			else {
																				swap(v0, v5); swap(v6, v7); // fbcdeahg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v4, v5); swap(v6, v7); // fbcdaehg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v6, v7); // bfecdahg
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v6, v7); // bfcedahg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v6, v7); // bfcdeahg
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v4, v5); swap(v6, v7); // bfcdaehg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v7, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v6, v7); // cfedhbag
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cfdehbag
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdfehbag
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfedbhag
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfebdhag
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfdebhag
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cdfebhag
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cfdbehag
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v5, v7); swap(v6, v7); // cdfbehag
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cfbedhag
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v5, v7); swap(v6, v7); // cfbdehag
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v6, v7); // cfedbahg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cfdebahg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v6, v7); // cdfebahg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v6, v7); // cfdbeahg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v6, v7); // cdfbeahg
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cfdbaehg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v6, v7); // cdfbaehg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cfebdahg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v6, v7); // cfbedahg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v6, v7); // cfbdeahg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v6, v7); // cfbdaehg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cbfedhag
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v5, v7); swap(v6, v7); // cbfdehag
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v6, v7); // cbfedahg
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v6, v7); // cbfdeahg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v4, v5); swap(v6, v7); // cbfdaehg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bcfedhag
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v5, v7); swap(v6, v7); // bcfdehag
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v6, v7); // bcfedahg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v6, v7); // bcfdeahg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v6, v7); // bcfdaehg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cdbfehag
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cbdfehag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bcdfehag
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v6, v7); // cdbfeahg
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v6, v7); // cbdfeahg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v6, v7); // bcdfeahg
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cdbfaehg
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cbdfaehg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bcdfaehg
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v6, v7); // cdbafehg
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v4, v5); swap(v6, v7); // cbdafehg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v6, v7); // bcdafehg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v7, v4)) {
										if (compare(v4, v0)) {
											if (compare(v7, v1)) {
												if (compare(v3, v1)) {
													if (compare(v6, v2)) {
														if (compare(v5, v2)) {
															if (compare(v7, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ghfecdba
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfhecdba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fghecdba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v6, v7); // ghfcedba
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gfhcedba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); // fghcedba
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ghfcdeba
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfhcdeba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fghcdeba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghfcdbea
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gfhcdbea
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fghcdbea
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gfchedba
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfchdeba
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfcdheba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gfchdbea
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v4, v7); swap(v6, v7); // gfcdhbea
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v5, v7); // fgchedba
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fgchdeba
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v4, v7); swap(v5, v7); // fgcdheba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgchdbea
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgcdhbea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v6, v7); // ghcfedba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v6, v7); // gchfedba
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gcfhedba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ghcfdeba
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gchfdeba
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcfhdeba
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcfdheba
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v5); swap(v6, v7); // ghcdfeba
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v6, v7); // gchdfeba
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gcdhfeba
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcdfheba
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghcfdbea
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gchfdbea
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gcfhdbea
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v6, v7); // gcfdhbea
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghcdfbea
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gchdfbea
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcdhfbea
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gcdfhbea
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v7); swap(v6, v7); // ghcdbfea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v6, v7); // gchdbfea
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gcdhbfea
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v7, v3)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); // cghfedba
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); // cgfhedba
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); // fcghedba
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cfghedba
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); // cghfdeba
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cgfhdeba
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v5); // cghdfeba
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fcghdeba
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfghdeba
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); // fcgdheba
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fcdgheba
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); // cgfdheba
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfgdheba
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfdgheba
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); // cgdhfeba
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cgdfheba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cdghfeba
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdgfheba
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cdfgheba
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cghfdbea
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgfhdbea
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcghdbea
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cfghdbea
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcgdhbea
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcdghbea
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgfdhbea
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v6, v7); // cfgdhbea
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cfdghbea
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); // cghdfbea
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cgdhfbea
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgdfhbea
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v6); // cghdbfea
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); // cgdhbfea
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdghfbea
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cdghbfea
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cdgfhbea
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdfghbea
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghfecbda
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ghfcebda
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghfcbeda
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ghfcbdea
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ghfebcda
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v6, v7); // ghfbecda
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ghfbceda
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghfbcdea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ghcfebda
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gchfebda
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); // cghfebda
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghcfbeda
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gchfbeda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cghfbeda
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ghcfbdea
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gchfbdea
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); // cghfbdea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ghcbfeda
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghcbfdea
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ghcbdfea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gchbfeda
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gchbfdea
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gchbdfea
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); // cghbfeda
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cghbfdea
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); // cghbdfea
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ghbfecda
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghbfceda
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ghbfcdea
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ghbcfeda
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghbcfdea
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ghbcdfea
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gfhecbda
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v6, v7); // gfhcebda
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gfhcbeda
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfhcbdea
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfhebcda
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gfhbecda
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfhbceda
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gfhbcdea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v2)) {
																		swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v6, v7); // gfchebda
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v6, v7); // gcfhebda
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cgfhebda
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gfchbeda
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gcfhbeda
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgfhbeda
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfchbdea
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcfhbdea
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cgfhbdea
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fghecbda
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fghcebda
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fghcbeda
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fghcbdea
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fghebcda
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); // fghbecda
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fghbceda
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fghbcdea
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fgchebda
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fcghebda
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v6, v7); // cfghebda
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgchbeda
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcghbeda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cfghbeda
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fgchbdea
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fcghbdea
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfghbdea
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v1)) {
													if (compare(v5, v1)) {
														if (compare(v6, v5)) {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfcbheda
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gfcbhdea
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfcdbhea
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfcbdhea
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v6, v7); // gfbhecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gfbhceda
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfbhcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfbcheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gfbchdea
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfbcdhea
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcfdbhea
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcdfbhea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); // cgfdbhea
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cgdfbhea
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdgfbhea
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcfbheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gcfbhdea
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcfbdhea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cgfbheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgfbhdea
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cgfbdhea
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fgcbheda
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgcbhdea
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v4, v6); swap(v5, v7); // fgcdbhea
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fgcbdhea
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fgbhecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgbhceda
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fgbhcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fgbcheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgbchdea
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fgbcdhea
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); // fcgdbhea
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fcdgbhea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfgdbhea
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfdgbhea
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cdfgbhea
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fcgbheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcgbhdea
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fcgbdhea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfgbheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cfgbhdea
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfgbdhea
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gcbhfeda
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcbhfdea
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v6, v7); // gcdbhfea
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gcbhdfea
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v4, v7); swap(v6, v7); // gcbdhfea
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cgbhfeda
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgbhfdea
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cgbhdfea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cgdbhfea
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v6, v7); // cdgbhfea
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v6, v7); // cgbdhfea
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gbhfecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbhfceda
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gbhfcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gbhcfeda
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbhcfdea
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gbhcdfea
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gbchfeda
																		}
																		else {
																			swap(v0, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbchfdea
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gbchdfea
																		}
																		else {
																			swap(v0, v6); swap(v4, v7); swap(v6, v7); // gbcdhfea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcbfheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gcbfhdea
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcbfdhea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcdbfhea
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcbdfhea
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgbfheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cgbfhdea
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgbfdhea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgdbfhea
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdgbfhea
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgbdfhea
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v6, v7); // gbfhecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gbfhceda
																			}
																			else {
																				swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbfhcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbfcheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gbfchdea
																			}
																			else {
																				swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbfcdhea
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbcfheda
																		}
																		else {
																			swap(v0, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gbcfhdea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbcfdhea
																		}
																		else {
																			swap(v0, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbcdfhea
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v7, v5)) {
															if (compare(v6, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bghfecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bghfceda
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bghfcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bghcfeda
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bghcfdea
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bghcdfea
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bgchfeda
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgchfdea
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bgchdfea
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v4, v7); swap(v6, v7); // bgcdhfea
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cbghfeda
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbghfdea
																			}
																			else {
																				swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cbghdfea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cdbghfea
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v4, v7); swap(v6, v7); // cbgdhfea
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cbdghfea
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bcghfeda
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcghfdea
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bcghdfea
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v6, v7); // bcgdhfea
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bcdghfea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v6, v7); // bgfhecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bgfhceda
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgfhcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgfcheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bgfchdea
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgfcdhea
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgcfheda
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bgcfhdea
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgcfdhea
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgcdfhea
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbgfheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cbgfhdea
																			}
																			else {
																				swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbgfdhea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdbgfhea
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbgdfhea
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbdgfhea
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcgfheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bcgfhdea
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcgfdhea
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcgdfhea
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcdgfhea
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fcbgheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcbghdea
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fcbgdhea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); // fcdbghea
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v4, v6); swap(v5, v7); // fcbdghea
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cfbgheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfbghdea
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cfbgdhea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cfdbghea
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); // cdfbghea
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); // cfbdghea
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fbghecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbghceda
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fbghcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fbgcheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbgchdea
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fbgcdhea
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fbcgheda
																		}
																		else {
																			swap(v0, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbcghdea
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fbcgdhea
																		}
																		else {
																			swap(v0, v5); swap(v4, v6); swap(v5, v7); // fbcdghea
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bfghecda
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfghceda
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bfghcdea
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bfgcheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfgchdea
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bfgcdhea
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bfcgheda
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfcghdea
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bfcgdhea
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v4, v6); swap(v5, v7); // bfcdghea
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cbfgheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbfghdea
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cbfgdhea
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cdbfghea
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); // cbfdghea
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cbdfghea
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bcfgheda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcfghdea
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bcfgdhea
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); // bcfdghea
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bcdfghea
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v3)) {
												if (compare(v5, v2)) {
													if (compare(v2, v1)) {
														if (compare(v3, v1)) {
															if (compare(v7, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ghfcdbae
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); // gfhcdbae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fghcdbae
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v7); // gfchdbae
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fgchdbae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fcghdbae
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v4, v7); // gfcdhbae
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v4, v7); swap(v5, v6); // fgcdhbae
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); // fcgdhbae
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fcdghbae
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v4, v5); swap(v5, v7); // gfcdbhae
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgcdbhae
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcgdbhae
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcdgbhae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcdbghae
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfcdbahe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v4, v6); swap(v6, v7); // fgcdbahe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v6, v7); // fcgdbahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fcdgbahe
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v6, v7); // fcdbgahe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcdbaghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); // ghfcbdae
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gfhcbdae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fghcbdae
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gfchbdae
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgchbdae
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v7); // gfcbhdae
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gfcbdhae
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfcbdahe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fgcbhdae
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgcbdhae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fgcbdahe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcghbdae
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fcgbhdae
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcgbdhae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fcbghdae
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcbgdhae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcbdghae
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fcgbdahe
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fcbgdahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v2); swap(v4, v6); swap(v6, v7); // fcbdgahe
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcbdaghe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ghfbcdae
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); // gfhbcdae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fghbcdae
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gfbhcdae
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgbhcdae
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); // gfbchdae
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gfbcdhae
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfbcdahe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fgbchdae
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgbcdhae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fgbcdahe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbghcdae
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fbgchdae
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbgcdhae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fbcghdae
																			}
																			else {
																				swap(v0, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbcgdhae
																			}
																		}
																		else {
																			swap(v0, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbcdghae
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fbgcdahe
																		}
																		else {
																			swap(v0, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fbcgdahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v4, v6); swap(v6, v7); // fbcdgahe
																		}
																		else {
																			swap(v0, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbcdaghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); // ghbfcdae
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); // gbhfcdae
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gbfhcdae
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); // gbfchdae
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); // bghfcdae
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bgfhcdae
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfghcdae
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); // bgfchdae
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bfgchdae
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bfcghdae
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gbfcdhae
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bgfcdhae
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfgcdhae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfcgdhae
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfcdghae
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbfcdahe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgfcdahe
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bfgcdahe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bfcgdahe
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v4, v6); swap(v6, v7); // bfcdgahe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfcdaghe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v3, v1)) {
															if (compare(v7, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ghcfdbae
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gchfdbae
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); // gcfhdbae
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v7); // gcfdhbae
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cghfdbae
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cgfhdbae
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); // cfghdbae
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); // cgfdhbae
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v7); // cfgdhbae
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); // cfdghbae
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); // gcfdbhae
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgfdbhae
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); // cfgdbhae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cfdgbhae
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfdbghae
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcfdbahe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v6, v7); // cgfdbahe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfgdbahe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfdgbahe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cfdbgahe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfdbaghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v7); // ghcfbdae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); // gchfbdae
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gcfhbdae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cghfbdae
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgfhbdae
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cfghbdae
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); // gcfbhdae
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cgfbhdae
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); // cfgbhdae
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cfbghdae
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gcfbdhae
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgfbdhae
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cfgbdhae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfbgdhae
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfbdghae
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcfbdahe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cgfbdahe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfgbdahe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cfbgdahe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v6, v7); // cfbdgahe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfbdaghe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v6, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ghcbfdae
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gchbfdae
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gcbhfdae
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v7); // gcbfhdae
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gcbfdhae
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcbfdahe
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cghbfdae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cgbhfdae
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); // cgbfhdae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cgbfdhae
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgbfdahe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cbghfdae
																			}
																			else {
																				swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); // cbgfhdae
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cbfghdae
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cbgfdhae
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbfgdhae
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbfdghae
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbgfdahe
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cbfgdahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v2, v5); swap(v4, v6); swap(v6, v7); // cbfdgahe
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbfdaghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ghbcfdae
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gbhcfdae
																			}
																			else {
																				swap(v0, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gbchfdae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v3, v5); swap(v4, v7); // gbcfhdae
																			}
																			else {
																				swap(v0, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gbcfdhae
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbcfdahe
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bghcfdae
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bgchfdae
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v7); // bgcfhdae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bgcfdhae
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgcfdahe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bcghfdae
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); // bcgfhdae
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bcfghdae
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bcgfdhae
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcfgdhae
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcfdghae
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcgfdahe
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bcfgdahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v6, v7); // bcfdgahe
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcfdaghe
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v5)) {
													if (compare(v6, v2)) {
														if (compare(v5, v0)) {
															if (compare(v7, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v5); swap(v5, v7); // ghcdfbae
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v4, v7); // ghcdbfae
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v7); // ghcbdfae
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); // ghbcdfae
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); // gchdfbae
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gcdhfbae
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v7); // gchdbfae
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); // gcdhbfae
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); // gchbdfae
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v7); // gcbhdfae
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); // gbhcdfae
																			}
																			else {
																				swap(v0, v6); swap(v3, v7); swap(v4, v7); // gbchdfae
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); // bghcdfae
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v7); // bgchdfae
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v7); // gcdbhfae
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v4, v7); // gcbdhfae
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v4, v7); // gbcdhfae
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v4, v7); // bgcdhfae
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v7); swap(v5, v6); // ghcdbafe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); // gchdbafe
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gcdhbafe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); // gcdbhafe
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcdbahfe
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ghcbdafe
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gchbdafe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gcbhdafe
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v4, v7); swap(v5, v6); // gcbdhafe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcbdahfe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ghbcdafe
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gbhcdafe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gbchdafe
																			}
																			else {
																				swap(v0, v6); swap(v4, v7); swap(v5, v6); // gbcdhafe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbcdahfe
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bghcdafe
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bgchdafe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v4, v7); swap(v5, v6); // bgcdhafe
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgcdahfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cghdfbae
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgdhfbae
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cdghfbae
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v6, v7); // cghdbfae
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cghbdfae
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cgdhbfae
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); // cdghbfae
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); // cgdbhfae
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v7); // cdgbhfae
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); // cgbhdfae
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v7); // cgbdhfae
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cghdbafe
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgdhbafe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cdghbafe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cgdbhafe
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); // cdgbhafe
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgdbahfe
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdgbahfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cghbdafe
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cgbhdafe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); // cgbdhafe
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgbdahfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); // cbghdfae
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v4, v7); // cbgdhfae
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cbghdafe
																			}
																			else {
																				swap(v0, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); // cbgdhafe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbgdahfe
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); // bcghdfae
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v7); // bcgdhfae
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bcghdafe
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); // bcgdhafe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcgdahfe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); // cdbghfae
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); // cbdghfae
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); // bcdghfae
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cdbghafe
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cbdghafe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bcdghafe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdbgahfe
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbdgahfe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcdgahfe
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdbaghfe
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbdaghfe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcdaghfe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v5, v1)) {
															if (compare(v7, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); // gcdfhbae
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cgdfhbae
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); // cdgfhbae
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cdfghbae
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gcdfbhae
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgdfbhae
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cdgfbhae
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdfgbhae
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdfbghae
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v6, v2)) {
																		swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcdfbahe
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cgdfbahe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdgfbahe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cdfgbahe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v6, v7); // cdfbgahe
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdfbaghe
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); // gcdbfhae
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cgdbfhae
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); // cdgbfhae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcdbfahe
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgdbfahe
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdgbfahe
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v6, v7); // gcdbafhe
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cgdbafhe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v6, v7); // cdgbafhe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cdbgfhae
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdbgfahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cdbgafhe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v6, v7); // cdbagfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdbfghae
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cdbfgahe
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdbfaghe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdbafghe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v6, v3)) {
																if (compare(v5, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v4, v5); swap(v5, v7); // gcbdfhae
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); // cgbdfhae
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); // cbgdfhae
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcbdfahe
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgbdfahe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbgdfahe
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v4, v6); swap(v6, v7); // gcbdafhe
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v6, v7); // cgbdafhe
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v6); swap(v4, v6); swap(v6, v7); // cbgdafhe
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cbdgfhae
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbdgfahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cbdgafhe
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v4, v6); swap(v6, v7); // cbdagfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbdfghae
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cbdfgahe
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbdfaghe
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbdafghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v5, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v4, v5); swap(v5, v7); // gbcdfhae
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v4, v5); swap(v5, v7); // bgcdfhae
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); // bcgdfhae
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbcdfahe
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgcdfahe
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcgdfahe
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		swap(v0, v6); swap(v4, v6); swap(v6, v7); // gbcdafhe
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v4, v6); swap(v6, v7); // bgcdafhe
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v6, v7); // bcgdafhe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bcdgfhae
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcdgfahe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bcdgafhe
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v6, v7); // bcdagfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcdfghae
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bcdfgahe
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcdfaghe
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcdafghe
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v7, v0)) {
											if (compare(v4, v1)) {
												if (compare(v3, v1)) {
													if (compare(v5, v2)) {
														if (compare(v6, v2)) {
															if (compare(v4, v2)) {
																if (compare(v7, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gfehcdba
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); // fgehcdba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // feghcdba
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfechdba
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fgechdba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fegchdba
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gfecdhba
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); // fgecdhba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fegcdhba
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v6, v7); // gfecdbha
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fgecdbha
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fegcdbha
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfcehdba
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gfcedhba
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v5, v7); swap(v6, v7); // gfcdehba
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v6, v7); // gfcedbha
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v6, v7); // gfcdebha
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fgcehdba
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v5, v7); // fgcedhba
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v5, v7); // fgcdehba
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fgcedbha
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v5, v6); swap(v6, v7); // fgcdebha
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fecghdba
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fcgehdba
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fceghdba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); // fcgedhba
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v5, v7); // fcgdehba
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fecgdhba
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fcegdhba
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v4, v6); swap(v5, v7); // fecdghba
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); // fcedghba
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); // fcdgehba
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fcdeghba
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fecgdbha
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fecdgbha
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v5, v6); swap(v6, v7); // fecdbgha
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fcgedbha
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcegdbha
																			}
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcedgbha
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v5, v6); swap(v6, v7); // fcedbgha
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v5, v6); swap(v6, v7); // fcgdebha
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fcdgebha
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcdegbha
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fcdebgha
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v4, v3)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcfehdba
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cgfehdba
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfgehdba
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cfeghdba
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gcfedhba
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); // cgfedhba
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cfgedhba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfegdhba
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfedghba
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v5, v7); swap(v6, v7); // gcfdehba
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v5, v7); // cgfdehba
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v5, v7); swap(v6, v7); // cfgdehba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cfdgehba
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfdeghba
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gcdfehba
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); // cgdfehba
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cdgfehba
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); // cdfgehba
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cdfeghba
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v6, v7); // gcfedbha
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cgfedbha
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v6, v7); // cfgedbha
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cfegdbha
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v6, v7); // gcfdebha
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v6, v7); // gcdfebha
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v5, v6); swap(v6, v7); // cgfdebha
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v6, v7); // cfgdebha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cgdfebha
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v6, v7); // cfedgbha
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v6, v7); // cfdgebha
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cfdegbha
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfedbgha
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfdebgha
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v6, v7); // cdgfebha
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cdfgebha
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdfegbha
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cdfebgha
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v6, v5)) {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v6, v7); // gfehcbda
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gfechbda
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfecbhda
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gfecbdha
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gfehbcda
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfebhcda
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gfebchda
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v6, v7); // gfebcdha
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v5, v2)) {
																		swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gfcehbda
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gcfehbda
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgfehbda
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfcebhda
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcfebhda
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cgfebhda
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gfcebdha
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gcfebdha
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgfebdha
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fgehcbda
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgechbda
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fgecbhda
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgecbdha
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgehbcda
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fgebhcda
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); // fgebchda
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fgebcdha
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgcehbda
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcgehbda
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cfgehbda
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fgcebhda
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fcgebhda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfgebhda
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgcebdha
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcgebdha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cfgebdha
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // feghcbda
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fegchbda
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); // fegcbhda
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fegcbdha
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // feghbcda
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fegbhcda
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fegbchda
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fegbcdha
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fecghbda
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fceghbda
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); // cfeghbda
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v5, v7); // fecgbhda
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); // fcegbhda
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cfegbhda
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fecgbdha
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fcegbdha
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfegbdha
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fecbghda
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fecbgdha
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fecbdgha
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fcebghda
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcebgdha
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fcebdgha
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfebghda
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cfebgdha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfebdgha
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // febghcda
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); // febgchda
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // febgcdha
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // febcghda
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // febcgdha
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // febcdgha
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v6, v1)) {
														if (compare(v6, v5)) {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gfcbehda
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v6, v7); // gfcbedha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v4, v5); swap(v6, v7); // gfcdbeha
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gfcbdeha
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gfbehcda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfbechda
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gfbecdha
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gfbcehda
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v6, v7); // gfbcedha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gfbcdeha
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v6, v7); // gcfdbeha
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gcdfbeha
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgfdbeha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgdfbeha
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cdgfbeha
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gcfbehda
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v6, v7); // gcfbedha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gcfbdeha
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); // cgfbehda
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cgfbedha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgfbdeha
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v5, v7); // fgcbehda
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fgcbedha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgcdbeha
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgcbdeha
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgbehcda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fgbechda
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgbecdha
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); // fgbcehda
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fgbcedha
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgbcdeha
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcgdbeha
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcdgbeha
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v6, v7); // cfgdbeha
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cfdgbeha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdfgbeha
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); // fcgbehda
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fcgbedha
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcgbdeha
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cfgbehda
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v6, v7); // cfgbedha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cfgbdeha
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v5, v7); // fcbgehda
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fcbgedha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcdbgeha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcbgdeha
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcbdgeha
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); // cfbgehda
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cfbgedha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfbgdeha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfdbgeha
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdfbgeha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfbdgeha
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbgehcda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fbgechda
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbgecdha
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); // fbgcehda
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fbgcedha
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbgcdeha
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v3, v6); swap(v5, v7); // fbcgehda
																		}
																		else {
																			swap(v0, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fbcgedha
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbcgdeha
																		}
																		else {
																			swap(v0, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbcdgeha
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fcbeghda
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcbegdha
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fcbedgha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v5, v6); swap(v6, v7); // fcdbegha
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v5, v6); swap(v6, v7); // fcbdegha
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cfbeghda
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfbegdha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cfbedgha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cfdbegha
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v5, v6); swap(v6, v7); // cdfbegha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v5, v6); swap(v6, v7); // cfbdegha
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fbeghcda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); // fbegchda
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fbegcdha
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fbecghda
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbecgdha
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fbecdgha
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fbceghda
																		}
																		else {
																			swap(v0, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbcegdha
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fbcedgha
																		}
																		else {
																			swap(v0, v5); swap(v5, v6); swap(v6, v7); // fbcdegha
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v6, v1)) {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gcbfehda
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v6, v7); // gcbfedha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v6, v7); // gcdbfeha
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gcbfdeha
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v4, v5); swap(v6, v7); // gcbdfeha
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cgbfehda
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v6, v7); // cgbfedha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cgbfdeha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cgdbfeha
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v6, v7); // cdgbfeha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v6, v7); // cgbdfeha
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gbfehcda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbfechda
																			}
																			else {
																				swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gbfecdha
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gbfcehda
																			}
																			else {
																				swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v6, v7); // gbfcedha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gbfcdeha
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gbcfehda
																		}
																		else {
																			swap(v0, v6); swap(v3, v5); swap(v6, v7); // gbcfedha
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gbcfdeha
																		}
																		else {
																			swap(v0, v6); swap(v4, v5); swap(v6, v7); // gbcdfeha
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bgfehcda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgfechda
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bgfecdha
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bgfcehda
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v6, v7); // bgfcedha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bgfcdeha
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bgcfehda
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v6, v7); // bgcfedha
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bgcfdeha
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v4, v5); swap(v6, v7); // bgcdfeha
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cbgfehda
																			}
																			else {
																				swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v6, v7); // cbgfedha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cbgfdeha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cdbgfeha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v4, v5); swap(v6, v7); // cbgdfeha
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cbdgfeha
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bcgfehda
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v6, v7); // bcgfedha
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bcgfdeha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v6, v7); // bcgdfeha
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bcdgfeha
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v4)) {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfgehcda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bfgechda
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfgecdha
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); // bfgcehda
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bfgcedha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfgcdeha
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v5, v7); // bfcgehda
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bfcgedha
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfcgdeha
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfcdgeha
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); // cbfgehda
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cbfgedha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbfgdeha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdbfgeha
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbfdgeha
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbdfgeha
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); // bcfgehda
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bcfgedha
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcfgdeha
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcfdgeha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcdfgeha
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bfeghcda
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); // bfegchda
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bfegcdha
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bfecghda
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfecgdha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bfecdgha
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bfceghda
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfcegdha
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bfcedgha
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v5, v6); swap(v6, v7); // bfcdegha
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cbfeghda
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbfegdha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cbfedgha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cdbfegha
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v5, v6); swap(v6, v7); // cbfdegha
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cbdfegha
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bcfeghda
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcfegdha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bcfedgha
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v5, v6); swap(v6, v7); // bcfdegha
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bcdfegha
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v3)) {
												if (compare(v6, v2)) {
													if (compare(v2, v1)) {
														if (compare(v3, v1)) {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); // gfecdbah
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // fgecdbah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fegcdbah
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); // gfcedbah
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v5, v6); // fgcedbah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); // gcfedbah
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); // gfcdebah
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v5, v6); // fgcdebah
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); // gcfdebah
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); // gcdfebah
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v4, v5); // gfcdbeah
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v4, v6); swap(v5, v6); // fgcdbeah
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v5); // gcfdbeah
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // gcdfbeah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v5); // gcdbfeah
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v4, v5); swap(v5, v6); // gfcdbaeh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v4, v6); // fgcdbaeh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // gcfdbaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v1)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gcdfbaeh
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // gcdbfaeh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v6); // gcdbafeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // gfecbdah
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fgecbdah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // fegcbdah
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v5); // gfcebdah
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fgcebdah
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v5); // gfcbedah
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v5); // gfcbdeah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gfcbdaeh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v5, v6); // fgcbedah
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fgcbdeah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v6); // fgcbdaeh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // gcfebdah
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); // gcfbedah
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // gcfbdeah
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v5); // gcbfedah
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v5); // gcbfdeah
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v4, v5); // gcbdfeah
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gcfbdaeh
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gcbfdaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v2); swap(v4, v5); swap(v5, v6); // gcbdfaeh
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v4, v6); // gcbdafeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); // gfebcdah
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // fgebcdah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fegbcdah
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // gfbecdah
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fgbecdah
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); // gfbcedah
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // gfbcdeah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gfbcdaeh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // fgbcedah
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fgbcdeah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // fgbcdaeh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // gbfecdah
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v5); // gbfcedah
																			}
																			else {
																				swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // gbfcdeah
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v3, v5); // gbcfedah
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v3, v5); swap(v4, v5); // gbcfdeah
																			}
																			else {
																				swap(v0, v6); swap(v4, v5); // gbcdfeah
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gbfcdaeh
																		}
																		else {
																			swap(v0, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gbcfdaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v4, v5); swap(v5, v6); // gbcdfaeh
																		}
																		else {
																			swap(v0, v6); swap(v4, v6); // gbcdafeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v4)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fbgecdah
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // fbgcedah
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // febgcdah
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // fbegcdah
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bgfecdah
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bfgecdah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // bfegcdah
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); // bgfcedah
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // bfgcedah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); // bgcfedah
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fbgcdeah
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bgfcdeah
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bfgcdeah
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v5); // bgcfdeah
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v4, v5); // bgcdfeah
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // fbgcdaeh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bgfcdaeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // bfgcdaeh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bgcfdaeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v4, v5); swap(v5, v6); // bgcdfaeh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v4, v6); // bgcdafeh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v3, v1)) {
															if (compare(v4, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v4)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // fcgedbah
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v5, v6); // fcgdebah
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fecgdbah
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fcegdbah
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // cgfedbah
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); // cfgedbah
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // cfegdbah
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v5, v6); // cgfdebah
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); // cfgdebah
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // cgdfebah
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // fcgdbeah
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // cgfdbeah
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v5); // cfgdbeah
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cgdfbeah
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // cgdbfeah
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v6); // fcgdbaeh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v6); // cgfdbaeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // cfgdbaeh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v1)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // cgdfbaeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cgdbfaeh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // cgdbafeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v4, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v5, v6); // fecgbdah
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fcgebdah
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // fcegbdah
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cgfebdah
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // cfgebdah
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cfegbdah
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // cgfbedah
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); // cgbfedah
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // fcgbedah
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); // cfgbedah
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fcgbdeah
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cgfbdeah
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // cfgbdeah
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // cgbfdeah
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v5); // cgbdfeah
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // fcgbdaeh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // cgfbdaeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cfgbdaeh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cgbfdaeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // cgbdfaeh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v6); // cgbdafeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v4)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v5, v6); // fcbgedah
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fcbgdeah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v6); // fcbgdaeh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fecbgdah
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fcebgdah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fcbegdah
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // cfebgdah
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // cfbgedah
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cfbegdah
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cfbgdeah
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // cfbgdaeh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v5); // cbgfedah
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // cbfgedah
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cbfegdah
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // cbgfdeah
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cbfgdeah
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v4, v5); // cbgdfeah
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cbgfdaeh
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // cbfgdaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // cbgdfaeh
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v4, v6); // cbgdafeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v4)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v3, v6); swap(v5, v6); // fbcgedah
																			}
																			else {
																				swap(v0, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fbcgdeah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v3, v6); swap(v4, v6); // fbcgdaeh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // febcgdah
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fbecgdah
																			}
																			else {
																				swap(v0, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fbcegdah
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bfecgdah
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v5, v6); // bfcgedah
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bfcegdah
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bfcgdeah
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v6); // bfcgdaeh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); // bcgfedah
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // bcfgedah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bcfegdah
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // bcgfdeah
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bcfgdeah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v5); // bcgdfeah
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bcgfdaeh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // bcfgdaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // bcgdfaeh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v6); // bcgdafeh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v4)) {
													if (compare(v3, v1)) {
														if (compare(v6, v1)) {
															if (compare(v4, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // fcdgebah
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); // cfdgebah
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); // cdgfebah
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // cdfgebah
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fcdgbeah
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // cfdgbeah
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // cdgfbeah
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cdfgbeah
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v5); // cdgbfeah
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cdgfbaeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // cdgbfaeh
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v6); // cdgbafeh
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // fcdgbaeh
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cfdgbaeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // cdfgbaeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // fcdbgeah
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cfdbgeah
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // cdfbgeah
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v6); // fcdbgaeh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // cfdbgaeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v6); // cdfbgaeh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // fcdbageh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cfdbageh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // cdfbageh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // cdbgfeah
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cdbgfaeh
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // cdbgafeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v6); // cdbagfeh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cdbfgeah
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // cdbfgaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cdbfageh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // cdbafgeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v4, v6); swap(v5, v6); // fcbdgeah
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // cfbdgeah
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // cbfdgeah
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v4, v6); // fcbdgaeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v6); // cfbdgaeh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v4, v6); // cbfdgaeh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v4, v5); swap(v5, v6); // fcbdageh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // cfbdageh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // cbfdageh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // cbdgfeah
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cbdgfaeh
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // cbdgafeh
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v4, v6); // cbdagfeh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cbdfgeah
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // cbdfgaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cbdfageh
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // cbdafgeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		swap(v0, v5); swap(v4, v6); swap(v5, v6); // fbcdgeah
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v4, v6); swap(v5, v6); // bfcdgeah
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // bcfdgeah
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v4, v6); // fbcdgaeh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v4, v6); // bfcdgaeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v6); // bcfdgaeh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v4, v5); swap(v5, v6); // fbcdageh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v4, v5); swap(v5, v6); // bfcdageh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // bcfdageh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // bcdgfeah
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bcdgfaeh
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // bcdgafeh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v6); // bcdagfeh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bcdfgeah
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // bcdfgaeh
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bcdfageh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // bcdafgeh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v6, v0)) {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v4, v6); swap(v5, v6); // fecdgbah
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v5, v6); // fecdbgah
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v5, v6); // fecbdgah
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // febcdgah
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // fcedgbah
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fcdegbah
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v5, v6); // fcedbgah
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // fcdebgah
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // fcebdgah
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v5, v6); // fcbedgah
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // fbecdgah
																			}
																			else {
																				swap(v0, v5); swap(v3, v4); swap(v5, v6); // fbcedgah
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // bfecdgah
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v5, v6); // bfcedgah
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v5, v6); // fcdbegah
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v5, v6); // fcbdegah
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v5, v6); // fbcdegah
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v5, v6); // bfcdegah
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); // fecdbagh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); // fcedbagh
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); // fcdebagh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); // fcdbeagh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v4, v5); // fcdbaegh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v4); // fecbdagh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); // fcebdagh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v4); // fcbedagh
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); // fcbdeagh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v4, v5); // fcbdaegh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); // febcdagh
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v4); // fbecdagh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v3, v4); // fbcedagh
																			}
																			else {
																				swap(v0, v5); // fbcdeagh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v4, v5); // fbcdaegh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); // bfecdagh
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); // bfcedagh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v5); // bfcdeagh
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v4, v5); // bfcdaegh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v6); // cfedgbah
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // cfdegbah
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cdfegbah
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // cfedbgah
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cfebdgah
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cfdebgah
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // cdfebgah
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // cfdbegah
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v5, v6); // cdfbegah
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // cfbedgah
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v5, v6); // cfbdegah
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v5); // cfedbagh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cfdebagh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); // cdfebagh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); // cfdbeagh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); // cdfbeagh
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cfdbaegh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v4, v5); // cdfbaegh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // cfebdagh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); // cfbedagh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); // cfbdeagh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v4, v5); // cfbdaegh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // cbfedgah
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v5, v6); // cbfdegah
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v3, v4); // cbfedagh
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); // cbfdeagh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v4, v5); // cbfdaegh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // bcfedgah
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v5, v6); // bcfdegah
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); // bcfedagh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); // bcfdeagh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v4, v5); // bcfdaegh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // cdbfegah
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // cbdfegah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // bcdfegah
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); // cdbfeagh
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v5); // cbdfeagh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); // bcdfeagh
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cdbfaegh
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cbdfaegh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // bcdfaegh
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v5); // cdbafegh
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v4, v5); // cbdafegh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v5); // bcdafegh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v7, v6)) {
									if (compare(v7, v4)) {
										if (compare(v4, v1)) {
											if (compare(v6, v1)) {
												if (compare(v5, v1)) {
													if (compare(v4, v2)) {
														if (compare(v5, v2)) {
															if (compare(v6, v5)) {
																if (compare(v6, v4)) {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); // hgefcdba
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // hgefcbda
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hgefbcda
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // hegfcdba
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hegfcbda
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // hegfbcda
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hefgcdba
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // hefcgdba
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hefcdgba
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // hefgcbda
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hefcgbda
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); // hefgbcda
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v6, v4)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // hgecfdba
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); // hgecdfba
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hgecfbda
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hegcfdba
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // hegcdfba
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // hegcfbda
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hecgfdba
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v5); // hecgfbda
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v6); // hecgdfba
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v4, v6); // hecdgfba
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v6); // hecfgdba
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hecfgbda
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hecfdgba
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v4, v5); swap(v5, v6); // hecdfgba
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v5); // hgcefdba
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // hcgefdba
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hcegfdba
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // hcefgdba
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hcefdgba
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hgcefbda
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hcgefbda
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // hcegfbda
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hcefgbda
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); // hgcedfba
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); // hcgedfba
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // hcegdfba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v6); // hcedgfba
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // hcedfgba
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); // hgcdefba
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); // hcgdefba
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); // hcdgefba
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // hcdegfba
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hcdefgba
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v6, v4)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // chgefdba
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chgefbda
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v6, v7); // chgedfba
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v6, v7); // chgdefba
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chegfdba
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // chefgdba
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // chegdfba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // chegfbda
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chefgbda
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v6, v7); // chedgfba
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v6, v7); // chdgefba
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // chdegfba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); // cdhgefba
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); // cdhegfba
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chefdgba
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chedfgba
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chdefgba
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cdhefgba
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v4)) {
														if (compare(v3, v1)) {
															if (compare(v5, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // hgecdbfa
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v5, v6); // hgcedbfa
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v5, v6); // hgcdebfa
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // hcgedbfa
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v5, v6); // hcgdebfa
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // hcdgebfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // chgedbfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v5, v7); swap(v6, v7); // chgdebfa
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // chdgebfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); // cdhgebfa
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hgecdbaf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hgcedbaf
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v5, v6); swap(v6, v7); // hgcdebaf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hcgedbaf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v5, v6); swap(v6, v7); // hcgdebaf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hcdgebaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); // chgedbaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v5, v7); // chgdebaf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); // chdgebaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cdhgebaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // hgecbfda
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hgecbdfa
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgecbdaf
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // hgebfcda
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); // hgebcfda
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // hgebcdfa
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hgebcdaf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v6); // hgcebfda
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // hcgebfda
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // chgebfda
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hgcebdfa
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hcgebdfa
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chgebdfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgcebdaf
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcgebdaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // chgebdaf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v5, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hegcdbfa
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hecgdbfa
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v4, v6); swap(v5, v6); // hecdgbfa
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hcegdbfa
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // hcedgbfa
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hcdegbfa
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chegdbfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chedgbfa
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chdegbfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cdhegbfa
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hegcdbaf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hecgdbaf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hecdgbaf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcegdbaf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcedgbaf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcdegbaf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // chegdbaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); // chedgbaf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // chdegbaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdhegbaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); // hegcbfda
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // hegcbdfa
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hegcbdaf
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hegbfcda
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // hegbcfda
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hegbcdfa
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hegbcdaf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v6); // hecgbfda
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); // hcegbfda
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); // chegbfda
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v5, v6); // hecgbdfa
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // hcegbdfa
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // chegbdfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hecgbdaf
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hcegbdaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chegbdaf
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v6, v0)) {
														if (compare(v7, v2)) {
															if (compare(v5, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // hefcdbga
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v5); // hecfdbga
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v4, v5); // hecdfbga
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // hcefdbga
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v5); // hcedfbga
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // hcdefbga
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v5, v6); // hecdbgfa
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v5, v6); // hcedbgfa
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // hcdebgfa
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); // hecdbfga
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); // hcedbfga
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); // hcdebfga
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v1, v4); swap(v5, v6); swap(v6, v7); // hecdbgaf
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v5, v6); swap(v6, v7); // hcedbgaf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hcdebgaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); // chedbgfa
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // chdebgfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); // cdhebgfa
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chedbgaf
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chdebgaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cdhebgaf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // chefdbga
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); // chedfbga
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // chdefbga
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); // cdhefbga
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v7); // chedbfga
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); // chdebfga
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); // cdhebfga
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v5, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hefcdbag
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hecfdbag
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v4, v5); swap(v6, v7); // hecdfbag
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v6, v7); // hecdbfag
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hcefdbag
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v6, v7); // hcedfbag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hcdefbag
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v6, v7); // hcedbfag
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v6, v7); // hcdebfag
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v1, v4); swap(v5, v7); // hecdbagf
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v5, v7); // hcedbagf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); // hcdebagf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v1, v4); swap(v5, v7); swap(v6, v7); // hecdbafg
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v5, v7); swap(v6, v7); // hcedbafg
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hcdebafg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chefdbag
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chedfbag
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chdefbag
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cdhefbag
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v6, v7); // chedbfag
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // chdebfag
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v6, v7); // cdhebfag
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); // chedbagf
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // chdebagf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); // cdhebagf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chedbafg
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chdebafg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cdhebafg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v6, v3)) {
															if (compare(v2, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // hefcbgda
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v5, v6); // hecfbgda
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // hcefbgda
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // chefbgda
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v6); // hecbgfda
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // hcebgfda
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // chebgfda
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hecbfgda
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hcebfgda
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chebfgda
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // hebgfcda
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); // hebgcfda
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // hefbgcda
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hebfgcda
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hefbcgda
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // hebfcgda
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // hebcgfda
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hebcfgda
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); // hefcbdga
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v3, v5); // hecfbdga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // hefbcdga
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); // hcefbdga
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); // chefbdga
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v5); // hecbfdga
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // hcebfdga
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // chebfdga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); // hebfcdga
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // hebcfdga
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v6, v7); // hefcbdag
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v6, v7); // hecfbdag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hefbcdag
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v6, v7); // hcefbdag
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // chefbdag
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hecbfdag
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hcebfdag
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chebfdag
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v6, v7); // hebfcdag
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hebcfdag
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hecbgdfa
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v5, v6); // hecbdgfa
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); // hecbdfga
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v6, v7); // hecbdfag
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // hebgcdfa
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hebcgdfa
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // hebcdgfa
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); // hebcdfga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v6, v7); // hebcdfag
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hcebgdfa
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // hcebdgfa
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); // hcebdfga
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v6, v7); // hcebdfag
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chebgdfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // chebdgfa
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); // chebdfga
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // chebdfag
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hecbgdaf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hecbdgaf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hebgcdaf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hebcgdaf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hebcdgaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcebgdaf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hcebdgaf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // chebgdaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chebdgaf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v5, v7); // hecbdagf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hecbdafg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); // hebcdagf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hebcdafg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); // hcebdagf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hcebdafg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // chebdagf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chebdafg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v0)) {
												if (compare(v6, v3)) {
													if (compare(v5, v3)) {
														if (compare(v7, v1)) {
															if (compare(v6, v4)) {
																if (compare(v6, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); // hgcbefda
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); // hcgbefda
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v6, v7); // chgbefda
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hgbefcda
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // hgbecfda
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); // hgbcefda
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hbgefcda
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); // hbgecfda
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v6); // hbgcefda
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v6); // hcbgefda
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); // chbgefda
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v6); // hbcgefda
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); // hbegfcda
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hbefgcda
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); // hbefcgda
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v6); // hbegcfda
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); // hbecgfda
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hbecfgda
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v6); // hcbegfda
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); // chbegfda
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v4); swap(v4, v6); // hbcegfda
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hcbefgda
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // chbefgda
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hbcefgda
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bhgefcda
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); // bhegfcda
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bhefgcda
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); // bhefcgda
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); // bhgecfda
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); // bhegcfda
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); // bhecgfda
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bhecfgda
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); // bhgcefda
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); // bhcgefda
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v6); // bhcegfda
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bhcefgda
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); // cbhgefda
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); // cbhegfda
																			}
																			else {
																				swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cbhefgda
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); // bchgefda
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); // bchegfda
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bchefgda
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v6, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		swap(v0, v7); swap(v1, v6); swap(v4, v6); swap(v5, v6); // hgcdbefa
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // hcgdbefa
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chgdbefa
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v5, v6); // hgcbedfa
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // hcgbedfa
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // chgbedfa
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hgcbdefa
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hcgbdefa
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chgbdefa
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v5, v6); // hcbgedfa
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hcbegdfa
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hcbgdefa
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); // chbgedfa
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // chbegdfa
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // chbgdefa
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); // cbhgedfa
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cbhegdfa
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cbhgdefa
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hgbecdfa
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hbgecdfa
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); // hbegcdfa
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hbecgdfa
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // hgbcedfa
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hgbcdefa
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); // hbgcedfa
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hbgcdefa
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v3, v6); swap(v5, v6); // hbcgedfa
																			}
																			else {
																				swap(v0, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hbcegdfa
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hbcgdefa
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bhgecdfa
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); // bhegcdfa
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); // bhgcedfa
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bhgcdefa
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v4)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v5, v6); // bhcgedfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bhcgdefa
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bhecgdfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bhcegdfa
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); // bchgedfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bchegdfa
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bchgdefa
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v4, v3)) {
															if (compare(v2, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v5); // hcbefdga
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v5, v6); // hcbedgfa
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v4); // hcbedfga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hcbefdag
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v6, v7); // hcbedfag
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); // chbefdga
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); // chbedgfa
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); // chbedfga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // chbefdag
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v6, v7); // chbedfag
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v5); // hbefcdga
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); // hbecfdga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v4); swap(v4, v5); // hbcefdga
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v6, v7); // hbefcdag
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hbecfdag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hbcefdag
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); // hbecdgfa
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v4); // hbecdfga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v6, v7); // hbecdfag
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v3, v4); swap(v5, v6); // hbcedgfa
																			}
																			else {
																				swap(v0, v7); swap(v3, v4); // hbcedfga
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v4); swap(v6, v7); // hbcedfag
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hcdgbefa
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // hcdbgefa
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v5, v6); // hcdbegfa
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v2); swap(v4, v6); swap(v5, v6); // hcbdgefa
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v5, v6); // hcbdegfa
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v4, v6); swap(v5, v6); // hbcdgefa
																			}
																			else {
																				swap(v0, v7); swap(v5, v6); // hbcdegfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); // hcdbefga
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); // hcbdefga
																			}
																		}
																		else {
																			swap(v0, v7); // hbcdefga
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v6, v7); // hcdbefag
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v6, v7); // hcbdefag
																			}
																		}
																		else {
																			swap(v0, v7); swap(v6, v7); // hbcdefag
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chdgbefa
																			}
																			else {
																				if (compare(v6, v4)) {
																					swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // chdbgefa
																				}
																				else {
																					swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); // chdbegfa
																				}
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); // chbdgefa
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v5, v6); // chbdegfa
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cdhgbefa
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); // cdhbgefa
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v5, v6); // cdhbegfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); // chdbefga
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); // cdhbefga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); // chbdefga
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v6, v7); // chdbefag
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v6, v7); // cdhbefag
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v6, v7); // chbdefag
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v7, v2)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); // bhefcdga
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); // bhecfdga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v5); // bhcefdga
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v6, v7); // bhefcdag
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bhecfdag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bhcefdag
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); // bhecdgfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); // bhecdfga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v6, v7); // bhecdfag
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v5, v6); // bhcedgfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v4); // bhcedfga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v6, v7); // bhcedfag
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); // cbhefdga
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); // cbhedgfa
																			}
																			else {
																				swap(v0, v2); swap(v2, v7); swap(v3, v4); // cbhedfga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cbhefdag
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v6, v7); // cbhedfag
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); // bchefdga
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); // bchedgfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); // bchedfga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bchefdag
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v6, v7); // bchedfag
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v4)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cdbhgefa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); // cbhdgefa
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cbdhgefa
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); // cdbhegfa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v5, v6); // cbhdegfa
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); // cbdhegfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); // cdbhefga
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v7); // cbhdefga
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); // cbdhefga
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v6, v7); // cdbhefag
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v6, v7); // cbhdefag
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v6, v7); // cbdhefag
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v6, v4)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v4, v6); swap(v5, v6); // bhcdgefa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); // bchdgefa
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bcdhgefa
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v5, v6); // bhcdegfa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v5, v6); // bchdegfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); // bcdhegfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v7); // bhcdefga
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); // bchdefga
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); // bcdhefga
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v6, v7); // bhcdefag
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v6, v7); // bchdefag
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v6, v7); // bcdhefag
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v4)) {
													if (compare(v4, v0)) {
														if (compare(v2, v1)) {
															if (compare(v6, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgcdbeaf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcgdbeaf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcdgbeaf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); // chgdbeaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // chdgbeaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdhgbeaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hgcbedaf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hcgbedaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); // chgbedaf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgcbdeaf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcgbdeaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // chgbdeaf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hcbgedaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // chbgedaf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cbhgedaf
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcbgdeaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chbgdeaf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbhgdeaf
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcdbgeaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chdbgeaf
																			}
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdhbgeaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdbhgeaf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcbdgeaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chbdgeaf
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbhdgeaf
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbdhgeaf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgbecdaf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hgbcedaf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbgecdaf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hbgcedaf
																			}
																			else {
																				swap(v0, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hbcgedaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgbcdeaf
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbgcdeaf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbcgdeaf
																		}
																		else {
																			swap(v0, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbcdgeaf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhgecdaf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bhgcedaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhgcdeaf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bhcgedaf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhcgdeaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhcdgeaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bchgedaf
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bchgdeaf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bchdgeaf
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcdhgeaf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v3, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v4, v6); swap(v5, v7); // hgcdbaef
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); // hcgdbaef
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hcdgbaef
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chgdbaef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chdgbaef
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cdhgbaef
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); // hcdbgaef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // chdbgaef
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); // cdhbgaef
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcdbagef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chdbagef
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdhbagef
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hgcbdaef
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hgbcdaef
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hcgbdaef
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chgbdaef
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hcbgdaef
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // chbgdaef
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hbgcdaef
																			}
																			else {
																				swap(v0, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hbcgdaef
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v4, v6); swap(v5, v7); // hcbdgaef
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); // chbdgaef
																			}
																		}
																		else {
																			swap(v0, v7); swap(v4, v6); swap(v5, v7); // hbcdgaef
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcbdagef
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chbdagef
																			}
																		}
																		else {
																			swap(v0, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbcdagef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cbhgdaef
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); // cbhdgaef
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cdbhgaef
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cbdhgaef
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bhgcdaef
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bhcgdaef
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v4, v6); swap(v5, v7); // bhcdgaef
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bchgdaef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); // bchdgaef
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bcdhgaef
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdbhagef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbhdagef
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbdhagef
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdbahgef
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbdahgef
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhcdagef
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bchdagef
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcdhagef
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcdahgef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v7, v2)) {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v5, v6); swap(v6, v7); // hcdbegaf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v5, v7); // hcdbeagf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v5, v7); swap(v6, v7); // hcdbeafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); // hcdbaegf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcdbaefg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcbegdaf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hcbedgaf
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v5, v7); // hcbedagf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hcbedafg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v0)) {
																				swap(v0, v7); swap(v1, v2); swap(v5, v6); swap(v6, v7); // hcbdegaf
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v7); swap(v1, v2); swap(v5, v7); // hcbdeagf
																				}
																				else {
																					swap(v0, v7); swap(v1, v2); swap(v5, v7); swap(v6, v7); // hcbdeafg
																				}
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v4, v7); swap(v5, v7); // hcbdaegf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcbdaefg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hbegcdaf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbecgdaf
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hbecdgaf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbcegdaf
																			}
																			else {
																				swap(v0, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hbcedgaf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v5, v6); swap(v6, v7); // hbcdegaf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); // hbecdagf
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hbecdafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v3, v4); swap(v5, v7); // hbcedagf
																			}
																			else {
																				swap(v0, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hbcedafg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v5, v7); // hbcdeagf
																			}
																			else {
																				swap(v0, v7); swap(v5, v7); swap(v6, v7); // hbcdeafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v4, v7); swap(v5, v7); // hbcdaegf
																			}
																			else {
																				swap(v0, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbcdaefg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // chdbegaf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); // chdbeagf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // chdbeafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // chdbaegf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chdbaefg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v5, v6); swap(v6, v7); // cdhbegaf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v5, v7); // cdhbeagf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v5, v7); swap(v6, v7); // cdhbeafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); // cdhbaegf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdhbaefg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chbegdaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // chbedgaf
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); // chbedagf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // chbedafg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v5, v6); swap(v6, v7); // chbdegaf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v5, v7); // chbdeagf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v5, v7); swap(v6, v7); // chbdeafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); // chbdaegf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chbdaefg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v4, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbhegdaf
																			}
																			else {
																				swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cbhedgaf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v5, v6); swap(v6, v7); // cbhdegaf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); // cbhedagf
																			}
																			else {
																				swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cbhedafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v2, v7); swap(v5, v7); // cbhdeagf
																			}
																			else {
																				swap(v0, v2); swap(v2, v7); swap(v5, v7); swap(v6, v7); // cbhdeafg
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cdbhegaf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); // cdbheagf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cdbheafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cbdhegaf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); // cbdheagf
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cbdheafg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cdbhaegf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); // cbhdaegf
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cbdhaegf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); // cdbahegf
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); // cbdahegf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdbhaefg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbhdaefg
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbdhaefg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdbahefg
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbdahefg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bhegcdaf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhecgdaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bhecdgaf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhcegdaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bhcedgaf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v5, v6); swap(v6, v7); // bhcdegaf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); // bhecdagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bhecdafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v5, v7); // bhcedagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bhcedafg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v5, v7); // bhcdeagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v5, v7); swap(v6, v7); // bhcdeafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v4, v7); swap(v5, v7); // bhcdaegf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhcdaefg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bchegdaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bchedgaf
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); // bchedagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bchedafg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v0)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v5, v6); swap(v6, v7); // bchdegaf
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v5, v7); // bchdeagf
																				}
																				else {
																					swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v5, v7); swap(v6, v7); // bchdeafg
																				}
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); // bchdaegf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bchdaefg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bcdhegaf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); // bcdheagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bcdheafg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bcdhaegf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcdhaefg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); // bcdahegf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcdahefg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v7, v1)) {
											if (compare(v5, v1)) {
												if (compare(v6, v1)) {
													if (compare(v7, v2)) {
														if (compare(v6, v2)) {
															if (compare(v6, v5)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ehgfcdba
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehgcfdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ehgcdfba
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehgfcbda
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ehgcfbda
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ehgfbcda
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehfgcdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ehfgcbda
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); // ehfgbcda
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // efhgcdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); // efhgcbda
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // efhgbcda
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ehfcgdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehfcdgba
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehfcgbda
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efhcgdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // efhcdgba
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // efhcgbda
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehcgfdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ehcgfbda
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ehcgdfba
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v4, v6); swap(v6, v7); // ehcdgfba
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ehcfgdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehcfgbda
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehcfdgba
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehcdfgba
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // echgfdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // echgdfba
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efchgdba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // echfgdba
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecfhgdba
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // echgfbda
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // echfgbda
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // efchgbda
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ecfhgbda
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // efchdgba
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v4, v7); swap(v5, v6); // efcdhgba
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // echfdgba
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ecfhdgba
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); // ecfdhgba
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v6, v7); // echdgfba
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // echdfgba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ecdhgfba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecdhfgba
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ecdfhgba
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cehgfdba
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); // cehfgdba
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cehfdgba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); // cehgfbda
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cehfgbda
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cefhgdba
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cefhgbda
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cefhdgba
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cefdhgba
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); // cehgdfba
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v6); // cehdgfba
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); // cedhgfba
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cdehgfba
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); // cehdfgba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cedhfgba
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cedfhgba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdehfgba
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cdefhgba
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v3, v1)) {
															if (compare(v6, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ehfcdbga
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ehcfdbga
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v4, v5); swap(v5, v7); // ehcdfbga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // echfdbga
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); // echdfbga
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ecdhfbga
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); // cehfdbga
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v5); // cehdfbga
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); // cedhfbga
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cdehfbga
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehfcdbag
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehcfdbag
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehcdfbag
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // echfdbag
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // echdfbag
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecdhfbag
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cehfdbag
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v6, v7); // cehdfbag
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cedhfbag
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdehfbag
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ehfcbgda
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); // ehfcbdga
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ehfcbdag
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ehfbgcda
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehfbcgda
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ehfbcdga
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehfbcdag
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ehcfbgda
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // echfbgda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); // cehfbgda
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v7); // ehcfbdga
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); // echfbdga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); // cehfbdga
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ehcfbdag
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // echfbdag
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v6, v7); // cehfbdag
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v6, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); // efhcdbga
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v7); // efchdbga
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v4, v7); // efcdhbga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); // ecfhdbga
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v7); // ecfdhbga
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); // ecdfhbga
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cefhdbga
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); // cefdhbga
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cedfhbga
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); // cdefhbga
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // efhcdbag
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // efchdbag
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v4, v7); swap(v6, v7); // efcdhbag
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ecfhdbag
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v6, v7); // ecfdhbag
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ecdfhbag
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cefhdbag
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cefdhbag
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cedfhbag
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cdefhbag
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efhcbgda
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // efhcbdga
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efhcbdag
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efhbgcda
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // efhbcgda
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); // efhbcdga
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // efhbcdag
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efchbgda
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecfhbgda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cefhbgda
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // efchbdga
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ecfhbdga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); // cefhbdga
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efchbdag
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecfhbdag
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cefhbdag
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v6, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehgcdbfa
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehcgdbfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehcdgbfa
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // echgdbfa
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // echdgbfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecdhgbfa
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v4, v7); swap(v5, v6); // ehcdbgfa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); // echdbgfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ecdhbgfa
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v4, v7); // ehcdbfga
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v7); // echdbfga
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); // ecdhbfga
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		swap(v0, v4); swap(v1, v7); swap(v4, v7); swap(v6, v7); // ehcdbfag
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v6, v7); // echdbfag
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ecdhbfag
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cehgdbfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); // cehdgbfa
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cedhgbfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdehgbfa
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v5, v6); // cehdbgfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); // cedhbgfa
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cdehbgfa
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); // cehdbfga
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); // cedhbfga
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); // cdehbfga
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v6, v7); // cehdbfag
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v6, v7); // cedhbfag
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cdehbfag
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ehgcdbaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ehcgdbaf
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v4, v6); swap(v5, v7); // ehcdgbaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehcdbgaf
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // echgdbaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); // echdgbaf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ecdhgbaf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // echdbgaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecdhbgaf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v2)) {
																		swap(v0, v4); swap(v1, v7); swap(v4, v7); swap(v5, v7); // ehcdbagf
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); // echdbagf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ecdhbagf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		swap(v0, v4); swap(v1, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehcdbafg
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // echdbafg
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecdhbafg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cehgdbaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cehdgbaf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cedhgbaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cdehgbaf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v5, v6); swap(v6, v7); // cehdbgaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cedhbgaf
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdehbgaf
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v5, v7); // cehdbagf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); // cedhbagf
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cdehbagf
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v5, v7); swap(v6, v7); // cehdbafg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cedhbafg
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdehbafg
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v3)) {
														if (compare(v5, v3)) {
															if (compare(v2, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); // ehgcbfda
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v7); // ehcgbfda
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); // echgbfda
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); // cehgbfda
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ehcbgfda
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // echbgfda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); // cehbgfda
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehcbfgda
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // echbfgda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cehbfgda
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehgbfcda
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ehbgfcda
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehbfgcda
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ehbfcgda
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ehgbcfda
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); // ehbgcfda
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ehbcgfda
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehbcfgda
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ehgcbdfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehgbcdfa
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ehcgbdfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // echgbdfa
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); // cehgbdfa
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehcbgdfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // echbgdfa
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cehbgdfa
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ehbgcdfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehbcgdfa
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehgcbdaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ehgbcdaf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehcgbdaf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // echgbdaf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cehgbdaf
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ehcbgdaf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // echbgdaf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cehbgdaf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehbgcdaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ehbcgdaf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ehcbdgfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehcbdgaf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ehcbfdga
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v7); // ehcbdfga
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); // ehbfcdga
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ehbcfdga
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ehbcdgfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); // ehbcdfga
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehbcdgaf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // echbdgfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // echbdgaf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // echbfdga
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); // echbdfga
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); // cehbdgfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cehbdgaf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); // cehbfdga
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); // cehbdfga
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehcbfdag
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ehcbdfag
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ehbfcdag
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehbcfdag
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ehbcdfag
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // echbfdag
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // echbdfag
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cehbfdag
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v6, v7); // cehbdfag
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ehcbdagf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehcbdafg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ehbcdagf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehbcdafg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // echbdagf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // echbdafg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); // cehbdagf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cehbdafg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v0)) {
												if (compare(v5, v3)) {
													if (compare(v6, v3)) {
														if (compare(v4, v1)) {
															if (compare(v7, v5)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ebhgfcda
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebhfgcda
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ebhfcgda
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); // ebhgcfda
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ebhcgfda
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebhcfgda
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ecbhgfda
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cebhgfda
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ebchgfda
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecbhfgda
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cebhfgda
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebchfgda
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // efcbhgda
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ecfbhgda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cefbhgda
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // efbhgcda
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efbhcgda
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // efbchgda
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ebfhgcda
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebfhcgda
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ebfchgda
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ecbfhgda
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cebfhgda
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ebcfhgda
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // behgfcda
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); // behgcfda
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // behfgcda
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // befhgcda
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // behfcgda
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // befhcgda
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // behcgfda
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // behcfgda
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bechgfda
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bechfgda
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // befchgda
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // becfhgda
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cbehgfda
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbehfgda
																			}
																			else {
																				swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cbefhgda
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bcehgfda
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcehfgda
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bcefhgda
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v5, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v4, v5); swap(v5, v7); // efcdbhga
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); // ecfdbhga
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v5, v7); // cefdbhga
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v7); // efcbhdga
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); // ecfbhdga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cefbhdga
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // efcbdhga
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ecfbdhga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); // cefbdhga
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ecbhfdga
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v7); // ecbfhdga
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ecbfdhga
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cebhfdga
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); // cebfhdga
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cebfdhga
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cbehfdga
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); // cbefhdga
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cbefdhga
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // efbhcdga
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); // efbchdga
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // efbcdhga
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); // ebhfcdga
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ebfhcdga
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); // ebfchdga
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ebfcdhga
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ebhcfdga
																		}
																		else {
																			swap(v0, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ebchfdga
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v3, v5); swap(v4, v7); // ebcfhdga
																		}
																		else {
																			swap(v0, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ebcfdhga
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); // behfcdga
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // befhcdga
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); // befchdga
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // befcdhga
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // behcfdga
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bechfdga
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v7); // becfhdga
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // becfdhga
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bcehfdga
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); // bcefhdga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bcefdhga
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v7, v3)) {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecbhgdfa
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ecbhdgfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v7); // ecbhdfga
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ecbhgdaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecbhdgaf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cebhgdfa
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cebhdgfa
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); // cebhdfga
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cebhgdaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cebhdgaf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ebhgcdfa
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebhcgdfa
																			}
																			else {
																				swap(v0, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebchgdfa
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebhgcdaf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ebhcgdaf
																			}
																			else {
																				swap(v0, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ebchgdaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ebhcdgfa
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); // ebhcdfga
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebhcdgaf
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ebchdgfa
																			}
																			else {
																				swap(v0, v4); swap(v3, v7); swap(v4, v7); // ebchdfga
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebchdgaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); // ecdbhgfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v4, v7); swap(v5, v6); // ecbdhgfa
																			}
																		}
																		else {
																			swap(v0, v4); swap(v4, v7); swap(v5, v6); // ebcdhgfa
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecdbhgaf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecbdhgaf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebcdhgaf
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ecdfbhga
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v7); // ecdbhfga
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); // ecdbfhga
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v4, v7); // ecbdhfga
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v4, v5); swap(v5, v7); // ecbdfhga
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v4, v7); // ebcdhfga
																			}
																			else {
																				swap(v0, v4); swap(v4, v5); swap(v5, v7); // ebcdfhga
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cedbhgfa
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); // cdebhgfa
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); // cebdhgfa
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cedbhgaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdebhgaf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cebdhgaf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v1)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); // cedfbhga
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); // cedbhfga
																				}
																				else {
																					swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cedbfhga
																				}
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v7); // cebdhfga
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); // cebdfhga
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cdefbhga
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v7); // cdebhfga
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); // cdebfhga
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v4, v2)) {
																if (compare(v6, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // behgcdfa
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // behcgdfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bechgdfa
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // behgcdaf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // behcgdaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bechgdaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // behcdgfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); // behcdfga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // behcdgaf
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bechdgfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v7); // bechdfga
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bechdgaf
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbehgdfa
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cbehdgfa
																			}
																			else {
																				swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); // cbehdfga
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cbehgdaf
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbehdgaf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcehgdfa
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bcehdgfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); // bcehdfga
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bcehgdaf
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcehdgaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cdbehgfa
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); // cbedhgfa
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cbdehgfa
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdbehgaf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbedhgaf
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbdehgaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); // cdbehfga
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v4, v7); // cbedhfga
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); // cbdehfga
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cdbefhga
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); // cbedfhga
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cbdefhga
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v4, v7); swap(v5, v6); // becdhgfa
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); // bcedhgfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bcdehgfa
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // becdhgaf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcedhgaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcdehgaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v4, v7); // becdhfga
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v7); // bcedhfga
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); // bcdehfga
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v4, v5); swap(v5, v7); // becdfhga
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); // bcedfhga
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bcdefhga
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v5)) {
													if (compare(v4, v1)) {
														if (compare(v4, v2)) {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v6, v7); // ecdbhfag
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); // ecdbhagf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecdbhafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v5, v7); // ecdbahgf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v5, v7); swap(v6, v7); // ecdbahfg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecbhfdag
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ecbhdfag
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ecbhdagf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecbhdafg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v4); swap(v1, v2); swap(v4, v7); swap(v6, v7); // ecbdhfag
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v4); swap(v1, v2); swap(v4, v7); swap(v5, v7); // ecbdhagf
																				}
																				else {
																					swap(v0, v4); swap(v1, v2); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecbdhafg
																				}
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v5, v7); // ecbdahgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v5, v7); swap(v6, v7); // ecbdahfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ebhfcdag
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebhcfdag
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ebhcdfag
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebchfdag
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ebchdfag
																			}
																			else {
																				swap(v0, v4); swap(v4, v7); swap(v6, v7); // ebcdhfag
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ebhcdagf
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebhcdafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ebchdagf
																			}
																			else {
																				swap(v0, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebchdafg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v4, v7); swap(v5, v7); // ebcdhagf
																			}
																			else {
																				swap(v0, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebcdhafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v5, v7); // ebcdahgf
																			}
																			else {
																				swap(v0, v4); swap(v5, v7); swap(v6, v7); // ebcdahfg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cedbhfag
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cedbhagf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cedbhafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); // cedbahgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cedbahfg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v6, v7); // cdebhfag
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); // cdebhagf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdebhafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v5, v7); // cdebahgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v5, v7); swap(v6, v7); // cdebahfg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cebhfdag
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cebhdfag
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cebhdagf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cebhdafg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v6, v7); // cebdhfag
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); // cebdhagf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cebdhafg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v5, v7); // cebdahgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v5, v7); swap(v6, v7); // cebdahfg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v7, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbehfdag
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cbehdfag
																			}
																			else {
																				swap(v0, v2); swap(v2, v4); swap(v4, v7); swap(v6, v7); // cbedhfag
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cbehdagf
																			}
																			else {
																				swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbehdafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); // cbedhagf
																			}
																			else {
																				swap(v0, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbedhafg
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cdbehfag
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cdbehagf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdbehafg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cbdehfag
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cbdehagf
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbdehafg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); // cdbeahgf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v5, v7); // cbedahgf
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); // cbdeahgf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v5, v7); // cdbaehgf
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v5, v7); // cbdaehgf
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cdbeahfg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v5, v7); swap(v6, v7); // cbedahfg
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cbdeahfg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v5, v7); swap(v6, v7); // cdbaehfg
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v5, v7); swap(v6, v7); // cbdaehfg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // behfcdag
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // behcfdag
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // behcdfag
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bechfdag
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bechdfag
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v4, v7); swap(v6, v7); // becdhfag
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // behcdagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // behcdafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bechdagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bechdafg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v4, v7); swap(v5, v7); // becdhagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // becdhafg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v5, v7); // becdahgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v5, v7); swap(v6, v7); // becdahfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcehfdag
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bcehdfag
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bcehdagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcehdafg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v6, v7); // bcedhfag
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); // bcedhagf
																				}
																				else {
																					swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcedhafg
																				}
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v5, v7); // bcedahgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v5, v7); swap(v6, v7); // bcedahfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bcdehfag
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bcdehagf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcdehafg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); // bcdeahgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bcdeahfg
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v5, v7); // bcdaehgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v5, v7); swap(v6, v7); // bcdaehfg
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v0)) {
														if (compare(v2, v1)) {
															if (compare(v5, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efcdbhag
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecfdbhag
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecdfbhag
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v5, v7); swap(v6, v7); // cefdbhag
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cedfbhag
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdefbhag
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // efcbhdag
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ecfbhdag
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cefbhdag
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efcbdhag
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecfbdhag
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cefbdhag
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ecbfhdag
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cebfhdag
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cbefhdag
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecbfdhag
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cebfdhag
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbefdhag
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecdbfhag
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cedbfhag
																			}
																		}
																		else {
																			if (compare(v4, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdebfhag
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdbefhag
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecbdfhag
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cebdfhag
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbedfhag
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbdefhag
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efbhcdag
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // efbchdag
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebfhcdag
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ebfchdag
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ebcfhdag
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efbcdhag
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebfcdhag
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebcfdhag
																		}
																		else {
																			swap(v0, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebcdfhag
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // befhcdag
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // befchdag
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // befcdhag
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // becfhdag
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // becfdhag
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // becdfhag
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bcefhdag
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcefdhag
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcedfhag
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcdefhag
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v3, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v4, v5); swap(v6, v7); // efcdbahg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v6, v7); // ecfdbahg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ecdfbahg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v6, v7); // cefdbahg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v6, v7); // cedfbahg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cdefbahg
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v6, v7); // ecdbfahg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cedbfahg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v6, v7); // cdebfahg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v6, v7); // ecdbafhg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v6, v7); // cedbafhg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v6, v7); // cdebafhg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // efcbdahg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // efbcdahg
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ecfbdahg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v6, v7); // cefbdahg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ecbfdahg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cebfdahg
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ebfcdahg
																			}
																			else {
																				swap(v0, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ebcfdahg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v4, v5); swap(v6, v7); // ecbdfahg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v6, v7); // cebdfahg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v4, v5); swap(v6, v7); // ebcdfahg
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v6, v7); // ecbdafhg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v6, v7); // cebdafhg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v6, v7); // ebcdafhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cbefdahg
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v4, v5); swap(v6, v7); // cbedfahg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cdbefahg
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cbdefahg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // befcdahg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // becfdahg
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v4, v5); swap(v6, v7); // becdfahg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bcefdahg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v6, v7); // bcedfahg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bcdefahg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v6, v7); // cdbeafhg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v6, v7); // cbedafhg
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v6, v7); // cbdeafhg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v6, v7); // cdbaefhg
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v6, v7); // cbdaefhg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v6, v7); // becdafhg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v6, v7); // bcedafhg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v6, v7); // bcdeafhg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v6, v7); // bcdaefhg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v4)) {
										if (compare(v4, v1)) {
											if (compare(v7, v1)) {
												if (compare(v5, v1)) {
													if (compare(v4, v2)) {
														if (compare(v5, v2)) {
															if (compare(v7, v5)) {
																if (compare(v7, v4)) {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v6, v7); // ghefcdba
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ghefcbda
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghefbcda
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gehfcdba
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gehfcbda
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gehfbcda
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gefhcdba
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gefchdba
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gefcdhba
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gefhcbda
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gefchbda
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v6, v7); // gefhbcda
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v7, v4)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ghecfdba
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v6, v7); // ghecdfba
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghecfbda
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gehcfdba
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gehcdfba
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gehcfbda
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gechfdba
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gechfbda
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gechdfba
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v4, v7); swap(v6, v7); // gecdhfba
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gecfhdba
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gecfhbda
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gecfdhba
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gecdfhba
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ghcefdba
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gchefdba
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcehfdba
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gcefhdba
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcefdhba
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghcefbda
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gchefbda
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gcehfbda
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcefhbda
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v6, v7); // ghcedfba
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v6, v7); // gchedfba
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gcehdfba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v6, v7); // gcedhfba
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcedfhba
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v6, v7); // ghcdefba
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v6, v7); // gchdefba
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v6, v7); // gcdhefba
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gcdehfba
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcdefhba
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v7, v4)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); // cghefdba
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cghefbda
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); // cghedfba
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); // cghdefba
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cgehfdba
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); // cgefhdba
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); // cgehdfba
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgehfbda
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cgefhbda
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v7); // cgedhfba
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); // cgdhefba
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); // cgdehfba
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v6, v7); // cdghefba
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cdgehfba
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cgefdhba
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); // cgedfhba
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cgdefhba
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdgefhba
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v4)) {
														if (compare(v3, v1)) {
															if (compare(v5, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ghecdbfa
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ghcedbfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v5, v7); swap(v6, v7); // ghcdebfa
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gchedbfa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v5, v7); swap(v6, v7); // gchdebfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gcdhebfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); // cghedbfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v5, v6); // cghdebfa
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); // cgdhebfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cdghebfa
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); // ghecdbaf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v5, v7); // ghcedbaf
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v5, v7); // ghcdebaf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); // gchedbaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v5, v7); // gchdebaf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); // gcdhebaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cghedbaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v5, v6); swap(v6, v7); // cghdebaf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cgdhebaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); // cdghebaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ghecbfda
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghecbdfa
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ghecbdaf
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ghebfcda
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v6, v7); // ghebcfda
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ghebcdfa
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); // ghebcdaf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ghcebfda
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gchebfda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); // cghebfda
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghcebdfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gchebdfa
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cghebdfa
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ghcebdaf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gchebdaf
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cghebdaf
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v5, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gehcdbfa
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gechdbfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gecdhbfa
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcehdbfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcedhbfa
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcdehbfa
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cgehdbfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); // cgedhbfa
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cgdehbfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdgehbfa
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gehcdbaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gechdbaf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v4, v7); swap(v5, v7); // gecdhbaf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gcehdbaf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); // gcedhbaf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gcdehbaf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgehdbaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgedhbaf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgdehbaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cdgehbaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v6, v7); // gehcbfda
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gehcbdfa
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); // gehcbdaf
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gehbfcda
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gehbcfda
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gehbcdfa
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gehbcdaf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v6, v7); // gechbfda
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v6, v7); // gcehbfda
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cgehbfda
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gechbdfa
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gcehbdfa
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgehbdfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v5, v7); // gechbdaf
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); // gcehbdaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cgehbdaf
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v7, v0)) {
														if (compare(v6, v2)) {
															if (compare(v5, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gefcdbha
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gecfdbha
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v4, v5); swap(v6, v7); // gecdfbha
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gcefdbha
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v6, v7); // gcedfbha
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gcdefbha
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v5, v7); swap(v6, v7); // gecdbhfa
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v5, v7); swap(v6, v7); // gcedbhfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gcdebhfa
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v6, v7); // gecdbfha
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v6, v7); // gcedbfha
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v6, v7); // gcdebfha
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v1, v4); swap(v5, v7); // gecdbhaf
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v5, v7); // gcedbhaf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); // gcdebhaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgedbhfa
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgdebhfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cdgebhfa
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); // cgedbhaf
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cgdebhaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); // cdgebhaf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgefdbha
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgedfbha
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgdefbha
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cdgefbha
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v6, v7); // cgedbfha
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cgdebfha
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v6, v7); // cdgebfha
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v5, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // gefcdbah
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v5); // gecfdbah
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v4, v5); // gecdfbah
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); // gecdbfah
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // gcefdbah
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v5); // gcedfbah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // gcdefbah
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); // gcedbfah
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); // gcdebfah
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v1, v4); swap(v5, v6); swap(v6, v7); // gecdbahf
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v5, v6); swap(v6, v7); // gcedbahf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gcdebahf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v1, v4); swap(v5, v6); // gecdbafh
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v5, v6); // gcedbafh
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // gcdebafh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cgefdbah
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // cgedfbah
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cgdefbah
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // cdgefbah
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v6); // cgedbfah
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // cgdebfah
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); // cdgebfah
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgedbahf
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgdebahf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cdgebahf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // cgedbafh
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cgdebafh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // cdgebafh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v7, v3)) {
															if (compare(v2, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gefcbhda
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gecfbhda
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gcefbhda
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgefbhda
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gecbhfda
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gcebhfda
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); // cgebhfda
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gecbfhda
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcebfhda
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cgebfhda
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gebhfcda
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v6, v7); // gebhcfda
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gefbhcda
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gebfhcda
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gefbchda
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gebfchda
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gebchfda
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gebcfhda
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v6, v7); // gefcbdha
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v6, v7); // gecfbdha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gefbcdha
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v6, v7); // gcefbdha
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cgefbdha
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gecbfdha
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gcebfdha
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgebfdha
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v6, v7); // gebfcdha
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gebcfdha
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); // gefcbdah
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v3, v5); // gecfbdah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // gefbcdah
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); // gcefbdah
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // cgefbdah
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v5); // gecbfdah
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // gcebfdah
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cgebfdah
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); // gebfcdah
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // gebcfdah
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gecbhdfa
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gecbdhfa
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v6, v7); // gecbdfha
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); // gecbdfah
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gebhcdfa
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gebchdfa
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gebcdhfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v6, v7); // gebcdfha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); // gebcdfah
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcebhdfa
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gcebdhfa
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v6, v7); // gcebdfha
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); // gcebdfah
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cgebhdfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgebdhfa
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cgebdfha
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // cgebdfah
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gecbhdaf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v5, v7); // gecbdhaf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); // gebhcdaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gebchdaf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); // gebcdhaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gcebhdaf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); // gcebdhaf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgebhdaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cgebdhaf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gecbdahf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v5, v6); // gecbdafh
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gebcdahf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // gebcdafh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gcebdahf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // gcebdafh
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgebdahf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cgebdafh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v0)) {
												if (compare(v7, v3)) {
													if (compare(v5, v3)) {
														if (compare(v6, v1)) {
															if (compare(v7, v4)) {
																if (compare(v7, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v6, v7); // ghcbefda
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v6, v7); // gchbefda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); // cghbefda
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghbefcda
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ghbecfda
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v6, v7); // ghbcefda
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbhefcda
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gbhecfda
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v6, v7); // gbhcefda
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v6, v7); // gcbhefda
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v6, v7); // cgbhefda
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v7); swap(v6, v7); // gbchefda
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gbehfcda
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbefhcda
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gbefchda
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v6, v7); // gbehcfda
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gbechfda
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbecfhda
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gcbehfda
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cgbehfda
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gbcehfda
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcbefhda
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgbefhda
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbcefhda
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bghefcda
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bgehfcda
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgefhcda
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bgefchda
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bghecfda
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v6, v7); // bgehcfda
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bgechfda
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgecfhda
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v6, v7); // bghcefda
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v6, v7); // bgchefda
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bgcehfda
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgcefhda
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v6, v7); // cbghefda
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cbgehfda
																			}
																			else {
																				swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbgefhda
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v6, v7); // bcghefda
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bcgehfda
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcgefhda
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v7, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghcdbefa
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gchdbefa
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); // cghdbefa
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ghcbedfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gchbedfa
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); // cghbedfa
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghcbdefa
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gchbdefa
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cghbdefa
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gcbhedfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcbehdfa
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcbhdefa
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cgbhedfa
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgbehdfa
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgbhdefa
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cbghedfa
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbgehdfa
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbghdefa
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghbecdfa
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbhecdfa
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gbehcdfa
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbechdfa
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ghbcedfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghbcdefa
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gbhcedfa
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbhcdefa
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gbchedfa
																			}
																			else {
																				swap(v0, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbcehdfa
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbchdefa
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bghecdfa
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bgehcdfa
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bghcedfa
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bghcdefa
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v4)) {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bgchedfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgchdefa
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgechdfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgcehdfa
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bcghedfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcgehdfa
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcghdefa
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v4, v3)) {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gcbefdha
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gcbedhfa
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v6, v7); // gcbedfha
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v5); // gcbefdah
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v4); // gcbedfah
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cgbefdha
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cgbedhfa
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v6, v7); // cgbedfha
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // cgbefdah
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); // cgbedfah
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v6, v7); // gbefcdha
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gbecfdha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gbcefdha
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v5); // gbefcdah
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // gbecfdah
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v4); swap(v4, v5); // gbcefdah
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gbecdhfa
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v6, v7); // gbecdfha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v4); // gbecdfah
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gbcedhfa
																			}
																			else {
																				swap(v0, v6); swap(v3, v4); swap(v6, v7); // gbcedfha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v4); // gbcedfah
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcdhbefa
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcdbhefa
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v5, v7); swap(v6, v7); // gcdbehfa
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v2); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcbdhefa
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v5, v7); swap(v6, v7); // gcbdehfa
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbcdhefa
																			}
																			else {
																				swap(v0, v6); swap(v5, v7); swap(v6, v7); // gbcdehfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v6, v7); // gcdbefha
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v6, v7); // gcbdefha
																			}
																		}
																		else {
																			swap(v0, v6); swap(v6, v7); // gbcdefha
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); // gcdbefah
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); // gcbdefah
																			}
																		}
																		else {
																			swap(v0, v6); // gbcdefah
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cgdhbefa
																			}
																			else {
																				if (compare(v7, v4)) {
																					swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgdbhefa
																				}
																				else {
																					swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cgdbehfa
																				}
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgbdhefa
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v5, v7); swap(v6, v7); // cgbdehfa
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdghbefa
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdgbhefa
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v5, v7); swap(v6, v7); // cdgbehfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v6, v7); // cgdbefha
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v6, v7); // cdgbefha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v6, v7); // cgbdefha
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); // cgdbefah
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); // cdgbefah
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); // cgbdefah
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v2)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v6, v7); // bgefcdha
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bgecfdha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bgcefdha
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); // bgefcdah
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // bgecfdah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v5); // bgcefdah
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bgecdhfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v6, v7); // bgecdfha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); // bgecdfah
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bgcedhfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v6, v7); // bgcedfha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v4); // bgcedfah
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cbgefdha
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cbgedhfa
																			}
																			else {
																				swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v6, v7); // cbgedfha
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // cbgefdah
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v4); // cbgedfah
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bcgefdha
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bcgedhfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v6, v7); // bcgedfha
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // bcgefdah
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); // bcgedfah
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v4)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdbghefa
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbgdhefa
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbdghefa
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cdbgehfa
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v5, v7); swap(v6, v7); // cbgdehfa
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cbdgehfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v6, v7); // cdbgefha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v6, v7); // cbgdefha
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v6, v7); // cbdgefha
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); // cdbgefah
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v6); // cbgdefah
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); // cbdgefah
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v7, v4)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgcdhefa
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcgdhefa
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcdghefa
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v5, v7); swap(v6, v7); // bgcdehfa
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v5, v7); swap(v6, v7); // bcgdehfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bcdgehfa
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v6, v7); // bgcdefha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v6, v7); // bcgdefha
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v6, v7); // bcdgefha
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v6); // bgcdefah
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); // bcgdefah
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); // bcdgefah
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v4)) {
													if (compare(v4, v0)) {
														if (compare(v2, v1)) {
															if (compare(v7, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v7); swap(v5, v7); // ghcdbeaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); // gchdbeaf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gcdhbeaf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cghdbeaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgdhbeaf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cdghbeaf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v5, v7); // ghcbedaf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); // gchbedaf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cghbedaf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ghcbdeaf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gchbdeaf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cghbdeaf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v5, v7); // gcbhedaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); // cgbhedaf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); // cbghedaf
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gcbhdeaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cgbhdeaf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cbghdeaf
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); // gcdbheaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cgdbheaf
																			}
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); // cdgbheaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cdbgheaf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v4, v7); swap(v5, v7); // gcbdheaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); // cgbdheaf
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); // cbgdheaf
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cbdgheaf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ghbecdaf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); // ghbcedaf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gbhecdaf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); // gbhcedaf
																			}
																			else {
																				swap(v0, v6); swap(v3, v7); swap(v5, v7); // gbchedaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ghbcdeaf
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gbhcdeaf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gbchdeaf
																		}
																		else {
																			swap(v0, v6); swap(v4, v7); swap(v5, v7); // gbcdheaf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bghecdaf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); // bghcedaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bghcdeaf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v5, v7); // bgchedaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bgchdeaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v4, v7); swap(v5, v7); // bgcdheaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); // bcghedaf
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bcghdeaf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); // bcgdheaf
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bcdgheaf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v3, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghcdbaef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gchdbaef
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcdhbaef
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); // cghdbaef
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cgdhbaef
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdghbaef
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcdbhaef
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgdbhaef
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdgbhaef
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); // gcdbahef
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cgdbahef
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); // cdgbahef
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghcbdaef
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghbcdaef
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gchbdaef
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cghbdaef
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcbhdaef
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgbhdaef
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbhcdaef
																			}
																			else {
																				swap(v0, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbchdaef
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcbdhaef
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgbdhaef
																			}
																		}
																		else {
																			swap(v0, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbcdhaef
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v4, v6); swap(v5, v7); // gcbdahef
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); // cgbdahef
																			}
																		}
																		else {
																			swap(v0, v6); swap(v4, v6); swap(v5, v7); // gbcdahef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbghdaef
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbgdhaef
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdbghaef
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbdghaef
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bghcdaef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgchdaef
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgcdhaef
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcghdaef
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcgdhaef
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcdghaef
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cdbgahef
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); // cbgdahef
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cbdgahef
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); // cdbaghef
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); // cbdaghef
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v4, v6); swap(v5, v7); // bgcdahef
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); // bcgdahef
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bcdgahef
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); // bcdaghef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v6, v2)) {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v5, v7); // gcdbehaf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v5, v6); swap(v6, v7); // gcdbeahf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v5, v6); // gcdbeafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcdbaehf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // gcdbaefh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gcbehdaf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v5, v7); // gcbedhaf
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gcbedahf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v5, v6); // gcbedafh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v0)) {
																				swap(v0, v6); swap(v1, v2); swap(v5, v7); // gcbdehaf
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v6); swap(v1, v2); swap(v5, v6); swap(v6, v7); // gcbdeahf
																				}
																				else {
																					swap(v0, v6); swap(v1, v2); swap(v5, v6); // gcbdeafh
																				}
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcbdaehf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v4, v6); swap(v5, v6); // gcbdaefh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); // gbehcdaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gbechdaf
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); // gbecdhaf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gbcehdaf
																			}
																			else {
																				swap(v0, v6); swap(v3, v4); swap(v5, v7); // gbcedhaf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v5, v7); // gbcdehaf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gbecdahf
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // gbecdafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gbcedahf
																			}
																			else {
																				swap(v0, v6); swap(v3, v4); swap(v5, v6); // gbcedafh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v5, v6); swap(v6, v7); // gbcdeahf
																			}
																			else {
																				swap(v0, v6); swap(v5, v6); // gbcdeafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbcdaehf
																			}
																			else {
																				swap(v0, v6); swap(v4, v6); swap(v5, v6); // gbcdaefh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); // cgdbehaf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cgdbeahf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // cgdbeafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgdbaehf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cgdbaefh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v5, v7); // cdgbehaf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v5, v6); swap(v6, v7); // cdgbeahf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v5, v6); // cdgbeafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdgbaehf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // cdgbaefh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cgbehdaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); // cgbedhaf
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cgbedahf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // cgbedafh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v5, v7); // cgbdehaf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v5, v6); swap(v6, v7); // cgbdeahf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v5, v6); // cgbdeafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgbdaehf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // cgbdaefh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v4, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cbgehdaf
																			}
																			else {
																				swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); // cbgedhaf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v5, v7); // cbgdehaf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cbgedahf
																			}
																			else {
																				swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // cbgedafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v6); swap(v5, v6); swap(v6, v7); // cbgdeahf
																			}
																			else {
																				swap(v0, v2); swap(v2, v6); swap(v5, v6); // cbgdeafh
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); // cdbgehaf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cdbgeahf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // cdbgeafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); // cbdgehaf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cbdgeahf
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // cbdgeafh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdbgaehf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbgdaehf
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbdgaehf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdbagehf
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbdagehf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cdbgaefh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // cbgdaefh
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cbdgaefh
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // cdbagefh
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // cbdagefh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); // bgehcdaf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bgechdaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); // bgecdhaf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bgcehdaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v5, v7); // bgcedhaf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v5, v7); // bgcdehaf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bgecdahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // bgecdafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bgcedahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v5, v6); // bgcedafh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v5, v6); swap(v6, v7); // bgcdeahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v5, v6); // bgcdeafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgcdaehf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v4, v6); swap(v5, v6); // bgcdaefh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bcgehdaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); // bcgedhaf
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bcgedahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // bcgedafh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v5, v7); // bcgdehaf
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v5, v6); swap(v6, v7); // bcgdeahf
																				}
																				else {
																					swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v5, v6); // bcgdeafh
																				}
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcgdaehf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // bcgdaefh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); // bcdgehaf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bcdgeahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // bcdgeafh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcdgaehf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bcdgaefh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcdagehf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // bcdagefh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v1)) {
											if (compare(v5, v1)) {
												if (compare(v7, v1)) {
													if (compare(v6, v2)) {
														if (compare(v7, v2)) {
															if (compare(v7, v5)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); // eghfcdba
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // eghcfdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); // eghcdfba
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // eghfcbda
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eghcfbda
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eghfbcda
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // egfhcdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egfhcbda
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // egfhbcda
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efghcdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // efghcbda
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efghbcda
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); // egfchdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // egfcdhba
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // egfchbda
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // efgchdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efgcdhba
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); // efgchbda
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // egchfdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egchfbda
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v7); // egchdfba
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v4, v7); // egcdhfba
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v7); // egcfhdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // egcfhbda
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // egcfdhba
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v4, v5); swap(v5, v7); // egcdfhba
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v7, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ecghfdba
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); // ecgfhdba
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); // ecghdfba
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // efcghdba
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ecfghdba
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecghfbda
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ecgfhbda
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v7); // efcghbda
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); // ecfghbda
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efcgdhba
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efcdghba
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ecgfdhba
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecfgdhba
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecfdghba
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v7); // ecgdhfba
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); // ecgdfhba
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); // ecdghfba
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ecdgfhba
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecdfghba
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ceghfdba
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cegfhdba
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cegfdhba
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ceghfbda
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cegfhbda
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cefghdba
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cefghbda
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cefgdhba
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); // cefdghba
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ceghdfba
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v6, v7); // cegdhfba
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cedghfba
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); // cdeghfba
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cegdfhba
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cedgfhba
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cedfghba
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cdegfhba
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdefghba
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v3, v1)) {
															if (compare(v7, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egfcdbha
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egcfdbha
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egcdfbha
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecgfdbha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecgdfbha
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecdgfbha
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cegfdbha
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v6, v7); // cegdfbha
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cedgfbha
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdegfbha
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // egfcdbah
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // egcfdbah
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v4, v5); swap(v5, v6); // egcdfbah
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ecgfdbah
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // ecgdfbah
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ecdgfbah
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // cegfdbah
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v5); // cegdfbah
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // cedgfbah
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cdegfbah
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egfcbhda
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // egfcbdha
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // egfcbdah
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); // egfbhcda
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // egfbchda
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egfbcdha
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // egfbcdah
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egcfbhda
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecgfbhda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cegfbhda
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // egcfbdha
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ecgfbdha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v6, v7); // cegfbdha
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v6); // egcfbdah
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // ecgfbdah
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); // cegfbdah
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v7, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // efgcdbha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // efcgdbha
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v4, v6); swap(v6, v7); // efcdgbha
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ecfgdbha
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v6, v7); // ecfdgbha
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ecdfgbha
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cefgdbha
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cefdgbha
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cedfgbha
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cdefgbha
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // efgcdbah
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v6); // efcgdbah
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v4, v6); // efcdgbah
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // ecfgdbah
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v6); // ecfdgbah
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // ecdfgbah
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cefgdbah
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // cefdgbah
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cedfgbah
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // cdefgbah
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // efgcbhda
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efgcbdha
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // efgcbdah
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // efgbhcda
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efgbchda
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // efgbcdha
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // efgbcdah
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // efcgbhda
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ecfgbhda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); // cefgbhda
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efcgbdha
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecfgbdha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cefgbdha
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // efcgbdah
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ecfgbdah
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // cefgbdah
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v7, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // eghcdbfa
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // egchdbfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v4, v7); swap(v5, v6); // egcdhbfa
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ecghdbfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); // ecgdhbfa
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ecdghbfa
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egcdbhfa
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecgdbhfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecdgbhfa
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v4, v6); swap(v6, v7); // egcdbfha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v6, v7); // ecgdbfha
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ecdgbfha
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		swap(v0, v4); swap(v1, v6); swap(v4, v6); // egcdbfah
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v6); // ecgdbfah
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // ecdgbfah
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ceghdbfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cegdhbfa
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cedghbfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cdeghbfa
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v5, v7); swap(v6, v7); // cegdbhfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cedgbhfa
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdegbhfa
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v6, v7); // cegdbfha
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v6, v7); // cedgbfha
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cdegbfha
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); // cegdbfah
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); // cedgbfah
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // cdegbfah
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v7, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eghcdbaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egchdbaf
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egcdhbaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v4, v6); swap(v5, v7); // egcdbhaf
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecghdbaf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecgdhbaf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecdghbaf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); // ecgdbhaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ecdgbhaf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v2)) {
																		swap(v0, v4); swap(v1, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egcdbahf
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecgdbahf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecdgbahf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		swap(v0, v4); swap(v1, v6); swap(v4, v6); swap(v5, v6); // egcdbafh
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // ecgdbafh
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ecdgbafh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ceghdbaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); // cegdhbaf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cedghbaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdeghbaf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v5, v7); // cegdbhaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); // cedgbhaf
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cdegbhaf
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v5, v6); swap(v6, v7); // cegdbahf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cedgbahf
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdegbahf
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v5, v6); // cegdbafh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // cedgbafh
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cdegbafh
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v3)) {
														if (compare(v5, v3)) {
															if (compare(v2, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // eghcbfda
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // egchbfda
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ecghbfda
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v6, v7); // ceghbfda
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v7); // egcbhfda
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); // ecgbhfda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cegbhfda
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // egcbfhda
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ecgbfhda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cegbfhda
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // eghbfcda
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egbhfcda
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // egbfhcda
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egbfchda
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); // eghbcfda
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // egbhcfda
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); // egbchfda
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // egbcfhda
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eghcbdfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egchbdfa
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // eghbcdfa
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecghbdfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ceghbdfa
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // egcbhdfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ecgbhdfa
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cegbhdfa
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egbhcdfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // egbchdfa
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // eghcbdaf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // egchbdaf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eghbcdaf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ecghbdaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); // ceghbdaf
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egcbhdaf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecgbhdaf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cegbhdaf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // egbhcdaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egbchdaf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v0)) {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egcbdhfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // egcbdhaf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egcbfdha
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // egcbdfha
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // egbfcdha
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egbcfdha
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egbcdhfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // egbcdfha
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // egbcdhaf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecgbdhfa
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ecgbdhaf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecgbfdha
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ecgbdfha
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cegbdhfa
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); // cegbdhaf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cegbfdha
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v6, v7); // cegbdfha
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // egcbfdah
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v6); // egcbdfah
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // egbfcdah
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // egbcfdah
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // egbcdfah
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ecgbfdah
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // ecgbdfah
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // cegbfdah
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); // cegbdfah
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egcbdahf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // egcbdafh
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egbcdahf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // egbcdafh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecgbdahf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ecgbdafh
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cegbdahf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // cegbdafh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v0)) {
												if (compare(v5, v3)) {
													if (compare(v7, v3)) {
														if (compare(v4, v1)) {
															if (compare(v6, v5)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebghfcda
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ebgfhcda
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebgfchda
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ebghcfda
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); // ebgchfda
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ebgcfhda
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v7); // ecbghfda
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); // cebghfda
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v4, v7); // ebcghfda
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ecbgfhda
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cebgfhda
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ebcgfhda
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efcbghda
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecfbghda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cefbghda
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); // efbghcda
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // efbgchda
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efbcghda
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); // ebfghcda
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ebfgchda
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebfcghda
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecbfghda
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cebfghda
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebcfghda
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // beghfcda
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // beghcfda
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // begfhcda
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); // befghcda
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // begfchda
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // befgchda
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); // begchfda
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // begcfhda
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v7); // becghfda
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // becgfhda
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // befcghda
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // becfghda
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); // cbeghfda
																			}
																			else {
																				swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cbegfhda
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbefghda
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); // bceghfda
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bcegfhda
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcefghda
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v5, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efcdbgha
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecfdbgha
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v5, v6); swap(v6, v7); // cefdbgha
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // efcbgdha
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ecfbgdha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cefbgdha
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efcbdgha
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecfbdgha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cefbdgha
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecbgfdha
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ecbfgdha
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecbfdgha
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cebgfdha
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cebfgdha
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cebfdgha
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbegfdha
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cbefgdha
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbefdgha
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efbgcdha
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // efbcgdha
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efbcdgha
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ebgfcdha
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebfgcdha
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ebfcgdha
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebfcdgha
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebgcfdha
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebcgfdha
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ebcfgdha
																		}
																		else {
																			swap(v0, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebcfdgha
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // begfcdha
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // befgcdha
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // befcgdha
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // befcdgha
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // begcfdha
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // becgfdha
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // becfgdha
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // becfdgha
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcegfdha
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bcefgdha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcefdgha
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v6, v3)) {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ecbghdfa
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecbgdhfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ecbgdfha
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecbghdaf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ecbgdhaf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cebghdfa
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cebgdhfa
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cebgdfha
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cebghdaf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cebgdhaf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebghcdfa
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ebgchdfa
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ebcghdfa
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ebghcdaf
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebgchdaf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebcghdaf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebgcdhfa
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ebgcdfha
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ebgcdhaf
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebcgdhfa
																			}
																			else {
																				swap(v0, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ebcgdfha
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ebcgdhaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecdbghfa
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecbdghfa
																			}
																		}
																		else {
																			swap(v0, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebcdghfa
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); // ecdbghaf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v4, v6); swap(v5, v7); // ecbdghaf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v4, v6); swap(v5, v7); // ebcdghaf
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecdfbgha
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v6, v7); // ecdbgfha
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecdbfgha
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v4, v6); swap(v6, v7); // ecbdgfha
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecbdfgha
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v4, v6); swap(v6, v7); // ebcdgfha
																			}
																			else {
																				swap(v0, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebcdfgha
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cedbghfa
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdebghfa
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cebdghfa
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cedbghaf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); // cdebghaf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); // cebdghaf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v1)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cedfbgha
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cedbgfha
																				}
																				else {
																					swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cedbfgha
																				}
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v6, v7); // cebdgfha
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cebdfgha
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdefbgha
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v6, v7); // cdebgfha
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdebfgha
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v4, v2)) {
																if (compare(v7, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // beghcdfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // begchdfa
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // becghdfa
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // beghcdaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // begchdaf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // becghdaf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // begcdhfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // begcdfha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // begcdhaf
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // becgdhfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // becgdfha
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // becgdhaf
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cbeghdfa
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbegdhfa
																			}
																			else {
																				swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cbegdfha
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbeghdaf
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cbegdhaf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bceghdfa
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcegdhfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bcegdfha
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bceghdaf
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bcegdhaf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdbeghfa
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbedghfa
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbdeghfa
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cdbeghaf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); // cbedghaf
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cbdeghaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cdbegfha
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v4, v6); swap(v6, v7); // cbedgfha
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cbdegfha
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdbefgha
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbedfgha
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbdefgha
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // becdghfa
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcedghfa
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcdeghfa
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v4, v6); swap(v5, v7); // becdghaf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); // bcedghaf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bcdeghaf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v4, v6); swap(v6, v7); // becdgfha
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v6, v7); // bcedgfha
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bcdegfha
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // becdfgha
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcedfgha
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcdefgha
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v5)) {
													if (compare(v4, v1)) {
														if (compare(v4, v2)) {
															if (compare(v2, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v6); // ecdbgfah
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecdbgahf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // ecdbgafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v5, v6); swap(v6, v7); // ecdbaghf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v5, v6); // ecdbagfh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ecbgfdah
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v6); // ecbgdfah
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecbgdahf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ecbgdafh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v4); swap(v1, v2); swap(v4, v6); // ecbdgfah
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v4); swap(v1, v2); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecbdgahf
																				}
																				else {
																					swap(v0, v4); swap(v1, v2); swap(v4, v6); swap(v5, v6); // ecbdgafh
																				}
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v5, v6); swap(v6, v7); // ecbdaghf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v5, v6); // ecbdagfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // ebgfcdah
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ebgcfdah
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // ebgcdfah
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ebcgfdah
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v3, v6); swap(v4, v6); // ebcgdfah
																			}
																			else {
																				swap(v0, v4); swap(v4, v6); // ebcdgfah
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebgcdahf
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ebgcdafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebcgdahf
																			}
																			else {
																				swap(v0, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ebcgdafh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebcdgahf
																			}
																			else {
																				swap(v0, v4); swap(v4, v6); swap(v5, v6); // ebcdgafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v5, v6); swap(v6, v7); // ebcdaghf
																			}
																			else {
																				swap(v0, v4); swap(v5, v6); // ebcdagfh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // cedbgfah
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cedbgahf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cedbgafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cedbaghf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // cedbagfh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v6); // cdebgfah
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdebgahf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // cdebgafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v5, v6); swap(v6, v7); // cdebaghf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v5, v6); // cdebagfh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cebgfdah
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // cebgdfah
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cebgdahf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cebgdafh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v6); // cebdgfah
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cebdgahf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // cebdgafh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v5, v6); swap(v6, v7); // cebdaghf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v5, v6); // cebdagfh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v6, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cbegfdah
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // cbegdfah
																			}
																			else {
																				swap(v0, v2); swap(v2, v4); swap(v4, v6); // cbedgfah
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbegdahf
																			}
																			else {
																				swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cbegdafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbedgahf
																			}
																			else {
																				swap(v0, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // cbedgafh
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // cdbegfah
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdbegahf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cdbegafh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // cbdegfah
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbdegahf
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cbdegafh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cdbeaghf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v5, v6); swap(v6, v7); // cbedaghf
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cbdeaghf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v5, v6); swap(v6, v7); // cdbaeghf
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v5, v6); swap(v6, v7); // cbdaeghf
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // cdbeagfh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v5, v6); // cbedagfh
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // cbdeagfh
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v5, v6); // cdbaegfh
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v5, v6); // cbdaegfh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // begfcdah
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // begcfdah
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // begcdfah
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // becgfdah
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v6); // becgdfah
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v4, v6); // becdgfah
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // begcdahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // begcdafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // becgdahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // becgdafh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // becdgahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v4, v6); swap(v5, v6); // becdgafh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v5, v6); swap(v6, v7); // becdaghf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v5, v6); // becdagfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bcegfdah
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // bcegdfah
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcegdahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bcegdafh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v6); // bcedgfah
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcedgahf
																				}
																				else {
																					swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // bcedgafh
																				}
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v5, v6); swap(v6, v7); // bcedaghf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v5, v6); // bcedagfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // bcdegfah
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcdegahf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bcdegafh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bcdeaghf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // bcdeagfh
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v5, v6); swap(v6, v7); // bcdaeghf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v5, v6); // bcdaegfh
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v0)) {
														if (compare(v2, v1)) {
															if (compare(v5, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v4, v5); swap(v5, v6); // efcdbgah
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // ecfdbgah
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ecdfbgah
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v5, v6); // cefdbgah
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // cedfbgah
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cdefbgah
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v6); // efcbgdah
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // ecfbgdah
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cefbgdah
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // efcbdgah
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ecfbdgah
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // cefbdgah
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v6); // ecbfgdah
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // cebfgdah
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // cbefgdah
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ecbfdgah
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cebfdgah
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cbefdgah
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // ecdbfgah
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cedbfgah
																			}
																		}
																		else {
																			if (compare(v4, v1)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // cdebfgah
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cdbefgah
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v4, v5); swap(v5, v6); // ecbdfgah
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // cebdfgah
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // cbedfgah
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cbdefgah
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // efbgcdah
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // efbcgdah
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ebfgcdah
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // ebfcgdah
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v5); swap(v4, v6); // ebcfgdah
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // efbcdgah
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ebfcdgah
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ebcfdgah
																		}
																		else {
																			swap(v0, v4); swap(v4, v5); swap(v5, v6); // ebcdfgah
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // befgcdah
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // befcgdah
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // befcdgah
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v6); // becfgdah
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // becfdgah
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v4, v5); swap(v5, v6); // becdfgah
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // bcefgdah
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bcefdgah
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // bcedfgah
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bcdefgah
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v3, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v4, v5); // efcdbagh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v5); // ecfdbagh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // ecdfbagh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); // cefdbagh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); // cedfbagh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // cdefbagh
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v4, v5); // ecdbfagh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cedbfagh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v4, v5); // cdebfagh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); // ecdbafgh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v4); // cedbafgh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); // cdebafgh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v5); // efcbdagh
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // efbcdagh
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // ecfbdagh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); // cefbdagh
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v5); // ecbfdagh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // cebfdagh
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // ebfcdagh
																			}
																			else {
																				swap(v0, v4); swap(v3, v5); swap(v4, v5); // ebcfdagh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v4, v5); // ecbdfagh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v4, v5); // cebdfagh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v4, v5); // ebcdfagh
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); // ecbdafgh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); // cebdafgh
																			}
																		}
																		else {
																			swap(v0, v4); // ebcdafgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // cbefdagh
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v4, v5); // cbedfagh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cdbefagh
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cbdefagh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // befcdagh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v5); // becfdagh
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v4, v5); // becdfagh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // bcefdagh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v4, v5); // bcedfagh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // bcdefagh
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); swap(v3, v4); // cdbeafgh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v2, v4); // cbedafgh
																			}
																			else {
																				swap(v0, v2); swap(v2, v3); swap(v3, v4); // cbdeafgh
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v3); // cdbaefgh
																		}
																		else {
																			swap(v0, v2); swap(v2, v3); // cbdaefgh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v4); // becdafgh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); // bcedafgh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); swap(v3, v4); // bcdeafgh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v3); // bcdaefgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
						else {
							if (compare(v5, v4)) {
								if (compare(v7, v6)) {
									if (compare(v6, v4)) {
										if (compare(v6, v0)) {
											if (compare(v6, v2)) {
												if (compare(v6, v5)) {
													if (compare(v6, v1)) {
														if (compare(v5, v1)) {
															if (compare(v4, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgfecbad
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hgfcebad
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hgcfebad
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgfcbead
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgcfbead
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hgfebcad
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v6, v7); // hgfbecad
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hgfbcead
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgfebacd
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hgfbeacd
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hgfcbaed
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgfcbade
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hgcfbaed
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgcfbade
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgfbcaed
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hgfbcade
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgfbaecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hgfbaced
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); // hgfbacde
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hgbfecad
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgbfcead
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hgbfcaed
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgbfcade
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v0)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); // hgbfeacd
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hgbfaecd
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgbfaced
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hgbfacde
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hgcbfead
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgcbfaed
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hgcbfade
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hgbcfead
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgbcfaed
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hgbcfade
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hgcbafed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v7); // hgcbafde
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hgcbadfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hgbcafed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); // hgbcafde
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hgbcadfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); // hgbafecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hgbafced
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgbafcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); // hgbacfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hgbacfde
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgbacdfe
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v7, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hbgfecad
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbgfcead
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hbgfcaed
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbgfcade
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hbgcfead
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbgcfaed
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hbgcfade
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hbgcafed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); // hbgcafde
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hbgcadfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bhgfecad
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhgfcead
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bhgfcaed
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhgfcade
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bhgcfead
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhgcfaed
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bhgcfade
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bhgcafed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); // bhgcafde
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bhgcadfe
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); // hbgfeacd
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hbgfaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbgfaced
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hbgfacde
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); // hbgafecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hbgafced
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbgafcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); // hbgacfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hbgacfde
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbgacdfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); // bhgfeacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bhgfaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhgfaced
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bhgfacde
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); // bhgafecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bhgafced
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhgafcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); // bhgacfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bhgacfde
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhgacdfe
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v2, v0)) {
															if (compare(v4, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hfgecbad
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v6, v7); // hfgcebad
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hfgcbead
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfgebcad
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hfgbecad
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfgbcead
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v6, v7); // hfbgecad
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v6, v7); // hbfgecad
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v6, v7); // bhfgecad
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hfbgcead
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hbfgcead
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bhfgcead
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfgcbaed
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hfgbcaed
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfbgcaed
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbfgcaed
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhfgcaed
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hfgcbade
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfgbcade
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hfbgcade
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hbfgcade
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bhfgcade
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hfgebacd
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); // hfgbeacd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hfgbaecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfgbaced
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hfgbacde
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hfbgeacd
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfbgaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hfbgaced
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); // hfbgacde
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hbfgeacd
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbfgaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hbfgaced
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); // hbfgacde
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bhfgeacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhfgaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bhfgaced
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); // bhfgacde
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v4, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhgecbad
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fhgcebad
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhgcbead
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fhgebcad
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); // fhgbecad
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fhgbcead
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fhbgecad
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fbhgecad
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bfhgecad
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhbgcead
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbhgcead
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfhgcead
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fhgcbaed
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); // fhgbcaed
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fhbgcaed
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fbhgcaed
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bfhgcaed
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); // fhgcbade
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fhgbcade
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); // fhbgcade
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); // fbhgcade
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); // bfhgcade
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); // fhgebacd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); // fhgbeacd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); // fhgbaecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fhgbaced
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhgbacde
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v6, v7); // fhbgeacd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fhbgaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhbgaced
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fhbgacde
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v6, v7); // fbhgeacd
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fbhgaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbhgaced
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fbhgacde
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v6, v7); // bfhgeacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bfhgaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfhgaced
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bfhgacde
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v2)) {
													if (compare(v2, v1)) {
														if (compare(v6, v1)) {
															if (compare(v6, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hcgfebad
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcgfbead
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hcgfbaed
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcgfbade
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hcgbfead
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcgbfaed
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hcgbfade
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hcgbafed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); // hcgbafde
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hcgbadfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v6, v7); // hfcgebad
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fhcgebad
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v6, v7); // hcfgebad
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hfcgbead
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhcgbead
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hcfgbead
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfcgbaed
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fhcgbaed
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcfgbaed
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hfcgbade
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v7); // fhcgbade
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hcfgbade
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfcbgead
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fhcbgead
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcfbgead
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hfcbgaed
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v6); // fhcbgaed
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hcfbgaed
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfcbgade
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fhcbgade
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcfbgade
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hcbgfead
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcbfgead
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcbgfaed
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hcbfgaed
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hcbgafed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hcbgfade
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcbfgade
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v7); // hcbgafde
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hcbgadfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfbcgead
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fhbcgead
																		}
																	}
																	else {
																		swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbfcgead
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hfbcgaed
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); // fhbcgaed
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hbfcgaed
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfbcgade
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fhbcgade
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbfcgade
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hbcgfead
																		}
																		else {
																			swap(v0, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbcfgead
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbcgfaed
																			}
																			else {
																				swap(v0, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hbcfgaed
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hbcgafed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hbcgfade
																		}
																		else {
																			swap(v0, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbcfgade
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v3, v6); swap(v4, v7); // hbcgafde
																		}
																		else {
																			swap(v0, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hbcgadfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fbhcgead
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhfcgead
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bfhcgead
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); // fbhcgaed
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bhfcgaed
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); // bfhcgaed
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fbhcgade
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhfcgade
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bfhcgade
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bhcgfead
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhcfgead
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhcgfaed
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bhcfgaed
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bhcgafed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bhcgfade
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhcfgade
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v7); // bhcgafde
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bhcgadfe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v6, v1)) {
															if (compare(v6, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); // chgfebad
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // chgfbead
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chgfbaed
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // chgfbade
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); // chgbfead
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // chgbfaed
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chgbfade
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); // chgbafed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // chgbafde
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chgbadfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fchgebad
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // chfgebad
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v6, v7); // cfhgebad
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fchgbead
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chfgbead
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cfhgbead
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fchgbaed
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // chfgbaed
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfhgbaed
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); // fchgbade
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); // chfgbade
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cfhgbade
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fchbgead
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // chfbgead
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfhbgead
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); // fchbgaed
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); // chfbgaed
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cfhbgaed
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fchbgade
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // chfbgade
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfhbgade
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // chbgfead
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chbfgead
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chbgfaed
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // chbfgaed
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // chbgafed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // chbgfade
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chbfgade
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); // chbgafde
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // chbgadfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v7, v5)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cbhgfead
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbhfgead
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbhgfaed
																			}
																			else {
																				swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cbhfgaed
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cbhgafed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cbhgfade
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbhfgade
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); // cbhgafde
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cbhgadfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fcbhgead
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cfbhgead
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cbfhgead
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v6); // fcbhgaed
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); // cfbhgaed
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); // cbfhgaed
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fcbhgade
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cfbhgade
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cbfhgade
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bchgfead
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bchfgead
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bchgfaed
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bchfgaed
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bchgafed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bchgfade
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bchfgade
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); // bchgafde
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bchgadfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		swap(v0, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fbchgead
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bfchgead
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bcfhgead
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v3, v7); swap(v4, v6); // fbchgaed
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v6); // bfchgaed
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); // bcfhgaed
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fbchgade
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bfchgade
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bcfhgade
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v0)) {
												if (compare(v5, v1)) {
													if (compare(v7, v1)) {
														if (compare(v2, v1)) {
															if (compare(v5, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfcbaged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hfcbagde
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v7); // hfcbadge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fhcbaged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhcbagde
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fhcbadge
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fchbaged
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fchbagde
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fchbadge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcfbaged
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hcfbagde
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); // hcfbadge
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // chfbaged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chfbagde
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // chfbadge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfhbaged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cfhbagde
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); // cfhbadge
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hfbagecd
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); // hfbagced
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfbcaged
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hfbacged
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hfbcagde
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); // hfbcadge
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hfbagcde
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfbacgde
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hfbacdge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhbagecd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fhbagced
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fhbcaged
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhbacged
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhbcagde
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fhbcadge
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhbagcde
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fhbacgde
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); // fhbacdge
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v0)) {
															if (compare(v7, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbhagecd
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fbhagced
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fbhcaged
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbhacged
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbhcagde
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fbhcadge
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbhagcde
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fbhacgde
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); // fbhacdge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fcbhaged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcbhagde
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fcbhadge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cfbhaged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfbhagde
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cfbhadge
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fbchaged
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbchagde
																		}
																		else {
																			swap(v0, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fbchadge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcbahged
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfbahged
																			}
																		}
																		else {
																			swap(v0, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbcahged
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fcbahgde
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cfbahgde
																			}
																		}
																		else {
																			swap(v0, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fbcahgde
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v7); // fcbahdge
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); // cfbahdge
																			}
																		}
																		else {
																			swap(v0, v5); swap(v3, v5); swap(v4, v7); // fbcahdge
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fcbadhge
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cfbadhge
																			}
																		}
																		else {
																			swap(v0, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fbcadhge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fbahgecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); // fbahgced
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fbahgcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fbahcged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbahcgde
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fbahcdge
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbachged
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fbachgde
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); // fbachdge
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fbacdhge
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v7, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hbfagecd
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); // hbfagced
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbfcaged
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hbfacged
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hbfcagde
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); // hbfcadge
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hbfagcde
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbfacgde
																			}
																			else {
																				swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hbfacdge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcbfaged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hcbfagde
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v7); // hcbfadge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chbfaged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // chbfagde
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); // chbfadge
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbcfaged
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hbcfagde
																		}
																		else {
																			swap(v0, v7); swap(v3, v5); swap(v4, v7); // hbcfadge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bhfagecd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); // bhfagced
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhfcaged
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bhfacged
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bhfcagde
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); // bhfcadge
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bhfagcde
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhfacgde
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bhfacdge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbhfaged
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cbhfagde
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); // cbhfadge
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhcfaged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bhcfagde
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v7); // bhcfadge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bchfaged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bchfagde
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); // bchfadge
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v0)) {
															if (compare(v7, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfhagecd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bfhagced
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bfhcaged
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfhacged
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfhcagde
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bfhcadge
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfhagcde
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bfhacgde
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); // bfhacdge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cbfhaged
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbfhagde
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cbfhadge
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bfchaged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfchagde
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bfchadge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bcfhaged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcfhagde
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bcfhadge
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbfahged
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfcahged
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcfahged
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cbfahgde
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bfcahgde
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bcfahgde
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); // cbfahdge
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v7); // bfcahdge
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); // bcfahdge
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cbfadhge
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bfcadhge
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bcfadhge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bfahgecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); // bfahgced
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bfahgcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bfahcged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfahcgde
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bfahcdge
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfachged
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bfachgde
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); // bfachdge
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bfacdhge
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v0)) {
													if (compare(v2, v1)) {
														if (compare(v6, v3)) {
															if (compare(v7, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v6); // hcbagfed
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hcbafged
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hcbagfde
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcbafgde
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcbagdfe
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); // chbagfed
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // chbafged
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // chbagfde
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chbafgde
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chbagdfe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); // cbhagfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cbhagfde
																			}
																			else {
																				swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbhagdfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v3, v7); swap(v4, v6); // cbahgfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cbahgfde
																			}
																			else {
																				swap(v0, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbahgdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cbhafged
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cbahfged
																			}
																			else {
																				swap(v0, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbafhged
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbhafgde
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbahfgde
																			}
																			else {
																				swap(v0, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cbafhgde
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hcbafdge
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hcbadgfe
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v7); // hcbadfge
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // chbafdge
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // chbadgfe
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); // chbadfge
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cbhafdge
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cbhadgfe
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); // cbhadfge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cbahfdge
																		}
																		else {
																			swap(v0, v2); swap(v3, v5); swap(v4, v7); // cbafhdge
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cbahdgfe
																		}
																		else {
																			swap(v0, v2); swap(v3, v7); swap(v4, v7); // cbahdfge
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v4, v7); swap(v5, v6); // cbadhgfe
																		}
																		else {
																			swap(v0, v2); swap(v4, v7); // cbadhfge
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cbafdhge
																		}
																		else {
																			swap(v0, v2); swap(v4, v5); swap(v5, v7); // cbadfhge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v7, v2)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v3, v7); swap(v4, v6); // hbcagfed
																		}
																		else {
																			swap(v0, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hbcafged
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hbcagfde
																			}
																			else {
																				swap(v0, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbcafgde
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbcagdfe
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v6); // bhcagfed
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bhcafged
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bhcagfde
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhcafgde
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhcagdfe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); // bchagfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bchagfde
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bchagdfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v6); // bcahgfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bcahgfde
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcahgdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bchafged
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bcahfged
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcafhged
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bchafgde
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcahfgde
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bcafhgde
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v5, v3)) {
																		swap(v0, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hbcafdge
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hbcadgfe
																		}
																		else {
																			swap(v0, v7); swap(v3, v7); swap(v4, v7); // hbcadfge
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bhcafdge
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bhcadgfe
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v7); // bhcadfge
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bchafdge
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bchadgfe
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); // bchadfge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bcahfdge
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v7); // bcafhdge
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bcahdgfe
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v7); // bcahdfge
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v4, v7); swap(v5, v6); // bcadhgfe
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v4, v7); // bcadhfge
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bcafdhge
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v4, v5); swap(v5, v7); // bcadfhge
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v0)) {
														if (compare(v7, v1)) {
															if (compare(v6, v5)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hbagfecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbagfced
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hbagfcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hbagcfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); // hbagcfde
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hbagcdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); // hbacgfed
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hbacgfde
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbacgdfe
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hbacdgfe
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbafgecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hbafgced
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbafgcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbafcged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hbafcgde
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); // hbafcdge
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hbacfged
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbacfgde
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hbacfdge
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); // hbacdfge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bhagfecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhagfced
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bhagfcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bhagcfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); // bhagcfde
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bhagcdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); // bhacgfed
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bhacgfde
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhacgdfe
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bhacdgfe
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhafgecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bhafgced
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhafgcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhafcged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bhafcgde
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); // bhafcdge
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bhacfged
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhacfgde
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bhacfdge
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); // bhacdfge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v5)) {
																		swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bahgfecd
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bahfgecd
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bafhgecd
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bahgfced
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bahfgced
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v6); // bafhgced
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bahgfcde
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bahfgcde
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bafhgcde
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bahfcged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bahfcgde
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v7); // bahfcdge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bafhcged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bafhcgde
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bafhcdge
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bafchged
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bafchgde
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v7); // bafchdge
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bafcdhge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v5, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bahgcfed
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v6); // bahcgfed
																			}
																			else {
																				swap(v0, v1); swap(v3, v7); swap(v4, v6); // bachgfed
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v7); // bahgcfde
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bahcgfde
																			}
																			else {
																				swap(v0, v1); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bachgfde
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bahgcdfe
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bahcgdfe
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bahcdgfe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bachgdfe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v3, v7); swap(v4, v7); swap(v5, v6); // bachdgfe
																			}
																			else {
																				swap(v0, v1); swap(v4, v7); swap(v5, v6); // bacdhgfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bahcfged
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bachfged
																			}
																			else {
																				swap(v0, v1); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bacfhged
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bahcfgde
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bachfgde
																			}
																			else {
																				swap(v0, v1); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bacfhgde
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bahcfdge
																			}
																			else {
																				swap(v0, v1); swap(v3, v7); swap(v4, v5); swap(v5, v7); // bachfdge
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v3, v5); swap(v4, v7); // bacfhdge
																			}
																			else {
																				swap(v0, v1); swap(v3, v5); swap(v4, v5); swap(v5, v7); // bacfdhge
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v7); // bahcdfge
																			}
																			else {
																				swap(v0, v1); swap(v3, v7); swap(v4, v7); // bachdfge
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v4, v7); // bacdhfge
																			}
																			else {
																				swap(v0, v1); swap(v4, v5); swap(v5, v7); // bacdfhge
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v0)) {
											if (compare(v4, v2)) {
												if (compare(v7, v4)) {
													if (compare(v7, v5)) {
														if (compare(v2, v0)) {
															if (compare(v6, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v6, v7); // hfegcbad
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hfecgbad
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfecbgad
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hfegbcad
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfebgcad
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hfebcgad
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hfbegcad
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hbfegcad
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bhfegcad
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfbecgad
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbfecgad
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhfecgad
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hfecbagd
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); // hfebcagd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hfbecagd
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hbfecagd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bhfecagd
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfecbadg
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hfebcadg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfbecadg
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbfecadg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhfecadg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfegbacd
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hfebgacd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfebagcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hfebacgd
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfebacdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfbegacd
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hfbeagcd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); // hfbeacgd
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hfbeacdg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbfegacd
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hbfeagcd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); // hbfeacgd
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hbfeacdg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhfegacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bhfeagcd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); // bhfeacgd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bhfeacdg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v6, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fhegcbad
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhecgbad
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fhecbgad
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhegbcad
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fhebgcad
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); // fhebcgad
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhbegcad
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbhegcad
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfhegcad
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fhbecgad
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fbhecgad
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bfhecgad
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); // fhecbagd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); // fhebcagd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); // fhbecagd
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); // fbhecagd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); // bfhecagd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fhecbadg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v6, v7); // fhebcadg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fhbecadg
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fbhecadg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bfhecadg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fhegbacd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); // fhebgacd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fhebagcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); // fhebacgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fhebacdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fhbegacd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhbeagcd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fhbeacgd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhbeacdg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fbhegacd
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbheagcd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fbheacgd
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbheacdg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bfhegacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfheagcd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bfheacgd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfheacdg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v7, v1)) {
															if (compare(v6, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fehgcbad
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fehcgbad
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fechgbad
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); // fehcbgad
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v5, v6); // fechbgad
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fehgbcad
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fehbgcad
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fehbcgad
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v6, v7); // fehgbacd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fehbgacd
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); // fehcbagd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v6, v7); // fehcbadg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); // fechbagd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v6, v7); // fechbadg
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); // fehbcagd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fehbcadg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fehbagcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fehbacgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fehbacdg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v2, v0)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // febhgcad
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); // febhcgad
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); // febhcagd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v6, v7); // febhcadg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v0)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); // febhgacd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // febhagcd
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); // febhacgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // febhacdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fecbhgad
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v7); // fecbhagd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fecbhadg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // febchgad
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); // febchagd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // febchadg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fecbahgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fecbahdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fecbadhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // febcahgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // febcahdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // febcadhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // febahgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // febahcgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // febahcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); // febachgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // febachdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v6, v7); // febacdhg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v5, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fbehgcad
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); // fbehcgad
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); // fbehcagd
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v6, v7); // fbehcadg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fbechgad
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); // fbechagd
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fbechadg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fbecahgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbecahdg
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fbecadhg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bfehgcad
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); // bfehcgad
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); // bfehcagd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v6, v7); // bfehcadg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bfechgad
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); // bfechagd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bfechadg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bfecahgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfecahdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bfecadhg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); // fbehgacd
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fbehagcd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); // fbehacgd
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fbehacdg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbeahgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fbeahcgd
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbeahcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); // fbeachgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fbeachdg
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v6, v7); // fbeacdhg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); // bfehgacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bfehagcd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); // bfehacgd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bfehacdg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfeahgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bfeahcgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfeahcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); // bfeachgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bfeachdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v6, v7); // bfeacdhg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v2, v1)) {
														if (compare(v4, v1)) {
															if (compare(v7, v4)) {
																if (compare(v6, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hfcegbad
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhcegbad
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fchegbad
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfcebgad
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fhcebgad
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fchebgad
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hfcebagd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v7); // fhcebagd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); // fchebagd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfcebadg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fhcebadg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fchebadg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fcehgbad
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); // fcehbgad
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); // fcehbagd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v6, v7); // fcehbadg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fcebhgad
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); // fcebhagd
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fcebhadg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fcebahgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcebahdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fcebadhg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hfcbegad
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v5, v6); // fhcbegad
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); // fchbegad
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v5, v7); // hfcbeagd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v7); // fhcbeagd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); // fchbeagd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hfcbeadg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v6, v7); // fhcbeadg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v6, v7); // fchbeadg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v5, v6); // fcbhegad
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fcbehgad
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v7); // fcbheagd
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v7); // fcbehagd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fcbeahgd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v6, v7); // fcbheadg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fcbehadg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcbeahdg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fcbeadhg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v7, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hfbcegad
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); // fhbcegad
																		}
																	}
																	else {
																		swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); // fbhcegad
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); // hfbceagd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); // fhbceagd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); // fbhceagd
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hfbceadg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v6, v7); // fhbceadg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v6, v7); // fbhceadg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v3, v7); swap(v5, v6); // fbchegad
																		}
																		else {
																			swap(v0, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fbcehgad
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v3, v7); // fbcheagd
																			}
																			else {
																				swap(v0, v5); swap(v3, v4); swap(v4, v7); // fbcehagd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fbceahgd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v3, v7); swap(v6, v7); // fbcheadg
																		}
																		else {
																			swap(v0, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fbcehadg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbceahdg
																		}
																		else {
																			swap(v0, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fbceadhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hbfcegad
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bhfcegad
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); // bfhcegad
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); // hbfceagd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); // bhfceagd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); // bfhceagd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hbfceadg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bhfceadg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v6, v7); // bfhceadg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v5, v6); // bfchegad
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bfcehgad
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v7); // bfcheagd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v7); // bfcehagd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bfceahgd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v6, v7); // bfcheadg
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bfcehadg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfceahdg
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bfceadhg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v1)) {
															if (compare(v7, v4)) {
																if (compare(v6, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hcfegbad
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chfegbad
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cfhegbad
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcfebgad
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // chfebgad
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfhebgad
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hcfebagd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); // chfebagd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cfhebagd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcfebadg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // chfebadg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfhebadg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); // cfehgbad
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cfehbgad
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); // cfehbagd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cfehbadg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfebhgad
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cfebhagd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfebhadg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); // cfebahgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cfebahdg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v6, v7); // cfebadhg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hcfbegad
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); // chfbegad
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cfhbegad
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); // hcfbeagd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); // chfbeagd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); // cfhbeagd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hcfbeadg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v6, v7); // chfbeadg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cfhbeadg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); // cfbhegad
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cfbehgad
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); // cfbheagd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); // cfbehagd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cfbeahgd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v6, v7); // cfbheadg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cfbehadg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfbeahdg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cfbeadhg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v7, v5)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hcbfegad
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // chbfegad
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cbhfegad
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v5, v7); // hcbfeagd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); // chbfeagd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); // cbhfeagd
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hcbfeadg
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // chbfeadg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cbhfeadg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); // cbfhegad
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cbfehgad
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v2, v5); swap(v3, v7); // cbfheagd
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); // cbfehagd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cbfeahgd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v6, v7); // cbfheadg
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cbfehadg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbfeahdg
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cbfeadhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		swap(v0, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hbcfegad
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bhcfegad
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bchfegad
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v3, v5); swap(v5, v7); // hbcfeagd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v5, v7); // bhcfeagd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); // bchfeagd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hbcfeadg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bhcfeadg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bchfeadg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); // bcfhegad
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bcfehgad
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); // bcfheagd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); // bcfehagd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bcfeahgd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v6, v7); // bcfheadg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bcfehadg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcfeahdg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bcfeadhg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v0)) {
												if (compare(v7, v1)) {
													if (compare(v5, v1)) {
														if (compare(v2, v1)) {
															if (compare(v5, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hfcbaegd
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfcbaedg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hfcbadeg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v5); // fhcbaegd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fhcbaedg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhcbadeg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); // fchbaegd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fchbaedg
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fchbadeg
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hcfbaegd
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcfbaedg
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hcfbadeg
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); // chfbaegd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // chfbaedg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chfbadeg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cfhbaegd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfhbaedg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cfhbadeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); // hfbaegcd
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); // hfbaecgd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hfbcaegd
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); // hfbacegd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfbcaedg
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hfbcadeg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v6, v7); // hfbaecdg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hfbacedg
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfbacdeg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fhbaegcd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); // fhbaecgd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); // fhbcaegd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fhbacegd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fhbcaedg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhbcadeg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fhbaecdg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhbacedg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fhbacdeg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v5, v2)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); // hbfaegcd
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); // hbfaecgd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hbfcaegd
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); // hbfacegd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbfcaedg
																		}
																		else {
																			swap(v0, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hbfcadeg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v6, v7); // hbfaecdg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hbfacedg
																			}
																			else {
																				swap(v0, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbfacdeg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hcbfaegd
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcbfaedg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hcbfadeg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // chbfaegd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chbfaedg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // chbfadeg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hbcfaegd
																		}
																		else {
																			swap(v0, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbcfaedg
																		}
																	}
																	else {
																		swap(v0, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hbcfadeg
																	}
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v5); // hcbafegd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); // chbafegd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v7); swap(v4, v5); // hbcafegd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hcbafedg
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // chbafedg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hbcafedg
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcbafdeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chbafdeg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbcafdeg
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hcbadfeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // chbadfeg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hbcadfeg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hbafegcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); // hbafecgd
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hbafecdg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hbafcegd
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbafcedg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hbafcdeg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); // hbacfegd
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hbacfedg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbacfdeg
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hbacdfeg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v5, v0)) {
															if (compare(v5, v2)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); // bhfaegcd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); // bhfaecgd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bhfcaegd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); // bhfacegd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhfcaedg
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bhfcadeg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v6, v7); // bhfaecdg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bhfacedg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhfacdeg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cbhfaegd
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbhfaedg
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cbhfadeg
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bhcfaegd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhcfaedg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bhcfadeg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bchfaegd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bchfaedg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bchfadeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); // cbhafegd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v5); // bhcafegd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); // bchafegd
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cbhafedg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bhcafedg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bchafedg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbhafdeg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhcafdeg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bchafdeg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cbhadfeg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bhcadfeg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bchadfeg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bhafegcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); // bhafecgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bhafecdg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bhafcegd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhafcedg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bhafcdeg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); // bhacfegd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bhacfedg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhacfdeg
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bhacdfeg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v7, v2)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fbhaegcd
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); // fbhaecgd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); // fbhcaegd
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fbhacegd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fbhcaedg
																		}
																		else {
																			swap(v0, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbhcadeg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fbhaecdg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbhacedg
																			}
																			else {
																				swap(v0, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fbhacdeg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v5); // fcbhaegd
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fcbhaedg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcbhadeg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); // cfbhaegd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cfbhaedg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfbhadeg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v3, v7); swap(v4, v5); // fbchaegd
																		}
																		else {
																			swap(v0, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fbchaedg
																		}
																	}
																	else {
																		swap(v0, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbchadeg
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bfhaegcd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); // bfhaecgd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); // bfhcaegd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bfhacegd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bfhcaedg
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfhcadeg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bfhaecdg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfhacedg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bfhacdeg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); // cbfhaegd
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cbfhaedg
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbfhadeg
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v5); // bfchaegd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bfchaedg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfchadeg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); // bcfhaegd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bcfhaedg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcfhadeg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v0)) {
													if (compare(v2, v1)) {
														if (compare(v4, v3)) {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fcbahegd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v5, v7); // fcbaehgd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcbahedg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fcbaehdg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v6, v7); // fcbaedhg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cfbahegd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); // cfbaehgd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfbahedg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cfbaehdg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v6, v7); // cfbaedhg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v4)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cbfahegd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v3, v7); swap(v4, v5); // cbahfegd
																			}
																			else {
																				swap(v0, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cbafhegd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbfahedg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cbahfedg
																			}
																			else {
																				swap(v0, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbafhedg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); // cbfaehgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cbfaehdg
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v6, v7); // cbfaedhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v3, v5); swap(v5, v7); // cbafehgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cbafehdg
																			}
																			else {
																				swap(v0, v2); swap(v3, v5); swap(v6, v7); // cbafedhg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fcbahdeg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcbadheg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fcbadehg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cfbahdeg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfbadheg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cfbadehg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cbfahdeg
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbfadheg
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cbfadehg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbahfdeg
																		}
																		else {
																			swap(v0, v2); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cbafhdeg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbafdheg
																		}
																		else {
																			swap(v0, v2); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cbafdehg
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cbahdfeg
																		}
																		else {
																			swap(v0, v2); swap(v4, v7); swap(v6, v7); // cbadhfeg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbadfheg
																		}
																		else {
																			swap(v0, v2); swap(v4, v5); swap(v6, v7); // cbadfehg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v2)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fbcahegd
																		}
																		else {
																			swap(v0, v5); swap(v3, v5); swap(v5, v7); // fbcaehgd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbcahedg
																			}
																			else {
																				swap(v0, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fbcaehdg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v3, v5); swap(v6, v7); // fbcaedhg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bfcahegd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v5, v7); // bfcaehgd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfcahedg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bfcaehdg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v6, v7); // bfcaedhg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v4)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bcfahegd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v5); // bcahfegd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bcafhegd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcfahedg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bcahfedg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcafhedg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); // bcfaehgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bcfaehdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v6, v7); // bcfaedhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v5, v7); // bcafehgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bcafehdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v6, v7); // bcafedhg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v7, v3)) {
																		swap(v0, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fbcahdeg
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbcadheg
																		}
																		else {
																			swap(v0, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fbcadehg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bfcahdeg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfcadheg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bfcadehg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bcfahdeg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcfadheg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bcfadehg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcahfdeg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bcafhdeg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcafdheg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bcafdehg
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bcahdfeg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v4, v7); swap(v6, v7); // bcadhfeg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v2); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcadfheg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v4, v5); swap(v6, v7); // bcadfehg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v5, v1)) {
															if (compare(v7, v4)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); // fbahegcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v7); // fbahecgd
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v6, v7); // fbahecdg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); // fbahcegd
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fbahcedg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbahcdeg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fbachegd
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbachedg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fbachdeg
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbacdheg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fbaehgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); // fbaehcgd
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fbaehcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fbaechgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbaechdg
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fbaecdhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); // fbacehgd
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fbacehdg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v6, v7); // fbacedhg
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fbacdehg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v4)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); // bfahegcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); // bfahecgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v6, v7); // bfahecdg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); // bfahcegd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bfahcedg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfahcdeg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bfachegd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfachedg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bfachdeg
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfacdheg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bfaehgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); // bfaehcgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bfaehcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bfaechgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfaechdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bfaecdhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); // bfacehgd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bfacehdg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v6, v7); // bfacedhg
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bfacdehg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																if (compare(v7, v4)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bahfegcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v5, v7); // bahfecgd
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bahfecdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v5, v6); // bafhegcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v7); // bafhecgd
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v6, v7); // bafhecdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bafehgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v7); // bafehcgd
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bafehcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bafechgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bafechdg
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bafecdhg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bahfcegd
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bahfcedg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bahfcdeg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v5); // bafhcegd
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bafhcedg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bafhcdeg
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bafchegd
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bafchedg
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bafchdeg
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bafcdheg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v5, v7); // bafcehgd
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bafcehdg
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v6, v7); // bafcedhg
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bafcdehg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v5); // bahcfegd
																		}
																		else {
																			swap(v0, v1); swap(v3, v7); swap(v4, v5); // bachfegd
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bacfhegd
																		}
																		else {
																			swap(v0, v1); swap(v3, v5); swap(v5, v7); // bacfehgd
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bahcfedg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bachfedg
																			}
																			else {
																				swap(v0, v1); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bacfhedg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bacfehdg
																		}
																		else {
																			swap(v0, v1); swap(v3, v5); swap(v6, v7); // bacfedhg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bahcfdeg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bachfdeg
																			}
																			else {
																				swap(v0, v1); swap(v3, v5); swap(v4, v7); swap(v6, v7); // bacfhdeg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bacfdheg
																		}
																		else {
																			swap(v0, v1); swap(v3, v5); swap(v4, v5); swap(v6, v7); // bacfdehg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bahcdfeg
																		}
																		else {
																			swap(v0, v1); swap(v3, v7); swap(v4, v7); swap(v6, v7); // bachdfeg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v4, v7); swap(v6, v7); // bacdhfeg
																			}
																			else {
																				swap(v0, v1); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bacdfheg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v4, v5); swap(v6, v7); // bacdfehg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v7, v4)) {
										if (compare(v7, v0)) {
											if (compare(v7, v2)) {
												if (compare(v7, v5)) {
													if (compare(v7, v1)) {
														if (compare(v5, v1)) {
															if (compare(v4, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ghfecbad
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); // ghfcebad
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v5, v7); // ghcfebad
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ghfcbead
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ghcfbead
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); // ghfebcad
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); // ghfbecad
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); // ghfbcead
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ghfebacd
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); // ghfbeacd
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghfcbaed
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ghfcbade
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghcfbaed
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ghcfbade
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ghfbcaed
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghfbcade
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ghfbaecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); // ghfbaced
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ghfbacde
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); // ghbfecad
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ghbfcead
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghbfcaed
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ghbfcade
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v0)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ghbfeacd
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghbfaecd
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ghbfaced
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghbfacde
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v5); // ghcbfead
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ghcbfaed
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghcbfade
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); // ghbcfead
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ghbcfaed
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghbcfade
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v6); // ghcbafed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ghcbafde
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghcbadfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); // ghbcafed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ghbcafde
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghbcadfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ghbafecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghbafced
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ghbafcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ghbacfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); // ghbacfde
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ghbacdfe
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v6, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); // gbhfecad
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gbhfcead
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbhfcaed
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gbhfcade
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); // gbhcfead
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gbhcfaed
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbhcfade
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); // gbhcafed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gbhcafde
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbhcadfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); // bghfecad
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bghfcead
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bghfcaed
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bghfcade
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); // bghcfead
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bghcfaed
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bghcfade
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); // bghcafed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bghcafde
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bghcadfe
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gbhfeacd
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbhfaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gbhfaced
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbhfacde
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gbhafecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbhafced
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gbhafcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gbhacfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); // gbhacfde
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gbhacdfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bghfeacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bghfaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bghfaced
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bghfacde
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bghafecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bghafced
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bghafcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bghacfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); // bghacfde
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bghacdfe
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v2, v0)) {
															if (compare(v4, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); // gfhecbad
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); // gfhcebad
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); // gfhcbead
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gfhebcad
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); // gfhbecad
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gfhbcead
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); // gfbhecad
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v7); // gbfhecad
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); // bgfhecad
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); // gfbhcead
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); // gbfhcead
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); // bgfhcead
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gfhcbaed
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfhbcaed
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gfbhcaed
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gbfhcaed
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bgfhcaed
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfhcbade
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gfhbcade
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfbhcade
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbfhcade
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgfhcade
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfhebacd
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gfhbeacd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfhbaecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gfhbaced
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfhbacde
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); // gfbheacd
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gfbhaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); // gfbhaced
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gfbhacde
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); // gbfheacd
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gbfhaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); // gbfhaced
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gbfhacde
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); // bgfheacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bgfhaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); // bgfhaced
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bgfhacde
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v4, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fghecbad
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); // fghcebad
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fghcbead
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fghebcad
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fghbecad
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fghbcead
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); // fgbhecad
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); // fbghecad
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); // bfghecad
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fgbhcead
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fbghcead
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bfghcead
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); // fghcbaed
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fghbcaed
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); // fgbhcaed
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); // fbghcaed
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); // bfghcaed
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fghcbade
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); // fghbcade
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fgbhcade
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fbghcade
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bfghcade
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fghebacd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v6, v7); // fghbeacd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fghbaecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fghbaced
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fghbacde
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); // fgbheacd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); // fgbhaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fgbhaced
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgbhacde
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); // fbgheacd
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); // fbghaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fbghaced
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbghacde
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); // bfgheacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); // bfghaecd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bfghaced
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfghacde
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v2)) {
													if (compare(v2, v1)) {
														if (compare(v7, v1)) {
															if (compare(v7, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); // gchfebad
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gchfbead
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gchfbaed
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gchfbade
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); // gchbfead
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gchbfaed
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gchbfade
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); // gchbafed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gchbafde
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gchbadfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v7); // gfchebad
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v5, v6); // fgchebad
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); // gcfhebad
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v5); // gfchbead
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fgchbead
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); // gcfhbead
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gfchbaed
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v6); // fgchbaed
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gcfhbaed
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfchbade
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fgchbade
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcfhbade
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gfcbhead
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgcbhead
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gcfbhead
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfcbhaed
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fgcbhaed
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcfbhaed
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gfcbhade
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v7); // fgcbhade
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gcfbhade
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v5); // gcbhfead
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gcbfhead
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gcbhfaed
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcbfhaed
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v6); // gcbhafed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcbhfade
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gcbfhade
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gcbhafde
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcbhadfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gfbchead
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgbchead
																		}
																	}
																	else {
																		swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gbfchead
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfbchaed
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fgbchaed
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbfchaed
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gfbchade
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); // fgbchade
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gbfchade
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v3, v7); swap(v4, v5); // gbchfead
																		}
																		else {
																			swap(v0, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gbcfhead
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gbchfaed
																			}
																			else {
																				swap(v0, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbcfhaed
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v7); swap(v4, v6); // gbchafed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbchfade
																		}
																		else {
																			swap(v0, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gbcfhade
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gbchafde
																		}
																		else {
																			swap(v0, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbchadfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbgchead
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bgfchead
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfgchead
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fbgchaed
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgfchaed
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bfgchaed
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); // fbgchade
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bgfchade
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); // bfgchade
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v5); // bgchfead
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bgcfhead
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bgchfaed
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgcfhaed
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v6); // bgchafed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgchfade
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bgcfhade
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bgchafde
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgchadfe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v7, v1)) {
															if (compare(v7, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cghfebad
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cghfbead
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cghfbaed
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cghfbade
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cghbfead
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cghbfaed
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cghbfade
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cghbafed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); // cghbafde
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cghbadfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); // fcghebad
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); // cgfhebad
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); // cfghebad
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fcghbead
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cgfhbead
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); // cfghbead
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); // fcghbaed
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); // cgfhbaed
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cfghbaed
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fcghbade
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cgfhbade
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfghbade
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcgbhead
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgfbhead
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cfgbhead
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fcgbhaed
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cgfbhaed
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfgbhaed
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); // fcgbhade
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); // cgfbhade
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cfgbhade
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); // cgbhfead
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cgbfhead
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cgbhfaed
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgbfhaed
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); // cgbhafed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgbhfade
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cgbfhade
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cgbhafde
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgbhadfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v6, v5)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); // cbghfead
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cbgfhead
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cbghfaed
																			}
																			else {
																				swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbgfhaed
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); // cbghafed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbghfade
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cbgfhade
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cbghafde
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbghadfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcbghead
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfbghead
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbfghead
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fcbghaed
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cfbghaed
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cbfghaed
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v7); // fcbghade
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); // cfbghade
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); // cbfghade
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); // bcghfead
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bcgfhead
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bcghfaed
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcgfhaed
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); // bcghafed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcghfade
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bcgfhade
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bcghafde
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcghadfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		swap(v0, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbcghead
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfcghead
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcfghead
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fbcghaed
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bfcghaed
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bcfghaed
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v5); swap(v3, v6); swap(v4, v7); // fbcghade
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v7); // bfcghade
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); // bcfghade
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v0)) {
												if (compare(v5, v1)) {
													if (compare(v6, v1)) {
														if (compare(v2, v1)) {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gfcbahed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfcbahde
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gfcbadhe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgcbahed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fgcbahde
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgcbadhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcgbahed
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fcgbahde
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcgbadhe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gcfbahed
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcfbahde
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gcfbadhe
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgfbahed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cgfbahde
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgfbadhe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cfgbahed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfgbahde
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cfgbadhe
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfbahecd
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gfbahced
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gfbcahed
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfbached
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfbcahde
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gfbcadhe
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); // gfbahcde
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gfbachde
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfbacdhe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fgbahecd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgbahced
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgbcahed
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fgbached
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fgbcahde
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgbcadhe
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fgbahcde
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgbachde
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fgbacdhe
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v6, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fbgahecd
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fbgahced
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbgcahed
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fbgached
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fbgcahde
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbgcadhe
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fbgahcde
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbgachde
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fbgacdhe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcbgahed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fcbgahde
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcbgadhe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfbgahed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cfbgahde
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfbgadhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbcgahed
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fbcgahde
																		}
																		else {
																			swap(v0, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbcgadhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fcbaghed
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cfbaghed
																			}
																		}
																		else {
																			swap(v0, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fbcaghed
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcbaghde
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfbaghde
																			}
																		}
																		else {
																			swap(v0, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbcaghde
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fcbagdhe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cfbagdhe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fbcagdhe
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcbadghe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfbadghe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbcadghe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbaghecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fbaghced
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); // fbaghcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbagched
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fbagchde
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbagcdhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fbacghed
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbacghde
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fbacgdhe
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbacdghe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v6, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbfahecd
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gbfahced
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gbfcahed
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbfached
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbfcahde
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gbfcadhe
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); // gbfahcde
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gbfachde
																			}
																			else {
																				swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbfacdhe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gcbfahed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcbfahde
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gcbfadhe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cgbfahed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgbfahde
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cgbfadhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gbcfahed
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbcfahde
																		}
																		else {
																			swap(v0, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gbcfadhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgfahecd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bgfahced
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bgfcahed
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgfached
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgfcahde
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bgfcadhe
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); // bgfahcde
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bgfachde
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgfacdhe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cbgfahed
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbgfahde
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cbgfadhe
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bgcfahed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgcfahde
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bgcfadhe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bcgfahed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcgfahde
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bcgfadhe
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v6, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bfgahecd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bfgahced
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfgcahed
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bfgached
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bfgcahde
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfgcadhe
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bfgahcde
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfgachde
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bfgacdhe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbfgahed
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cbfgahde
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbfgadhe
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfcgahed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bfcgahde
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfcgadhe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcfgahed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bcfgahde
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcfgadhe
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cbfaghed
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bfcaghed
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bcfaghed
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbfaghde
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfcaghde
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcfaghde
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cbfagdhe
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bfcagdhe
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bcfagdhe
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbfadghe
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfcadghe
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcfadghe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfaghecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bfaghced
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); // bfaghcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfagched
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bfagchde
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfagcdhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bfacghed
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfacghde
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bfacgdhe
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfacdghe
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v0)) {
													if (compare(v2, v1)) {
														if (compare(v7, v3)) {
															if (compare(v6, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gcbahfed
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcbafhed
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v7); // gcbahfde
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gcbafhde
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gcbahdfe
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cgbahfed
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgbafhed
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); // cgbahfde
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cgbafhde
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cgbahdfe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cbgahfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); // cbgahfde
																			}
																			else {
																				swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cbgahdfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cbaghfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v3, v6); swap(v4, v7); // cbaghfde
																			}
																			else {
																				swap(v0, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cbaghdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbgafhed
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbagfhed
																			}
																			else {
																				swap(v0, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cbafghed
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cbgafhde
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cbagfhde
																			}
																			else {
																				swap(v0, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbafghde
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcbafdhe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcbadhfe
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gcbadfhe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgbafdhe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgbadhfe
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cgbadfhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbgafdhe
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbgadhfe
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cbgadfhe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbagfdhe
																		}
																		else {
																			swap(v0, v2); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cbafgdhe
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbagdhfe
																		}
																		else {
																			swap(v0, v2); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cbagdfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbadghfe
																		}
																		else {
																			swap(v0, v2); swap(v4, v6); swap(v6, v7); // cbadgfhe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbafdghe
																		}
																		else {
																			swap(v0, v2); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbadfghe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v6, v2)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gbcahfed
																		}
																		else {
																			swap(v0, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbcafhed
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v3, v6); swap(v4, v7); // gbcahfde
																			}
																			else {
																				swap(v0, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gbcafhde
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gbcahdfe
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bgcahfed
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgcafhed
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v7); // bgcahfde
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bgcafhde
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bgcahdfe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bcgahfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); // bcgahfde
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bcgahdfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bcaghfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v7); // bcaghfde
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bcaghdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcgafhed
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcagfhed
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bcafghed
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bcgafhde
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bcagfhde
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcafghde
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v5, v3)) {
																		swap(v0, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbcafdhe
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbcadhfe
																		}
																		else {
																			swap(v0, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gbcadfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgcafdhe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgcadhfe
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bgcadfhe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcgafdhe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcgadhfe
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bcgadfhe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcagfdhe
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bcafgdhe
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcagdhfe
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bcagdfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcadghfe
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v4, v6); swap(v6, v7); // bcadgfhe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcafdghe
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcadfghe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v0)) {
														if (compare(v6, v1)) {
															if (compare(v7, v5)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); // gbahfecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gbahfced
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbahfcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); // gbahcfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gbahcfde
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbahcdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gbachfed
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); // gbachfde
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gbachdfe
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbacdhfe
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gbafhecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbafhced
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gbafhcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gbafched
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbafchde
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gbafcdhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbacfhed
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gbacfhde
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbacfdhe
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gbacdfhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); // bgahfecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bgahfced
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgahfcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); // bgahcfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // bgahcfde
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgahcdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bgachfed
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); // bgachfde
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bgachdfe
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgacdhfe
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bgafhecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgafhced
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bgafhcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bgafched
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgafchde
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bgafcdhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgacfhed
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bgacfhde
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgacfdhe
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bgacdfhe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v7, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v5); // baghfecd
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bagfhecd
																		}
																	}
																	else {
																		swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bafghecd
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // baghfced
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bagfhced
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bafghced
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // baghfcde
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // bagfhcde
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v7); // bafghcde
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bagfched
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bagfchde
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bagfcdhe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bafgched
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bafgchde
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bafgcdhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bafcghed
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bafcghde
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bafcgdhe
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bafcdghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v5, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v6); // baghcfed
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bagchfed
																			}
																		}
																		else {
																			swap(v0, v1); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bacghfed
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // baghcfde
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v7); // bagchfde
																			}
																		}
																		else {
																			swap(v0, v1); swap(v3, v6); swap(v4, v7); // bacghfde
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // baghcdfe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bagchdfe
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bagcdhfe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v3, v6); swap(v4, v7); swap(v5, v6); // bacghdfe
																			}
																			else {
																				swap(v0, v1); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bacgdhfe
																			}
																		}
																		else {
																			swap(v0, v1); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bacdghfe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bagcfhed
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bacgfhed
																			}
																			else {
																				swap(v0, v1); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bacfghed
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bagcfhde
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v3, v6); swap(v4, v5); swap(v5, v7); // bacgfhde
																			}
																			else {
																				swap(v0, v1); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bacfghde
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bagcfdhe
																			}
																			else {
																				swap(v0, v1); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bacgfdhe
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v3, v5); swap(v4, v6); swap(v6, v7); // bacfgdhe
																			}
																			else {
																				swap(v0, v1); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bacfdghe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bagcdfhe
																			}
																			else {
																				swap(v0, v1); swap(v3, v6); swap(v4, v6); swap(v6, v7); // bacgdfhe
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v4, v6); swap(v6, v7); // bacdgfhe
																			}
																			else {
																				swap(v0, v1); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bacdfghe
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v0)) {
											if (compare(v4, v2)) {
												if (compare(v6, v4)) {
													if (compare(v6, v5)) {
														if (compare(v2, v0)) {
															if (compare(v7, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); // gfehcbad
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); // gfechbad
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gfecbhad
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); // gfehbcad
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gfebhcad
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); // gfebchad
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); // gfbehcad
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); // gbfehcad
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); // bgfehcad
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gfbechad
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gbfechad
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bgfechad
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfecbahd
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gfebcahd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfbecahd
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbfecahd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgfecahd
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gfecbadh
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // gfebcadh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gfbecadh
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gbfecadh
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bgfecadh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gfehbacd
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfebhacd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gfebahcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfebachd
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gfebacdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gfbehacd
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfbeahcd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gfbeachd
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // gfbeacdh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gbfehacd
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbfeahcd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gbfeachd
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // gbfeacdh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bgfehacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgfeahcd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bgfeachd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // bgfeacdh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v7, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); // fgehcbad
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fgechbad
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgecbhad
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fgehbcad
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgebhcad
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fgebchad
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fgbehcad
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fbgehcad
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bfgehcad
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgbechad
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbgechad
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfgechad
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fgecbahd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v6, v7); // fgebcahd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fgbecahd
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fbgecahd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bfgecahd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // fgecbadh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); // fgebcadh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // fgbecadh
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // fbgecadh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // bfgecadh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); // fgehbacd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fgebhacd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgebahcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fgebachd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // fgebacdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); // fgbehacd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fgbeahcd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgbeachd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fgbeacdh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); // fbgehacd
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fbgeahcd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbgeachd
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fbgeacdh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); // bfgehacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bfgeahcd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfgeachd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bfgeacdh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v6, v1)) {
															if (compare(v7, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // feghcbad
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fegchbad
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fecghbad
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fegcbhad
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fecgbhad
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); // feghbcad
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fegbhcad
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fegbchad
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); // feghbacd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); // fegbhacd
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v6, v7); // fegcbahd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); // fegcbadh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v6, v7); // fecgbahd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); // fecgbadh
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fegbcahd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // fegbcadh
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fegbahcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fegbachd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fegbacdh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v2, v0)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // febghcad
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // febgchad
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v6, v7); // febgcahd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); // febgcadh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v0)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // febghacd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // febgahcd
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // febgachd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // febgacdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fecbghad
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fecbgahd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v6); // fecbgadh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // febcghad
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // febcgahd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // febcgadh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fecbaghd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fecbagdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v3, v4); swap(v4, v5); // fecbadgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // febcaghd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // febcagdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // febcadgh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // febaghcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // febagchd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // febagcdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // febacghd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // febacgdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v4); swap(v3, v5); // febacdgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v5, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fbeghcad
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fbegchad
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v6, v7); // fbegcahd
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); // fbegcadh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbecghad
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fbecgahd
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // fbecgadh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbecaghd
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fbecagdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // fbecadgh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bfeghcad
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bfegchad
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v6, v7); // bfegcahd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); // bfegcadh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfecghad
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bfecgahd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // bfecgadh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfecaghd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bfecagdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // bfecadgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fbeghacd
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fbegahcd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fbegachd
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // fbegacdh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fbeaghcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbeagchd
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fbeagcdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fbeacghd
																			}
																			else {
																				swap(v0, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // fbeacgdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v4); swap(v3, v5); // fbeacdgh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bfeghacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bfegahcd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bfegachd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // bfegacdh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bfeaghcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfeagchd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bfeagcdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bfeacghd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // bfeacgdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v4); swap(v3, v5); // bfeacdgh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v2, v1)) {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																if (compare(v7, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v7); // gfcehbad
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fgcehbad
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fcgehbad
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gfcebhad
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgcebhad
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcgebhad
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfcebahd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fgcebahd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fcgebahd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gfcebadh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v6); // fgcebadh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // fcgebadh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fceghbad
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fcegbhad
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v6, v7); // fcegbahd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); // fcegbadh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcebghad
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fcebgahd
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // fcebgadh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcebaghd
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fcebagdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // fcebadgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v5, v7); // gfcbehad
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fgcbehad
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fcgbehad
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gfcbeahd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v6, v7); // fgcbeahd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v6, v7); // fcgbeahd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v5, v6); // gfcbeadh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v6); // fgcbeadh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); // fcgbeadh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fcbgehad
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcbeghad
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v6, v7); // fcbgeahd
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fcbegahd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcbeaghd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v6); // fcbgeadh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v6); // fcbegadh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fcbeagdh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v4); swap(v4, v5); // fcbeadgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); // gfbcehad
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fgbcehad
																		}
																	}
																	else {
																		swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fbgcehad
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gfbceahd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v6, v7); // fgbceahd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v6, v7); // fbgceahd
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // gfbceadh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); // fgbceadh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); // fbgceadh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fbcgehad
																		}
																		else {
																			swap(v0, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbceghad
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v3, v6); swap(v6, v7); // fbcgeahd
																			}
																			else {
																				swap(v0, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fbcegahd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbceaghd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v3, v6); // fbcgeadh
																		}
																		else {
																			swap(v0, v5); swap(v3, v4); swap(v4, v6); // fbcegadh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fbceagdh
																		}
																		else {
																			swap(v0, v5); swap(v3, v4); swap(v4, v5); // fbceadgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); // gbfcehad
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); // bgfcehad
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bfgcehad
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gbfceahd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bgfceahd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v6, v7); // bfgceahd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // gbfceadh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // bgfceadh
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); // bfgceadh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bfcgehad
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfceghad
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v6, v7); // bfcgeahd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bfcegahd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfceaghd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); // bfcgeadh
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v6); // bfcegadh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bfceagdh
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v4); swap(v4, v5); // bfceadgh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v4, v1)) {
															if (compare(v6, v4)) {
																if (compare(v7, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); // gcfehbad
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cgfehbad
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); // cfgehbad
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gcfebhad
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgfebhad
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cfgebhad
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcfebahd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cgfebahd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfgebahd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gcfebadh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // cgfebadh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cfgebadh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cfeghbad
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfegbhad
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cfegbahd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // cfegbadh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cfebghad
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfebgahd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cfebgadh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cfebaghd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // cfebagdh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v5); // cfebadgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); // gcfbehad
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cgfbehad
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); // cfgbehad
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gcfbeahd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v6, v7); // cgfbeahd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cfgbeahd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // gcfbeadh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); // cgfbeadh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // cfgbeadh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cfbgehad
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfbeghad
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v6, v7); // cfbgeahd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cfbegahd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfbeaghd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); // cfbgeadh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // cfbegadh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cfbeagdh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cfbeadgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v6, v5)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v5, v7); // gcbfehad
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); // cgbfehad
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); // cbgfehad
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gcbfeahd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cgbfeahd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cbgfeahd
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v5, v6); // gcbfeadh
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // cgbfeadh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // cbgfeadh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cbfgehad
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbfeghad
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v6, v7); // cbfgeahd
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cbfegahd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbfeaghd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v6); // cbfgeadh
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // cbfegadh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cbfeagdh
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cbfeadgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		swap(v0, v6); swap(v3, v5); swap(v5, v7); // gbcfehad
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v5, v7); // bgcfehad
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); // bcgfehad
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gbcfeahd
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bgcfeahd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bcgfeahd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v3, v5); swap(v5, v6); // gbcfeadh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v5, v6); // bgcfeadh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // bcgfeadh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bcfgehad
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcfeghad
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v6, v7); // bcfgeahd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bcfegahd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcfeaghd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); // bcfgeadh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // bcfegadh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bcfeagdh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bcfeadgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v0)) {
												if (compare(v6, v1)) {
													if (compare(v5, v1)) {
														if (compare(v2, v1)) {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfcbaehd
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gfcbaedh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v3, v5); swap(v4, v6); // gfcbadeh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fgcbaehd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v5); // fgcbaedh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fgcbadeh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fcgbaehd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // fcgbaedh
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fcgbadeh
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcfbaehd
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gcfbaedh
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // gcfbadeh
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cgfbaehd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // cgfbaedh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cgfbadeh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfgbaehd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cfgbaedh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // cfgbadeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gfbaehcd
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v6, v7); // gfbaechd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfbcaehd
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gfbacehd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gfbcaedh
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // gfbcadeh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); // gfbaecdh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // gfbacedh
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gfbacdeh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); // fgbaehcd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fgbaechd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fgbcaehd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgbacehd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // fgbcaedh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fgbcadeh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // fgbaecdh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fgbacedh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // fgbacdeh
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v5, v2)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gbfaehcd
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v6, v7); // gbfaechd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbfcaehd
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gbfacehd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gbfcaedh
																		}
																		else {
																			swap(v0, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // gbfcadeh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v5); swap(v3, v6); // gbfaecdh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // gbfacedh
																			}
																			else {
																				swap(v0, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gbfacdeh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcbfaehd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gcbfaedh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v5); swap(v4, v6); // gcbfadeh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgbfaehd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cgbfaedh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // cgbfadeh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbcfaehd
																		}
																		else {
																			swap(v0, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gbcfaedh
																		}
																	}
																	else {
																		swap(v0, v6); swap(v3, v5); swap(v4, v6); // gbcfadeh
																	}
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gcbafehd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cgbafehd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gbcafehd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v5); // gcbafedh
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // cgbafedh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v6); swap(v4, v5); // gbcafedh
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gcbafdeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cgbafdeh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gbcafdeh
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v6); // gcbadfeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // cgbadfeh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v6); swap(v4, v6); // gbcadfeh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); // gbafehcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gbafechd
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // gbafecdh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbafcehd
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gbafcedh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // gbafcdeh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gbacfehd
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // gbacfedh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gbacfdeh
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // gbacdfeh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v5, v0)) {
															if (compare(v5, v2)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bgfaehcd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v6, v7); // bgfaechd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgfcaehd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bgfacehd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bgfcaedh
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // bgfcadeh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); // bgfaecdh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // bgfacedh
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bgfacdeh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbgfaehd
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cbgfaedh
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // cbgfadeh
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgcfaehd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bgcfaedh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v5); swap(v4, v6); // bgcfadeh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcgfaehd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bcgfaedh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // bcgfadeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cbgafehd
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bgcafehd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bcgafehd
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // cbgafedh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v5); // bgcafedh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // bcgafedh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cbgafdeh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bgcafdeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bcgafdeh
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // cbgadfeh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v6); // bgcadfeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // bcgadfeh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); // bgafehcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bgafechd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // bgafecdh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgafcehd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bgafcedh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // bgafcdeh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bgacfehd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // bgacfedh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bgacfdeh
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // bgacdfeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v2)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); // fbgaehcd
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fbgaechd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fbgcaehd
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbgacehd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // fbgcaedh
																		}
																		else {
																			swap(v0, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fbgcadeh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // fbgaecdh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fbgacedh
																			}
																			else {
																				swap(v0, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // fbgacdeh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fcbgaehd
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v5); // fcbgaedh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fcbgadeh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cfbgaehd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // cfbgaedh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cfbgadeh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fbcgaehd
																		}
																		else {
																			swap(v0, v5); swap(v3, v6); swap(v4, v5); // fbcgaedh
																		}
																	}
																	else {
																		swap(v0, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fbcgadeh
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); // bfgaehcd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bfgaechd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bfgcaehd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfgacehd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // bfgcaedh
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bfgcadeh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // bfgaecdh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bfgacedh
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // bfgacdeh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cbfgaehd
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // cbfgaedh
																		}
																	}
																	else {
																		swap(v0, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cbfgadeh
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bfcgaehd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v5); // bfcgaedh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bfcgadeh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bcfgaehd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // bcfgaedh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bcfgadeh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v0)) {
													if (compare(v2, v1)) {
														if (compare(v4, v3)) {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcbagehd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fcbaeghd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fcbagedh
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v5, v6); // fcbaegdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v3, v5); // fcbaedgh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfbagehd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cfbaeghd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cfbagedh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // cfbaegdh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); // cfbaedgh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v4)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbfagehd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cbagfehd
																			}
																			else {
																				swap(v0, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbafgehd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cbfagedh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v3, v6); swap(v4, v5); // cbagfedh
																			}
																			else {
																				swap(v0, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cbafgedh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cbfaeghd
																			}
																			else {
																				swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // cbfaegdh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); // cbfaedgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cbafeghd
																			}
																			else {
																				swap(v0, v2); swap(v3, v5); swap(v5, v6); // cbafegdh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v3, v5); // cbafedgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v6); // fcbagdeh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fcbadgeh
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v3, v5); swap(v4, v5); // fcbadegh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // cfbagdeh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cfbadgeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cfbadegh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // cbfagdeh
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cbfadgeh
																		}
																		else {
																			swap(v0, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cbfadegh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cbagfdeh
																		}
																		else {
																			swap(v0, v2); swap(v3, v5); swap(v4, v6); // cbafgdeh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cbafdgeh
																		}
																		else {
																			swap(v0, v2); swap(v3, v5); swap(v4, v5); // cbafdegh
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v3, v6); swap(v4, v6); // cbagdfeh
																		}
																		else {
																			swap(v0, v2); swap(v4, v6); // cbadgfeh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v4, v5); swap(v5, v6); // cbadfgeh
																		}
																		else {
																			swap(v0, v2); swap(v4, v5); // cbadfegh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v2)) {
																if (compare(v5, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbcagehd
																		}
																		else {
																			swap(v0, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fbcaeghd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fbcagedh
																			}
																			else {
																				swap(v0, v5); swap(v3, v5); swap(v5, v6); // fbcaegdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v3, v5); // fbcaedgh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfcagehd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bfcaeghd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bfcagedh
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v5, v6); // bfcaegdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v3, v5); // bfcaedgh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v4)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcfagehd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bcagfehd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcafgehd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bcfagedh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v5); // bcagfedh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bcafgedh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bcfaeghd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // bcfaegdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); // bcfaedgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bcafeghd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v5, v6); // bcafegdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v5); // bcafedgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v3)) {
																		swap(v0, v5); swap(v3, v5); swap(v4, v6); // fbcagdeh
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fbcadgeh
																		}
																		else {
																			swap(v0, v5); swap(v3, v5); swap(v4, v5); // fbcadegh
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v6); // bfcagdeh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bfcadgeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v3, v5); swap(v4, v5); // bfcadegh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // bcfagdeh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bcfadgeh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bcfadegh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bcagfdeh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v6); // bcafgdeh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bcafdgeh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v5); swap(v4, v5); // bcafdegh
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v6); // bcagdfeh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v4, v6); // bcadgfeh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v2); swap(v4, v5); swap(v5, v6); // bcadfgeh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v4, v5); // bcadfegh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fbagehcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v6, v7); // fbagechd
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v6); // fbagecdh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fbagcehd
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // fbagcedh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fbagcdeh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fbacgehd
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fbacgedh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // fbacgdeh
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fbacdgeh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fbaeghcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fbaegchd
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // fbaegcdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fbaecghd
																			}
																			else {
																				swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fbaecgdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fbaecdgh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fbaceghd
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // fbacegdh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); // fbacedgh
																		}
																		else {
																			swap(v0, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fbacdegh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bfagehcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v6, v7); // bfagechd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); // bfagecdh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bfagcehd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // bfagcedh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bfagcdeh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bfacgehd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bfacgedh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // bfacgdeh
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bfacdgeh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bfaeghcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bfaegchd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // bfaegcdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bfaecghd
																			}
																			else {
																				swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bfaecgdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bfaecdgh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bfaceghd
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // bfacegdh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); // bfacedgh
																		}
																		else {
																			swap(v0, v1); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bfacdegh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v5, v7); // bagfehcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bagfechd
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v5, v6); // bagfecdh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bafgehcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v6, v7); // bafgechd
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v6); // bafgecdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bafeghcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bafegchd
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v6); // bafegcdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bafecghd
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bafecgdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v4); swap(v4, v5); // bafecdgh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bagfcehd
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bagfcedh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v3, v5); swap(v4, v6); // bagfcdeh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bafgcehd
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v5); // bafgcedh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bafgcdeh
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bafcgehd
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bafcgedh
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v6); // bafcgdeh
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bafcdgeh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bafceghd
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v5, v6); // bafcegdh
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); // bafcedgh
																			}
																			else {
																				swap(v0, v1); swap(v2, v5); swap(v3, v5); swap(v4, v5); // bafcdegh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bagcfehd
																		}
																		else {
																			swap(v0, v1); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bacgfehd
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bacfgehd
																		}
																		else {
																			swap(v0, v1); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bacfeghd
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v5); // bagcfedh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v3, v6); swap(v4, v5); // bacgfedh
																			}
																			else {
																				swap(v0, v1); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bacfgedh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v3, v5); swap(v5, v6); // bacfegdh
																		}
																		else {
																			swap(v0, v1); swap(v3, v5); // bacfedgh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bagcfdeh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v3, v6); swap(v4, v5); swap(v5, v6); // bacgfdeh
																			}
																			else {
																				swap(v0, v1); swap(v3, v5); swap(v4, v6); // bacfgdeh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v3, v5); swap(v4, v5); swap(v5, v6); // bacfdgeh
																		}
																		else {
																			swap(v0, v1); swap(v3, v5); swap(v4, v5); // bacfdegh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v6); // bagcdfeh
																		}
																		else {
																			swap(v0, v1); swap(v3, v6); swap(v4, v6); // bacgdfeh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v4, v6); // bacdgfeh
																			}
																			else {
																				swap(v0, v1); swap(v4, v5); swap(v5, v6); // bacdfgeh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v4, v5); // bacdfegh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v7, v6)) {
									if (compare(v7, v4)) {
										if (compare(v6, v0)) {
											if (compare(v6, v2)) {
												if (compare(v6, v1)) {
													if (compare(v5, v0)) {
														if (compare(v5, v1)) {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v4)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hgefcbad
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hegfcbad
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hefgcbad
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgecfbad
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hegcfbad
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgcefbad
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v6, v4)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgefbcad
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hegfbcad
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v6, v7); // hefgbcad
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hgefbacd
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); // hegfbacd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hefgbacd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v2, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hgebfcad
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hegbfcad
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgbefcad
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgebfacd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hegbfacd
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hgbefacd
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hgecbfad
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v6, v7); // hegcbfad
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hgcebfad
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v6, v7); // hgcbefad
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v6, v7); // hgebcfad
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hegbcfad
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hgbecfad
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v6, v7); // hgbcefad
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v2, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgecbafd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hegcbafd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgcebafd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hgcbeafd
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgcbaefd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hgebcafd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hegbcafd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgbecafd
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hgbceafd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgbcaefd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hgebafcd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); // hegbafcd
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgebacfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hegbacfd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); // hgbeafcd
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hgbeacfd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); // hgbaefcd
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); // hgbaecfd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hgbacefd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v2, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hgecbadf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); // hegcbadf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); // hgebcadf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hegbcadf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hgbecadf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hgebacdf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hegbacdf
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgbeacdf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hgbaecdf
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hgcebadf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v5, v7); // hgcbeadf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hgcbaedf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgcbadef
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); // hgbceadf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hgbcaedf
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgbcadef
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgbacedf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hgbacdef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v0)) {
														if (compare(v5, v0)) {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hebgfcad
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbgefcad
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hbegfcad
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhgefcad
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bhegfcad
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hefbgcad
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hebfgcad
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbefgcad
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhefgcad
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v4)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hbgecfad
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v6, v7); // hbgcefad
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v6, v7); // hebgcfad
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v6, v7); // hbegcfad
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bhgecfad
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v6, v7); // bhegcfad
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v6, v7); // bhgcefad
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hebgcafd
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbgecafd
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hbegcafd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhgecafd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bhegcafd
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); // hebgcadf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hbgecadf
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); // hbegcadf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bhgecadf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); // bhegcadf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hbgceafd
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); // hbgceadf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbgcaefd
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hbgcaedf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbgcadef
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bhgceafd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); // bhgceadf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhgcaefd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bhgcaedf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhgcadef
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hebgfacd
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hbgefacd
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbegfacd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bhgefacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhegfacd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hefbgacd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hebfgacd
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hbefgacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bhefgacd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v4)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); // hbgeafcd
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v7); // hbgaefcd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hebgafcd
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hbegafcd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); // bhgeafcd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bhegafcd
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); // bhgaefcd
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hebgacfd
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hbgeacfd
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbegacfd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bhgeacfd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhegacfd
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hebgacdf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbgeacdf
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hbegacdf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhgeacdf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bhegacdf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); // hbgaecfd
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hbgaecdf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hbgacefd
																			}
																			else {
																				swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbgacedf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hbgacdef
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); // bhgaecfd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bhgaecdf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bhgacefd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhgacedf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bhgacdef
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v4)) {
													if (compare(v7, v2)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcgefbad
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hcgebfad
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v6, v7); // hcgbefad
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v6, v7); // hcbgefad
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v3, v6); swap(v6, v7); // hbcgefad
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v6, v7); // bhcgefad
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcgebafd
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hcgbeafd
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hcbgeafd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hbcgeafd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bhcgeafd
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcgbaefd
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcbgaefd
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbcgaefd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhcgaefd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hcgebadf
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); // hcgbeadf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v5, v7); // hcbgeadf
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hcgbaedf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcgbadef
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hcbgaedf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcbgadef
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v0)) {
																		swap(v0, v7); swap(v3, v6); swap(v5, v7); // hbcgeadf
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hbcgaedf
																		}
																		else {
																			swap(v0, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbcgadef
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v5, v7); // bhcgeadf
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bhcgaedf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhcgadef
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // chgefbad
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); // chgebfad
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); // chgbefad
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v6, v7); // chbgefad
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v6, v7); // cbhgefad
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v6, v7); // bchgefad
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // chgebafd
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); // chgbeafd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // chbgeafd
																			}
																			else {
																				swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cbhgeafd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bchgeafd
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // chgbaefd
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chbgaefd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbhgaefd
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bchgaefd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chgebadf
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // chgbeadf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); // chbgeadf
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chgbaedf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // chgbadef
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // chbgaedf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chbgadef
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v0)) {
																		swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); // cbhgeadf
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cbhgaedf
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbhgadef
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); // bchgeadf
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bchgaedf
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bchgadef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v4, v2)) {
															if (compare(v6, v5)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hecgfbad
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v6, v7); // hecgbfad
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hecbgfad
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hebcgfad
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hecgbafd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hecbgafd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hebcgafd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v5, v7); // hecgbadf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hecbgadf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hebcgadf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hefcgbad
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hefcbgad
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hefbcgad
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hebfcgad
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hecfgbad
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hecfbgad
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hecbfgad
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hebcfgad
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hcegfbad
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcefgbad
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v6, v7); // hcegbfad
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hcegbafd
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); // hcegbadf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chegfbad
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // chefgbad
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // chegbfad
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chegbafd
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // chegbadf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hcefbgad
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hcebgfad
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcebfgad
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcebgafd
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hcebgadf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chefbgad
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); // chebgfad
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // chebfgad
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // chebgafd
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chebgadf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hcbegfad
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcbefgad
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // chbegfad
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chbefgad
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hbefcgad
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hbecgfad
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbecfgad
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hbcegfad
																		}
																		else {
																			swap(v0, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbcefgad
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcbegafd
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hcbegadf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chbegafd
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // chbegadf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbecgafd
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hbecgadf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbcegafd
																		}
																		else {
																			swap(v0, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hbcegadf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bhefcgad
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bhecgfad
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhecfgad
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bhcegfad
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhcefgad
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cbhegfad
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbhefgad
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bchegfad
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bchefgad
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhecgafd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bhecgadf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhcegafd
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bhcegadf
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbhegafd
																		}
																		else {
																			swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cbhegadf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bchegafd
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bchegadf
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v1)) {
												if (compare(v4, v2)) {
													if (compare(v4, v1)) {
														if (compare(v2, v0)) {
															if (compare(v5, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); // hefcbagd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v5, v7); // hecfbagd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hecbfagd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hefcbadg
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hecfbadg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hecbfadg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hefbcagd
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); // hebfcagd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hebcfagd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hefbcadg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hebfcadg
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hebcfadg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hecbagfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v7); // hecbafgd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hecbafdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hecbagdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hecbadgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hecbadfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hebcagfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); // hebcafgd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hebcafdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hebcagdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hebcadgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hebcadfg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hefbagcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); // hefbacgd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hefbacdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hebfagcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hebfacgd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hebfacdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); // hebagfcd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hebafgcd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); // hebafcgd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hebafcdg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hebagcfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); // hebacgfd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); // hebacfgd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v6, v7); // hebacfdg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hebagcdf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hebacgdf
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); // hebacdgf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hebacdfg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v4, v0)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); // hbeagfcd
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hbeagcfd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hbecagfd
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); // hbeacgfd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbecagdf
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hbecadgf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbeagcdf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hbeacgdf
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); // hbeacdgf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v6, v7); // hbagefcd
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hbaegfcd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hbagecfd
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hbaegcfd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hbaecgfd
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); // hbagecdf
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hbaegcdf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbaecgdf
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hbaecdgf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v2, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); // hbefcagd
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hbefcadg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbefagcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hbefacgd
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbefacdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hbeafgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); // hbeafcgd
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hbeafcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hbaefgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hbaefcgd
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbaefcdg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hbecfagd
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbecfadg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); // hbecafgd
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hbecafdg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbecadfg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v4); swap(v3, v7); // hbeacfgd
																			}
																			else {
																				swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v6, v7); // hbeacfdg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hbeacdfg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); // hbaecfgd
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hbaecfdg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbaecdfg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v4, v0)) {
															if (compare(v5, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); // hcefbagd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // chefbagd
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hcebfagd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); // chebfagd
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hcbefagd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // chbefagd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hbcefagd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hcefbadg
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chefbadg
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcebfadg
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // chebfadg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcbefadg
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chbefadg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hbcefadg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hcebagfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); // hcebafgd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hcebafdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); // chebagfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); // chebafgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v6, v7); // chebafdg
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v6, v3)) {
																				if (compare(v6, v5)) {
																					swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hcbeagfd
																				}
																				else {
																					swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v7); // hcbeafgd
																				}
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hcbeafdg
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				if (compare(v6, v5)) {
																					swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // chbeagfd
																				}
																				else {
																					swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); // chbeafgd
																				}
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // chbeafdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hbceagfd
																			}
																			else {
																				swap(v0, v7); swap(v3, v4); swap(v4, v7); // hbceafgd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hbceafdg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hcbagefd
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v5, v6); // hcbaegfd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v7); // hcbaefgd
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v6, v7); // hcbaefdg
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // chbagefd
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); // chbaegfd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); // chbaefgd
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v6, v7); // chbaefdg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hbcagefd
																		}
																		else {
																			swap(v0, v7); swap(v3, v7); swap(v5, v6); // hbcaegfd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v3, v7); // hbcaefgd
																		}
																		else {
																			swap(v0, v7); swap(v3, v7); swap(v6, v7); // hbcaefdg
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbagcefd
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hbacgefd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); // hbacegfd
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v7); // hbacefgd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v6, v7); // hbacefdg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v7, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcebagdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hcebadgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcebadfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcbeagdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hcbeadgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcbeadfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcbagedf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hcbaegdf
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v5, v7); // hcbaedgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hcbaedfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hcbagdef
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcbadgef
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hcbadegf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcbadefg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // chebagdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); // chebadgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // chebadfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chbeagdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // chbeadgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chbeadfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chbagedf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // chbaegdf
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); // chbaedgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // chbaedfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // chbagdef
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chbadgef
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // chbadegf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chbadefg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbceagdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hbceadgf
																			}
																			else {
																				swap(v0, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbceadfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbcagedf
																			}
																			else {
																				swap(v0, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hbcaegdf
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v3, v7); swap(v5, v7); // hbcaedgf
																			}
																			else {
																				swap(v0, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hbcaedfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hbcagdef
																		}
																		else {
																			swap(v0, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbcadgef
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hbcadegf
																		}
																		else {
																			swap(v0, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbcadefg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hbagcedf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hbacgedf
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hbacegdf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); // hbacedgf
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hbacedfg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbagcdef
																		}
																		else {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hbacgdef
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hbacdgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hbacdegf
																			}
																			else {
																				swap(v0, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hbacdefg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v0)) {
													if (compare(v6, v3)) {
														if (compare(v5, v3)) {
															if (compare(v6, v5)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cbheagfd
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cbhagefd
																			}
																			else {
																				swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); // cbhaegfd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cbahgefd
																		}
																		else {
																			swap(v0, v2); swap(v3, v7); swap(v5, v6); // cbahegfd
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bhecagfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bhceagfd
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bhcagefd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v5, v6); // bhcaegfd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v4, v0)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bcheagfd
																			}
																			else {
																				if (compare(v6, v4)) {
																					swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bchagefd
																				}
																				else {
																					swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); // bchaegfd
																				}
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bcahgefd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v5, v6); // bcahegfd
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); // bhefcagd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bhecfagd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bhcefagd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); // bhecafgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v7); // bhceafgd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v7); // bhcaefgd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cbhefagd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bchefagd
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); // cbheafgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); // bcheafgd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v7); // cbhaefgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); // bchaefgd
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v3, v7); // cbahefgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v7); // bcahefgd
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbhagedf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhcagedf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bchagedf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbahgedf
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcahgedf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cbhagdef
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bhcagdef
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bchagdef
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cbahgdef
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bcahgdef
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhecagdf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhceagdf
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbheagdf
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcheagdf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cbhaegdf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bhcaegdf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bchaegdf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cbahegdf
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bcahegdf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v7, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bhecadgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bhceadgf
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cbheadgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bcheadgf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); // cbhaedgf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v5, v7); // bhcaedgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); // bchaedgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbhadgef
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhcadgef
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bchadgef
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cbhadegf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bhcadegf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bchadegf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v3, v7); swap(v5, v7); // cbahedgf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbahdgef
																			}
																			else {
																				swap(v0, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cbahdegf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbadhgef
																		}
																		else {
																			swap(v0, v2); swap(v4, v7); swap(v5, v7); // cbadhegf
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v5, v7); // bcahedgf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcahdgef
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bcahdegf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v2); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcadhgef
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v4, v7); swap(v5, v7); // bcadhegf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bhefcadg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhecfadg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhcefadg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bhecafdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bhceafdg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v6, v7); // bhcaefdg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbhefadg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bchefadg
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cbheafdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bcheafdg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v6, v7); // cbhaefdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v6, v7); // bchaefdg
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v3, v7); swap(v6, v7); // cbahefdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v6, v7); // bcahefdg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhecadfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhceadfg
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbheadfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcheadfg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cbhaedfg
																			}
																			else {
																				if (compare(v7, v2)) {
																					swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bhcaedfg
																				}
																				else {
																					swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bchaedfg
																				}
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cbahedfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bcahedfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbhadefg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhcadefg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bchadefg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbahdefg
																			}
																			else {
																				swap(v0, v2); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbadhefg
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcahdefg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcadhefg
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v0)) {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhefagcd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); // bheagfcd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bheafgcd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v6, v7); // bhagefcd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bhaegfcd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bhaefgcd
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bhefacgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); // bheafcgd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bhaefcgd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhefacdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bheafcdg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bhaefcdg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bhagecfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhagcefd
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bheagcfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bhaegcfd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v0)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); // bheacgfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bhaecgfd
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bhacgefd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); // bhacegfd
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); // bheacfgd
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); // bhaecfgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); // bhacefgd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v6, v7); // bheacfdg
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bhaecfdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v6, v7); // bhacefdg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bheagcdf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bheacgdf
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); // bheacdgf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bheacdfg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); // bhagecdf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bhaegcdf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhaecgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bhaecdgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhaecdfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bhagcedf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bhacgedf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bhacegdf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); // bhacedgf
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bhacedfg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhagcdef
																		}
																		else {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bhacgdef
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bhacdgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bhacdegf
																			}
																			else {
																				swap(v0, v1); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bhacdefg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v6, v7); // bahgefcd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bahegfcd
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bahefgcd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bahefcgd
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bahefcdg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bahgecfd
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bahegcfd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bahecgfd
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v7); // bahecfgd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bahecfdg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bahgcefd
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bahcgefd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v5, v6); // bahcegfd
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v7); // bahcefgd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v6, v7); // bahcefdg
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v1); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bachgefd
																		}
																		else {
																			swap(v0, v1); swap(v3, v7); swap(v5, v6); // bachegfd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v3, v7); // bachefgd
																		}
																		else {
																			swap(v0, v1); swap(v3, v7); swap(v6, v7); // bachefdg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v5, v7); // bahgecdf
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bahgcedf
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bahcgedf
																			}
																			else {
																				swap(v0, v1); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bachgedf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bahegcdf
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bahecgdf
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bahcegdf
																			}
																			else {
																				swap(v0, v1); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bachegdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bahecdgf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v5, v7); // bahcedgf
																			}
																			else {
																				swap(v0, v1); swap(v3, v7); swap(v5, v7); // bachedgf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bahecdfg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bahcedfg
																			}
																			else {
																				swap(v0, v1); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bachedfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v4)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bahgcdef
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bahcgdef
																			}
																			else {
																				swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bahcdgef
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bachgdef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bachdgef
																			}
																			else {
																				swap(v0, v1); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bacdhgef
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bahcdegf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v3, v7); swap(v4, v7); swap(v5, v7); // bachdegf
																			}
																			else {
																				swap(v0, v1); swap(v4, v7); swap(v5, v7); // bacdhegf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bahcdefg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bachdefg
																			}
																			else {
																				swap(v0, v1); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bacdhefg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v5, v0)) {
											if (compare(v5, v2)) {
												if (compare(v5, v1)) {
													if (compare(v6, v0)) {
														if (compare(v6, v1)) {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ehgfcbad
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehfgcbad
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // efhgcbad
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ehfcgbad
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); // efhcgbad
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v6); // efchgbad
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v6, v5)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehgfbcad
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ehfgbcad
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efhgbcad
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ehgfbacd
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehfgbacd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // efhgbacd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v2, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); // ehfbgcad
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // efhbgcad
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); // efbhgcad
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ehfbgacd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efhbgacd
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // efbhgacd
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehfcbgad
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // efhcbgad
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // efchbgad
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efcbhgad
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ehfbcgad
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efhbcgad
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // efbhcgad
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efbchgad
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v2, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ehfcbagd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); // efhcbagd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v5); // efchbagd
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // efcbhagd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v5, v7); // efcbahgd
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); // ehfbcagd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // efhbcagd
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); // efbhcagd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // efbchagd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); // efbcahgd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); // ehfbagcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // efhbagcd
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); // ehfbacgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); // efhbacgd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); // efbhagcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); // efbhacgd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // efbahgcd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); // efbahcgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // efbachgd
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v2, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehfcbadg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // efhcbadg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ehfbcadg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efhbcadg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // efbhcadg
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v6, v7); // ehfbacdg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // efhbacdg
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v6, v7); // efbhacdg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // efbahcdg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // efchbadg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efcbhadg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // efcbahdg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v6, v7); // efcbadhg
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efbchadg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // efbcahdg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v6, v7); // efbcadhg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efbachdg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // efbacdhg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v0)) {
														if (compare(v6, v0)) {
															if (compare(v6, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ehgbfcad
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehbgfcad
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebhgfcad
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // behgfcad
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ehbfgcad
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ebhfgcad
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); // ebfhgcad
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // behfgcad
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); // befhgcad
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehbfcgad
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebhfcgad
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ebfhcgad
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebfchgad
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // behfcgad
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // befhcgad
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // befchgad
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ehbfcagd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ebhfcagd
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); // ebfhcagd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // behfcagd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); // befhcagd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehbfcadg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebhfcadg
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ebfhcadg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // behfcadg
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // befhcadg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ebfchagd
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebfchadg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); // ebfcahgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ebfcahdg
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v6, v7); // ebfcadhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // befchagd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // befchadg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); // befcahgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // befcahdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v6, v7); // befcadhg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v6, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); // ehgbfacd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ehbgfacd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ebhgfacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // behgfacd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehbfgacd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebhfgacd
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ebfhgacd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // behfgacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // befhgacd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ehbfagcd
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ebhfagcd
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); // ebfhagcd
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ebfahgcd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // behfagcd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); // befhagcd
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // befahgcd
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); // ehbfacgd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); // ebhfacgd
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); // ebfhacgd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); // behfacgd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); // befhacgd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ehbfacdg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ebhfacdg
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v7); swap(v6, v7); // ebfhacdg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // behfacdg
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v6, v7); // befhacdg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); // ebfahcgd
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ebfahcdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ebfachgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebfachdg
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ebfacdhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); // befahcgd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // befahcdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // befachgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // befachdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // befacdhg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v5)) {
													if (compare(v7, v1)) {
														if (compare(v7, v2)) {
															if (compare(v6, v5)) {
																if (compare(v6, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehgcfbad
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ehgcbfad
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); // ehgbcfad
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ehbgcfad
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehcgfbad
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ehcgbfad
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v6); // ehcbgfad
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); // ehbcgfad
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ehcfgbad
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehcfbgad
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ehcbfgad
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ehbcfgad
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v1)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ehcfbagd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v5); // ehcbfagd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); // ehbcfagd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v5, v1)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehcfbadg
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ehcbfadg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ehbcfadg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // echgfbad
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // echfgbad
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // echfbgad
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // echfbagd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // echfbadg
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cehgfbad
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cehfgbad
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cehfbgad
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); // cehfbagd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cehfbadg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // echgbfad
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); // echbgfad
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // echbfgad
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); // echbfagd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // echbfadg
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v6, v7); // cehgbfad
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cehbgfad
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cehbfgad
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cehbfagd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cehbfadg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v6); // ecbhgfad
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ecbhfgad
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); // cebhgfad
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cebhfgad
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ebhgcfad
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); // ebhcgfad
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ebhcfgad
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v3, v7); swap(v4, v6); // ebchgfad
																		}
																		else {
																			swap(v0, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ebchfgad
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v5); // ecbhfagd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ecbhfadg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); // cebhfagd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cebhfadg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); // ebhcfagd
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ebhcfadg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v3, v7); swap(v4, v5); // ebchfagd
																		}
																		else {
																			swap(v0, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ebchfadg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // behgcfad
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); // behcgfad
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // behcfgad
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v6); // bechgfad
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bechfgad
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); // cbehgfad
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cbehfgad
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); // bcehgfad
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bcehfgad
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); // behcfagd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // behcfadg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v5); // bechfagd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bechfadg
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); // cbehfagd
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cbehfadg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); // bcehfagd
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // bcehfadg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v6, v3)) {
															if (compare(v6, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v1)) {
																		swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); // ecfhgbad
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ecfhbgad
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecfbhgad
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecbfhgad
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebcfhgad
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // becfhgad
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); // ecfhbagd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ecfbhagd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ecbfhagd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ebcfhagd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // becfhagd
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); // ecfbahgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v5, v7); // ecbfahgd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v3, v5); swap(v5, v7); // ebcfahgd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v5, v7); // becfahgd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ecfhbadg
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecfbhadg
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecbfhadg
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ecfbahdg
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v6, v7); // ecfbadhg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ecbfahdg
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v6, v7); // ecbfadhg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v0)) {
																		swap(v0, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebcfhadg
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ebcfahdg
																		}
																		else {
																			swap(v0, v4); swap(v3, v5); swap(v6, v7); // ebcfadhg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // becfhadg
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // becfahdg
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v6, v7); // becfadhg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v6, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v1)) {
																		swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cefhgbad
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); // cefhbgad
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cefbhgad
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cebfhgad
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbefhgad
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcefhgad
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); // cefhbagd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); // cefbhagd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cebfhagd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cbefhagd
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bcefhagd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cefbahgd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); // cebfahgd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); // cbefahgd
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); // bcefahgd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v6, v7); // cefhbadg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cefbhadg
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cebfhadg
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cefbahdg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cefbadhg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cebfahdg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v6, v7); // cebfadhg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v0)) {
																		swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbefhadg
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cbefahdg
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v6, v7); // cbefadhg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcefhadg
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // bcefahdg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v6, v7); // bcefadhg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v1)) {
												if (compare(v7, v2)) {
													if (compare(v7, v1)) {
														if (compare(v2, v0)) {
															if (compare(v6, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehgcbafd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehcgbafd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ehcbgafd
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ehgcbadf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ehcgbadf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehcbgadf
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ehgbcafd
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehbgcafd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ehbcgafd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehgbcadf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ehbgcadf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehbcgadf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v5, v6); // ehcbagfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v3, v7); // ehcbafgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v6, v7); // ehcbafdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ehcbagdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v5, v7); // ehcbadgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ehcbadfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); // ehbcagfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); // ehbcafgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v6, v7); // ehbcafdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ehbcagdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); // ehbcadgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ehbcadfg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); // ehgbafcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); // ehgbacfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ehgbacdf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v6, v7); // ehbgafcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ehbgacfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); // ehbgacdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ehbagfcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehbafgcd
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehbagcfd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ehbagcdf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ehbacgfd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehbacgdf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ehbafcgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); // ehbacfgd
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehbafcdg
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ehbacfdg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ehbacdgf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehbacdfg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v6, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v2, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebhgcafd
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ebhgcadf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v6, v7); // ebhgafcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ebhgacfd
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); // ebhgacdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ebhagfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebhagcfd
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ebhagcdf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); // ebahgfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ebahgcfd
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebahgcdf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ebhcgafd
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebhcgadf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); // ebhcagfd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ebhcagdf
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); // ebhcadgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ebhacgfd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebhacgdf
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ebhacdgf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); // ebahcgfd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ebahcgdf
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); // ebahcdgf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebhafgcd
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ebhafcgd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v7); // ebhcafgd
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); // ebhacfgd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v6, v7); // ebhcafdg
																		}
																		else {
																			swap(v0, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ebhcadfg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebhafcdg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ebhacfdg
																			}
																			else {
																				swap(v0, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebhacdfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ebahfgcd
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebafhgcd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); // ebahfcgd
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ebafhcgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v7); // ebahcfgd
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ebahfcdg
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebafhcdg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v6, v7); // ebahcfdg
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ebahcdfg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v7, v0)) {
															if (compare(v7, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // echgbafd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // echbgafd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); // echbagfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); // echbafgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v6, v7); // echbafdg
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cehgbafd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cehbgafd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cehbagfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); // cehbafgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cehbafdg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ecbhgafd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cebhgafd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ebchgafd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v5, v6); // ecbhagfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); // cebhagfd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v7); swap(v5, v6); // ebchagfd
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); // ecbhafgd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); // cebhafgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v7); // ebchafgd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v6, v7); // ecbhafdg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v6, v7); // cebhafdg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v7); swap(v6, v7); // ebchafdg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ecbahgfd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v7); // ecbahfgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ecbafhgd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cebahgfd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); // cebahfgd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cebafhgd
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ebcahgfd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v3, v4); swap(v4, v7); // ebcahfgd
																			}
																			else {
																				swap(v0, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ebcafhgd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ebachgfd
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); // ebachfgd
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); // ebafchgd
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ebacfhgd
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ecbahfdg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecbafhdg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ecbafdhg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cebahfdg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cebafhdg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cebafdhg
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ebcahfdg
																			}
																			else {
																				swap(v0, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebcafhdg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ebcafdhg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ebafchdg
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v6, v7); // ebafcdhg
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				if (compare(v7, v5)) {
																					swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ebachfdg
																				}
																				else {
																					swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebacfhdg
																				}
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ebacfdhg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v0)) {
															if (compare(v7, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // echgbadf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // echbgadf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // echbagdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); // echbadgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // echbadfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); // cehgbadf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cehbgadf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cehbagdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cehbadgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cehbadfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecbhgadf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cebhgadf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebchgadf
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ecbhagdf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cebhagdf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ebchagdf
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v5, v7); // ecbhadgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); // cebhadgf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v7); swap(v5, v7); // ebchadgf
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ecbhadfg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cebhadfg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ebchadfg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecbahgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ecbahdgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecbahdfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v5, v7); // ecbadhgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ecbadhfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v6, v7); // ecbadfhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cebahgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cebahdgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cebahdfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); // cebadhgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cebadhfg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v6, v7); // cebadfhg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebcahgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ebcahdgf
																			}
																			else {
																				swap(v0, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebcahdfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v3, v4); swap(v5, v7); // ebcadhgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ebcadhfg
																			}
																			else {
																				swap(v0, v4); swap(v3, v4); swap(v6, v7); // ebcadfhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebachgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ebachdgf
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebachdfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); // ebacdhgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ebacdhfg
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v6, v7); // ebacdfhg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v0)) {
													if (compare(v5, v3)) {
														if (compare(v6, v3)) {
															if (compare(v6, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // behgcafd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // behcgafd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bechgafd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); // behcagfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v5, v6); // bechagfd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // becahgfd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cbehgafd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bcehgafd
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); // cbehagfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); // bcehagfd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cbeahgfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bceahgfd
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cbaehgfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bcaehgfd
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v7); // cbehafgd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); // cbeahfgd
																			}
																			else {
																				swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cbeafhgd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v3, v4); swap(v4, v7); // cbaehfgd
																		}
																		else {
																			swap(v0, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cbaefhgd
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); // behcafgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); // bechafgd
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v7); // becahfgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // becafhgd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); // bcehafgd
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); // bceahfgd
																				}
																				else {
																					swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bceafhgd
																				}
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v7); // bcaehfgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bcaefhgd
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v7, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v6, v7); // behcafdg
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v6, v7); // bechafdg
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v6, v7); // cbehafdg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v6, v7); // bcehafdg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cbeahfdg
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // becahfdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bceahfdg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cbaehfdg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bcaehfdg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbeafhdg
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // becafhdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bceafhdg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbaefhdg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcaefhdg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cbeafdhg
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // becafdhg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bceafdhg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cbaefdhg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bcaefdhg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v6, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // behgcadf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // behcgadf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bechgadf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // behcagdf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bechagdf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // becahgdf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbehgadf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcehgadf
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cbehagdf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bcehagdf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbeahgdf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bceahgdf
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbaehgdf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcaehgdf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); // behcadgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v5, v7); // bechadgf
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); // cbehadgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); // bcehadgf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cbeahdgf
																			}
																			else {
																				if (compare(v4, v2)) {
																					swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // becahdgf
																				}
																				else {
																					swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bceahdgf
																				}
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cbaehdgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bcaehdgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); // cbeadhgf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v5, v7); // becadhgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); // bceadhgf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v3, v4); swap(v5, v7); // cbaedhgf
																			}
																			else {
																				swap(v0, v2); swap(v5, v7); // cbadehgf
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v5, v7); // bcaedhgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v5, v7); // bcadehgf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // behcadfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bechadfg
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cbehadfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // bcehadfg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbeahdfg
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // becahdfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bceahdfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cbeadhfg
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // becadhfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bceadhfg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v6, v7); // cbeadfhg
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v6, v7); // becadfhg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v6, v7); // bceadfhg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbaehdfg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cbaedhfg
																			}
																			else {
																				swap(v0, v2); swap(v3, v4); swap(v6, v7); // cbaedfhg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v5, v7); swap(v6, v7); // cbadehfg
																		}
																		else {
																			swap(v0, v2); swap(v6, v7); // cbadefhg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcaehdfg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bcaedhfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v6, v7); // bcaedfhg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v5, v7); swap(v6, v7); // bcadehfg
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v6, v7); // bcadefhg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v3)) {
															if (compare(v6, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v6, v7); // behgafcd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // behagfcd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // behafgcd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); // beahgfcd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // beahfgcd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // beafhgcd
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // behgacfd
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // behagcfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // beahgcfd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); // behgacdf
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // behagcdf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // beahgcdf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // behacgfd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); // beahcgfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // beachgfd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // behacgdf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // beahcgdf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // beachgdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // behafcgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); // beahfcgd
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // beafhcgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); // beafchgd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); // behacfgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); // beahcfgd
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); // beachfgd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // beacfhgd
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // behafcdg
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // behacfdg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // behacdgf
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // behacdfg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // beahfcdg
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // beafhcdg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v6, v7); // beahcfdg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); // beahcdgf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // beahcdfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // beafchdg
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v6, v7); // beafcdhg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // beachfdg
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // beacfhdg
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // beacfdhg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // beachdgf
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // beachdfg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); // beacdhgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // beacdhfg
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v6, v7); // beacdfhg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v7, v5)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v6); // baehgfcd
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // baehfgcd
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v5); // baehfcgd
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // baehfcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // baehgcfd
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v5, v6); // baehcgfd
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); // baehcfgd
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v6, v7); // baehcfdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // baechgfd
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v7); // baechfgd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // baechfdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bacehgfd
																			}
																			else {
																				swap(v0, v1); swap(v3, v4); swap(v4, v7); // bacehfgd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v3, v4); swap(v4, v7); swap(v6, v7); // bacehfdg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // baefhgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // baefhcgd
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // baefhcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v5, v7); // baefchgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // baefchdg
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v6, v7); // baefcdhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // baecfhgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // baecfhdg
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // baecfdhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bacefhgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bacefhdg
																			}
																			else {
																				swap(v0, v1); swap(v3, v4); swap(v4, v5); swap(v6, v7); // bacefdhg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // baehgcdf
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // baehcgdf
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v5, v7); // baehcdgf
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // baehcdfg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // baechgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // baechdgf
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // baechdfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v5, v7); // baecdhgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // baecdhfg
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v6, v7); // baecdfhg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bacehgdf
																		}
																		else {
																			swap(v0, v1); swap(v3, v4); swap(v4, v7); swap(v5, v7); // bacehdgf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v3, v4); swap(v5, v7); // bacedhgf
																		}
																		else {
																			swap(v0, v1); swap(v5, v7); // bacdehgf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bacehdfg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v3, v4); swap(v5, v7); swap(v6, v7); // bacedhfg
																			}
																			else {
																				swap(v0, v1); swap(v3, v4); swap(v6, v7); // bacedfhg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v5, v7); swap(v6, v7); // bacdehfg
																		}
																		else {
																			swap(v0, v1); swap(v6, v7); // bacdefhg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v4)) {
										if (compare(v7, v0)) {
											if (compare(v7, v2)) {
												if (compare(v7, v1)) {
													if (compare(v5, v0)) {
														if (compare(v5, v1)) {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v4)) {
																		swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); // ghefcbad
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gehfcbad
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); // gefhcbad
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ghecfbad
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); // gehcfbad
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ghcefbad
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v7, v4)) {
																		swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ghefbcad
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); // gehfbcad
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); // gefhbcad
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghefbacd
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gehfbacd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); // gefhbacd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v2, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); // ghebfcad
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gehbfcad
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ghbefcad
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ghebfacd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gehbfacd
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghbefacd
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); // ghecbfad
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); // gehcbfad
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v7); // ghcebfad
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v3, v7); // ghcbefad
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); // ghebcfad
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); // gehbcfad
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); // ghbecfad
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); // ghbcefad
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v2, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ghecbafd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); // gehcbafd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ghcebafd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v5, v6); // ghcbeafd
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ghcbaefd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); // ghebcafd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gehbcafd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ghbecafd
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); // ghbceafd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ghbcaefd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); // ghebafcd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gehbafcd
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ghebacfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gehbacfd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ghbeafcd
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghbeacfd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v6, v7); // ghbaefcd
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ghbaecfd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghbacefd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v2, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghecbadf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gehcbadf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ghebcadf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gehbcadf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghbecadf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghebacdf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gehbacdf
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ghbeacdf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); // ghbaecdf
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghcebadf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ghcbeadf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghcbaedf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ghcbadef
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ghbceadf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghbcaedf
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ghbcadef
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ghbacedf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghbacdef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v0)) {
														if (compare(v5, v0)) {
															if (compare(v5, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); // gebhfcad
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gbhefcad
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); // gbehfcad
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bghefcad
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); // bgehfcad
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); // gefbhcad
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gebfhcad
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gbefhcad
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // bgefhcad
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v4)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); // gbhecfad
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v7); // gbhcefad
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); // gebhcfad
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v7); // gbehcfad
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); // bghecfad
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); // bgehcfad
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); // bghcefad
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); // gebhcafd
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gbhecafd
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); // gbehcafd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bghecafd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); // bgehcafd
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gebhcadf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbhecadf
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gbehcadf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bghecadf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bgehcadf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); // gbhceafd
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gbhceadf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gbhcaefd
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbhcaedf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gbhcadef
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); // bghceafd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bghceadf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bghcaefd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bghcaedf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bghcadef
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gebhfacd
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbhefacd
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gbehfacd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bghefacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // bgehfacd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gefbhacd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gebfhacd
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbefhacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgefhacd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v4)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gbheafcd
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v6, v7); // gbhaefcd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); // gebhafcd
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); // gbehafcd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bgheafcd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); // bgehafcd
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v6, v7); // bghaefcd
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gebhacfd
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbheacfd
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gbehacfd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgheacfd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bgehacfd
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gebhacdf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gbheacdf
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbehacdf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bgheacdf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgehacdf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gbhaecfd
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); // gbhaecdf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbhacefd
																			}
																			else {
																				swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gbhacedf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbhacdef
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bghaecfd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); // bghaecdf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bghacefd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bghacedf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bghacdef
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v4)) {
													if (compare(v6, v2)) {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gchefbad
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); // gchebfad
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); // gchbefad
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		swap(v0, v6); swap(v1, v2); swap(v3, v7); // gcbhefad
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v3, v7); // gbchefad
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); // bgchefad
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gchebafd
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); // gchbeafd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v5, v6); // gcbheafd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v3, v7); swap(v5, v6); // gbcheafd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v5, v6); // bgcheafd
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gchbaefd
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gcbhaefd
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gbchaefd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bgchaefd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gchebadf
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gchbeadf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gcbheadf
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gchbaedf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gchbadef
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcbhaedf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gcbhadef
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v0)) {
																		swap(v0, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gbcheadf
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbchaedf
																		}
																		else {
																			swap(v0, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gbchadef
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bgcheadf
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgchaedf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bgchadef
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cghefbad
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cghebfad
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v6, v7); // cghbefad
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); // cgbhefad
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); // cbghefad
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); // bcghefad
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cghebafd
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cghbeafd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); // cgbheafd
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); // cbgheafd
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); // bcgheafd
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cghbaefd
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cgbhaefd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cbghaefd
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bcghaefd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cghebadf
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); // cghbeadf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cgbheadf
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cghbaedf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cghbadef
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgbhaedf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cgbhadef
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v0)) {
																		swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cbgheadf
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbghaedf
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cbghadef
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bcgheadf
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcghaedf
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bcghadef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v4, v2)) {
															if (compare(v7, v5)) {
																if (compare(v5, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v5); // gechfbad
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v7); // gechbfad
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v7); // gecbhfad
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); // gebchfad
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v5, v6); // gechbafd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gecbhafd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gebchafd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gechbadf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gecbhadf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gebchadf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gefchbad
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); // gefcbhad
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gefbchad
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); // gebfchad
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gecfhbad
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v5, v7); // gecfbhad
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gecbfhad
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gebcfhad
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); // gcehfbad
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gcefhbad
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); // gcehbfad
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); // gcehbafd
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gcehbadf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cgehfbad
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgefhbad
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); // cgehbfad
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cgehbafd
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgehbadf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); // gcefbhad
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); // gcebhfad
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gcebfhad
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gcebhafd
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcebhadf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cgefbhad
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cgebhfad
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgebfhad
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgebhafd
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cgebhadf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v7); // gcbehfad
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gcbefhad
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); // cgbehfad
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cgbefhad
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); // gbefchad
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); // gbechfad
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gbecfhad
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v3, v4); swap(v4, v7); // gbcehfad
																		}
																		else {
																			swap(v0, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gbcefhad
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gcbehafd
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcbehadf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cgbehafd
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgbehadf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gbechafd
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbechadf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gbcehafd
																		}
																		else {
																			swap(v0, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbcehadf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); // bgefchad
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); // bgechfad
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bgecfhad
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v7); // bgcehfad
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bgcefhad
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); // cbgehfad
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cbgefhad
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); // bcgehfad
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bcgefhad
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bgechafd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgechadf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bgcehafd
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgcehadf
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cbgehafd
																		}
																		else {
																			swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbgehadf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bcgehafd
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcgehadf
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v1)) {
												if (compare(v4, v2)) {
													if (compare(v4, v1)) {
														if (compare(v2, v0)) {
															if (compare(v5, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gefcbahd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gecfbahd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gecbfahd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // gefcbadh
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v5, v6); // gecfbadh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gecbfadh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gefbcahd
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gebfcahd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gebcfahd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gefbcadh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // gebfcadh
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gebcfadh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gecbahfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gecbafhd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v6); // gecbafdh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gecbahdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gecbadhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gecbadfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gebcahfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gebcafhd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // gebcafdh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gebcahdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gebcadhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gebcadfh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gefbahcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gefbachd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // gefbacdh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gebfahcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gebfachd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gebfacdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gebahfcd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gebafhcd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gebafchd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // gebafcdh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gebahcfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gebachfd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v6, v7); // gebacfhd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); // gebacfdh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gebahcdf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); // gebachdf
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gebacdhf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // gebacdfh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v4, v0)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gbeahfcd
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbeahcfd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbecahfd
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gbeachfd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gbecahdf
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbecadhf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gbeahcdf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); // gbeachdf
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gbeacdhf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); // gbahefcd
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); // gbaehfcd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); // gbahecfd
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gbaehcfd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbaechfd
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gbahecdf
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbaehcdf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gbaechdf
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbaecdhf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v2, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gbefcahd
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // gbefcadh
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gbefahcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbefachd
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gbefacdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gbeafhcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gbeafchd
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // gbeafcdh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gbaefhcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbaefchd
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gbaefcdh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbecfahd
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gbecfadh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gbecafhd
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // gbecafdh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gbecadfh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v6, v7); // gbeacfhd
																			}
																			else {
																				swap(v0, v6); swap(v2, v4); swap(v3, v6); // gbeacfdh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // gbeacdfh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gbaecfhd
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // gbaecfdh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gbaecdfh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v4, v0)) {
															if (compare(v5, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gcefbahd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgefbahd
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcebfahd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cgebfahd
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcbefahd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgbefahd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gbcefahd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // gcefbadh
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cgefbadh
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gcebfadh
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // cgebfadh
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gcbefadh
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cgbefadh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gbcefadh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcebahfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gcebafhd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // gcebafdh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cgebahfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v6, v7); // cgebafhd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); // cgebafdh
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v3)) {
																				if (compare(v7, v5)) {
																					swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcbeahfd
																				}
																				else {
																					swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gcbeafhd
																				}
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v6); // gcbeafdh
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				if (compare(v7, v5)) {
																					swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgbeahfd
																				}
																				else {
																					swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cgbeafhd
																				}
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // cgbeafdh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gbceahfd
																			}
																			else {
																				swap(v0, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gbceafhd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v3, v4); swap(v4, v6); // gbceafdh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcbahefd
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gcbaehfd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v6, v7); // gcbaefhd
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v3, v6); // gcbaefdh
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgbahefd
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cgbaehfd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v6, v7); // cgbaefhd
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); // cgbaefdh
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbcahefd
																		}
																		else {
																			swap(v0, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gbcaehfd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v3, v6); swap(v6, v7); // gbcaefhd
																		}
																		else {
																			swap(v0, v6); swap(v3, v6); // gbcaefdh
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gbahcefd
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gbachefd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gbacehfd
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v6, v7); // gbacefhd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); // gbacefdh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v6, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gcebahdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcebadhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gcebadfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gcbeahdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcbeadhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gcbeadfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gcbahedf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v5, v7); // gcbaehdf
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gcbaedhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v5, v6); // gcbaedfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcbahdef
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gcbadhef
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcbadehf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gcbadefh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); // cgebahdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cgebadhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // cgebadfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cgbeahdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgbeadhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cgbeadfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cgbahedf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); // cgbaehdf
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cgbaedhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // cgbaedfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgbahdef
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cgbadhef
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgbadehf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cgbadefh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gbceahdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbceadhf
																			}
																			else {
																				swap(v0, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gbceadfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gbcahedf
																			}
																			else {
																				swap(v0, v6); swap(v3, v6); swap(v5, v7); // gbcaehdf
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gbcaedhf
																			}
																			else {
																				swap(v0, v6); swap(v3, v6); swap(v5, v6); // gbcaedfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbcahdef
																		}
																		else {
																			swap(v0, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gbcadhef
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbcadehf
																		}
																		else {
																			swap(v0, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gbcadefh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbahcedf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gbachedf
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); // gbacehdf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gbacedhf
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // gbacedfh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gbahcdef
																		}
																		else {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gbachdef
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gbacdhef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gbacdehf
																			}
																			else {
																				swap(v0, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gbacdefh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v0)) {
													if (compare(v7, v3)) {
														if (compare(v5, v3)) {
															if (compare(v7, v5)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbgeahfd
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbgahefd
																			}
																			else {
																				swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cbgaehfd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbaghefd
																		}
																		else {
																			swap(v0, v2); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cbagehfd
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgecahfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgceahfd
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgcahefd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bgcaehfd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v4, v0)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcgeahfd
																			}
																			else {
																				if (compare(v7, v4)) {
																					swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcgahefd
																				}
																				else {
																					swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bcgaehfd
																				}
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bcaghefd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bcagehfd
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bgefcahd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgecfahd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgcefahd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bgecafhd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bgceafhd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v6, v7); // bgcaefhd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbgefahd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcgefahd
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cbgeafhd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bcgeafhd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v6, v7); // cbgaefhd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v6, v7); // bcgaefhd
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v3, v6); swap(v6, v7); // cbagefhd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v6, v7); // bcagefhd
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v4)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cbgahedf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bgcahedf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bcgahedf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cbaghedf
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bcaghedf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbgahdef
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgcahdef
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcgahdef
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cbaghdef
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bcaghdef
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bgecahdf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bgceahdf
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cbgeahdf
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bcgeahdf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); // cbgaehdf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v5, v7); // bgcaehdf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); // bcgaehdf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v3, v6); swap(v5, v7); // cbagehdf
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v5, v7); // bcagehdf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v6, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgecadhf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgceadhf
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbgeadhf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcgeadhf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cbgaedhf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bgcaedhf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bcgaedhf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cbgadhef
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bgcadhef
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bcgadhef
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbgadehf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgcadehf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcgadehf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cbagedhf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cbagdhef
																			}
																			else {
																				swap(v0, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbagdehf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v4, v6); swap(v5, v7); // cbadghef
																		}
																		else {
																			swap(v0, v2); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbadgehf
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bcagedhf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bcagdhef
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcagdehf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v2); swap(v4, v6); swap(v5, v7); // bcadghef
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcadgehf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // bgefcadh
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bgecfadh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bgcefadh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // bgecafdh
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v6); // bgceafdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v3, v6); // bgcaefdh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cbgefadh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bcgefadh
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // cbgeafdh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // bcgeafdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v6); // cbgaefdh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); // bcgaefdh
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v3, v6); // cbagefdh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v6); // bcagefdh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bgecadfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bgceadfh
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cbgeadfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bcgeadfh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // cbgaedfh
																			}
																			else {
																				if (compare(v6, v2)) {
																					swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v5, v6); // bgcaedfh
																				}
																				else {
																					swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // bcgaedfh
																				}
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v3, v6); swap(v5, v6); // cbagedfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v5, v6); // bcagedfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cbgadefh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bgcadefh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bcgadefh
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cbagdefh
																			}
																			else {
																				swap(v0, v2); swap(v4, v6); swap(v5, v6); // cbadgefh
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bcagdefh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v4, v6); swap(v5, v6); // bcadgefh
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v0)) {
														if (compare(v5, v3)) {
															if (compare(v5, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bgefahcd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bgeahfcd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bgeafhcd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); // bgahefcd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); // bgaehfcd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bgaefhcd
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgefachd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bgeafchd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bgaefchd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v5, v0)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bgefacdh
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // bgeafcdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bgaefcdh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); // bgahecfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // bgahcefd
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgeahcfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bgaehcfd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v0)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bgeachfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bgaechfd
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bgachefd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bgacehfd
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v6, v7); // bgeacfhd
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bgaecfhd
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v6, v7); // bgacefhd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); // bgeacfdh
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // bgaecfdh
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); // bgacefdh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bgeahcdf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); // bgeachdf
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bgeacdhf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // bgeacdfh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // bgahecdf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgaehcdf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bgaechdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgaecdhf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bgaecdfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgahcedf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bgachedf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); // bgacehdf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bgacedhf
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // bgacedfh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // bgahcdef
																		}
																		else {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bgachdef
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bgacdhef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bgacdehf
																			}
																			else {
																				swap(v0, v1); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bgacdefh
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v7); // baghefcd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v7); // bagehfcd
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // bagefhcd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bagefchd
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bagefcdh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v5, v6); // baghecfd
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // bagehcfd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bagechfd
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bagecfhd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v6); // bagecfdh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // baghcefd
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bagchefd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bagcehfd
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v6, v7); // bagcefhd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v6); swap(v3, v6); // bagcefdh
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v1); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bacghefd
																		}
																		else {
																			swap(v0, v1); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bacgehfd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v3, v6); swap(v6, v7); // bacgefhd
																		}
																		else {
																			swap(v0, v1); swap(v3, v6); // bacgefdh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v7, v3)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // baghecdf
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // baghcedf
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bagchedf
																			}
																			else {
																				swap(v0, v1); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bacghedf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bagehcdf
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bagechdf
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v5, v7); // bagcehdf
																			}
																			else {
																				swap(v0, v1); swap(v3, v6); swap(v5, v7); // bacgehdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bagecdhf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bagcedhf
																			}
																			else {
																				swap(v0, v1); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bacgedhf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bagecdfh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v5, v6); // bagcedfh
																			}
																			else {
																				swap(v0, v1); swap(v3, v6); swap(v5, v6); // bacgedfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v4)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // baghcdef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bagchdef
																			}
																			else {
																				swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bagcdhef
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // bacghdef
																			}
																			else {
																				swap(v0, v1); swap(v3, v6); swap(v4, v6); swap(v5, v7); // bacgdhef
																			}
																		}
																		else {
																			swap(v0, v1); swap(v4, v6); swap(v5, v7); // bacdghef
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bagcdehf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bacgdehf
																			}
																			else {
																				swap(v0, v1); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bacdgehf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bagcdefh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v1); swap(v3, v6); swap(v4, v6); swap(v5, v6); // bacgdefh
																			}
																			else {
																				swap(v0, v1); swap(v4, v6); swap(v5, v6); // bacdgefh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v5, v0)) {
											if (compare(v5, v2)) {
												if (compare(v5, v1)) {
													if (compare(v7, v0)) {
														if (compare(v7, v1)) {
															if (compare(v2, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eghfcbad
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // egfhcbad
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); // efghcbad
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egfchbad
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // efgchbad
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // efcghbad
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // eghfbcad
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); // egfhbcad
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // efghbcad
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eghfbacd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // egfhbacd
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); // efghbacd
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v2, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // egfbhcad
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efgbhcad
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // efbghcad
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egfbhacd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // efgbhacd
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efbghacd
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // egfcbhad
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efgcbhad
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efcgbhad
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // efcbghad
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egfbchad
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // efgbchad
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efbgchad
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // efbcghad
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v2, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egfcbahd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // efgcbahd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // efcgbahd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efcbgahd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // efcbaghd
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // egfbcahd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efgbcahd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // efbgcahd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efbcgahd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // efbcaghd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // egfbahcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); // efgbahcd
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v6, v7); // egfbachd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // efgbachd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // efbgahcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v6, v7); // efbgachd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efbaghcd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // efbagchd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efbacghd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v2, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // egfcbadh
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // efgcbadh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // egfbcadh
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // efgbcadh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // efbgcadh
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v6); // egfbacdh
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // efgbacdh
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v6); // efbgacdh
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // efbagcdh
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v5); // efcgbadh
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // efcbgadh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v5); swap(v5, v6); // efcbagdh
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v3, v5); // efcbadgh
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // efbcgadh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // efbcagdh
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v5); // efbcadgh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // efbacgdh
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // efbacdgh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v0)) {
														if (compare(v7, v0)) {
															if (compare(v7, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eghbfcad
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // egbhfcad
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ebghfcad
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // beghfcad
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egbfhcad
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebgfhcad
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ebfghcad
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // begfhcad
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // befghcad
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // egbfchad
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ebgfchad
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebfgchad
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ebfcghad
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // begfchad
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // befgchad
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // befcghad
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egbfcahd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebgfcahd
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ebfgcahd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // begfcahd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // befgcahd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // egbfcadh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ebgfcadh
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // ebfgcadh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // begfcadh
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // befgcadh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebfcgahd
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ebfcgadh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ebfcaghd
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // ebfcagdh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v5); // ebfcadgh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // befcgahd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // befcgadh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // befcaghd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // befcagdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v5); // befcadgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v7, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // eghbfacd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); // egbhfacd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); // ebghfacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); // beghfacd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // egbfhacd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ebgfhacd
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebfghacd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // begfhacd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // befghacd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); // egbfahcd
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); // ebgfahcd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ebfgahcd
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebfaghcd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); // begfahcd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // befgahcd
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // befaghcd
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // egbfachd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ebgfachd
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); swap(v6, v7); // ebfgachd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // begfachd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v6, v7); // befgachd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // egbfacdh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // ebgfacdh
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v6); // ebfgacdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // begfacdh
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v6); // befgacdh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ebfagchd
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // ebfagcdh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebfacghd
																			}
																			else {
																				swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ebfacgdh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // ebfacdgh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // befagchd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // befagcdh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // befacghd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // befacgdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // befacdgh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v5)) {
													if (compare(v6, v1)) {
														if (compare(v6, v2)) {
															if (compare(v7, v5)) {
																if (compare(v7, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // eghcfbad
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); // eghcbfad
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // eghbcfad
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); // egbhcfad
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // egchfbad
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v6); // egchbfad
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // egcbhfad
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // egbchfad
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egcfhbad
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // egcfbhad
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egcbfhad
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egbcfhad
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v1)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egcfbahd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // egcbfahd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // egbcfahd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v5, v1)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // egcfbadh
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v5); // egcbfadh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // egbcfadh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ecghfbad
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecgfhbad
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ecgfbhad
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecgfbahd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ecgfbadh
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); // ceghfbad
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cegfhbad
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); // cegfbhad
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cegfbahd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // cegfbadh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); // ecghbfad
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ecgbhfad
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecgbfhad
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ecgbfahd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // ecgbfadh
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); // ceghbfad
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); // cegbhfad
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cegbfhad
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cegbfahd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cegbfadh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v7, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ecbghfad
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecbgfhad
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cebghfad
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cebgfhad
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); // ebghcfad
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ebgchfad
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebgcfhad
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ebcghfad
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebcgfhad
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ecbgfahd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v5); // ecbgfadh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cebgfahd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // cebgfadh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ebgcfahd
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // ebgcfadh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ebcgfahd
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v4, v5); // ebcgfadh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); // beghcfad
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // begchfad
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // begcfhad
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // becghfad
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // becgfhad
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cbeghfad
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cbegfhad
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // bceghfad
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // bcegfhad
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // begcfahd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // begcfadh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // becgfahd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v5); // becgfadh
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cbegfahd
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // cbegfadh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // bcegfahd
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // bcegfadh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v7, v3)) {
															if (compare(v7, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ecfghbad
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecfgbhad
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ecfbghad
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ecbfghad
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ebcfghad
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // becfghad
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ecfgbahd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecfbgahd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecbfgahd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebcfgahd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // becfgahd
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ecfbaghd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ecbfaghd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ebcfaghd
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // becfaghd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // ecfgbadh
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ecfbgadh
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ecbfgadh
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // ecfbagdh
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v5); // ecfbadgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v3, v5); swap(v5, v6); // ecbfagdh
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v5); // ecbfadgh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v0)) {
																		swap(v0, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ebcfgadh
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v3, v5); swap(v5, v6); // ebcfagdh
																		}
																		else {
																			swap(v0, v4); swap(v3, v5); // ebcfadgh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // becfgadh
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v3, v5); swap(v5, v6); // becfagdh
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v5); // becfadgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v7, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cefghbad
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cefgbhad
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cefbghad
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cebfghad
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cbefghad
																		}
																	}
																	else {
																		swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // bcefghad
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v6, v7); // cefgbahd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cefbgahd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cebfgahd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbefgahd
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcefgahd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cefbaghd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cebfaghd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cbefaghd
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // bcefaghd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); // cefgbadh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // cefbgadh
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cebfgadh
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cefbagdh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cefbadgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // cebfagdh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v5); // cebfadgh
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v0)) {
																		swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cbefgadh
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // cbefagdh
																		}
																		else {
																			swap(v0, v2); swap(v2, v4); swap(v3, v5); // cbefadgh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // bcefgadh
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // bcefagdh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v5); // bcefadgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v1)) {
												if (compare(v6, v2)) {
													if (compare(v6, v1)) {
														if (compare(v2, v0)) {
															if (compare(v7, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // eghcbafd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // egchbafd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egcbhafd
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eghcbadf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egchbadf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // egcbhadf
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eghbcafd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // egbhcafd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egbchafd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // eghbcadf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egbhcadf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // egbchadf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // egcbahfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v6, v7); // egcbafhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v6); // egcbafdh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v5, v7); // egcbahdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // egcbadhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v3, v6); swap(v5, v6); // egcbadfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // egbcahfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v6, v7); // egbcafhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); // egbcafdh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); // egbcahdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // egbcadhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // egbcadfh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v6, v7); // eghbafcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // eghbacfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); // eghbacdf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); // egbhafcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); // egbhacfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // egbhacdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); // egbahfcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // egbafhcd
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // egbahcfd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egbahcdf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egbachfd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // egbachdf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egbafchd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // egbacfhd
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // egbafcdh
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // egbacfdh
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egbacdhf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // egbacdfh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v7, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v2, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ebghcafd
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebghcadf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v7); // ebghafcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); // ebghacfd
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ebghacdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); // ebgahfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ebgahcfd
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ebgahcdf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ebaghfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebaghcfd
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ebaghcdf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebgchafd
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ebgchadf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ebgcahfd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); // ebgcahdf
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ebgcadhf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebgachfd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ebgachdf
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebgacdhf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ebagchfd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); // ebagchdf
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ebagcdhf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ebgafhcd
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebgafchd
																		}
																	}
																	else {
																		if (compare(v2, v0)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v6, v7); // ebgcafhd
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ebgacfhd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v6); // ebgcafdh
																		}
																		else {
																			swap(v0, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // ebgcadfh
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ebgafcdh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // ebgacfdh
																			}
																			else {
																				swap(v0, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ebgacdfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ebagfhcd
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ebafghcd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ebagfchd
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebafgchd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v6, v7); // ebagcfhd
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // ebagfcdh
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ebafgcdh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v6); // ebagcfdh
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // ebagcdfh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v6, v0)) {
															if (compare(v6, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ecghbafd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecgbhafd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ecgbahfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v6, v7); // ecgbafhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); // ecgbafdh
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); // ceghbafd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cegbhafd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cegbahfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cegbafhd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // cegbafdh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecbghafd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cebghafd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ebcghafd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ecbgahfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cebgahfd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ebcgahfd
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v6, v7); // ecbgafhd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v6, v7); // cebgafhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v6, v7); // ebcgafhd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); // ecbgafdh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); // cebgafdh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); // ebcgafdh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecbaghfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ecbagfhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecbafghd
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cebaghfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cebagfhd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cebafghd
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebcaghfd
																			}
																			else {
																				swap(v0, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ebcagfhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebcafghd
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ebacghfd
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ebacgfhd
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ebafcghd
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ebacfghd
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v6); // ecbagfdh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ecbafgdh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v5); // ecbafdgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // cebagfdh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cebafgdh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cebafdgh
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v3, v4); swap(v4, v6); // ebcagfdh
																			}
																			else {
																				swap(v0, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ebcafgdh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v4); swap(v4, v5); // ebcafdgh
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // ebafcgdh
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v5); // ebafcdgh
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				if (compare(v6, v5)) {
																					swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // ebacgfdh
																				}
																				else {
																					swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ebacfgdh
																				}
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // ebacfdgh
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v6, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecghbadf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ecgbhadf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); // ecgbahdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ecgbadhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // ecgbadfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ceghbadf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cegbhadf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cegbahdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cegbadhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cegbadfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ecbghadf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cebghadf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ebcghadf
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v5, v7); // ecbgahdf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); // cebgahdf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v5, v7); // ebcgahdf
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ecbgadhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cebgadhf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ebcgadhf
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v6); swap(v5, v6); // ecbgadfh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // cebgadfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v6); swap(v5, v6); // ebcgadfh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ecbaghdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecbagdhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ecbagdfh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ecbadghf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v3, v4); swap(v5, v6); // ecbadgfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v3, v4); // ecbadfgh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cebaghdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cebagdhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cebagdfh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cebadghf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // cebadgfh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v4); swap(v3, v4); // cebadfgh
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ebcaghdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebcagdhf
																			}
																			else {
																				swap(v0, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ebcagdfh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ebcadghf
																			}
																			else {
																				swap(v0, v4); swap(v3, v4); swap(v5, v6); // ebcadgfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v3, v4); // ebcadfgh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ebacghdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ebacgdhf
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ebacgdfh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ebacdghf
																			}
																			else {
																				swap(v0, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // ebacdgfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v2, v4); swap(v3, v4); // ebacdfgh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v0)) {
													if (compare(v5, v3)) {
														if (compare(v7, v3)) {
															if (compare(v7, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // beghcafd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // begchafd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // becghafd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // begcahfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // becgahfd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // becaghfd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cbeghafd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // bceghafd
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cbegahfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // bcegahfd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbeaghfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bceaghfd
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cbaeghfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // bcaeghfd
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v6, v7); // cbegafhd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cbeagfhd
																			}
																			else {
																				swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbeafghd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cbaegfhd
																		}
																		else {
																			swap(v0, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cbaefghd
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v6, v7); // begcafhd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v6, v7); // becgafhd
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // becagfhd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // becafghd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v0)) {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v6, v7); // bcegafhd
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bceagfhd
																				}
																				else {
																					swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bceafghd
																				}
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bcaegfhd
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bcaefghd
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v6, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); // begcafdh
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v6); // becgafdh
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v6); // cbegafdh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); // bcegafdh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // cbeagfdh
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v6); // becagfdh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // bceagfdh
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v3, v4); swap(v4, v6); // cbaegfdh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v6); // bcaegfdh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cbeafgdh
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // becafgdh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bceafgdh
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cbaefgdh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bcaefgdh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cbeafdgh
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v5); // becafdgh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // bceafdgh
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v3, v4); swap(v4, v5); // cbaefdgh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v5); // bcaefdgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v7, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // beghcadf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // begchadf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // becghadf
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); // begcahdf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v5, v7); // becgahdf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // becaghdf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cbeghadf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // bceghadf
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); // cbegahdf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); // bcegahdf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cbeaghdf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bceaghdf
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cbaeghdf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); // bcaeghdf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // begcadhf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // becgadhf
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cbegadhf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // bcegadhf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbeagdhf
																			}
																			else {
																				if (compare(v4, v2)) {
																					swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // becagdhf
																				}
																				else {
																					swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bceagdhf
																				}
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cbaegdhf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bcaegdhf
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cbeadghf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // becadghf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bceadghf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cbaedghf
																			}
																			else {
																				swap(v0, v2); swap(v5, v6); swap(v6, v7); // cbadeghf
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bcaedghf
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v5, v6); swap(v6, v7); // bcadeghf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // begcadfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v3, v6); swap(v5, v6); // becgadfh
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // cbegadfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // bcegadfh
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cbeagdfh
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // becagdfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bceagdfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // cbeadgfh
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); swap(v5, v6); // becadgfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // bceadgfh
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v2); swap(v2, v4); swap(v3, v4); // cbeadfgh
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v3, v4); // becadfgh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v2, v4); swap(v3, v4); // bceadfgh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cbaegdfh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v3, v4); swap(v5, v6); // cbaedgfh
																			}
																			else {
																				swap(v0, v2); swap(v3, v4); // cbaedfgh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v5, v6); // cbadegfh
																		}
																		else {
																			swap(v0, v2); // cbadefgh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bcaegdfh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); swap(v5, v6); // bcaedgfh
																			}
																			else {
																				swap(v0, v1); swap(v1, v2); swap(v3, v4); // bcaedfgh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v2); swap(v5, v6); // bcadegfh
																		}
																		else {
																			swap(v0, v1); swap(v1, v2); // bcadefgh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v7, v3)) {
															if (compare(v7, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); // beghafcd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); // begahfcd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // begafhcd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // beaghfcd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // beagfhcd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // beafghcd
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); // beghacfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // begahcfd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // beaghcfd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v7, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // beghacdf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // begahcdf
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // beaghcdf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // begachfd
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // beagchfd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // beacghfd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // begachdf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); // beagchdf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // beacghdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v6, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // begafchd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // beagfchd
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // beafgchd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // beafcghd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v0)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // begacfhd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v6, v7); // beagcfhd
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // beacgfhd
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // beacfghd
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // begafcdh
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // begacfdh
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // begacdhf
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // begacdfh
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // beagfcdh
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // beafgcdh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); // beagcfdh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // beagcdhf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // beagcdfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // beafcgdh
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v5); // beafcdgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // beacgfdh
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // beacfgdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // beacfdgh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // beacgdhf
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // beacgdfh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // beacdghf
																			}
																			else {
																				swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // beacdgfh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v1, v4); swap(v2, v4); swap(v3, v4); // beacdfgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v5)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // baeghfcd
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // baegfhcd
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // baegfchd
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v5); // baegfcdh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			if (compare(v7, v2)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // baeghcfd
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // baegchfd
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v6, v7); // baegcfhd
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v6); // baegcfdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // baecghfd
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // baecgfhd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v6); // baecgfdh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // baceghfd
																			}
																			else {
																				swap(v0, v1); swap(v3, v4); swap(v4, v6); swap(v6, v7); // bacegfhd
																			}
																		}
																		else {
																			swap(v0, v1); swap(v3, v4); swap(v4, v6); // bacegfdh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // baefghcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // baefgchd
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // baefgcdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // baefcghd
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v5); swap(v5, v6); // baefcgdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v3, v5); // baefcdgh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // baecfghd
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // baecfgdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v5); // baecfdgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v1); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // bacefghd
																			}
																			else {
																				swap(v0, v1); swap(v3, v4); swap(v4, v5); swap(v5, v6); // bacefgdh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v3, v4); swap(v4, v5); // bacefdgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // baeghcdf
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v5, v7); // baegchdf
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // baegcdhf
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v3, v6); swap(v5, v6); // baegcdfh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // baecghdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // baecgdhf
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // baecgdfh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // baecdghf
																			}
																			else {
																				swap(v0, v1); swap(v2, v4); swap(v3, v4); swap(v5, v6); // baecdgfh
																			}
																		}
																		else {
																			swap(v0, v1); swap(v2, v4); swap(v3, v4); // baecdfgh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v1); swap(v3, v4); swap(v4, v6); swap(v5, v7); // baceghdf
																		}
																		else {
																			swap(v0, v1); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // bacegdhf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v1); swap(v3, v4); swap(v5, v6); swap(v6, v7); // bacedghf
																		}
																		else {
																			swap(v0, v1); swap(v5, v6); swap(v6, v7); // bacdeghf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v1); swap(v3, v4); swap(v4, v6); swap(v5, v6); // bacegdfh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v1); swap(v3, v4); swap(v5, v6); // bacedgfh
																			}
																			else {
																				swap(v0, v1); swap(v3, v4); // bacedfgh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v1); swap(v5, v6); // bacdegfh
																		}
																		else {
																			swap(v0, v1); // bacdefgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
				else {
					if (compare(v3, v2)) {
						if (compare(v3, v0)) {
							if (compare(v5, v4)) {
								if (compare(v7, v6)) {
									if (compare(v6, v4)) {
										if (compare(v5, v0)) {
											if (compare(v4, v1)) {
												if (compare(v2, v0)) {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v6, v7); // hgfedcab
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hfgedcab
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v6, v7); // hgfdecab
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v5, v6); swap(v6, v7); // hfgdecab
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hfdgecab
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); // fhgedcab
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v5, v6); // fhgdecab
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); // fhdgecab
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fdhgecab
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v6, v7); // hgfdceab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfgdceab
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); // fhgdceab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfdgceab
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fhdgceab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdhgceab
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfdcgeab
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); // fhdcgeab
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fdhcgeab
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fdchgeab
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v6, v7); // hgdfecab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hdgfecab
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); // dhgfecab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hgdfceab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdgfceab
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dhgfceab
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v6, v7); // hgdcfeab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hdgcfeab
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); // dhgcfeab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hdcgfeab
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v5); // dhcgfeab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); // dchgfeab
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v6, v7); // hdfgecab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); // dhfgecab
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); // dfhgecab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hdfgceab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); // dhfgceab
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dfhgceab
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdfcgeab
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dhfcgeab
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfhcgeab
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfchgeab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdcfgeab
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dhcfgeab
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dchfgeab
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcfhgeab
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v6, v2)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgfdcaeb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); // hfgdcaeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v6); // fhgdcaeb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hfdgcaeb
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); // fhdgcaeb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgdfcaeb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hdgfcaeb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdfgcaeb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgdcfaeb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdgcfaeb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); // hfdcgaeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v6); // fhdcgaeb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hdfcgaeb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdcgfaeb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hdcfgaeb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfdcageb
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); // fhdcageb
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdfcageb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdcfageb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); // dhgfcaeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dhfgcaeb
																			}
																		}
																		else {
																			if (compare(v6, v0)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); // dhfcgaeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dhfcageb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dhgcfaeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dhcgfaeb
																			}
																		}
																		else {
																			if (compare(v6, v0)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v6); // dhcfgaeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dhcfageb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fdhgcaeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); // fdhcgaeb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fdhcageb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfhgcaeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dfhcgaeb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfhcageb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v6); // fdchgaeb
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fdchageb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdcahgeb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dfchgaeb
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfchageb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfcahgeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dchgfaeb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); // dchfgaeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dcfhgaeb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dchfageb
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcfhageb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcfahgeb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v7, v5)) {
															if (compare(v4, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hgfedacb
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); // hfgedacb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v5, v7); swap(v6, v7); // hgfdeacb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v5, v7); // hfgdeacb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); // hfdgeacb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); // hgfedabc
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hfgedabc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v5, v7); // hgfdeabc
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v5, v7); swap(v6, v7); // hfgdeabc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hfdgeabc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgfdaecb
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v6, v7); // hgfdaceb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); // hgfdaebc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); // hfgdaecb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfgdaceb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfgdaebc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hfdgaecb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfdgaceb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfdgaebc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfdagecb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hfdagebc
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hfdagceb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfdacgeb
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); // fhgedacb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); // fhgdeacb
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v6, v7); // fhgedabc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v6, v7); // fhgdeabc
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v5); // fhgdaecb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); // fhgdaceb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v6, v7); // fhgdaebc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); // fhdgeacb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v6, v7); // fhdgeabc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				if (compare(v4, v2)) {
																					swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); // fhdgaecb
																				}
																				else {
																					swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fhdgaceb
																				}
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fhdgaebc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fhdagecb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhdagebc
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); // fhdagceb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fhdacgeb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v6, v7); // fdhgeacb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); // fdhgeabc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fdhgaecb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); // fdhgaebc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdhagecb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fdhagebc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fdahgecb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdahgebc
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdhgaceb
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fdhagceb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdhacgeb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); // fdahgceb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fdahcgeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdachgeb
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v6, v5)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hgdfeacb
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); // hgdfeabc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgdfaecb
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hgdfaceb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hgdfaebc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); // hdgfeacb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hdgfeabc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hdgfaecb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdgfaceb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdgfaebc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hdfgeacb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); // hdfgeabc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdfgaecb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hdfgaceb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hdfgaebc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hdfagecb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdfagebc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); // hdfagceb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hdfacgeb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); // dhgfeacb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); // dhfgeacb
																			}
																		}
																		else {
																			if (compare(v6, v0)) {
																				if (compare(v6, v5)) {
																					swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); // dhgfaecb
																				}
																				else {
																					swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dhfgaecb
																				}
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); // dhfagecb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dhgfaceb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); // dhfgaceb
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v6); // dhfagceb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); // dhfacgeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v6, v7); // dhgfeabc
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dhfgeabc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dhgfaebc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhfgaebc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhfagebc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dfhgeacb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); // dfhgeabc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfhgaecb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dfhgaebc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfhagecb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); // dfhagebc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfahgecb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dfahgebc
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dfhgaceb
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); // dfhagceb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfhacgeb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dfahgceb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfahcgeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfachgeb
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v2)) {
													if (compare(v5, v3)) {
														if (compare(v6, v0)) {
															if (compare(v2, v1)) {
																if (compare(v2, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); // hgfdcabe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfgdcabe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v6, v7); // fhgdcabe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfdgcabe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fhdgcabe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); // fdhgcabe
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v7); // hgfdacbe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); // hfgdacbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhgdacbe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hfdgacbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhdgacbe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fdhgacbe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgfdabec
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v6, v7); // hfgdabec
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhgdabec
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hfdgabec
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhdgabec
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fdhgabec
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); // hgfdabce
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v4, v7); // hfgdabce
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); // fhgdabce
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); // hfdgabce
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fhdgabce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdhgabce
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfdagcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); // hfdagbec
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hfdagbce
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fhdagcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fhdagbec
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhdagbce
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdhagcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); // fdhagbec
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fdhagbce
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fdahgcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fdahgbec
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdahgbce
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hfdabgec
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhdabgec
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fdhabgec
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdahbgec
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdabhgec
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfdabgce
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fhdabgce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdhabgce
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fdahbgce
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); // fdabhgce
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v7, v3)) {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v5, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hgdfcabe
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdgfcabe
																				}
																				else {
																					swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hdfgcabe
																				}
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); // hgdcfabe
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hdgcfabe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); // hgdfacbe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hdgfacbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); // hdfgacbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgdfabec
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hdgfabec
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdfgabec
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hgdfabce
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); // hdgfabce
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hdfgabce
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v6, v5)) {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dhgfcabe
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhgcfabe
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhfgcabe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dfhgcabe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhgfacbe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dhfgacbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfhgacbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhgfabec
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dhfgabec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); // dfhgabec
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dhgfabce
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhfgabce
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dfhgabce
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hdfagcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hdfagbec
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdfagbce
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v6, v7); // dhfagcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); // dhfagbec
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhfagbce
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfhagcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v6); // dfhagbec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v6, v7); // dfhagbce
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfahgcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); // dfahgbec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dfahgbce
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdfabgec
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhfabgec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); // dfhabgec
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dfahbgec
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfabhgec
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hdfabgce
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); // dhfabgce
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfhabgce
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfahbgce
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dfabhgce
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v7, v5)) {
															if (compare(v6, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfdcgabe
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdfcgabe
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hdcgfabe
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdcfgabe
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dhfcgabe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhcgfabe
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dhcfgabe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dchgfabe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dchfgabe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); // hfdcagbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hdfcagbe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hdcfagbe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hfdacgbe
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdfacgbe
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhfcagbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhfacgbe
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhcfagbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dchfagbe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v2, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v6, v7); // fhdcgabe
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fdhcgabe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfhcgabe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhdcagbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdhcagbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dfhcagbe
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhdacgbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fdhacgbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); // dfhacgbe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdahcgbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dfahcgbe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fdchgabe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfchgabe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcfhgabe
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdchagbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dfchagbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dcfhagbe
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fdcahgbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v4, v7); swap(v5, v6); // dfcahgbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); // dcfahgbe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fdachgbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); // dfachgbe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v7, v2)) {
																if (compare(v7, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v7); // hfdcabge
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); // fhdcabge
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); // hdfcabge
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v7); // hdcfabge
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dhfcabge
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dhcfabge
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fdhcabge
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); // dfhcabge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fdchabge
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v7); // dfchabge
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dchfabge
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); // dcfhabge
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v7); // fdcahbge
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v4, v7); // dfcahbge
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v7); // dcfahbge
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fdcabhge
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v4, v5); swap(v5, v7); // dfcabhge
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); // dcfabhge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); // hfdacbge
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fhdacbge
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); // fdhacbge
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hdfacbge
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); // dhfacbge
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v7); // dfhacbge
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hfdabcge
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); // fhdabcge
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fdhabcge
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); // hdfabcge
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v7); // dhfabcge
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); // dfhabcge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fdahcbge
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); // fdachbge
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fdacbhge
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); // dfahcbge
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v7); // dfachbge
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); // dfacbhge
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); // fdahbcge
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v7); // fdabhcge
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); // fdabchge
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dfahbcge
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); // dfabhcge
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dfabchge
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v6, v5)) {
													if (compare(v4, v2)) {
														if (compare(v7, v3)) {
															if (compare(v4, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hgdafecb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); // hdgafecb
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hdagfecb
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); // hgdafebc
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hdgafebc
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); // hdagfebc
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hgdafbec
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdgafbec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); // hgdabfec
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hdgabfec
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hdagfbec
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); // hdagbfec
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v6); // hdabgfec
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v6, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v5); // dhgafecb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v6, v7); // dhgafebc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhgafbec
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v6, v7); // dhgabfec
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); // dhagfecb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dhagfebc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhagfbec
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dhagbfec
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); // dhabgfec
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v2, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dahgfecb
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); // dahgfebc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dahgfbec
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); // dahgbfec
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v6); // dahbgfec
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); // dabhgfec
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v2, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgdafceb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hdgafceb
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdagfceb
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hgdafcbe
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdgafcbe
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hdagfcbe
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgdafbce
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hdgafbce
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hgdabfce
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); // hdgabfce
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdagfbce
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hdagbfce
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v6, v7); // hdabgfce
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v6, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); // dhgafceb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhgafcbe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); // dhgafbce
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v7); // dhgabfce
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dhagfceb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhagfcbe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dhagfbce
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); // dhagbfce
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dhabgfce
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dahgfceb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dahgfcbe
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dahgfbce
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dahgbfce
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v6, v7); // dahbgfce
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dabhgfce
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v4, v1)) {
															if (compare(v7, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdafgecb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dhafgecb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hdafgebc
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhafgebc
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hdafgceb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); // dhafgceb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdafcgeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dhafcgeb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dahfgecb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dafhgecb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dahfgebc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dafhgebc
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dahfgceb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); // dafhgceb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dahfcgeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dafhcgeb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dafchgeb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdafgcbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dhafgcbe
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hdafcgbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhafcgbe
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); // hdafgbec
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dhafgbec
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hdafgbce
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhafgbce
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dahfgcbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dafhgcbe
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dahfcgbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dafhcgbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dafchgbe
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); // dahfgbec
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dafhgbec
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dahfgbce
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dafhgbce
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v5, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hdafbgec
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhafbgec
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dahfbgec
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dafhbgec
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dafbhgec
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdafbgce
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dhafbgce
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dahfbgce
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dafhbgce
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); // dafbhgce
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); // hdabfgec
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dhabfgec
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); // dahbfgec
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dabhfgec
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dabfhgec
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdabfgce
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhabfgce
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dahbfgce
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dabhfgce
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dabfhgce
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v7, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); // hdafcbge
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hdafbcge
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); // hdabfcge
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dhafcbge
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); // dhafbcge
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dhabfcge
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); // dahfcbge
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dahfbcge
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); // dahbfcge
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dabhfcge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dafhcbge
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); // dafhbcge
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v7); // dafbhcge
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); // dabfhcge
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); // dafchbge
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dafcbhge
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); // dafbchge
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dabfchge
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v2, v0)) {
														if (compare(v7, v2)) {
															if (compare(v6, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v6, v7); // hgdcafeb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hdgcafeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); // dhgcafeb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hdcgafeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v6); // dhcgafeb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v7); // hgdcafbe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); // hdgcafbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dhgcafbe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v7); // hdcgafbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dhcgafbe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v6); // hdcagfeb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hdcafgeb
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hdcagfbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdcafgbe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hdcafbge
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v4, v6); // dhcagfeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v4, v5); swap(v5, v6); // dhcafgeb
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v4, v6); swap(v6, v7); // dhcagfbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhcafgbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v4, v5); swap(v5, v7); // dhcafbge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v0)) {
																		swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); // dchgafeb
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v6); // dchagfeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); // dchafgeb
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dchgafbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v6, v7); // dchagfbe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dchafgbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); // dchafbge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); // dcahgfeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dcahfgeb
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dcahgfbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcahfgbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dcahfbge
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcafhgeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dcafhgbe
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); // dcafhbge
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dcafbhge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v0)) {
															if (compare(v6, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hgdacfeb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); // hdgacfeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v6); // dhgacfeb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hdagcfeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); // dhagcfeb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); // hgdacfbe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hdgacfbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v6, v7); // dhgacfbe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); // hdagcfbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dhagcfbe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); // hdacgfeb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hdacfgeb
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hdacgfbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdacfgbe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hdacfbge
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v6); // dhacgfeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); // dhacfgeb
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v6, v7); // dhacgfbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhacfgbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); // dhacfbge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v2)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dahgcfeb
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); // dahcgfeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dahcfgeb
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); // dahgcfbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dahcgfbe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dahcfgbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dahcfbge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v6); // dachgfeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dachfgeb
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dachgfbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dachfgbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dachfbge
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dacfhgeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dacfhgbe
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v7); // dacfhbge
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dacfbhge
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v0)) {
														if (compare(v7, v3)) {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); // hgdcabfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hdgcabfe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hgdacbfe
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdgacbfe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hdagcbfe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgdabcfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hdgabcfe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdagbcfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdabgcfe
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hdcgabfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdcagbfe
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hdcabgfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v7); // hdcabfge
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdacgbfe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hdacbgfe
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); // hdacbfge
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); // hdabcgfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v7); // hdabcfge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhgcabfe
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhcgabfe
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dchgabfe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhcagbfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dchagbfe
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v4, v7); swap(v5, v6); // dhcabgfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v4, v7); // dhcabfge
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); // dchabgfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v7); // dchabfge
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhgacbfe
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhagcbfe
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhacgbfe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); // dhacbgfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v7); // dhacbfge
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); // dhgabcfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dhagbcfe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhabgcfe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dhabcgfe
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); // dhabcfge
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dahgcbfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dahcgbfe
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dahcbgfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); // dahcbfge
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dahgbcfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dahbgcfe
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); // dahbcgfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v7); // dahbcfge
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v6, v1)) {
																		swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcahgbfe
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dcahbgfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); // dcahbfge
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dachgbfe
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dachbgfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v7); // dachbfge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v2, v0)) {
																	if (compare(v6, v5)) {
																		swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); // dcabhgfe
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v7); // dcabhfge
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); // dcabfhge
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		swap(v0, v3); swap(v1, v3); swap(v4, v7); swap(v5, v6); // dacbhgfe
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v4, v7); // dacbhfge
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v4, v5); swap(v5, v7); // dacbfhge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dabhgcfe
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dabhcgfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); // dabhcfge
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); // dabchgfe
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v7); // dabchfge
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); // dabcfhge
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v7, v0)) {
											if (compare(v6, v1)) {
												if (compare(v2, v0)) {
													if (compare(v6, v0)) {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																if (compare(v7, v4)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfegdcab
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfedgcab
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v5, v6); swap(v6, v7); // hfedcgab
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fhegdcab
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); // fhedgcab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v5, v6); // fhedcgab
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fehgdcab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fehdgcab
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fedhgcab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); // fehdcgab
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fedhcgab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); // fedchgab
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfdegcab
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hfdecgab
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v5, v6); swap(v6, v7); // hfdcegab
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fhdegcab
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); // fhdecgab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v5, v6); // fhdcegab
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdhegcab
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fdhecgab
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); // fdhcegab
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v5, v6); // fdchegab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fdehgcab
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); // fdehcgab
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fdechgab
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fdcehgab
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hdfegcab
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); // dhfegcab
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dfhegcab
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfehgcab
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdfecgab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dhfecgab
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfhecgab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dfehcgab
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfechgab
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hdfcegab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); // dhfcegab
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dfhcegab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dfchegab
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfcehgab
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hdcfegab
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v5, v6); // dhcfegab
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); // dchfegab
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dcfhegab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcfehgab
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v5, v7); // hfedcagb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); // fhedcagb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v7); // fehdcagb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); // fedhcagb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v7); // fedchagb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); // hfdecagb
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); // fhdecagb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); // fdhecagb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); // fdehcagb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); // fdechagb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v5, v7); // hfdceagb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); // fhdceagb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); // fdhceagb
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); // fdcheagb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v7); // fdcehagb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); // hfdcaegb
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v5); // fhdcaegb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); // fdhcaegb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v5); // fdchaegb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hdfecagb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); // hdfceagb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hdfcaegb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dhfecagb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); // dhfceagb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dhfcaegb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dfhecagb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); // dfehcagb
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); // dfhceagb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dfhcaegb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v4, v0)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v5, v7); // dfcheagb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dfchaegb
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dfechagb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dfcehagb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v5, v7); // hdcfeagb
																		}
																		else {
																			if (compare(v5, v0)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hdcfaegb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v5); // hdcafegb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v5); // dhcfeagb
																		}
																		else {
																			if (compare(v5, v0)) {
																				swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v5); // dhcfaegb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v4, v5); // dhcafegb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); // dchfeagb
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); // dcfheagb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dcfehagb
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); // dchfaegb
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dcfhaegb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v5); // dchafegb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v3)) {
														if (compare(v7, v5)) {
															if (compare(v5, v3)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hfegdacb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); // hfedgacb
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfegdabc
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfedgabc
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfedagcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); // hfedacgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); // hfedagbc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hfdegacb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfdegabc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				if (compare(v6, v2)) {
																					swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfdeagcb
																				}
																				else {
																					swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hfdeacgb
																				}
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hfdeagbc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hfdaegcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); // hfdaegbc
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); // hfdaecgb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hfdacegb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdfegacb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hdfegabc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hdfeagcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdfeagbc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); // hdfaegcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hdfaegbc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hdafegcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); // hdafegbc
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); // hdfeacgb
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); // hdfaecgb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); // hdfacegb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); // hdafecgb
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hdafcegb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); // hdacfegb
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); // fhegdacb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fehgdacb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v6); // fhedgacb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v6, v7); // fehdgacb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); // fhdegacb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fhegdabc
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); // fehgdabc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v6, v7); // fhedgabc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v6); // fehdgabc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fhdegabc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v4)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); // fhedagcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v5); // fhedacgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhedagbc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fehdagcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); // fehdacgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); // fehdagbc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fhdeagcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fhdeacgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhdeagbc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); // fhdaegcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fhdaegbc
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); // fhdaecgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fhdacegb
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v7, v4)) {
																if (compare(v4, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fdhegacb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); // fdhegabc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdheagcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fdheacgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fdheagbc
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fdhaegcb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); // fdhaegbc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); // fdhaecgb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fdhacegb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fedhgacb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); // fedhgabc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fedhagcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fedhacgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fedhagbc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); // fdehgacb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fdehgabc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fdehagcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); // fdehacgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdehagbc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dhfegacb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhfegabc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); // dhfeagcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dhfeagbc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v5, v6); // dhfaegcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v5, v6); swap(v6, v7); // dhfaegbc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); // dhafegcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dhafegbc
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); // dhfeacgb
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); // dhfaecgb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v5); // dhfacegb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); // dhafecgb
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dhafcegb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v5); // dhacfegb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfhegacb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dfehgacb
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				if (compare(v7, v4)) {
																					swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dfheagcb
																				}
																				else {
																					swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfehagcb
																				}
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v5, v6); swap(v6, v7); // dfhaegcb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); // dfheacgb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dfehacgb
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v5, v7); // dfhaecgb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); // dfhacegb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dfhegabc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfehgabc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); // dfheagbc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dfehagbc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v5, v6); // dfhaegbc
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v5, v3)) {
														if (compare(v2, v1)) {
															if (compare(v2, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v5, v7); swap(v6, v7); // hfedcabg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v6, v7); // fhedcabg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v6, v7); // fehdcabg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fedhcabg
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v6, v7); // fedchabg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hfdecabg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v6, v7); // fhdecabg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fdhecabg
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v6, v7); // fdehcabg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fdechabg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfedacbg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v6, v7); // fhedacbg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fehdacbg
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fedhacbg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfdeacbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fhdeacbg
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdheacbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fdehacbg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hfdaecbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v6, v7); // fhdaecbg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fdhaecbg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v7); // hfedabgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); // fhedabgc
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v5); // fehdabgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); // fedhabgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); // hfdeabgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fhdeabgc
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); // fdheabgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fdehabgc
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); // hfdaebgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); // fhdaebgc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); // fdhaebgc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); // hfdabegc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fhdabegc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); // fdhabegc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v6, v7); // hfedabcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhedabcg
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v6, v7); // fehdabcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fedhabcg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hfdeabcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhdeabcg
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fdheabcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdehabcg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v6, v7); // hfdaebcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fhdaebcg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v6, v7); // fdhaebcg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hfdabecg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhdabecg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fdhabecg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v2, v1)) {
																if (compare(v2, v0)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdfecabg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dhfecabg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfhecabg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dfehcabg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfechabg
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hdfeacbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v6, v7); // dhfeacbg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dfheacbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfehacbg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hdfeabgc
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); // dhfeabgc
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); // dfheabgc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); // dfehabgc
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdfeabcg
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dhfeabcg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v6, v7); // dfheabcg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dfehabcg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v6, v7); // hdfaecbg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v6, v7); // dhfaecbg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v5, v7); swap(v6, v7); // dfhaecbg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); // hdfaebgc
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v5, v7); // dhfaebgc
																				}
																				else {
																					swap(v0, v3); swap(v1, v5); swap(v2, v7); // dfhaebgc
																				}
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hdfaebcg
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v5, v7); swap(v6, v7); // dhfaebcg
																				}
																				else {
																					swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v6, v7); // dfhaebcg
																				}
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hdfabegc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); // dhfabegc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v5); // dfhabegc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdfabecg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhfabecg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v6, v7); // dfhabecg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hdafecbg
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); // hdafebgc
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v6, v7); // hdafebcg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); // hdafbegc
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hdafbecg
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v5); // hdabfegc
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v6, v7); // hdabfecg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v6, v7); // dhafecbg
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); // dhafebgc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dhafebcg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dhafbegc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhafbecg
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); // dhabfegc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dhabfecg
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v7, v5)) {
															if (compare(v5, v2)) {
																if (compare(v2, v0)) {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v5, v7); swap(v6, v7); // hfdceabg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hdfceabg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v6, v7); // dhfceabg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfdcaebg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdfcaebg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dhfcaebg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfdacebg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hdfacebg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v6, v7); // dhfacebg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdafcebg
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dhafcebg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hdcfeabg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v6, v7); // dhcfeabg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v6, v7); // dchfeabg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdcfaebg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dhcfaebg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dchfaebg
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hdcafebg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v4, v5); swap(v6, v7); // dhcafebg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v6, v7); // dchafebg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hdacfebg
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v6, v7); // dhacfebg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v6, v7); // fhdceabg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v6, v7); // fdhceabg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v6, v7); // fdcheabg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fdcehabg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dfhceabg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dfcheabg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfcehabg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dcfheabg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcfehabg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v6, v7); // fhdcaebg
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fdhcaebg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fdchaebg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fhdacebg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdhacebg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfhcaebg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfhacebg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfchaebg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcfhaebg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v6, v7); // hfdcabeg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhdcabeg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdhcabeg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdchabeg
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hdfcabeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhfcabeg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dfhcabeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dfchabeg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hdcfabeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhcfabeg
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dchfabeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dcfhabeg
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdcafbeg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhcafbeg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dchafbeg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hdcabfeg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v4, v7); swap(v6, v7); // dhcabfeg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v6, v7); // dchabfeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hfdacbeg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhdacbeg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fdhacbeg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdfacbeg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhfacbeg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v6, v7); // dfhacbeg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfdabceg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fhdabceg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdhabceg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hdfabceg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v6, v7); // dhfabceg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfhabceg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hdafcbeg
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdacfbeg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hdacbfeg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhafcbeg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhacfbeg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v6, v7); // dhacbfeg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdafbceg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdabfceg
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v6, v7); // hdabcfeg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dhafbceg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhabfceg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dhabcfeg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v2)) {
												if (compare(v7, v4)) {
													if (compare(v4, v1)) {
														if (compare(v6, v1)) {
															if (compare(v5, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); // fdahegcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dfahegcb
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fdahegbc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); // dfahegbc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); // fdahecgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); // dfahecgb
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); // fdahcegb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dfahcegb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dahfegcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); // dafhegcb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); // dahfegbc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dafhegbc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); // dahfecgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); // dafhecgb
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dahfcegb
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); // dafhcegb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); // dahcfegb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v6, v7); // fdahecbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dfahecbg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fdahcebg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfahcebg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); // fdahebgc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); // dfahebgc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fdahebcg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v6, v7); // dfahebcg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dahfecbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v6, v7); // dafhecbg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dahfcebg
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dafhcebg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dahcfebg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); // dahfebgc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); // dafhebgc
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v6, v7); // dahfebcg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dafhebcg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v7, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fdahbegc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); // dfahbegc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); // dahfbegc
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dafhbegc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v5); // dahbfegc
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdahbecg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dfahbecg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dahfbecg
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dafhbecg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v6, v7); // dahbfecg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); // fdabhegc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dfabhegc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); // dafbhegc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); // dabhfegc
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dabfhegc
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdabhecg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfabhecg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dafbhecg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dabhfecg
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dabfhecg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v5, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dahfcbeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dahfbceg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dahbfceg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dabhfceg
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dahcfbeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dahcbfeg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v6, v7); // dahbcfeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dabhcfeg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdahcbeg
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fdahbceg
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v6, v7); // fdabhceg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dfahcbeg
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfahbceg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dfabhceg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dafhcbeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dafhbceg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v6, v7); // dafbhceg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dabfhceg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v2)) {
														if (compare(v5, v3)) {
															if (compare(v6, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fedahgcb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdeahgcb
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fdaehgcb
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fedahgbc
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fdeahgbc
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdaehgbc
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fedahbgc
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); // fdeahbgc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); // fedabhgc
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fdeabhgc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fdaehbgc
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); // fdaebhgc
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v5, v7); // fdabehgc
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfeahgcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); // dfeahgbc
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v7); // dfeahbgc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); // dfeabhgc
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfaehgcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dfaehgbc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); // dfaehbgc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dfaebhgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); // dfabehgc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v2, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dafehgcb
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dafehgbc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dafehbgc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); // dafebhgc
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v5, v7); // dafbehgc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); // dabfehgc
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v2, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); // fedahcgb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fdeahcgb
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); // fdaehcgb
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fedahcbg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdeahcbg
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fdaehcbg
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fedahbcg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fdeahbcg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fedabhcg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdeabhcg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdaehbcg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fdaebhcg
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v5, v7); swap(v6, v7); // fdabehcg
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); // dfeahcgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfeahcbg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v6, v7); // dfeahbcg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfeabhcg
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dfaehcgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfaehcbg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dfaehbcg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfaebhcg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dfabehcg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); // dafehcgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dafehcbg
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dafehbcg
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dafebhcg
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v5, v7); swap(v6, v7); // dafbehcg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dabfehcg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v1)) {
													if (compare(v2, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); // fedcahgb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fdecahgb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); // dfecahgb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fdceahgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v5, v7); // dfceahgb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fedcahbg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdecahbg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dfecahbg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdceahbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dfceahbg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fdcahegb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v5, v7); // fdcaehgb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdcahebg
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fdcaehbg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fdcahbeg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v4, v7); swap(v5, v7); // dfcahegb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v5, v7); // dfcaehgb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfcahebg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v5, v7); swap(v6, v7); // dfcaehbg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v4, v7); swap(v6, v7); // dfcahbeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v0)) {
																		swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); // dcfeahgb
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); // dcfahegb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v5, v7); // dcfaehgb
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcfahebg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v6, v7); // dcfahbeg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dcfeahbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v5, v7); swap(v6, v7); // dcfaehbg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); // dcahfegb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dcahfebg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcahfbeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dcahbfeg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dcafhegb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); // dcafehgb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcafhebg
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dcafehbg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dcafhbeg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v2)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fedachgb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); // fdeachgb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v5, v7); // dfeachgb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fdaechgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); // dfaechgb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fedachbg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fdeachbg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v5, v7); swap(v6, v7); // dfeachbg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdaechbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dfaechbg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fdachegb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); // fdacehgb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdachebg
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fdacehbg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fdachbeg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); // dfachegb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v5, v7); // dfacehgb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfachebg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v5, v7); swap(v6, v7); // dfacehbg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v6, v7); // dfachbeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dafechgb
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dafchegb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); // dafcehgb
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dafchebg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dafchbeg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dafechbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dafcehbg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v5); // dachfegb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dachfebg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dachfbeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dachbfeg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dacfhegb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v5, v7); // dacfehgb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dacfhebg
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dacfehbg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dacfhbeg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v6, v7); // fedcabhg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fdecabhg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fedacbhg
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v6, v7); // fdeacbhg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fdaecbhg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v6, v7); // fedabchg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fdeabchg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v6, v7); // fdaebchg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v6, v7); // fdabechg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fdceabhg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v6, v7); // fdcaebhg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdcabheg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fdcabehg
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v6, v7); // fdacebhg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdacbheg
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fdacbehg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdabcheg
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v6, v7); // fdabcehg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v6, v7); // dfecabhg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v6, v7); // dfceabhg
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v6, v7); // dcfeabhg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v6, v7); // dfcaebhg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v6, v7); // dcfaebhg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfcabheg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v4, v5); swap(v6, v7); // dfcabehg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcfabheg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v6, v7); // dcfabehg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v6, v7); // dfeacbhg
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v6, v7); // dfaecbhg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v6, v7); // dfacebhg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfacbheg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v6, v7); // dfacbehg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v6, v7); // dfeabchg
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dfaebchg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v6, v7); // dfabechg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfabcheg
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dfabcehg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dafecbhg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v6, v7); // dafcebhg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dafcbheg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dafcbehg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v6, v7); // dafebchg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v6, v7); // dafbechg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dafbcheg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v6, v7); // dafbcehg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v4, v1)) {
																		swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v6, v7); // dcafebhg
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcafbheg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dcafbehg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v6, v7); // dacfebhg
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dacfbheg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dacfbehg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v2, v0)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v6, v7); // dcabhfeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcabfheg
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v6, v7); // dcabfehg
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v4, v7); swap(v6, v7); // dacbhfeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dacbfheg
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v4, v5); swap(v6, v7); // dacbfehg
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v6, v7); // dabfechg
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dabfcheg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dabfcehg
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v6, v7); // dabchfeg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dabcfheg
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v6, v7); // dabcfehg
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v7, v4)) {
										if (compare(v5, v0)) {
											if (compare(v4, v1)) {
												if (compare(v2, v0)) {
													if (compare(v4, v0)) {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); // ghfedcab
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); // gfhedcab
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); // ghfdecab
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v5, v7); // gfhdecab
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); // gfdhecab
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fghedcab
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v5, v7); swap(v6, v7); // fghdecab
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fgdhecab
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); // fdghecab
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v5); // ghfdceab
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); // gfhdceab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fghdceab
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gfdhceab
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgdhceab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fdghceab
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); // gfdcheab
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgdcheab
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdgcheab
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdcgheab
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); // ghdfecab
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); // gdhfecab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dghfecab
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); // ghdfceab
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gdhfceab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dghfceab
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v5); // ghdcfeab
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); // gdhcfeab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dghcfeab
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v5); // gdchfeab
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dgchfeab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dcghfeab
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); // gdfhecab
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v6, v7); // dgfhecab
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dfghecab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); // gdfhceab
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dgfhceab
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfghceab
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gdfcheab
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgfcheab
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dfgcheab
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dfcgheab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gdcfheab
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgcfheab
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcgfheab
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dcfgheab
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v7, v2)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); // ghfdcaeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfhdcaeb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v6, v7); // fghdcaeb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfdhcaeb
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fgdhcaeb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ghdfcaeb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdhfcaeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gdfhcaeb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); // ghdcfaeb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gdhcfaeb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfdchaeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v6, v7); // fgdchaeb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdfchaeb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gdchfaeb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdcfhaeb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); // gfdcaheb
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgdcaheb
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gdfcaheb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gdcfaheb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dghfcaeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgfhcaeb
																			}
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dgfchaeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgfcaheb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dghcfaeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgchfaeb
																			}
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dgcfhaeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgcfaheb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); // fdghcaeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fdgchaeb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdgcaheb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dfghcaeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfgchaeb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dfgcaheb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v0)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fdcghaeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdcgaheb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fdcagheb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v7, v0)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfcghaeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dfcgaheb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v4, v6); swap(v5, v7); // dfcagheb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcghfaeb
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dcgfhaeb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcfghaeb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcgfaheb
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dcfgaheb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); // dcfagheb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v6, v5)) {
															if (compare(v4, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); // ghfedacb
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gfhedacb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v5, v6); // ghfdeacb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v5, v6); swap(v6, v7); // gfhdeacb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gfdheacb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ghfedabc
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); // gfhedabc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v5, v6); swap(v6, v7); // ghfdeabc
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v5, v6); // gfhdeabc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); // gfdheabc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); // ghfdaecb
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v6); // ghfdaceb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghfdaebc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfhdaecb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); // gfhdaceb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); // gfhdaebc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfdhaecb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gfdhaceb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gfdhaebc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gfdahecb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfdahebc
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfdahceb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gfdacheb
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v6, v7); // fghedacb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v6, v7); // fghdeacb
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); // fghedabc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); // fghdeabc
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v6, v7); // fghdaecb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fghdaceb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v5); // fghdaebc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v6, v7); // fgdheacb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); // fgdheabc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				if (compare(v4, v2)) {
																					swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fgdhaecb
																				}
																				else {
																					swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgdhaceb
																				}
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); // fgdhaebc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgdahecb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fgdahebc
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fgdahceb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgdacheb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); // fdgheacb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v6, v7); // fdgheabc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); // fdghaecb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fdghaebc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fdgahecb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdgahebc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdaghecb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fdaghebc
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fdghaceb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdgahceb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fdgacheb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fdaghceb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdagcheb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fdacgheb
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v7, v5)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); // ghdfeacb
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ghdfeabc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ghdfaecb
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); // ghdfaceb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghdfaebc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gdhfeacb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); // gdhfeabc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdhfaecb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gdhfaceb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gdhfaebc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); // gdfheacb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gdfheabc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gdfhaecb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); // gdfhaceb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdfhaebc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdfahecb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gdfahebc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gdfahceb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdfacheb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v6, v7); // dghfeacb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dgfheacb
																			}
																		}
																		else {
																			if (compare(v7, v0)) {
																				if (compare(v7, v5)) {
																					swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dghfaecb
																				}
																				else {
																					swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgfhaecb
																				}
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgfahecb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dghfaceb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dgfhaceb
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v6, v7); // dgfahceb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgfacheb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); // dghfeabc
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); // dgfheabc
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); // dghfaebc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dgfhaebc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); // dgfahebc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); // dfgheacb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dfgheabc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dfghaecb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfghaebc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); // dfgahecb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfgahebc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dfaghecb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfaghebc
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfghaceb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfgahceb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); // dfgacheb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfaghceb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dfagcheb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); // dfacgheb
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v2)) {
													if (compare(v5, v3)) {
														if (compare(v7, v0)) {
															if (compare(v2, v1)) {
																if (compare(v2, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghfdcabe
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); // gfhdcabe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v7); // fghdcabe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gfdhcabe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); // fgdhcabe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fdghcabe
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v6, v7); // ghfdacbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfhdacbe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); // fghdacbe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfdhacbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fgdhacbe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdghacbe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); // ghfdabec
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v6); // gfhdabec
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); // fghdabec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); // gfdhabec
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fgdhabec
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdghabec
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghfdabce
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v6, v7); // gfhdabce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fghdabce
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gfdhabce
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgdhabce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fdghabce
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gfdahcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gfdahbec
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); // gfdahbce
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); // fgdahcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgdahbec
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fgdahbce
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fdgahcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fdgahbec
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); // fdgahbce
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); // fdaghcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fdaghbec
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fdaghbce
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfdabhec
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fgdabhec
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdgabhec
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fdagbhec
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); // fdabghec
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gfdabhce
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgdabhce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fdgabhce
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdagbhce
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdabghce
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v0)) {
															if (compare(v6, v3)) {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghdfcabe
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gdhfcabe
																				}
																				else {
																					swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdfhcabe
																				}
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghdcfabe
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdhcfabe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ghdfacbe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdhfacbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gdfhacbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ghdfabec
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); // gdhfabec
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gdfhabec
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghdfabce
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gdhfabce
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdfhabce
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v7, v5)) {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); // dghfcabe
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dghcfabe
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dgfhcabe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfghcabe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dghfacbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); // dgfhacbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dfghacbe
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dghfabec
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgfhabec
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dfghabec
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dghfabce
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dgfhabce
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); // dfghabce
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); // gdfahcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdfahbec
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gdfahbce
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v7); // dgfahcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgfahbec
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); // dgfahbce
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); // dfgahcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v6, v7); // dfgahbec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v7); // dfgahbce
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dfaghcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dfaghbec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); // dfaghbce
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gdfabhec
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); // dgfabhec
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfgabhec
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dfagbhec
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dfabghec
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdfabhce
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgfabhce
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); // dfgabhce
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dfagbhce
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfabghce
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v6, v5)) {
															if (compare(v7, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); // gfdchabe
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gdfchabe
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdchfabe
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gdcfhabe
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); // dgfchabe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dgchfabe
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v7); // dgcfhabe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dcghfabe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); // dcgfhabe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfdcahbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdfcahbe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdcfahbe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfdachbe
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gdfachbe
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dgfcahbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); // dgfachbe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dgcfahbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dcgfahbe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v2, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v7); // fgdchabe
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); // fdgchabe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dfgchabe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); // fgdcahbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fdgcahbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfgcahbe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fgdachbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdgachbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfgachbe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fdagchbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfagchbe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v7); // fdcghabe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dfcghabe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dcfghabe
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fdcgahbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfcgahbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcfgahbe
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdcaghbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfcaghbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcfaghbe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdacghbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfacghbe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v6, v2)) {
																if (compare(v6, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v6, v7); // gfdcabhe
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgdcabhe
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gdfcabhe
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gdcfabhe
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgfcabhe
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgcfabhe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdgcabhe
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dfgcabhe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdcgabhe
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dfcgabhe
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcgfabhe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dcfgabhe
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fdcagbhe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v4, v6); swap(v6, v7); // dfcagbhe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v6, v7); // dcfagbhe
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdcabghe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfcabghe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcfabghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gfdacbhe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgdacbhe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fdgacbhe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdfacbhe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgfacbhe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v6, v7); // dfgacbhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfdabche
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fgdabche
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdgabche
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gdfabche
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v6, v7); // dgfabche
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfgabche
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdagcbhe
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fdacgbhe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdacbghe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dfagcbhe
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v6, v7); // dfacgbhe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfacbghe
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fdagbche
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v6, v7); // fdabgche
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdabcghe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfagbche
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dfabgche
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfabcghe
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v2)) {
												if (compare(v7, v5)) {
													if (compare(v4, v2)) {
														if (compare(v6, v3)) {
															if (compare(v4, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); // ghdafecb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gdhafecb
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); // gdahfecb
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ghdafebc
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); // gdhafebc
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gdahfebc
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghdafbec
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gdhafbec
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ghdabfec
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); // gdhabfec
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdahfbec
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gdahbfec
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v6, v7); // gdabhfec
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v7, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v6, v7); // dghafecb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v5); // dghafebc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); // dghafbec
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v6); // dghabfec
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dgahfecb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); // dgahfebc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // dgahfbec
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); // dgahbfec
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dgabhfec
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v2, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); // daghfecb
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // daghfebc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // daghfbec
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // daghbfec
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v6, v7); // dagbhfec
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dabghfec
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v2, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ghdafceb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdhafceb
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gdahfceb
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghdafcbe
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gdhafcbe
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdahfcbe
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ghdafbce
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdhafbce
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); // ghdabfce
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gdhabfce
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gdahfbce
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); // gdahbfce
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v7); // gdabhfce
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v7, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dghafceb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); // dghafcbe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dghafbce
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v6, v7); // dghabfce
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgahfceb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dgahfcbe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgahfbce
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dgahbfce
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); // dgabhfce
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // daghfceb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // daghfcbe
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // daghfbce
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); // daghbfce
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v7); // dagbhfce
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); // dabghfce
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v4, v1)) {
															if (compare(v6, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gdafhecb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgafhecb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdafhebc
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dgafhebc
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdafhceb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dgafhceb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gdafcheb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgafcheb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // dagfhecb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dafghecb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dagfhebc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dafghebc
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dagfhceb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dafghceb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dagfcheb
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dafgcheb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dafcgheb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gdafhcbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); // dgafhcbe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdafchbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dgafchbe
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gdafhbec
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgafhbec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); // gdafhbce
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dgafhbce
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dagfhcbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); // dafghcbe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dagfchbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dafgchbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dafcghbe
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dagfhbec
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dafghbec
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); // dagfhbce
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dafghbce
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v5, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdafbhec
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dgafbhec
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dagfbhec
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dafgbhec
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); // dafbghec
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gdafbhce
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgafbhce
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dagfbhce
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dafgbhce
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dafbghce
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdabfhec
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgabfhec
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dagbfhec
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dabgfhec
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dabfghec
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); // gdabfhce
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dgabfhce
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); // dagbfhce
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dabgfhce
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dabfghce
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v6, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gdafcbhe
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdafbche
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdabfche
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgafcbhe
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dgafbche
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgabfche
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dagfcbhe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dagfbche
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dagbfche
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dabgfche
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dafgcbhe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dafgbche
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v6, v7); // dafbgche
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dabfgche
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dafcgbhe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dafcbghe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dafbcghe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dabfcghe
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v2, v0)) {
														if (compare(v6, v2)) {
															if (compare(v7, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v6); // ghdcafeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); // gdhcafeb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dghcafeb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v6); // gdchafeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dgchafeb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v6, v7); // ghdcafbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gdhcafbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); // dghcafbe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gdchafbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v7); // dgchafbe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gdcahfeb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdcafheb
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v7); // gdcahfbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gdcafhbe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdcafbhe
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v4, v7); swap(v6, v7); // dgcahfeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgcafheb
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v4, v7); // dgcahfbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v4, v5); swap(v5, v7); // dgcafhbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgcafbhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v0)) {
																		swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dcghafeb
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v6, v7); // dcgahfeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcgafheb
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); // dcghafbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v7); // dcgahfbe
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); // dcgafhbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcgafbhe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dcaghfeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcagfheb
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); // dcaghfbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dcagfhbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcagfbhe
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dcafgheb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcafghbe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dcafgbhe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcafbghe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v7, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); // ghdacfeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gdhacfeb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v6, v7); // dghacfeb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); // gdahcfeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dgahcfeb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ghdacfbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); // gdhacfbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v7); // dghacfbe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gdahcfbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); // dgahcfbe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gdachfeb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdacfheb
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); // gdachfbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gdacfhbe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdacfbhe
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v6, v7); // dgachfeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgacfheb
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v7); // dgachfbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); // dgacfhbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgacfbhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v2)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); // daghcfeb
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dagchfeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dagcfheb
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // daghcfbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); // dagchfbe
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dagcfhbe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dagcfbhe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // dacghfeb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dacgfheb
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v7); // dacghfbe
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dacgfhbe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dacgfbhe
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // dacfgheb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dacfghbe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dacfgbhe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dacfbghe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v0)) {
														if (compare(v6, v3)) {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghdcabfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdhcabfe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghdacbfe
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gdhacbfe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdahcbfe
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ghdabcfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdhabcfe
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gdahbcfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); // gdabhcfe
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdchabfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gdcahbfe
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdcabhfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gdcabfhe
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gdachbfe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdacbhfe
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gdacbfhe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdabchfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v6, v7); // gdabcfhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dghcabfe
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dgchabfe
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dcghabfe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v4, v7); swap(v5, v6); // dgcahbfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); // dcgahbfe
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgcabhfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v4, v6); swap(v6, v7); // dgcabfhe
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcgabhfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v6, v7); // dcgabfhe
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); // dghacbfe
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dgahcbfe
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); // dgachbfe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgacbhfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v6, v7); // dgacbfhe
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dghabcfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgahbcfe
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dgabhcfe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgabchfe
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dgabcfhe
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // daghcbfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dagchbfe
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dagcbhfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dagcbfhe
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // daghbcfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); // dagbhcfe
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dagbchfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v6, v7); // dagbcfhe
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v7, v1)) {
																		swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dcaghbfe
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcagbhfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dcagbfhe
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dacghbfe
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dacgbhfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dacgbfhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v2, v0)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcabghfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v6, v7); // dcabgfhe
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcabfghe
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dacbghfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v4, v6); swap(v6, v7); // dacbgfhe
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dacbfghe
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dabghcfe
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dabgchfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dabgcfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dabcghfe
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v6, v7); // dabcgfhe
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dabcfghe
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v0)) {
											if (compare(v7, v1)) {
												if (compare(v2, v0)) {
													if (compare(v7, v0)) {
														if (compare(v5, v3)) {
															if (compare(v4, v3)) {
																if (compare(v6, v4)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gfehdcab
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); // gfedhcab
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v5, v7); // gfedchab
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgehdcab
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgedhcab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v5, v7); swap(v6, v7); // fgedchab
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // feghdcab
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); // fegdhcab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fedghcab
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fegdchab
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fedgchab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fedcghab
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gfdehcab
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); // gfdechab
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v5, v7); // gfdcehab
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgdehcab
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fgdechab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v5, v7); swap(v6, v7); // fgdcehab
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fdgehcab
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdgechab
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fdgcehab
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fdcgehab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fdeghcab
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fdegchab
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdecghab
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdceghab
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); // gdfehcab
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dgfehcab
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfgehcab
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dfeghcab
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gdfechab
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgfechab
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dfgechab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); // dfegchab
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dfecghab
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); // gdfcehab
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dgfcehab
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); // dfgcehab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v5, v7); // dfcgehab
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dfceghab
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v5, v7); // gdcfehab
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dgcfehab
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dcgfehab
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); // dcfgehab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dcfeghab
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v5, v6); swap(v6, v7); // gfedcahb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v6, v7); // fgedcahb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v6, v7); // fegdcahb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fedgcahb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v6, v7); // fedcgahb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gfdecahb
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v6, v7); // fgdecahb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fdgecahb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v6, v7); // fdegcahb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fdecgahb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v5, v6); swap(v6, v7); // gfdceahb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v6, v7); // fgdceahb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v6, v7); // fdgceahb
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v6, v7); // fdcgeahb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fdcegahb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfdcaehb
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v6, v7); // fgdcaehb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fdgcaehb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fdcgaehb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdfecahb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gdfceahb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdfcaehb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dgfecahb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v6, v7); // dgfceahb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dgfcaehb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfgecahb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dfegcahb
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dfgceahb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfgcaehb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v4, v0)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dfcgeahb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfcgaehb
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfecgahb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfcegahb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gdcfeahb
																		}
																		else {
																			if (compare(v5, v0)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdcfaehb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gdcafehb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v6, v7); // dgcfeahb
																		}
																		else {
																			if (compare(v5, v0)) {
																				swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dgcfaehb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v4, v5); swap(v6, v7); // dgcafehb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v6, v7); // dcgfeahb
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dcfgeahb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcfegahb
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dcgfaehb
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcfgaehb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v6, v7); // dcgafehb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v3)) {
														if (compare(v6, v5)) {
															if (compare(v5, v3)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfehdacb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfedhacb
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gfehdabc
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); // gfedhabc
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); // gfedahcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfedachb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfedahbc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfdehacb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gfdehabc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				if (compare(v7, v2)) {
																					swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gfdeahcb
																				}
																				else {
																					swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfdeachb
																				}
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfdeahbc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); // gfdaehcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gfdaehbc
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gfdaechb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfdacehb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gdfehacb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdfehabc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdfeahcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gdfeahbc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gdfaehcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); // gdfaehbc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); // gdafehcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gdafehbc
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gdfeachb
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v6, v7); // gdfaechb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gdfacehb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gdafechb
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdafcehb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gdacfehb
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fgehdacb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); // feghdacb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v6, v7); // fgedhacb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v7); // fegdhacb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fgdehacb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); // fgehdabc
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // feghdabc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v7); // fgedhabc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v6, v7); // fegdhabc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); // fgdehabc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v4)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgedahcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v6, v7); // fgedachb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); // fgedahbc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); // fegdahcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fegdachb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fegdahbc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgdeahcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fgdeachb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fgdeahbc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fgdaehcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); // fgdaehbc
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v6, v7); // fgdaechb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fgdacehb
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v6, v4)) {
																if (compare(v4, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); // fdgehacb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fdgehabc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fdgeahcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdgeachb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdgeahbc
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); // fdgaehcb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fdgaehbc
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fdgaechb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdgacehb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); // fedghacb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fedghabc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fedgahcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fedgachb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fedgahbc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fdeghacb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); // fdeghabc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fdegahcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fdegachb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fdegahbc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgfehacb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dgfehabc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dgfeahcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); // dgfeahbc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v5, v7); swap(v6, v7); // dgfaehcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v5, v7); // dgfaehbc
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dgafehcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); // dgafehbc
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v6, v7); // dgfeachb
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v6, v7); // dgfaechb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v6, v7); // dgfacehb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v6, v7); // dgafechb
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dgafcehb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v6, v7); // dgacfehb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dfgehacb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dfeghacb
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				if (compare(v6, v4)) {
																					swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); // dfgeahcb
																				}
																				else {
																					swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dfegahcb
																				}
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v5, v7); // dfgaehcb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dfgeachb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfegachb
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v5, v6); swap(v6, v7); // dfgaechb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfgacehb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dfgehabc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dfeghabc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dfgeahbc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfegahbc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v5, v7); swap(v6, v7); // dfgaehbc
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v2)) {
													if (compare(v5, v3)) {
														if (compare(v2, v1)) {
															if (compare(v2, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v5, v6); // gfedcabh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); // fgedcabh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v6); // fegdcabh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // fedgcabh
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v6); // fedcgabh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // gfdecabh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); // fgdecabh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // fdgecabh
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); // fdegcabh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // fdecgabh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // gfedacbh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v5); // fgedacbh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // fegdacbh
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fedgacbh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gfdeacbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fgdeacbh
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fdgeacbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // fdegacbh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // gfdaecbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); // fgdaecbh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // fdgaecbh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v6, v7); // gfedabhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgedabhc
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v6, v7); // fegdabhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fedgabhc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gfdeabhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgdeabhc
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fdgeabhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdegabhc
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v6, v7); // gfdaebhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fgdaebhc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v6, v7); // fdgaebhc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gfdabehc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgdabehc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fdgabehc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v4, v6); // gfedabch
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // fgedabch
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v4, v5); // fegdabch
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // fedgabch
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // gfdeabch
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fgdeabch
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // fdgeabch
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fdegabch
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); // gfdaebch
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // fgdaebch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); // fdgaebch
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // gfdabech
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fgdabech
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // fdgabech
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v2, v1)) {
																if (compare(v2, v0)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gdfecabh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dgfecabh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dfgecabh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // dfegcabh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dfecgabh
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // gdfeacbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); // dgfeacbh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // dfgeacbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dfegacbh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdfeabhc
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dgfeabhc
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v6, v7); // dfgeabhc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dfegabhc
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gdfeabch
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // dgfeabch
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v4); // dfgeabch
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // dfegabch
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); // gdfaecbh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); // dgfaecbh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v5, v6); // dfgaecbh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gdfaebhc
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v5, v6); swap(v6, v7); // dgfaebhc
																				}
																				else {
																					swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v6, v7); // dfgaebhc
																				}
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // gdfaebch
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v5, v6); // dgfaebch
																				}
																				else {
																					swap(v0, v3); swap(v1, v5); swap(v2, v6); // dfgaebch
																				}
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdfabehc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgfabehc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v6, v7); // dfgabehc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gdfabech
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // dgfabech
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v5); // dfgabech
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // gdafecbh
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v6, v7); // gdafebhc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); // gdafebch
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gdafbehc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // gdafbech
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v6, v7); // gdabfehc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v5); // gdabfech
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); // dgafecbh
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dgafebhc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // dgafebch
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgafbehc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dgafbech
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dgabfehc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // dgabfech
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v6, v5)) {
															if (compare(v5, v2)) {
																if (compare(v2, v0)) {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v5, v6); // gfdceabh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // gdfceabh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); // dgfceabh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // gfdcaebh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gdfcaebh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dgfcaebh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gfdacebh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // gdfacebh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v5); // dgfacebh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gdafcebh
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dgafcebh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v5, v6); // gdcfeabh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v5); // dgcfeabh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); // dcgfeabh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gdcfaebh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v5); // dgcfaebh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // dcgfaebh
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v5); // gdcafebh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v4, v5); // dgcafebh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v5); // dcgafebh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // gdacfebh
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v5); // dgacfebh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); // fgdceabh
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); // fdgceabh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); // fdcgeabh
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v6); // fdcegabh
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // dfgceabh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v5, v6); // dfcgeabh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dfcegabh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // dcfgeabh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dcfegabh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v5); // fgdcaebh
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // fdgcaebh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v5); // fdcgaebh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fgdacebh
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fdgacebh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dfgcaebh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // dfgacebh
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dfcgaebh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dcfgaebh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v5, v2)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v4, v6); // gfdcabeh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // fgdcabeh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fdgcabeh
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fdcgabeh
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // gdfcabeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dgfcabeh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // dfgcabeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v6); swap(v4, v6); // dfcgabeh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v5); swap(v4, v6); // gdcfabeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dgcfabeh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dcgfabeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // dcfgabeh
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gdcafbeh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v4, v5); swap(v5, v6); // dgcafbeh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // dcgafbeh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v6); // gdcabfeh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v4, v6); // dgcabfeh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v6); // dcgabfeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // gfdacbeh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fgdacbeh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // fdgacbeh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gdfacbeh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // dgfacbeh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v6); // dfgacbeh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gfdabceh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // fgdabceh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fdgabceh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // gdfabceh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v5); swap(v4, v6); // dgfabceh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // dfgabceh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // gdafcbeh
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gdacfbeh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // gdacbfeh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dgafcbeh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // dgacfbeh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v6); // dgacbfeh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gdafbceh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // gdabfceh
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v6); // gdabcfeh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // dgafbceh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dgabfceh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // dgabcfeh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v2)) {
												if (compare(v6, v4)) {
													if (compare(v4, v1)) {
														if (compare(v7, v1)) {
															if (compare(v5, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fdagehcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); // dfagehcb
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); // fdagehbc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dfagehbc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v6, v7); // fdagechb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dfagechb
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fdagcehb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfagcehb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); // dagfehcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dafgehcb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dagfehbc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); // dafgehbc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dagfechb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v6, v7); // dafgechb
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dagfcehb
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dafgcehb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dagcfehb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); // fdagecbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // dfagecbh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // fdagcebh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dfagcebh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fdagebhc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v6, v7); // dfagebhc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // fdagebch
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); // dfagebch
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // dagfecbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); // dafgecbh
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dagfcebh
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // dafgcebh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // dagcfebh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v6, v7); // dagfebhc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dafgebhc
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); // dagfebch
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // dafgebch
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdagbehc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dfagbehc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dagfbehc
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dafgbehc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v6, v7); // dagbfehc
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fdagbech
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // dfagbech
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // dagfbech
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dafgbech
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v5); // dagbfech
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdabgehc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfabgehc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dafbgehc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dabgfehc
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dabfgehc
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // fdabgech
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dfabgech
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // dafbgech
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // dabgfech
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dabfgech
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v5, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // dagfcbeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dagfbceh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // dagbfceh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dabgfceh
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dagcfbeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // dagcbfeh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v6); // dagbcfeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // dabgcfeh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fdagcbeh
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // fdagbceh
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v6); // fdabgceh
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // dfagcbeh
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dfagbceh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // dfabgceh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dafgcbeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // dafgbceh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v6); // dafbgceh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // dabfgceh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v2)) {
														if (compare(v5, v3)) {
															if (compare(v7, v1)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fedaghcb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fdeaghcb
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdaeghcb
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fedaghbc
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fdeaghbc
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fdaeghbc
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fedagbhc
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fdeagbhc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fedabghc
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdeabghc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdaegbhc
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fdaebghc
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v5, v6); swap(v6, v7); // fdabeghc
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); // dfeaghcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfeaghbc
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v6, v7); // dfeagbhc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfeabghc
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dfaeghcb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dfaeghbc
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dfaegbhc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dfaebghc
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dfabeghc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v2, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dafeghcb
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dafeghbc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dafegbhc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dafebghc
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v5, v6); swap(v6, v7); // dafbeghc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dabfeghc
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v2, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fedagchb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdeagchb
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fdaegchb
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // fedagcbh
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fdeagcbh
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // fdaegcbh
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fedagbch
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // fdeagbch
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // fedabgch
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fdeabgch
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fdaegbch
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // fdaebgch
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v5, v6); // fdabegch
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v0)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfeagchb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // dfeagcbh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v6); // dfeagbch
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // dfeabgch
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfaegchb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dfaegcbh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // dfaegbch
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dfaebgch
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // dfabegch
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dafegchb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // dafegcbh
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dafegbch
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // dafebgch
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v5, v6); // dafbegch
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // dabfegch
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v1)) {
													if (compare(v2, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fedcaghb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdecaghb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dfecaghb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdceaghb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dfceaghb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // fedcagbh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fdecagbh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // dfecagbh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fdceagbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v3, v4); swap(v5, v6); // dfceagbh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdcagehb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fdcaeghb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fdcagebh
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v5, v6); // fdcaegbh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v6); // fdcagbeh
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfcagehb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v5, v6); swap(v6, v7); // dfcaeghb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v4, v6); swap(v5, v6); // dfcagebh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v5, v6); // dfcaegbh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v4, v6); // dfcagbeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v0)) {
																		swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dcfeaghb
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcfagehb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v5, v6); swap(v6, v7); // dcfaeghb
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // dcfagebh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v6); // dcfagbeh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // dcfeagbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v5, v6); // dcfaegbh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dcagfehb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // dcagfebh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dcagfbeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // dcagbfeh
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcafgehb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dcafeghb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dcafgebh
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // dcafegbh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // dcafgbeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v2)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fedacghb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fdeacghb
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v5, v6); swap(v6, v7); // dfeacghb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fdaecghb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dfaecghb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fedacgbh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // fdeacgbh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v5, v6); // dfeacgbh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fdaecgbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // dfaecgbh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fdacgehb
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fdaceghb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fdacgebh
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // fdacegbh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // fdacgbeh
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dfacgehb
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v5, v6); swap(v6, v7); // dfaceghb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // dfacgebh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v5, v6); // dfacegbh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v6); // dfacgbeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dafecghb
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dafcgehb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dafceghb
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dafcgebh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // dafcgbeh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dafecgbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // dafcegbh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // dacgfehb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v5); // dacgfebh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dacgfbeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v6); // dacgbfeh
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dacfgehb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // dacfeghb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dacfgebh
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v5, v6); // dacfegbh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v6); // dacfgbeh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v4, v5); // fedcabgh
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // fdecabgh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fedacbgh
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); // fdeacbgh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fdaecbgh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v3); swap(v3, v5); // fedabcgh
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // fdeabcgh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v4); // fdaebcgh
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); // fdabecgh
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v4); swap(v4, v5); // fdceabgh
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v5); // fdcaebgh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fdcabgeh
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v3, v5); swap(v4, v5); // fdcabegh
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); // fdacebgh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fdacbgeh
																			}
																			else {
																				swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fdacbegh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // fdabcgeh
																		}
																		else {
																			swap(v0, v5); swap(v1, v3); swap(v2, v5); swap(v4, v5); // fdabcegh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v3, v4); // dfecabgh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v3, v4); // dfceabgh
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v3, v4); // dcfeabgh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v5); // dfcaebgh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); // dcfaebgh
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v4, v5); swap(v5, v6); // dfcabgeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v4, v5); // dfcabegh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // dcfabgeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v5); swap(v4, v5); // dcfabegh
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); // dfeacbgh
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); // dfaecbgh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); // dfacebgh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // dfacbgeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v4, v5); // dfacbegh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v4); swap(v4, v5); // dfeabcgh
																		}
																		else {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dfaebcgh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); // dfabecgh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dfabcgeh
																			}
																			else {
																				swap(v0, v3); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dfabcegh
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // dafecbgh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); // dafcebgh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dafcbgeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // dafcbegh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v3, v4); // dafebcgh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); // dafbecgh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // dafbcgeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v5); swap(v4, v5); // dafbcegh
																		}
																	}
																}
															}
															else {
																if (compare(v2, v0)) {
																	if (compare(v4, v1)) {
																		swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); // dcafebgh
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dcafbgeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dcafbegh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v3, v5); // dacfebgh
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dacfbgeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v5); swap(v4, v5); // dacfbegh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v2, v0)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v6); // dcabgfeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // dcabfgeh
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v5); // dcabfegh
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v4, v6); // dacbgfeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v4, v5); swap(v5, v6); // dacbfgeh
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v4, v5); // dacbfegh
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); // dabfecgh
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dabfcgeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // dabfcegh
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v6); // dabcgfeh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // dabcfgeh
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v5); // dabcfegh
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v7, v6)) {
									if (compare(v7, v4)) {
										if (compare(v4, v0)) {
											if (compare(v6, v0)) {
												if (compare(v2, v0)) {
													if (compare(v5, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hgefdcab
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hegfdcab
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hefgdcab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v6, v7); // hefdgcab
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hefdcgab
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v6, v7); // hgedfcab
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hegdfcab
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hedgfcab
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hedfgcab
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hedfcgab
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hgdefcab
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdgefcab
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hdegfcab
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdefgcab
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hdefcgab
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dhgefcab
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); // dhegfcab
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // dhefgcab
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); // dhefcgab
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v6, v7); // hgedcfab
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v6, v7); // hgdecfab
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hdgecfab
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); // dhgecfab
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v6, v7); // hegdcfab
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hedgcfab
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v6, v7); // hdegcfab
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); // dhegcfab
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v6, v7); // hedcgfab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hdecgfab
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); // dhecgfab
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hedcfgab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdecfgab
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dhecfgab
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v4)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v6, v7); // hgdcefab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v6, v7); // hdgcefab
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); // dhgcefab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v6, v7); // hdcgefab
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); // dhcgefab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); // dchgefab
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hdcegfab
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v6); // dhcegfab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); // dchegfab
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdcefgab
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dhcefgab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dchefgab
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v5, v7); swap(v6, v7); // hgedcafb
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hgdecafb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdgecafb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dhgecafb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v5, v7); swap(v6, v7); // hgdceafb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hdgceafb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); // dhgceafb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hdcgeafb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v5, v6); // dhcgeafb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); // dchgeafb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hegdcafb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hedgcafb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hdegcafb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); // dhegcafb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hedcgafb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdecgafb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dhecgafb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdcegafb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dhcegafb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dchegafb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v5, v7); // hgedcabf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); // hgdecabf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hdgecabf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhgecabf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v5, v7); // hgdceabf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); // hdgceabf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dhgceabf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v5, v7); // hdcgeabf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dhcgeabf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dchgeabf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); // hegdcabf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hedgcabf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); // hdegcabf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dhegcabf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); // hedcgabf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hdecgabf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhecgabf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hdcegabf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhcegabf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dchegabf
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v4, v3)) {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgefdacb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hegfdacb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hefgdacb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hefdgacb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgedfacb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); // hegdfacb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hedgfacb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hedfgacb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v6, v7); // hgedafcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v7); // hegdafcb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); // hedgafcb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgedacfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); // hegdacfb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hedgacfb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); // hgedacbf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hegdacbf
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hedgacbf
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v4)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgdefacb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hdgefacb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); // dhgefacb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hgdeafcb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); // hdgeafcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); // dhgeafcb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdegfacb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hdefgacb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hdegafcb
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dhegfacb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); // dhefgacb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); // dhegafcb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgdeacfb
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hdgeacfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdegacfb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); // dhgeacfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dhegacfb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hgdeacbf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdgeacbf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hdegacbf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dhgeacbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhegacbf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hgefdabc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hegfdabc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hefgdabc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); // hefdgabc
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); // hgedfabc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hegdfabc
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hedgfabc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hedfgabc
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v7); // hgedafbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v6, v7); // hegdafbc
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hedgafbc
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); // hgedabfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hegdabfc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hedgabfc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgedabcf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); // hegdabcf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hedgabcf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v4)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hgdefabc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdgefabc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dhgefabc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); // hgdeafbc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hdgeafbc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v6, v7); // dhgeafbc
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hdegfabc
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdefgabc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); // hdegafbc
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhegfabc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dhefgabc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dhegafbc
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hgdeabfc
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdgeabfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hdegabfc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dhgeabfc
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhegabfc
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgdeabcf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hdgeabcf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdegabcf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); // dhgeabcf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dhegabcf
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v1)) {
													if (compare(v6, v5)) {
														if (compare(v5, v1)) {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hedagfcb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); // hdeagfcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v6); // dheagfcb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hedagcfb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hdeagcfb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); // dheagcfb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); // hedagfbc
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hdeagfbc
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v6, v7); // dheagfbc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); // hedcagfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hdecagfb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hdceagfb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hedacgfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); // hdeacgfb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); // dhecagfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v5, v6); // dheacgfb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v5, v6); // dhceagfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); // dcheagfb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hedagcbf
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdeagcbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dheagcbf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hedagbfc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdeagbfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dheagbfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hedagbcf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hdeagbcf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); // dheagbcf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hedcagbf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdecagbf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdceagbf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hedacgbf
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hdeacgbf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dhecagbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v5, v6); swap(v6, v7); // dheacgbf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dhceagbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dcheagbf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v2, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); // hefdcagb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hedfcagb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); // hdefcagb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); // dhefcagb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); // hedcfagb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hdecfagb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); // dhecfagb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hdcefagb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v5); // dhcefagb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); // dchefagb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); // hefdagcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v7); // hefdacgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hefdagbc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hedfagcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); // hedfacgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hedfagbc
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdefagcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hdefacgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hdefagbc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dhefagcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); // dhefacgb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhefagbc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hedafgcb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hdeafgcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); // dheafgcb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hedafcgb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); // hdeafcgb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v5); // dheafcgb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hedafgbc
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdeafgbc
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dheafgbc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v7); // hedcafgb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); // hdecafgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v7); // hdceafgb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); // hedacfgb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); // hdeacfgb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); // dhecafgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); // dheacfgb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); // dhceafgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); // dcheafgb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v5, v0)) {
															if (compare(v2, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hefdcabg
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hedfcabg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hdefcabg
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v6, v7); // dhefcabg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hedcfabg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdecfabg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dhecfabg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdcefabg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dhcefabg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dchefabg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v6, v7); // hefdacbg
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); // hefdabgc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hefdabcg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hedfacbg
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hedfabgc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hedfabcg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdefacbg
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); // hdefabgc
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hdefabcg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dhefacbg
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dhefabgc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhefabcg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hedafcbg
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hdeafcbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v6, v7); // dheafcbg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); // hedafbgc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hdeafbgc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); // dheafbgc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hedafbcg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdeafbcg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dheafbcg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v6, v7); // hedcafbg
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hdecafbg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hdceafbg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hedacfbg
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v6, v7); // hdeacfbg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v6, v7); // dhecafbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v6, v7); // dheacfbg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v6, v7); // dhceafbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v6, v7); // dcheafbg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v6, v5)) {
																if (compare(v7, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); // hedcabgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hdecabgf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hdceabgf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hedacbgf
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); // hdeacbgf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); // dhecabgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v5, v7); // dheacbgf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v5, v7); // dhceabgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); // dcheabgf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hedcabfg
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdecabfg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdceabfg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hedacbfg
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hdeacbfg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dhecabfg
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v5, v7); swap(v6, v7); // dheacbfg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dhceabfg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dcheabfg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); // hedabgfc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hdeabgfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); // dheabgfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); // hedabfgc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); // hdeabfgc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v7); // dheabfgc
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v6, v7); // hedabfcg
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hdeabfcg
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v6, v7); // dheabfcg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hedabgcf
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdeabgcf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dheabgcf
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); // hedabcgf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hdeabcgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); // dheabcgf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hedabcfg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdeabcfg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dheabcfg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v2)) {
												if (compare(v5, v2)) {
													if (compare(v5, v1)) {
														if (compare(v2, v1)) {
															if (compare(v6, v5)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v6, v7); // hgdaefcb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); // hdgaefcb
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v6, v7); // hdagefcb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hdaegfcb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); // dhgaefcb
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); // dhagefcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); // dhaegfcb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v6, v7); // dahgefcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dahegfcb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hdaefgcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hdaefcgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hdaefcbg
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dhaefgcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dhaefcgb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dhaefcbg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dahefgcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dahefcgb
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dahefcbg
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); // hgdaefbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v6, v7); // hdgaefbc
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); // hdagefbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); // hdaegfbc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v6, v7); // dhgaefbc
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v6, v7); // dhagefbc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dhaegfbc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); // dahgefbc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); // dahegfbc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hdaefgbc
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); // hdaefbgc
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hdaefbcg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dhaefgbc
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dhaefbgc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dhaefbcg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dahefgbc
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); // dahefbgc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dahefbcg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); // hgdaebfc
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hdgaebfc
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); // hdagebfc
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hdaegbfc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v5, v7); swap(v6, v7); // dhgaebfc
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dhagebfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dhaegbfc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); // dahgebfc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dahegbfc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); // hdaebgfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); // hdaebfgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v6, v7); // hdaebfcg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dhaebgfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); // dhaebfgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dhaebfcg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); // dahebgfc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); // dahebfgc
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v6, v7); // dahebfcg
																	}
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hgdabefc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdgabefc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhgabefc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hdagbefc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhagbefc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dahgbefc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); // hdabgefc
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dhabgefc
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); // dahbgefc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dabhgefc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v5, v6); // hdabegfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); // hdabefgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v6, v7); // hdabefcg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); // dhabegfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); // dhabefgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v6, v7); // dhabefcg
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v5, v6); // dahbegfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); // dahbefgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v6, v7); // dahbefcg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); // dabhegfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); // dabhefgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v6, v7); // dabhefcg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v5, v1)) {
															if (compare(v6, v2)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hgdaecfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); // hdgaecfb
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hdagecfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hdaegcfb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v5, v6); // dhgaecfb
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); // dhagecfb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dhaegcfb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dahgecfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dahegcfb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hdaecgfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); // hdaecfgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hdaecfbg
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); // dhaecgfb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); // dhaecfgb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v6, v7); // dhaecfbg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dahecgfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); // dahecfgb
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dahecfbg
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); // hgdaecbf
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hdgaecbf
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); // hdagecbf
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hdaegcbf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v5, v6); swap(v6, v7); // dhgaecbf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dhagecbf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhaegcbf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); // dahgecbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dahegcbf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdaecgbf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hdaecbgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdaecbfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dhaecgbf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); // dhaecbgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dhaecbfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dahecgbf
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dahecbgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dahecbfg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hgdaebcf
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); // hdgaebcf
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hdagebcf
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdaegbcf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v5, v7); // dhgaebcf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); // dhagebcf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dhaegbcf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dahgebcf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dahegbcf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hdaebgcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); // hdaebcgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hdaebcfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhaebgcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dhaebcgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhaebcfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dahebgcf
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); // dahebcgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dahebcfg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgdabecf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hdgabecf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); // dhgabecf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdagbecf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dhagbecf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dahgbecf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdabgecf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhabgecf
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dahbgecf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dabhgecf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v5, v6); swap(v6, v7); // hdabegcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v5, v7); // hdabecgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v5, v7); swap(v6, v7); // hdabecfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dhabegcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); // dhabecgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dhabecfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v5, v6); swap(v6, v7); // dahbegcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v5, v7); // dahbecgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v5, v7); swap(v6, v7); // dahbecfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dabhegcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); // dabhecgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dabhecfg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v4)) {
													if (compare(v4, v1)) {
														if (compare(v2, v0)) {
															if (compare(v5, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgdcaefb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdgcaefb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dhgcaefb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdcgaefb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dhcgaefb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dchgaefb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hdcagefb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v4, v6); swap(v5, v6); // dhcagefb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); // dchagefb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dcahgefb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); // hgdcaebf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hdgcaebf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhgcaebf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hdcgaebf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhcgaebf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dchgaebf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdcagebf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhcagebf
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dchagebf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcahgebf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgdacefb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hdgacefb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); // dhgacefb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdagcefb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dhagcefb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dahgcefb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hdacgefb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); // dhacgefb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dahcgefb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dachgefb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hgdacebf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdgacebf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhgacebf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hdagcebf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhagcebf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dahgcebf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hdacgebf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dhacgebf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dahcgebf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dachgebf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v6, v0)) {
																if (compare(v2, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgdcabef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdgcabef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dhgcabef
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdcgabef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dhcgabef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dchgabef
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgdacbef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hdgacbef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); // dhgacbef
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hgdabcef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdgabcef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhgabcef
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdagcbef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dhagcbef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dahgcbef
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hdagbcef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhagbcef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dahgbcef
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hdcagbef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v4, v6); swap(v5, v7); // dhcagbef
																			}
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); // dchagbef
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dcahgbef
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hdacgbef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); // dhacgbef
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dahcgbef
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dachgbef
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdcabgef
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhcabgef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dchabgef
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdacbgef
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhacbgef
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); // hdabgcef
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dhabgcef
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hdabcgef
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dhabcgef
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcahbgef
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcabhgef
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dahcbgef
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dachbgef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dacbhgef
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); // dahbgcef
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dahbcgef
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dabhgcef
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dabhcgef
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dabchgef
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v2, v0)) {
															if (compare(v6, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v5, v6); // hdcaegfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v5, v6); // dhcaegfb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v5, v6); // dchaegfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); // dcahegfb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hdcaegbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v5, v6); swap(v6, v7); // dhcaegbf
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v5, v6); swap(v6, v7); // dchaegbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dcahegbf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v5, v7); // hdcaebgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v5, v7); // dhcaebgf
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v5, v7); // dchaebgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); // dcahebgf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hdcabegf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v4, v7); swap(v5, v7); // dhcabegf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); // dchabegf
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dcahbegf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); // dcabhegf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); // hdacegfb
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hdacegbf
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); // hdacebgf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hdacbegf
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); // hdabcegf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v5, v6); // dhacegfb
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v5, v6); swap(v6, v7); // dhacegbf
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v5, v7); // dhacebgf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); // dhacbegf
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dhabcegf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); // dahcegfb
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dahcegbf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); // dahcebgf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v5, v6); // dachegfb
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dachegbf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v5, v7); // dachebgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dahcbegf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dachbegf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v4, v7); swap(v5, v7); // dacbhegf
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); // dahbcegf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dabhcegf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); // dabchegf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v5, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v7); // hdcaefgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); // dhcaefgb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); // dchaefgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); // dcahefgb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v6, v7); // hdcaefbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v6, v7); // dhcaefbg
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v6, v7); // dchaefbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v6, v7); // dcahefbg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hdcaebfg
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v5, v7); swap(v6, v7); // dhcaebfg
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v5, v7); swap(v6, v7); // dchaebfg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dcahebfg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v7); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdcabefg
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhcabefg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dchabefg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcahbefg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcabhefg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); // hdacefgb
																			}
																			else {
																				swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v6, v7); // hdacefbg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hdacebfg
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdacbefg
																		}
																		else {
																			swap(v0, v7); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hdabcefg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); // dhacefgb
																			}
																			else {
																				swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v6, v7); // dhacefbg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v5, v7); swap(v6, v7); // dhacebfg
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhacbefg
																		}
																		else {
																			swap(v0, v3); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dhabcefg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); // dahcefgb
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v6, v7); // dahcefbg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dahcebfg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v3, v7); // dachefgb
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v6, v7); // dachefbg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dachebfg
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dahcbefg
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dachbefg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dacbhefg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dahbcefg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dabhcefg
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dabchefg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v7, v0)) {
											if (compare(v5, v0)) {
												if (compare(v2, v0)) {
													if (compare(v6, v0)) {
														if (compare(v5, v2)) {
															if (compare(v7, v5)) {
																if (compare(v7, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ehgfdcab
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); // ehfgdcab
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v6); // ehfdgcab
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); // ehfdcgab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); // ehgdfcab
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ehdgfcab
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); // ehdfgcab
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ehdfcgab
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edhgfcab
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // edhfgcab
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edhfcgab
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dehgfcab
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dehfgcab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dehfcgab
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // efhgdcab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); // efhdgcab
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // efdhgcab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efhdcgab
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efdhcgab
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efdchgab
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); // edfhgcab
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // edfhcgab
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edfchgab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // defhgcab
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // defhcgab
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // defchgab
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v6); // ehgdcfab
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); // ehdgcfab
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // edhgcfab
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); // dehgcfab
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v6); // ehdcgfab
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); // edhcgfab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dehcgfab
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); // ehdcfgab
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // edhcfgab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dehcfgab
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v6); // edchgfab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // edchfgab
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edcfhgab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dechgfab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dechfgab
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // decfhgab
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // dcehgfab
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcehfgab
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // dcefhgab
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v7, v5)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v5); // ehfdcagb
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); // ehdfcagb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // edhfcagb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); // dehfcagb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v5); // ehdcfagb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); // edhcfagb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dehcfagb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v5); // edchfagb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dechfagb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dcehfagb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); // efhdcagb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // efdhcagb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); // edfhcagb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // defhcagb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); // efdchagb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // edfchagb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); // defchagb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // edcfhagb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v7); // decfhagb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); // dcefhagb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v6, v7); // ehfdcabg
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ehdfcabg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edhfcabg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dehfcabg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v6, v7); // ehdcfabg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // edhcfabg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dehcfabg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // edchfabg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dechfabg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcehfabg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efhdcabg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efdhcabg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // edfhcabg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // defhcabg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efdchabg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edfchabg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // defchabg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edcfhabg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // decfhabg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dcefhabg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v7, v3)) {
															if (compare(v6, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); // ehgfdacb
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v5); // ehgdfacb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ehfgdacb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efhgdacb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); // ehfdgacb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efhdgacb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); // ehdgfacb
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ehdfgacb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v5, v6); // ehfdagcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v5, v6); swap(v6, v7); // efhdagcb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); // ehdfagcb
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); // ehfdacgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v5, v7); // efhdacgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); // ehdfacgb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v6, v7); // ehfdacbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v5, v7); swap(v6, v7); // efhdacbg
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v6, v7); // ehdfacbg
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // edhgfacb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edhfgacb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // edhfagcb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dehgfacb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dehfgacb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dehfagcb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efdhgacb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // edfhgacb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // defhgacb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // efdhagcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); // edfhagcb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // defhagcb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); // efdhacgb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); // edhfacgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); // edfhacgb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dehfacgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); // defhacgb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // efdhacbg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // edhfacbg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v6, v7); // edfhacbg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dehfacbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // defhacbg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v6, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ehgfdabc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v6, v7); // ehgdfabc
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehfgdabc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // efhgdabc
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehfdgabc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); // efhdgabc
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ehdgfabc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehdfgabc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v5, v6); swap(v6, v7); // ehfdagbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v5, v6); // efhdagbc
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ehdfagbc
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v5, v7); // ehfdabgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); // efhdabgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); // ehdfabgc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v5, v7); swap(v6, v7); // ehfdabcg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v6, v7); // efhdabcg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ehdfabcg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); // edhgfabc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // edhfgabc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); // edhfagbc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dehgfabc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); // dehfgabc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dehfagbc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // efdhgabc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edfhgabc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // defhgabc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); // efdhagbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // edfhagbc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // defhagbc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); // efdhabgc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); // edhfabgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); // edfhabgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); // dehfabgc
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // defhabgc
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v6, v7); // efdhabcg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v6, v7); // edhfabcg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // edfhabcg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dehfabcg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // defhabcg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v6, v5)) {
														if (compare(v6, v0)) {
															if (compare(v2, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); // ehgdcafb
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ehdgcafb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edhgcafb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dehgcafb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); // ehdcgafb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // edhcgafb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dehcgafb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // edchgafb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dechgafb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcehgafb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); // ehgdafcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v5, v6); // ehgdacfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v6, v7); // ehgdafbc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); // ehdgafcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); // ehdgacfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v6, v7); // ehdgafbc
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v6, v7); // edhgafcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // edhgacfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); // edhgafbc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dehgafcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dehgacfb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); // dehgafbc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); // ehdagfcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // edhagfcb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v6, v7); // dehagfcb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ehdagcfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edhagcfb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dehagcfb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ehdagfbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); // edhagfbc
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v6); // dehagfbc
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v5, v6); // ehdcagfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); // edhcagfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v5, v6); // edchagfb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); // ehdacgfb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // edhacgfb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dehcagfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); // dehacgfb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dechagfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dcehagfb
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ehdafgcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edhafgcb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dehafgcb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); // ehdafcgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // edhafcgb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); // dehafcgb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehdafgbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // edhafgbc
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); // dehafgbc
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); // ehdcafgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); // edhcafgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v7); // edchafgb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); // ehdacfgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); // edhacfgb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); // dehcafgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v7); // dehacfgb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v7); // dechafgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); // dcehafgb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ehdafcbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edhafcbg
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dehafcbg
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ehdafbgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); // edhafbgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v5); // dehafbgc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehdafbcg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // edhafbcg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v6, v7); // dehafbcg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v6, v7); // ehdcafbg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v6, v7); // edhcafbg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v6, v7); // edchafbg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v6, v7); // ehdacfbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // edhacfbg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dehcafbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v6, v7); // dehacfbg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dechafbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dcehafbg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v6, v0)) {
															if (compare(v2, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehgdcabf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehdgcabf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // edhgcabf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dehgcabf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehdcgabf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edhcgabf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dehcgabf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edchgabf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dechgabf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dcehgabf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v5, v6); swap(v6, v7); // ehgdacbf
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v5, v7); swap(v6, v7); // ehgdabfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v5, v7); // ehgdabcf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ehdgacbf
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ehdgabfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); // ehdgabcf
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); // edhgacbf
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); // edhgabfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // edhgabcf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dehgacbf
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dehgabfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dehgabcf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehdagcbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // edhagcbf
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); // dehagcbf
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehdagbfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // edhagbfc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); // dehagbfc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ehdagbcf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edhagbcf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dehagbcf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v5, v6); swap(v6, v7); // ehdcagbf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // edhcagbf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // edchagbf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ehdacgbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edhacgbf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dehcagbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dehacgbf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dechagbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcehagbf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v6, v5)) {
																if (compare(v4, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v5, v7); // ehdcabgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); // edhcabgf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v5, v7); // edchabgf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); // ehdacbgf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // edhacbgf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dehcabgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); // dehacbgf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dechabgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dcehabgf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v5, v7); swap(v6, v7); // ehdcabfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // edhcabfg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // edchabfg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ehdacbfg
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edhacbfg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dehcabfg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dehacbfg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dechabfg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcehabfg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ehdabgfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); // edhabgfc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v5, v6); // dehabgfc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); // ehdabfgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); // edhabfgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); // dehabfgc
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ehdabfcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v6, v7); // edhabfcg
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v6, v7); // dehabfcg
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehdabgcf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // edhabgcf
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v5, v6); swap(v6, v7); // dehabgcf
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ehdabcgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); // edhabcgf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v5, v7); // dehabcgf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehdabcfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // edhabcfg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v7); swap(v5, v7); swap(v6, v7); // dehabcfg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v2)) {
												if (compare(v6, v2)) {
													if (compare(v6, v1)) {
														if (compare(v2, v1)) {
															if (compare(v6, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); // edahgfcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // edahgcfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edahgcbf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // deahgfcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // deahgcfb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // deahgcbf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); // daehgfcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // daehgcfb
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // daehgcbf
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efdahgcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // edfahgcb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // edahfgcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edafhgcb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); // defahgcb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // deahfgcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // deafhgcb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // daehfgcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // daefhgcb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // edahgfbc
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edahgbfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // edahgbcf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); // deahgfbc
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // deahgbfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // deahgbcf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // daehgfbc
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // daehgbfc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // daehgbcf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // efdahgbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edfahgbc
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edahfgbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // edafhgbc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // defahgbc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // deahfgbc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // deafhgbc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // daehfgbc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // daefhgbc
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); // efdahbgc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // edfahbgc
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // edahfbgc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); // edafhbgc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); // defahbgc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); // deahfbgc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // deafhbgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // daehfbgc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); // daefhbgc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // edahbgfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); // edahbfgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edahbgcf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); // deahbgfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); // deahbfgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // deahbgcf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // daehbgfc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); // daehbfgc
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // daehbgcf
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); // edabhgfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v7); // edabhfgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edabhgcf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // deabhgfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); // deabhfgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // deabhgcf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); // daebhgfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v7); // daebhfgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // daebhgcf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dabehgfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); // dabehfgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dabehgcf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // efdabhgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); // edfabhgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v5, v7); // defabhgc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // edafbhgc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); // deafbhgc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // daefbhgc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); // edabfhgc
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // deabfhgc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); // daebfhgc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dabefhgc
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v6, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // efdahcgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); // edfahcgb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); // edahfcgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // edafhcgb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v7); // defahcgb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // deahfcgb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); // deafhcgb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); // daehfcgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // daefhcgb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); // edahcgfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); // edahcfgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // edahcgbf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // deahcgfb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); // deahcfgb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // deahcgbf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); // daehcgfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); // daehcfgb
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // daehcgbf
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efdahcbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // edfahcbg
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // edahfcbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edafhcbg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v6, v7); // defahcbg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // deahfcbg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // deafhcbg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // daehfcbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // daefhcbg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v6, v7); // edahcfbg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); // edahcbgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // edahcbfg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // deahcfbg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // deahcbgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // deahcbfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v6, v7); // daehcfbg
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); // daehcbgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // daehcbfg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // efdahbcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edfahbcg
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edahfbcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // edafhbcg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // defahbcg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // deahfbcg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // deafhbcg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // daehfbcg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // daefhbcg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // edahbfcg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // edahbcgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edahbcfg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v6, v7); // deahbfcg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); // deahbcgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // deahbcfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // daehbfcg
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // daehbcgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // daehbcfg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v6, v7); // edabhfcg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); // edabhcgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edabhcfg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // deabhfcg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // deabhcgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // deabhcfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v6, v7); // daebhfcg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); // daebhcgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // daebhcfg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dabehfcg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dabehcgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dabehcfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efdabhcg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // edfabhcg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v5, v7); swap(v6, v7); // defabhcg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edafbhcg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // deafbhcg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // daefbhcg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edabfhcg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // deabfhcg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // daebfhcg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dabefhcg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v5)) {
													if (compare(v6, v5)) {
														if (compare(v2, v0)) {
															if (compare(v6, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // edcahgfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v7); swap(v5, v6); // decahgfb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); // dceahgfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dcaehgfb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edcahgbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // decahgbf
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dceahgbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcaehgbf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // edcahbgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v7); swap(v5, v7); // decahbgf
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); // dceahbgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dcaehbgf
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v5, v7); // edcabhgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v5, v7); // decabhgf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v5, v7); // dceabhgf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); // dcaebhgf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v5, v7); // dcabehgf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // edachgfb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edachgbf
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // edachbgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); // edacbhgf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v5, v7); // edabchgf
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); // deachgfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // deachgbf
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); // deachbgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v5, v7); // deacbhgf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); // deabchgf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // daechgfb
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // daechgbf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // daechbgf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dacehgfb
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dacehgbf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dacehbgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); // daecbhgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v5, v7); // dacebhgf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v5, v7); // dacbehgf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v5, v7); // daebchgf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); // dabechgf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v5, v7); // dabcehgf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v5, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v7); // edcahfgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v7); // decahfgb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v7); // dceahfgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); // dcaehfgb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // edcahfbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v7); swap(v6, v7); // decahfbg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v6, v7); // dceahfbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dcaehfbg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edcahbfg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // decahbfg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dceahbfg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcaehbfg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // edcabhfg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v5, v7); swap(v6, v7); // decabhfg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v5, v7); swap(v6, v7); // dceabhfg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dcaebhfg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v5, v7); swap(v6, v7); // dcabehfg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); // edachfgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // edachfbg
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edachbfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // edacbhfg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v5, v7); swap(v6, v7); // edabchfg
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v7); // deachfgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v6, v7); // deachfbg
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // deachbfg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v5, v7); swap(v6, v7); // deacbhfg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // deabchfg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); // daechfgb
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // daechfbg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // daechbfg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v7); // dacehfgb
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dacehfbg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dacehbfg
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // daecbhfg
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dacebhfg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v5, v7); swap(v6, v7); // dacbehfg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v5, v7); swap(v6, v7); // daebchfg
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dabechfg
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v5, v7); swap(v6, v7); // dabcehfg
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v2, v0)) {
															if (compare(v6, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v5, v7); // efdcahgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); // edfcahgb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // defcahgb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v5, v7); // edcfahgb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // decfahgb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dcefahgb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // edcafhgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v5); swap(v5, v7); // decafhgb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); // dceafhgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dcaefhgb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v5, v7); swap(v6, v7); // efdcahbg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // edfcahbg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // defcahbg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // edcfahbg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // decfahbg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcefahbg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edcafhbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // decafhbg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dceafhbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcaefhbg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); // efdachgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // edfachgb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); // defachgb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); // edafchgb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // deafchgb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); // daefchgb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // edacfhgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); // deacfhgb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // daecfhgb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dacefhgb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // efdachbg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edfachbg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // defachbg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // edafchbg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // deafchbg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // daefchbg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edacfhbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // deacfhbg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // daecfhbg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dacefhbg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																if (compare(v2, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v6, v7); // efdcabhg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v6, v7); // edfcabhg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // defcabhg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v6, v7); // edcfabhg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // decfabhg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dcefabhg
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v6, v7); // efdacbhg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // edfacbhg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v6, v7); // defacbhg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // efdabchg
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v6, v7); // edfabchg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v6, v7); // defabchg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v6, v7); // edafcbhg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // deafcbhg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v6, v7); // daefcbhg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // edafbchg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v6, v7); // deafbchg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // daefbchg
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // edcafbhg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v4, v5); swap(v6, v7); // decafbhg
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v6, v7); // dceafbhg
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dcaefbhg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // edacfbhg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v6, v7); // deacfbhg
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // daecfbhg
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dacefbhg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v6, v7); // edcabfhg
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v6, v7); // decabfhg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v6, v7); // dceabfhg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v6, v7); // edacbfhg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v6, v7); // deacbfhg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v6, v7); // edabfchg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // deabfchg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v6, v7); // edabcfhg
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v6, v7); // deabcfhg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v6, v7); // dcaebfhg
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v6, v7); // dcabefhg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v6, v7); // daecbfhg
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v6, v7); // dacebfhg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v6, v7); // dacbefhg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v6, v7); // daebfchg
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v6, v7); // daebcfhg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dabefchg
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v6, v7); // dabecfhg
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v6, v7); // dabcefhg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v4)) {
										if (compare(v4, v0)) {
											if (compare(v7, v0)) {
												if (compare(v2, v0)) {
													if (compare(v5, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); // ghefdcab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gehfdcab
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); // gefhdcab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v7); // gefdhcab
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); // gefdchab
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v5); // ghedfcab
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); // gehdfcab
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gedhfcab
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); // gedfhcab
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gedfchab
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); // ghdefcab
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gdhefcab
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); // gdehfcab
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gdefhcab
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); // gdefchab
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dghefcab
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // dgehfcab
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgefhcab
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // dgefchab
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); // ghedcfab
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); // ghdecfab
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); // gdhecfab
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dghecfab
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v7); // gehdcfab
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); // gedhcfab
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); // gdehcfab
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v6, v7); // dgehcfab
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v7); // gedchfab
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); // gdechfab
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dgechfab
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); // gedcfhab
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gdecfhab
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgecfhab
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v4)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); // ghdcefab
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); // gdhcefab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v6, v7); // dghcefab
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v7); // gdchefab
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v6, v7); // dgchefab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v6, v7); // dcghefab
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v7); // gdcehfab
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dgcehfab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dcgehfab
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gdcefhab
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgcefhab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dcgefhab
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v7, v4)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v5, v6); // ghedcafb
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); // ghdecafb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gdhecafb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dghecafb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v5, v6); // ghdceafb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); // gdhceafb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dghceafb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v5, v6); // gdcheafb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dgcheafb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dcgheafb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); // gehdcafb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gedhcafb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); // gdehcafb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dgehcafb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); // gedchafb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gdechafb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgechafb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gdcehafb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgcehafb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcgehafb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v4)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v5, v6); swap(v6, v7); // ghedcabf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ghdecabf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdhecabf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dghecabf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v5, v6); swap(v6, v7); // ghdceabf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gdhceabf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); // dghceabf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gdcheabf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v5, v7); // dgcheabf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); // dcgheabf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gehdcabf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gedhcabf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gdehcabf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); // dgehcabf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gedchabf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdechabf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dgechabf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdcehabf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dgcehabf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dcgehabf
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v4, v3)) {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ghefdacb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gehfdacb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gefhdacb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); // gefdhacb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); // ghedfacb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gehdfacb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gedhfacb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gedfhacb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v6); // ghedafcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v6, v7); // gehdafcb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gedhafcb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); // ghedacfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gehdacfb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gedhacfb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghedacbf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); // gehdacbf
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gedhacbf
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v7, v4)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ghdefacb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdhefacb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dghefacb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); // ghdeafcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gdheafcb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v6, v7); // dgheafcb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gdehfacb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdefhacb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); // gdehafcb
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgehfacb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // dgefhacb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // dgehafcb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ghdeacfb
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdheacfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gdehacfb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dgheacfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgehacfb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghdeacbf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gdheacbf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdehacbf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); // dgheacbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dgehacbf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v5, v0)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghefdabc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gehfdabc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gefhdabc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gefdhabc
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghedfabc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); // gehdfabc
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gedhfabc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gedfhabc
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v6, v7); // ghedafbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v6); // gehdafbc
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); // gedhafbc
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghedabfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); // gehdabfc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gedhabfc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); // ghedabcf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gehdabcf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gedhabcf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v7, v4)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghdefabc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gdhefabc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); // dghefabc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ghdeafbc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); // gdheafbc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); // dgheafbc
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdehfabc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gdefhabc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gdehafbc
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // dgehfabc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); // dgefhabc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); // dgehafbc
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghdeabfc
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gdheabfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdehabfc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); // dgheabfc
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // dgehabfc
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ghdeabcf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdheabcf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gdehabcf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dgheabcf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgehabcf
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v1)) {
													if (compare(v7, v5)) {
														if (compare(v5, v1)) {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); // gedahfcb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gdeahfcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v6, v7); // dgeahfcb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gedahcfb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdeahcfb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgeahcfb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gedahfbc
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); // gdeahfbc
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v7); // dgeahfbc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gedcahfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdecahfb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdceahfb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gedachfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gdeachfb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dgecahfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v5, v7); swap(v6, v7); // dgeachfb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dgceahfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dcgeahfb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gedahcbf
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gdeahcbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); // dgeahcbf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gedahbfc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gdeahbfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); // dgeahbfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gedahbcf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdeahbcf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgeahbcf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); // gedcahbf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gdecahbf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gdceahbf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gedachbf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); // gdeachbf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); // dgecahbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v5, v7); // dgeachbf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v5, v7); // dgceahbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); // dcgeahbf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v2, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gefdcahb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gedfcahb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gdefcahb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v6, v7); // dgefcahb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gedcfahb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdecfahb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dgecfahb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdcefahb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dgcefahb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dcgefahb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gefdahcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v6, v7); // gefdachb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); // gefdahbc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gedfahcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gedfachb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gedfahbc
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gdefahcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdefachb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdefahbc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgefahcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // dgefachb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // dgefahbc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gedafhcb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdeafhcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgeafhcb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gedafchb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gdeafchb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v6, v7); // dgeafchb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gedafhbc
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gdeafhbc
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); // dgeafhbc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v6, v7); // gedcafhb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gdecafhb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gdceafhb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gedacfhb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v6, v7); // gdeacfhb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v6, v7); // dgecafhb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v6, v7); // dgeacfhb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v6, v7); // dgceafhb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v6, v7); // dcgeafhb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v5, v0)) {
															if (compare(v2, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // gefdcabh
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gedfcabh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // gdefcabh
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); // dgefcabh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // gedcfabh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gdecfabh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // dgecfabh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gdcefabh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v4, v5); // dgcefabh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // dcgefabh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v6); // gefdacbh
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gefdabhc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // gefdabch
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // gedfacbh
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gedfabhc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gedfabch
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gdefacbh
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gdefabhc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // gdefabch
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // dgefacbh
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgefabhc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dgefabch
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gedafcbh
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // gdeafcbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v5); // dgeafcbh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gedafbhc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdeafbhc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgeafbhc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // gedafbch
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gdeafbch
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // dgeafbch
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v6); // gedcafbh
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // gdecafbh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v6); // gdceafbh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // gedacfbh
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); // gdeacfbh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); // dgecafbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); // dgeacfbh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); // dgceafbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); // dcgeafbh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v7, v5)) {
																if (compare(v6, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gedcabhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdecabhf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdceabhf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gedacbhf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gdeacbhf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dgecabhf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v5, v6); swap(v6, v7); // dgeacbhf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dgceabhf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dcgeabhf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // gedcabfh
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gdecabfh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gdceabfh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gedacbfh
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // gdeacbfh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // dgecabfh
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v5, v6); // dgeacbfh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v3, v4); swap(v5, v6); // dgceabfh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // dcgeabfh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gedabhfc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdeabhfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgeabhfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v6, v7); // gedabfhc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gdeabfhc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v6, v7); // dgeabfhc
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); // gedabfch
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // gdeabfch
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v6); // dgeabfch
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); // gedabhcf
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gdeabhcf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); // dgeabhcf
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gedabchf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdeabchf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgeabchf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // gedabcfh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gdeabcfh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // dgeabcfh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v2)) {
												if (compare(v5, v2)) {
													if (compare(v5, v1)) {
														if (compare(v2, v1)) {
															if (compare(v7, v5)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); // ghdaefcb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v6, v7); // gdhaefcb
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); // gdahefcb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); // gdaehfcb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v6, v7); // dghaefcb
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v6, v7); // dgahefcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dgaehfcb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); // daghefcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); // dagehfcb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gdaefhcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gdaefchb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gdaefcbh
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dgaefhcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dgaefchb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dgaefcbh
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dagefhcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dagefchb
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dagefcbh
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v6, v7); // ghdaefbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); // gdhaefbc
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v6, v7); // gdahefbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gdaehfbc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); // dghaefbc
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); // dgahefbc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); // dgaehfbc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v6, v7); // daghefbc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // dagehfbc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gdaefhbc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gdaefbhc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // gdaefbch
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // dgaefhbc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dgaefbhc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dgaefbch
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // dagefhbc
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // dagefbhc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // dagefbch
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v7, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ghdaebfc
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); // gdhaebfc
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gdahebfc
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdaehbfc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v5, v6); // dghaebfc
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); // dgahebfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dgaehbfc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // daghebfc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dagehbfc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gdaebhfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v6, v7); // gdaebfhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); // gdaebfch
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dgaebhfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dgaebfhc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // dgaebfch
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dagebhfc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v6, v7); // dagebfhc
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); // dagebfch
																	}
																}
															}
														}
														else {
															if (compare(v7, v4)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghdabefc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gdhabefc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); // dghabefc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdahbefc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // dgahbefc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // daghbefc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdabhefc
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgabhefc
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dagbhefc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dabghefc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v5, v7); swap(v6, v7); // gdabehfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v6, v7); // gdabefhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); // gdabefch
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dgabehfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v6, v7); // dgabefhc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); // dgabefch
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v5, v7); swap(v6, v7); // dagbehfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v6, v7); // dagbefhc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); // dagbefch
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dabgehfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v6, v7); // dabgefhc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); // dabgefch
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v5, v1)) {
															if (compare(v7, v2)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); // ghdaecfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gdhaecfb
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); // gdahecfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gdaehcfb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v5, v7); swap(v6, v7); // dghaecfb
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // dgahecfb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgaehcfb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); // daghecfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // dagehcfb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gdaechfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gdaecfhb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // gdaecfbh
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // dgaechfb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v6, v7); // dgaecfhb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); // dgaecfbh
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dagechfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dagecfhb
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // dagecfbh
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ghdaecbf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); // gdhaecbf
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gdahecbf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdaehcbf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v5, v7); // dghaecbf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); // dgahecbf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dgaehcbf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // daghecbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dagehcbf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gdaechbf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdaecbhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gdaecbfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); // dgaechbf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dgaecbhf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // dgaecbfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dagechbf
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dagecbhf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dagecbfh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v7, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); // ghdaebcf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gdhaebcf
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); // gdahebcf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gdaehbcf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v5, v6); swap(v6, v7); // dghaebcf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // dgahebcf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgaehbcf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); // daghebcf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // dagehbcf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); // gdaebhcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gdaebchf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // gdaebcfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dgaebhcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgaebchf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dgaebcfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); // dagebhcf
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dagebchf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // dagebcfh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v4)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ghdabecf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdhabecf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dghabecf
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gdahbecf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgahbecf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // daghbecf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); // gdabhecf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dgabhecf
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); // dagbhecf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dabghecf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v5, v7); // gdabehcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v5, v6); swap(v6, v7); // gdabechf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v5, v6); // gdabecfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); // dgabehcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dgabechf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // dgabecfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v5, v7); // dagbehcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v5, v6); swap(v6, v7); // dagbechf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v5, v6); // dagbecfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); // dabgehcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dabgechf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // dabgecfh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v4)) {
													if (compare(v4, v1)) {
														if (compare(v2, v0)) {
															if (compare(v5, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); // ghdcaefb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gdhcaefb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dghcaefb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gdchaefb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgchaefb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcghaefb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdcahefb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgcahefb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcgahefb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dcaghefb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghdcaebf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdhcaebf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dghcaebf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdchaebf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dgchaebf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dcghaebf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gdcahebf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v4, v7); swap(v5, v7); // dgcahebf
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); // dcgahebf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dcaghebf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ghdacefb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdhacefb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dghacefb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gdahcefb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgahcefb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // daghcefb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gdachefb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dgachefb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dagchefb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // dacghefb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghdacebf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gdhacebf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); // dghacebf
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdahcebf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // dgahcebf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // daghcebf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gdachebf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); // dgachebf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dagchebf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // dacghebf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v7, v0)) {
																if (compare(v2, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); // ghdcabef
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gdhcabef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dghcabef
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gdchabef
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgchabef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcghabef
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ghdacbef
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdhacbef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dghacbef
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghdabcef
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gdhabcef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); // dghabcef
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gdahcbef
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgahcbef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // daghcbef
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdahbcef
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // dgahbcef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // daghbcef
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdcahbef
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgcahbef
																			}
																		}
																		else {
																			if (compare(v6, v0)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcgahbef
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dcaghbef
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdachbef
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgachbef
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dagchbef
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dacghbef
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gdcabhef
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v4, v6); swap(v5, v7); // dgcabhef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); // dcgabhef
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gdacbhef
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); // dgacbhef
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gdabhcef
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dgabhcef
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); // gdabchef
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dgabchef
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dcagbhef
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); // dcabghef
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dagcbhef
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dacgbhef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v4, v6); swap(v5, v7); // dacbghef
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dagbhcef
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); // dagbchef
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dabghcef
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dabgchef
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); // dabcghef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v2, v0)) {
															if (compare(v7, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gdcaehfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v5, v7); swap(v6, v7); // dgcaehfb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v5, v7); swap(v6, v7); // dcgaehfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dcagehfb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v5, v7); // gdcaehbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v5, v7); // dgcaehbf
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v5, v7); // dcgaehbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); // dcagehbf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gdcaebhf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v5, v6); swap(v6, v7); // dgcaebhf
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v5, v6); swap(v6, v7); // dcgaebhf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dcagebhf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdcabehf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgcabehf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcgabehf
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcagbehf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcabgehf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gdacehfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); // gdacehbf
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gdacebhf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdacbehf
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gdabcehf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v5, v7); swap(v6, v7); // dgacehfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v5, v7); // dgacehbf
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v5, v6); swap(v6, v7); // dgacebhf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgacbehf
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dgabcehf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dagcehfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); // dagcehbf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dagcebhf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // dacgehfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v5, v7); // dacgehbf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // dacgebhf
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dagcbehf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dacgbehf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dacbgehf
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dagbcehf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dabgcehf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dabcgehf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v5, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v6, v7); // gdcaefhb
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v6, v7); // dgcaefhb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v6, v7); // dcgaefhb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v6, v7); // dcagefhb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v6); // gdcaefbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); // dgcaefbh
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); // dcgaefbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); // dcagefbh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v5, v6); // gdcaebfh
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v5, v6); // dgcaebfh
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v5, v6); // dcgaebfh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // dcagebfh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v6); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gdcabefh
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v4, v6); swap(v5, v6); // dgcabefh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // dcgabefh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dcagbefh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // dcabgefh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v6, v7); // gdacefhb
																			}
																			else {
																				swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); // gdacefbh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // gdacebfh
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gdacbefh
																		}
																		else {
																			swap(v0, v6); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // gdabcefh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v6, v7); // dgacefhb
																			}
																			else {
																				swap(v0, v3); swap(v1, v6); swap(v2, v3); // dgacefbh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v5, v6); // dgacebfh
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // dgacbefh
																		}
																		else {
																			swap(v0, v3); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dgabcefh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v6, v7); // dagcefhb
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); // dagcefbh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // dagcebfh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v6, v7); // dacgefhb
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v6); // dacgefbh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v5, v6); // dacgebfh
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dagcbefh
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dacgbefh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v4, v6); swap(v5, v6); // dacbgefh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // dagbcefh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dabgcefh
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // dabcgefh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v0)) {
											if (compare(v5, v0)) {
												if (compare(v2, v0)) {
													if (compare(v7, v0)) {
														if (compare(v5, v2)) {
															if (compare(v6, v5)) {
																if (compare(v6, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eghfdcab
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // egfhdcab
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v6, v7); // egfdhcab
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egfdchab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eghdfcab
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egdhfcab
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // egdfhcab
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egdfchab
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // edghfcab
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edgfhcab
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // edgfchab
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // deghfcab
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // degfhcab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // degfchab
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efghdcab
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efgdhcab
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efdghcab
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); // efgdchab
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // efdgchab
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); // efdcghab
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // edfghcab
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edfgchab
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // edfcghab
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); // defghcab
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // defgchab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // defcghab
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v6, v7); // eghdcfab
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // egdhcfab
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); // edghcfab
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // deghcfab
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v6, v7); // egdchfab
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // edgchfab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); // degchfab
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egdcfhab
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edgcfhab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // degcfhab
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // edcghfab
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edcgfhab
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // edcfghab
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v7); // decghfab
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // decgfhab
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // decfghab
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); // dceghfab
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // dcegfhab
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcefghab
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v6, v5)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v6, v7); // egfdcahb
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // egdfcahb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edgfcahb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // degfcahb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v6, v7); // egdcfahb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // edgcfahb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // degcfahb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // edcgfahb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // decgfahb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcegfahb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efgdcahb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efdgcahb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // edfgcahb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // defgcahb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efdcgahb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edfcgahb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // defcgahb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edcfgahb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // decfgahb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // dcefgahb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v5); // egfdcabh
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // egdfcabh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // edgfcabh
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // degfcabh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v5); // egdcfabh
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // edgcfabh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // degcfabh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v5); // edcgfabh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // decgfabh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // dcegfabh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // efgdcabh
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // efdgcabh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // edfgcabh
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // defgcabh
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // efdcgabh
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // edfcgabh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // defcgabh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // edcfgabh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v6); // decfgabh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // dcefgabh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v6, v3)) {
															if (compare(v7, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // eghfdacb
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v6, v7); // eghdfacb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egfhdacb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // efghdacb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egfdhacb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); // efgdhacb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // egdhfacb
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egdfhacb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v5, v7); swap(v6, v7); // egfdahcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v5, v7); // efgdahcb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // egdfahcb
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v6, v7); // egfdachb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v5, v6); swap(v6, v7); // efgdachb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v6, v7); // egdfachb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); // egfdacbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v5, v6); // efgdacbh
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); // egdfacbh
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); // edghfacb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // edgfhacb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); // edgfahcb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // deghfacb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); // degfhacb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // degfahcb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // efdghacb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edfghacb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // defghacb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); // efdgahcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // edfgahcb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // defgahcb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // efdgachb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // edgfachb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v6, v7); // edfgachb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // degfachb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // defgachb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // efdgacbh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // edgfacbh
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); // edfgacbh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // degfacbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // defgacbh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v7, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); // eghfdabc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v5); // eghdfabc
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // egfhdabc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efghdabc
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); // egfdhabc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efgdhabc
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); // egdhfabc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // egdfhabc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v5, v7); // egfdahbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v5, v7); swap(v6, v7); // efgdahbc
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); // egdfahbc
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v5, v6); swap(v6, v7); // egfdabhc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v6, v7); // efgdabhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // egdfabhc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v5, v6); // egfdabch
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); // efgdabch
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // egdfabch
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // edghfabc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edgfhabc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // edgfahbc
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // deghfabc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // degfhabc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // degfahbc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efdghabc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // edfghabc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // defghabc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // efdgahbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); // edfgahbc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // defgahbc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v6, v7); // efdgabhc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v6, v7); // edgfabhc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // edfgabhc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // degfabhc
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // defgabhc
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v6); // efdgabch
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v5); // edgfabch
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // edfgabch
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // degfabch
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // defgabch
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v7, v5)) {
														if (compare(v7, v0)) {
															if (compare(v2, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eghdcafb
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egdhcafb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // edghcafb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // deghcafb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egdchafb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edgchafb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // degchafb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edcghafb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // decghafb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // dceghafb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v6, v7); // eghdafcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v5, v7); swap(v6, v7); // eghdacfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); // eghdafbc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v6, v7); // egdhafcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // egdhacfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); // egdhafbc
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); // edghafcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); // edghacfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v6, v7); // edghafbc
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); // deghafcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // deghacfb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // deghafbc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // egdahfcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); // edgahfcb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v7); // degahfcb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egdahcfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // edgahcfb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); // degahcfb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); // egdahfbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // edgahfbc
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v6, v7); // degahfbc
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v5, v7); swap(v6, v7); // egdcahfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // edgcahfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // edcgahfb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // egdachfb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edgachfb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // degcahfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // degachfb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // decgahfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcegahfb
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egdafhcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // edgafhcb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); // degafhcb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // egdafchb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edgafchb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // degafchb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // egdafhbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edgafhbc
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // degafhbc
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v6, v7); // egdcafhb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v6, v7); // edgcafhb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v6, v7); // edcgafhb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v6, v7); // egdacfhb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // edgacfhb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // degcafhb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v6, v7); // degacfhb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // decgafhb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // dcegafhb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // egdafcbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // edgafcbh
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // degafcbh
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egdafbhc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // edgafbhc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v6, v7); // degafbhc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // egdafbch
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // edgafbch
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v5); // degafbch
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); // egdcafbh
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); // edgcafbh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v6); // edcgafbh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); // egdacfbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // edgacfbh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // degcafbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v6); // degacfbh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v6); // decgafbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // dcegafbh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v7, v0)) {
															if (compare(v2, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); // eghdcabf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // egdhcabf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edghcabf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // deghcabf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); // egdchabf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // edgchabf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // degchabf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // edcghabf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // decghabf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // dceghabf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v5, v7); // eghdacbf
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v5, v6); // eghdabfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v5, v6); swap(v6, v7); // eghdabcf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); // egdhacbf
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); // egdhabfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // egdhabcf
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // edghacbf
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // edghabfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); // edghabcf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // deghacbf
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // deghabfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // deghabcf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // egdahcbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edgahcbf
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // degahcbf
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // egdahbfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edgahbfc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // degahbfc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egdahbcf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // edgahbcf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); // degahbcf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v5, v7); // egdcahbf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); // edgcahbf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v5, v7); // edcgahbf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); // egdachbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // edgachbf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // degcahbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); // degachbf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // decgahbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // dcegahbf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v7, v5)) {
																if (compare(v4, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v5, v6); swap(v6, v7); // egdcabhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // edgcabhf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // edcgabhf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // egdacbhf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edgacbhf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // degcabhf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // degacbhf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // decgabhf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcegabhf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v2, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v5, v6); // egdcabfh
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // edgcabfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v3, v6); swap(v5, v6); // edcgabfh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // egdacbfh
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // edgacbfh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v2, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // degcabfh
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // degacbfh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // decgabfh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // dcegabfh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egdabhfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // edgabhfc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v5, v7); swap(v6, v7); // degabhfc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // egdabfhc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v6, v7); // edgabfhc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v6, v7); // degabfhc
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // egdabfch
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); // edgabfch
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v6); // degabfch
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // egdabhcf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); // edgabhcf
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v5, v7); // degabhcf
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egdabchf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // edgabchf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v5, v6); swap(v6, v7); // degabchf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // egdabcfh
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // edgabcfh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v6); swap(v5, v6); // degabcfh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v2)) {
												if (compare(v7, v2)) {
													if (compare(v7, v1)) {
														if (compare(v2, v1)) {
															if (compare(v7, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // edaghfcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // edaghcfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // edaghcbf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); // deaghfcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // deaghcfb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // deaghcbf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // daeghfcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // daeghcfb
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // daeghcbf
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // efdaghcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edfaghcb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // edagfhcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // edafghcb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // defaghcb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // deagfhcb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // deafghcb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // daegfhcb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // daefghcb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); // edaghfbc
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // edaghbfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // edaghbcf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // deaghfbc
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // deaghbfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // deaghbcf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); // daeghfbc
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // daeghbfc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // daeghbcf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efdaghbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // edfaghbc
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // edagfhbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edafghbc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); // defaghbc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // deagfhbc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // deafghbc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // daegfhbc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // daefghbc
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // efdagbhc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edfagbhc
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edagfbhc
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // edafgbhc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // defagbhc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // deagfbhc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // deafgbhc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // daegfbhc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // daefgbhc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edagbhfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // edagbfhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // edagbhcf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // deagbhfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v6, v7); // deagbfhc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); // deagbhcf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // daegbhfc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // daegbfhc
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // daegbhcf
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edabghfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v6, v7); // edabgfhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); // edabghcf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // deabghfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // deabgfhc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // deabghcf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // daebghfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v6, v7); // daebgfhc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); // daebghcf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dabeghfc
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dabegfhc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dabeghcf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efdabghc
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // edfabghc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v5, v6); swap(v6, v7); // defabghc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edafbghc
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // deafbghc
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // daefbghc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edabfghc
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // deabfghc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // daebfghc
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dabefghc
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v7, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efdagchb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // edfagchb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // edagfchb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edafgchb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v6, v7); // defagchb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // deagfchb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // deafgchb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // daegfchb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // daefgchb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // edagchfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v6, v7); // edagcfhb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); // edagchbf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // deagchfb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // deagcfhb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // deagchbf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // daegchfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v6, v7); // daegcfhb
																		}
																	}
																	else {
																		swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); // daegchbf
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // efdagcbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // edfagcbh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // edagfcbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // edafgcbh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v6); // defagcbh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // deagfcbh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // deafgcbh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // daegfcbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // daefgcbh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); // edagcfbh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // edagcbhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // edagcbfh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // deagcfbh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // deagcbhf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // deagcbfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); // daegcfbh
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // daegcbhf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // daegcbfh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // efdagbch
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // edfagbch
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // edagfbch
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // edafgbch
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // defagbch
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // deagfbch
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // deafgbch
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // daegfbch
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // daefgbch
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // edagbfch
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edagbchf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // edagbcfh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); // deagbfch
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // deagbchf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // deagbcfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // daegbfch
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // daegbchf
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // daegbcfh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v6); // edabgfch
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edabgchf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // edabgcfh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // deabgfch
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // deabgchf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // deabgcfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v6); // daebgfch
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // daebgchf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // daebgcfh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // dabegfch
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dabegchf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dabegcfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // efdabgch
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // edfabgch
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v5, v6); // defabgch
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // edafbgch
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // deafbgch
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // daefbgch
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // edabfgch
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // deabfgch
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // daebfgch
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dabefgch
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v5)) {
													if (compare(v7, v5)) {
														if (compare(v2, v0)) {
															if (compare(v7, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edcaghfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // decaghfb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dceaghfb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // dcaeghfb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // edcaghbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v6); swap(v5, v7); // decaghbf
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); // dceaghbf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // dcaeghbf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edcagbhf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // decagbhf
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dceagbhf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dcaegbhf
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // edcabghf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v5, v6); swap(v6, v7); // decabghf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v5, v6); swap(v6, v7); // dceabghf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dcaebghf
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v5, v6); swap(v6, v7); // dcabeghf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // edacghfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // edacghbf
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // edacgbhf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // edacbghf
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v5, v6); swap(v6, v7); // edabcghf
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // deacghfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); // deacghbf
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // deacgbhf
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v5, v6); swap(v6, v7); // deacbghf
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // deabcghf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // daecghfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // daecghbf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // daecgbhf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // daceghfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // daceghbf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // dacegbhf
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // daecbghf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dacebghf
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v5, v6); swap(v6, v7); // dacbeghf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v5, v6); swap(v6, v7); // daebcghf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // dabecghf
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v5, v6); swap(v6, v7); // dabceghf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v0)) {
															if (compare(v5, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // edcagfhb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v6); swap(v6, v7); // decagfhb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v6, v7); // dceagfhb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dcaegfhb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v6); // edcagfbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v6); // decagfbh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v6); // dceagfbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // dcaegfbh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // edcagbfh
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v6); swap(v5, v6); // decagbfh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // dceagbfh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dcaegbfh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v5, v6); // edcabgfh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v5, v6); // decabgfh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v5, v6); // dceabgfh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // dcaebgfh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v5, v6); // dcabegfh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // edacgfhb
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // edacgfbh
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // edacgbfh
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // edacbgfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v5, v6); // edabcgfh
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v6, v7); // deacgfhb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v6); // deacgfbh
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // deacgbfh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v5, v6); // deacbgfh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // deabcgfh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // daecgfhb
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // daecgfbh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // daecgbfh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // dacegfhb
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v6); // dacegfbh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // dacegbfh
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // daecbgfh
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v5, v6); // dacebgfh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v5, v6); // dacbegfh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v5, v6); // daebcgfh
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // dabecgfh
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v5, v6); // dabcegfh
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v2, v0)) {
															if (compare(v7, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v5, v6); swap(v6, v7); // efdcaghb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // edfcaghb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // defcaghb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // edcfaghb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // decfaghb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcefaghb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edcafghb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // decafghb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dceafghb
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dcaefghb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v5, v6); // efdcagbh
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // edfcagbh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // defcagbh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); swap(v5, v6); // edcfagbh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // decfagbh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // dcefagbh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // edcafgbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v4, v5); swap(v5, v6); // decafgbh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // dceafgbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dcaefgbh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // efdacghb
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edfacghb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // defacghb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // edafcghb
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // deafcghb
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // daefcghb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // edacfghb
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // deacfghb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // daecfghb
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // dacefghb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // efdacgbh
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // edfacgbh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // defacgbh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // edafcgbh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // deafcgbh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // daefcgbh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // edacfgbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // deacfgbh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // daecfgbh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // dacefgbh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																if (compare(v2, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); // efdcabgh
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v5); // edfcabgh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // defcabgh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v5); // edcfabgh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v3, v5); swap(v4, v5); // decfabgh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // dcefabgh
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); // efdacbgh
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // edfacbgh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); swap(v4, v5); // defacbgh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // efdabcgh
																			}
																			else {
																				swap(v0, v4); swap(v1, v3); swap(v2, v5); swap(v3, v4); // edfabcgh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v5); // defabcgh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); // edafcbgh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // deafcbgh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); // daefcbgh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // edafbcgh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v5); // deafbcgh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // daefbcgh
																		}
																	}
																}
																else {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v4); swap(v4, v5); // edcafbgh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v4, v5); // decafbgh
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				swap(v0, v3); swap(v1, v2); swap(v2, v4); swap(v4, v5); // dceafbgh
																			}
																			else {
																				swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dcaefbgh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // edacfbgh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v4, v5); // deacfbgh
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // daecfbgh
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); swap(v4, v5); // dacefbgh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v4); swap(v1, v3); swap(v3, v4); // edcabfgh
																			}
																			else {
																				swap(v0, v3); swap(v1, v4); // decabfgh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v4); // dceabfgh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v3, v4); // edacbfgh
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); // deacbfgh
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); swap(v4, v5); // edabfcgh
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // deabfcgh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v4); swap(v1, v3); swap(v2, v4); // edabcfgh
																		}
																		else {
																			swap(v0, v3); swap(v1, v4); swap(v2, v3); swap(v3, v4); // deabcfgh
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v2, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); swap(v3, v4); // dcaebfgh
																		}
																		else {
																			swap(v0, v3); swap(v1, v2); swap(v2, v3); // dcabefgh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v3, v4); // daecbfgh
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v3, v4); // dacebfgh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); // dacbefgh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); swap(v4, v5); // daebfcgh
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v4); // daebcfgh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // dabefcgh
																			}
																			else {
																				swap(v0, v3); swap(v1, v3); swap(v2, v3); swap(v3, v4); // dabecfgh
																			}
																		}
																		else {
																			swap(v0, v3); swap(v1, v3); swap(v2, v3); // dabcefgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
						else {
							if (compare(v5, v4)) {
								if (compare(v7, v6)) {
									if (compare(v6, v4)) {
										if (compare(v6, v1)) {
											if (compare(v7, v0)) {
												if (compare(v6, v5)) {
													if (compare(v6, v0)) {
														if (compare(v4, v1)) {
															if (compare(v5, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgfeadcb
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hgfaedcb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgfadecb
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hgfadceb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hgfeadbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgfeabdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); // hgfaedbc
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hgfadebc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hgfaebdc
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v6, v7); // hgafedcb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hgafdecb
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v6, v7); // hgadfecb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgafdceb
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgadfceb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v6, v7); // hgadcfeb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); // hgafedbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); // hgafebdc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); // hgafdebc
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v5); // hgadfebc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); // hgfadcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgfadbec
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hgfadbce
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hgafdcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hgafdbec
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgafdbce
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); // hgadfcbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v7); // hgadcfbe
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); // hgadfbec
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgadfbce
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v6); // hgadbfec
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v6, v7); // hgadbfce
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); // hgadcbfe
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgadbcfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgfabedc
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hgfabdec
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgfabdce
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hgafbedc
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); // hgafbdec
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hgafbdce
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); // hgabfedc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hgabfdec
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); // hgabdfec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgabfdce
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hgabdfce
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgabdcfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); // hagfedcb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); // hagfdecb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v5); // hagdfecb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); // hadgfecb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hagfdceb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); // hagdfceb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hadgfceb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v6); // hagdcfeb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); // hadgcfeb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v6); // hadcgfeb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hagfdcbe
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hagdfcbe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v6, v7); // hagdcfbe
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hagdcbfe
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hadgfcbe
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hadgcfbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v6, v7); // hadcgfbe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hadgcbfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hadcgbfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v6, v7); // hagfedbc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hagfdebc
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hagfdbec
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hagfdbce
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v4, v1)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hagfebdc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hagfbedc
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hagfbdec
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); // hagfbdce
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v6, v7); // hagdfebc
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hagdfbec
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); // hagdfbce
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hadgfebc
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hadgfbec
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hadgfbce
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v6, v7); // hagdbfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v7); // hagdbfce
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); // hagdbcfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hadgbfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); // hadgbfce
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hadgbcfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hagbfedc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hagbfdec
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hagbdfec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hagbfdce
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); // hagbdfce
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hagbdcfe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v2, v1)) {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hfgeadcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); // fhgeadcb
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); // hfgaedcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); // fhgaedcb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hfagedcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); // fhagedcb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); // hafgedcb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hfgadecb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); // fhgadecb
																			}
																		}
																		else {
																			if (compare(v5, v0)) {
																				if (compare(v7, v5)) {
																					swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfagdecb
																				}
																				else {
																					swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fhagdecb
																				}
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hafgdecb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfadgecb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); // fhadgecb
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); // hafdgecb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hadfgecb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfgadceb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fhgadceb
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hfagdceb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); // fhagdceb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); // hfadgceb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v6); // fhadgceb
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfadcgeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); // fhadcgeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); // hafgdceb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v6); // hafdgceb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); // hadfgceb
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); // hafdcgeb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hadfcgeb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); // hadcfgeb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfgeadbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hfgeabdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hfgaedbc
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfgadebc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v6, v7); // hfgaebdc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fhgeadbc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhgeabdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v6, v7); // fhgaedbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fhgadebc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fhgaebdc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); // hfagedbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fhagedbc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hafgedbc
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); // hfagebdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); // fhagebdc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); // hafgebdc
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hfagdebc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhagdebc
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); // hfadgebc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhadgebc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hafgdebc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hafdgebc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hadfgebc
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v3, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hfgadbec
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhgadbec
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); // hfagdbec
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fhagdbec
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hafgdbec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v6); // hfadgbec
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); // fhadgbec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); // hafdgbec
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hadfgbec
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hfgabedc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhgabedc
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); // hfagbedc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fhagbedc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hafgbedc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfgabdec
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fhgabdec
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hfagbdec
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhagbdec
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hafgbdec
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hfgadcbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhgadcbe
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfagdcbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fhagdcbe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfadgcbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v6, v7); // fhadgcbe
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); // hfadcgbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhadcgbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hafgdcbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v6, v7); // hafdgcbe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hadfgcbe
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hafdcgbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hadfcgbe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hadcfgbe
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); // hfgadbce
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fhgadbce
																			}
																		}
																		else {
																			if (compare(v5, v0)) {
																				if (compare(v7, v5)) {
																					swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hfagdbce
																				}
																				else {
																					swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhagdbce
																				}
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hafgdbce
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v6, v7); // hfadgbce
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhadgbce
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hafdgbce
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hadfgbce
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hfgabdce
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); // fhgabdce
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfagbdce
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fhagbdce
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hafgbdce
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v4, v1)) {
														if (compare(v6, v3)) {
															if (compare(v2, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); // ahgfedcb
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); // ahgfdecb
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ahgfdceb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); // ahfgedcb
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ahfgdecb
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); // ahfgdceb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fahgedcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fahgdecb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fahgdceb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // afhgedcb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afhgdecb
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // afhgdceb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v5)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v6, v7); // ahgfedbc
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ahgfdebc
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ahgfebdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ahfgedbc
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahfgdebc
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); // ahfgebdc
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); // fahgedbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fahgdebc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); // fahgebdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); // afhgedbc
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // afhgdebc
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v6); // afhgebdc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fahdgecb
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fadhgecb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); // ahfdgecb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afhdgecb
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afdhgecb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); // fahdgebc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fadhgebc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahfdgebc
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); // afhdgebc
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // afdhgebc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); // fahdgceb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v4, v6); // ahfdgceb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); // afhdgceb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fahdcgeb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); // ahfdcgeb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afhdcgeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fadhgceb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fadhcgeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fadchgeb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // afdhgceb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afdhcgeb
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afdchgeb
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v4, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahgfbedc
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ahfgbedc
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ahgbfedc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahgfdbec
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ahfgdbec
																			}
																		}
																		else {
																			if (compare(v5, v1)) {
																				if (compare(v6, v5)) {
																					swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ahgfbdec
																				}
																				else {
																					swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahfgbdec
																				}
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahgbfdec
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); // fahgdbec
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); // fahgbedc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fahgbdec
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); // afhgdbec
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); // afhgbedc
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // afhgbdec
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v3, v1)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahgfdcbe
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ahfgdcbe
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ahgfdbce
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahfgdbce
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); // ahgfbdce
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ahfgbdce
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ahgbfdce
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fahgdcbe
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fahgdbce
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fahgbdce
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afhgdcbe
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // afhgdbce
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afhgbdce
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fahdgcbe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v6, v7); // ahfdgcbe
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afhdgcbe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); // fahdcgbe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahfdcgbe
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); // afhdcgbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fadhgcbe
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fadhcgbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); // fadchgbe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afdhgcbe
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // afdhcgbe
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); // afdchgbe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v6); // fahdgbec
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); // fadhgbec
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); // ahfdgbec
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v4, v6); // afhdgbec
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); // afdhgbec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v6, v7); // fahdgbce
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fadhgbce
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahfdgbce
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v6, v7); // afhdgbce
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // afdhgbce
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v3)) {
														if (compare(v2, v1)) {
															if (compare(v4, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v4, v5); // ahgdfecb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); // ahdgfecb
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ahdfgecb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); // ahgdfceb
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ahdgfceb
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); // ahdfgceb
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ahdfcgeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v4, v6); // ahgdcfeb
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); // ahdgcfeb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v7); swap(v2, v3); swap(v4, v6); // ahdcgfeb
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); // ahdcfgeb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahgdfcbe
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahdgfcbe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ahdfgcbe
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahdfcgbe
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v6, v7); // ahgdcfbe
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ahdgcfbe
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v6, v7); // ahdcgfbe
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahdcfgbe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahgdcbfe
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahdgcbfe
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahdcgbfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v3)) {
																		swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v6, v7); // ahgdfebc
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ahdgfebc
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahdfgebc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahgdfbec
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahdgfbec
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ahdfgbec
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); // ahgdfbce
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ahdgfbce
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahdfgbce
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v6, v7); // ahgdbfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v7); swap(v2, v6); swap(v4, v7); // ahgdbfce
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); // ahgdbcfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ahdgbfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); // ahdgbfce
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ahdgbcfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ahgbdfec
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); // ahgbdfce
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ahgbdcfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v5)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // adhgfecb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adhfgecb
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // adfhgecb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); // adhgfebc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // adhfgebc
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adfhgebc
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adhgfceb
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // adhgcfeb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // adhfgceb
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); // adfhgceb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adhfcgeb
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // adfhcgeb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); // adhcgfeb
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // adhcfgeb
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v1, v3); swap(v3, v7); swap(v4, v6); // adchgfeb
																		}
																		else {
																			swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // adchfgeb
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adfchgeb
																		}
																		else {
																			swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adcfhgeb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // adhgfcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // adhgfbec
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adhgfbce
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); // adhgcfbe
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // adhcgfbe
																			}
																			else {
																				swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // adchgfbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // adhgcbfe
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adhcgbfe
																			}
																			else {
																				swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adchgbfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); // adhgbfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // adhgbfce
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adhgbcfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v2, v1)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adhfgcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); // adhfgbec
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // adhfgbce
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // adfhgcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // adfhgbec
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adfhgbce
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // adhfcgbe
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adfhcgbe
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // adfchgbe
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adhcfgbe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adchfgbe
																			}
																			else {
																				swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // adcfhgbe
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v1)) {
												if (compare(v3, v1)) {
													if (compare(v7, v3)) {
														if (compare(v7, v0)) {
															if (compare(v5, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); // hfadbgec
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhadbgec
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfadbgce
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); // fhadbgce
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v7); // hfadcbge
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); // fhadcbge
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); // hfadbcge
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v7); // fhadbcge
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hafdbgec
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hadfbgec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); // hafdbgce
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hadfbgce
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); // hafdcbge
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hadfcbge
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); // hadcfbge
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v7); // hafdbcge
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); // hadfbcge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahfdbgec
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahdfbgec
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); // fahdbgec
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); // afhdbgec
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v6); // ahfdbgce
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ahdfbgce
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fahdbgce
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afhdbgce
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v7); // fahdcbge
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); // ahfdcbge
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v4, v7); // afhdcbge
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ahdfcbge
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); // ahdcfbge
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v5); swap(v4, v7); // ahfdbcge
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); // ahdfbcge
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); // fahdbcge
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); // afhdbcge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); // fadhcbge
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v7); // fadchbge
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fadhbgec
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fadhbgce
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fadhbcge
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fadbhgec
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fadbhgce
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); // fadcbhge
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); // fadbhcge
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fadbchge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); // afdhcbge
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v4, v7); // afdchbge
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // afdhbgec
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afdhbgce
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // afdhbcge
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afdbhgec
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // afdbhgce
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); // afdcbhge
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); // afdbhcge
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // afdbchge
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); // adhfcbge
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // adfhcbge
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); // adfchbge
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // adfcbhge
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // adhcfbge
																		}
																		else {
																			swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // adchfbge
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v1, v3); swap(v3, v5); swap(v4, v7); // adcfhbge
																		}
																		else {
																			swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // adcfbhge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // adhfbgec
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adhfbgce
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // adhfbcge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adfhbgec
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // adfhbgce
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); // adfhbcge
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adfbhgec
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v6); // adfbhgce
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v4, v7); // adfbhcge
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); // adfbchge
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v7, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hfabgedc
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); // hfabgdec
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hfabdgec
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhabgedc
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fhabgdec
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhabdgec
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v3)) {
																		swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fahbgedc
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); // fahbgdec
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fahbdgec
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fabhgedc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fabhgdec
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fabhdgec
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fabdhgec
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hfabgdce
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfabdgce
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hfabdcge
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhabgdce
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fhabdgce
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); // fhabdcge
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fahbgdce
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fahbdgce
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fabhgdce
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fabhdgce
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); // fabdhgce
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fahbdcge
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); // fabhdcge
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v7); // fabdhcge
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); // fabdchge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v7, v5)) {
																if (compare(v7, v0)) {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hafbgedc
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hafbgdec
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hafbdgec
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahfbgedc
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ahfbgdec
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahfbdgec
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v3)) {
																		swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // afhbgedc
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); // afhbgdec
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // afhbdgec
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afbhgedc
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // afbhgdec
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afbhdgec
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afbdhgec
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hafbgdce
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hafbdgce
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); // hafbdcge
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahfbgdce
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ahfbdgce
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); // ahfbdcge
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // afhbgdce
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afhbdgce
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // afhbdcge
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v3)) {
																		swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afbhgdce
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // afbhdgce
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v6); // afbdhgce
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); // afbhdcge
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v4, v7); // afbdhcge
																		}
																	}
																	else {
																		swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); // afbdchge
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v1)) {
													if (compare(v6, v5)) {
														if (compare(v3, v1)) {
															if (compare(v6, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); // hadbgfec
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hadbgfce
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hadbgcfe
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); // ahdbgfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ahdbgfce
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahdbgcfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v4, v6); // adhbgfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v6, v7); // adhbgfce
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adhbgcfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); // hadcbgfe
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); // ahdcbgfe
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // adhcbgfe
																		}
																		else {
																			swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // adchbgfe
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hadbcgfe
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ahdbcgfe
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); // adhbcgfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); // habgfedc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // habgfdec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); // habgdfec
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v6); // habdgfec
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // habgfdce
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // habgdfce
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v6, v7); // habdgfce
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // habgdcfe
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // habdgcfe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); // habdcgfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); // ahbgfedc
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ahbgfdec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); // ahbgdfec
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v4, v6); // ahbdgfec
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahbgfdce
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ahbgdfce
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v6, v7); // ahbdgfce
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahbgdcfe
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahbdgcfe
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); // ahbdcgfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v5, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hadbfgec
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hadbfgce
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hadbfcge
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ahdbfgec
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahdbfgce
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ahdbfcge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); // adhbfgec
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adhbfgce
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); // adhbfcge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v7); // hadcbfge
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v4, v7); // ahdcbfge
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); // adhcbfge
																		}
																		else {
																			swap(v1, v3); swap(v3, v7); swap(v4, v7); // adchbfge
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); // hadbcfge
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); // ahdbcfge
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v4, v7); // adhbcfge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // habfgedc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); // habfgdec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // habfdgec
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); // habdfgec
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // habfgdce
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // habfdgce
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // habdfgce
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // habfdcge
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); // habdfcge
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v7); // habdcfge
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ahbfgedc
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); // ahbfgdec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ahbfdgec
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); // ahbdfgec
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ahbfgdce
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahbfdgce
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahbdfgce
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ahbfdcge
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); // ahbdfcge
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v4, v7); // ahbdcfge
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v2)) {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v2, v7); swap(v3, v6); swap(v4, v5); // abhgfedc
																		}
																		else {
																			swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // abhfgedc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // abhgfdec
																			}
																			else {
																				swap(v2, v7); swap(v3, v5); swap(v4, v6); // abhfgdec
																			}
																		}
																		else {
																			swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // abhfdgec
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abfhgedc
																		}
																		else {
																			swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // abfhgdec
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abfhdgec
																		}
																		else {
																			swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abfdhgec
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v2, v7); swap(v3, v6); swap(v4, v6); // abhgdfec
																		}
																		else {
																			swap(v2, v7); swap(v4, v6); // abhdgfec
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); // adbhgfec
																		}
																		else {
																			swap(v2, v3); swap(v3, v7); swap(v4, v6); // abdhgfec
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v5)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // adbhfgec
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adbfhgec
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v7); swap(v4, v5); swap(v5, v6); // abhdfgec
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // abdhfgec
																			}
																			else {
																				swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abdfhgec
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abhgfdce
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // abhfgdce
																			}
																			else {
																				swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abfhgdce
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abhfdgce
																		}
																		else {
																			swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // abfhdgce
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // abhgdfce
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v2, v7); swap(v4, v6); swap(v6, v7); // abhdgfce
																			}
																			else {
																				swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abhdfgce
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abhgdcfe
																		}
																		else {
																			swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abhdgcfe
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // adbhgfce
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adbhgcfe
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adbhfgce
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // adbfhgce
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // abdhgfce
																		}
																		else {
																			swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abdhgcfe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v5); swap(v4, v7); swap(v5, v6); // abfdhgce
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abdhfgce
																			}
																			else {
																				swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // abdfhgce
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v5, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // abhfdcge
																		}
																		else {
																			swap(v2, v7); swap(v4, v5); swap(v5, v7); // abhdfcge
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // adbhfcge
																		}
																		else {
																			swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // abdhfcge
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v2, v5); swap(v3, v7); swap(v4, v7); // abfhdcge
																		}
																		else {
																			swap(v2, v5); swap(v4, v7); // abfdhcge
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); // adbfhcge
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); swap(v4, v7); // abdfhcge
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v3, v1)) {
																		swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // adbhcgfe
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v7); swap(v4, v7); swap(v5, v6); // abhdcgfe
																		}
																		else {
																			swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // abdhcgfe
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); // adbhcfge
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v7); swap(v4, v7); // abhdcfge
																		}
																		else {
																			swap(v2, v3); swap(v3, v7); swap(v4, v7); // abdhcfge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v6, v5)) {
																	if (compare(v2, v1)) {
																		swap(v1, v3); swap(v4, v7); swap(v5, v6); // adcbhgfe
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); // adbchgfe
																		}
																		else {
																			swap(v2, v3); swap(v4, v7); swap(v5, v6); // abdchgfe
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		swap(v1, v3); swap(v4, v7); // adcbhfge
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v4, v7); // adbchfge
																		}
																		else {
																			swap(v2, v3); swap(v4, v7); // abdchfge
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v2, v1)) {
																		swap(v1, v3); swap(v4, v5); swap(v5, v7); // adcbfhge
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // adbfchge
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); // adbcfhge
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v2, v5); swap(v4, v5); swap(v5, v7); // abfdchge
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // abdfchge
																		}
																	}
																	else {
																		swap(v2, v3); swap(v4, v5); swap(v5, v7); // abdcfhge
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v1)) {
											if (compare(v5, v0)) {
												if (compare(v7, v4)) {
													if (compare(v6, v1)) {
														if (compare(v2, v1)) {
															if (compare(v6, v2)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfegadcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fhegadcb
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hfeagdcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); // fheagdcb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfaegdcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fhaegdcb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fahegdcb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfeadgcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fheadgcb
																			}
																		}
																		else {
																			if (compare(v7, v0)) {
																				if (compare(v7, v5)) {
																					swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hfaedgcb
																				}
																				else {
																					swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); // fhaedgcb
																				}
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fahedgcb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v5, v6); swap(v6, v7); // hfadegcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v5, v6); // fhadegcb
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v5, v6); swap(v6, v7); // fahdegcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fadhegcb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hfeadcgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); // fheadcgb
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); // hfaedcgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); // fhaedcgb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v5, v7); // hfadecgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); // fhadecgb
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); // hfadcegb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v5); // fhadcegb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); // fahedcgb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v5, v7); // fahdecgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); // fadhecgb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); // fahdcegb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fadhcegb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); // fadchegb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hfegadbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); // hfegabdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfeagdbc
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hfeadgbc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hfeagbdc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhegadbc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fhegabdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fheagdbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fheadgbc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fheagbdc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hfaegdbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhaegdbc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fahegdbc
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); // hfaegbdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fhaegbdc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); // fahegbdc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); // hfaedgbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fhaedgbc
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v5, v6); // hfadegbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v5, v6); swap(v6, v7); // fhadegbc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); // fahedgbc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v5, v6); // fahdegbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); // fadhegbc
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v3, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); // hfeadbgc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fheadbgc
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); // hfaedbgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); // fhaedbgc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); // fahedbgc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); // hfadebgc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v5, v7); // fhadebgc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); // fahdebgc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); // fadhebgc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfeabgdc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fheabgdc
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hfaebgdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhaebgdc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fahebgdc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hfeabdgc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); // fheabdgc
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); // hfaebdgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fhaebdgc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); // fahebdgc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfeadcbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fheadcbg
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hfaedcbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v6, v7); // fhaedcbg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v5, v7); swap(v6, v7); // hfadecbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v6, v7); // fhadecbg
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfadcebg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v6, v7); // fhadcebg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fahedcbg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v5, v7); swap(v6, v7); // fahdecbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fadhecbg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fahdcebg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fadhcebg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fadchebg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hfeadbcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fheadbcg
																			}
																		}
																		else {
																			if (compare(v7, v0)) {
																				if (compare(v7, v5)) {
																					swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v6, v7); // hfaedbcg
																				}
																				else {
																					swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fhaedbcg
																				}
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v6, v7); // fahedbcg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v6, v7); // hfadebcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v5, v7); swap(v6, v7); // fhadebcg
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v6, v7); // fahdebcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v6, v7); // fadhebcg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfeabdcg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fheabdcg
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hfaebdcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhaebdcg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fahebdcg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v1)) {
															if (compare(v7, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fehgadcb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fehagdcb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fehadgcb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fehadcgb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fehgadbc
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); // fehgabdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); // fehagdbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fehadgbc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fehagbdc
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // feahgdcb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // feahdgcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); // feadhgcb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); // feahdcgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v7); // feadhcgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); // feadchgb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // feahgdbc
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // feahgbdc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // feahdgbc
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // feadhgbc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fehadcbg
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); // fehadbgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fehadbcg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // feahdcbg
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // feahdbgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // feahdbcg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v6, v7); // feadhcbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // feadchbg
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); // feadhbgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // feadhbcg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v5, v7); // feadbhgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v5, v7); swap(v6, v7); // feadbhcg
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v6, v7); // feadcbhg
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v6, v7); // feadbchg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); // fehabgdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); // fehabdgc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v6, v7); // fehabdcg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // feahbgdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); // feahbdgc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // feahbdcg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // feabhgdc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // feabhdgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); // feabdhgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // feabhdcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // feabdhcg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v6, v7); // feabdchg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v6, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // faehgdcb
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // faehdgcb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // faedhgcb
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fadehgcb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // faehdcgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); // faedhcgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fadehcgb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v5, v7); // faedchgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); // fadechgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v5, v7); // fadcehgb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // faehdcbg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // faedhcbg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v5, v7); swap(v6, v7); // faedchbg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v6, v7); // faedcbhg
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fadehcbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fadechbg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v5, v7); swap(v6, v7); // fadcehbg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v6, v7); // fadecbhg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v6, v7); // fadcebhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // faehgdbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // faehdgbc
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); // faehdbgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // faehdbcg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // faehgbdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // faehbgdc
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // faehbdgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // faehbdcg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); // faedhgbc
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v7); // faedhbgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v6, v7); // faedhbcg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fadehgbc
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); // fadehbgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fadehbcg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); // faedbhgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // faedbhcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v6, v7); // faedbchg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fadebhgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fadebhcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fadebchg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // faebhgdc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); // faebhdgc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // faebdhgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // faebhdcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // faebdhcg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // faebdchg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v3)) {
													if (compare(v6, v1)) {
														if (compare(v4, v3)) {
															if (compare(v2, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hafegdcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); // hafedgcb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); // hafedcgb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ahfegdcb
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); // ahfedgcb
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v4); // ahfedcgb
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afhegdcb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // afhedgcb
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); // afhedcgb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // afehgdcb
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afehdgcb
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // afehdcgb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hafegdbc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hafedgbc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hafegbdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahfegdbc
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ahfedgbc
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ahfegbdc
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // afhegdbc
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); // afhedgbc
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); // afhegbdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afehgdbc
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // afehdgbc
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // afehgbdc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v5, v6); // hafdegcb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); // hadfegcb
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v5, v6); // ahfdegcb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v5, v6); swap(v6, v7); // afhdegcb
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); // ahdfegcb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v5, v6); swap(v6, v7); // hafdegbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hadfegbc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v5, v6); swap(v6, v7); // ahfdegbc
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v5, v6); // afhdegbc
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ahdfegbc
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); // hafdecgb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); // ahfdecgb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v5, v7); // afhdecgb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v5); // hafdcegb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v4, v5); // ahfdcegb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); // afhdcegb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); // hadfecgb
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); // hadfcegb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v5); // hadcfegb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); // ahdfecgb
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); // ahdfcegb
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v4, v5); // ahdcfegb
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); // hafedbgc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hafebgdc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hafebdgc
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); // ahfedbgc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahfebgdc
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ahfebdgc
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // afhebgdc
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afehbgdc
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // afebhgdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); // afhedbgc
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); // afehdbgc
																			}
																		}
																		else {
																			if (compare(v7, v1)) {
																				if (compare(v7, v4)) {
																					swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); // afhebdgc
																				}
																				else {
																					swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // afehbdgc
																				}
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); // afebhdgc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v6, v7); // hafedcbg
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hafedbcg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hafebdcg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v6, v7); // ahfedcbg
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ahfedbcg
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahfebdcg
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // afhedcbg
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afehdcbg
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v6, v7); // afhedbcg
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // afehdbcg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // afhebdcg
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afehbdcg
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // afebhdcg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v6, v7); // hafdecbg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v6, v7); // ahfdecbg
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v5, v7); swap(v6, v7); // afhdecbg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v6, v7); // hafdcebg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v6, v7); // ahfdcebg
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afhdcebg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v6, v7); // hadfecbg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hadfcebg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v6, v7); // hadcfebg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v6, v7); // ahdfecbg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ahdfcebg
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v6, v7); // ahdcfebg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v5, v7); // hafdebgc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); // hadfebgc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v5, v7); // ahfdebgc
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); // afhdebgc
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); // ahdfebgc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v5, v7); swap(v6, v7); // hafdebcg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hadfebcg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v5, v7); swap(v6, v7); // ahfdebcg
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v6, v7); // afhdebcg
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ahdfebcg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v2, v1)) {
															if (compare(v6, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afedhgcb
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // afdhegcb
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afdehgcb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v4, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); // afdhecgb
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // afdhcegb
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); // afedhcgb
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // afdehcgb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v4); swap(v5, v7); // afedchgb
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); // afdechgb
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); // afdchegb
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v5, v7); // afdcehgb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afedhcbg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // afdhecbg
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afdehcbg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v5, v7); swap(v6, v7); // afedchbg
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // afdechbg
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v6, v7); // afedcbhg
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v6, v7); // afdecbhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afdhcebg
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afdchebg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v1, v5); swap(v2, v3); swap(v5, v7); swap(v6, v7); // afdcehbg
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v6, v7); // afdcebhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v3)) {
																		swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); // afedhgbc
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); // afdhegbc
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // afdehgbc
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v4); swap(v4, v7); // afedhbgc
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v7); // afdhebgc
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); // afdehbgc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v6, v7); // afedhbcg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v6, v7); // afdhebcg
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // afdehbcg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); // afedbhgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afedbhcg
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v6, v7); // afedbchg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // afdebhgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afdebhcg
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // afdebchg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // afebdhgc
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afebdhcg
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // afebdchg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // adhfegcb
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); // adfhegcb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // adfehgcb
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); // adhfegbc
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // adfhegbc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adfehgbc
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v5)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); // adhfecgb
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v7); // adfhecgb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); // adfehcgb
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // adfechgb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // adhfcegb
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); // adfhcegb
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // adfchegb
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); // adfcehgb
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); // adhcfegb
																		}
																		else {
																			swap(v1, v3); swap(v3, v7); swap(v4, v5); // adchfegb
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // adcfhegb
																		}
																		else {
																			swap(v1, v3); swap(v3, v5); swap(v5, v7); // adcfehgb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v4)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v2, v1)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // adhfecbg
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v3); swap(v2, v7); swap(v3, v5); // adhfebgc
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v6, v7); // adhfebcg
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v6, v7); // adfhecbg
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); // adfhebgc
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // adfhebcg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adhfcebg
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // adfhcebg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adfchebg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // adhcfebg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // adchfebg
																			}
																			else {
																				swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adcfhebg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // adfehcbg
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // adfehbgc
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adfehbcg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adfechbg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // adfcehbg
																			}
																			else {
																				swap(v1, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // adcfehbg
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // adfecbhg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v6, v7); // adfcebhg
																			}
																			else {
																				swap(v1, v3); swap(v3, v5); swap(v6, v7); // adcfebhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); // adfebhgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // adfebhcg
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v6, v7); // adfebchg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v1)) {
												if (compare(v3, v1)) {
													if (compare(v5, v3)) {
														if (compare(v5, v0)) {
															if (compare(v7, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v5); // hfadbegc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); // fhadbegc
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v6, v7); // hfadbecg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhadbecg
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v6, v7); // hfadcbeg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhadcbeg
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfadbceg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v6, v7); // fhadbceg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v5); // fahdbegc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); // fadhbegc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v6, v7); // fahdbecg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fadhbecg
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v6, v7); // fahdcbeg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fadhcbeg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v6, v7); // fadchbeg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fahdbceg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fadhbceg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); // hafdbegc
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); // ahfdbegc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v7); swap(v4, v5); // afhdbegc
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); // afdhbegc
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hafdbecg
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahfdbecg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v6, v7); // afhdbecg
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // afdhbecg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hafdcbeg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahfdcbeg
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v6, v7); // afhdcbeg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // afdhcbeg
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v6, v7); // afdchbeg
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v6, v7); // hafdbceg
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v4, v7); swap(v6, v7); // ahfdbceg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afhdbceg
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afdhbceg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v7, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hadfcbeg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hadcfbeg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hadfbegc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hadfbecg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hadfbceg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); // hadbfegc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hadbfecg
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v6, v7); // hadcbfeg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hadbfceg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hadbcfeg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahdfcbeg
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahdcfbeg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ahdfbegc
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahdfbecg
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ahdfbceg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); // ahdbfegc
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ahdbfecg
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v6, v7); // ahdcbfeg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahdbfceg
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ahdbcfeg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // adhfcbeg
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adfhcbeg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adhcfbeg
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // adhcbfeg
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v1)) {
																			swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adchfbeg
																		}
																		else {
																			swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // adchbfeg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // adfchbeg
																		}
																		else {
																			swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // adcfhbeg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); // adhfbegc
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // adhfbecg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adhfbceg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // adfhbegc
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adfhbecg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // adfhbceg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v4, v5); // adhbfegc
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v6, v7); // adhbfecg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adhbfceg
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v6, v7); // adhbcfeg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v0)) {
														if (compare(v6, v2)) {
															if (compare(v5, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); // hfabegdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); // hfabedgc
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); // hfabdegc
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fhabegdc
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); // fhabedgc
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fhabdegc
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hafbegdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); // hafbedgc
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hafbdegc
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); // habfegdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); // habfedgc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); // habfdegc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v5); // habdfegc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v6, v7); // hfabedcg
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hfabdecg
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfabdceg
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fhabedcg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhabdecg
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fhabdceg
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hafbedcg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hafbdecg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v6, v7); // habfedcg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // habfdecg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v6, v7); // habdfecg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hafbdceg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // habfdceg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // habdfceg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v6, v7); // habdcfeg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v7, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ahfbegdc
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); // ahfbedgc
																		}
																	}
																	else {
																		swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ahfbdegc
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); // ahbfegdc
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); // ahbfedgc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); // ahbfdegc
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v4, v5); // ahbdfegc
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); // fahbegdc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); // fahbedgc
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); // fahbdegc
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); // afhbegdc
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v5); // afhbedgc
																		}
																	}
																	else {
																		swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); // afhbdegc
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v6, v7); // fahbedcg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fahbdecg
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fahbdceg
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ahfbedcg
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahfbdecg
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ahfbdceg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v6, v7); // afhbedcg
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // afhbdecg
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afhbdceg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v3)) {
																		swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v6, v7); // ahbfedcg
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ahbfdecg
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v6, v7); // ahbdfecg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahbfdceg
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahbdfceg
																		}
																	}
																	else {
																		swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v6, v7); // ahbdcfeg
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v7, v4)) {
														if (compare(v3, v1)) {
															if (compare(v7, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fadbhegc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fadbhecg
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fadbhceg
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // afdbhegc
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afdbhecg
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // afdbhceg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); // adfbhegc
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adfbhecg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v4, v7); swap(v6, v7); // adfbhceg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fadcbheg
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afdcbheg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adfcbheg
																		}
																		else {
																			swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adcfbheg
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fadbcheg
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afdbcheg
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adfbcheg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fabhegdc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); // fabhedgc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fabhdegc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); // fabdhegc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fabhedcg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fabhdecg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fabdhecg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fabhdceg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v6, v7); // fabdhceg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fabdcheg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // afbhegdc
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); // afbhedgc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // afbhdegc
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); // afbdhegc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // afbhedcg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afbhdecg
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afbdhecg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // afbhdceg
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v4, v7); swap(v6, v7); // afbdhceg
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afbdcheg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v4, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v2)) {
																		swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); // fadbehgc
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fadbehcg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v6, v7); // fadbechg
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); // afdbehgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // afdbehcg
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v6, v7); // afdbechg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v5, v7); // adfbehgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v5, v7); swap(v6, v7); // adfbehcg
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v6, v7); // adfbechg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v6, v7); // fadcbehg
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v6, v7); // afdcbehg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // adfcbehg
																		}
																		else {
																			swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // adcfbehg
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fadbcehg
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // afdbcehg
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v6, v7); // adfbcehg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fabehgdc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fabehdgc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); // fabedhgc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v5, v7); // fabdehgc
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fabehdcg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fabedhcg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v5, v7); swap(v6, v7); // fabdehcg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v6, v7); // fabedchg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v6, v7); // fabdechg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v6, v7); // fabdcehg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afbehgdc
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // afbehdgc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); // afbedhgc
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v5, v7); // afbdehgc
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afbehdcg
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // afbedhcg
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v5, v7); swap(v6, v7); // afbdehcg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v6, v7); // afbedchg
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v6, v7); // afbdechg
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v6, v7); // afbdcehg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v6, v2)) {
															if (compare(v7, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v2, v7); swap(v3, v5); swap(v5, v6); // abhfegdc
																		}
																		else {
																			swap(v2, v7); swap(v3, v5); // abhfedgc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v7); swap(v3, v5); swap(v4, v5); // abhfdegc
																		}
																		else {
																			swap(v2, v7); swap(v4, v5); // abhdfegc
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v4)) {
																			swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // abfhegdc
																		}
																		else {
																			swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abfehgdc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v2, v5); swap(v3, v7); swap(v5, v7); // abfhedgc
																			}
																			else {
																				swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // abfehdgc
																			}
																		}
																		else {
																			swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // abfhdegc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v4)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v5)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); // adbhfegc
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // adbfhegc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v5); swap(v4, v7); swap(v5, v7); // abfdhegc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v3); swap(v3, v7); swap(v4, v5); // abdhfegc
																			}
																			else {
																				swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // abdfhegc
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v2, v5); swap(v3, v4); swap(v5, v7); // abfedhgc
																		}
																		else {
																			swap(v2, v5); swap(v5, v7); // abfdehgc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); // adbfehgc
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); swap(v5, v7); // abdfehgc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v2, v7); swap(v3, v5); swap(v6, v7); // abhfedcg
																			}
																			else {
																				swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // abfhedcg
																			}
																		}
																		else {
																			swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abfehdcg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // abfedhcg
																		}
																		else {
																			swap(v2, v5); swap(v3, v4); swap(v6, v7); // abfedchg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // abhfdecg
																		}
																		else {
																			swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abfhdecg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abfdhecg
																			}
																			else {
																				swap(v2, v5); swap(v5, v7); swap(v6, v7); // abfdehcg
																			}
																		}
																		else {
																			swap(v2, v5); swap(v6, v7); // abfdechg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v5)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // adbhfecg
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adbfhecg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // adbfehcg
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v6, v7); // adbfechg
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			swap(v2, v7); swap(v4, v5); swap(v6, v7); // abhdfecg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // abdhfecg
																			}
																			else {
																				swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abdfhecg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // abdfehcg
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); swap(v6, v7); // abdfechg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v7, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abhfdceg
																		}
																		else {
																			swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abhdfceg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adbhfceg
																		}
																		else {
																			swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abdhfceg
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // abfhdceg
																		}
																		else {
																			swap(v2, v5); swap(v4, v7); swap(v6, v7); // abfdhceg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // adbfhceg
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // abdfhceg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v4)) {
																	if (compare(v3, v1)) {
																		swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adbfcheg
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abfdcheg
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abdfcheg
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // adbfcehg
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v5); swap(v4, v5); swap(v6, v7); // abfdcehg
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // abdfcehg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v3, v1)) {
																	if (compare(v2, v1)) {
																		swap(v1, v3); swap(v4, v7); swap(v6, v7); // adcbhfeg
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // adbhcfeg
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v6, v7); // adbchfeg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v2, v7); swap(v4, v7); swap(v6, v7); // abhdcfeg
																		}
																		else {
																			swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // abdhcfeg
																		}
																	}
																	else {
																		swap(v2, v3); swap(v4, v7); swap(v6, v7); // abdchfeg
																	}
																}
															}
															else {
																if (compare(v7, v4)) {
																	if (compare(v2, v1)) {
																		swap(v1, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adcbfheg
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adbcfheg
																		}
																		else {
																			swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abdcfheg
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		swap(v1, v3); swap(v4, v5); swap(v6, v7); // adcbfehg
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v6, v7); // adbcfehg
																		}
																		else {
																			swap(v2, v3); swap(v4, v5); swap(v6, v7); // abdcfehg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v7, v4)) {
										if (compare(v7, v1)) {
											if (compare(v6, v0)) {
												if (compare(v7, v5)) {
													if (compare(v7, v0)) {
														if (compare(v4, v1)) {
															if (compare(v5, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ghfeadcb
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); // ghfaedcb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ghfadecb
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); // ghfadceb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghfeadbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ghfeabdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ghfaedbc
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghfadebc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); // ghfaebdc
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); // ghafedcb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); // ghafdecb
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v5); // ghadfecb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ghafdceb
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); // ghadfceb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v6); // ghadcfeb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v6, v7); // ghafedbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ghafebdc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ghafdebc
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v6, v7); // ghadfebc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ghfadcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ghfadbec
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghfadbce
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghafdcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghafdbec
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ghafdbce
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v2, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghadfcbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v6, v7); // ghadcfbe
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghadfbec
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v7); // ghadfbce
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v6, v7); // ghadbfec
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v7); // ghadbfce
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghadcbfe
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); // ghadbcfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ghfabedc
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghfabdec
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ghfabdce
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghafbedc
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ghafbdec
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); // ghafbdce
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ghabfedc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghabfdec
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ghabdfec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ghabfdce
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); // ghabdfce
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ghabdcfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v6, v7); // gahfedcb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gahfdecb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v6, v7); // gahdfecb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gadhfecb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gahfdceb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gahdfceb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gadhfceb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v6, v7); // gahdcfeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gadhcfeb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v6, v7); // gadchfeb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gahfdcbe
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); // gahdfcbe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v7); // gahdcfbe
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); // gahdcbfe
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gadhfcbe
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); // gadhcfbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v7); // gadchfbe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gadhcbfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); // gadchbfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); // gahfedbc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); // gahfdebc
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gahfdbec
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gahfdbce
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v4, v1)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); // gahfebdc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gahfbedc
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); // gahfbdec
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gahfbdce
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v5); // gahdfebc
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); // gahdfbec
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gahdfbce
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); // gadhfebc
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gadhfbec
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gadhfbce
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v6); // gahdbfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v6, v7); // gahdbfce
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gahdbcfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); // gadhbfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gadhbfce
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gadhbcfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); // gahbfedc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gahbfdec
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); // gahbdfec
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gahbfdce
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gahbdfce
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gahbdcfe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v2, v1)) {
															if (compare(v4, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfheadcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fgheadcb
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gfhaedcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v6, v7); // fghaedcb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); // gfahedcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fgahedcb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gafhedcb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfhadecb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fghadecb
																			}
																		}
																		else {
																			if (compare(v5, v0)) {
																				if (compare(v6, v5)) {
																					swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gfahdecb
																				}
																				else {
																					swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgahdecb
																				}
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gafhdecb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); // gfadhecb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgadhecb
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gafdhecb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gadfhecb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gfhadceb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fghadceb
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfahdceb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fgahdceb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfadhceb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v6, v7); // fgadhceb
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); // gfadcheb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgadcheb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gafhdceb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v6, v7); // gafdhceb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gadfhceb
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gafdcheb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gadfcheb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gadcfheb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gfheadbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); // gfheabdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); // gfhaedbc
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gfhadebc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); // gfhaebdc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); // fgheadbc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fgheabdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); // fghaedbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); // fghadebc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); // fghaebdc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gfahedbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); // fgahedbc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); // gafhedbc
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v6, v7); // gfahebdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fgahebdc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gafhebdc
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfahdebc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fgahdebc
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfadhebc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); // fgadhebc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gafhdebc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); // gafdhebc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gadfhebc
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v3, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); // gfhadbec
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fghadbec
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gfahdbec
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgahdbec
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gafhdbec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v6, v7); // gfadhbec
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgadhbec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gafdhbec
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gadfhbec
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); // gfhabedc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fghabedc
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gfahbedc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgahbedc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gafhbedc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gfhabdec
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); // fghabdec
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfahbdec
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fgahbdec
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gafhbdec
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfhadcbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fghadcbe
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gfahdcbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); // fgahdcbe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); // gfadhcbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v7); // fgadhcbe
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfadchbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); // fgadchbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); // gafhdcbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v7); // gafdhcbe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); // gadfhcbe
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); // gafdchbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gadfchbe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); // gadcfhbe
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gfhadbce
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fghadbce
																			}
																		}
																		else {
																			if (compare(v5, v0)) {
																				if (compare(v6, v5)) {
																					swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); // gfahdbce
																				}
																				else {
																					swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fgahdbce
																				}
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gafhdbce
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v7); // gfadhbce
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); // fgadhbce
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); // gafdhbce
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gadfhbce
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfhabdce
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fghabdce
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gfahbdce
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgahbdce
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gafhbdce
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v4, v1)) {
														if (compare(v7, v3)) {
															if (compare(v2, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v6, v7); // aghfedcb
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // aghfdecb
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aghfdceb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // agfhedcb
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agfhdecb
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // agfhdceb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); // faghedcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // faghdecb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // faghdceb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); // afghedcb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // afghdecb
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afghdceb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v5)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); // aghfedbc
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); // aghfdebc
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); // aghfebdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); // agfhedbc
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // agfhdebc
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // agfhebdc
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // faghedbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // faghdebc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v6, v7); // faghebdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // afghedbc
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afghdebc
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v6, v7); // afghebdc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); // fagdhecb
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fadghecb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agfdhecb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); // afgdhecb
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // afdghecb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fagdhebc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fadghebc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); // agfdhebc
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afgdhebc
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afdghebc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fagdhceb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v6, v7); // agfdhceb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afgdhceb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); // fagdcheb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agfdcheb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); // afgdcheb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fadghceb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fadgcheb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); // fadcgheb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afdghceb
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // afdgcheb
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); // afdcgheb
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v4, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // aghfbedc
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agfhbedc
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); // aghbfedc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // aghfdbec
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agfhdbec
																			}
																		}
																		else {
																			if (compare(v5, v1)) {
																				if (compare(v7, v5)) {
																					swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); // aghfbdec
																				}
																				else {
																					swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // agfhbdec
																				}
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // aghbfdec
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // faghdbec
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // faghbedc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // faghbdec
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // afghdbec
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // afghbedc
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afghbdec
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v3, v1)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // aghfdcbe
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); // agfhdcbe
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aghfdbce
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // agfhdbce
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // aghfbdce
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agfhbdce
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aghbfdce
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // faghdcbe
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); // faghdbce
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // faghbdce
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // afghdcbe
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); // afghdbce
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // afghbdce
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); // fagdhcbe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v4, v7); // agfdhcbe
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); // afgdhcbe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fagdchbe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); // agfdchbe
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afgdchbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fadghcbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fadgchbe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fadcghbe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // afdghcbe
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afdgchbe
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afdcghbe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v6, v7); // fagdhbec
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fadghbec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agfdhbec
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v6, v7); // afgdhbec
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // afdghbec
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v7); // fagdhbce
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); // fadghbce
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v6); // agfdhbce
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v4, v7); // afgdhbce
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); // afdghbce
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v3)) {
														if (compare(v2, v1)) {
															if (compare(v4, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v6, v7); // aghdfecb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // agdhfecb
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agdfhecb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aghdfceb
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agdhfceb
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // agdfhceb
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agdfcheb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v6, v7); // aghdcfeb
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // agdhcfeb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v6, v7); // agdchfeb
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agdcfheb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); // aghdfcbe
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // agdhfcbe
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); // agdfhcbe
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // agdfchbe
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v4, v7); // aghdcfbe
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); // agdhcfbe
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v3); swap(v4, v7); // agdchfbe
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); // agdcfhbe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); // aghdcbfe
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // agdhcbfe
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); // agdchbfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v3)) {
																		swap(v1, v6); swap(v2, v7); swap(v4, v5); // aghdfebc
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); // agdhfebc
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // agdfhebc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); // aghdfbec
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // agdhfbec
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agdfhbec
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aghdfbce
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agdhfbce
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // agdfhbce
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v7); swap(v4, v6); // aghdbfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v6, v7); // aghdbfce
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aghdbcfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); // agdhbfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // agdhbfce
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agdhbcfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); // aghbdfec
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // aghbdfce
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aghbdcfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v7, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); // adghfecb
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // adgfhecb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adfghecb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // adghfebc
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adgfhebc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // adfghebc
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // adghfceb
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); // adghcfeb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adgfhceb
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // adfghceb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // adgfcheb
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adfgcheb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // adgchfeb
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adgcfheb
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // adcghfeb
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adcgfheb
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // adfcgheb
																		}
																		else {
																			swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // adcfgheb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adghfcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adghfbec
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // adghfbce
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // adghcfbe
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); // adgchfbe
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v4, v7); // adcghfbe
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adghcbfe
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // adgchbfe
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // adcghbfe
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // adghbfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); // adghbfce
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // adghbcfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v2, v1)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // adgfhcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // adgfhbec
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); // adgfhbce
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); // adfghcbe
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adfghbec
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // adfghbce
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adgfchbe
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // adfgchbe
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adfcghbe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // adgcfhbe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // adcgfhbe
																			}
																			else {
																				swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adcfghbe
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v1)) {
												if (compare(v3, v1)) {
													if (compare(v6, v3)) {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfadbhec
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); // fgadbhec
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); // gfadbhce
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgadbhce
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v6, v7); // gfadcbhe
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgadcbhe
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfadbche
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v6, v7); // fgadbche
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); // gafdbhec
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gadfbhec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gafdbhce
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gadfbhce
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gafdcbhe
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gadfcbhe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gadcfbhe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v6, v7); // gafdbche
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gadfbche
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); // agfdbhec
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // agdfbhec
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fagdbhec
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afgdbhec
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agfdbhce
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agdfbhce
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); // fagdbhce
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v7); // afgdbhce
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v6, v7); // fagdcbhe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agfdcbhe
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v6, v7); // afgdcbhe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agdfcbhe
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agdcfbhe
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v6, v7); // agfdbche
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // agdfbche
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fagdbche
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afgdbche
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v5, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fadgcbhe
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v6, v7); // fadcgbhe
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fadgbhec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fadgbhce
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fadgbche
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fadbghec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fadbghce
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fadbgche
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fadcbghe
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fadbcghe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // afdgcbhe
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v6, v7); // afdcgbhe
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afdgbhec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // afdgbhce
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afdgbche
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // afdbghec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afdbghce
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // afdbgche
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afdcbghe
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afdbcghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // adgfcbhe
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adfgcbhe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // adfcgbhe
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adfcbghe
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adgcfbhe
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adcgfbhe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // adcfgbhe
																		}
																		else {
																			swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adcfbghe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adgfbhec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // adgfbhce
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adgfbche
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // adfgbhec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adfgbhce
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // adfgbche
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); // adfbghec
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adfbghce
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v6, v7); // adfbgche
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adfbcghe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v2)) {
															if (compare(v6, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfabhedc
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gfabhdec
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfabdhec
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fgabhedc
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgabhdec
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fgabdhec
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v3)) {
																		swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fagbhedc
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fagbhdec
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fagbdhec
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fabghedc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fabghdec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fabgdhec
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); // fabdghec
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); // gfabhdce
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gfabdhce
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfabdche
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fgabhdce
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgabdhce
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fgabdche
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); // fagbhdce
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fagbdhce
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fabghdce
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fabgdhce
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fabdghce
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fagbdche
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fabgdche
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v6, v7); // fabdgche
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fabdcghe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v5)) {
																if (compare(v6, v0)) {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gafbhedc
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gafbhdec
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gafbdhec
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // agfbhedc
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agfbhdec
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // agfbdhec
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v3)) {
																		swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afgbhedc
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // afgbhdec
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afgbdhec
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // afbghedc
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afbghdec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // afbgdhec
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); // afbdghec
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gafbhdce
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gafbdhce
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gafbdche
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // agfbhdce
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agfbdhce
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // agfbdche
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); // afgbhdce
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // afgbdhce
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afgbdche
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // afbghdce
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afbgdhce
																		}
																	}
																	else {
																		swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afbdghce
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // afbgdche
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v6, v7); // afbdgche
																		}
																	}
																	else {
																		swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afbdcghe
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v1)) {
													if (compare(v7, v5)) {
														if (compare(v3, v1)) {
															if (compare(v7, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gadbhfec
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); // gadbhfce
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gadbhcfe
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // agdbhfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); // agdbhfce
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // agdbhcfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v6, v7); // adgbhfec
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v3); swap(v2, v6); swap(v4, v7); // adgbhfce
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); // adgbhcfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gadcbhfe
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agdcbhfe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adgcbhfe
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adcgbhfe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gadbchfe
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agdbchfe
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adgbchfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gabhfedc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gabhfdec
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gabhdfec
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v6, v7); // gabdhfec
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gabhfdce
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); // gabhdfce
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v7); // gabdhfce
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gabhdcfe
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); // gabdhcfe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gabdchfe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // agbhfedc
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agbhfdec
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // agbhdfec
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v6, v7); // agbdhfec
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // agbhfdce
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); // agbhdfce
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v4, v7); // agbdhfce
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // agbhdcfe
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); // agbdhcfe
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agbdchfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v5, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gadbfhec
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gadbfhce
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gadbfche
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agdbfhec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // agdbfhce
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agdbfche
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adgbfhec
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v7); // adgbfhce
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adgbfche
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v6, v7); // gadcbfhe
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v6, v7); // agdcbfhe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // adgcbfhe
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // adcgbfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gadbcfhe
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // agdbcfhe
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v6, v7); // adgbcfhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gabfhedc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gabfhdec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gabfdhec
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gabdfhec
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); // gabfhdce
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gabfdhce
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); // gabdfhce
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gabfdche
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gabdfche
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v6, v7); // gabdcfhe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agbfhedc
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // agbfhdec
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agbfdhec
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agbdfhec
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); // agbfhdce
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // agbfdhce
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v7); // agbdfhce
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agbfdche
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agbdfche
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v6, v7); // agbdcfhe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v2)) {
														if (compare(v4, v2)) {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // abghfedc
																		}
																		else {
																			swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abgfhedc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abghfdec
																			}
																			else {
																				swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // abgfhdec
																			}
																		}
																		else {
																			swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abgfdhec
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // abfghedc
																		}
																		else {
																			swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abfghdec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // abfgdhec
																		}
																		else {
																			swap(v2, v5); swap(v4, v6); swap(v5, v7); // abfdghec
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // abghdfec
																		}
																		else {
																			swap(v2, v6); swap(v4, v7); swap(v6, v7); // abgdhfec
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // adbghfec
																		}
																		else {
																			swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // abdghfec
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adbgfhec
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // adbfghec
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abgdfhec
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abdgfhec
																			}
																			else {
																				swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // abdfghec
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // abghfdce
																			}
																			else {
																				swap(v2, v6); swap(v3, v5); swap(v4, v7); // abgfhdce
																			}
																		}
																		else {
																			swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // abfghdce
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // abgfdhce
																		}
																		else {
																			swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abfgdhce
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v2, v6); swap(v3, v7); swap(v4, v7); // abghdfce
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v6); swap(v4, v7); // abgdhfce
																			}
																			else {
																				swap(v2, v6); swap(v4, v5); swap(v5, v7); // abgdfhce
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // abghdcfe
																		}
																		else {
																			swap(v2, v6); swap(v4, v7); swap(v5, v6); // abgdhcfe
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); // adbghfce
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // adbghcfe
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // adbgfhce
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adbfghce
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v2, v3); swap(v3, v6); swap(v4, v7); // abdghfce
																			}
																			else {
																				swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // abdgfhce
																			}
																		}
																		else {
																			swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // abdghcfe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abfdghce
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abdfghce
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abgfdche
																		}
																		else {
																			swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abgdfche
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adbgfche
																		}
																		else {
																			swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abdgfche
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // abfgdche
																		}
																		else {
																			swap(v2, v5); swap(v4, v6); swap(v6, v7); // abfdgche
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // adbfgche
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // abdfgche
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v3, v1)) {
																		swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adbgchfe
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abgdchfe
																		}
																		else {
																			swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abdgchfe
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // adbgcfhe
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v6); swap(v4, v6); swap(v6, v7); // abgdcfhe
																		}
																		else {
																			swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // abdgcfhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v7, v5)) {
																	if (compare(v2, v1)) {
																		swap(v1, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adcbghfe
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adbcghfe
																		}
																		else {
																			swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abdcghfe
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		swap(v1, v3); swap(v4, v6); swap(v6, v7); // adcbgfhe
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v6, v7); // adbcgfhe
																		}
																		else {
																			swap(v2, v3); swap(v4, v6); swap(v6, v7); // abdcgfhe
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v2, v1)) {
																		swap(v1, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adcbfghe
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adbfcghe
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adbcfghe
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abfdcghe
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abdfcghe
																		}
																	}
																	else {
																		swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abdcfghe
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v1)) {
											if (compare(v5, v0)) {
												if (compare(v6, v4)) {
													if (compare(v7, v1)) {
														if (compare(v2, v1)) {
															if (compare(v7, v2)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gfehadcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgehadcb
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfeahdcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fgeahdcb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gfaehdcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgaehdcb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fagehdcb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gfeadhcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgeadhcb
																			}
																		}
																		else {
																			if (compare(v6, v0)) {
																				if (compare(v6, v5)) {
																					swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); // gfaedhcb
																				}
																				else {
																					swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fgaedhcb
																				}
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); // fagedhcb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v5, v7); // gfadehcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v5, v7); swap(v6, v7); // fgadehcb
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v5, v7); // fagdehcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); // fadgehcb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfeadchb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fgeadchb
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gfaedchb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v6, v7); // fgaedchb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v5, v6); swap(v6, v7); // gfadechb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v6, v7); // fgadechb
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfadcehb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v6, v7); // fgadcehb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fagedchb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v5, v6); swap(v6, v7); // fagdechb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fadgechb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fagdcehb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fadgcehb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fadcgehb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfehadbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gfehabdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gfeahdbc
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfeadhbc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); // gfeahbdc
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fgehadbc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgehabdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); // fgeahdbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fgeadhbc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fgeahbdc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfaehdbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fgaehdbc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fagehdbc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gfaehbdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgaehbdc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fagehbdc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gfaedhbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); // fgaedhbc
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v5, v7); swap(v6, v7); // gfadehbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v5, v7); // fgadehbc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fagedhbc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v5, v7); swap(v6, v7); // fagdehbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fadgehbc
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v3, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gfeadbhc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgeadbhc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v6, v7); // gfaedbhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fgaedbhc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v6, v7); // fagedbhc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v6, v7); // gfadebhc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v5, v6); swap(v6, v7); // fgadebhc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v6, v7); // fagdebhc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v6, v7); // fadgebhc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gfeabhdc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgeabhdc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfaebhdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fgaebhdc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fagebhdc
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfeabdhc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fgeabdhc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gfaebdhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgaebdhc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fagebdhc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gfeadcbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // fgeadcbh
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // gfaedcbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); // fgaedcbh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v5, v6); // gfadecbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); // fgadecbh
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // gfadcebh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v5); // fgadcebh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // fagedcbh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v5, v6); // fagdecbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // fadgecbh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // fagdcebh
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fadgcebh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // fadcgebh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // gfeadbch
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fgeadbch
																			}
																		}
																		else {
																			if (compare(v6, v0)) {
																				if (compare(v6, v5)) {
																					swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); // gfaedbch
																				}
																				else {
																					swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // fgaedbch
																				}
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); // fagedbch
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); // gfadebch
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v5, v6); // fgadebch
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); // fagdebch
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); // fadgebch
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gfeabdch
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // fgeabdch
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // gfaebdch
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fgaebdch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // fagebdch
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v7, v1)) {
															if (compare(v6, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // feghadcb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); // fegahdcb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fegadhcb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fegadchb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // feghadbc
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // feghabdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fegahdbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fegadhbc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fegahbdc
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // feaghdcb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // feagdhcb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // feadghcb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // feagdchb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v6, v7); // feadgchb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // feadcghb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // feaghdbc
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // feaghbdc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // feagdhbc
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); // feadghbc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fegadcbh
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fegadbhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // fegadbch
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // feagdcbh
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // feagdbhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // feagdbch
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v6); // feadgcbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // feadcgbh
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // feadgbhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // feadgbch
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v5, v6); swap(v6, v7); // feadbghc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v5, v6); // feadbgch
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v4, v5); // feadcbgh
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); // feadbcgh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fegabhdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v6, v7); // fegabdhc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); // fegabdch
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); // feagbhdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // feagbdhc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // feagbdch
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // feabghdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // feabgdhc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // feabdghc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // feabgdch
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // feabdgch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); // feabdcgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v7, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // faeghdcb
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // faegdhcb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); // faedghcb
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fadeghcb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // faegdchb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // faedgchb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fadegchb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v5, v6); swap(v6, v7); // faedcghb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fadecghb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v5, v6); swap(v6, v7); // fadceghb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // faegdcbh
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // faedgcbh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v5, v6); // faedcgbh
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); // faedcbgh
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fadegcbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // fadecgbh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v5, v6); // fadcegbh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); // fadecbgh
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); // fadcebgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // faeghdbc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // faegdhbc
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // faegdbhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // faegdbch
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); // faeghbdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // faegbhdc
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // faegbdhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // faegbdch
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // faedghbc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v6, v7); // faedgbhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v6); // faedgbch
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fadeghbc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fadegbhc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // fadegbch
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // faedbghc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // faedbgch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v4, v5); // faedbcgh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fadebghc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fadebgch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fadebcgh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // faebghdc
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // faebgdhc
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // faebdghc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // faebgdch
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // faebdgch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // faebdcgh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v3)) {
													if (compare(v7, v1)) {
														if (compare(v4, v3)) {
															if (compare(v2, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gafehdcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gafedhcb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v6, v7); // gafedchb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agfehdcb
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // agfedhcb
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v6, v7); // agfedchb
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // afgehdcb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); // afgedhcb
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // afgedchb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afeghdcb
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // afegdhcb
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afegdchb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gafehdbc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); // gafedhbc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gafehbdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // agfehdbc
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); // agfedhbc
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agfehbdc
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afgehdbc
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // afgedhbc
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // afgehbdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // afeghdbc
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afegdhbc
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); // afeghbdc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v5, v7); swap(v6, v7); // gafdehcb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gadfehcb
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v5, v7); swap(v6, v7); // agfdehcb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v5, v7); // afgdehcb
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // agdfehcb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v5, v7); // gafdehbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); // gadfehbc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v5, v7); // agfdehbc
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v5, v7); swap(v6, v7); // afgdehbc
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); // agdfehbc
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v6, v7); // gafdechb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v6, v7); // agfdechb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v5, v6); swap(v6, v7); // afgdechb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v6, v7); // gafdcehb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v6, v7); // agfdcehb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afgdcehb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v6, v7); // gadfechb
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gadfcehb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v6, v7); // gadcfehb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v6, v7); // agdfechb
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // agdfcehb
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v6, v7); // agdcfehb
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v7, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gafedbhc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gafebhdc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gafebdhc
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // agfedbhc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // agfebhdc
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agfebdhc
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afgebhdc
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // afegbhdc
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afebghdc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v6, v7); // afgedbhc
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // afegdbhc
																			}
																		}
																		else {
																			if (compare(v6, v1)) {
																				if (compare(v6, v4)) {
																					swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // afgebdhc
																				}
																				else {
																					swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afegbdhc
																				}
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // afebgdhc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); // gafedcbh
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // gafedbch
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gafebdch
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v6); swap(v2, v5); swap(v3, v4); // agfedcbh
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // agfedbch
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // agfebdch
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // afgedcbh
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // afegdcbh
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); // afgedbch
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // afegdbch
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // afgebdch
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // afegbdch
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // afebgdch
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); // gafdecbh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); // agfdecbh
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v5, v6); // afgdecbh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v5); // gafdcebh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v4, v5); // agfdcebh
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // afgdcebh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); // gadfecbh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // gadfcebh
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v5); // gadcfebh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); // agdfecbh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // agdfcebh
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v4, v5); // agdcfebh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v5, v6); swap(v6, v7); // gafdebhc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gadfebhc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v5, v6); swap(v6, v7); // agfdebhc
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v6, v7); // afgdebhc
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // agdfebhc
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v5, v6); // gafdebch
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // gadfebch
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v5, v6); // agfdebch
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); // afgdebch
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v6); // agdfebch
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v2, v1)) {
															if (compare(v7, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); // afedghcb
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); // afdgehcb
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // afdeghcb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v4, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // afdgechb
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afdgcehb
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afedgchb
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afdegchb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v4); swap(v5, v6); swap(v6, v7); // afedcghb
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // afdecghb
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afdcgehb
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v5, v6); swap(v6, v7); // afdceghb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // afedgcbh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // afdgecbh
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // afdegcbh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v5, v6); // afedcgbh
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // afdecgbh
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v4); // afedcbgh
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); // afdecbgh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // afdgcebh
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // afdcgebh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v1, v5); swap(v2, v3); swap(v5, v6); // afdcegbh
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); // afdcebgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v3)) {
																		swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afedghbc
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // afdgehbc
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afdeghbc
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v6, v7); // afedgbhc
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v6, v7); // afdgebhc
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // afdegbhc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v4); swap(v4, v6); // afedgbch
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v6); // afdgebch
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); // afdegbch
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afedbghc
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // afedbgch
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v4, v5); // afedbcgh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afdebghc
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // afdebgch
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v5); // afdebcgh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afebdghc
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // afebdgch
																		}
																	}
																	else {
																		swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // afebdcgh
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v4, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); // adgfehcb
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // adfgehcb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adfeghcb
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // adgfehbc
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); // adfgehbc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // adfeghbc
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v5)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // adgfechb
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v6, v7); // adfgechb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // adfegchb
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adfecghb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adgfcehb
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // adfgcehb
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adfcgehb
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // adfceghb
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // adgcfehb
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // adcgfehb
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adcfgehb
																		}
																		else {
																			swap(v1, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // adcfeghb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v2, v1)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // adgfecbh
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v6, v7); // adgfebhc
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v5); // adgfebch
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v6); // adfgecbh
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // adfgebhc
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v6); // adfgebch
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // adgfcebh
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // adfgcebh
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // adfcgebh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // adgcfebh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v3); swap(v3, v6); swap(v4, v5); // adcgfebh
																			}
																			else {
																				swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // adcfgebh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // adfegcbh
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adfegbhc
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // adfegbch
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // adfecgbh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // adfcegbh
																			}
																			else {
																				swap(v1, v3); swap(v3, v5); swap(v5, v6); // adcfegbh
																			}
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // adfecbgh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v5); // adfcebgh
																			}
																			else {
																				swap(v1, v3); swap(v3, v5); // adcfebgh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // adfebghc
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v5, v6); // adfebgch
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v4); // adfebcgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v1)) {
												if (compare(v3, v1)) {
													if (compare(v5, v3)) {
														if (compare(v5, v0)) {
															if (compare(v6, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v6, v7); // gfadbehc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgadbehc
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v5); // gfadbech
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // fgadbech
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v6); // gfadcbeh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // fgadcbeh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // gfadbceh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v4, v6); // fgadbceh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v6, v7); // fagdbehc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fadgbehc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v5); // fagdbech
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // fadgbech
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v6); // fagdcbeh
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // fadgcbeh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v6); // fadcgbeh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // fagdbceh
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fadgbceh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gafdbehc
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agfdbehc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v6, v7); // afgdbehc
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // afdgbehc
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // gafdbech
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v4, v6); swap(v5, v6); // agfdbech
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v6); swap(v4, v5); // afgdbech
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); // afdgbech
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // gafdcbeh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // agfdcbeh
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v4, v6); // afgdcbeh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // afdgcbeh
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v4, v6); // afdcgbeh
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v4, v6); // gafdbceh
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v4, v6); // agfdbceh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v6); swap(v4, v5); swap(v5, v6); // afgdbceh
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // afdgbceh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v6, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gadfcbeh
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // gadcfbeh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gadfbehc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gadfbech
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // gadfbceh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gadbfehc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // gadbfech
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v6); // gadcbfeh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gadbfceh
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // gadbcfeh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // agdfcbeh
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // agdcfbeh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agdfbehc
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // agdfbech
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v6); // agdfbceh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // agdbfehc
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); // agdbfech
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v6); swap(v2, v3); swap(v4, v6); // agdcbfeh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // agdbfceh
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); // agdbcfeh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // adgfcbeh
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // adfgcbeh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // adgcfbeh
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // adgcbfeh
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v1)) {
																			swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // adcgfbeh
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v4, v6); // adcgbfeh
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // adfcgbeh
																		}
																		else {
																			swap(v1, v3); swap(v3, v5); swap(v4, v6); // adcfgbeh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // adgfbehc
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); // adgfbech
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // adgfbceh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adfgbehc
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // adfgbech
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v6); // adfgbceh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v6, v7); // adgbfehc
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v4, v5); // adgbfech
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v6); swap(v4, v5); swap(v5, v6); // adgbfceh
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v4, v6); // adgbcfeh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v0)) {
														if (compare(v7, v2)) {
															if (compare(v5, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gfabehdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v6, v7); // gfabedhc
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gfabdehc
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); // fgabehdc
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fgabedhc
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgabdehc
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); // gafbehdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gafbedhc
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gafbdehc
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gabfehdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v6, v7); // gabfedhc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gabfdehc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v6, v7); // gabdfehc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); // gfabedch
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // gfabdech
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gfabdceh
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // fgabedch
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fgabdech
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // fgabdceh
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // gafbedch
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gafbdech
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); // gabfedch
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // gabfdech
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v5); // gabdfech
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // gafbdceh
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gabfdceh
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // gabdfceh
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v6); // gabdcfeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v6, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); // agfbehdc
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // agfbedhc
																		}
																	}
																	else {
																		swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agfbdehc
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // agbfehdc
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v6, v7); // agbfedhc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // agbfdehc
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v6, v7); // agbdfehc
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fagbehdc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v6, v7); // fagbedhc
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fagbdehc
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // afgbehdc
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v6, v7); // afgbedhc
																		}
																	}
																	else {
																		swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // afgbdehc
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); // fagbedch
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // fagbdech
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fagbdceh
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v6); // agfbedch
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // agfbdech
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v6); // agfbdceh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v5); // afgbedch
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); // afgbdech
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // afgbdceh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v3)) {
																		swap(v1, v6); swap(v2, v6); swap(v3, v5); // agbfedch
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); // agbfdech
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v4, v5); // agbdfech
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // agbfdceh
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v4, v5); swap(v5, v6); // agbdfceh
																		}
																	}
																	else {
																		swap(v1, v6); swap(v2, v6); swap(v4, v6); // agbdcfeh
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v6, v4)) {
														if (compare(v3, v1)) {
															if (compare(v6, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fadbgehc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fadbgech
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // fadbgceh
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afdbgehc
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // afdbgech
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v6); // afdbgceh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adfbgehc
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v4, v6); swap(v5, v6); // adfbgech
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v4, v6); // adfbgceh
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // fadcbgeh
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // afdcbgeh
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // adfcbgeh
																		}
																		else {
																			swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // adcfbgeh
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fadbcgeh
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // afdbcgeh
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v4, v5); swap(v5, v6); // adfbcgeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); // fabgehdc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fabgedhc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fabgdehc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fabdgehc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // fabgedch
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fabgdech
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // fabdgech
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // fabgdceh
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v6); // fabdgceh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // fabdcgeh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); // afbgehdc
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // afbgedhc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afbgdehc
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afbdgehc
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v6); // afbgedch
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // afbgdech
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v4, v6); swap(v5, v6); // afbdgech
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v6); // afbgdceh
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v4, v6); // afbdgceh
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v4, v5); swap(v5, v6); // afbdcgeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v4, v2)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fadbeghc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // fadbegch
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); // fadbecgh
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // afdbeghc
																			}
																			else {
																				swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v5, v6); // afdbegch
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v3, v5); // afdbecgh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v5); swap(v5, v6); swap(v6, v7); // adfbeghc
																			}
																			else {
																				swap(v1, v3); swap(v2, v5); swap(v5, v6); // adfbegch
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); // adfbecgh
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v4, v5); // fadcbegh
																		}
																		else {
																			swap(v1, v5); swap(v2, v3); swap(v4, v5); // afdcbegh
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // adfcbegh
																		}
																		else {
																			swap(v1, v3); swap(v3, v5); swap(v4, v5); // adcfbegh
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		swap(v0, v5); swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fadbcegh
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v5); swap(v2, v3); swap(v3, v5); swap(v4, v5); // afdbcegh
																		}
																		else {
																			swap(v1, v3); swap(v2, v5); swap(v4, v5); // adfbcegh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fabeghdc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fabegdhc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fabedghc
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v5, v6); swap(v6, v7); // fabdeghc
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fabegdch
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // fabedgch
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v5, v6); // fabdegch
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); // fabedcgh
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); // fabdecgh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v4, v5); // fabdcegh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // afbeghdc
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afbegdhc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // afbedghc
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v5, v6); swap(v6, v7); // afbdeghc
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // afbegdch
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v5, v6); // afbedgch
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v5, v6); // afbdegch
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v4); // afbedcgh
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); // afbdecgh
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v4, v5); // afbdcegh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v2)) {
														if (compare(v7, v2)) {
															if (compare(v6, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // abgfehdc
																		}
																		else {
																			swap(v2, v6); swap(v3, v5); swap(v6, v7); // abgfedhc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // abgfdehc
																		}
																		else {
																			swap(v2, v6); swap(v4, v5); swap(v6, v7); // abgdfehc
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v4)) {
																			swap(v2, v5); swap(v3, v6); swap(v5, v7); // abfgehdc
																		}
																		else {
																			swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // abfeghdc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // abfgedhc
																			}
																			else {
																				swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abfegdhc
																			}
																		}
																		else {
																			swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abfgdehc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v4)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // adbgfehc
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adbfgehc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abfdgehc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // abdgfehc
																			}
																			else {
																				swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abdfgehc
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // abfedghc
																		}
																		else {
																			swap(v2, v5); swap(v5, v6); swap(v6, v7); // abfdeghc
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // adbfeghc
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // abdfeghc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v2, v6); swap(v3, v5); // abgfedch
																			}
																			else {
																				swap(v2, v5); swap(v3, v6); swap(v5, v6); // abfgedch
																			}
																		}
																		else {
																			swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // abfegdch
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v2, v5); swap(v3, v4); swap(v5, v6); // abfedgch
																		}
																		else {
																			swap(v2, v5); swap(v3, v4); // abfedcgh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v2, v6); swap(v3, v5); swap(v4, v5); // abgfdech
																		}
																		else {
																			swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // abfgdech
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v2, v5); swap(v4, v6); swap(v5, v6); // abfdgech
																			}
																			else {
																				swap(v2, v5); swap(v5, v6); // abfdegch
																			}
																		}
																		else {
																			swap(v2, v5); // abfdecgh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v5)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); // adbgfech
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // adbfgech
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v5, v6); // adbfegch
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); // adbfecgh
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			swap(v2, v6); swap(v4, v5); // abgdfech
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v2, v3); swap(v3, v6); swap(v4, v5); // abdgfech
																			}
																			else {
																				swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // abdfgech
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v2, v3); swap(v3, v5); swap(v5, v6); // abdfegch
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); // abdfecgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // abgfdceh
																		}
																		else {
																			swap(v2, v6); swap(v4, v5); swap(v5, v6); // abgdfceh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // adbgfceh
																		}
																		else {
																			swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // abdgfceh
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v2, v5); swap(v3, v6); swap(v4, v6); // abfgdceh
																		}
																		else {
																			swap(v2, v5); swap(v4, v6); // abfdgceh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v6); // adbfgceh
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); swap(v4, v6); // abdfgceh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v4)) {
																	if (compare(v3, v1)) {
																		swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // adbfcgeh
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v5); swap(v4, v5); swap(v5, v6); // abfdcgeh
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // abdfcgeh
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		swap(v1, v3); swap(v2, v3); swap(v3, v5); swap(v4, v5); // adbfcegh
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v5); swap(v4, v5); // abfdcegh
																		}
																		else {
																			swap(v2, v3); swap(v3, v5); swap(v4, v5); // abdfcegh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v3, v1)) {
																	if (compare(v2, v1)) {
																		swap(v1, v3); swap(v4, v6); // adcbgfeh
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); // adbgcfeh
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v4, v6); // adbcgfeh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v2, v6); swap(v4, v6); // abgdcfeh
																		}
																		else {
																			swap(v2, v3); swap(v3, v6); swap(v4, v6); // abdgcfeh
																		}
																	}
																	else {
																		swap(v2, v3); swap(v4, v6); // abdcgfeh
																	}
																}
															}
															else {
																if (compare(v6, v4)) {
																	if (compare(v2, v1)) {
																		swap(v1, v3); swap(v4, v5); swap(v5, v6); // adcbfgeh
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v4, v5); swap(v5, v6); // adbcfgeh
																		}
																		else {
																			swap(v2, v3); swap(v4, v5); swap(v5, v6); // abdcfgeh
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		swap(v1, v3); swap(v4, v5); // adcbfegh
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v3); swap(v2, v3); swap(v4, v5); // adbcfegh
																		}
																		else {
																			swap(v2, v3); swap(v4, v5); // abdcfegh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v7, v6)) {
									if (compare(v7, v4)) {
										if (compare(v7, v0)) {
											if (compare(v6, v3)) {
												if (compare(v6, v4)) {
													if (compare(v5, v2)) {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hgefadcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); // hgefadbc
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hgefabdc
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgeafdcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hgeafdbc
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgeafbdc
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hgaefdcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); // hgaefdbc
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hgaefbdc
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); // hagefdcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hagefdbc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hagefbdc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hgeadfcb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v6, v7); // hgaedfcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v6, v7); // hgadefcb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); // hagedfcb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); // hagdefcb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); // hgeadfbc
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); // hgaedfbc
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); // hgadefbc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v6, v7); // hagedfbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v6, v7); // hagdefbc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hgeadbfc
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); // hgaedbfc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v5, v6); // hgadebfc
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); // hgadbefc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hgeabfdc
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgeabdfc
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); // hgaebfdc
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hgaebdfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); // hgabefdc
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); // hgabedfc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hgabdefc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hagedbfc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v5, v7); swap(v6, v7); // hagdebfc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hagebfdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hagebdfc
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v6, v7); // hagbefdc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hagbedfc
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hagdbefc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hagbdefc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); // habgefdc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); // habgedfc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // habgdefc
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v2, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgeadcfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hgaedcfb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v5, v7); swap(v6, v7); // hgadecfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgadcefb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hgeadcbf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); // hgaedcbf
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v5, v7); // hgadecbf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); // hgadcebf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgadcbef
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgeadbcf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hgaedbcf
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v5, v6); swap(v6, v7); // hgadebcf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgadbecf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); // hgadbcef
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hgeabdcf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgaebdcf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hgabedcf
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgabdecf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hgabdcef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); // hagedcfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hagedcbf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); // hagedbcf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hagebdcf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v5, v6); // hagdecfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); // hagdcefb
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v5, v6); swap(v6, v7); // hagdecbf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hagdcebf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v5, v7); // hagdebcf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); // hagdbecf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); // hagbedcf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hagbdecf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // habgedcf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // habgdecf
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); // hagdcbef
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hagdbcef
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hagbdcef
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // habgdcef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v2, v1)) {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); // hegfadcb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hegafdcb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // heagfdcb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // haegfdcb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hefgadcb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hefagdcb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // heafgdcb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); // haefgdcb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); // hegadfcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hegadcfb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hegadcbf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // heagdfcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // heagdcfb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // heagdcbf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); // haegdfcb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // haegdcfb
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // haegdcbf
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hegfadbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hegafdbc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // heagfdbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // haegfdbc
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hefgadbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hefagdbc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); // heafgdbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // haefgdbc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hegadfbc
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hegadbfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hegadbcf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); // heagdfbc
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // heagdbfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // heagdbcf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // haegdfbc
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // haegdbfc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // haegdbcf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hegfabdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hegafbdc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); // heagfbdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // haegfbdc
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hefgabdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hefagbdc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // heafgbdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // haefgbdc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v6, v7); // hegabfdc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hegabdfc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); // hegabdcf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); // heagbfdc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); // heagbdfc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // heagbdcf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); // haegbfdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // haegbdfc
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // haegbdcf
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); // heabgfdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // heabgdfc
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // heabgdcf
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hefabgdc
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); // heafbgdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // heabfgdc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // haefbgdc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // haebgfdc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // haebfgdc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // haebgdfc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // haebgdcf
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); // habegfdc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // habefgdc
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // habegdfc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // habegdcf
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v2)) {
													if (compare(v6, v1)) {
														if (compare(v2, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hefadgcb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // heafdgcb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v6, v7); // headgfcb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // headfgcb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // haefdgcb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v6); // haedgfcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); // haedfgcb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); // hadgefcb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); // hadegfcb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hadefgcb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // headgcfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); // haedgcfb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); // hadgecfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hadegcfb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hadgcefb
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hadgecbf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hadgcebf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hadgcbef
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); // headgcbf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // haedgcbf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hadegcbf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hefadgbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // heafdgbc
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v6); // headgfbc
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); // headfgbc
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // haefdgbc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v6, v7); // haedgfbc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // haedfgbc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v6, v7); // hadgefbc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hadegfbc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hadefgbc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); // headgbfc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // haedgbfc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hadgebfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hadegbfc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hadgbefc
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); // hadgebcf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hadgbecf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hadgbcef
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // headgbcf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); // haedgbcf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hadegbcf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v4, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hefadbgc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); // heafdbgc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v5); // headfbgc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v5, v6); // headbgfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); // headbfgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v5, v6); swap(v6, v7); // headbgcf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // haefdbgc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); // haedfbgc
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); // haedbgfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v7); // haedbfgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // haedbgcf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hadebgfc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hadebgcf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hadefbgc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); // hadebfgc
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hadbgefc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); // hadbegfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); // hadbefgc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hadbgecf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hadbegcf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hadbgcef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); // hefabdgc
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); // heafbdgc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); // heabfdgc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); // heabdgfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); // heabdfgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // heabdgcf
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); // haefbdgc
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // haebfdgc
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // haebdgfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); // haebdfgc
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // haebdgcf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); // habefdgc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); // habedgfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); // habedfgc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); // habdgefc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v5, v6); // habdegfc
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); // habdefgc
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // habdgecf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); // habdgcef
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // habedgcf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v5, v6); swap(v6, v7); // habdegcf
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v6, v1)) {
															if (compare(v4, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); // hefadcgb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // heafdcgb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); // headfcgb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); // headcgfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v7); // headcfgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // headcgbf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); // haefdcgb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v5); // haedfcgb
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v5, v6); // haedcgfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); // haedcfgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v5, v6); swap(v6, v7); // haedcgbf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); // hadefcgb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); // hadecgfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); // hadecfgb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); // hadcgefb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v5, v6); // hadcegfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); // hadcefgb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hadcgebf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); // hadcgbef
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hadecgbf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v5, v6); swap(v6, v7); // hadcegbf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hefadcbg
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // heafdcbg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // headfcbg
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v6, v7); // headcfbg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // haefdcbg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v6, v7); // haedfcbg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hadefcbg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v6, v7); // haedcfbg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v6, v7); // hadecfbg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v6, v7); // hadcefbg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); // headcbgf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v5, v7); // haedcbgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); // hadecbgf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // headcbfg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v5, v7); swap(v6, v7); // haedcbfg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hadecbfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v5, v7); // hadcebgf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v5, v7); swap(v6, v7); // hadcebfg
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hadcbgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); // hadcbegf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hadcbefg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hefadbcg
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // heafdbcg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v6, v7); // headfbcg
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v6, v7); // headbfcg
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // haefdbcg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // haedfbcg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hadefbcg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v6, v7); // haedbfcg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hadebfcg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v6, v7); // hadbefcg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v5, v7); // headbcgf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); // haedbcgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hadebcgf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v5, v7); swap(v6, v7); // headbcfg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // haedbcfg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hadebcfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); // hadbecgf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hadbecfg
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hadbcgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hadbcegf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hadbcefg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hefabdcg
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v6, v7); // heafbdcg
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // heabfdcg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v6, v7); // heabdfcg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); // heabdcgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // heabdcfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // haefbdcg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // haebfdcg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // haebdfcg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // haebdcgf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // haebdcfg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // habefdcg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v6, v7); // habedfcg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); // habedcgf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // habedcfg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v6, v7); // habdefcg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v5, v7); // habdecgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v5, v7); swap(v6, v7); // habdecfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // habdcgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); // habdcegf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // habdcefg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v1)) {
												if (compare(v5, v1)) {
													if (compare(v6, v5)) {
														if (compare(v6, v4)) {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); // ahgefdcb
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); // ahgedfcb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v6); // ahgdefcb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); // ahdgefcb
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v6, v7); // adhgefcb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ahgefdbc
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v6, v7); // ahgedfbc
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahgefbdc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v6, v7); // ahgdefbc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v6, v7); // ahdgefbc
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v6); // adhgefbc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); // ahgedcfb
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v5, v6); // ahgdecfb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); // ahdgecfb
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // adhgecfb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); // ahgdcefb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ahdgcefb
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adhgcefb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); // ahdcgefb
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // adhcgefb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // adchgefb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ahegfdcb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); // ahegdfcb
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v4, v6); // ahedgfcb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); // ahdegfcb
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // adhegfcb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahegfdbc
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ahegdfbc
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ahegfbdc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v6, v7); // ahedgfbc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ahdegfbc
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); // adhegfbc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ahegdcfb
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); // ahedgcfb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ahdegcfb
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adhegcfb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v5, v6); // ahedcgfb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); // ahdecgfb
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // adhecgfb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v5, v6); // ahdcegfb
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); // adhcegfb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v7); swap(v5, v6); // adchegfb
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); // ahefgdcb
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ahefdgcb
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); // ahedfgcb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ahdefgcb
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adhefgcb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ahefgdbc
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahefdgbc
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahefgbdc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahedfgbc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahdefgbc
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // adhefgbc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); // ahefdcgb
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v4, v5); // ahedfcgb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); // ahdefcgb
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // adhefcgb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v4); // ahedcfgb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v4); // ahdecfgb
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); // adhecfgb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); // ahdcefgb
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v7); // adhcefgb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v7); // adchefgb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ahefdcbg
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v6, v7); // ahedfcbg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ahdefcbg
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adhefcbg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v6, v7); // ahedcfbg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v6, v7); // ahdecfbg
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // adhecfbg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v6, v7); // ahdcefbg
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v6, v7); // adhcefbg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v7); swap(v6, v7); // adchefbg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ahefdbgc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ahefbgdc
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); // ahefbdgc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); // ahedfbgc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ahdefbgc
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); // adhefbgc
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahefdbcg
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahedfbcg
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ahefbdcg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahdefbcg
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // adhefbcg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v6, v4)) {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ahgedcbf
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v5, v6); swap(v6, v7); // ahgdecbf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ahdgecbf
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); // adhgecbf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahgdcebf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahdgcebf
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // adhgcebf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahdcgebf
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adhcgebf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adchgebf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ahgedbfc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ahgebfdc
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahgebdfc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v5, v7); swap(v6, v7); // ahgdebfc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ahdgebfc
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); // adhgebfc
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); // ahgedbcf
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ahgebdcf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v5, v7); // ahgdebcf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v7); // ahdgebcf
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // adhgebcf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahegdcbf
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahedgcbf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahdegcbf
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // adhegcbf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v5, v6); swap(v6, v7); // ahedcgbf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ahdecgbf
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adhecgbf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v5, v6); swap(v6, v7); // ahdcegbf
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // adhcegbf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // adchegbf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahegdbfc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); // ahegbfdc
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ahegbdfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahedgbfc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahdegbfc
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // adhegbfc
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ahegdbcf
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); // ahedgbcf
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahegbdcf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ahdegbcf
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adhegbcf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v6, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); // ahedbgfc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ahdebgfc
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); // adhebgfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahedbgcf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahdebgcf
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // adhebgcf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ahebgfdc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahebgdfc
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ahebdgfc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ahebgdcf
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahebdgcf
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v5, v7); // ahedcbgf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); // ahdecbgf
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // adhecbgf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v5, v7); // ahdcebgf
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); // adhcebgf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v7); swap(v5, v7); // adchebgf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); // ahedbcgf
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ahebdcgf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ahdebcgf
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); // adhebcgf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v4, v7); // ahedbfgc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); // ahdebfgc
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v4); // adhebfgc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v6, v7); // ahedbfcg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ahdebfcg
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v6, v7); // adhebfcg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahebfgdc
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ahebfdgc
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); // ahebdfgc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahebfdcg
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ahebdfcg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v5, v7); swap(v6, v7); // ahedcbfg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ahdecbfg
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adhecbfg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v5, v7); swap(v6, v7); // ahdcebfg
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // adhcebfg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // adchebfg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahedbcfg
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahebdcfg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahdebcfg
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // adhebcfg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v6, v4)) {
														if (compare(v6, v1)) {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v3)) {
																		swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); // ahgdcbef
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ahdgcbef
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adhgcbef
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); // ahdcgbef
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // adhcgbef
																		}
																	}
																	else {
																		swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // adchgbef
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahgdbefc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahdgbefc
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // adhgbefc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); // ahgdbecf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ahdgbecf
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adhgbecf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahgdbcef
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahdgbcef
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // adhgbcef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ahdbgefc
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahdbgecf
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ahdbgcef
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); // adhbgefc
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adhbgecf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v4, v6); swap(v5, v7); // adhbgcef
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // adbhgefc
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adbhgecf
																		}
																	}
																	else {
																		swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // adbhgcef
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahdcbgef
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adhcbgef
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adchbgef
																		}
																		else {
																			swap(v1, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adcbhgef
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahdbcgef
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adhbcgef
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adbhcgef
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adbchgef
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v4, v2)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); // ahdbegfc
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ahdbegcf
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); // ahdbecgf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v7); swap(v5, v6); // adhbegfc
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v3); swap(v2, v7); swap(v5, v6); swap(v6, v7); // adhbegcf
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v5, v7); // adhbecgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); // adbhegfc
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // adbhegcf
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); // adbhecgf
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); // ahdcbegf
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // adhcbegf
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // adchbegf
																		}
																		else {
																			swap(v1, v3); swap(v4, v7); swap(v5, v7); // adcbhegf
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ahdbcegf
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); // adhbcegf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // adbhcegf
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); // adbchegf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v6, v2)) {
																				swap(v1, v7); swap(v2, v3); swap(v3, v7); // ahdbefgc
																			}
																			else {
																				swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v6, v7); // ahdbefcg
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ahdbecfg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v2)) {
																				swap(v1, v3); swap(v2, v7); // adhbefgc
																			}
																			else {
																				swap(v1, v3); swap(v2, v7); swap(v6, v7); // adhbefcg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v5, v7); swap(v6, v7); // adhbecfg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); // adbhefgc
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v6, v7); // adbhefcg
																		}
																	}
																	else {
																		swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // adbhecfg
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahdcbefg
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adhcbefg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adchbefg
																		}
																		else {
																			swap(v1, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adcbhefg
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v1, v7); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahdbcefg
																		}
																		else {
																			swap(v1, v3); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adhbcefg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adbhcefg
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adbchefg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v7, v1)) {
															if (compare(v6, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v1)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v6, v7); // ahgbefdc
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v6); // ahbgefdc
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); // ahbegfdc
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ahbefgdc
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ahgbedfc
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahgbdefc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); // ahbgedfc
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ahbegdfc
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ahbgdefc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); // ahbefdgc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); // ahbedgfc
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v3, v4); // ahbedfgc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ahbefdcg
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v6, v7); // ahbedfcg
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); // ahbdgefc
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v5, v6); // ahbdegfc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v7); // ahbdefgc
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v6, v7); // ahbdefcg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			swap(v2, v7); swap(v3, v6); // abhgefdc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v2, v7); swap(v3, v4); swap(v4, v6); // abhegfdc
																			}
																			else {
																				swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // abhefgdc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v2, v7); swap(v3, v4); swap(v4, v5); // abhefdgc
																		}
																		else {
																			swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // abhefdcg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v4)) {
																			swap(v2, v7); swap(v3, v6); swap(v5, v6); // abhgedfc
																		}
																		else {
																			swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // abhegdfc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v2, v7); swap(v3, v4); swap(v5, v6); // abhedgfc
																			}
																			else {
																				swap(v2, v7); swap(v3, v4); // abhedfgc
																			}
																		}
																		else {
																			swap(v2, v7); swap(v3, v4); swap(v6, v7); // abhedfcg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // abhgdefc
																		}
																		else {
																			swap(v2, v7); swap(v4, v6); swap(v5, v6); // abhdgefc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v2, v7); swap(v5, v6); // abhdegfc
																			}
																			else {
																				swap(v2, v7); // abhdefgc
																			}
																		}
																		else {
																			swap(v2, v7); swap(v6, v7); // abhdefcg
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // abdhgefc
																		}
																		else {
																			swap(v2, v3); swap(v3, v7); swap(v5, v6); // abdhegfc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v2, v3); swap(v3, v7); // abdhefgc
																		}
																		else {
																			swap(v2, v3); swap(v3, v7); swap(v6, v7); // abdhefcg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v7, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v1)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); // ahgbedcf
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ahbgedcf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahbegdcf
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ahbedgcf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v1)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ahgbdecf
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahbgdecf
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahbdgecf
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v5, v6); swap(v6, v7); // ahbdegcf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahgbdcef
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ahbgdcef
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v4, v6); swap(v5, v7); // ahbdgcef
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // abhgedcf
																		}
																		else {
																			swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abhegdcf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abhgdecf
																		}
																		else {
																			swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // abhgdcef
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v4)) {
																			if (compare(v4, v2)) {
																				swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abhdgecf
																			}
																			else {
																				swap(v2, v7); swap(v4, v6); swap(v5, v7); // abhdgcef
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // abhedgcf
																			}
																			else {
																				swap(v2, v7); swap(v5, v6); swap(v6, v7); // abhdegcf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v4)) {
																				swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abdhgecf
																			}
																			else {
																				swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // abdhegcf
																			}
																		}
																		else {
																			swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // abdhgcef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); // ahbedcgf
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v5, v7); // ahbdecgf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v2, v7); swap(v3, v4); swap(v5, v7); // abhedcgf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v2, v7); swap(v5, v7); // abhdecgf
																			}
																			else {
																				swap(v2, v3); swap(v3, v7); swap(v5, v7); // abdhecgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ahbedcfg
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v5, v7); swap(v6, v7); // ahbdecfg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // abhedcfg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v2, v7); swap(v5, v7); swap(v6, v7); // abhdecfg
																			}
																			else {
																				swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // abdhecfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahbdcgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); // ahbdcegf
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahbdcefg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abhdcgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v2, v7); swap(v4, v7); swap(v5, v7); // abhdcegf
																			}
																			else {
																				swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abhdcefg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v4)) {
																			swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abdhcgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // abdhcegf
																			}
																			else {
																				swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abdhcefg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abdchgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v2, v3); swap(v4, v7); swap(v5, v7); // abdchegf
																			}
																			else {
																				swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abdchefg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v0)) {
											if (compare(v5, v3)) {
												if (compare(v7, v5)) {
													if (compare(v3, v1)) {
														if (compare(v2, v1)) {
															if (compare(v6, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); // ehgfadcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); // ehgafdcb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ehagfdcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eahgfdcb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); // ehfgadcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ehfagdcb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); // ehafgdcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // eahfgdcb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); // ehfadgcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); // ehfadcgb
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v6, v7); // ehfadcbg
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ehafdgcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); // ehafdcgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ehafdcbg
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eahfdgcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // eahfdcgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eahfdcbg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v6, v7); // ehgfadbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ehgafdbc
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehagfdbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // eahgfdbc
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ehfgadbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehfagdbc
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ehafgdbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); // eahfgdbc
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v6, v1)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ehfadgbc
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); // ehfadbgc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ehfadbcg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehafdgbc
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ehafdbgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehafdbcg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // eahfdgbc
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); // eahfdbgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // eahfdbcg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ehgfabdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehgafbdc
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ehagfbdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); // eahgfbdc
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); // ehfgabdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ehfagbdc
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehafgbdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // eahfgbdc
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehfabgdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ehfabdgc
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehfabdcg
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ehafbgdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); // ehafbdgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ehafbdcg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); // eahfbgdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); // eahfbdgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v6, v7); // eahfbdcg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ehgabfdc
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); // ehagbfdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ehabgfdc
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehabfgdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ehabfdgc
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehabfdcg
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); // eahgbfdc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); // eahbgfdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // eahbfgdc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); // eahbfdgc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // eahbfdcg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // eabhgfdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eabhfgdc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // eabhfdgc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eabhfdcg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v2)) {
														if (compare(v6, v1)) {
															if (compare(v6, v3)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // efhgadcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); // efhgadbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); // efhgabdc
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efhagdcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // efhagdbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); // efhagbdc
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // efahgdcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efahgdbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // efahgbdc
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eafhgdcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // eafhgdbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eafhgbdc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // efhadgcb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efahdgcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efadhgcb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // eafhdgcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); // eafdhgcb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); // efhadgbc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // efahdgbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); // efadhgbc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eafhdgbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eafdhgbc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); // efhadbgc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); // efahdbgc
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v7); // efadhbgc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); // efadbhgc
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // efhabgdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); // efhabdgc
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efahbgdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // efahbdgc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // efabhgdc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); // efabhdgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // efabdhgc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // eafhdbgc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); // eafdhbgc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // eafhbgdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); // eafhbdgc
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eafbhgdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // eafbhdgc
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v5, v7); // eafdbhgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); // eafbdhgc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // eabfhgdc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); // eabfhdgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // eabfdhgc
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v2, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); // efhadcgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // efahdcgb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); // efadhcgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v5, v7); // efadchgb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // efhadcbg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efahdcbg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efadhcbg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v5, v7); swap(v6, v7); // efadchbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v6, v7); // efadcbhg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v6, v7); // efhadbcg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // efahdbcg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v6, v7); // efadhbcg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efadbhcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v6, v7); // efadbchg
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // efhabdcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efahbdcg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // efabhdcg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efabdhcg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // efabdchg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); // eafhdcgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // eafhdcbg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eafhdbcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // eafhbdcg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v7); // eafdhcgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); // eafdchgb
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v6, v7); // eafdhcbg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eafdchbg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eafdhbcg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v5, v7); swap(v6, v7); // eafdbhcg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eafbhdcg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // eafbdhcg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // eabfhdcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eabfdhcg
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v6, v7); // eafdcbhg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v6, v7); // eafdbchg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v6, v7); // eafbdchg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // eabfdchg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v1)) {
														if (compare(v2, v1)) {
															if (compare(v6, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); // ehgadfcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); // ehagdfcb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v6); // ehadgfcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); // ehadfgcb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // eahgdfcb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v6, v7); // eahdgfcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eahdfgcb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // eadhgfcb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eadhfgcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // eadfhgcb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v5); // ehadfcgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); // eahdfcgb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // eadhfcgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); // eadfhcgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // eadfchgb
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v6, v7); // ehadfcbg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eahdfcbg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eadhfcbg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // eadfhcbg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eadfchbg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // eadfcbhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v6, v7); // ehgadfbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ehagdfbc
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v6, v7); // ehadgfbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehadfgbc
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); // eahgdfbc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v6); // eahdgfbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); // eahdfgbc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); // eadhgfbc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // eadhfgbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eadfhgbc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); // ehadfbgc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v5); // eahdfbgc
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); // eadhfbgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // eadfhbgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); // eadfbhgc
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehadfbcg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v6, v7); // eahdfbcg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // eadhfbcg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eadfhbcg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // eadfbhcg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v6, v7); // eadfbchg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v7, v3)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ehgadbfc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehagdbfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehadgbfc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); // ehadbgfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v7); // ehadbfgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v6, v7); // ehadbfcg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // eahgdbfc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); // eahdgbfc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v5, v6); // eahdbgfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); // eahdbfgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v6, v7); // eahdbfcg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // eadhgbfc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); // eadhbgfc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); // eadhbfgc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v6, v7); // eadhbfcg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // eadbhgfc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); // eadbhfgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // eadbfhgc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // eadbhfcg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eadbfhcg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // eadbfchg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehgabdfc
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ehagbdfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehabgdfc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ehabdgfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); // ehabdfgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ehabdfcg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); // eahgbdfc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // eahbgdfc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); // eahbdgfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); // eahbdfgc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v6, v7); // eahbdfcg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eabhgdfc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // eabhdgfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); // eabhdfgc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); // eabdhgfc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v7); // eabdhfgc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); // eabdfhgc
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // eabhdfcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v6, v7); // eabdhfcg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eabdfhcg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v6, v7); // eabdfchg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v5, v1)) {
															if (compare(v7, v3)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); // ehgadcfb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ehagdcfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); // ehadgcfb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v5, v6); // ehadcgfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); // ehadcfgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v6, v7); // ehadcfbg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eahgdcfb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eahdgcfb
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); // eahdcgfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v7); // eahdcfgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v6, v7); // eahdcfbg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eadhgcfb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // eadhcgfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); // eadhcfgb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); // eadchgfb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v7); // eadchfgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); // eadcfhgb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // eadhcfbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v6, v7); // eadchfbg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eadcfhbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v6, v7); // eadcfbhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ehgadcbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehagdcbf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehadgcbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v5, v6); swap(v6, v7); // ehadcgbf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // eahgdcbf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); // eahdgcbf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // eadhgcbf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eahdcgbf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eadhcgbf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eadchgbf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v5, v7); // ehadcbgf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); // eahdcbgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // eadhcbgf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v5, v7); swap(v6, v7); // ehadcbfg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eahdcbfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eadhcbfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); // eadchbgf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eadchbfg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v5, v7); // eadcbhgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v5, v7); swap(v6, v7); // eadcbhfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v6, v7); // eadcbfhg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v6, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); // ehgadbcf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ehagdbcf
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v7); // ehadgbcf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehadbgcf
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eahgdbcf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eahdgbcf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eadhgbcf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v5, v6); swap(v6, v7); // eahdbgcf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // eadhbgcf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eadbhgcf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); // ehadbcgf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v5, v7); // eahdbcgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); // eadhbcgf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehadbcfg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v5, v7); swap(v6, v7); // eahdbcfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // eadhbcfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // eadbhcgf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eadbhcfg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); // eadbchgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // eadbchfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v6, v7); // eadbcfhg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ehgabdcf
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehagbdcf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ehabgdcf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehabdgcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ehabdcgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehabdcfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // eahgbdcf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eahbgdcf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // eahbdgcf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); // eahbdcgf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // eahbdcfg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // eabhgdcf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eabhdgcf
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // eabhdcgf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eabhdcfg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eabdhgcf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); // eabdhcgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eabdhcfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v5, v7); // eabdchgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v5, v7); swap(v6, v7); // eabdchfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v6, v7); // eabdcfhg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v1)) {
												if (compare(v6, v1)) {
													if (compare(v6, v5)) {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aehgfdcb
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // aehgdfcb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v6, v7); // aehdgfcb
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // aedhgfcb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); // adehgfcb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // aehgfdbc
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); // aehgdfbc
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); // aehgfbdc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v4, v6); // aehdgfbc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); // aedhgfbc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // adehgfbc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aehgdcfb
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aehdgcfb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aedhgcfb
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // adehgcfb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); // aehdcgfb
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // aedhcgfb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); // adehcgfb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); // aedchgfb
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // adechgfb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // adcehgfb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // aehgdcbf
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); // aehdgcbf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // aedhgcbf
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adehgcbf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aehdcgbf
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aedhcgbf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // adehcgbf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aedchgbf
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adechgbf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adcehgbf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // aehgdbfc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v6); // aehgbfdc
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); // aehgbdfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); // aehdgbfc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // aedhgbfc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adehgbfc
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aehgdbcf
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // aehgbdcf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aehdgbcf
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aedhgbcf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // adehgbcf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // aehfgdcb
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aehfdgcb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aehdfgcb
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aedhfgcb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // adehfgcb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); // aehfgdbc
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // aehfdgbc
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // aehfgbdc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); // aehdfgbc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // aedhfgbc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adehfgbc
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // aehfdcgb
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); // aehdfcgb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // aedhfcgb
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); // adehfcgb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v4, v7); // aehdcfgb
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); // aedhcfgb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); // adehcfgb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v7); // aedchfgb
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); // adechfgb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v7); // adcehfgb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aefhgdcb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // aefhdgcb
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); // aefdhgcb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // aedfhgcb
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adefhgcb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // aefhgdbc
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aefhdgbc
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aefhgbdc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aefdhgbc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aedfhgbc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // adefhgbc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); // aefhdcgb
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v4, v7); // aefdhcgb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); // aedfhcgb
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // adefhcgb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); // aefdchgb
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // aedfchgb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); // adefchgb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); // aedcfhgb
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // adecfhgb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // adcefhgb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v7, v5)) {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aehfdcbg
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aehdfcbg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aedhfcbg
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // adehfcbg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v6, v7); // aehdcfbg
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // aedhcfbg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v6, v7); // adehcfbg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v6, v7); // aedchfbg
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // adechfbg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // adcehfbg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); // aehfdbgc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); // aehfbgdc
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v5); // aehfbdgc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v4, v5); // aehdfbgc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); // aedhfbgc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // adehfbgc
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // aehfdbcg
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v6, v7); // aehfbdcg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v6, v7); // aehdfbcg
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // aedhfbcg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adehfbcg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // aefhdcbg
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v6, v7); // aefdhcbg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // aedfhcbg
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adefhcbg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aefdchbg
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aedfchbg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // adefchbg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aedcfhbg
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adecfhbg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adcefhbg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // aefhdbgc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // aefhbgdc
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); // aefhbdgc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); // aefdhbgc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // aedfhbgc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); // adefhbgc
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aefhdbcg
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aefdhbcg
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // aefhbdcg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aedfhbcg
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // adefhbcg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v6, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v5, v6); // aehdbgfc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); // aedhbgfc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // adehbgfc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v5, v6); swap(v6, v7); // aehdbgcf
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // aedhbgcf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adehbgcf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); // aehbgfdc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // aehbgdfc
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); // aehbdgfc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aehbgdcf
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // aehbdgcf
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); // aehdcbgf
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // aedhcbgf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); // adehcbgf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); // aedchbgf
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // adechbgf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // adcehbgf
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v7); swap(v5, v7); // aehdbcgf
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); // aehbdcgf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); // aedhbcgf
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // adehbcgf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); // aehdbfgc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); // aedhbfgc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); // adehbfgc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v6, v7); // aehdbfcg
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v6, v7); // aedhbfcg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // adehbfcg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // aehbfgdc
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); // aehbfdgc
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v4); // aehbdfgc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // aehbfdcg
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v6, v7); // aehbdfcg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aehdcbfg
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aedhcbfg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // adehcbfg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aedchbfg
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adechbfg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adcehbfg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v7); swap(v5, v7); swap(v6, v7); // aehdbcfg
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // aehbdcfg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // aedhbcfg
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adehbcfg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v7, v5)) {
														if (compare(v6, v5)) {
															if (compare(v7, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // aedbhgfc
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aedbhgcf
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // aedbhcgf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); // adebhgfc
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adebhgcf
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); // adebhcgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // adbehgfc
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adbehgcf
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // adbehcgf
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v5, v7); // aedcbhgf
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); // adecbhgf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v1, v3); swap(v3, v4); swap(v5, v7); // adcebhgf
																		}
																		else {
																			swap(v1, v3); swap(v5, v7); // adcbehgf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); // aedbchgf
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v5, v7); // adebchgf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); // adbechgf
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v5, v7); // adbcehgf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v6, v2)) {
																				swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); // aedbhfgc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // aedbhfcg
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aedbhcfg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v2)) {
																				swap(v1, v3); swap(v2, v4); swap(v4, v7); // adebhfgc
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v6, v7); // adebhfcg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adebhcfg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); // adbehfgc
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // adbehfcg
																		}
																	}
																	else {
																		swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adbehcfg
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v5, v7); swap(v6, v7); // aedcbhfg
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // adecbhfg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v1, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // adcebhfg
																		}
																		else {
																			swap(v1, v3); swap(v5, v7); swap(v6, v7); // adcbehfg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // aedbchfg
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v5, v7); swap(v6, v7); // adebchfg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // adbechfg
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v5, v7); swap(v6, v7); // adbcehfg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v6, v7); // aefdcbhg
																		}
																		else {
																			swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // aedfcbhg
																		}
																	}
																	else {
																		swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v6, v7); // adefcbhg
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v6, v7); // aedcfbhg
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // adecfbhg
																		}
																	}
																	else {
																		swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // adcefbhg
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v5); swap(v5, v7); // aefdbhgc
																		}
																		else {
																			swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); // aedfbhgc
																		}
																	}
																	else {
																		swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // adefbhgc
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v5, v7); swap(v6, v7); // aefdbhcg
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // aedfbhcg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adefbhcg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v6, v7); // aefdbchg
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v6, v7); // aedfbchg
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // adefbchg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // aedbfhgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aedbfhcg
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // aedbfchg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); // adebfhgc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adebfhcg
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v6, v7); // adebfchg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // adbefhgc
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adbefhcg
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // adbefchg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v6, v7); // aedcbfhg
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v6, v7); // adecbfhg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v1, v3); swap(v3, v4); swap(v6, v7); // adcebfhg
																		}
																		else {
																			swap(v1, v3); swap(v6, v7); // adcbefhg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v6, v7); // aedbcfhg
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v6, v7); // adebcfhg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v6, v7); // adbecfhg
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v6, v7); // adbcefhg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v2)) {
														if (compare(v4, v1)) {
															if (compare(v5, v3)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // aebhgfdc
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aebhfgdc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aefbhgdc
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // aebfhgdc
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // aefbhdgc
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); // aefbdhgc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // aebhfdgc
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); // aebfhdgc
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // aebfdhgc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aebhgdfc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // aebhdgfc
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); // aebhdfgc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // aebhgdcf
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aebhdgcf
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); // aebdhgfc
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aebdhgcf
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v1, v4); swap(v2, v4); swap(v4, v7); // aebdhfgc
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); // aebdfhgc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // abehgfdc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abehfgdc
																			}
																			else {
																				swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // abefhgdc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // abehfdgc
																		}
																		else {
																			swap(v2, v4); swap(v3, v5); swap(v4, v7); // abefhdgc
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abehgdfc
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // abehdgfc
																			}
																			else {
																				swap(v2, v4); swap(v3, v7); swap(v4, v7); // abehdfgc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // abehgdcf
																		}
																		else {
																			swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abehdgcf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v2, v4); swap(v4, v7); swap(v5, v6); // abedhgfc
																		}
																		else {
																			swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abedhgcf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // abefdhgc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v4); swap(v4, v7); // abedhfgc
																			}
																			else {
																				swap(v2, v4); swap(v4, v5); swap(v5, v7); // abedfhgc
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // abdehgfc
																		}
																		else {
																			swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abdehgcf
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v2, v3); swap(v3, v4); swap(v4, v7); // abdehfgc
																		}
																		else {
																			swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // abdefhgc
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aefbhdcg
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // aefbdhcg
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v6, v7); // aefbdchg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aebhfdcg
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // aebfhdcg
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aebfdhcg
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // aebfdchg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // aebhdfcg
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v6, v7); // aebdhfcg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aebdfhcg
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v6, v7); // aebdfchg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abehfdcg
																		}
																		else {
																			swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // abefhdcg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abefdhcg
																		}
																		else {
																			swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // abefdchg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			if (compare(v7, v3)) {
																				swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // abehdfcg
																			}
																			else {
																				swap(v2, v4); swap(v4, v7); swap(v6, v7); // abedhfcg
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abedfhcg
																			}
																			else {
																				swap(v2, v4); swap(v4, v5); swap(v6, v7); // abedfchg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // abdehfcg
																			}
																			else {
																				swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abdefhcg
																			}
																		}
																		else {
																			swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // abdefchg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // aebhdcgf
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); // aebdhcgf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // abehdcgf
																			}
																			else {
																				swap(v2, v4); swap(v4, v7); swap(v5, v7); // abedhcgf
																			}
																		}
																		else {
																			swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // abdehcgf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aebhdcfg
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aebdhcfg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abehdcfg
																			}
																			else {
																				swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abedhcfg
																			}
																		}
																		else {
																			swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abdehcfg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v1, v4); swap(v2, v4); swap(v5, v7); // aebdchgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v5, v7); swap(v6, v7); // aebdchfg
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v6, v7); // aebdcfhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v2, v4); swap(v5, v7); // abedchgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v4); swap(v5, v7); swap(v6, v7); // abedchfg
																			}
																			else {
																				swap(v2, v4); swap(v6, v7); // abedcfhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v2, v3); swap(v3, v4); swap(v5, v7); // abdechgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // abdechfg
																			}
																			else {
																				swap(v2, v3); swap(v3, v4); swap(v6, v7); // abdecfhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v2, v3); swap(v5, v7); // abdcehgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v3); swap(v5, v7); swap(v6, v7); // abdcehfg
																			}
																			else {
																				swap(v2, v3); swap(v6, v7); // abdcefhg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v4)) {
										if (compare(v6, v0)) {
											if (compare(v7, v3)) {
												if (compare(v7, v4)) {
													if (compare(v5, v2)) {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); // ghefadcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ghefadbc
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghefabdc
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gheafdcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gheafdbc
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gheafbdc
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); // ghaefdcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ghaefdbc
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghaefbdc
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gahefdcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); // gahefdbc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gahefbdc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); // gheadfcb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); // ghaedfcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); // ghadefcb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v6, v7); // gahedfcb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v6, v7); // gahdefcb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gheadfbc
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v6, v7); // ghaedfbc
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v6, v7); // ghadefbc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); // gahedfbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); // gahdefbc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gheadbfc
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ghaedbfc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v5, v7); swap(v6, v7); // ghadebfc
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghadbefc
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); // gheabfdc
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gheabdfc
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ghaebfdc
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghaebdfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v6, v7); // ghabefdc
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ghabedfc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghabdefc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); // gahedbfc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v5, v6); // gahdebfc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); // gahebfdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gahebdfc
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); // gahbefdc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); // gahbedfc
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); // gahdbefc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gahbdefc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v6, v7); // gabhefdc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gabhedfc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gabhdefc
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v0)) {
															if (compare(v2, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gheadcfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); // ghaedcfb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v5, v6); // ghadecfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); // ghadcefb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gheadcbf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ghaedcbf
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v5, v6); swap(v6, v7); // ghadecbf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghadcebf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); // ghadcbef
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gheadbcf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v7); // ghaedbcf
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v5, v7); // ghadebcf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v7); // ghadbecf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghadbcef
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gheabdcf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ghaebdcf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v7); // ghabedcf
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ghabdecf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghabdcef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gahedcfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); // gahedcbf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gahedbcf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gahebdcf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v5, v7); swap(v6, v7); // gahdecfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gahdcefb
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v5, v7); // gahdecbf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); // gahdcebf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v5, v6); swap(v6, v7); // gahdebcf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gahdbecf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gahbedcf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gahbdecf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); // gabhedcf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gabhdecf
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gahdcbef
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); // gahdbcef
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gahbdcef
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gabhdcef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v2, v1)) {
															if (compare(v5, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gehfadcb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gehafdcb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // geahfdcb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gaehfdcb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gefhadcb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gefahdcb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); // geafhdcb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gaefhdcb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gehadfcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gehadcfb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gehadcbf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); // geahdfcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // geahdcfb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // geahdcbf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gaehdfcb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gaehdcfb
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gaehdcbf
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); // gehfadbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gehafdbc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // geahfdbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gaehfdbc
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gefhadbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gefahdbc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // geafhdbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); // gaefhdbc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); // gehadfbc
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gehadbfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gehadbcf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // geahdfbc
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // geahdbfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // geahdbcf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); // gaehdfbc
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gaehdbfc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gaehdbcf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gehfabdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); // gehafbdc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // geahfbdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gaehfbdc
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gefhabdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gefahbdc
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // geafhbdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gaefhbdc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); // gehabfdc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); // gehabdfc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gehabdcf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v6, v7); // geahbfdc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // geahbdfc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); // geahbdcf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gaehbfdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gaehbdfc
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gaehbdcf
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // geabhfdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // geabhdfc
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // geabhdcf
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); // gefabhdc
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // geafbhdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // geabfhdc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gaefbhdc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); // gaebhfdc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gaebfhdc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gaebhdfc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gaebhdcf
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gabehfdc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gabefhdc
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gabehdfc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gabehdcf
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v2)) {
													if (compare(v7, v1)) {
														if (compare(v2, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gefadhcb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // geafdhcb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v7); // geadhfcb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); // geadfhcb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gaefdhcb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v6, v7); // gaedhfcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gaedfhcb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v6, v7); // gadhefcb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gadehfcb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gadefhcb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); // geadhcfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gaedhcfb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gadhecfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gadehcfb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gadhcefb
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); // gadhecbf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gadhcebf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gadhcbef
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // geadhcbf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); // gaedhcbf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gadehcbf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gefadhbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // geafdhbc
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v6, v7); // geadhfbc
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // geadfhbc
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gaefdhbc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v7); // gaedhfbc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); // gaedfhbc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); // gadhefbc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); // gadehfbc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gadefhbc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // geadhbfc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); // gaedhbfc
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); // gadhebfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gadehbfc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gadhbefc
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gadhebcf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gadhbecf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gadhbcef
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); // geadhbcf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gaedhbcf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gadehbcf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v4, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gefadbhc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // geafdbhc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v6, v7); // geadfbhc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v5, v7); swap(v6, v7); // geadbhfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v6, v7); // geadbfhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v5, v7); // geadbhcf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gaefdbhc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gaedfbhc
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gaedbhfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v6, v7); // gaedbfhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); // gaedbhcf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gadebhfc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gadebhcf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gadefbhc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gadebfhc
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gadbhefc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gadbehfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v6, v7); // gadbefhc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gadbhecf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); // gadbehcf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gadbhcef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gefabdhc
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v6, v7); // geafbdhc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // geabfdhc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // geabdhfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v6, v7); // geabdfhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); // geabdhcf
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gaefbdhc
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gaebfdhc
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gaebdhfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gaebdfhc
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gaebdhcf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gabefdhc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gabedhfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v6, v7); // gabedfhc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gabdhefc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v5, v7); swap(v6, v7); // gabdehfc
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v6, v7); // gabdefhc
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); // gabdhecf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gabdhcef
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); // gabedhcf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v5, v7); // gabdehcf
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v7, v1)) {
															if (compare(v4, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gefadchb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // geafdchb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // geadfchb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // geadchfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v6, v7); // geadcfhb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); // geadchbf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gaefdchb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v6, v7); // gaedfchb
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v5, v7); swap(v6, v7); // gaedchfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v6, v7); // gaedcfhb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v5, v7); // gaedchbf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gadefchb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gadechfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v6, v7); // gadecfhb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gadchefb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v5, v7); swap(v6, v7); // gadcehfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v6, v7); // gadcefhb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); // gadchebf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gadchbef
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); // gadechbf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v5, v7); // gadcehbf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // gefadcbh
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // geafdcbh
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // geadfcbh
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v6); // geadcfbh
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // gaefdcbh
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v5); // gaedfcbh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // gadefcbh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); // gaedcfbh
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); // gadecfbh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); // gadcefbh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // geadcbhf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v5, v6); swap(v6, v7); // gaedcbhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gadecbhf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // geadcbfh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v5, v6); // gaedcbfh
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // gadecbfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v5, v6); swap(v6, v7); // gadcebhf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v5, v6); // gadcebfh
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); // gadcbhef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gadcbehf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // gadcbefh
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gefadbch
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // geafdbch
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v4, v5); // geadfbch
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); // geadbfch
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gaefdbch
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // gaedfbch
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gadefbch
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v6); // gaedbfch
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // gadebfch
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); // gadbefch
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v5, v6); swap(v6, v7); // geadbchf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gaedbchf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gadebchf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v5, v6); // geadbcfh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // gaedbcfh
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gadebcfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gadbechf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // gadbecfh
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gadbchef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gadbcehf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gadbcefh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // gefabdch
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); // geafbdch
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // geabfdch
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); // geabdfch
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // geabdchf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // geabdcfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // gaefbdch
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gaebfdch
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // gaebdfch
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gaebdchf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gaebdcfh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // gabefdch
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); // gabedfch
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gabedchf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // gabedcfh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); // gabdefch
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v5, v6); swap(v6, v7); // gabdechf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v5, v6); // gabdecfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); // gabdchef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gabdcehf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // gabdcefh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v1)) {
												if (compare(v5, v1)) {
													if (compare(v7, v5)) {
														if (compare(v7, v4)) {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // aghefdcb
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v6, v7); // aghedfcb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v6, v7); // aghdefcb
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v6, v7); // agdhefcb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v7); // adghefcb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); // aghefdbc
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v4); // aghedfbc
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // aghefbdc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v7); // aghdefbc
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); // agdhefbc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v6, v7); // adghefbc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // aghedcfb
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v5, v7); swap(v6, v7); // aghdecfb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // agdhecfb
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); // adghecfb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aghdcefb
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agdhcefb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // adghcefb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agdchefb
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adgchefb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adcghefb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agehfdcb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // agehdfcb
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v6, v7); // agedhfcb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // agdehfcb
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); // adgehfcb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // agehfdbc
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); // agehdfbc
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agehfbdc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v4, v7); // agedhfbc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); // agdehfbc
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // adgehfbc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agehdcfb
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agedhcfb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agdehcfb
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // adgehcfb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v5, v7); swap(v6, v7); // agedchfb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // agdechfb
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adgechfb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v5, v7); swap(v6, v7); // agdcehfb
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // adgcehfb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // adcgehfb
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // agefhdcb
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agefdhcb
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agedfhcb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agdefhcb
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // adgefhcb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); // agefhdbc
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // agefdhbc
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // agefhbdc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); // agedfhbc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // agdefhbc
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adgefhbc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // agefdchb
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v6, v7); // agedfchb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // agdefchb
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adgefchb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v6, v7); // agedcfhb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v6, v7); // agdecfhb
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // adgecfhb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v6, v7); // agdcefhb
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v6, v7); // adgcefhb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v6, v7); // adcgefhb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // agefdcbh
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v4, v5); // agedfcbh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // agdefcbh
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // adgefcbh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v4); // agedcfbh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v4); // agdecfbh
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // adgecfbh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); // agdcefbh
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v6); // adgcefbh
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); // adcgefbh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agefdbhc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agefbhdc
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // agefbdhc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agedfbhc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agdefbhc
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // adgefbhc
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // agefdbch
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // agedfbch
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // agefbdch
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // agdefbch
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); // adgefbch
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v7, v4)) {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); // aghedcbf
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v5, v7); // aghdecbf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); // agdhecbf
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // adghecbf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); // aghdcebf
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // agdhcebf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adghcebf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); // agdchebf
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // adgchebf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // adcghebf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); // aghedbfc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); // aghebfdc
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // aghebdfc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v5, v6); // aghdebfc
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); // agdhebfc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // adghebfc
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // aghedbcf
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aghebdcf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v5, v6); swap(v6, v7); // aghdebcf
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // agdhebcf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v7); // adghebcf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // agehdcbf
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); // agedhcbf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // agdehcbf
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adgehcbf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v5, v7); // agedchbf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); // agdechbf
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // adgechbf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v5, v7); // agdcehbf
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); // adgcehbf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v5, v7); // adcgehbf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // agehdbfc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // agehbfdc
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agehbdfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); // agedhbfc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // agdehbfc
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adgehbfc
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agehdbcf
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agedhbcf
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // agehbdcf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agdehbcf
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // adgehbcf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v7, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agedbhfc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agdebhfc
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // adgebhfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); // agedbhcf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // agdebhcf
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v7); // adgebhcf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); // agebhfdc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // agebhdfc
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agebdhfc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agebhdcf
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // agebdhcf
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v5, v6); swap(v6, v7); // agedcbhf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // agdecbhf
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adgecbhf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v5, v6); swap(v6, v7); // agdcebhf
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // adgcebhf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // adcgebhf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agedbchf
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agebdchf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agdebchf
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // adgebchf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v6, v7); // agedbfhc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // agdebfhc
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v6, v7); // adgebfhc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v4, v6); // agedbfch
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); // agdebfch
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v4); // adgebfch
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // agebfhdc
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agebfdhc
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // agebdfhc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // agebfdch
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // agebdfch
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v5, v6); // agedcbfh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // agdecbfh
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // adgecbfh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v6); swap(v2, v3); swap(v5, v6); // agdcebfh
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // adgcebfh
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v6); swap(v5, v6); // adcgebfh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // agedbcfh
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // agebdcfh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // agdebcfh
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v5, v6); // adgebcfh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v7, v4)) {
														if (compare(v7, v1)) {
															if (compare(v2, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aghdcbef
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agdhcbef
																		}
																	}
																	else {
																		swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // adghcbef
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agdchbef
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adgchbef
																		}
																	}
																	else {
																		swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adcghbef
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); // aghdbefc
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // agdhbefc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adghbefc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aghdbecf
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agdhbecf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // adghbecf
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v7); swap(v4, v6); swap(v5, v7); // aghdbcef
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // agdhbcef
																		}
																	}
																	else {
																		swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adghbcef
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agdbhefc
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // agdbhecf
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agdbhcef
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adgbhefc
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v7); // adgbhecf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adgbhcef
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adbghefc
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // adbghecf
																		}
																	}
																	else {
																		swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adbghcef
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); // agdcbhef
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // adgcbhef
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // adcgbhef
																		}
																		else {
																			swap(v1, v3); swap(v4, v6); swap(v5, v7); // adcbghef
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // agdbchef
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v7); // adgbchef
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // adbgchef
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v7); // adbcghef
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v4, v2)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // agdbehfc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v7); // agdbehcf
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // agdbechf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v6); swap(v5, v7); swap(v6, v7); // adgbehfc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v6); swap(v5, v7); // adgbehcf
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); swap(v5, v6); swap(v6, v7); // adgbechf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // adbgehfc
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v7); // adbgehcf
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // adbgechf
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agdcbehf
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adgcbehf
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adcgbehf
																		}
																		else {
																			swap(v1, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adcbgehf
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agdbcehf
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adgbcehf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adbgcehf
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adbcgehf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v6, v7); // agdbefhc
																			}
																			else {
																				swap(v1, v6); swap(v2, v3); swap(v3, v6); // agdbefch
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v5, v6); // agdbecfh
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v6); swap(v6, v7); // adgbefhc
																			}
																			else {
																				swap(v1, v3); swap(v2, v6); // adgbefch
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v5, v6); // adgbecfh
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v6, v7); // adbgefhc
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); // adbgefch
																		}
																	}
																	else {
																		swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v5, v6); // adbgecfh
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // agdcbefh
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // adgcbefh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // adcgbefh
																		}
																		else {
																			swap(v1, v3); swap(v4, v6); swap(v5, v6); // adcbgefh
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v1, v6); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // agdbcefh
																		}
																		else {
																			swap(v1, v3); swap(v2, v6); swap(v4, v6); swap(v5, v6); // adgbcefh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // adbgcefh
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v4, v6); swap(v5, v6); // adbcgefh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v6, v1)) {
															if (compare(v7, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v1)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); // aghbefdc
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v6, v7); // agbhefdc
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // agbehfdc
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agbefhdc
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); // aghbedfc
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // aghbdefc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // agbhedfc
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agbehdfc
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agbhdefc
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // agbefdhc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // agbedhfc
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v6, v7); // agbedfhc
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); // agbefdch
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); // agbedfch
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agbdhefc
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v5, v7); swap(v6, v7); // agbdehfc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v6); swap(v6, v7); // agbdefhc
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); // agbdefch
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			swap(v2, v6); swap(v3, v7); swap(v6, v7); // abghefdc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // abgehfdc
																			}
																			else {
																				swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abgefhdc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // abgefdhc
																		}
																		else {
																			swap(v2, v6); swap(v3, v4); swap(v4, v5); // abgefdch
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v4)) {
																			swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // abghedfc
																		}
																		else {
																			swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abgehdfc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // abgedhfc
																			}
																			else {
																				swap(v2, v6); swap(v3, v4); swap(v6, v7); // abgedfhc
																			}
																		}
																		else {
																			swap(v2, v6); swap(v3, v4); // abgedfch
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abghdefc
																		}
																		else {
																			swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abgdhefc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v2, v6); swap(v5, v7); swap(v6, v7); // abgdehfc
																			}
																			else {
																				swap(v2, v6); swap(v6, v7); // abgdefhc
																			}
																		}
																		else {
																			swap(v2, v6); // abgdefch
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abdghefc
																		}
																		else {
																			swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // abdgehfc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v2, v3); swap(v3, v6); swap(v6, v7); // abdgefhc
																		}
																		else {
																			swap(v2, v3); swap(v3, v6); // abdgefch
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v6, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v1)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // aghbedcf
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v7); // agbhedcf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // agbehdcf
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v7); // agbedhcf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v1)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aghbdecf
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // agbhdecf
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v7); // agbdhecf
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v5, v7); // agbdehcf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // aghbdcef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agbhdcef
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agbdhcef
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			swap(v2, v6); swap(v3, v7); swap(v5, v7); // abghedcf
																		}
																		else {
																			swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // abgehdcf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // abghdecf
																		}
																		else {
																			swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abghdcef
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v4)) {
																			if (compare(v4, v2)) {
																				swap(v2, v6); swap(v4, v7); swap(v5, v7); // abgdhecf
																			}
																			else {
																				swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abgdhcef
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v2, v6); swap(v3, v4); swap(v5, v7); // abgedhcf
																			}
																			else {
																				swap(v2, v6); swap(v5, v7); // abgdehcf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v4)) {
																				swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // abdghecf
																			}
																			else {
																				swap(v2, v3); swap(v3, v6); swap(v5, v7); // abdgehcf
																			}
																		}
																		else {
																			swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abdghcef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // agbedchf
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v5, v6); swap(v6, v7); // agbdechf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // abgedchf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v2, v6); swap(v5, v6); swap(v6, v7); // abgdechf
																			}
																			else {
																				swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // abdgechf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v5, v6); // agbedcfh
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v5, v6); // agbdecfh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v2, v6); swap(v3, v4); swap(v5, v6); // abgedcfh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v2, v6); swap(v5, v6); // abgdecfh
																			}
																			else {
																				swap(v2, v3); swap(v3, v6); swap(v5, v6); // abdgecfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v7); // agbdchef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agbdcehf
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v4, v6); swap(v5, v6); // agbdcefh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v2, v6); swap(v4, v6); swap(v5, v7); // abgdchef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abgdcehf
																			}
																			else {
																				swap(v2, v6); swap(v4, v6); swap(v5, v6); // abgdcefh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v4)) {
																			swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // abdgchef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abdgcehf
																			}
																			else {
																				swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // abdgcefh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v2, v3); swap(v4, v6); swap(v5, v7); // abdcghef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abdcgehf
																			}
																			else {
																				swap(v2, v3); swap(v4, v6); swap(v5, v6); // abdcgefh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v0)) {
											if (compare(v5, v3)) {
												if (compare(v6, v5)) {
													if (compare(v3, v1)) {
														if (compare(v2, v1)) {
															if (compare(v7, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v6, v7); // eghfadcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // eghafdcb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egahfdcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // eaghfdcb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // egfhadcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egfahdcb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // egafhdcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); // eagfhdcb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // egfadhcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v6, v7); // egfadchb
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); // egfadcbh
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egafdhcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // egafdchb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // egafdcbh
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // eagfdhcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eagfdchb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // eagfdcbh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); // eghfadbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); // eghafdbc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // egahfdbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eaghfdbc
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); // egfhadbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // egfahdbc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); // egafhdbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // eagfhdbc
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v7, v1)) {
																		swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); // egfadhbc
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // egfadbhc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v6); // egfadbch
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // egafdhbc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egafdbhc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // egafdbch
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eagfdhbc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // eagfdbhc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // eagfdbch
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v7, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); // eghfabdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // eghafbdc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egahfbdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // eaghfbdc
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // egfhabdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egfahbdc
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // egafhbdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eagfhbdc
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // egfabhdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egfabdhc
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // egfabdch
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egafbhdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // egafbdhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); // egafbdch
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // eagfbhdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v6, v7); // eagfbdhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); // eagfbdch
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); // eghabfdc
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // egahbfdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); // egabhfdc
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // egabfhdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egabfdhc
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // egabfdch
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v6, v7); // eaghbfdc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // eagbhfdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eagbfhdc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // eagbfdhc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // eagbfdch
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); // eabghfdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // eabgfhdc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eabgfdhc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // eabgfdch
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v2)) {
														if (compare(v7, v1)) {
															if (compare(v7, v3)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); // efghadcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // efghadbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v6, v7); // efghabdc
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // efgahdcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efgahdbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // efgahbdc
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efaghdcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // efaghdbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); // efaghbdc
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // eafghdcb
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eafghdbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // eafghbdc
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); // efgadhcb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // efagdhcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); // efadghcb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eafgdhcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eafdghcb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // efgadhbc
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efagdhbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efadghbc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // eafgdhbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); // eafdghbc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v6, v7); // efgadbhc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // efagdbhc
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v6, v7); // efadgbhc
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efadbghc
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efgabhdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // efgabdhc
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // efagbhdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efagbdhc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efabghdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // efabgdhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efabdghc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eafgdbhc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eafdgbhc
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); // eafgbhdc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // eafgbdhc
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // eafbghdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eafbgdhc
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v5, v6); swap(v6, v7); // eafdbghc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // eafbdghc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eabfghdc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // eabfgdhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eabfdghc
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v2, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // efgadchb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efagdchb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efadgchb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v5, v6); swap(v6, v7); // efadcghb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // efgadcbh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // efagdcbh
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // efadgcbh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v5, v6); // efadcgbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); // efadcbgh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); // efgadbch
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); // efagdbch
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v6); // efadgbch
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v5); swap(v5, v6); // efadbgch
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v4, v5); // efadbcgh
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); // efgabdch
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // efagbdch
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); // efabgdch
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // efabdgch
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v5); // efabdcgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // eafgdchb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // eafgdcbh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // eafgdbch
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // eafgbdch
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v6, v7); // eafdgchb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eafdcghb
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v6); // eafdgcbh
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // eafdcgbh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // eafdgbch
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v5, v6); // eafdbgch
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // eafbgdch
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // eafbdgch
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // eabfgdch
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // eabfdgch
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v4, v5); // eafdcbgh
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); // eafdbcgh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); // eafbdcgh
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // eabfdcgh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v5, v1)) {
														if (compare(v2, v1)) {
															if (compare(v7, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v6, v7); // eghadfcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // egahdfcb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v6, v7); // egadhfcb
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egadfhcb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); // eaghdfcb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v7); // eagdhfcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); // eagdfhcb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); // eadghfcb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // eadgfhcb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eadfghcb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v6, v7); // egadfchb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eagdfchb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eadgfchb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // eadfgchb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eadfcghb
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v5); // egadfcbh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // eagdfcbh
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // eadgfcbh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // eadfgcbh
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // eadfcgbh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // eadfcbgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); // eghadfbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); // egahdfbc
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v7); // egadhfbc
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); // egadfhbc
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // eaghdfbc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v6, v7); // eagdhfbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eagdfhbc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // eadghfbc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eadgfhbc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // eadfghbc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egadfbhc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v6, v7); // eagdfbhc
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // eadgfbhc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eadfgbhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // eadfbghc
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v6); // egadfbch
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v5); // eagdfbch
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // eadgfbch
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // eadfgbch
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // eadfbgch
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v5); // eadfbcgh
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v6, v3)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); // eghadbfc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // egahdbfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); // egadhbfc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egadbhfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v6, v7); // egadbfhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v6); // egadbfch
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eaghdbfc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eagdhbfc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v5, v7); swap(v6, v7); // eagdbhfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v6, v7); // eagdbfhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); // eagdbfch
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eadghbfc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // eadgbhfc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v6, v7); // eadgbfhc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); // eadgbfch
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eadbghfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // eadbgfhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eadbfghc
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // eadbgfch
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // eadbfgch
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // eadbfcgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // eghabdfc
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egahbdfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // egabhdfc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egabdhfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // egabdfhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); // egabdfch
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // eaghbdfc
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eagbhdfc
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // eagbdhfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v6, v7); // eagbdfhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); // eagbdfch
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // eabghdfc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eabgdhfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // eabgdfhc
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eabdghfc
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v6, v7); // eabdgfhc
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eabdfghc
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // eabgdfch
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v6); // eabdgfch
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // eabdfgch
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v5); // eabdfcgh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v5, v1)) {
															if (compare(v6, v3)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // eghadcfb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egahdcfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egadhcfb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v5, v7); swap(v6, v7); // egadchfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v6, v7); // egadcfhb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); // egadcfbh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // eaghdcfb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); // eagdhcfb
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eagdchfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v6, v7); // eagdcfhb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v6); // eagdcfbh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // eadghcfb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eadgchfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // eadgcfhb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eadcghfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v6, v7); // eadcgfhb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eadcfghb
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // eadgcfbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v6); // eadcgfbh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // eadcfgbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v5); // eadcfbgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); // eghadcbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // egahdcbf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); // egadhcbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v5, v7); // egadchbf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eaghdcbf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eagdhcbf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eadghcbf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); // eagdchbf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // eadgchbf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); // eadcghbf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v5, v6); swap(v6, v7); // egadcbhf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eagdcbhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eadgcbhf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v5, v6); // egadcbfh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // eagdcbfh
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // eadgcbfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eadcgbhf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // eadcgbfh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v5, v6); swap(v6, v7); // eadcbghf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v5, v6); // eadcbgfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); // eadcbfgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v7, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // eghadbcf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egahdbcf
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egadhbcf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v7); // egadbhcf
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // eaghdbcf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); // eagdhbcf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // eadghbcf
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v5, v7); // eagdbhcf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); // eadgbhcf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // eadbghcf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egadbchf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v5, v6); swap(v6, v7); // eagdbchf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // eadgbchf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v4, v6); swap(v5, v6); // egadbcfh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v5, v6); // eagdbcfh
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // eadgbcfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eadbgchf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // eadbgcfh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // eadbcghf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // eadbcgfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v3); swap(v3, v4); // eadbcfgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eghabdcf
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // egahbdcf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egabhdcf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // egabdhcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egabdchf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // egabdcfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); // eaghbdcf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // eagbhdcf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); // eagbdhcf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // eagbdchf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // eagbdcfh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eabghdcf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // eabgdhcf
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eabgdchf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // eabgdcfh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); // eabdghcf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eabdgchf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // eabdgcfh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v5, v6); swap(v6, v7); // eabdcghf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v5, v6); // eabdcgfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); // eabdcfgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v1)) {
												if (compare(v7, v1)) {
													if (compare(v7, v5)) {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // aeghfdcb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); // aeghdfcb
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v4, v7); // aegdhfcb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); // aedghfcb
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // adeghfcb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aeghfdbc
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // aeghdfbc
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // aeghfbdc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v6, v7); // aegdhfbc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // aedghfbc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); // adeghfbc
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // aeghdcfb
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); // aegdhcfb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // aedghcfb
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // adeghcfb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aegdchfb
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aedgchfb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // adegchfb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aedcghfb
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adecghfb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adceghfb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aeghdcbf
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aegdhcbf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aedghcbf
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // adeghcbf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); // aegdchbf
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // aedgchbf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); // adegchbf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); // aedcghbf
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // adecghbf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // adceghbf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aeghdbfc
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v6, v7); // aeghbfdc
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // aeghbdfc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aegdhbfc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aedghbfc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // adeghbfc
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // aeghdbcf
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v7); // aegdhbcf
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v7); // aeghbdcf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // aedghbcf
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // adeghbcf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); // aegfhdcb
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // aegfdhcb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); // aegdfhcb
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // aedgfhcb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // adegfhcb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // aegfhdbc
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aegfdhbc
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aegfhbdc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aegdfhbc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aedgfhbc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // adegfhbc
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aegfdchb
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aegdfchb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aedgfchb
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // adegfchb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v6, v7); // aegdcfhb
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // aedgcfhb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v6, v7); // adegcfhb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v6, v7); // aedcgfhb
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // adecgfhb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // adcegfhb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // aefghdcb
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aefgdhcb
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aefdghcb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aedfghcb
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // adefghcb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aefghdbc
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // aefgdhbc
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // aefghbdc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); // aefdghbc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // aedfghbc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adefghbc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // aefgdchb
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v6, v7); // aefdgchb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // aedfgchb
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adefgchb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aefdcghb
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aedfcghb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // adefcghb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aedcfghb
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adecfghb
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adcefghb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v6, v5)) {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // aegfdcbh
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // aegdfcbh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // aedgfcbh
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // adegfcbh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v4, v6); // aegdcfbh
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // aedgcfbh
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); // adegcfbh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v6); // aedcgfbh
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // adecgfbh
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v6); // adcegfbh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // aegfdbhc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // aegfbhdc
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v6, v7); // aegfbdhc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v6, v7); // aegdfbhc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // aedgfbhc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adegfbhc
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); // aegfdbch
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); // aegfbdch
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v4, v5); // aegdfbch
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); // aedgfbch
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // adegfbch
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // aefgdcbh
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v4, v6); // aefdgcbh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // aedfgcbh
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // adefgcbh
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // aefdcgbh
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // aedfcgbh
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // adefcgbh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // aedcfgbh
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // adecfgbh
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // adcefgbh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aefgdbhc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); // aefgbhdc
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // aefgbdhc
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aefdgbhc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aedfgbhc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // adefgbhc
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // aefgdbch
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v6); // aefdgbch
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v6); // aefgbdch
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // aedfgbch
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); // adefgbch
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v7, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v5, v7); swap(v6, v7); // aegdbhfc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // aedgbhfc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adegbhfc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v5, v7); // aegdbhcf
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v7); // aedgbhcf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // adegbhcf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // aegbhfdc
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aegbhdfc
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // aegbdhfc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // aegbhdcf
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v7); // aegbdhcf
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aegdcbhf
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aedgcbhf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // adegcbhf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aedcgbhf
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adecgbhf
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adcegbhf
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v6); swap(v5, v6); swap(v6, v7); // aegdbchf
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // aegbdchf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // aedgbchf
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adegbchf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v6, v7); // aegdbfhc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v6, v7); // aedgbfhc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // adegbfhc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); // aegdbfch
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); // aedgbfch
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); // adegbfch
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aegbfhdc
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // aegbfdhc
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v6, v7); // aegbdfhc
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); // aegbfdch
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v4); // aegbdfch
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // aegdcbfh
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // aedgcbfh
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // adegcbfh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v3)) {
																				swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // aedcgbfh
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // adecgbfh
																			}
																		}
																		else {
																			swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // adcegbfh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v6); swap(v5, v6); // aegdbcfh
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v5, v6); // aegbdcfh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v5, v6); // aedgbcfh
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // adegbcfh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v3, v1)) {
													if (compare(v6, v5)) {
														if (compare(v7, v5)) {
															if (compare(v6, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aedbghfc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // aedbghcf
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aedbgchf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adebghfc
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v7); // adebghcf
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adebgchf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // adbeghfc
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // adbeghcf
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // adbegchf
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v5, v6); swap(v6, v7); // aedcbghf
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // adecbghf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v1, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // adcebghf
																		}
																		else {
																			swap(v1, v3); swap(v5, v6); swap(v6, v7); // adcbeghf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // aedbcghf
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v5, v6); swap(v6, v7); // adebcghf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // adbecghf
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v5, v6); swap(v6, v7); // adbceghf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // aedbgfhc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); // aedbgfch
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // aedbgcfh
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v6, v7); // adebgfhc
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v4, v6); // adebgfch
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v4, v6); swap(v5, v6); // adebgcfh
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // adbegfhc
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); // adbegfch
																		}
																	}
																	else {
																		swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // adbegcfh
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v5, v6); // aedcbgfh
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // adecbgfh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v1, v3); swap(v3, v4); swap(v5, v6); // adcebgfh
																		}
																		else {
																			swap(v1, v3); swap(v5, v6); // adcbegfh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v5, v6); // aedbcgfh
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v5, v6); // adebcgfh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v5, v6); // adbecgfh
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v5, v6); // adbcegfh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v5); swap(v4, v5); // aefdcbgh
																		}
																		else {
																			swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // aedfcbgh
																		}
																	}
																	else {
																		swap(v1, v3); swap(v2, v4); swap(v3, v5); // adefcbgh
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v4, v5); // aedcfbgh
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // adecfbgh
																		}
																	}
																	else {
																		swap(v1, v3); swap(v3, v4); swap(v4, v5); // adcefbgh
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v5, v6); swap(v6, v7); // aefdbghc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // aedfbghc
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adefbghc
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v5, v6); // aefdbgch
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v5, v6); // aedfbgch
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // adefbgch
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v5); // aefdbcgh
																		}
																		else {
																			swap(v1, v4); swap(v2, v3); swap(v3, v5); // aedfbcgh
																		}
																	}
																	else {
																		swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v5); // adefbcgh
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aedbfghc
																			}
																			else {
																				swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // aedbfgch
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v3); swap(v3, v4); swap(v4, v5); // aedbfcgh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adebfghc
																			}
																			else {
																				swap(v1, v3); swap(v2, v4); swap(v4, v5); swap(v5, v6); // adebfgch
																			}
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v4, v5); // adebfcgh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // adbefghc
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // adbefgch
																		}
																	}
																	else {
																		swap(v1, v3); swap(v2, v3); swap(v3, v4); swap(v4, v5); // adbefcgh
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); // aedcbfgh
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); swap(v3, v4); // adecbfgh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v1, v3); swap(v3, v4); // adcebfgh
																		}
																		else {
																			swap(v1, v3); // adcbefgh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v1, v4); swap(v2, v3); swap(v3, v4); // aedbcfgh
																		}
																		else {
																			swap(v1, v3); swap(v2, v4); // adebcfgh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v3); swap(v2, v3); swap(v3, v4); // adbecfgh
																		}
																		else {
																			swap(v1, v3); swap(v2, v3); // adbcefgh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v2)) {
														if (compare(v4, v1)) {
															if (compare(v5, v3)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); // aebghfdc
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // aebgfhdc
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // aefbghdc
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aebfghdc
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aefbgdhc
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // aefbdghc
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aebgfdhc
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // aebfgdhc
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aebfdghc
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // aebghdfc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aebgdhfc
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // aebgdfhc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aebghdcf
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // aebgdhcf
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aebdghfc
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v7); // aebdghcf
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v6, v7); // aebdgfhc
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aebdfghc
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v2, v4); swap(v3, v6); swap(v4, v7); // abeghfdc
																			}
																			else {
																				swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // abegfhdc
																			}
																		}
																		else {
																			swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abefghdc
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abegfdhc
																		}
																		else {
																			swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // abefgdhc
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // abeghdfc
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abegdhfc
																			}
																			else {
																				swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // abegdfhc
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abeghdcf
																		}
																		else {
																			swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // abegdhcf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abedghfc
																			}
																			else {
																				swap(v2, v4); swap(v4, v6); swap(v6, v7); // abedgfhc
																			}
																		}
																		else {
																			swap(v2, v4); swap(v4, v6); swap(v5, v7); // abedghcf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abefdghc
																		}
																		else {
																			swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abedfghc
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abdeghfc
																		}
																		else {
																			swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // abdeghcf
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // abdegfhc
																		}
																		else {
																			swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abdefghc
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v4, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // aefbgdch
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v5, v6); // aefbdgch
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); // aefbdcgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // aebgfdch
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); // aebfgdch
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // aebfdgch
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v5); // aebfdcgh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); // aebgdfch
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v4, v6); // aebdgfch
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v4); swap(v4, v5); swap(v5, v6); // aebdfgch
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v4, v5); // aebdfcgh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // abegfdch
																		}
																		else {
																			swap(v2, v4); swap(v3, v5); swap(v4, v6); // abefgdch
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // abefdgch
																		}
																		else {
																			swap(v2, v4); swap(v3, v5); swap(v4, v5); // abefdcgh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v6, v3)) {
																				swap(v2, v4); swap(v3, v6); swap(v4, v6); // abegdfch
																			}
																			else {
																				swap(v2, v4); swap(v4, v6); // abedgfch
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v2, v4); swap(v4, v5); swap(v5, v6); // abedfgch
																			}
																			else {
																				swap(v2, v4); swap(v4, v5); // abedfcgh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v2, v3); swap(v3, v4); swap(v4, v6); // abdegfch
																			}
																			else {
																				swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // abdefgch
																			}
																		}
																		else {
																			swap(v2, v3); swap(v3, v4); swap(v4, v5); // abdefcgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aebgdchf
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aebdgchf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abegdchf
																			}
																			else {
																				swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abedgchf
																			}
																		}
																		else {
																			swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abdegchf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // aebgdcfh
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v4, v6); swap(v5, v6); // aebdgcfh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // abegdcfh
																			}
																			else {
																				swap(v2, v4); swap(v4, v6); swap(v5, v6); // abedgcfh
																			}
																		}
																		else {
																			swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // abdegcfh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v5, v6); swap(v6, v7); // aebdcghf
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v5, v6); // aebdcgfh
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); // aebdcfgh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v2, v4); swap(v5, v6); swap(v6, v7); // abedcghf
																			}
																			else {
																				swap(v2, v4); swap(v5, v6); // abedcgfh
																			}
																		}
																		else {
																			swap(v2, v4); // abedcfgh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // abdecghf
																			}
																			else {
																				swap(v2, v3); swap(v3, v4); swap(v5, v6); // abdecgfh
																			}
																		}
																		else {
																			swap(v2, v3); swap(v3, v4); // abdecfgh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v2, v3); swap(v5, v6); swap(v6, v7); // abdceghf
																			}
																			else {
																				swap(v2, v3); swap(v5, v6); // abdcegfh
																			}
																		}
																		else {
																			swap(v2, v3); // abdcefgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
					else {
						if (compare(v2, v0)) {
							if (compare(v5, v4)) {
								if (compare(v7, v6)) {
									if (compare(v6, v4)) {
										if (compare(v5, v0)) {
											if (compare(v4, v1)) {
												if (compare(v3, v0)) {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v6, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v5)) {
																		swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hgfecdab
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfgecdab
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fhgecdab
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v6, v7); // hgfcedab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hfgcedab
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); // fhgcedab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hgfcdeab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfgcdeab
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fhgcdeab
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hfcgedab
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfcgdeab
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfcdgeab
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v5, v6); // fhcgedab
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fhcgdeab
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v4, v6); swap(v5, v6); // fhcdgeab
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); // fchgedab
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fchgdeab
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); // fchdgeab
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fcdhgeab
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v5, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v6, v7); // hgcfedab
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v6, v7); // hcgfedab
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); // chgfedab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hgcfdeab
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hcgfdeab
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); // chgfdeab
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v4, v5); swap(v6, v7); // hgcdfeab
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v6, v7); // hcgdfeab
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hcdgfeab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v5); // chgdfeab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); // chdgfeab
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cdhgfeab
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hcfgedab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v6); // chfgedab
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cfhgedab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcfgdeab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // chfgdeab
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfhgdeab
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcfdgeab
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v5, v6); // chfdgeab
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfhdgeab
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfdhgeab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcdfgeab
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // chdfgeab
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdhfgeab
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cdfhgeab
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v5, v2)) {
																if (compare(v6, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgfcdaeb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hfgcdaeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); // fhgcdaeb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hfcgdaeb
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v6); // fhcgdaeb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v4, v6); swap(v5, v7); // hfcdgaeb
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v4, v6); // fhcdgaeb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfcdageb
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v4, v5); swap(v5, v6); // fhcdageb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgcfdaeb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcgfdaeb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hcfgdaeb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); // hcfdgaeb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcfdageb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgcdfaeb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcgdfaeb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcdgfaeb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hcdfgaeb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcdfageb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // chgfdaeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); // chfgdaeb
																			}
																		}
																		else {
																			if (compare(v6, v0)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v6); // chfdgaeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); // chfdageb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); // chgdfaeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // chdgfaeb
																			}
																		}
																		else {
																			if (compare(v6, v0)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); // chdfgaeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // chdfageb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); // fchgdaeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v6); // fchdgaeb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); // fchdageb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cfhgdaeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); // cfhdgaeb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfhdageb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); // fcdhgaeb
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fcdhageb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcdahgeb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cfdhgaeb
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfdhageb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfdahgeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdhgfaeb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cdhfgaeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); // cdfhgaeb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdhfageb
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cdfhageb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdfahgeb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v7, v5)) {
															if (compare(v4, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgfecadb
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hfgecadb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hgfceadb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); // hfgceadb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v5, v7); // hfcgeadb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hgfecabd
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfgecabd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); // hgfceabd
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hfgceabd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hfcgeabd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgfcaedb
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hgfcadeb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hgfcaebd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hfgcaedb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfgcadeb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfgcaebd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hfcgaedb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfcgadeb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfcgaebd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfcagedb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hfcagebd
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hfcagdeb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfcadgeb
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); // fhgecadb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); // fhgceadb
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fhgecabd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v6, v7); // fhgceabd
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); // fhgcaedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fhgcadeb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fhgcaebd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); // fhcgeadb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v6, v7); // fhcgeabd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				if (compare(v4, v3)) {
																					swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v5); // fhcgaedb
																				}
																				else {
																					swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fhcgadeb
																				}
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fhcgaebd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fhcagedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhcagebd
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v6); // fhcagdeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fhcadgeb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); // fchgeadb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v6, v7); // fchgeabd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); // fchgaedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fchgaebd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fchagedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fchagebd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcahgedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fcahgebd
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fchgadeb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); // fchagdeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fchadgeb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fcahgdeb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcahdgeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcadhgeb
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v6, v5)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hgcfeadb
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v5, v7); // hgcfeabd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgcfaedb
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hgcfadeb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hgcfaebd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hcgfeadb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); // hcgfeabd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcgfaedb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hcgfadeb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hcgfaebd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); // hcfgeadb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hcfgeabd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hcfgaedb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcfgadeb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcfgaebd
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcfagedb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hcfagebd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hcfagdeb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcfadgeb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); // chgfeadb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); // chfgeadb
																			}
																		}
																		else {
																			if (compare(v6, v0)) {
																				if (compare(v6, v5)) {
																					swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // chgfaedb
																				}
																				else {
																					swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); // chfgaedb
																				}
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // chfagedb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); // chgfadeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // chfgadeb
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); // chfagdeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // chfadgeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // chgfeabd
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v6, v7); // chfgeabd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chgfaebd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // chfgaebd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chfagebd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); // cfhgeadb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cfhgeabd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cfhgaedb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfhgaebd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfhagedb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cfhagebd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfahgedb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cfahgebd
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfhgadeb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cfhagdeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfhadgeb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cfahgdeb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfahdgeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfadhgeb
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v3)) {
													if (compare(v5, v2)) {
														if (compare(v6, v0)) {
															if (compare(v3, v1)) {
																if (compare(v3, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hgfcdabe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfgcdabe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fhgcdabe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfcgdabe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fhcgdabe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fchgdabe
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); // hgfcadbe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hfgcadbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhgcadbe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hfcgadbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhcgadbe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fchgadbe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgfcabed
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hfgcabed
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhgcabed
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hgfcabde
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); // hfgcabde
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fhgcabde
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hfcgabed
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhcgabed
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fchgabed
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v6); swap(v4, v7); // hfcgabde
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fhcgabde
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fchgabde
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfcagdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v6); // hfcagbed
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hfcagbde
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fhcagdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fhcagbed
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhcagbde
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fchagdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fchagbed
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fchagbde
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcahgdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); // fcahgbed
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fcahgbde
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hfcabged
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhcabged
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fchabged
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fcahbged
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcabhged
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfcabgde
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fhcabgde
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fchabgde
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcahbgde
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fcabhgde
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v7, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hgcfdabe
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v4, v5); swap(v5, v7); // hgcdfabe
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				if (compare(v6, v5)) {
																					swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hcgfdabe
																				}
																				else {
																					swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcfgdabe
																				}
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); // hcgdfabe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v7); // hgcfadbe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); // hcgfadbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hcfgadbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgcfabed
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcgfabed
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hcfgabed
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hgcfabde
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hcgfabde
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); // hcfgabde
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v6, v5)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chgfdabe
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chgdfabe
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // chfgdabe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfhgdabe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // chgfadbe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chfgadbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cfhgadbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // chgfabed
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chfgabed
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cfhgabed
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chgfabde
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // chfgabde
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); // cfhgabde
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcfagdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); // hcfagbed
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hcfagbde
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // chfagdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // chfagbed
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chfagbde
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfhagdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); // cfhagbed
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cfhagbde
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfahgdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v6); // cfahgbed
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cfahgbde
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hcfabged
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chfabged
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cfhabged
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cfahbged
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfabhged
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcfabgde
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // chfabgde
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfhabgde
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfahbgde
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cfabhgde
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v7, v5)) {
															if (compare(v6, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfcdgabe
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcfdgabe
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hcdgfabe
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcdfgabe
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v6); swap(v6, v7); // chfdgabe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chdgfabe
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // chdfgabe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cdhgfabe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdhfgabe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v4, v7); swap(v5, v6); // hfcdagbe
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); // hcfdagbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hcdfagbe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hfcadgbe
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hcfadgbe
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chfdagbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chfadgbe
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chdfagbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cdhfagbe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v3, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v4, v6); swap(v6, v7); // fhcdgabe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v6, v7); // fchdgabe
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fcdhgabe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhcdagbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fchdagbe
																			}
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcdhagbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fcdahgbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhcadgbe
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fchadgbe
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fcahdgbe
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); // fcadhgbe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfhdgabe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfdhgabe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cdfhgabe
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v3, v0)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v6); // cfhdagbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cfhadgbe
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cfdhagbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdfhagbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v6); // cfdahgbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cdfahgbe
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cfahdgbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v4, v7); swap(v5, v6); // cfadhgbe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v7, v3)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v4, v7); // hfcdabge
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v4, v5); swap(v5, v7); // fhcdabge
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v7); // hcfdabge
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); // hcdfabge
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); // chfdabge
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // chdfabge
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); // fchdabge
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v7); // cfhdabge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fcdhabge
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); // cfdhabge
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); // cdhfabge
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cdfhabge
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); // fcdahbge
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v7); // cfdahbge
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); // cdfahbge
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fcdabhge
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); // cfdabhge
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cdfabhge
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v7); // hfcadbge
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fhcadbge
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fchadbge
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); // hcfadbge
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // chfadbge
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); // cfhadbge
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hfcabdge
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v7); // fhcabdge
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); // fchabdge
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hcfabdge
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); // chfabdge
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cfhabdge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); // fcahdbge
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v7); // fcadhbge
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); // fcadbhge
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v7); // cfahdbge
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v4, v7); // cfadhbge
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v4, v5); swap(v5, v7); // cfadbhge
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fcahbdge
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); // fcabhdge
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fcabdhge
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cfahbdge
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v7); // cfabhdge
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cfabdhge
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v3)) {
												if (compare(v6, v5)) {
													if (compare(v4, v3)) {
														if (compare(v7, v2)) {
															if (compare(v4, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hgcafedb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hcgafedb
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); // hcagfedb
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v5); // hgcafebd
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); // hcgafebd
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hcagfebd
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hgcafbed
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hcgafbed
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v6); // hgcabfed
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); // hcgabfed
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcagfbed
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hcagbfed
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); // hcabgfed
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v6, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); // chgafedb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // chgafebd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chgafbed
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // chgabfed
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v5); // chagfedb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // chagfebd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chagfbed
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // chagbfed
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v6); // chabgfed
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v3, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); // cahgfedb
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cahgfebd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cahgfbed
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cahgbfed
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); // cahbgfed
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v6); // cabhgfed
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v3, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgcafdeb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcgafdeb
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hcagfdeb
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hgcafdbe
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hcgafdbe
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcagfdbe
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgcafbde
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcgafbde
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hgcabfde
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hcgabfde
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hcagfbde
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); // hcagbfde
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hcabgfde
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v6, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // chgafdeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chgafdbe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // chgafbde
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); // chgabfde
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // chagfdeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chagfdbe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // chagfbde
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v7); // chagbfde
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // chabgfde
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cahgfdeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cahgfdbe
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cahgfbde
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); // cahgbfde
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cahbgfde
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cabhgfde
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v4, v1)) {
															if (compare(v7, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hcafgedb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcafgebd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); // hcafgdeb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hcafdgeb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // chafgedb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chafgebd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v6); // chafgdeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // chafdgeb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cahfgedb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cafhgedb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cahfgebd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cafhgebd
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); // cahfgdeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cafhgdeb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cahfdgeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cafhdgeb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cafdhgeb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hcafgdbe
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcafdgbe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // chafgdbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chafdgbe
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hcafgbed
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcafgbde
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // chafgbed
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chafgbde
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cahfgdbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cafhgdbe
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cahfdgbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cafhdgbe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); // cafdhgbe
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cahfgbed
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); // cafhgbed
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cahfgbde
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cafhgbde
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v5, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcafbged
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chafbged
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cahfbged
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cafhbged
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cafbhged
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hcafbgde
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // chafbgde
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cahfbgde
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cafhbgde
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cafbhgde
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hcabfged
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // chabfged
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cahbfged
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cabhfged
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cabfhged
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcabfgde
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chabfgde
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cahbfgde
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cabhfgde
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cabfhgde
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hcafdbge
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); // hcafbdge
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hcabfdge
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // chafdbge
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v7); // chafbdge
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // chabfdge
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cahfdbge
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); // cahfbdge
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cahbfdge
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cabhfdge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); // cafhdbge
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cafhbdge
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); // cafbhdge
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v7); // cabfhdge
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v7); // cafdhbge
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); // cafdbhge
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cafbdhge
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cabfdhge
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v3, v0)) {
														if (compare(v7, v3)) {
															if (compare(v6, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v4, v7); swap(v6, v7); // hgcdafeb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v6, v7); // hcgdafeb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hcdgafeb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v6); // chgdafeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); // chdgafeb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v4, v7); // hgcdafbe
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v7); // hcgdafbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); // hcdgafbe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v6, v7); // chgdafbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // chdgafbe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); // hcdagfeb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hcdafgeb
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hcdagfbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcdafgbe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hcdafbge
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v6); // chdagfeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); // chdafgeb
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v6, v7); // chdagfbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chdafgbe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); // chdafbge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v0)) {
																		swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cdhgafeb
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); // cdhagfeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cdhafgeb
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); // cdhgafbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cdhagfbe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdhafgbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cdhafbge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v6); // cdahgfeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cdahfgeb
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cdahgfbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdahfgbe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cdahfbge
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdafhgeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cdafhgbe
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v7); // cdafhbge
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cdafbhge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v0)) {
															if (compare(v6, v3)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hgcadfeb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hcgadfeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); // chgadfeb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); // hcagdfeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v6); // chagdfeb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v7); // hgcadfbe
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); // hcgadfbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // chgadfbe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hcagdfbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // chagdfbe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v6); // hcadgfeb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); // hcadfgeb
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v6, v7); // hcadgfbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcadfgbe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); // hcadfbge
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v4, v6); // chadgfeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v4, v5); swap(v5, v6); // chadfgeb
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v4, v6); swap(v6, v7); // chadgfbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chadfgbe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v4, v5); swap(v5, v7); // chadfbge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v3)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); // cahgdfeb
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v6); // cahdgfeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); // cahdfgeb
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cahgdfbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v6, v7); // cahdgfbe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cahdfgbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); // cahdfbge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); // cadhgfeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cadhfgeb
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cadhgfbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cadhfgbe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cadhfbge
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cadfhgeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cadfhgbe
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); // cadfhbge
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cadfbhge
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v0)) {
														if (compare(v7, v2)) {
															if (compare(v6, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v4, v7); swap(v5, v6); // hgcdabfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); // hcgdabfe
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hgcadbfe
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hcgadbfe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcagdbfe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgcabdfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcgabdfe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hcagbdfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcabgdfe
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hcdgabfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcdagbfe
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hcdabgfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); // hcdabfge
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcadgbfe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); // hcadbgfe
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v7); // hcadbfge
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hcabdgfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); // hcabdfge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chgdabfe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chdgabfe
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cdhgabfe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chdagbfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdhagbfe
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); // chdabgfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v7); // chdabfge
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cdhabgfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); // cdhabfge
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chgadbfe
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chagdbfe
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chadgbfe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v7); swap(v4, v7); swap(v5, v6); // chadbgfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v4, v7); // chadbfge
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // chgabdfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // chagbdfe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chabgdfe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // chabdgfe
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v7); // chabdfge
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v7, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cahgdbfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cahdgbfe
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); // cahdbgfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v7); // cahdbfge
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cahgbdfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cahbgdfe
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cahbdgfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); // cahbdfge
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v6, v1)) {
																		swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdahgbfe
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cdahbgfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v7); // cdahbfge
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cadhgbfe
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cadhbgfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); // cadhbfge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v3, v0)) {
																	if (compare(v6, v5)) {
																		swap(v0, v2); swap(v1, v3); swap(v4, v7); swap(v5, v6); // cdabhgfe
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v4, v7); // cdabhfge
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v4, v5); swap(v5, v7); // cdabfhge
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); // cadbhgfe
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v7); // cadbhfge
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); // cadbfhge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v3)) {
																		swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cabhgdfe
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cabhdgfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v7); // cabhdfge
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		swap(v0, v2); swap(v1, v2); swap(v4, v7); swap(v5, v6); // cabdhgfe
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v4, v7); // cabdhfge
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v4, v5); swap(v5, v7); // cabdfhge
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v7, v0)) {
											if (compare(v6, v1)) {
												if (compare(v3, v0)) {
													if (compare(v6, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																if (compare(v7, v4)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hfegcdab
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfecgdab
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hfecdgab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); // fhegcdab
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fhecgdab
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); // fhecdgab
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fehgcdab
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fehcgdab
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fehcdgab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fechgdab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fechdgab
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v4, v7); swap(v5, v6); // fecdhgab
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfcegdab
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hfcedgab
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v5, v6); swap(v6, v7); // hfcdegab
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fhcegdab
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v5, v6); // fhcedgab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v5, v6); // fhcdegab
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fchegdab
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); // fchedgab
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v5, v6); // fchdegab
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); // fcdhegab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcehgdab
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fcehdgab
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); // fcedhgab
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fcdehgab
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcfegdab
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // chfegdab
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfhegdab
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cfehgdab
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hcfedgab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v5, v6); // chfedgab
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cfhedgab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfehdgab
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfedhgab
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v5, v6); swap(v6, v7); // hcfdegab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v5, v6); // chfdegab
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v5, v6); swap(v6, v7); // cfhdegab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cfdhegab
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfdehgab
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hcdfegab
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v5, v6); // chdfegab
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cdhfegab
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v5, v6); // cdfhegab
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cdfehgab
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v7, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); // hfecdagb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); // fhecdagb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); // fehcdagb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v5, v7); // hfcedagb
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v4); // fhcedagb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v5, v7); // hfcdeagb
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); // fhcdeagb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v4, v7); swap(v5, v7); // hfcdaegb
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v4, v5); // fhcdaegb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v7); // fechdagb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v4, v7); // fecdhagb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); // fchedagb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); // fcehdagb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v7); // fcedhagb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); // fchdeagb
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); // fcdheagb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); // fcdehagb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v5); // fchdaegb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); // fcdhaegb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); // hcfedagb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v5, v7); // hcfdeagb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); // hcfdaegb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); // chfedagb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); // chfdeagb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v5); // chfdaegb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); // cfhedagb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cfehdagb
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v5, v7); // cfhdeagb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); // cfhdaegb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v4, v0)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); // cfdheagb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cfdhaegb
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); // cfedhagb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cfdehagb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); // hcdfeagb
																		}
																		else {
																			if (compare(v5, v0)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hcdfaegb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); // hcdafegb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); // chdfeagb
																		}
																		else {
																			if (compare(v5, v0)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); // chdfaegb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v5); // chdafegb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); // cdhfeagb
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); // cdfheagb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); // cdfehagb
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cdhfaegb
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); // cdfhaegb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); // cdhafegb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v2)) {
														if (compare(v7, v5)) {
															if (compare(v5, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); // hfegcadb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hfecgadb
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hfegcabd
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfecgabd
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfecagdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hfecadgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hfecagbd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hfcegadb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfcegabd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				if (compare(v6, v3)) {
																					swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfceagdb
																				}
																				else {
																					swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hfceadgb
																				}
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hfceagbd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hfcaegdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v5, v6); // hfcaegbd
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v5, v7); // hfcaedgb
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hfcadegb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hcfegadb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcfegabd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcfeagdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hcfeagbd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hcfaegdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); // hcfaegbd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); // hcafegdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hcafegbd
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hcfeadgb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); // hcfaedgb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hcfadegb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); // hcafedgb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); // hcafdegb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v5); // hcadfegb
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); // fhegcadb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); // fehgcadb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); // fhecgadb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fehcgadb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v6); // fhcegadb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v6, v7); // fhegcabd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fehgcabd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fhecgabd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); // fehcgabd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fhcegabd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v4)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fhecagdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); // fhecadgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhecagbd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fehcagdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fehcadgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fehcagbd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fhceagdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v5); // fhceadgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhceagbd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v5, v6); // fhcaegdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fhcaegbd
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); // fhcaedgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v5); // fhcadegb
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v7, v4)) {
																if (compare(v4, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); // fchegadb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fchegabd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fcheagdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); // fcheadgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcheagbd
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); // fchaegdb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fchaegbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); // fchaedgb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); // fchadegb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fechgadb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v6); // fechgabd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fechagdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fechadgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fechagbd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fcehgadb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); // fcehgabd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcehagdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fcehadgb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fcehagbd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); // chfegadb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // chfegabd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // chfeagdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chfeagbd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); // chfaegdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // chfaegbd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v5, v6); // chafegdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // chafegbd
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); // chfeadgb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); // chfaedgb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); // chfadegb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); // chafedgb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v5); // chafdegb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v4, v5); // chadfegb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cfhegadb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfehgadb
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				if (compare(v7, v4)) {
																					swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfheagdb
																				}
																				else {
																					swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cfehagdb
																				}
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cfhaegdb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cfheadgb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); // cfehadgb
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); // cfhaedgb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cfhadegb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfhegabd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cfehgabd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cfheagbd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); // cfehagbd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); // cfhaegbd
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v5, v2)) {
														if (compare(v3, v1)) {
															if (compare(v3, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hfecdabg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v6, v7); // fhecdabg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fehcdabg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fechdabg
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v4, v7); swap(v6, v7); // fecdhabg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hfcedabg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v6, v7); // fhcedabg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v6, v7); // fchedabg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fcehdabg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v6, v7); // fcedhabg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfecadbg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fhecadbg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fehcadbg
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fechadbg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfceadbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fhceadbg
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fcheadbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcehadbg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hfcaedbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v6, v7); // fhcaedbg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v6, v7); // fchaedbg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); // hfecabgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fhecabgd
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); // fehcabgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v5); // fechabgd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v7); // hfceabgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fhceabgd
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fcheabgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); // fcehabgd
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); // hfcaebgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v5, v7); // fhcaebgd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); // fchaebgd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v5); // hfcabegd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fhcabegd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fchabegd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hfecabdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhecabdg
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fehcabdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fechabdg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hfceabdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhceabdg
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcheabdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fcehabdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v6, v7); // hfcaebdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fhcaebdg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fchaebdg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hfcabedg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhcabedg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fchabedg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v3, v1)) {
																if (compare(v3, v0)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hcfedabg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v6, v7); // chfedabg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cfhedabg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfehdabg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfedhabg
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcfeadbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // chfeadbg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfheadbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cfehadbg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); // hcfeabgd
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // chfeabgd
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); // cfheabgd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); // cfehabgd
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hcfeabdg
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chfeabdg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cfheabdg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v6, v7); // cfehabdg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hcfaedbg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v6, v7); // chfaedbg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cfhaedbg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); // hcfaebgd
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); // chfaebgd
																				}
																				else {
																					swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); // cfhaebgd
																				}
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v6, v7); // hcfaebdg
																			}
																			else {
																				if (compare(v7, v5)) {
																					swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // chfaebdg
																				}
																				else {
																					swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v6, v7); // cfhaebdg
																				}
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); // hcfabegd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // chfabegd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); // cfhabegd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hcfabedg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chfabedg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cfhabedg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v6, v7); // hcafedbg
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); // hcafebgd
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hcafebdg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hcafbegd
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcafbedg
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); // hcabfegd
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hcabfedg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v6, v7); // chafedbg
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v5, v7); // chafebgd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // chafebdg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // chafbegd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chafbedg
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v5); // chabfegd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // chabfedg
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v7, v5)) {
															if (compare(v7, v2)) {
																if (compare(v3, v0)) {
																	if (compare(v4, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v5, v7); swap(v6, v7); // hfcdeabg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v5, v7); swap(v6, v7); // hcfdeabg
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hcdfeabg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfcdaebg
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcfdaebg
																			}
																		}
																		else {
																			if (compare(v5, v0)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcdfaebg
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hcdafebg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfcadebg
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcfadebg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hcafdebg
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v6, v7); // hcadfebg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v6, v7); // chfdeabg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v6, v7); // chdfeabg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cdhfeabg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v3, v0)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v6, v7); // chfdaebg
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // chfadebg
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // chdfaebg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdhfaebg
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v6, v7); // chdafebg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cdhafebg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // chafdebg
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v4, v5); swap(v6, v7); // chadfebg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v6, v7); // fhcdeabg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v6, v7); // fchdeabg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v6, v7); // fcdheabg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fcdehabg
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v5, v7); swap(v6, v7); // cfhdeabg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cfdheabg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfdehabg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v6, v7); // cdfheabg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cdfehabg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v4, v5); swap(v6, v7); // fhcdaebg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v6, v7); // fchdaebg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fcdhaebg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fhcadebg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fchadebg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfhdaebg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfhadebg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfdhaebg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cdfhaebg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v5, v3)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v4, v7); swap(v6, v7); // hfcdabeg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhcdabeg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fchdabeg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcdhabeg
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v6, v7); // hcfdabeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chfdabeg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v4, v7); swap(v6, v7); // cfhdabeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cfdhabeg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hcdfabeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chdfabeg
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cdhfabeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdfhabeg
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcdafbeg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chdafbeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdhafbeg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hcdabfeg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v6, v7); // chdabfeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cdhabfeg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hfcadbeg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhcadbeg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fchadbeg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hcfadbeg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chfadbeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cfhadbeg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfcabdeg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fhcabdeg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fchabdeg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcfabdeg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // chfabdeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfhabdeg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcafdbeg
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcadfbeg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v6, v7); // hcadbfeg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chafdbeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chadfbeg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v4, v7); swap(v6, v7); // chadbfeg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hcafbdeg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcabfdeg
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hcabdfeg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // chafbdeg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chabfdeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // chabdfeg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v3)) {
												if (compare(v7, v4)) {
													if (compare(v4, v1)) {
														if (compare(v6, v1)) {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fcahegdb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); // fcahegbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); // fcahedgb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fcahdegb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cfahegdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v5, v6); // cfahegbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v5, v7); // cfahedgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cfahdegb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); // cahfegdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cafhegdb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cahfegbd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); // cafhegbd
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); // cahfedgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); // cafhedgb
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); // cahfdegb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cafhdegb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v5); // cahdfegb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fcahedbg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcahdebg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cfahedbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfahdebg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); // fcahebgd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v6, v7); // fcahebdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); // cfahebgd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v6, v7); // cfahebdg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v6, v7); // cahfedbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cafhedbg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cahfdebg
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cafhdebg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v6, v7); // cahdfebg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); // cahfebgd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); // cafhebgd
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cahfebdg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v6, v7); // cafhebdg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v7, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); // fcahbegd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v5); // cfahbegd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cahfbegd
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); // cafhbegd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); // cahbfegd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fcahbedg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cfahbedg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cahfbedg
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cafhbedg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cahbfedg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fcabhegd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cfabhegd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cafbhegd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v5); // cabhfegd
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cabfhegd
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcabhedg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfabhedg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cafbhedg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cabhfedg
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cabfhedg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v5, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cahfdbeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cahfbdeg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cahbfdeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cabhfdeg
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cahdfbeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v6, v7); // cahdbfeg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cahbdfeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cabhdfeg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fcahdbeg
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcahbdeg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fcabhdeg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cfahdbeg
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfahbdeg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cfabhdeg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cafhdbeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cafhbdeg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cafbhdeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cabfhdeg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v3)) {
														if (compare(v5, v2)) {
															if (compare(v6, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fecahgdb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fceahgdb
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcaehgdb
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fecahgbd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fceahgbd
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fcaehgbd
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fecahbgd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fceahbgd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v5, v7); // fecabhgd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); // fceabhgd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); // fcaehbgd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fcaebhgd
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); // fcabehgd
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfeahgdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cfeahgbd
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); // cfeahbgd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cfeabhgd
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfaehgdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cfaehgbd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v7); // cfaehbgd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cfaebhgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v5, v7); // cfabehgd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v3, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cafehgdb
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cafehgbd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); // cafehbgd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cafebhgd
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); // cafbehgd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v5, v7); // cabfehgd
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v3, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v7); // fecahdgb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); // fceahdgb
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fcaehdgb
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fecahdbg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fceahdbg
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcaehdbg
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fecahbdg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fceahbdg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fecabhdg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fceabhdg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fcaehbdg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcaebhdg
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fcabehdg
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cfeahdgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfeahdbg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cfeahbdg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfeabhdg
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cfaehdgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfaehdbg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cfaehbdg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfaebhdg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cfabehdg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cafehdgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cafehdbg
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cafehbdg
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cafebhdg
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cafbehdg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cabfehdg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v1)) {
													if (compare(v3, v0)) {
														if (compare(v5, v3)) {
															if (compare(v4, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v4, v5); swap(v5, v7); // fecdahgb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); // fcedahgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fcdeahgb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v5, v7); // cfedahgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); // cfdeahgb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fecdahbg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcedahbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcdeahbg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v5, v7); swap(v6, v7); // cfedahbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cfdeahbg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fcdahegb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); // fcdaehgb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcdahebg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fcdaehbg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fcdahbeg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); // cfdahegb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v5, v7); // cfdaehgb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfdahebg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v5, v7); swap(v6, v7); // cfdaehbg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v7); swap(v6, v7); // cfdahbeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v0)) {
																		swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cdfeahgb
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cdfahegb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); // cdfaehgb
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdfahebg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cdfahbeg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdfeahbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cdfaehbg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v5); // cdahfegb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cdahfebg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdahfbeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cdahbfeg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cdafhegb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v5, v7); // cdafehgb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdafhebg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cdafehbg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cdafhbeg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fecadhgb
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fceadhgb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); // cfeadhgb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); // fcaedhgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v5, v7); // cfaedhgb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fecadhbg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fceadhbg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cfeadhbg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fcaedhbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cfaedhbg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); // fcadhegb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v5, v7); // fcadehgb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcadhebg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v5, v7); swap(v6, v7); // fcadehbg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v6, v7); // fcadhbeg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v4, v7); swap(v5, v7); // cfadhegb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v5, v7); // cfadehgb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfadhebg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v5, v7); swap(v6, v7); // cfadehbg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v4, v7); swap(v6, v7); // cfadhbeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); // cafedhgb
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); // cafdhegb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v5, v7); // cafdehgb
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cafdhebg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v6, v7); // cafdhbeg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cafedhbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v5, v7); swap(v6, v7); // cafdehbg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); // cadhfegb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cadhfebg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cadhfbeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cadhbfeg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cadfhegb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); // cadfehgb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cadfhebg
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cadfehbg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cadfhbeg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v4, v5); swap(v6, v7); // fecdabhg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v6, v7); // fcedabhg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fecadbhg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fceadbhg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v6, v7); // fcaedbhg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v6, v7); // fecabdhg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v6, v7); // fceabdhg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fcaebdhg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v6, v7); // fcabedhg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fcdeabhg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v6, v7); // fcdaebhg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcdabheg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fcdabehg
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v6, v7); // fcadebhg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcadbheg
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v6, v7); // fcadbehg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcabdheg
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fcabdehg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v6, v7); // cfedabhg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v6, v7); // cfdeabhg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cdfeabhg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v6, v7); // cfdaebhg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v6, v7); // cdfaebhg
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfdabheg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v6, v7); // cfdabehg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdfabheg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cdfabehg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v6, v7); // cfeadbhg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v6, v7); // cfaedbhg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v6, v7); // cfadebhg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfadbheg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v4, v5); swap(v6, v7); // cfadbehg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cfeabdhg
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cfaebdhg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v6, v7); // cfabedhg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfabdheg
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cfabdehg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v6, v7); // cafedbhg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v6, v7); // cafdebhg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cafdbheg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v6, v7); // cafdbehg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cafebdhg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v6, v7); // cafbedhg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cafbdheg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cafbdehg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v4, v1)) {
																		swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v6, v7); // cdafebhg
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdafbheg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cdafbehg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v6, v7); // cadfebhg
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cadfbheg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cadfbehg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v3, v0)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v4, v7); swap(v6, v7); // cdabhfeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdabfheg
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v4, v5); swap(v6, v7); // cdabfehg
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v6, v7); // cadbhfeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cadbfheg
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v6, v7); // cadbfehg
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v6, v7); // cabfedhg
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cabfdheg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cabfdehg
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v4, v7); swap(v6, v7); // cabdhfeg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cabdfheg
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v4, v5); swap(v6, v7); // cabdfehg
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v7, v4)) {
										if (compare(v5, v0)) {
											if (compare(v4, v1)) {
												if (compare(v3, v0)) {
													if (compare(v4, v0)) {
														if (compare(v5, v2)) {
															if (compare(v7, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); // ghfecdab
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gfhecdab
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fghecdab
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); // ghfcedab
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); // gfhcedab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fghcedab
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); // ghfcdeab
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gfhcdeab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fghcdeab
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v5, v7); // gfchedab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gfchdeab
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v4, v7); swap(v5, v7); // gfcdheab
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fgchedab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgchdeab
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgcdheab
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fcghedab
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcghdeab
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcgdheab
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcdgheab
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v5, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v5); // ghcfedab
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); // gchfedab
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v6, v7); // cghfedab
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v5); // ghcfdeab
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); // gchfdeab
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cghfdeab
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v5); // ghcdfeab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v5); // gchdfeab
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); // gcdhfeab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v6, v7); // cghdfeab
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cgdhfeab
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); // cdghfeab
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); // gcfhedab
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cgfhedab
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v7); // cfghedab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gcfhdeab
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgfhdeab
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cfghdeab
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); // gcfdheab
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgfdheab
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v7); // cfgdheab
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cfdgheab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gcdfheab
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgdfheab
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cdgfheab
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdfgheab
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v5, v2)) {
																if (compare(v7, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ghfcdaeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfhcdaeb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fghcdaeb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfchdaeb
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fgchdaeb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfcdhaeb
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v4, v7); swap(v6, v7); // fgcdhaeb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v4, v6); swap(v5, v7); // gfcdaheb
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgcdaheb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ghcfdaeb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gchfdaeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcfhdaeb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcfdhaeb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); // gcfdaheb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v5); swap(v5, v6); // ghcdfaeb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); // gchdfaeb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gcdhfaeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcdfhaeb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gcdfaheb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cghfdaeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cgfhdaeb
																			}
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v7); swap(v6, v7); // cgfdhaeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgfdaheb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cghdfaeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgdhfaeb
																			}
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cgdfhaeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgdfaheb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fcghdaeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v6, v7); // fcgdhaeb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcgdaheb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfghdaeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfgdhaeb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); // cfgdaheb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v0)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fcdghaeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcdgaheb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fcdagheb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v7, v0)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfdghaeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cfdgaheb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); // cfdagheb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cdghfaeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdgfhaeb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cdfghaeb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cdgfaheb
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdfgaheb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cdfagheb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v6, v5)) {
															if (compare(v4, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ghfecadb
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfhecadb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); // ghfceadb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gfhceadb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gfcheadb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghfecabd
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gfhecabd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ghfceabd
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); // gfhceabd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v5, v6); // gfcheabd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ghfcaedb
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); // ghfcadeb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghfcaebd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfhcaedb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gfhcadeb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gfhcaebd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfchaedb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gfchadeb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gfchaebd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gfcahedb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfcahebd
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfcahdeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gfcadheb
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fghecadb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v6, v7); // fghceadb
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); // fghecabd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); // fghceabd
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fghcaedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fghcadeb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); // fghcaebd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v6, v7); // fgcheadb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); // fgcheabd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				if (compare(v4, v3)) {
																					swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fgchaedb
																				}
																				else {
																					swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgchadeb
																				}
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v5); // fgchaebd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgcahedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fgcahebd
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fgcahdeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgcadheb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v6, v7); // fcgheadb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); // fcgheabd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fcghaedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); // fcghaebd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcgahedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fcgahebd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fcaghedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcaghebd
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcghadeb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fcgahdeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcgadheb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcaghdeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fcagdheb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); // fcadgheb
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v7, v5)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v5, v6); // ghcfeadb
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ghcfeabd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ghcfaedb
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v6); // ghcfadeb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghcfaebd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); // gchfeadb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gchfeabd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gchfaedb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); // gchfadeb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gchfaebd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gcfheadb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); // gcfheabd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcfhaedb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gcfhadeb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gcfhaebd
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gcfahedb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcfahebd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcfahdeb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gcfadheb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cghfeadb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v6, v7); // cgfheadb
																			}
																		}
																		else {
																			if (compare(v7, v0)) {
																				if (compare(v7, v5)) {
																					swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cghfaedb
																				}
																				else {
																					swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cgfhaedb
																				}
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgfahedb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cghfadeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgfhadeb
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cgfahdeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgfadheb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); // cghfeabd
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); // cgfheabd
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cghfaebd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); // cgfhaebd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cgfahebd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cfgheadb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); // cfgheabd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfghaedb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cfghaebd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cfgahedb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfgahebd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cfaghedb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfaghebd
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cfghadeb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfgahdeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cfgadheb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfaghdeb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cfagdheb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v4, v6); swap(v5, v7); // cfadgheb
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v3)) {
													if (compare(v5, v2)) {
														if (compare(v7, v0)) {
															if (compare(v3, v1)) {
																if (compare(v3, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghfcdabe
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gfhcdabe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); // fghcdabe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gfchdabe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v7); // fgchdabe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); // fcghdabe
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ghfcadbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfhcadbe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fghcadbe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfchadbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fgchadbe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fcghadbe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ghfcabed
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); // gfhcabed
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fghcabed
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghfcabde
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gfhcabde
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fghcabde
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v6); // gfchabed
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fgchabed
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fcghabed
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gfchabde
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgchabde
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcghabde
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gfcahdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gfcahbed
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v7); // gfcahbde
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v7); // fgcahdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgcahbed
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fgcahbde
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); // fcgahdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fcgahbed
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fcgahbde
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fcaghdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fcaghbed
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); // fcaghbde
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfcabhed
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fgcabhed
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fcgabhed
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcagbhed
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fcabghed
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gfcabhde
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgcabhde
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcgabhde
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fcagbhde
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcabghde
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v0)) {
															if (compare(v6, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghcfdabe
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghcdfabe
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				if (compare(v7, v5)) {
																					swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gchfdabe
																				}
																				else {
																					swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gcfhdabe
																				}
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gchdfabe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ghcfadbe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gchfadbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcfhadbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ghcfabed
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gchfabed
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); // gcfhabed
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghcfabde
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gchfabde
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gcfhabde
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v7, v5)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cghfdabe
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v5); swap(v5, v7); // cghdfabe
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v7); // cgfhdabe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cfghdabe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); // cghfadbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cgfhadbe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfghadbe
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cghfabed
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cgfhabed
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); // cfghabed
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cghfabde
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgfhabde
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cfghabde
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gcfahdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gcfahbed
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); // gcfahbde
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); // cgfahdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgfahbed
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cgfahbde
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cfgahdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cfgahbed
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); // cfgahbde
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cfaghdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cfaghbed
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v7); // cfaghbde
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcfabhed
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cgfabhed
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfgabhed
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cfagbhed
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cfabghed
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gcfabhde
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgfabhde
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cfgabhde
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cfagbhde
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfabghde
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v6, v5)) {
															if (compare(v7, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v4, v7); swap(v5, v6); // gfcdhabe
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); // gcfdhabe
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcdhfabe
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gcdfhabe
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v7); // cgfdhabe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cgdhfabe
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v7); // cgdfhabe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdghfabe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cdgfhabe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfcdahbe
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcfdahbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcdfahbe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfcadhbe
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcfadhbe
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v7); // cgfdahbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cgfadhbe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cgdfahbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdgfahbe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v3, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v4, v7); // fgcdhabe
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v7); // fcgdhabe
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); // fcdghabe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v4, v5); swap(v5, v7); // fgcdahbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); // fcgdahbe
																			}
																		}
																		else {
																			if (compare(v6, v0)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fcdgahbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcdaghbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fgcadhbe
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fcgadhbe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcagdhbe
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcadghbe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v7); swap(v5, v6); // cfgdhabe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cfdghabe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v7); // cdfghabe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v3, v0)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfgdahbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfgadhbe
																			}
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfdgahbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cdfgahbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfdaghbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdfaghbe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfagdhbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfadghbe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v6, v3)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v4, v6); swap(v6, v7); // gfcdabhe
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgcdabhe
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v6, v7); // gcfdabhe
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gcdfabhe
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgfdabhe
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgdfabhe
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcgdabhe
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v6, v7); // cfgdabhe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcdgabhe
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cfdgabhe
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cdgfabhe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdfgabhe
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fcdagbhe
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v6, v7); // cfdagbhe
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cdfagbhe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcdabghe
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfdabghe
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdfabghe
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gfcadbhe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgcadbhe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcgadbhe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gcfadbhe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgfadbhe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cfgadbhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfcabdhe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fgcabdhe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fcgabdhe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcfabdhe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cgfabdhe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfgabdhe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fcagdbhe
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v6, v7); // fcadgbhe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcadbghe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cfagdbhe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v4, v6); swap(v6, v7); // cfadgbhe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfadbghe
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcagbdhe
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fcabgdhe
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcabdghe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfagbdhe
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cfabgdhe
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfabdghe
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v3)) {
												if (compare(v7, v5)) {
													if (compare(v4, v3)) {
														if (compare(v6, v2)) {
															if (compare(v4, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v5); // ghcafedb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); // gchafedb
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gcahfedb
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ghcafebd
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gchafebd
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); // gcahfebd
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghcafbed
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gchafbed
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ghcabfed
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gchabfed
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gcahfbed
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); // gcahbfed
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gcabhfed
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v7, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cghafedb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); // cghafebd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cghafbed
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); // cghabfed
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cgahfedb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v5); // cgahfebd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cgahfbed
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v6); // cgahbfed
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cgabhfed
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v3, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // caghfedb
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); // caghfebd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // caghfbed
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); // caghbfed
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cagbhfed
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cabghfed
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v3, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ghcafdeb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gchafdeb
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcahfdeb
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghcafdbe
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gchafdbe
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gcahfdbe
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ghcafbde
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gchafbde
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v7); // ghcabfde
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); // gchabfde
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcahfbde
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gcahbfde
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); // gcabhfde
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v7, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cghafdeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cghafdbe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cghafbde
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cghabfde
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgahfdeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cgahfdbe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgahfbde
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cgahbfde
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v7); // cgabhfde
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // caghfdeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // caghfdbe
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // caghfbde
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // caghbfde
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); // cagbhfde
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v7); // cabghfde
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v4, v1)) {
															if (compare(v6, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcafhedb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gcafhebd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gcafhdeb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcafdheb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgafhedb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cgafhebd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cgafhdeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgafdheb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cagfhedb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cafghedb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cagfhebd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cafghebd
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cagfhdeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cafghdeb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cagfdheb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cafgdheb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); // cafdgheb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); // gcafhdbe
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gcafdhbe
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v7); // cgafhdbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cgafdhbe
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcafhbed
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gcafhbde
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgafhbed
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cgafhbde
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); // cagfhdbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cafghdbe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cagfdhbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cafgdhbe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cafdghbe
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cagfhbed
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cafghbed
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cagfhbde
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); // cafghbde
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v5, v1)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gcafbhed
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cgafbhed
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cagfbhed
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cafgbhed
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cafbghed
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcafbhde
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgafbhde
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cagfbhde
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cafgbhde
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cafbghde
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcabfhed
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgabfhed
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cagbfhed
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cabgfhed
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cabfghed
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gcabfhde
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cgabfhde
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cagbfhde
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cabgfhde
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cabfghde
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcafdbhe
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gcafbdhe
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcabfdhe
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgafdbhe
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cgafbdhe
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgabfdhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cagfdbhe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cagfbdhe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cagbfdhe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cabgfdhe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cafgdbhe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cafgbdhe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cafbgdhe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cabfgdhe
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v6, v7); // cafdgbhe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cafdbghe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cafbdghe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cabfdghe
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v3, v0)) {
														if (compare(v6, v3)) {
															if (compare(v7, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v6); // ghcdafeb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v6); // gchdafeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); // gcdhafeb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v6, v7); // cghdafeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cgdhafeb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v6); swap(v6, v7); // ghcdafbe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v6, v7); // gchdafbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gcdhafbe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v7); // cghdafbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); // cgdhafbe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gcdahfeb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcdafheb
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); // gcdahfbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gcdafhbe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcdafbhe
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v6, v7); // cgdahfeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgdafheb
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v7); // cgdahfbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v7); // cgdafhbe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgdafbhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v0)) {
																		swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); // cdghafeb
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cdgahfeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdgafheb
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cdghafbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); // cdgahfbe
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cdgafhbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdgafbhe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cdaghfeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdagfheb
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v7); // cdaghfbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cdagfhbe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdagfbhe
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cdafgheb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdafghbe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cdafgbhe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdafbghe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v7, v3)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v6); // ghcadfeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); // gchadfeb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cghadfeb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gcahdfeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cgahdfeb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ghcadfbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gchadfbe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); // cghadfbe
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); // gcahdfbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v7); // cgahdfbe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v6, v7); // gcadhfeb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcadfheb
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v7); // gcadhfbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); // gcadfhbe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcadfbhe
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v4, v7); swap(v6, v7); // cgadhfeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgadfheb
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v4, v7); // cgadhfbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v4, v5); swap(v5, v7); // cgadfhbe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgadfbhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v3)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // caghdfeb
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v6, v7); // cagdhfeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cagdfheb
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); // caghdfbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v7); // cagdhfbe
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); // cagdfhbe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cagdfbhe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cadghfeb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cadgfheb
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); // cadghfbe
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cadgfhbe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cadgfbhe
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cadfgheb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cadfghbe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cadfgbhe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cadfbghe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v0)) {
														if (compare(v6, v2)) {
															if (compare(v7, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghcdabfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gchdabfe
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghcadbfe
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gchadbfe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gcahdbfe
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ghcabdfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gchabdfe
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcahbdfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gcabhdfe
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcdhabfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gcdahbfe
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcdabhfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gcdabfhe
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); // gcadhbfe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcadbhfe
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v6, v7); // gcadbfhe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcabdhfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gcabdfhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); // cghdabfe
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cgdhabfe
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdghabfe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); // cgdahbfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cdgahbfe
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgdabhfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v6, v7); // cgdabfhe
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdgabhfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cdgabfhe
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cghadbfe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cgahdbfe
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v4, v7); swap(v5, v6); // cgadhbfe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgadbhfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v4, v6); swap(v6, v7); // cgadbfhe
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cghabdfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgahbdfe
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cgabhdfe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgabdhfe
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cgabdfhe
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v6, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // caghdbfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); // cagdhbfe
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cagdbhfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v6, v7); // cagdbfhe
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // caghbdfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cagbhdfe
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cagbdhfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cagbdfhe
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v7, v1)) {
																		swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cdaghbfe
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdagbhfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cdagbfhe
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cadghbfe
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cadgbhfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cadgbfhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v3, v0)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdabghfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v4, v6); swap(v6, v7); // cdabgfhe
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdabfghe
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cadbghfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v6, v7); // cadbgfhe
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cadbfghe
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v3)) {
																		swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cabghdfe
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cabgdhfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cabgdfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cabdghfe
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v4, v6); swap(v6, v7); // cabdgfhe
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cabdfghe
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v0)) {
											if (compare(v7, v1)) {
												if (compare(v3, v0)) {
													if (compare(v7, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v2)) {
																if (compare(v6, v4)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); // gfehcdab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gfechdab
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); // gfecdhab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fgehcdab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgechdab
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fgecdhab
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // feghcdab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fegchdab
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fegcdhab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fecghdab
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fecgdhab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fecdghab
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gfcehdab
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v5, v7); // gfcedhab
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v5, v7); // gfcdehab
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgcehdab
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fgcedhab
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v5, v7); swap(v6, v7); // fgcdehab
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fcgehdab
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // fcgedhab
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v5, v7); swap(v6, v7); // fcgdehab
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fcdgehab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fceghdab
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcegdhab
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcedghab
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcdeghab
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gcfehdab
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgfehdab
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cfgehdab
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfeghdab
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); // gcfedhab
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cgfedhab
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v7); // cfgedhab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cfegdhab
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v7); // cfedghab
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v5, v7); // gcfdehab
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v5, v7); swap(v6, v7); // cgfdehab
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v5, v7); // cfgdehab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v7); // cfdgehab
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cfdeghab
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); // gcdfehab
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cgdfehab
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v7); // cdgfehab
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cdfgehab
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdfeghab
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gfecdahb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v6, v7); // fgecdahb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fegcdahb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gfcedahb
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v6, v7); // fgcedahb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v5, v6); swap(v6, v7); // gfcdeahb
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v6, v7); // fgcdeahb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfcdaehb
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v4, v5); swap(v6, v7); // fgcdaehb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fecgdahb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v4, v6); swap(v6, v7); // fecdgahb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v6, v7); // fcgedahb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fcegdahb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v6, v7); // fcedgahb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v6, v7); // fcgdeahb
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v6, v7); // fcdgeahb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fcdegahb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v6, v7); // fcgdaehb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fcdgaehb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gcfedahb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v5, v6); swap(v6, v7); // gcfdeahb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcfdaehb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v6, v7); // cgfedahb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v6, v7); // cgfdeahb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v6, v7); // cgfdaehb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cfgedahb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfegdahb
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v5, v6); swap(v6, v7); // cfgdeahb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfgdaehb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v4, v0)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cfdgeahb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfdgaehb
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfedgahb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfdegahb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gcdfeahb
																		}
																		else {
																			if (compare(v5, v0)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcdfaehb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gcdafehb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v6, v7); // cgdfeahb
																		}
																		else {
																			if (compare(v5, v0)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cgdfaehb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v6, v7); // cgdafehb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cdgfeahb
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v6, v7); // cdfgeahb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cdfegahb
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdgfaehb
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cdfgaehb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cdgafehb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v2)) {
														if (compare(v6, v5)) {
															if (compare(v5, v2)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gfehcadb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfechadb
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); // gfehcabd
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gfechabd
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gfecahdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfecadhb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfecahbd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfcehadb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gfcehabd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				if (compare(v7, v3)) {
																					swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gfceahdb
																				}
																				else {
																					swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfceadhb
																				}
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfceahbd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v5, v7); // gfcaehdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gfcaehbd
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gfcaedhb
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfcadehb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcfehadb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gcfehabd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gcfeahdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcfeahbd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); // gcfaehdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gcfaehbd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gcafehdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); // gcafehbd
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcfeadhb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gcfaedhb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcfadehb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v6, v7); // gcafedhb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gcafdehb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v6, v7); // gcadfehb
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v6, v7); // fgehcadb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // feghcadb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fgechadb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); // fegchadb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fgcehadb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); // fgehcabd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); // feghcabd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); // fgechabd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fegchabd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v7); // fgcehabd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v4)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgecahdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fgecadhb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fgecahbd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fegcahdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fegcadhb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fegcahbd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgceahdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fgceadhb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fgceahbd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fgcaehdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v5, v7); // fgcaehbd
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v6, v7); // fgcaedhb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fgcadehb
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v6, v4)) {
																if (compare(v4, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fcgehadb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); // fcgehabd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcgeahdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fcgeadhb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fcgeahbd
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fcgaehdb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); // fcgaehbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v6, v7); // fcgaedhb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fcgadehb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v7); // fecghadb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fecghabd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fecgahdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fecgadhb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fecgahbd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); // fceghadb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fceghabd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fcegahdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcegadhb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fcegahbd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cgfehadb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); // cgfehabd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgfeahdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cgfeahbd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cgfaehdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); // cgfaehbd
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cgafehdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v5, v7); // cgafehbd
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cgfeadhb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v6, v7); // cgfaedhb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cgfadehb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v6, v7); // cgafedhb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cgafdehb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v4, v5); swap(v6, v7); // cgadfehb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cfgehadb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cfeghadb
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				if (compare(v6, v4)) {
																					swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cfgeahdb
																				}
																				else {
																					swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); // cfegahdb
																				}
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); // cfgaehdb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfgeadhb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cfegadhb
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cfgaedhb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfgadehb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cfgehabd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cfeghabd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfgeahbd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cfegahbd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cfgaehbd
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v4, v3)) {
													if (compare(v5, v2)) {
														if (compare(v3, v1)) {
															if (compare(v3, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // gfecdabh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); // fgecdabh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // fegcdabh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v6); // fecgdabh
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v4, v6); // fecdgabh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v5, v6); // gfcedabh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v3, v4); // fgcedabh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); // fcgedabh
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // fcegdabh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v6); // fcedgabh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gfecadbh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // fgecadbh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fegcadbh
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fecgadbh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gfceadbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v5); // fgceadbh
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // fcgeadbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fcegadbh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v5, v6); // gfcaedbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); // fgcaedbh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); // fcgaedbh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gfecabhd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgecabhd
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fegcabhd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fecgabhd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gfceabhd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgceabhd
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcgeabhd
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fcegabhd
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v6, v7); // gfcaebhd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fgcaebhd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fcgaebhd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gfcabehd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgcabehd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcgabehd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // gfecabdh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fgecabdh
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // fegcabdh
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v3, v6); swap(v4, v5); // fecgabdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v4); swap(v4, v6); // gfceabdh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fgceabdh
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fcgeabdh
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // fcegabdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); // gfcaebdh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v5, v6); // fgcaebdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // fcgaebdh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v5); // gfcabedh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fgcabedh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fcgabedh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v0)) {
															if (compare(v3, v1)) {
																if (compare(v3, v0)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // gcfedabh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); // cgfedabh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v5, v6); // cfgedabh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cfegdabh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v4, v6); swap(v5, v6); // cfedgabh
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gcfeadbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cgfeadbh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cfgeadbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // cfegadbh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gcfeabhd
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgfeabhd
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cfgeabhd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v6, v7); // cfegabhd
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // gcfeabdh
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cgfeabdh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // cfgeabdh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v6); // cfegabdh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // gcfaedbh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); // cgfaedbh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // cfgaedbh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v6, v7); // gcfaebhd
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cgfaebhd
																				}
																				else {
																					swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v6, v7); // cfgaebhd
																				}
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); // gcfaebdh
																			}
																			else {
																				if (compare(v6, v5)) {
																					swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // cgfaebdh
																				}
																				else {
																					swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); // cfgaebdh
																				}
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gcfabehd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgfabehd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cfgabehd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // gcfabedh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cgfabedh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // cfgabedh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); // gcafedbh
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gcafebhd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // gcafebdh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcafbehd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gcafbedh
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gcabfehd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // gcabfedh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); // cgafedbh
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cgafebhd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v5, v6); // cgafebdh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgafbehd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cgafbedh
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cgabfehd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v5); // cgabfedh
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v6, v5)) {
															if (compare(v6, v2)) {
																if (compare(v3, v0)) {
																	if (compare(v4, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v5, v6); // gfcdeabh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v5, v6); // gcfdeabh
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // gcdfeabh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v4, v6); swap(v5, v6); // gfcdaebh
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // gcfdaebh
																			}
																		}
																		else {
																			if (compare(v5, v0)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gcdfaebh
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // gcdafebh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gfcadebh
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gcfadebh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // gcafdebh
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v5); // gcadfebh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v0)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); // cgfdeabh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); // cgdfeabh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v5, v6); // cdgfeabh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v3, v0)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v5); // cgfdaebh
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cgfadebh
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cgdfaebh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cdgfaebh
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v5); // cgdafebh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); // cdgafebh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v5); // cgafdebh
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v4, v5); // cgadfebh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v6); // fgcdeabh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); // fcgdeabh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); // fcdgeabh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // fcdegabh
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v5, v6); // cfgdeabh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v5, v6); // cfdgeabh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cfdegabh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); // cdfgeabh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v6); // cdfegabh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v4, v5); // fgcdaebh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v5); // fcgdaebh
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // fcdgaebh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v5); // fgcadebh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // fcgadebh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v6); swap(v5, v6); // cfgdaebh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cfgadebh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cfdgaebh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); // cdfgaebh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v5, v3)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v4, v6); // gfcdabeh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v4, v5); swap(v5, v6); // fgcdabeh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // fcgdabeh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fcdgabeh
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v4, v6); // gcfdabeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v4, v5); swap(v5, v6); // cgfdabeh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v4, v6); // cfgdabeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v6); swap(v4, v6); // cfdgabeh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // gcdfabeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cgdfabeh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v5); swap(v4, v6); // cdgfabeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cdfgabeh
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gcdafbeh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v5); swap(v5, v6); // cgdafbeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cdgafbeh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // gcdabfeh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v6); // cgdabfeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); // cdgabfeh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v6); // gfcadbeh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fgcadbeh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fcgadbeh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // gcfadbeh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cgfadbeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // cfgadbeh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gfcabdeh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v3, v5); swap(v4, v6); // fgcabdeh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // fcgabdeh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gcfabdeh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // cgfabdeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cfgabdeh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gcafdbeh
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // gcadfbeh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v6); // gcadbfeh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cgafdbeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v4, v5); swap(v5, v6); // cgadfbeh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v4, v6); // cgadbfeh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // gcafbdeh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gcabfdeh
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // gcabdfeh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v5); swap(v4, v6); // cgafbdeh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cgabfdeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v6); // cgabdfeh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v3)) {
												if (compare(v6, v4)) {
													if (compare(v4, v1)) {
														if (compare(v7, v1)) {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); // fcagehdb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fcagehbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fcagedhb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcagdehb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v5, v7); // cfagehdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cfagehbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cfagedhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfagdehb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cagfehdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); // cafgehdb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); // cagfehbd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cafgehbd
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v6, v7); // cagfedhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cafgedhb
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cagfdehb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cafgdehb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v6, v7); // cagdfehb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // fcagedbh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fcagdebh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v5, v6); // cfagedbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cfagdebh
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v6, v7); // fcagebhd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); // fcagebdh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v6, v7); // cfagebhd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); // cfagebdh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); // cagfedbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // cafgedbh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // cagfdebh
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cafgdebh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v5); // cagdfebh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cagfebhd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v6, v7); // cafgebhd
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // cagfebdh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); // cafgebdh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fcagbehd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cfagbehd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cagfbehd
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cafgbehd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cagbfehd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // fcagbedh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v5); // cfagbedh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cagfbedh
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // cafgbedh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // cagbfedh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcabgehd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfabgehd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cafbgehd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cabgfehd
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cabfgehd
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fcabgedh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cfabgedh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cafbgedh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v5); // cabgfedh
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cabfgedh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v5, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cagfdbeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // cagfbdeh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cagbfdeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cabgfdeh
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // cagdfbeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v6); // cagdbfeh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // cagbdfeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v6); // cabgdfeh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // fcagdbeh
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fcagbdeh
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // fcabgdeh
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v6); // cfagdbeh
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cfagbdeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v6); // cfabgdeh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // cafgdbeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cafgbdeh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // cafbgdeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v6); // cabfgdeh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v3)) {
														if (compare(v5, v2)) {
															if (compare(v7, v1)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fecaghdb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fceaghdb
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fcaeghdb
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fecaghbd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fceaghbd
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fcaeghbd
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fecagbhd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fceagbhd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fecabghd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fceabghd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fcaegbhd
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcaebghd
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fcabeghd
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cfeaghdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfeaghbd
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cfeagbhd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfeabghd
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cfaeghdb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cfaeghbd
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cfaegbhd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cfaebghd
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cfabeghd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v3, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cafeghdb
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cafeghbd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cafegbhd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cafebghd
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cafbeghd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cabfeghd
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v3, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fecagdhb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fceagdhb
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcaegdhb
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v6); // fecagdbh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // fceagdbh
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fcaegdbh
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fecagbdh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fceagbdh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v5, v6); // fecabgdh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // fceabgdh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // fcaegbdh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fcaebgdh
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // fcabegdh
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v0)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfeagdhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cfeagdbh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // cfeagbdh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cfeabgdh
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfaegdhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cfaegdbh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v6); // cfaegbdh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cfaebgdh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v5, v6); // cfabegdh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cafegdhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cafegdbh
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // cafegbdh
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cafebgdh
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // cafbegdh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v5, v6); // cabfegdh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v1)) {
													if (compare(v3, v0)) {
														if (compare(v5, v3)) {
															if (compare(v4, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fecdaghb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcedaghb
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fcdeaghb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v5, v6); swap(v6, v7); // cfedaghb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cfdeaghb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v4, v5); swap(v5, v6); // fecdagbh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // fcedagbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fcdeagbh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v5, v6); // cfedagbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); swap(v5, v6); // cfdeagbh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcdagehb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fcdaeghb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fcdagebh
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // fcdaegbh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // fcdagbeh
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfdagehb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v5, v6); swap(v6, v7); // cfdaeghb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v6); swap(v5, v6); // cfdagebh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v5, v6); // cfdaegbh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v6); // cfdagbeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v0)) {
																		swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdfeaghb
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdfagehb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cdfaeghb
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cdfagebh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v6); // cdfagbeh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cdfeagbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v5, v6); // cdfaegbh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cdagfehb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v5); // cdagfebh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cdagfbeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v6); // cdagbfeh
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdafgehb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cdafeghb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cdafgebh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v5, v6); // cdafegbh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v6); // cdafgbeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v4, v3)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fecadghb
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fceadghb
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cfeadghb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // fcaedghb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cfaedghb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fecadgbh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fceadgbh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // cfeadgbh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // fcaedgbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v5, v6); // cfaedgbh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fcadgehb
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v5, v6); swap(v6, v7); // fcadeghb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // fcadgebh
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v5, v6); // fcadegbh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v6); // fcadgbeh
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cfadgehb
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v5, v6); swap(v6, v7); // cfadeghb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v4, v6); swap(v5, v6); // cfadgebh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v5, v6); // cfadegbh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v4, v6); // cfadgbeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cafedghb
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cafdgehb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v5, v6); swap(v6, v7); // cafdeghb
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // cafdgebh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v6); // cafdgbeh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // cafedgbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v5, v6); // cafdegbh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cadgfehb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // cadgfebh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cadgfbeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // cadgbfeh
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cadfgehb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cadfeghb
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cadfgebh
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // cadfegbh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // cadfgbeh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v5, v2)) {
															if (compare(v4, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v4, v5); // fecdabgh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v4, v5); // fcedabgh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v3, v5); swap(v4, v5); // fecadbgh
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // fceadbgh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); // fcaedbgh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v3, v5); // fecabdgh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v4); swap(v3, v5); // fceabdgh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fcaebdgh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); // fcabedgh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // fcdeabgh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); // fcdaebgh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fcdabgeh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // fcdabegh
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); // fcadebgh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // fcadbgeh
																			}
																			else {
																				swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v4, v5); // fcadbegh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fcabdgeh
																		}
																		else {
																			swap(v0, v5); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fcabdegh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v0)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); // cfedabgh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v3, v4); // cfdeabgh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cdfeabgh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); // cfdaebgh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); // cdfaebgh
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v5); swap(v5, v6); // cfdabgeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v2, v3); swap(v4, v5); // cfdabegh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cdfabgeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cdfabegh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); // cfeadbgh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v4); // cfaedbgh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); // cfadebgh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v5); swap(v4, v5); swap(v5, v6); // cfadbgeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v4, v5); // cfadbegh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cfeabdgh
																		}
																		else {
																			swap(v0, v2); swap(v1, v5); swap(v3, v4); swap(v4, v5); // cfaebdgh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v5); swap(v3, v5); // cfabedgh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cfabdgeh
																			}
																			else {
																				swap(v0, v2); swap(v1, v5); swap(v3, v5); swap(v4, v5); // cfabdegh
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); // cafedbgh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); // cafdebgh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // cafdbgeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v4, v5); // cafdbegh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // cafebdgh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); // cafbedgh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cafbdgeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cafbdegh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v0)) {
																	if (compare(v4, v1)) {
																		swap(v0, v2); swap(v1, v3); swap(v3, v5); // cdafebgh
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cdafbgeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v5); swap(v4, v5); // cdafbegh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); // cadfebgh
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cadfbgeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cadfbegh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v3, v0)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v4, v6); // cdabgfeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v4, v5); swap(v5, v6); // cdabfgeh
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v3); swap(v4, v5); // cdabfegh
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v6); // cadbgfeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // cadbfgeh
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v5); // cadbfegh
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v4, v3)) {
																		swap(v0, v2); swap(v1, v2); swap(v3, v5); // cabfedgh
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cabfdgeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v5); swap(v4, v5); // cabfdegh
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v4, v6); // cabdgfeh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v4, v5); swap(v5, v6); // cabdfgeh
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v4, v5); // cabdfegh
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v7, v6)) {
									if (compare(v7, v4)) {
										if (compare(v4, v0)) {
											if (compare(v6, v0)) {
												if (compare(v3, v0)) {
													if (compare(v5, v0)) {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v6, v7); // hgefcdab
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hegfcdab
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hefgcdab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hefcgdab
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hefcdgab
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hgecfdab
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hegcfdab
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hecgfdab
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hecfgdab
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hecfdgab
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hgcefdab
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hcgefdab
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcegfdab
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hcefgdab
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcefdgab
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); // chgefdab
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // chegfdab
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); // chefgdab
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // chefdgab
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v6, v7); // hgecdfab
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v6, v7); // hgcedfab
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v6, v7); // hcgedfab
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); // chgedfab
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hegcdfab
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hecgdfab
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hcegdfab
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); // chegdfab
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v4, v6); swap(v6, v7); // hecdgfab
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v6, v7); // hcedgfab
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v6); // chedgfab
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hecdfgab
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcedfgab
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v5, v6); // chedfgab
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v4)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v6, v7); // hgcdefab
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v6, v7); // hcgdefab
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v6, v7); // hcdgefab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); // chgdefab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); // chdgefab
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v6, v7); // cdhgefab
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hcdegfab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); // chdegfab
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cdhegfab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcdefgab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // chdefgab
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdhefgab
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v4)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hgecdafb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hgcedafb
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v5, v7); swap(v6, v7); // hgcdeafb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hcgedafb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v5, v7); swap(v6, v7); // hcgdeafb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hcdgeafb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); // chgedafb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v5, v6); // chgdeafb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); // chdgeafb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cdhgeafb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hegcdafb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hecgdafb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hecdgafb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcegdafb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcedgafb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcdegafb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // chegdafb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); // chedgafb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // chdegafb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdhegafb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); // hgecdabf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v5, v7); // hgcedabf
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v5, v7); // hgcdeabf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); // hcgedabf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v5, v7); // hcgdeabf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); // hcdgeabf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // chgedabf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v5, v6); swap(v6, v7); // chgdeabf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // chdgeabf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v5, v7); // cdhgeabf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hegcdabf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hecgdabf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v4, v6); swap(v5, v7); // hecdgabf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hcegdabf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); // hcedgabf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hcdegabf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chegdabf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chedgabf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chdegabf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cdhegabf
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v4, v2)) {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hgefcadb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hegfcadb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hefgcadb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hefcgadb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgecfadb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hegcfadb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hecgfadb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hecfgadb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hgecafdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); // hegcafdb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v7); // hecgafdb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgecadfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hegcadfb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hecgadfb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hgecadbf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hegcadbf
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hecgadbf
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v6, v4)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgcefadb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcgefadb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // chgefadb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hgceafdb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hcgeafdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); // chgeafdb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hcegfadb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcefgadb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); // hcegafdb
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); // chegfadb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // chefgadb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); // chegafdb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgceadfb
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcgeadfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hcegadfb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // chgeadfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); // chegadfb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hgceadbf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hcgeadbf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcegadbf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chgeadbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // chegadbf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); // hgefcabd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hegfcabd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hefgcabd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hefcgabd
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hgecfabd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hegcfabd
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hecgfabd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hecfgabd
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); // hgecafbd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hegcafbd
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hecgafbd
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hgecabfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hegcabfd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hecgabfd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgecabdf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hegcabdf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hecgabdf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v4)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hgcefabd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hcgefabd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chgefabd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v7); // hgceafbd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); // hcgeafbd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // chgeafbd
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcegfabd
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hcefgabd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hcegafbd
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // chegfabd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chefgabd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v6, v7); // chegafbd
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hgceabfd
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hcgeabfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcegabfd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chgeabfd
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // chegabfd
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgceabdf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcgeabdf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hcegabdf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // chgeabdf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); // chegabdf
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v1)) {
													if (compare(v6, v5)) {
														if (compare(v5, v1)) {
															if (compare(v6, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hecagfdb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hceagfdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); // cheagfdb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hecagdfb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hceagdfb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cheagdfb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v6); // hecagfbd
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); // hceagfbd
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cheagfbd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v4, v7); swap(v5, v6); // hecdagfb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); // hcedagfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hcdeagfb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hecadgfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hceadgfb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v5, v6); // chedagfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); // cheadgfb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); // chdeagfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cdheagfb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hecagdbf
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hceagdbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cheagdbf
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hecagbfd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hceagbfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cheagbfd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hecagbdf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hceagbdf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cheagbdf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hecdagbf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcedagbf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcdeagbf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hecadgbf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hceadgbf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v5, v6); swap(v6, v7); // chedagbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cheadgbf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // chdeagbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdheagbf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v3, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hefcdagb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hecfdagb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v4, v5); swap(v5, v7); // hecdfagb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hcefdagb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); // hcedfagb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hcdefagb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); // chefdagb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v5); // chedfagb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); // chdefagb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cdhefagb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hefcagdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); // hefcadgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hefcagbd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hecfagdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v7); // hecfadgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hecfagbd
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hcefagdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); // hcefadgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcefagbd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); // chefagdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); // chefadgb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // chefagbd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hecafgdb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hceafgdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cheafgdb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hecafdgb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hceafdgb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cheafdgb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hecafgbd
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hceafgbd
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cheafgbd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v4, v7); // hecdafgb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v7); // hcedafgb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); // hcdeafgb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v7); // hecadfgb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); // hceadfgb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); // chedafgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); // cheadfgb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); // chdeafgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); // cdheafgb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v5, v0)) {
															if (compare(v3, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hefcdabg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hecfdabg
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hecdfabg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcefdabg
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcedfabg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcdefabg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // chefdabg
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v4, v5); swap(v6, v7); // chedfabg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // chdefabg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdhefabg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hefcadbg
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hefcabgd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hefcabdg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hecfadbg
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hecfabgd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hecfabdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hcefadbg
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hcefabgd
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcefabdg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v6, v7); // chefadbg
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); // chefabgd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // chefabdg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hecafdbg
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hceafdbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cheafdbg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v5); // hecafbgd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); // hceafbgd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cheafbgd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hecafbdg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hceafbdg
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cheafbdg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v4, v7); swap(v6, v7); // hecdafbg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v6, v7); // hcedafbg
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hcdeafbg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hecadfbg
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hceadfbg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v6, v7); // chedafbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v6, v7); // cheadfbg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v6, v7); // chdeafbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cdheafbg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v6, v5)) {
																if (compare(v7, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v4, v7); swap(v5, v7); // hecdabgf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); // hcedabgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hcdeabgf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hecadbgf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hceadbgf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v5, v7); // chedabgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); // cheadbgf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); // chdeabgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cdheabgf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hecdabfg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcedabfg
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcdeabfg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hecadbfg
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hceadbfg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v5, v7); swap(v6, v7); // chedabfg
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cheadbfg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // chdeabfg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdheabfg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v5, v6); // hecabgfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); // hecabfgd
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v6, v7); // hecabfdg
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); // hceabgfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); // hceabfgd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v6, v7); // hceabfdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cheabgfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); // cheabfgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cheabfdg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hecabgdf
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v5, v7); // hecabdgf
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hecabdfg
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hceabgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); // hceabdgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hceabdfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cheabgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cheabdgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cheabdfg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v3)) {
												if (compare(v5, v3)) {
													if (compare(v5, v1)) {
														if (compare(v3, v1)) {
															if (compare(v6, v5)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v6, v7); // hgcaefdb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v6, v7); // hcgaefdb
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); // hcagefdb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); // hcaegfdb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); // chgaefdb
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); // chagefdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v6); // chaegfdb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); // cahgefdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); // cahegfdb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hcaefgdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); // hcaefdgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hcaefdbg
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // chaefgdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v5); // chaefdgb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // chaefdbg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cahefgdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); // cahefdgb
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cahefdbg
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); // hgcaefbd
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); // hcgaefbd
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v6, v7); // hcagefbd
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hcaegfbd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v6, v7); // chgaefbd
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v6, v7); // chagefbd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // chaegfbd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v6, v7); // cahgefbd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cahegfbd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hcaefgbd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hcaefbgd
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hcaefbdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // chaefgbd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // chaefbgd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // chaefbdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cahefgbd
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cahefbgd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cahefbdg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v5, v6); // hgcaebfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); // hcgaebfd
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hcagebfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hcaegbfd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // chgaebfd
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // chagebfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // chaegbfd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cahgebfd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cahegbfd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hcaebgfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); // hcaebfgd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hcaebfdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // chaebgfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v7); // chaebfgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // chaebfdg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cahebgfd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); // cahebfgd
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cahebfdg
																	}
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hgcabefd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hcgabefd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chgabefd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcagbefd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chagbefd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cahgbefd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hcabgefd
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // chabgefd
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cahbgefd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cabhgefd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); // hcabegfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); // hcabefgd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v6, v7); // hcabefdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v5, v6); // chabegfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v7); // chabefgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v6, v7); // chabefdg
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); // cahbegfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); // cahbefgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v6, v7); // cahbefdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v5, v6); // cabhegfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v7); // cabhefgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v6, v7); // cabhefdg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v5, v1)) {
															if (compare(v6, v3)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hgcaedfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hcgaedfb
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); // hcagedfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hcaegdfb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); // chgaedfb
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v5, v6); // chagedfb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // chaegdfb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); // cahgedfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cahegdfb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); // hcaedgfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); // hcaedfgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v6, v7); // hcaedfbg
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v5, v6); // chaedgfb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); // chaedfgb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v6, v7); // chaedfbg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); // cahedgfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); // cahedfgb
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v6, v7); // cahedfbg
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v5, v7); // hgcaedbf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); // hcgaedbf
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hcagedbf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcaegdbf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // chgaedbf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // chagedbf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chaegdbf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cahgedbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cahegdbf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hcaedgbf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); // hcaedbgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hcaedbfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // chaedgbf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v5, v7); // chaedbgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // chaedbfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cahedgbf
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); // cahedbgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cahedbfg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hgcaebdf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hcgaebdf
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); // hcagebdf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hcaegbdf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); // chgaebdf
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v5, v7); // chagebdf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // chaegbdf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); // cahgebdf
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cahegbdf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcaebgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hcaebdgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcaebdfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chaebgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // chaebdgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chaebdfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cahebgdf
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cahebdgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cahebdfg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgcabedf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcgabedf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // chgabedf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hcagbedf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // chagbedf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cahgbedf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcabgedf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chabgedf
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cahbgedf
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cabhgedf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hcabegdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); // hcabedgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hcabedfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // chabegdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v5, v7); // chabedgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // chabedfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cahbegdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); // cahbedgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cahbedfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cabhegdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v5, v7); // cabhedgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cabhedfg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v4)) {
													if (compare(v4, v1)) {
														if (compare(v3, v0)) {
															if (compare(v5, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgcdaefb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcgdaefb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcdgaefb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); // chgdaefb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // chdgaefb
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdhgaefb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hcdagefb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); // chdagefb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cdhagefb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cdahgefb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v4, v7); swap(v5, v7); // hgcdaebf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); // hcgdaebf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hcdgaebf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chgdaebf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chdgaebf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cdhgaebf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcdagebf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chdagebf
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdhagebf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdahgebf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgcadefb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcgadefb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hcagdefb
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); // hcadgefb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // chgadefb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // chagdefb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v4, v6); swap(v5, v6); // chadgefb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cahgdefb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); // cahdgefb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cadhgefb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hgcadebf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hcgadebf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcagdebf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hcadgebf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chgadebf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chagdebf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // chadgebf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cahgdebf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cahdgebf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cadhgebf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v6, v0)) {
																if (compare(v3, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgcdabef
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcgdabef
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcdgabef
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v4, v6); swap(v5, v7); // chgdabef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // chdgabef
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdhgabef
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgcadbef
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcgadbef
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // chgadbef
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hgcabdef
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hcgabdef
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chgabdef
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hcagdbef
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // chagdbef
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cahgdbef
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcagbdef
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chagbdef
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cahgbdef
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hcdagbef
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v6); swap(v5, v7); // chdagbef
																			}
																		}
																		else {
																			if (compare(v7, v0)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cdhagbef
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cdahgbef
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); // hcadgbef
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v4, v6); swap(v5, v7); // chadgbef
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); // cahdgbef
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cadhgbef
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcdabgef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chdabgef
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdhabgef
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcadbgef
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chadbgef
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hcabgdef
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hcabdgef
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // chabgdef
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // chabdgef
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdahbgef
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdabhgef
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cahdbgef
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cadhbgef
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cadbhgef
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cahbgdef
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cahbdgef
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cabhgdef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cabhdgef
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cabdhgef
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v5)) {
														if (compare(v3, v0)) {
															if (compare(v6, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); // hcdaegfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v5, v6); // chdaegfb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); // cdhaegfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v5, v6); // cdahegfb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hcdaegbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v5, v6); swap(v6, v7); // chdaegbf
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cdhaegbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cdahegbf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); // hcdaebgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v5, v7); // chdaebgf
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); // cdhaebgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v5, v7); // cdahebgf
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hcdabegf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); // chdabegf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cdhabegf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cdahbegf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v4, v7); swap(v5, v7); // cdabhegf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v5, v6); // hcadegfb
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v5, v6); swap(v6, v7); // hcadegbf
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v5, v7); // hcadebgf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); // hcadbegf
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hcabdegf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v5, v6); // chadegfb
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v7); swap(v5, v6); swap(v6, v7); // chadegbf
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v5, v7); // chadebgf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v4, v7); swap(v5, v7); // chadbegf
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // chabdegf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v5, v6); // cahdegfb
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v5, v6); swap(v6, v7); // cahdegbf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v5, v7); // cahdebgf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); // cadhegfb
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cadhegbf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); // cadhebgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); // cahdbegf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cadhbegf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); // cadbhegf
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cahbdegf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cabhdegf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v4, v7); swap(v5, v7); // cabdhegf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v5, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); // hcdaefgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); // chdaefgb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); // cdhaefgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); // cdahefgb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v6, v7); // hcdaefbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v6, v7); // chdaefbg
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v6, v7); // cdhaefbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v6, v7); // cdahefbg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hcdaebfg
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v5, v7); swap(v6, v7); // chdaebfg
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cdhaebfg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cdahebfg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcdabefg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v7); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chdabefg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdhabefg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdahbefg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdabhefg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); // hcadefgb
																			}
																			else {
																				swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v6, v7); // hcadefbg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v5, v7); swap(v6, v7); // hcadebfg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcadbefg
																		}
																		else {
																			swap(v0, v7); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hcabdefg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v7); // chadefgb
																			}
																			else {
																				swap(v0, v2); swap(v1, v7); swap(v6, v7); // chadefbg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v5, v7); swap(v6, v7); // chadebfg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chadbefg
																		}
																		else {
																			swap(v0, v2); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // chabdefg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); // cahdefgb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v6, v7); // cahdefbg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v5, v7); swap(v6, v7); // cahdebfg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); // cadhefgb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v6, v7); // cadhefbg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cadhebfg
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cahdbefg
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cadhbefg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cadbhefg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cahbdefg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cabhdefg
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cabdhefg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v7, v0)) {
											if (compare(v5, v0)) {
												if (compare(v3, v0)) {
													if (compare(v6, v0)) {
														if (compare(v5, v3)) {
															if (compare(v7, v5)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); // ehgfcdab
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ehfgcdab
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); // ehfcgdab
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ehfcdgab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ehgcfdab
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ehcgfdab
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v6); // ehcfgdab
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ehcfdgab
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // echgfdab
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); // echfgdab
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // echfdgab
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cehgfdab
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cehfgdab
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cehfdgab
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efhgcdab
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // efhcgdab
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efhcdgab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // efchgdab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efchdgab
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efcdhgab
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ecfhgdab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecfhdgab
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecfdhgab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cefhgdab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cefhdgab
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v5, v6); // cefdhgab
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); // ehgcdfab
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v6); // ehcgdfab
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v4, v6); // ehcdgfab
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v4, v5); swap(v5, v6); // ehcdfgab
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); // echgdfab
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v6); // echdgfab
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); // echdfgab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); // ecdhgfab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ecdhfgab
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecdfhgab
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cehgdfab
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v6, v7); // cehdgfab
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cedhgfab
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); // cdehgfab
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cehdfgab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cedhfgab
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cedfhgab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cdehfgab
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdefhgab
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v7, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); // ehfcdagb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v5); // ehcfdagb
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v4, v5); // ehcdfagb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); // echfdagb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v5); // echdfagb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); // ecdhfagb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cehfdagb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); // cehdfagb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cedhfagb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); // cdehfagb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // efhcdagb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // efchdagb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v4, v7); swap(v5, v7); // efcdhagb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ecfhdagb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); // ecfdhagb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ecdfhagb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); // cefhdagb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v7); // cefdhagb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); // cedfhagb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cdefhagb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ehfcdabg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ehcfdabg
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v4, v5); swap(v6, v7); // ehcdfabg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // echfdabg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v6, v7); // echdfabg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ecdhfabg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cehfdabg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cehdfabg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cedhfabg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cdehfabg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efhcdabg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efchdabg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efcdhabg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecfhdabg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecfdhabg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecdfhabg
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cefhdabg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v7); swap(v6, v7); // cefdhabg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cedfhabg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdefhabg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v7, v2)) {
															if (compare(v6, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ehgfcadb
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); // ehgcfadb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); // ehfgcadb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // efhgcadb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ehfcgadb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efhcgadb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v5); // ehcgfadb
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ehcfgadb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); // ehfcagdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // efhcagdb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v5, v6); // ehcfagdb
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); // ehfcadgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); // efhcadgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v5); // ehcfadgb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v6, v7); // ehfcadbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // efhcadbg
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v6, v7); // ehcfadbg
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); // echgfadb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // echfgadb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); // echfagdb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cehgfadb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cehfgadb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cehfagdb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efchgadb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecfhgadb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cefhgadb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // efchagdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ecfhagdb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cefhagdb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v5, v7); // efchadgb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); // echfadgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); // ecfhadgb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); // cehfadgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cefhadgb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // efchadbg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v6, v7); // echfadbg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ecfhadbg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cehfadbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cefhadbg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v6, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehgfcabd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ehgcfabd
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ehfgcabd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efhgcabd
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehfcgabd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // efhcgabd
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ehcgfabd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehcfgabd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ehfcagbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); // efhcagbd
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ehcfagbd
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); // ehfcabgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); // efhcabgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v5, v7); // ehcfabgd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ehfcabdg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v6, v7); // efhcabdg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ehcfabdg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // echgfabd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // echfgabd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // echfagbd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cehgfabd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cehfgabd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cehfagbd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // efchgabd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ecfhgabd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); // cefhgabd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v5, v6); // efchagbd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); // ecfhagbd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cefhagbd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v7); // efchabgd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); // echfabgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); // ecfhabgd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cehfabgd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); // cefhabgd
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v7); swap(v6, v7); // efchabdg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // echfabdg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v6, v7); // ecfhabdg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cehfabdg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cefhabdg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v6, v5)) {
														if (compare(v6, v0)) {
															if (compare(v3, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ehgcdafb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ehcgdafb
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v4, v6); swap(v5, v6); // ehcdgafb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // echgdafb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); // echdgafb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ecdhgafb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cehgdafb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cehdgafb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cedhgafb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cdehgafb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); // ehgcafdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); // ehgcadfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v6, v7); // ehgcafbd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v6); // ehcgafdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v5, v6); // ehcgadfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v6, v7); // ehcgafbd
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); // echgafdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); // echgadfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v6, v7); // echgafbd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); // cehgafdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cehgadfb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cehgafbd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v6); // ehcagfdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); // echagfdb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // cehagfdb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ehcagdfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // echagdfb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cehagdfb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ehcagfbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // echagfbd
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); // cehagfbd
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v5, v6); // ehcdagfb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v5, v6); // echdagfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); // ecdhagfb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v5, v6); // ehcadgfb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); // echadgfb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); // cehdagfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cehadgfb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // cedhagfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); // cdehagfb
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ehcafgdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // echafgdb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cehafgdb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v5); // ehcafdgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); // echafdgb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // cehafdgb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehcafgbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // echafgbd
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // cehafgbd
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); // ehcdafgb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); // echdafgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); // ecdhafgb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v4); // ehcadfgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); // echadfgb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v7); // cehdafgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); // cehadfgb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); // cedhafgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); // cdehafgb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ehcafdbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // echafdbg
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cehafdbg
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ehcafbgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // echafbgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); // cehafbgd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehcafbdg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // echafbdg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cehafbdg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v6, v7); // ehcdafbg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v6, v7); // echdafbg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v6, v7); // ecdhafbg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v6, v7); // ehcadfbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v6, v7); // echadfbg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v6, v7); // cehdafbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cehadfbg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cedhafbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v6, v7); // cdehafbg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v6, v0)) {
															if (compare(v3, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehgcdabf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehcgdabf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehcdgabf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // echgdabf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // echdgabf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecdhgabf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cehgdabf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v6); swap(v5, v7); // cehdgabf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cedhgabf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdehgabf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ehgcadbf
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ehgcabfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); // ehgcabdf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ehcgadbf
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ehcgabfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v3, v6); swap(v5, v7); // ehcgabdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // echgadbf
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // echgabfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); // echgabdf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cehgadbf
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cehgabfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cehgabdf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehcagdbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // echagdbf
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cehagdbf
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehcagbfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // echagbfd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cehagbfd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ehcagbdf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // echagbdf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cehagbdf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v5, v6); swap(v6, v7); // ehcdagbf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v5, v6); swap(v6, v7); // echdagbf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ecdhagbf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ehcadgbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // echadgbf
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cehdagbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cehadgbf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cedhagbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cdehagbf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v6, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v5, v7); // ehcdabgf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v5, v7); // echdabgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); // ecdhabgf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v5, v7); // ehcadbgf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); // echadbgf
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); // cehdabgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cehadbgf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cedhabgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); // cdehabgf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v5, v7); swap(v6, v7); // ehcdabfg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v5, v7); swap(v6, v7); // echdabfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ecdhabfg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ehcadbfg
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // echadbfg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cehdabfg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cehadbfg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cedhabfg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cdehabfg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ehcabgfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v7); // ehcabfgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ehcabfdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // echabgfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); // echabfgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // echabfdg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); // cehabgfd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); // cehabfgd
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v6, v7); // cehabfdg
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehcabgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ehcabdgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehcabdfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // echabgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // echabdgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // echabdfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cehabgdf
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); // cehabdgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cehabdfg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v3)) {
												if (compare(v6, v3)) {
													if (compare(v6, v1)) {
														if (compare(v3, v1)) {
															if (compare(v6, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ecahgfdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecahgdfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ecahgdbf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ceahgfdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ceahgdfb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ceahgdbf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // caehgfdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // caehgdfb
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // caehgdbf
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efcahgdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecfahgdb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecahfgdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ecafhgdb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cefahgdb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ceahfgdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ceafhgdb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // caehfgdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // caefhgdb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); // ecahgfbd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ecahgbfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecahgbdf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v6); // ceahgfbd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ceahgbfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ceahgbdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); // caehgfbd
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // caehgbfd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // caehgbdf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // efcahgbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ecfahgbd
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ecahfgbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecafhgbd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cefahgbd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ceahfgbd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ceafhgbd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // caehfgbd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // caefhgbd
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v7); // efcahbgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); // ecfahbgd
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); // ecahfbgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ecafhbgd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cefahbgd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v5); // ceahfbgd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ceafhbgd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); // caehfbgd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // caefhbgd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); // ecahbgfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); // ecahbfgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ecahbgdf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v5, v6); // ceahbgfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); // ceahbfgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ceahbgdf
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); // caehbgfd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); // caehbfgd
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // caehbgdf
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ecabhgfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); // ecabhfgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecabhgdf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ceabhgfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v7); // ceabhfgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ceabhgdf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // caebhgfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); // caebhfgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // caebhgdf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cabehgfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v7); // cabehfgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cabehgdf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // efcabhgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ecfabhgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); // cefabhgd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); // ecafbhgd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v5, v7); // ceafbhgd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); // caefbhgd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ecabfhgd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ceabfhgd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // caebfhgd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cabefhgd
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v6, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // efcahdgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ecfahdgb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ecahfdgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); // ecafhdgb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); // cefahdgb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ceahfdgb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v7); // ceafhdgb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // caehfdgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); // caefhdgb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ecahdgfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); // ecahdfgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecahdgbf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ceahdgfb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v7); // ceahdfgb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ceahdgbf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // caehdgfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); // caehdfgb
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // caehdgbf
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efcahdbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecfahdbg
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecahfdbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ecafhdbg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cefahdbg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ceahfdbg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ceafhdbg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // caehfdbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // caefhdbg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ecahdfbg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ecahdbgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecahdbfg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ceahdfbg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ceahdbgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ceahdbfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // caehdfbg
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // caehdbgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // caehdbfg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // efcahbdg
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ecfahbdg
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ecahfbdg
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecafhbdg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cefahbdg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ceahfbdg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ceafhbdg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // caehfbdg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // caefhbdg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v6, v7); // ecahbfdg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); // ecahbdgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ecahbdfg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v6, v7); // ceahbfdg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v5, v7); // ceahbdgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ceahbdfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v6, v7); // caehbfdg
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); // caehbdgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // caehbdfg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ecabhfdg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ecabhdgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecabhdfg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ceabhfdg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ceabhdgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ceabhdfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // caebhfdg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // caebhdgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // caebhdfg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cabehfdg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cabehdgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cabehdfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efcabhdg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecfabhdg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cefabhdg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ecafbhdg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ceafbhdg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // caefbhdg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecabfhdg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ceabfhdg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // caebfhdg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cabefhdg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v5)) {
													if (compare(v6, v5)) {
														if (compare(v3, v0)) {
															if (compare(v6, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ecdahgfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); // cedahgfb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cdeahgfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cdaehgfb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecdahgbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cedahgbf
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdeahgbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdaehgbf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ecdahbgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); // cedahbgf
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cdeahbgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cdaehbgf
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); // ecdabhgf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v5, v7); // cedabhgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); // cdeabhgf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v5, v7); // cdaebhgf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v5, v7); // cdabehgf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); // ecadhgfb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecadhgbf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); // ecadhbgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v5, v7); // ecadbhgf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); // ecabdhgf
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v4, v7); swap(v5, v6); // ceadhgfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ceadhgbf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v4); swap(v4, v7); swap(v5, v7); // ceadhbgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v5, v7); // ceadbhgf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v5, v7); // ceabdhgf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); // caedhgfb
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // caedhgbf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); // caedhbgf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cadehgfb
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cadehgbf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cadehbgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v5, v7); // caedbhgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); // cadebhgf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v5, v7); // cadbehgf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); // caebdhgf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v5, v7); // cabedhgf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v5, v7); // cabdehgf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v5, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); // ecdahfgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v7); // cedahfgb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); // cdeahfgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v7); // cdaehfgb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ecdahfbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v6, v7); // cedahfbg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cdeahfbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cdaehfbg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecdahbfg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cedahbfg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdeahbfg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdaehbfg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ecdabhfg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v5, v7); swap(v6, v7); // cedabhfg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cdeabhfg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cdaebhfg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v5, v7); swap(v6, v7); // cdabehfg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v7); // ecadhfgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v6, v7); // ecadhfbg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecadhbfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v5, v7); swap(v6, v7); // ecadbhfg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ecabdhfg
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v4, v7); // ceadhfgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v4, v7); swap(v6, v7); // ceadhfbg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ceadhbfg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v5, v7); swap(v6, v7); // ceadbhfg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ceabdhfg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v7); // caedhfgb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v6, v7); // caedhfbg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // caedhbfg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); // cadehfgb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cadehfbg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cadehbfg
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v5, v7); swap(v6, v7); // caedbhfg
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cadebhfg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v5, v7); swap(v6, v7); // cadbehfg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // caebdhfg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cabedhfg
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v5, v7); swap(v6, v7); // cabdehfg
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v3, v0)) {
															if (compare(v6, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v5, v7); // efcdahgb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v5, v7); // ecfdahgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); // ecdfahgb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); // cefdahgb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cedfahgb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); // cdefahgb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ecdafhgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); // cedafhgb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cdeafhgb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cdaefhgb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v5, v7); swap(v6, v7); // efcdahbg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v5, v7); swap(v6, v7); // ecfdahbg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ecdfahbg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cefdahbg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cedfahbg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cdefahbg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecdafhbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cedafhbg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdeafhbg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdaefhbg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v5, v7); // efcadhgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); // ecfadhgb
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ecafdhgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); // ecadfhgb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cefadhgb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ceafdhgb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v4, v5); swap(v5, v7); // ceadfhgb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // caefdhgb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); // caedfhgb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cadefhgb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // efcadhbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ecfadhbg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecafdhbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecadfhbg
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cefadhbg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ceafdhbg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ceadfhbg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // caefdhbg
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // caedfhbg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cadefhbg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																if (compare(v3, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v6, v7); // efcdabhg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v6, v7); // ecfdabhg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v6, v7); // ecdfabhg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v6, v7); // cefdabhg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cedfabhg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v6, v7); // cdefabhg
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v6, v7); // efcadbhg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v6, v7); // ecfadbhg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cefadbhg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // efcabdhg
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ecfabdhg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v6, v7); // cefabdhg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ecafdbhg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ceafdbhg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // caefdbhg
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v6, v7); // ecafbdhg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v6, v7); // ceafbdhg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v6, v7); // caefbdhg
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ecdafbhg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v6, v7); // cedafbhg
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cdeafbhg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cdaefbhg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v6, v7); // ecadfbhg
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v4, v5); swap(v6, v7); // ceadfbhg
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v6, v7); // caedfbhg
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cadefbhg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v6, v7); // ecdabfhg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v6, v7); // cedabfhg
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v6, v7); // cdeabfhg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v6, v7); // ecadbfhg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v6, v7); // ceadbfhg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ecabfdhg
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v6, v7); // ecabdfhg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ceabfdhg
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v6, v7); // ceabdfhg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v6, v7); // cdaebfhg
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v6, v7); // cdabefhg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v6, v7); // caedbfhg
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v6, v7); // cadebfhg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v6, v7); // cadbefhg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // caebfdhg
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v6, v7); // caebdfhg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cabefdhg
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v6, v7); // cabedfhg
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v6, v7); // cabdefhg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v4)) {
										if (compare(v4, v0)) {
											if (compare(v7, v0)) {
												if (compare(v3, v0)) {
													if (compare(v5, v0)) {
														if (compare(v5, v3)) {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); // ghefcdab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); // gehfcdab
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gefhcdab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); // gefchdab
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gefcdhab
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); // ghecfdab
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gehcfdab
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gechfdab
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v7); // gecfhdab
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gecfdhab
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v5); // ghcefdab
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); // gchefdab
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gcehfdab
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); // gcefhdab
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gcefdhab
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cghefdab
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgehfdab
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // cgefhdab
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgefdhab
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v7, v3)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); // ghecdfab
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v3, v4); // ghcedfab
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); // gchedfab
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v6, v7); // cghedfab
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); // gehcdfab
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v7); // gechdfab
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); // gcehdfab
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // cgehdfab
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v4, v7); // gecdhfab
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v7); // gcedhfab
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v6, v7); // cgedhfab
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v4, v5); swap(v5, v7); // gecdfhab
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); // gcedfhab
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgedfhab
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v4)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); // ghcdefab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); // gchdefab
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); // gcdhefab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v6, v7); // cghdefab
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v6, v7); // cgdhefab
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); // cdghefab
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); // gcdehfab
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cgdehfab
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); // cdgehfab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gcdefhab
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgdefhab
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cdgefhab
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v7, v4)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); // ghecdafb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v5, v6); // ghcedafb
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v5, v6); // ghcdeafb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); // gchedafb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v5, v6); // gchdeafb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); // gcdheafb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cghedafb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v5, v7); swap(v6, v7); // cghdeafb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cgdheafb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); // cdgheafb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gehcdafb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gechdafb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v4, v7); swap(v5, v6); // gecdhafb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gcehdafb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); // gcedhafb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gcdehafb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgehdafb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgedhafb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgdehafb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cdgehafb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v4)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ghecdabf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ghcedabf
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v5, v6); swap(v6, v7); // ghcdeabf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gchedabf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v5, v6); swap(v6, v7); // gchdeabf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gcdheabf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v5, v7); // cghedabf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v5, v7); // cghdeabf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v5, v7); // cgdheabf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cdgheabf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gehcdabf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gechdabf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gecdhabf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcehdabf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcedhabf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcdehabf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cgehdabf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v7); swap(v5, v7); // cgedhabf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cgdehabf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdgehabf
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v4, v2)) {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); // ghefcadb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gehfcadb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gefhcadb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gefchadb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ghecfadb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gehcfadb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gechfadb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gecfhadb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); // ghecafdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gehcafdb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gechafdb
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ghecadfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gehcadfb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gechadfb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghecadbf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gehcadbf
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gechadbf
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v7, v4)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ghcefadb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gchefadb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cghefadb
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v6); // ghceafdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); // gcheafdb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cgheafdb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcehfadb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gcefhadb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gcehafdb
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // cgehfadb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgefhadb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v6, v7); // cgehafdb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ghceadfb
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gcheadfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcehadfb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgheadfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cgehadfb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghceadbf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcheadbf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gcehadbf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cgheadbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); // cgehadbf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v5, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ghefcabd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gehfcabd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gefhcabd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gefchabd
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghecfabd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gehcfabd
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gechfabd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gecfhabd
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ghecafbd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); // gehcafbd
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v6); // gechafbd
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghecabfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gehcabfd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gechabfd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ghecabdf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gehcabdf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gechabdf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v7, v4)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghcefabd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gchefabd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cghefabd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ghceafbd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gcheafbd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); // cgheafbd
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gcehfabd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcefhabd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); // gcehafbd
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); // cgehfabd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // cgefhabd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); // cgehafbd
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghceabfd
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcheabfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gcehabfd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cgheabfd
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); // cgehabfd
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ghceabdf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gcheabdf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcehabdf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgheabdf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cgehabdf
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v1)) {
													if (compare(v7, v5)) {
														if (compare(v5, v1)) {
															if (compare(v7, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v7); // gecahfdb
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); // gceahfdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cgeahfdb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gecahdfb
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gceahdfb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgeahdfb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gecahfbd
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gceahfbd
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); // cgeahfbd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gecdahfb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcedahfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcdeahfb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gecadhfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gceadhfb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v5, v7); swap(v6, v7); // cgedahfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cgeadhfb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cgdeahfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdgeahfb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gecahdbf
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gceahdbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cgeahdbf
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gecahbfd
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gceahbfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cgeahbfd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gecahbdf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gceahbdf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgeahbdf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v4, v6); swap(v5, v7); // gecdahbf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); // gcedahbf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gcdeahbf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gecadhbf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gceadhbf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v5, v7); // cgedahbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); // cgeadhbf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v7); // cgdeahbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cdgeahbf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v3, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gefcdahb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gecfdahb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gecdfahb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcefdahb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcedfahb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcdefahb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // cgefdahb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v5); swap(v6, v7); // cgedfahb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cgdefahb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdgefahb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gefcahdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gefcadhb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gefcahbd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gecfahdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gecfadhb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gecfahbd
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcefahdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gcefadhb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gcefahbd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // cgefahdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v6, v7); // cgefadhb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); // cgefahbd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gecafhdb
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gceafhdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgeafhdb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gecafdhb
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gceafdhb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cgeafdhb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gecafhbd
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gceafhbd
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cgeafhbd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v4, v6); swap(v6, v7); // gecdafhb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v6, v7); // gcedafhb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gcdeafhb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gecadfhb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gceadfhb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v6, v7); // cgedafhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v6, v7); // cgeadfhb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v6, v7); // cgdeafhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cdgeafhb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v5, v0)) {
															if (compare(v3, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gefcdabh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gecfdabh
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v4, v5); swap(v5, v6); // gecdfabh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gcefdabh
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // gcedfabh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gcdefabh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v5); // cgefdabh
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v4, v5); // cgedfabh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cgdefabh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cdgefabh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // gefcadbh
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gefcabhd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gefcabdh
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v6); // gecfadbh
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gecfabhd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gecfabdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // gcefadbh
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcefabhd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gcefabdh
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); // cgefadbh
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cgefabhd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // cgefabdh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gecafdbh
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gceafdbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cgeafdbh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gecafbhd
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gceafbhd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgeafbhd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v5); // gecafbdh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // gceafbdh
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cgeafbdh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v4, v6); // gecdafbh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v6); // gcedafbh
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // gcdeafbh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v6); // gecadfbh
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // gceadfbh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); // cgedafbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); // cgeadfbh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); // cgdeafbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); // cdgeafbh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v7, v5)) {
																if (compare(v6, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gecdabhf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcedabhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcdeabhf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gecadbhf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gceadbhf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v5, v6); swap(v6, v7); // cgedabhf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cgeadbhf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cgdeabhf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdgeabhf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v4, v6); swap(v5, v6); // gecdabfh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // gcedabfh
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gcdeabfh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gecadbfh
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gceadbfh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v5, v6); // cgedabfh
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // cgeadbfh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v4); swap(v5, v6); // cgdeabfh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cdgeabfh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gecabhfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v6, v7); // gecabfhd
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v4); swap(v3, v6); // gecabfdh
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gceabhfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v6, v7); // gceabfhd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); // gceabfdh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgeabhfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cgeabfhd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // cgeabfdh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v5, v7); // gecabhdf
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gecabdhf
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v3, v6); swap(v5, v6); // gecabdfh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); // gceabhdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gceabdhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // gceabdfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cgeabhdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgeabdhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cgeabdfh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v3)) {
												if (compare(v5, v3)) {
													if (compare(v5, v1)) {
														if (compare(v3, v1)) {
															if (compare(v7, v5)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); // ghcaefdb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); // gchaefdb
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v6, v7); // gcahefdb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gcaehfdb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v6, v7); // cghaefdb
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v6, v7); // cgahefdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cgaehfdb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v6, v7); // caghefdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // cagehfdb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gcaefhdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gcaefdhb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // gcaefdbh
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cgaefhdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cgaefdhb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v5); // cgaefdbh
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cagefhdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // cagefdhb
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // cagefdbh
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v6, v7); // ghcaefbd
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v6, v7); // gchaefbd
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); // gcahefbd
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); // gcaehfbd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); // cghaefbd
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); // cgahefbd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v7); // cgaehfbd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); // caghefbd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); // cagehfbd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gcaefhbd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gcaefbhd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gcaefbdh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cgaefhbd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cgaefbhd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cgaefbdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // cagefhbd
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cagefbhd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cagefbdh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v7, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ghcaebfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gchaebfd
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); // gcahebfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gcaehbfd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); // cghaebfd
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v5, v6); // cgahebfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cgaehbfd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); // caghebfd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // cagehbfd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gcaebhfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gcaebfhd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // gcaebfdh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cgaebhfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cgaebfhd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v6); // cgaebfdh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cagebhfd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cagebfhd
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // cagebfdh
																	}
																}
															}
														}
														else {
															if (compare(v7, v4)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghcabefd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gchabefd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cghabefd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gcahbefd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cgahbefd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // caghbefd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcabhefd
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgabhefd
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cagbhefd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cabghefd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gcabehfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v6, v7); // gcabefhd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); // gcabefdh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cgabehfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v6, v7); // cgabefhd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v6); // cgabefdh
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cagbehfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v6, v7); // cagbefhd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); // cagbefdh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cabgehfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v6, v7); // cabgefhd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v6); // cabgefdh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v5, v1)) {
															if (compare(v7, v3)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v5, v6); // ghcaedfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); // gchaedfb
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gcahedfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcaehdfb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cghaedfb
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // cgahedfb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgaehdfb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // caghedfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cagehdfb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gcaedhfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v6, v7); // gcaedfhb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); // gcaedfbh
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cgaedhfb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v6, v7); // cgaedfhb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v4); // cgaedfbh
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // cagedhfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v6, v7); // cagedfhb
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); // cagedfbh
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ghcaedbf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gchaedbf
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); // gcahedbf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gcaehdbf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); // cghaedbf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v5, v7); // cgahedbf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cgaehdbf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); // caghedbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // cagehdbf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); // gcaedhbf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gcaedbhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // gcaedbfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v5, v7); // cgaedhbf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cgaedbhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v5, v6); // cgaedbfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); // cagedhbf
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cagedbhf
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // cagedbfh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v7, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v5, v7); // ghcaebdf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); // gchaebdf
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gcahebdf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcaehbdf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cghaebdf
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // cgahebdf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgaehbdf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // caghebdf
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cagehbdf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gcaebhdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcaebdhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gcaebdfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cgaebhdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgaebdhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cgaebdfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cagebhdf
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cagebdhf
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cagebdfh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v4)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ghcabedf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gchabedf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cghabedf
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcahbedf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgahbedf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // caghbedf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gcabhedf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cgabhedf
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cagbhedf
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cabghedf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); // gcabehdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gcabedhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // gcabedfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v5, v7); // cgabehdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cgabedhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v5, v6); // cgabedfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); // cagbehdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cagbedhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // cagbedfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v5, v7); // cabgehdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cabgedhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v5, v6); // cabgedfh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v4)) {
													if (compare(v4, v1)) {
														if (compare(v3, v0)) {
															if (compare(v5, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v6); swap(v5, v6); // ghcdaefb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); // gchdaefb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gcdhaefb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cghdaefb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgdhaefb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // cdghaefb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcdahefb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgdahefb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdgahefb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdaghefb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghcdaebf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gchdaebf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcdhaebf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v7); // cghdaebf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cgdhaebf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdghaebf
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gcdahebf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v7); // cgdahebf
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cdgahebf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cdaghebf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ghcadefb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gchadefb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcahdefb
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gcadhefb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cghadefb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgahdefb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cgadhefb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // caghdefb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cagdhefb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cadghefb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghcadebf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gchadebf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gcahdebf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); // gcadhebf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cghadebf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // cgahdebf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v4, v7); swap(v5, v7); // cgadhebf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // caghdebf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); // cagdhebf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cadghebf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v7, v0)) {
																if (compare(v3, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v4, v6); swap(v5, v7); // ghcdabef
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); // gchdabef
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gcdhabef
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cghdabef
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgdhabef
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cdghabef
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ghcadbef
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gchadbef
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cghadbef
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghcabdef
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gchabdef
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cghabdef
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcahdbef
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgahdbef
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // caghdbef
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gcahbdef
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // cgahbdef
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // caghbdef
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcdahbef
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgdahbef
																			}
																		}
																		else {
																			if (compare(v6, v0)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdgahbef
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cdaghbef
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcadhbef
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgadhbef
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cagdhbef
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cadghbef
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gcdabhef
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v7); // cgdabhef
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cdgabhef
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); // gcadbhef
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v4, v6); swap(v5, v7); // cgadbhef
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gcabhdef
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gcabdhef
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cgabhdef
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cgabdhef
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cdagbhef
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v4, v6); swap(v5, v7); // cdabghef
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); // cagdbhef
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cadgbhef
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); // cadbghef
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cagbhdef
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cagbdhef
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cabghdef
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cabgdhef
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v4, v6); swap(v5, v7); // cabdghef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v5)) {
														if (compare(v3, v0)) {
															if (compare(v7, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gcdaehfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v5, v7); swap(v6, v7); // cgdaehfb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cdgaehfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cdagehfb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); // gcdaehbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v5, v7); // cgdaehbf
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v7); // cdgaehbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v5, v7); // cdagehbf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gcdaebhf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v5, v6); swap(v6, v7); // cgdaebhf
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cdgaebhf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cdagebhf
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcdabehf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgdabehf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdgabehf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdagbehf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdabgehf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v5, v7); swap(v6, v7); // gcadehfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v5, v7); // gcadehbf
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v5, v6); swap(v6, v7); // gcadebhf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcadbehf
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gcabdehf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v5, v7); swap(v6, v7); // cgadehfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v6); swap(v5, v7); // cgadehbf
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); swap(v5, v6); swap(v6, v7); // cgadebhf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgadbehf
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cgabdehf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v5, v7); swap(v6, v7); // cagdehfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v5, v7); // cagdehbf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v5, v6); swap(v6, v7); // cagdebhf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cadgehfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); // cadgehbf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cadgebhf
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cagdbehf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cadgbehf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cadbgehf
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cagbdehf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cabgdehf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cabdgehf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v5, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v6, v7); // gcdaefhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v6, v7); // cgdaefhb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v6, v7); // cdgaefhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v6, v7); // cdagefhb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); // gcdaefbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); // cgdaefbh
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); // cdgaefbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); // cdagefbh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // gcdaebfh
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v5, v6); // cgdaebfh
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v5, v6); // cdgaebfh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v5, v6); // cdagebfh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gcdabefh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v6); swap(v2, v3); swap(v4, v6); swap(v5, v6); // cgdabefh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cdgabefh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cdagbefh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v4, v6); swap(v5, v6); // cdabgefh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v6, v7); // gcadefhb
																			}
																			else {
																				swap(v0, v6); swap(v1, v2); swap(v2, v6); // gcadefbh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v5, v6); // gcadebfh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // gcadbefh
																		}
																		else {
																			swap(v0, v6); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gcabdefh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v6); swap(v6, v7); // cgadefhb
																			}
																			else {
																				swap(v0, v2); swap(v1, v6); // cgadefbh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v5, v6); // cgadebfh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v6); swap(v4, v6); swap(v5, v6); // cgadbefh
																		}
																		else {
																			swap(v0, v2); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cgabdefh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v6, v7); // cagdefhb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); // cagdefbh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v5, v6); // cagdebfh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v6, v7); // cadgefhb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); // cadgefbh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // cadgebfh
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // cagdbefh
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cadgbefh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // cadbgefh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cagbdefh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cabgdefh
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v4, v6); swap(v5, v6); // cabdgefh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v6, v0)) {
											if (compare(v5, v0)) {
												if (compare(v3, v0)) {
													if (compare(v7, v0)) {
														if (compare(v5, v3)) {
															if (compare(v6, v5)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // eghfcdab
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egfhcdab
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // egfchdab
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egfcdhab
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eghcfdab
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egchfdab
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // egcfhdab
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egcfdhab
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecghfdab
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ecgfhdab
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecgfdhab
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ceghfdab
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); // cegfhdab
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // cegfdhab
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // efghcdab
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efgchdab
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // efgcdhab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efcghdab
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // efcgdhab
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v4, v6); swap(v5, v7); // efcdghab
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecfghdab
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ecfgdhab
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); // ecfdghab
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cefghdab
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cefgdhab
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cefdghab
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // eghcdfab
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // egchdfab
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v4, v7); swap(v6, v7); // egcdhfab
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egcdfhab
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ecghdfab
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v6, v7); // ecgdhfab
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecgdfhab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ecdghfab
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecdgfhab
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ecdfghab
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); // ceghdfab
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v7); // cegdhfab
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); // cedghfab
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cdeghfab
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v7); // cegdfhab
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cedgfhab
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cedfghab
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cdegfhab
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cdefghab
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v6, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // egfcdahb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // egcfdahb
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v4, v5); swap(v6, v7); // egcdfahb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ecgfdahb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v6, v7); // ecgdfahb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ecdgfahb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cegfdahb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cegdfahb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cedgfahb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cdegfahb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efgcdahb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efcgdahb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efcdgahb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecfgdahb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecfdgahb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecdfgahb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cefgdahb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v6); swap(v6, v7); // cefdgahb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cedfgahb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdefgahb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // egfcdabh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v5); // egcfdabh
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v4, v5); // egcdfabh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // ecgfdabh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v5); // ecgdfabh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // ecdgfabh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cegfdabh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v5); swap(v5, v6); // cegdfabh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cedgfabh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v5); // cdegfabh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // efgcdabh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // efcgdabh
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v4, v6); swap(v5, v6); // efcdgabh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ecfgdabh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // ecfdgabh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ecdfgabh
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v6); // cefgdabh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v6); // cefdgabh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v6); // cedfgabh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cdefgabh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v6, v2)) {
															if (compare(v7, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eghfcadb
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // eghcfadb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // egfhcadb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efghcadb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egfchadb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // efgchadb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // egchfadb
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egcfhadb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // egfcahdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); // efgcahdb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // egcfahdb
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v6, v7); // egfcadhb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // efgcadhb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v6, v7); // egcfadhb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); // egfcadbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // efgcadbh
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v6); swap(v3, v5); // egcfadbh
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ecghfadb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecgfhadb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ecgfahdb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ceghfadb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // cegfhadb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cegfahdb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // efcghadb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ecfghadb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); // cefghadb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v5, v7); // efcgahdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); // ecfgahdb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cefgahdb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // efcgadhb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v6, v7); // ecgfadhb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ecfgadhb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cegfadhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cefgadhb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v5, v6); // efcgadbh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); // ecgfadbh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // ecfgadbh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // cegfadbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cefgadbh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v7, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // eghfcabd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); // eghcfabd
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); // egfhcabd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // efghcabd
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // egfchabd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efgchabd
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v5); // egchfabd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // egcfhabd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); // egfcahbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // efgcahbd
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v5, v7); // egcfahbd
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // egfcabhd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v6, v7); // efgcabhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // egcfabhd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // egfcabdh
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v6); // efgcabdh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v5); swap(v5, v6); // egcfabdh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); // ecghfabd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ecgfhabd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); // ecgfahbd
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ceghfabd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cegfhabd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cegfahbd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efcghabd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecfghabd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cefghabd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // efcgahbd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ecfgahbd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cefgahbd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v6); swap(v6, v7); // efcgabhd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ecgfabhd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v6, v7); // ecfgabhd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cegfabhd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cefgabhd
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v6); // efcgabdh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // ecgfabdh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v6); // ecfgabdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cegfabdh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // cefgabdh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v7, v5)) {
														if (compare(v7, v0)) {
															if (compare(v3, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eghcdafb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egchdafb
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egcdhafb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecghdafb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecgdhafb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecdghafb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ceghdafb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); // cegdhafb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cedghafb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cdeghafb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v6, v7); // eghcafdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // eghcadfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); // eghcafbd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v6, v7); // egchafdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // egchadfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v7); // egchafbd
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v6, v7); // ecghafdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ecghadfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); // ecghafbd
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ceghafdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ceghadfb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); // ceghafbd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // egcahfdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ecgahfdb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); // cegahfdb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egcahdfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecgahdfb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // cegahdfb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v7); // egcahfbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); // ecgahfbd
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // cegahfbd
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v5, v7); swap(v6, v7); // egcdahfb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v5, v7); swap(v6, v7); // ecgdahfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ecdgahfb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // egcadhfb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ecgadhfb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cegdahfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cegadhfb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cedgahfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cdegahfb
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egcafhdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecgafhdb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // cegafhdb
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // egcafdhb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ecgafdhb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cegafdhb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // egcafhbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ecgafhbd
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // cegafhbd
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v6, v7); // egcdafhb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v6, v7); // ecgdafhb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v6, v7); // ecdgafhb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v6, v7); // egcadfhb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v6, v7); // ecgadfhb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v6, v7); // cegdafhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cegadfhb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // cedgafhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v6, v7); // cdegafhb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v5); // egcafdbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // ecgafdbh
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // cegafdbh
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egcafbhd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecgafbhd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // cegafbhd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // egcafbdh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ecgafbdh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // cegafbdh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); // egcdafbh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); // ecgdafbh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); // ecdgafbh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v4); // egcadfbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); // ecgadfbh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v6); // cegdafbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // cegadfbh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); // cedgafbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); // cdegafbh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v7, v0)) {
															if (compare(v3, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // eghcdabf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // egchdabf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v4, v7); swap(v5, v7); // egcdhabf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ecghdabf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); // ecgdhabf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ecdghabf
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ceghdabf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cegdhabf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cedghabf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cdeghabf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); // eghcadbf
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); // eghcabfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // eghcabdf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v5, v7); // egchadbf
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v5, v6); // egchabfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // egchabdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); // ecghadbf
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); // ecghabfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ecghabdf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ceghadbf
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ceghabfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ceghabdf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // egcahdbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ecgahdbf
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // cegahdbf
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // egcahbfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ecgahbfd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // cegahbfd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egcahbdf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecgahbdf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // cegahbdf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v5, v7); // egcdahbf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v5, v7); // ecgdahbf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); // ecdgahbf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v5, v7); // egcadhbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); // ecgadhbf
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v7); // cegdahbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cegadhbf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // cedgahbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v7); // cdegahbf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v7, v5)) {
																if (compare(v4, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v5, v6); swap(v6, v7); // egcdabhf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v5, v6); swap(v6, v7); // ecgdabhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ecdgabhf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // egcadbhf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ecgadbhf
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cegdabhf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cegadbhf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cedgabhf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cdegabhf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v3, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v5, v6); // egcdabfh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v5, v6); // ecgdabfh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // ecdgabfh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v5, v6); // egcadbfh
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // ecgadbfh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v4, v6); swap(v5, v6); // cegdabfh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cegadbfh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // cedgabfh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v6); swap(v5, v6); // cdegabfh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egcabhfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // egcabfhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v6); // egcabfdh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecgabhfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ecgabfhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // ecgabfdh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // cegabhfd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v6, v7); // cegabfhd
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); // cegabfdh
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // egcabhdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egcabdhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // egcabdfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ecgabhdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecgabdhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ecgabdfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); // cegabhdf
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // cegabdhf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // cegabdfh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v3)) {
												if (compare(v7, v3)) {
													if (compare(v7, v1)) {
														if (compare(v3, v1)) {
															if (compare(v7, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); // ecaghfdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ecaghdfb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ecaghdbf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v7); // ceaghfdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ceaghdfb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ceaghdbf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); // caeghfdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // caeghdfb
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // caeghdbf
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // efcaghdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ecfaghdb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ecagfhdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecafghdb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cefaghdb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ceagfhdb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ceafghdb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // caegfhdb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // caefghdb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ecaghfbd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ecaghbfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ecaghbdf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ceaghfbd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ceaghbfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ceaghbdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // caeghfbd
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // caeghbfd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // caeghbdf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efcaghbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecfaghbd
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ecagfhbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ecafghbd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // cefaghbd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ceagfhbd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ceafghbd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // caegfhbd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // caefghbd
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // efcagbhd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ecfagbhd
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ecagfbhd
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecafgbhd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cefagbhd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ceagfbhd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ceafgbhd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // caegfbhd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // caefgbhd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ecagbhfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v6, v7); // ecagbfhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); // ecagbhdf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ceagbhfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v6, v7); // ceagbfhd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v5, v7); // ceagbhdf
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // caegbhfd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v6, v7); // caegbfhd
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); // caegbhdf
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecabghfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ecabgfhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ecabghdf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ceabghfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ceabgfhd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ceabghdf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // caebghfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // caebgfhd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // caebghdf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cabeghfd
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cabegfhd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cabeghdf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efcabghd
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecfabghd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cefabghd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ecafbghd
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ceafbghd
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // caefbghd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecabfghd
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ceabfghd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // caebfghd
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cabefghd
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v7, v1)) {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efcagdhb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecfagdhb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecagfdhb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ecafgdhb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // cefagdhb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ceagfdhb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ceafgdhb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // caegfdhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // caefgdhb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecagdhfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ecagdfhb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ecagdhbf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ceagdhfb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ceagdfhb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ceagdhbf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // caegdhfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // caegdfhb
																		}
																	}
																	else {
																		swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // caegdhbf
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // efcagdbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ecfagdbh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ecagfdbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // ecafgdbh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // cefagdbh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ceagfdbh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v6); // ceafgdbh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // caegfdbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // caefgdbh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // ecagdfbh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecagdbhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ecagdbfh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v6); // ceagdfbh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ceagdbhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ceagdbfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // caegdfbh
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // caegdbhf
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // caegdbfh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v6); // efcagbdh
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // ecfagbdh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // ecagfbdh
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ecafgbdh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // cefagbdh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v4, v5); // ceagfbdh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ceafgbdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // caegfbdh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // caefgbdh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); // ecagbfdh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ecagbdhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // ecagbdfh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v3, v6); // ceagbfdh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ceagbdhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v6); swap(v5, v6); // ceagbdfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); // caegbfdh
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // caegbdhf
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // caegbdfh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v4, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // ecabgfdh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecabgdhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ecabgdfh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v6); // ceabgfdh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ceabgdhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ceabgdfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // caebgfdh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // caebgdhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // caebgdfh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v6); // cabegfdh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cabegdhf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cabegdfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // efcabgdh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ecfabgdh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // cefabgdh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // ecafbgdh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v5, v6); // ceafbgdh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // caefbgdh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ecabfgdh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ceabfgdh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // caebfgdh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cabefgdh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v5)) {
													if (compare(v7, v5)) {
														if (compare(v3, v0)) {
															if (compare(v7, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecdaghfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cedaghfb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdeaghfb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cdaeghfb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ecdaghbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v7); // cedaghbf
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cdeaghbf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cdaeghbf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecdagbhf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cedagbhf
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdeagbhf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cdaegbhf
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ecdabghf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v5, v6); swap(v6, v7); // cedabghf
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cdeabghf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cdaebghf
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v5, v6); swap(v6, v7); // cdabeghf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ecadghfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); // ecadghbf
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ecadgbhf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v5, v6); swap(v6, v7); // ecadbghf
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ecabdghf
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ceadghfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v4); swap(v4, v6); swap(v5, v7); // ceadghbf
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ceadgbhf
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v5, v6); swap(v6, v7); // ceadbghf
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ceabdghf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // caedghfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); // caedghbf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // caedgbhf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // cadeghfb
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // cadeghbf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // cadegbhf
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v5, v6); swap(v6, v7); // caedbghf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cadebghf
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v5, v6); swap(v6, v7); // cadbeghf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // caebdghf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v5, v6); swap(v6, v7); // cabedghf
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v5, v6); swap(v6, v7); // cabdeghf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v0)) {
															if (compare(v5, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ecdagfhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v6, v7); // cedagfhb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cdeagfhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cdaegfhb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // ecdagfbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v6); // cedagfbh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); // cdeagfbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v6); // cdaegfbh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ecdagbfh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v6); swap(v5, v6); // cedagbfh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cdeagbfh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cdaegbfh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // ecdabgfh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v5, v6); // cedabgfh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v5, v6); // cdeabgfh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v5, v6); // cdaebgfh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v5, v6); // cdabegfh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v6, v7); // ecadgfhb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v6); // ecadgfbh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // ecadgbfh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v5, v6); // ecadbgfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // ecabdgfh
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v2); swap(v1, v4); swap(v4, v6); swap(v6, v7); // ceadgfhb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v4, v6); // ceadgfbh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v1)) {
																				swap(v0, v2); swap(v1, v4); swap(v4, v6); swap(v5, v6); // ceadgbfh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v5, v6); // ceadbgfh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v5, v6); // ceabdgfh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v6, v7); // caedgfhb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v6); // caedgfbh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // caedgbfh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v1)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // cadegfhb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // cadegfbh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // cadegbfh
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v5, v6); // caedbgfh
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // cadebgfh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v5, v6); // cadbegfh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // caebdgfh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v5, v6); // cabedgfh
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v5, v6); // cabdegfh
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v3, v0)) {
															if (compare(v7, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v5, v6); swap(v6, v7); // efcdaghb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v5, v6); swap(v6, v7); // ecfdaghb
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ecdfaghb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cefdaghb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cedfaghb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // cdefaghb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecdafghb
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cedafghb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdeafghb
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cdaefghb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v5, v6); // efcdagbh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v5, v6); // ecfdagbh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // ecdfagbh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v5); swap(v5, v6); // cefdagbh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cedfagbh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); swap(v5, v6); // cdefagbh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ecdafgbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v5); swap(v5, v6); // cedafgbh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cdeafgbh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cdaefgbh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // efcadghb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ecfadghb
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecafdghb
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ecadfghb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cefadghb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ceafdghb
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ceadfghb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // caefdghb
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // caedfghb
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // cadefghb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v5, v6); // efcadgbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // ecfadgbh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ecafdgbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // ecadfgbh
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // cefadgbh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ceafdgbh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v4, v5); swap(v5, v6); // ceadfgbh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // caefdgbh
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // caedfgbh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // cadefgbh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v5, v0)) {
																if (compare(v3, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v4); swap(v1, v5); // efcdabgh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); // ecfdabgh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v5); // ecdfabgh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v4, v5); // cefdabgh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v3, v5); swap(v4, v5); // cedfabgh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v5); // cdefabgh
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v4); // efcadbgh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); // ecfadbgh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // cefadbgh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v3, v4); swap(v4, v5); // efcabdgh
																			}
																			else {
																				swap(v0, v4); swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // ecfabdgh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v2, v5); swap(v3, v5); // cefabdgh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // ecafdbgh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); swap(v4, v5); // ceafdbgh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // caefdbgh
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v5); // ecafbdgh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v3, v5); // ceafbdgh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v5); // caefbdgh
																		}
																	}
																}
																else {
																	if (compare(v3, v0)) {
																		if (compare(v4, v3)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // ecdafbgh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); swap(v4, v5); // cedafbgh
																			}
																		}
																		else {
																			if (compare(v4, v0)) {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); swap(v4, v5); // cdeafbgh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v3, v4); swap(v4, v5); // cdaefbgh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			if (compare(v4, v2)) {
																				swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v4, v5); // ecadfbgh
																			}
																			else {
																				swap(v0, v2); swap(v1, v4); swap(v4, v5); // ceadfbgh
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v4, v5); // caedfbgh
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // cadefbgh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v3); swap(v3, v4); // ecdabfgh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v4); swap(v2, v3); // cedabfgh
																			}
																			else {
																				swap(v0, v2); swap(v1, v3); swap(v2, v4); swap(v3, v4); // cdeabfgh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); // ecadbfgh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); // ceadbfgh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // ecabfdgh
																		}
																		else {
																			swap(v0, v4); swap(v1, v2); swap(v2, v4); swap(v3, v4); // ecabdfgh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); swap(v4, v5); // ceabfdgh
																		}
																		else {
																			swap(v0, v2); swap(v1, v4); swap(v3, v4); // ceabdfgh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v3, v0)) {
																		if (compare(v4, v1)) {
																			swap(v0, v2); swap(v1, v3); swap(v3, v4); // cdaebfgh
																		}
																		else {
																			swap(v0, v2); swap(v1, v3); // cdabefgh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v2, v4); // caedbfgh
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v2, v3); swap(v3, v4); // cadebfgh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v3); // cadbefgh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // caebfdgh
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); swap(v2, v4); swap(v3, v4); // caebdfgh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); swap(v4, v5); // cabefdgh
																			}
																			else {
																				swap(v0, v2); swap(v1, v2); swap(v3, v4); // cabedfgh
																			}
																		}
																		else {
																			swap(v0, v2); swap(v1, v2); // cabdefgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
						else {
							if (compare(v5, v4)) {
								if (compare(v7, v6)) {
									if (compare(v6, v4)) {
										if (compare(v6, v1)) {
											if (compare(v7, v0)) {
												if (compare(v6, v5)) {
													if (compare(v6, v0)) {
														if (compare(v4, v1)) {
															if (compare(v5, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hgfeacdb
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v6, v7); // hgfaecdb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hgfacedb
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgfacdeb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); // hgfeacbd
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hgfeabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); // hgfaecbd
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); // hgfacebd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); // hgfaebcd
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hgafecdb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgafcedb
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hgafcdeb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hgacfedb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgacfdeb
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hgacdfeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); // hgafecbd
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hgafebcd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hgafcebd
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); // hgacfebd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hgfacdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hgfacbed
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgfacbde
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); // hgafcdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgafcbed
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hgafcbde
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hgacfdbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); // hgacdfbe
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hgacfbed
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgacfbde
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); // hgacbfed
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hgacbfde
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hgacdbfe
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgacbdfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hgfabecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); // hgfabced
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hgfabcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgafbecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hgafbced
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgafbcde
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hgabfecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgabfced
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hgabfcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hgabcfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); // hgabcfde
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hgabcdfe
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v4, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); // hagfecdb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hagfcedb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); // hagfcdeb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); // hagcfedb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hagcfdeb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); // hagcdfeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v5); // hacgfedb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hacgfdeb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v6); // hacgdfeb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v4, v6); // hacdgfeb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hagfcdbe
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hagcfdbe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hagcdfbe
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hagcdbfe
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hacgfdbe
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // hacgdfbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v4, v6); swap(v6, v7); // hacdgfbe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hacgdbfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hacdgbfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hagfecbd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hagfcebd
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hagfcbed
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hagfcbde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v1)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); // hagfebcd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hagfbecd
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hagfbced
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hagfbcde
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hagcfebd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hagcfbed
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hagcfbde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hacgfebd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hacgfbed
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hacgfbde
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hagcbfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); // hagcbfde
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hagcbdfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hacgbfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v7); // hacgbfde
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hacgbdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); // hagbfecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hagbfced
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hagbfcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); // hagbcfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hagbcfde
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hagbcdfe
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v3, v1)) {
															if (compare(v4, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfgeacdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fhgeacdb
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hfgaecdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); // fhgaecdb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); // hfagecdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); // fhagecdb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); // hafgecdb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfgacedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fhgacedb
																			}
																		}
																		else {
																			if (compare(v5, v0)) {
																				if (compare(v7, v5)) {
																					swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hfagcedb
																				}
																				else {
																					swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); // fhagcedb
																				}
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); // hafgcedb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfacgedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fhacgedb
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hafcgedb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // hacfgedb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hfgacdeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); // fhgacdeb
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfagcdeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fhagcdeb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hfacgdeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); // fhacgdeb
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfacdgeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fhacdgeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // hafgcdeb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); // hafcgdeb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hafcdgeb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v6); // hacfgdeb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // hacfdgeb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v4, v5); swap(v5, v6); // hacdfgeb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hfgeacbd
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); // hfgeabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); // hfgaecbd
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hfgacebd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); // hfgaebcd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhgeacbd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fhgeabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fhgaecbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhgacebd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); // fhgaebcd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hfagecbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v6, v7); // fhagecbd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v6, v7); // hafgecbd
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v6, v7); // hfagebcd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fhagebcd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hafgebcd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfagcebd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fhagcebd
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hfacgebd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fhacgebd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hafgcebd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hafcgebd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hacfgebd
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); // hfgacbed
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fhgacbed
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hfagcbed
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhagcbed
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hafgcbed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); // hfacgbed
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fhacgbed
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hafcgbed
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hacfgbed
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); // hfgabecd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fhgabecd
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // hfagbecd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhagbecd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hafgbecd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hfgabced
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhgabced
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfagbced
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fhagbced
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hafgbced
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfgacdbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fhgacdbe
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // hfagcdbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhagcdbe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfacgdbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fhacgdbe
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hfacdgbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhacdgbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hafgcdbe
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hafcgdbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hafcdgbe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // hacfgdbe
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hacfdgbe
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hacdfgbe
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hfgacbde
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhgacbde
																			}
																		}
																		else {
																			if (compare(v5, v0)) {
																				if (compare(v7, v5)) {
																					swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); // hfagcbde
																				}
																				else {
																					swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fhagcbde
																				}
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hafgcbde
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hfacgbde
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhacgbde
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hafcgbde
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hacfgbde
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfgabcde
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fhgabcde
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // hfagbcde
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); // fhagbcde
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); // hafgbcde
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v4, v1)) {
														if (compare(v6, v2)) {
															if (compare(v3, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); // ahgfecdb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ahgfcedb
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); // ahgfcdeb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v5); swap(v3, v6); // ahfgecdb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); // ahfgcedb
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ahfgcdeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); // fahgecdb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fahgcedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fahgcdeb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); // afhgecdb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // afhgcedb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afhgcdeb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v5)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ahgfecbd
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahgfcebd
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); // ahgfebcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v6, v7); // ahfgecbd
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ahfgcebd
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ahfgebcd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fahgecbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fahgcebd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v6, v7); // fahgebcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // afhgecbd
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afhgcebd
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v6, v7); // afhgebcd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fahcgedb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ahfcgedb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afhcgedb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fahcgebd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahfcgebd
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // afhcgebd
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fahcgdeb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); // ahfcgdeb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // afhcgdeb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fahcdgeb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ahfcdgeb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afhcdgeb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fachgedb
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fachgebd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // fachgdeb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fachdgeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // facdhgeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afchgedb
																		}
																		else {
																			swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // afchgebd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // afchgdeb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afchdgeb
																			}
																			else {
																				swap(v1, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afcdhgeb
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v4, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ahgfbecd
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahfgbecd
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); // ahgbfecd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ahgfcbed
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahfgcbed
																			}
																		}
																		else {
																			if (compare(v5, v1)) {
																				if (compare(v6, v5)) {
																					swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahgfbced
																				}
																				else {
																					swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ahfgbced
																				}
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ahgbfced
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fahgcbed
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fahgbecd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fahgbced
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // afhgcbed
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // afhgbecd
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afhgbced
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ahgfcdbe
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahfgcdbe
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahgfcbde
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ahfgcbde
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ahgfbcde
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); // ahfgbcde
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahgbfcde
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fahgcdbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); // fahgcbde
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fahgbcde
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // afhgcdbe
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); // afhgcbde
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // afhgbcde
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fahcgdbe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ahfcgdbe
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afhcgdbe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fahcdgbe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahfcdgbe
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // afhcdgbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fachgdbe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // fachdgbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v4, v7); swap(v5, v6); // facdhgbe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afchgdbe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // afchdgbe
																			}
																			else {
																				swap(v1, v5); swap(v4, v7); swap(v5, v6); // afcdhgbe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); // fahcgbed
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v6); // fachgbed
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ahfcgbed
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); // afhcgbed
																			}
																		}
																		else {
																			swap(v1, v5); swap(v3, v7); swap(v4, v6); // afchgbed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fahcgbde
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fachgbde
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahfcgbde
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // afhcgbde
																			}
																		}
																		else {
																			swap(v1, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // afchgbde
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v2)) {
														if (compare(v3, v1)) {
															if (compare(v4, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); // ahgcfedb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v3, v6); swap(v4, v5); // ahcgfedb
																			}
																			else {
																				swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ahcfgedb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v6, v2)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ahgcfdeb
																			}
																			else {
																				swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ahcgfdeb
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v3, v5); swap(v4, v6); // ahcfgdeb
																			}
																			else {
																				swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // ahcfdgeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); // ahgcdfeb
																		}
																		else {
																			swap(v1, v7); swap(v3, v6); swap(v4, v6); // ahcgdfeb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v7); swap(v4, v6); // ahcdgfeb
																		}
																		else {
																			swap(v1, v7); swap(v4, v5); swap(v5, v6); // ahcdfgeb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahgcfdbe
																		}
																		else {
																			swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahcgfdbe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ahcfgdbe
																		}
																		else {
																			swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahcfdgbe
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v2)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ahgcdfbe
																			}
																			else {
																				swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ahcgdfbe
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v4, v6); swap(v6, v7); // ahcdgfbe
																			}
																			else {
																				swap(v1, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahcdfgbe
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahgcdbfe
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahcgdbfe
																			}
																			else {
																				swap(v1, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahcdgbfe
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v6, v2)) {
																		swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ahgcfebd
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ahcgfebd
																		}
																		else {
																			swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahcfgebd
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahgcfbed
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahcgfbed
																			}
																			else {
																				swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ahcfgbed
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ahgcfbde
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ahcgfbde
																			}
																			else {
																				swap(v1, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahcfgbde
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ahgcbfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); // ahgcbfde
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ahgcbdfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ahcgbfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v3, v6); swap(v4, v7); // ahcgbfde
																			}
																			else {
																				swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ahcgbdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); // ahgbcfed
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ahgbcfde
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahgbcdfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v6, v3)) {
																if (compare(v4, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); // achgfedb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); // achfgedb
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acfhgedb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // achgfebd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // achfgebd
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // acfhgebd
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); // achgfdeb
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); // achgdfeb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); // achfgdeb
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); // acfhgdeb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); // achfdgeb
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acfhdgeb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v2); swap(v2, v7); swap(v4, v6); // achdgfeb
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); // achdfgeb
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); // acdhgfeb
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); // acdhfgeb
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acfdhgeb
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acdfhgeb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // achgfdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // achgfbed
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // achgfbde
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v6, v7); // achgdfbe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v6, v7); // achdgfbe
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v6, v7); // acdhgfbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // achgdbfe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // achdgbfe
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acdhgbfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // achgbfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); // achgbfde
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // achgbdfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v6, v7); // achfgdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // achfgbed
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // achfgbde
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acfhgdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); // acfhgbed
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // acfhgbde
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // achfdgbe
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v6); // acfhdgbe
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v6); // acfdhgbe
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // achdfgbe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acdhfgbe
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v6); // acdfhgbe
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v1)) {
												if (compare(v2, v1)) {
													if (compare(v7, v2)) {
														if (compare(v7, v0)) {
															if (compare(v5, v0)) {
																if (compare(v6, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hfacbged
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fhacbged
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfacbgde
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fhacbgde
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); // hfacdbge
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fhacdbge
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hfacbdge
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); // fhacbdge
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hafcbged
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hafcbgde
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hafcdbge
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); // hafcbdge
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // hacfdbge
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v4, v5); swap(v5, v7); // hacdfbge
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hacfbged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hacfbgde
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v7); // hacfbdge
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahfcbged
																		}
																		else {
																			swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahcfbged
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fahcbged
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // afhcbged
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ahfcbgde
																		}
																		else {
																			swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ahcfbgde
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fahcbgde
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afhcbgde
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); // fahcdbge
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ahfcdbge
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); // afhcdbge
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // ahcfdbge
																		}
																		else {
																			swap(v1, v7); swap(v4, v5); swap(v5, v7); // ahcdfbge
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); // ahfcbdge
																		}
																		else {
																			swap(v1, v7); swap(v3, v5); swap(v4, v7); // ahcfbdge
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fahcbdge
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // afhcbdge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																if (compare(v7, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v7); // fachdbge
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v4, v7); // facdhbge
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fachbged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fachbgde
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fachbdge
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // facbhged
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // facbhgde
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v4, v5); swap(v5, v7); // facdbhge
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v7); // facbhdge
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // facbdhge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v3, v7); swap(v4, v7); // afchdbge
																		}
																		else {
																			swap(v1, v5); swap(v4, v7); // afcdhbge
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // afchbged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afchbgde
																			}
																			else {
																				swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // afchbdge
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afcbhged
																		}
																		else {
																			swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // afcbhgde
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v5); swap(v4, v5); swap(v5, v7); // afcdbhge
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v3, v5); swap(v4, v7); // afcbhdge
																			}
																			else {
																				swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // afcbdhge
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); // achfdbge
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); // acfhdbge
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v1, v2); swap(v2, v5); swap(v4, v7); // acfdhbge
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); // acfdbhge
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); // achdfbge
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); // acdhfbge
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); // acdfhbge
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); // acdfbhge
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // achfbged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // achfbgde
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); // achfbdge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // acfhbged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acfhbgde
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // acfhbdge
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acfbhged
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // acfbhgde
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); // acfbhdge
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // acfbdhge
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v7, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfabgecd
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // hfabgced
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfabcged
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fhabgecd
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); // fhabgced
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fhabcged
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fahbgecd
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fahbgced
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fahbcged
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fabhgecd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); // fabhgced
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fabhcged
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fabchged
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v2)) {
																		swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfabgcde
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hfabcgde
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); // hfabcdge
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fhabgcde
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhabcgde
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fhabcdge
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fahbgcde
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fahbcgde
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fabhgcde
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fabhcgde
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fabchgde
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); // fahbcdge
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fabhcdge
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); // fabchdge
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fabcdhge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v7, v5)) {
																if (compare(v7, v0)) {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hafbgecd
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); // hafbgced
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hafbcged
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ahfbgecd
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); // ahfbgced
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ahfbcged
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afhbgecd
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // afhbgced
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afhbcged
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // afbhgecd
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); // afbhgced
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // afbhcged
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afbchged
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hafbgcde
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hafbcgde
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hafbcdge
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ahfbgcde
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahfbcgde
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ahfbcdge
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afhbgcde
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // afhbcgde
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); // afhbcdge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // afbhgcde
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afbhcgde
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // afbhcdge
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // afbchgde
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); // afbchdge
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // afbcdhge
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v1)) {
													if (compare(v6, v5)) {
														if (compare(v2, v1)) {
															if (compare(v6, v3)) {
																if (compare(v7, v0)) {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v6); // hacbgfed
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hacbgfde
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hacbgdfe
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v3, v7); swap(v4, v6); // ahcbgfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ahcbgfde
																			}
																			else {
																				swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahcbgdfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); // achbgfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // achbgfde
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // achbgdfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v4, v7); swap(v5, v6); // hacdbgfe
																		}
																		else {
																			swap(v1, v7); swap(v4, v7); swap(v5, v6); // ahcdbgfe
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); // achdbgfe
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); // acdhbgfe
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // hacbdgfe
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ahcbdgfe
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // achbdgfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // habgfecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // habgfced
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // habgfcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // habgcfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); // habgcfde
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // habgcdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); // habcgfed
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // habcgfde
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // habcgdfe
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // habcdgfe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ahbgfecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahbgfced
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ahbgfcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ahbgcfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); // ahbgcfde
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ahbgcdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); // ahbcgfed
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ahbcgfde
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahbcgdfe
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // ahbcdgfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v5, v3)) {
																if (compare(v7, v0)) {
																	if (compare(v4, v3)) {
																		swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hacbfged
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hacbfgde
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hacbfdge
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ahcbfged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahcbfgde
																			}
																			else {
																				swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ahcbfdge
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // achbfged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // achbfgde
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // achbfdge
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v4, v7); // hacdbfge
																		}
																		else {
																			swap(v1, v7); swap(v4, v7); // ahcdbfge
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v4, v7); // achdbfge
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); // acdhbfge
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v7); // hacbdfge
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v7); swap(v3, v7); swap(v4, v7); // ahcbdfge
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); // achbdfge
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // habfgecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // habfgced
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // habfgcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // habfcged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // habfcgde
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); // habfcdge
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // habcfged
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // habcfgde
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // habcfdge
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); // habcdfge
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahbfgecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ahbfgced
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahbfgcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahbfcged
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ahbfcgde
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); // ahbfcdge
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ahbcfged
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahbcfgde
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // ahbcfdge
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); // ahbcdfge
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v2)) {
														if (compare(v6, v2)) {
															if (compare(v6, v5)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // abhgfecd
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abhgfced
																		}
																		else {
																			swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // abhgfcde
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // abhgcfed
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v7); swap(v3, v6); swap(v4, v7); // abhgcfde
																		}
																		else {
																			swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // abhgcdfe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v4, v2)) {
																		swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abhfgecd
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // abhfgced
																		}
																		else {
																			swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abhfgcde
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // abfhgecd
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v2, v5); swap(v3, v7); swap(v4, v6); // abfhgced
																		}
																		else {
																			swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // abfhgcde
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abhfcged
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // abhfcgde
																		}
																		else {
																			swap(v2, v7); swap(v3, v5); swap(v4, v7); // abhfcdge
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // abfhcged
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abfhcgde
																		}
																		else {
																			swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // abfhcdge
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v2, v7); swap(v3, v7); swap(v4, v6); // abhcgfed
																		}
																		else {
																			swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // abhcgfde
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abhcgdfe
																		}
																		else {
																			swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // abhcdgfe
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // abhcfged
																		}
																		else {
																			swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abhcfgde
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // abhcfdge
																		}
																		else {
																			swap(v2, v7); swap(v3, v7); swap(v4, v7); // abhcdfge
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v5, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v1, v2); swap(v3, v7); swap(v4, v6); // acbhgfed
																		}
																		else {
																			swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); // acbhfged
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v6, v7); // acbhgfde
																			}
																			else {
																				swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acbhfgde
																			}
																		}
																		else {
																			swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); // acbhfdge
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acbfhged
																		}
																		else {
																			swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v6); // acbfhgde
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v3, v5); swap(v4, v7); // acbfhdge
																		}
																		else {
																			swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); // acbfdhge
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acbhgdfe
																		}
																		else {
																			swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); // acbhdgfe
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); // acdbhgfe
																		}
																		else {
																			swap(v1, v2); swap(v4, v7); swap(v5, v6); // acbdhgfe
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v5)) {
																			swap(v1, v2); swap(v2, v3); swap(v4, v7); // acdbhfge
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); // acdbfhge
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v3, v7); swap(v4, v7); // acbhdfge
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v2); swap(v4, v7); // acbdhfge
																			}
																			else {
																				swap(v1, v2); swap(v4, v5); swap(v5, v7); // acbdfhge
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v3, v7); swap(v4, v6); // abchgfed
																		}
																		else {
																			swap(v3, v7); swap(v4, v5); swap(v5, v6); // abchfged
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abfchged
																		}
																		else {
																			swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abcfhged
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v3, v7); swap(v4, v6); swap(v6, v7); // abchgfde
																		}
																		else {
																			swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abchgdfe
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // abfchgde
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abchfgde
																			}
																			else {
																				swap(v3, v5); swap(v4, v7); swap(v5, v6); // abcfhgde
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v2, v5); swap(v3, v5); swap(v4, v7); // abfchdge
																		}
																		else {
																			swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // abfcdhge
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v3, v7); swap(v4, v5); swap(v5, v7); // abchfdge
																			}
																			else {
																				swap(v3, v5); swap(v4, v7); // abcfhdge
																			}
																		}
																		else {
																			swap(v3, v5); swap(v4, v5); swap(v5, v7); // abcfdhge
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			swap(v3, v7); swap(v4, v7); swap(v5, v6); // abchdgfe
																		}
																		else {
																			swap(v3, v7); swap(v4, v7); // abchdfge
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v4, v7); swap(v5, v6); // abcdhgfe
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v4, v7); // abcdhfge
																			}
																			else {
																				swap(v4, v5); swap(v5, v7); // abcdfhge
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v1)) {
											if (compare(v5, v0)) {
												if (compare(v7, v4)) {
													if (compare(v6, v1)) {
														if (compare(v3, v1)) {
															if (compare(v6, v3)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hfegacdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); // fhegacdb
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hfeagcdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fheagcdb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hfaegcdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); // fhaegcdb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fahegcdb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hfeacgdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); // fheacgdb
																			}
																		}
																		else {
																			if (compare(v7, v0)) {
																				if (compare(v7, v5)) {
																					swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hfaecgdb
																				}
																				else {
																					swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fhaecgdb
																				}
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fahecgdb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hfacegdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); // fhacegdb
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fahcegdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // fachegdb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); // hfeacdgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); // fheacdgb
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hfaecdgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fhaecdgb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); // hfacedgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); // fhacedgb
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hfacdegb
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fhacdegb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // fahecdgb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); // fahcedgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fahcdegb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v5, v7); // fachedgb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // fachdegb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v4, v7); swap(v5, v7); // facdhegb
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfegacbd
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hfegabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hfeagcbd
																			}
																			else {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); // hfeacgbd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); // hfeagbcd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fhegacbd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhegabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fheagcbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fheacgbd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fheagbcd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hfaegcbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fhaegcbd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fahegcbd
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hfaegbcd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fhaegbcd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fahegbcd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hfaecgbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fhaecgbd
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); // hfacegbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fhacegbd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fahecgbd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); // fahcegbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v5, v6); // fachegbd
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); // hfeacbgd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); // fheacbgd
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); // hfaecbgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fhaecbgd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); // fahecbgd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); // hfacebgd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); // fhacebgd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); // fahcebgd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v7); // fachebgd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hfeabgcd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fheabgcd
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hfaebgcd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fhaebgcd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fahebgcd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); // hfeabcgd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fheabcgd
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hfaebcgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); // fhaebcgd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fahebcgd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hfeacdbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v6, v7); // fheacdbg
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfaecdbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fhaecdbg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hfacedbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v6, v7); // fhacedbg
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfacdebg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fhacdebg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fahecdbg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fahcedbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fahcdebg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // fachedbg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fachdebg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // facdhebg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v6, v7); // hfeacbdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fheacbdg
																			}
																		}
																		else {
																			if (compare(v7, v0)) {
																				if (compare(v7, v5)) {
																					swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hfaecbdg
																				}
																				else {
																					swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhaecbdg
																				}
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fahecbdg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v6, v7); // hfacebdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fhacebdg
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v6, v7); // fahcebdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v6, v7); // fachebdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hfeabcdg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fheabcdg
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			if (compare(v7, v5)) {
																				swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfaebcdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fhaebcdg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fahebcdg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v6, v1)) {
															if (compare(v7, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fehgacdb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fehagcdb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fehacgdb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); // fehacdgb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fehgacbd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fehgabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fehagcbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fehacgbd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fehagbcd
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // feahgcdb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // feahcgdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // feahcdgb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // feachgdb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); // feachdgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // feacdhgb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); // feahgcbd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // feahgbcd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // feahcgbd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // feachgbd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fehacdbg
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fehacbgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fehacbdg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // feahcdbg
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); // feahcbgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // feahcbdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // feachdbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // feacdhbg
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // feachbgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // feachbdg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); // feacbhgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // feacbhdg
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // feacdbhg
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v6, v7); // feacbdhg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fehabgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); // fehabcgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fehabcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); // feahbgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); // feahbcgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v6, v7); // feahbcdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // feabhgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); // feabhcgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // feabhcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // feabchgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // feabchdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // feabcdhg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v6, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // faehgcdb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // faehcgdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); // faehcdgb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // faechgdb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // faechdgb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); // faecdhgb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // facehgdb
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // facehdgb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v5, v7); // facedhgb
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v5, v7); // facdehgb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // faehcdbg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // faechdbg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // faecdhbg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v6, v7); // faecdbhg
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // facehdbg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // facedhbg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v5, v7); swap(v6, v7); // facdehbg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v6, v7); // facedbhg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v6, v7); // facdebhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // faehgcbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); // faehcgbd
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); // faehcbgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v6, v7); // faehcbdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v1)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); // faehgbcd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // faehbgcd
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); // faehbcgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // faehbcdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // faechgbd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); // faechbgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // faechbdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // facehgbd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v7); // facehbgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // facehbdg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // faecbhgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // faecbhdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // faecbdhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // facebhgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // facebhdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // facebdhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // faebhgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // faebhcgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // faebhcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); // faebchgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // faebchdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v6, v7); // faebcdhg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v2)) {
													if (compare(v6, v1)) {
														if (compare(v4, v2)) {
															if (compare(v3, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); // hafegcdb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hafecgdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); // hafecdgb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); // ahfegcdb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ahfecgdb
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); // ahfecdgb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // afhegcdb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afhecgdb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // afhecdgb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afehgcdb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // afehcgdb
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); // afehcdgb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hafegcbd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hafecgbd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hafegbcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ahfegcbd
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahfecgbd
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahfegbcd
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afhegcbd
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // afhecgbd
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // afhegbcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v2)) {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // afehgcbd
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); // afehcgbd
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); // afehgbcd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); // hafcegdb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); // ahfcegdb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // afhcegdb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hafcegbd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ahfcegbd
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v6); // afhcegbd
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); // hafcedgb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); // ahfcedgb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); // afhcedgb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); // hafcdegb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); // ahfcdegb
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // afhcdegb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v5, v6); // hacfegdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v5); // hacfedgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hacfegbd
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v5); // hacfdegb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v4, v5); // hacdfegb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v3, v5); swap(v5, v6); // ahcfegdb
																			}
																			else {
																				swap(v1, v7); swap(v3, v5); // ahcfedgb
																			}
																		}
																		else {
																			swap(v1, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ahcfegbd
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v3, v5); swap(v4, v5); // ahcfdegb
																		}
																		else {
																			swap(v1, v7); swap(v4, v5); // ahcdfegb
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v6, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hafecbgd
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hafebgcd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); // hafebcgd
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ahfecbgd
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ahfebgcd
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); // ahfebcgd
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afhebgcd
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // afehbgcd
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afebhgcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); // afhecbgd
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); // afehcbgd
																			}
																		}
																		else {
																			if (compare(v7, v1)) {
																				if (compare(v7, v4)) {
																					swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // afhebcgd
																				}
																				else {
																					swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); // afehbcgd
																				}
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // afebhcgd
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hafecdbg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hafecbdg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hafebcdg
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ahfecdbg
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahfecbdg
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ahfebcdg
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afhecdbg
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // afehcdbg
																			}
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // afhecbdg
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v6, v7); // afehcbdg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afhebcdg
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // afehbcdg
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afebhcdg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v6, v7); // hafcedbg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v6, v7); // ahfcedbg
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // afhcedbg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hafcdebg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ahfcdebg
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afhcdebg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v6, v7); // hacfedbg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // hacfdebg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v4, v5); swap(v6, v7); // hacdfebg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v3, v5); swap(v6, v7); // ahcfedbg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // ahcfdebg
																			}
																			else {
																				swap(v1, v7); swap(v4, v5); swap(v6, v7); // ahcdfebg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); // hafcebgd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v5, v7); // hacfebgd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); // ahfcebgd
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); // afhcebgd
																			}
																		}
																		else {
																			swap(v1, v7); swap(v3, v5); swap(v5, v7); // ahcfebgd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hafcebdg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hacfebdg
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ahfcebdg
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v6, v7); // afhcebdg
																			}
																		}
																		else {
																			swap(v1, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ahcfebdg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v3, v1)) {
															if (compare(v6, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afechgdb
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // afchegdb
																			}
																			else {
																				swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afcehgdb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v3, v7); swap(v5, v7); // afchedgb
																			}
																			else {
																				swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // afchdegb
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // afechdgb
																			}
																			else {
																				swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // afcehdgb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); // afecdhgb
																		}
																		else {
																			swap(v1, v5); swap(v3, v4); swap(v5, v7); // afcedhgb
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v1, v5); swap(v4, v7); swap(v5, v7); // afcdhegb
																		}
																		else {
																			swap(v1, v5); swap(v5, v7); // afcdehgb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afechdbg
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // afecdhbg
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v6, v7); // afecdbhg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // afchedbg
																			}
																			else {
																				swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afcehdbg
																			}
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v1, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // afcedhbg
																			}
																			else {
																				swap(v1, v5); swap(v3, v4); swap(v6, v7); // afcedbhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afchdebg
																		}
																		else {
																			swap(v1, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afcdhebg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v1, v5); swap(v5, v7); swap(v6, v7); // afcdehbg
																		}
																		else {
																			swap(v1, v5); swap(v6, v7); // afcdebhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // afechgbd
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v1, v5); swap(v3, v7); swap(v5, v6); // afchegbd
																		}
																		else {
																			swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // afcehgbd
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); // afechbgd
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v3, v7); // afchebgd
																			}
																			else {
																				swap(v1, v5); swap(v3, v4); swap(v4, v7); // afcehbgd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // afechbdg
																		}
																		else {
																			if (compare(v7, v4)) {
																				swap(v1, v5); swap(v3, v7); swap(v6, v7); // afchebdg
																			}
																			else {
																				swap(v1, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // afcehbdg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // afecbhgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afecbhdg
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // afecbdhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // afcebhgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afcebhdg
																			}
																			else {
																				swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // afcebdhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); // afebchgd
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // afebchdg
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v6, v7); // afebcdhg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); // achfegdb
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); swap(v6, v7); // acfhegdb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acfehgdb
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // achfegbd
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v6); // acfhegbd
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // acfehgbd
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v5)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v5); // achfedgb
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); // acfhedgb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); // acfehdgb
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); // acfedhgb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); // achfdegb
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); // acfhdegb
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); // acfdhegb
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v5, v7); // acfdehgb
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v4, v5); // achdfegb
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); // acdhfegb
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); // acdfhegb
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); // acdfehgb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v4)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v6, v7); // achfedbg
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); // achfebgd
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // achfebdg
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v5, v7); swap(v6, v7); // acfhedbg
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v7); // acfhebgd
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v6, v7); // acfhebdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v6, v7); // achfdebg
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acfhdebg
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acfdhebg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v6, v7); // achdfebg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v6, v7); // acdhfebg
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acdfhebg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acfehdbg
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); // acfehbgd
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // acfehbdg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v7); swap(v6, v7); // acfedhbg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v5, v7); swap(v6, v7); // acfdehbg
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v7); swap(v6, v7); // acdfehbg
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v6, v7); // acfedbhg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v6, v7); // acfdebhg
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v6, v7); // acdfebhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // acfebhgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acfebhdg
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // acfebdhg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v1)) {
												if (compare(v2, v1)) {
													if (compare(v5, v2)) {
														if (compare(v5, v0)) {
															if (compare(v7, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); // hfacbegd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fhacbegd
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hfacbedg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fhacbedg
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hfacdbeg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhacdbeg
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hfacbdeg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fhacbdeg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); // fahcbegd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fahcbedg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fahcdbeg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fahcbdeg
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // fachdbeg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v4, v7); swap(v6, v7); // facdhbeg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v5); // fachbegd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fachbedg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fachbdeg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hafcbegd
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ahfcbegd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); // afhcbegd
																		}
																		else {
																			swap(v1, v5); swap(v3, v7); swap(v4, v5); // afchbegd
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hafcbedg
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahfcbedg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // afhcbedg
																		}
																		else {
																			swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // afchbedg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hafcdbeg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahfcdbeg
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // afhcdbeg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // afchdbeg
																		}
																		else {
																			swap(v1, v5); swap(v4, v7); swap(v6, v7); // afcdhbeg
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hafcbdeg
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ahfcbdeg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afhcbdeg
																		}
																		else {
																			swap(v1, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afchbdeg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v7, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hacfdbeg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hacdfbeg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hacfbegd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hacfbedg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hacfbdeg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v5); // hacbfegd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hacbfedg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v4, v7); swap(v6, v7); // hacdbfeg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hacbfdeg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // hacbdfeg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahcfdbeg
																		}
																		else {
																			swap(v1, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahcdfbeg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ahcfbegd
																			}
																			else {
																				swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahcfbedg
																			}
																		}
																		else {
																			swap(v1, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ahcfbdeg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v3, v7); swap(v4, v5); // ahcbfegd
																		}
																		else {
																			swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ahcbfedg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v7); swap(v4, v7); swap(v6, v7); // ahcdbfeg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahcbfdeg
																			}
																			else {
																				swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ahcbdfeg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v7, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v5)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // achfdbeg
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v7); swap(v6, v7); // acfhdbeg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v1, v2); swap(v2, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // achdfbeg
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v6, v7); // achdbfeg
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acdhfbeg
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v6, v7); // acdhbfeg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v4, v7); swap(v6, v7); // acfdhbeg
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v7); swap(v6, v7); // acdfhbeg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // achfbegd
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // achfbedg
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // achfbdeg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); // acfhbegd
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // acfhbedg
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acfhbdeg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); // achbfegd
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // achbfedg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // achbfdeg
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // achbdfeg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v0)) {
														if (compare(v6, v3)) {
															if (compare(v5, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // hfabegcd
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); // hfabecgd
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hfabcegd
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); // fhabegcd
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); // fhabecgd
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); // fhabcegd
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); // hafbegcd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); // hafbecgd
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); // hafbcegd
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // habfegcd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); // habfecgd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // habfcegd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); // habcfegd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v2)) {
																		swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // hfabecdg
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hfabcedg
																		}
																		else {
																			swap(v0, v7); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // hfabcdeg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v6, v7); // fhabecdg
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fhabcedg
																		}
																		else {
																			swap(v0, v5); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fhabcdeg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v6, v7); // hafbecdg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hafbcedg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // habfecdg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // habfcedg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // habcfedg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hafbcdeg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // habfcdeg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // habcfdeg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // habcdfeg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v7, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v5, v6); // ahfbegcd
																		}
																		else {
																			swap(v1, v7); swap(v2, v5); swap(v3, v7); // ahfbecgd
																		}
																	}
																	else {
																		swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); // ahfbcegd
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ahbfegcd
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); // ahbfecgd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ahbfcegd
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); // ahbcfegd
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fahbegcd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); // fahbecgd
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fahbcegd
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // afhbegcd
																		}
																		else {
																			swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); // afhbecgd
																		}
																	}
																	else {
																		swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // afhbcegd
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fahbecdg
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fahbcedg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fahbcdeg
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v6, v7); // ahfbecdg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ahfbcedg
																			}
																			else {
																				swap(v1, v7); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahfbcdeg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // afhbecdg
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afhbcedg
																			}
																			else {
																				swap(v1, v5); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // afhbcdeg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ahbfecdg
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahbfcedg
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // ahbfcdeg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ahbcfedg
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahbcfdeg
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // ahbcdfeg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v7, v4)) {
														if (compare(v2, v1)) {
															if (compare(v7, v3)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // facbhegd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // facbhedg
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // facbhdeg
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // afcbhegd
																			}
																			else {
																				swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afcbhedg
																			}
																		}
																		else {
																			swap(v1, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // afcbhdeg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // acfbhegd
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acfbhedg
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // acfbhdeg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // facdbheg
																		}
																		else {
																			swap(v1, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afcdbheg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acfdbheg
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acdfbheg
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // facbdheg
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afcbdheg
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acfbdheg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); // fabhegcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); // fabhecgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v6, v7); // fabhecdg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); // fabhcegd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fabhcedg
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fabhcdeg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fabchegd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fabchedg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fabchdeg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fabcdheg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v5, v6); // afbhegcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v7); // afbhecgd
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v6, v7); // afbhecdg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); // afbhcegd
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // afbhcedg
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afbhcdeg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // afbchegd
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afbchedg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // afbchdeg
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afbcdheg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v4, v3)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v3)) {
																		swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v5, v7); // facbehgd
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // facbehdg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v6, v7); // facbedhg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v3, v5); swap(v5, v7); // afcbehgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // afcbehdg
																			}
																			else {
																				swap(v1, v5); swap(v3, v5); swap(v6, v7); // afcbedhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); // acfbehgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // acfbehdg
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v6, v7); // acfbedhg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v4, v5); swap(v6, v7); // facdbehg
																		}
																		else {
																			swap(v1, v5); swap(v4, v5); swap(v6, v7); // afcdbehg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v6, v7); // acfdbehg
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v6, v7); // acdfbehg
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // facbdehg
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // afcbdehg
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // acfbdehg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fabehgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); // fabehcgd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fabehcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fabechgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fabechdg
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fabecdhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); // fabcehgd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fabcehdg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v6, v7); // fabcedhg
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fabcdehg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // afbehgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); // afbehcgd
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // afbehcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // afbechgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afbechdg
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // afbecdhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); // afbcehgd
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // afbcehdg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v6, v7); // afbcedhg
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // afbcdehg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v2)) {
															if (compare(v7, v4)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v2)) {
																		swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // abhfegcd
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v7); swap(v3, v5); swap(v5, v7); // abhfecgd
																		}
																		else {
																			swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // abhfecdg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		swap(v2, v5); swap(v3, v7); swap(v5, v6); // abfhegcd
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v5); swap(v3, v7); // abfhecgd
																		}
																		else {
																			swap(v2, v5); swap(v3, v7); swap(v6, v7); // abfhecdg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // abfehgcd
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v5); swap(v3, v4); swap(v4, v7); // abfehcgd
																		}
																		else {
																			swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // abfehcdg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // abfechgd
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abfechdg
																		}
																		else {
																			swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // abfecdhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // abhfcegd
																		}
																		else {
																			swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abhfcedg
																		}
																	}
																	else {
																		swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // abhfcdeg
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v2, v5); swap(v3, v7); swap(v4, v5); // abfhcegd
																		}
																		else {
																			swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // abfhcedg
																		}
																	}
																	else {
																		swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abfhcdeg
																	}
																}
															}
															else {
																if (compare(v7, v4)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // abfchegd
																		}
																		else {
																			swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abfchedg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // abfchdeg
																		}
																		else {
																			swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abfcdheg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v2, v5); swap(v3, v5); swap(v5, v7); // abfcehgd
																		}
																		else {
																			swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // abfcehdg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v2, v5); swap(v3, v5); swap(v6, v7); // abfcedhg
																		}
																		else {
																			swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // abfcdehg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v7, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v3, v7); swap(v4, v5); // acbhfegd
																		}
																		else {
																			swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v6, v7); // acbhfedg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acbhfdeg
																		}
																		else {
																			swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v6, v7); // acbhdfeg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v4)) {
																			swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); // acbfhegd
																		}
																		else {
																			swap(v1, v2); swap(v3, v5); swap(v5, v7); // acbfehgd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acbfhedg
																			}
																			else {
																				swap(v1, v2); swap(v3, v5); swap(v5, v7); swap(v6, v7); // acbfehdg
																			}
																		}
																		else {
																			swap(v1, v2); swap(v3, v5); swap(v4, v7); swap(v6, v7); // acbfhdeg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v4)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v5)) {
																			swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v6, v7); // acdbhfeg
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acdbfheg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acbfdheg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v2); swap(v4, v7); swap(v6, v7); // acbdhfeg
																			}
																			else {
																				swap(v1, v2); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acbdfheg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v3, v5); swap(v6, v7); // acbfedhg
																		}
																		else {
																			swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v6, v7); // acbfdehg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v6, v7); // acdbfehg
																		}
																		else {
																			swap(v1, v2); swap(v4, v5); swap(v6, v7); // acbdfehg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v2, v7); swap(v3, v7); swap(v4, v5); // abhcfegd
																		}
																		else {
																			swap(v3, v7); swap(v4, v5); // abchfegd
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v3, v5); swap(v4, v7); swap(v5, v7); // abcfhegd
																		}
																		else {
																			swap(v3, v5); swap(v5, v7); // abcfehgd
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // abhcfedg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v3, v7); swap(v4, v5); swap(v6, v7); // abchfedg
																			}
																			else {
																				swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abcfhedg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v3, v5); swap(v5, v7); swap(v6, v7); // abcfehdg
																		}
																		else {
																			swap(v3, v5); swap(v6, v7); // abcfedhg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abhcfdeg
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abchfdeg
																			}
																			else {
																				swap(v3, v5); swap(v4, v7); swap(v6, v7); // abcfhdeg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abcfdheg
																		}
																		else {
																			swap(v3, v5); swap(v4, v5); swap(v6, v7); // abcfdehg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // abhcdfeg
																		}
																		else {
																			swap(v3, v7); swap(v4, v7); swap(v6, v7); // abchdfeg
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			if (compare(v7, v5)) {
																				swap(v4, v7); swap(v6, v7); // abcdhfeg
																			}
																			else {
																				swap(v4, v5); swap(v5, v7); swap(v6, v7); // abcdfheg
																			}
																		}
																		else {
																			swap(v4, v5); swap(v6, v7); // abcdfehg
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v7, v4)) {
										if (compare(v7, v1)) {
											if (compare(v6, v0)) {
												if (compare(v7, v5)) {
													if (compare(v7, v0)) {
														if (compare(v4, v1)) {
															if (compare(v5, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); // ghfeacdb
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); // ghfaecdb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); // ghfacedb
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ghfacdeb
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ghfeacbd
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghfeabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v6, v7); // ghfaecbd
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ghfacebd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ghfaebcd
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); // ghafecdb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ghafcedb
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); // ghafcdeb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); // ghacfedb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // ghacfdeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); // ghacdfeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ghafecbd
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); // ghafebcd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghafcebd
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ghacfebd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v5, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghfacdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghfacbed
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ghfacbde
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // ghafcdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ghafcbed
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghafcbde
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghacfdbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // ghacdfbe
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ghacfbed
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // ghacfbde
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ghacbfed
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); // ghacbfde
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghacdbfe
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // ghacbdfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghfabecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ghfabced
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v4, v7); // ghfabcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ghafbecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghafbced
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // ghafbcde
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); // ghabfecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ghabfced
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghabfcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); // ghabcfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // ghabcfde
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghabcdfe
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v4, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gahfecdb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gahfcedb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gahfcdeb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gahcfedb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gahcfdeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gahcdfeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gachfedb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gachfdeb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // gachdfeb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v4, v7); swap(v6, v7); // gacdhfeb
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); // gahfcdbe
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gahcfdbe
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); // gahcdfbe
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gahcdbfe
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gachfdbe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v7); // gachdfbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v4, v7); // gacdhfbe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // gachdbfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v4, v7); swap(v5, v6); // gacdhbfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); // gahfecbd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gahfcebd
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gahfcbed
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gahfcbde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v4, v1)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gahfebcd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gahfbecd
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gahfbced
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gahfbcde
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); // gahcfebd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gahcfbed
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gahcfbde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v5); // gachfebd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gachfbed
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gachfbde
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); // gahcbfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gahcbfde
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gahcbdfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v6); // gachbfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gachbfde
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gachbdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gahbfecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gahbfced
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gahbfcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gahbcfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); // gahbcfde
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gahbcdfe
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v1)) {
														if (compare(v3, v1)) {
															if (compare(v4, v3)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gfheacdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgheacdb
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); // gfhaecdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fghaecdb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gfahecdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v6, v7); // fgahecdb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v6, v7); // gafhecdb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gfhacedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fghacedb
																			}
																		}
																		else {
																			if (compare(v5, v0)) {
																				if (compare(v6, v5)) {
																					swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfahcedb
																				}
																				else {
																					swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fgahcedb
																				}
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gafhcedb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gfachedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgachedb
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gafchedb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gacfhedb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfhacdeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fghacdeb
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gfahcdeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgahcdeb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfachdeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // fgachdeb
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gfacdheb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgacdheb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gafhcdeb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gafchdeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gafcdheb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // gacfhdeb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gacfdheb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gacdfheb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfheacbd
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gfheabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gfhaecbd
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfhacebd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v6, v7); // gfhaebcd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fgheacbd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgheabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); // fghaecbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fghacebd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fghaebcd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); // gfahecbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); // fgahecbd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); // gafhecbd
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); // gfahebcd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); // fgahebcd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); // gafhebcd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gfahcebd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); // fgahcebd
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfachebd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fgachebd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); // gafhcebd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gafchebd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gacfhebd
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v2, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gfhacbed
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fghacbed
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); // gfahcbed
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fgahcbed
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gafhcbed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gfachbed
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgachbed
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gafchbed
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gacfhbed
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gfhabecd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fghabecd
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); // gfahbecd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // fgahbecd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gafhbecd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfhabced
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fghabced
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gfahbced
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); // fgahbced
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); // gafhbced
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gfhacdbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); // fghacdbe
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfahcdbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // fgahcdbe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gfachdbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); // fgachdbe
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfacdhbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // fgacdhbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // gafhcdbe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); // gafchdbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gafcdhbe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v7); // gacfhdbe
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // gacfdhbe
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v4, v5); swap(v5, v7); // gacdfhbe
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v7); // gfhacbde
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fghacbde
																			}
																		}
																		else {
																			if (compare(v5, v0)) {
																				if (compare(v6, v5)) {
																					swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gfahcbde
																				}
																				else {
																					swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgahcbde
																				}
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gafhcbde
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); // gfachbde
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fgachbde
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gafchbde
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gacfhbde
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gfhabcde
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fghabcde
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfahbcde
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // fgahbcde
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gafhbcde
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v2)) {
													if (compare(v4, v1)) {
														if (compare(v7, v2)) {
															if (compare(v3, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // aghfecdb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aghfcedb
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // aghfcdeb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v6, v7); // agfhecdb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // agfhcedb
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agfhcdeb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // faghecdb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // faghcedb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // faghcdeb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // afghecdb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afghcedb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // afghcdeb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v7, v5)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); // aghfecbd
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // aghfcebd
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // aghfebcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); // agfhecbd
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); // agfhcebd
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); // agfhebcd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); // faghecbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // faghcebd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); // faghebcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v4, v2)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); // afghecbd
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // afghcebd
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v7); // afghebcd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v4, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // fagchedb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agfchedb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // afgchedb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fagchebd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // agfchebd
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afgchebd
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fagchdeb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // agfchdeb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afgchdeb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // fagcdheb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agfcdheb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // afgcdheb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // facghedb
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // facghebd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // facghdeb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // facgdheb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v4, v6); swap(v5, v7); // facdgheb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // afcghedb
																		}
																		else {
																			swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afcghebd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afcghdeb
																			}
																			else {
																				swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // afcgdheb
																			}
																		}
																		else {
																			swap(v1, v5); swap(v4, v6); swap(v5, v7); // afcdgheb
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v2)) {
															if (compare(v4, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aghfbecd
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // agfhbecd
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // aghbfecd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aghfcbed
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // agfhcbed
																			}
																		}
																		else {
																			if (compare(v5, v1)) {
																				if (compare(v7, v5)) {
																					swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // aghfbced
																				}
																				else {
																					swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); // agfhbced
																				}
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aghbfced
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); // faghcbed
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); // faghbecd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // faghbced
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); // afghcbed
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); // afghbecd
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // afghbced
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); // aghfcdbe
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // agfhcdbe
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // aghfcbde
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agfhcbde
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aghfbcde
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // agfhbcde
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // aghbfcde
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // faghcdbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // faghcbde
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // faghbcde
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afghcdbe
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // afghcbde
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afghbcde
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // fagchdbe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); // agfchdbe
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // afgchdbe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fagcdhbe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // agfcdhbe
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afgcdhbe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // facghdbe
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // facgdhbe
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // facdghbe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // afcghdbe
																			}
																			else {
																				swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afcgdhbe
																			}
																		}
																		else {
																			swap(v1, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afcdghbe
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fagchbed
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // facghbed
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agfchbed
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // afgchbed
																			}
																		}
																		else {
																			swap(v1, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // afcghbed
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); // fagchbde
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v7); // facghbde
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // agfchbde
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v7); // afgchbde
																			}
																		}
																		else {
																			swap(v1, v5); swap(v3, v6); swap(v4, v7); // afcghbde
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v2)) {
														if (compare(v3, v1)) {
															if (compare(v4, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // aghcfedb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // agchfedb
																			}
																			else {
																				swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agcfhedb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			if (compare(v7, v2)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aghcfdeb
																			}
																			else {
																				swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agchfdeb
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // agcfhdeb
																			}
																			else {
																				swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agcfdheb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // aghcdfeb
																		}
																		else {
																			swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // agchdfeb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v1, v6); swap(v4, v7); swap(v6, v7); // agcdhfeb
																		}
																		else {
																			swap(v1, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agcdfheb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // aghcfdbe
																		}
																		else {
																			swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // agchfdbe
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v3, v5); swap(v4, v7); // agcfhdbe
																		}
																		else {
																			swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // agcfdhbe
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v3)) {
																			if (compare(v7, v2)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); // aghcdfbe
																			}
																			else {
																				swap(v1, v6); swap(v3, v7); swap(v4, v7); // agchdfbe
																			}
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v4, v7); // agcdhfbe
																			}
																			else {
																				swap(v1, v6); swap(v4, v5); swap(v5, v7); // agcdfhbe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // aghcdbfe
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // agchdbfe
																			}
																			else {
																				swap(v1, v6); swap(v4, v7); swap(v5, v6); // agcdhbfe
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v1)) {
																	if (compare(v7, v2)) {
																		swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); // aghcfebd
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v1, v6); swap(v3, v7); swap(v4, v5); // agchfebd
																		}
																		else {
																			swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // agcfhebd
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // aghcfbed
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // agchfbed
																			}
																			else {
																				swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agcfhbed
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aghcfbde
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agchfbde
																			}
																			else {
																				swap(v1, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // agcfhbde
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); // aghcbfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // aghcbfde
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aghcbdfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v3, v7); swap(v4, v6); // agchbfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // agchbfde
																			}
																			else {
																				swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agchbdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // aghbcfed
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); // aghbcfde
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // aghbcdfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v1)) {
															if (compare(v7, v3)) {
																if (compare(v4, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v6, v7); // acghfedb
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acgfhedb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); // acfghedb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); // acghfebd
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // acgfhebd
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acfghebd
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acghfdeb
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v6, v7); // acghdfeb
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v6, v7); // acgfhdeb
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acfghdeb
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acgfdheb
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); // acfgdheb
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v6, v7); // acgdhfeb
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acgdfheb
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v6, v7); // acdghfeb
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acdgfheb
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); // acfdgheb
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); // acdfgheb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v7); // acghfdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // acghfbed
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acghfbde
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); // acghdfbe
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v4, v7); // acgdhfbe
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); // acdghfbe
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); // acghdbfe
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); // acgdhbfe
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); // acdghbfe
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); // acghbfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // acghbfde
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acghbdfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); // acgfhdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acgfhbed
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // acgfhbde
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v6); // acfghdbe
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // acfghbed
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v7); // acfghbde
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v7); // acgfdhbe
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acfgdhbe
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acfdghbe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v7); // acgdfhbe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v7); // acdgfhbe
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acdfghbe
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v5, v1)) {
												if (compare(v2, v1)) {
													if (compare(v6, v2)) {
														if (compare(v6, v0)) {
															if (compare(v5, v0)) {
																if (compare(v7, v3)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfacbhed
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fgacbhed
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gfacbhde
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgacbhde
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gfacdbhe
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgacdbhe
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfacbdhe
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fgacbdhe
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gafcbhed
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gafcbhde
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gafcdbhe
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gafcbdhe
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gacfdbhe
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gacdfbhe
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gacfbhed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gacfbhde
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gacfbdhe
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // agfcbhed
																		}
																		else {
																			swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // agcfbhed
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fagcbhed
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afgcbhed
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agfcbhde
																		}
																		else {
																			swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agcfbhde
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fagcbhde
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // afgcbhde
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // fagcdbhe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agfcdbhe
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // afgcdbhe
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agcfdbhe
																		}
																		else {
																			swap(v1, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agcdfbhe
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // agfcbdhe
																		}
																		else {
																			swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // agcfbdhe
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fagcbdhe
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afgcbdhe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v2)) {
															if (compare(v5, v0)) {
																if (compare(v6, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // facgdbhe
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v4, v6); swap(v6, v7); // facdgbhe
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // facgbhed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // facgbhde
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // facgbdhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // facbghed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // facbghde
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // facbgdhe
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // facdbghe
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // facbdghe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // afcgdbhe
																		}
																		else {
																			swap(v1, v5); swap(v4, v6); swap(v6, v7); // afcdgbhe
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afcgbhed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // afcgbhde
																			}
																			else {
																				swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afcgbdhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // afcbghed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afcbghde
																			}
																			else {
																				swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // afcbgdhe
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afcdbghe
																		}
																		else {
																			swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afcbdghe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acgfdbhe
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v6, v7); // acfgdbhe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v6, v7); // acfdgbhe
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acfdbghe
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acgdfbhe
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acdgfbhe
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v6, v7); // acdfgbhe
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acdfbghe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // acgfbhed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acgfbhde
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // acgfbdhe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acfgbhed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // acfgbhde
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acfgbdhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // acfbghed
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acfbghde
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // acfbgdhe
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acfbdghe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v0)) {
														if (compare(v4, v3)) {
															if (compare(v6, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gfabhecd
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfabhced
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gfabched
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgabhecd
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fgabhced
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgabched
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fagbhecd
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fagbhced
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fagbched
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fabghecd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // fabghced
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fabgched
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fabcghed
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v2)) {
																		swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gfabhcde
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfabchde
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gfabcdhe
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); // fgabhcde
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fgabchde
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgabcdhe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fagbhcde
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fagbchde
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); // fabghcde
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // fabgchde
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fabcghde
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fagbcdhe
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fabgcdhe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fabcgdhe
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fabcdghe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v3)) {
															if (compare(v6, v5)) {
																if (compare(v6, v0)) {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gafbhecd
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gafbhced
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gafbched
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agfbhecd
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // agfbhced
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agfbched
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v4, v2)) {
																		swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // afgbhecd
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afgbhced
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // afgbched
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afbghecd
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // afbghced
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afbgched
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // afbcghed
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); // gafbhcde
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gafbchde
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gafbcdhe
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v7); // agfbhcde
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // agfbchde
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agfbcdhe
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // afgbhcde
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afgbchde
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // afgbcdhe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v7); // afbghcde
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // afbgchde
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afbgcdhe
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afbcghde
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // afbcgdhe
																		}
																	}
																	else {
																		swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afbcdghe
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v1)) {
													if (compare(v7, v5)) {
														if (compare(v2, v1)) {
															if (compare(v7, v3)) {
																if (compare(v6, v0)) {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gacbhfed
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v7); // gacbhfde
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gacbhdfe
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // agcbhfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v3, v6); swap(v4, v7); // agcbhfde
																			}
																			else {
																				swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // agcbhdfe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // acgbhfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); // acgbhfde
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // acgbhdfe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gacdbhfe
																		}
																		else {
																			swap(v1, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agcdbhfe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acgdbhfe
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acdgbhfe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gacbdhfe
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agcbdhfe
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acgbdhfe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); // gabhfecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gabhfced
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gabhfcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); // gabhcfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gabhcfde
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gabhcdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gabchfed
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); // gabchfde
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gabchdfe
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gabcdhfe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); // agbhfecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // agbhfced
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agbhfcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); // agbhcfed
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // agbhcfde
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agbhcdfe
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // agbchfed
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); // agbchfde
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // agbchdfe
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agbcdhfe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v5, v3)) {
																if (compare(v6, v0)) {
																	if (compare(v4, v3)) {
																		swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gacbfhed
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gacbfhde
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gacbfdhe
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agcbfhed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // agcbfhde
																			}
																			else {
																				swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agcbfdhe
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acgbfhed
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // acgbfhde
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acgbfdhe
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v4, v6); swap(v6, v7); // gacdbfhe
																		}
																		else {
																			swap(v1, v6); swap(v4, v6); swap(v6, v7); // agcdbfhe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v6, v7); // acgdbfhe
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v6, v7); // acdgbfhe
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gacbdfhe
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // agcbdfhe
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // acgbdfhe
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gabfhecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gabfhced
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // gabfhcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gabfched
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gabfchde
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // gabfcdhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gabcfhed
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gabcfhde
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gabcfdhe
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // gabcdfhe
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // agbfhecd
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agbfhced
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // agbfhcde
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // agbfched
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agbfchde
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // agbfcdhe
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agbcfhed
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // agbcfhde
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agbcfdhe
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // agbcdfhe
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v2)) {
														if (compare(v7, v2)) {
															if (compare(v7, v5)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		swap(v2, v6); swap(v3, v7); swap(v4, v5); // abghfecd
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // abghfced
																		}
																		else {
																			swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abghfcde
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v2, v6); swap(v3, v7); swap(v4, v6); // abghcfed
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // abghcfde
																		}
																		else {
																			swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abghcdfe
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v4, v2)) {
																		swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // abgfhecd
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abgfhced
																		}
																		else {
																			swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // abgfhcde
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abfghecd
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // abfghced
																		}
																		else {
																			swap(v2, v5); swap(v3, v6); swap(v4, v7); // abfghcde
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // abgfched
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abgfchde
																		}
																		else {
																			swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // abgfcdhe
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abfgched
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // abfgchde
																		}
																		else {
																			swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abfgcdhe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // abgchfed
																		}
																		else {
																			swap(v2, v6); swap(v3, v6); swap(v4, v7); // abgchfde
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // abgchdfe
																		}
																		else {
																			swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abgcdhfe
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abgcfhed
																		}
																		else {
																			swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // abgcfhde
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abgcfdhe
																		}
																		else {
																			swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // abgcdfhe
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v5, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v5)) {
																			swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v6, v7); // acbghfed
																		}
																		else {
																			swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acbgfhed
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v2); swap(v3, v6); swap(v4, v7); // acbghfde
																			}
																			else {
																				swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v7); // acbgfhde
																			}
																		}
																		else {
																			swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acbgfdhe
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); // acbfghed
																		}
																		else {
																			swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acbfghde
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v6, v7); // acbfgdhe
																		}
																		else {
																			swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acbfdghe
																		}
																	}
																}
															}
															else {
																if (compare(v7, v5)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); // acbghdfe
																		}
																		else {
																			swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acbgdhfe
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acdbghfe
																		}
																		else {
																			swap(v1, v2); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acbdghfe
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v6, v7); // acdbgfhe
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acdbfghe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v6, v7); // acbgdfhe
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v2); swap(v4, v6); swap(v6, v7); // acbdgfhe
																			}
																			else {
																				swap(v1, v2); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acbdfghe
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v3, v6); swap(v4, v7); swap(v6, v7); // abcghfed
																		}
																		else {
																			swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abcgfhed
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // abfcghed
																		}
																		else {
																			swap(v3, v5); swap(v4, v6); swap(v5, v7); // abcfghed
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v3, v6); swap(v4, v7); // abcghfde
																			}
																			else {
																				swap(v3, v6); swap(v4, v5); swap(v5, v7); // abcgfhde
																			}
																		}
																		else {
																			swap(v3, v6); swap(v4, v7); swap(v5, v6); // abcghdfe
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abfcghde
																		}
																		else {
																			swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abcfghde
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // abfcgdhe
																		}
																		else {
																			swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abfcdghe
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abcgfdhe
																			}
																			else {
																				swap(v3, v5); swap(v4, v6); swap(v6, v7); // abcfgdhe
																			}
																		}
																		else {
																			swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abcfdghe
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v5)) {
																			swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abcgdhfe
																		}
																		else {
																			swap(v3, v6); swap(v4, v6); swap(v6, v7); // abcgdfhe
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v4, v6); swap(v5, v7); swap(v6, v7); // abcdghfe
																			}
																			else {
																				swap(v4, v6); swap(v6, v7); // abcdgfhe
																			}
																		}
																		else {
																			swap(v4, v5); swap(v5, v6); swap(v6, v7); // abcdfghe
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v1)) {
											if (compare(v5, v0)) {
												if (compare(v6, v4)) {
													if (compare(v7, v1)) {
														if (compare(v3, v1)) {
															if (compare(v7, v3)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v7, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfehacdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // fgehacdb
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gfeahcdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // fgeahcdb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gfaehcdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // fgaehcdb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fagehcdb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); // gfeachdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fgeachdb
																			}
																		}
																		else {
																			if (compare(v6, v0)) {
																				if (compare(v6, v5)) {
																					swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gfaechdb
																				}
																				else {
																					swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // fgaechdb
																				}
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // fagechdb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); // gfacehdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // fgacehdb
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); // fagcehdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v5, v7); // facgehdb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gfeacdhb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v6, v7); // fgeacdhb
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfaecdhb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // fgaecdhb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gfacedhb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v6, v7); // fgacedhb
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfacdehb
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // fgacdehb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fagecdhb
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // fagcedhb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fagcdehb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // facgedhb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // facgdehb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // facdgehb
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gfehacbd
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); // gfehabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gfeahcbd
																			}
																			else {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gfeachbd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gfeahbcd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); // fgehacbd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // fgehabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fgeahcbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); // fgeachbd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fgeahbcd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gfaehcbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); // fgaehcbd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fagehcbd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); // gfaehbcd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // fgaehbcd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); // fagehbcd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gfaechbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fgaechbd
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gfacehbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); // fgacehbd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fagechbd
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fagcehbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // facgehbd
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v2, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v6, v7); // gfeacbhd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fgeacbhd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gfaecbhd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fgaecbhd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fagecbhd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v6, v7); // gfacebhd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fgacebhd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v6, v7); // fagcebhd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v6, v7); // facgebhd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gfeabhcd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fgeabhcd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gfaebhcd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fgaebhcd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // fagebhcd
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gfeabchd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgeabchd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gfaebchd
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fgaebchd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fagebchd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // gfeacdbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); // fgeacdbh
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gfaecdbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fgaecdbh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // gfacedbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); // fgacedbh
																			}
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gfacdebh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fgacdebh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // fagecdbh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // fagcedbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // fagcdebh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v5, v6); // facgedbh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // facgdebh
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v4, v6); swap(v5, v6); // facdgebh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); // gfeacbdh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // fgeacbdh
																			}
																		}
																		else {
																			if (compare(v6, v0)) {
																				if (compare(v6, v5)) {
																					swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // gfaecbdh
																				}
																				else {
																					swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fgaecbdh
																				}
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // fagecbdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); // gfacebdh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // fgacebdh
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); // fagcebdh
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v6); // facgebdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v0)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // gfeabcdh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fgeabcdh
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			if (compare(v6, v5)) {
																				swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gfaebcdh
																			}
																			else {
																				swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // fgaebcdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fagebcdh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v4, v0)) {
														if (compare(v7, v1)) {
															if (compare(v6, v0)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // feghacdb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // fegahcdb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fegachdb
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // fegacdhb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // feghacbd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); // feghabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // fegahcbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // fegachbd
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // fegahbcd
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); // feaghcdb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // feagchdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // feagcdhb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // feacghdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // feacgdhb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // feacdghb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v2, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // feaghcbd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // feaghbcd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // feagchbd
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // feacghbd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v2, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v6, v0)) {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // fegacdbh
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fegacbhd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fegacbdh
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // feagcdbh
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // feagcbhd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // feagcbdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // feacgdbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // feacdgbh
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // feacgbhd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // feacgbdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // feacbghd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // feacbgdh
																			}
																		}
																		else {
																			if (compare(v3, v1)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // feacdbgh
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v5); // feacbdgh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); // fegabhcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fegabchd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // fegabcdh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // feagbhcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v6, v7); // feagbchd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v6); // feagbcdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // feabghcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // feabgchd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // feabgcdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // feabcghd
																			}
																			else {
																				swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // feabcgdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // feabcdgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v3, v1)) {
															if (compare(v7, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // faeghcdb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); // faegchdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // faegcdhb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // faecghdb
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // faecgdhb
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // faecdghb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // faceghdb
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // facegdhb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // facedghb
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v5, v6); swap(v6, v7); // facdeghb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // faegcdbh
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // faecgdbh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // faecdgbh
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); // faecdbgh
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // facegdbh
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v5, v6); // facedgbh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v5, v6); // facdegbh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v4); // facedbgh
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); // facdebgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // faeghcbd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // faegchbd
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v6, v7); // faegcbhd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); // faegcbdh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v1)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // faeghbcd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // faegbhcd
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // faegbchd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // faegbcdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // faecghbd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // faecgbhd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // faecgbdh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // faceghbd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // facegbhd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v6); // facegbdh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // faecbghd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // faecbgdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // faecbdgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // facebghd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // facebgdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v4); swap(v4, v5); // facebdgh
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // faebghcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // faebgchd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // faebgcdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // faebcghd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // faebcgdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v4); swap(v3, v5); // faebcdgh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v2)) {
													if (compare(v7, v1)) {
														if (compare(v4, v2)) {
															if (compare(v3, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gafehcdb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gafechdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gafecdhb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // agfehcdb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agfechdb
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // agfecdhb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // afgehcdb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // afgechdb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afgecdhb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // afeghcdb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); // afegchdb
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // afegcdhb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); // gafehcbd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gafechbd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gafehbcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); // agfehcbd
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // agfechbd
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // agfehbcd
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // afgehcbd
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afgechbd
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); // afgehbcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v7, v2)) {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // afeghcbd
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // afegchbd
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // afeghbcd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v7, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gafcehdb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // agfcehdb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); // afgcehdb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); // gafcehbd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v7); // agfcehbd
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // afgcehbd
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v6, v7); // gafcedhb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v6, v7); // agfcedhb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // afgcedhb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gafcdehb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // agfcdehb
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afgcdehb
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gacfehdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v6, v7); // gacfedhb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v5, v7); // gacfehbd
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // gacfdehb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v4, v5); swap(v6, v7); // gacdfehb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // agcfehdb
																			}
																			else {
																				swap(v1, v6); swap(v3, v5); swap(v6, v7); // agcfedhb
																			}
																		}
																		else {
																			swap(v1, v6); swap(v3, v5); swap(v5, v7); // agcfehbd
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // agcfdehb
																		}
																		else {
																			swap(v1, v6); swap(v4, v5); swap(v6, v7); // agcdfehb
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v4, v2)) {
															if (compare(v7, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gafecbhd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gafebhcd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gafebchd
																			}
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agfecbhd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agfebhcd
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // agfebchd
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // afgebhcd
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // afegbhcd
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // afebghcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // afgecbhd
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v6, v7); // afegcbhd
																			}
																		}
																		else {
																			if (compare(v6, v1)) {
																				if (compare(v6, v4)) {
																					swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afgebchd
																				}
																				else {
																					swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // afegbchd
																				}
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afebgchd
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // gafecdbh
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gafecbdh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // gafebcdh
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // agfecdbh
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // agfecbdh
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // agfebcdh
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // afgecdbh
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // afegcdbh
																			}
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // afgecbdh
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); // afegcbdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // afgebcdh
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // afegbcdh
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // afebgcdh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); // gafcedbh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); // agfcedbh
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v5, v6); // afgcedbh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // gafcdebh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); // agfcdebh
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // afgcdebh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v5); // gacfedbh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v5); // gacfdebh
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v4, v5); // gacdfebh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v3, v5); // agcfedbh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v3, v5); swap(v4, v5); // agcfdebh
																			}
																			else {
																				swap(v1, v6); swap(v4, v5); // agcdfebh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gafcebhd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gacfebhd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // agfcebhd
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v6, v7); // afgcebhd
																			}
																		}
																		else {
																			swap(v1, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // agcfebhd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // gafcebdh
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v5, v6); // gacfebdh
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v5, v6); // agfcebdh
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); // afgcebdh
																			}
																		}
																		else {
																			swap(v1, v6); swap(v3, v5); swap(v5, v6); // agcfebdh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v3, v1)) {
															if (compare(v7, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // afecghdb
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v3, v6); swap(v5, v7); // afcgehdb
																			}
																			else {
																				swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // afceghdb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // afcgedhb
																			}
																			else {
																				swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afcgdehb
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afecgdhb
																			}
																			else {
																				swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afcegdhb
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // afecdghb
																		}
																		else {
																			swap(v1, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // afcedghb
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v1, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afcdgehb
																		}
																		else {
																			swap(v1, v5); swap(v5, v6); swap(v6, v7); // afcdeghb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // afecgdbh
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // afecdgbh
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v4); // afecdbgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v3, v6); swap(v5, v6); // afcgedbh
																			}
																			else {
																				swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // afcegdbh
																			}
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v1, v5); swap(v3, v4); swap(v5, v6); // afcedgbh
																			}
																			else {
																				swap(v1, v5); swap(v3, v4); // afcedbgh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // afcgdebh
																		}
																		else {
																			swap(v1, v5); swap(v4, v6); swap(v5, v6); // afcdgebh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v1, v5); swap(v5, v6); // afcdegbh
																		}
																		else {
																			swap(v1, v5); // afcdebgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v4, v2)) {
																		swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afecghbd
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v1, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // afcgehbd
																		}
																		else {
																			swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afceghbd
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // afecgbhd
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v3, v6); swap(v6, v7); // afcgebhd
																			}
																			else {
																				swap(v1, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // afcegbhd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // afecgbdh
																		}
																		else {
																			if (compare(v6, v4)) {
																				swap(v1, v5); swap(v3, v6); // afcgebdh
																			}
																			else {
																				swap(v1, v5); swap(v3, v4); swap(v4, v6); // afcegbdh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afecbghd
																			}
																			else {
																				swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // afecbgdh
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // afecbdgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afcebghd
																			}
																			else {
																				swap(v1, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // afcebgdh
																			}
																		}
																		else {
																			swap(v1, v5); swap(v3, v4); swap(v4, v5); // afcebdgh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // afebcghd
																		}
																		else {
																			swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // afebcgdh
																		}
																	}
																	else {
																		swap(v1, v5); swap(v2, v4); swap(v3, v5); // afebcdgh
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v4, v3)) {
																if (compare(v7, v3)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); swap(v6, v7); // acgfehdb
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); // acfgehdb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); // acfeghdb
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v7); // acgfehbd
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // acfgehbd
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acfeghbd
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v5)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v6, v7); // acgfedhb
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); swap(v6, v7); // acfgedhb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acfegdhb
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); swap(v6, v7); // acfedghb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v6, v7); // acgfdehb
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acfgdehb
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acfdgehb
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v5, v6); swap(v6, v7); // acfdeghb
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v6, v7); // acgdfehb
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v6, v7); // acdgfehb
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acdfgehb
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); swap(v6, v7); // acdfeghb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v4)) {
																if (compare(v4, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v5); // acgfedbh
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // acgfebhd
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v5, v6); // acgfebdh
																			}
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v5, v6); // acfgedbh
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v6, v7); // acfgebhd
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v6); // acfgebdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); // acgfdebh
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); swap(v5, v6); // acfgdebh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v4, v6); swap(v5, v6); // acfdgebh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v4, v5); // acgdfebh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); // acdgfebh
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); swap(v5, v6); // acdfgebh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v6); // acfegdbh
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // acfegbhd
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v6); // acfegbdh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v5, v6); // acfedgbh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v5, v6); // acfdegbh
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v5, v6); // acdfegbh
																			}
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); // acfedbgh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v5); // acfdebgh
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v5); // acdfebgh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acfebghd
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // acfebgdh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v4); swap(v4, v5); // acfebdgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v1)) {
												if (compare(v2, v1)) {
													if (compare(v5, v2)) {
														if (compare(v5, v0)) {
															if (compare(v6, v0)) {
																if (compare(v4, v3)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gfacbehd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fgacbehd
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // gfacbedh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fgacbedh
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // gfacdbeh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fgacdbeh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gfacbdeh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // fgacbdeh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fagcbehd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // fagcbedh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // fagcdbeh
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fagcbdeh
																		}
																	}
																}
																else {
																	if (compare(v3, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v6); // facgdbeh
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v4, v6); // facdgbeh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // facgbehd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v5); // facgbedh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // facgbdeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gafcbehd
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agfcbehd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // afgcbehd
																		}
																		else {
																			swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // afcgbehd
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gafcbedh
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // agfcbedh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); // afgcbedh
																		}
																		else {
																			swap(v1, v5); swap(v3, v6); swap(v4, v5); // afcgbedh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gafcdbeh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // agfcdbeh
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v6); // afgcdbeh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v3, v6); swap(v4, v6); // afcgdbeh
																		}
																		else {
																			swap(v1, v5); swap(v4, v6); // afcdgbeh
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // gafcbdeh
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v5); swap(v4, v6); // agfcbdeh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v5); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // afgcbdeh
																		}
																		else {
																			swap(v1, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // afcgbdeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v2)) {
															if (compare(v6, v0)) {
																if (compare(v5, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // gacfdbeh
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v4, v5); swap(v5, v6); // gacdfbeh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gacfbehd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gacfbedh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v5); swap(v4, v6); // gacfbdeh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gacbfehd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v5); // gacbfedh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v4, v6); // gacdbfeh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gacbfdeh
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v6); // gacbdfeh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // agcfdbeh
																		}
																		else {
																			swap(v1, v6); swap(v4, v5); swap(v5, v6); // agcdfbeh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agcfbehd
																			}
																			else {
																				swap(v1, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // agcfbedh
																			}
																		}
																		else {
																			swap(v1, v6); swap(v3, v5); swap(v4, v6); // agcfbdeh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // agcbfehd
																		}
																		else {
																			swap(v1, v6); swap(v3, v6); swap(v4, v5); // agcbfedh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v6); swap(v4, v6); // agcdbfeh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // agcbfdeh
																			}
																			else {
																				swap(v1, v6); swap(v3, v6); swap(v4, v6); // agcbdfeh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v6, v3)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v5)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v5); swap(v5, v6); // acgfdbeh
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v6); // acfgdbeh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v1, v2); swap(v2, v6); swap(v4, v5); swap(v5, v6); // acgdfbeh
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v4, v6); // acgdbfeh
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v5); swap(v5, v6); // acdgfbeh
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); // acdgbfeh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v4, v6); // acfdgbeh
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v6); // acdfgbeh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acgfbehd
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // acgfbedh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v5); swap(v4, v6); // acgfbdeh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // acfgbehd
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); // acfgbedh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // acfgbdeh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // acgbfehd
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); // acgbfedh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // acgbfdeh
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); // acgbdfeh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v0)) {
														if (compare(v7, v3)) {
															if (compare(v5, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); // gfabehcd
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gfabechd
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gfabcehd
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fgabehcd
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v6, v7); // fgabechd
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fgabcehd
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gafbehcd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v6, v7); // gafbechd
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gafbcehd
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); // gabfehcd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gabfechd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gabfcehd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gabcfehd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v2)) {
																		swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // gfabecdh
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gfabcedh
																		}
																		else {
																			swap(v0, v6); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // gfabcdeh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); // fgabecdh
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // fgabcedh
																		}
																		else {
																			swap(v0, v5); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fgabcdeh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); // gafbecdh
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // gafbcedh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // gabfecdh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gabfcedh
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // gabcfedh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gafbcdeh
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // gabfcdeh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gabcfdeh
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // gabcdfeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v6, v5)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // agfbehcd
																		}
																		else {
																			swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v6, v7); // agfbechd
																		}
																	}
																	else {
																		swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // agfbcehd
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v7); // agbfehcd
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // agbfechd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agbfcehd
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // agbcfehd
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); // fagbehcd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fagbechd
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fagbcehd
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v7); // afgbehcd
																		}
																		else {
																			swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // afgbechd
																		}
																	}
																	else {
																		swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afgbcehd
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v2)) {
																		swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // fagbecdh
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fagbcedh
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // fagbcdeh
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v5); swap(v3, v6); // agfbecdh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); // agfbcedh
																			}
																			else {
																				swap(v1, v6); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // agfbcdeh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v5, v6); // afgbecdh
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // afgbcedh
																			}
																			else {
																				swap(v1, v5); swap(v2, v6); swap(v3, v5); swap(v4, v6); // afgbcdeh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v4, v2)) {
																		swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v5, v6); // agbfecdh
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // agbfcedh
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v5); swap(v4, v6); // agbfcdeh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); // agbcfedh
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // agbcfdeh
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); // agbcdfeh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v1)) {
													if (compare(v6, v4)) {
														if (compare(v2, v1)) {
															if (compare(v6, v3)) {
																if (compare(v5, v0)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // facbgehd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // facbgedh
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v6); // facbgdeh
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afcbgehd
																			}
																			else {
																				swap(v1, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // afcbgedh
																			}
																		}
																		else {
																			swap(v1, v5); swap(v3, v5); swap(v4, v6); // afcbgdeh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acfbgehd
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // acfbgedh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v6); // acfbgdeh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v4, v5); swap(v5, v6); // facdbgeh
																		}
																		else {
																			swap(v1, v5); swap(v4, v5); swap(v5, v6); // afcdbgeh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v4, v5); swap(v5, v6); // acfdbgeh
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); swap(v5, v6); // acdfbgeh
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // facbdgeh
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // afcbdgeh
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // acfbdgeh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // fabgehcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v6, v7); // fabgechd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); // fabgecdh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // fabgcehd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // fabgcedh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // fabgcdeh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // fabcgehd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // fabcgedh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // fabcgdeh
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // fabcdgeh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // afbgehcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v6, v7); // afbgechd
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v3, v6); // afbgecdh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // afbgcehd
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); // afbgcedh
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // afbgcdeh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // afbcgehd
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // afbcgedh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v6); // afbcgdeh
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // afbcdgeh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v4, v3)) {
																if (compare(v5, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // facbeghd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v5, v6); // facbegdh
																		}
																	}
																	else {
																		swap(v0, v5); swap(v1, v5); swap(v3, v5); // facbedgh
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // afcbeghd
																			}
																			else {
																				swap(v1, v5); swap(v3, v5); swap(v5, v6); // afcbegdh
																			}
																		}
																		else {
																			swap(v1, v5); swap(v3, v5); // afcbedgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // acfbeghd
																			}
																			else {
																				swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v5, v6); // acfbegdh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); // acfbedgh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v5); swap(v1, v5); swap(v4, v5); // facdbegh
																		}
																		else {
																			swap(v1, v5); swap(v4, v5); // afcdbegh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v5); swap(v4, v5); // acfdbegh
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v5); swap(v4, v5); // acdfbegh
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		swap(v0, v5); swap(v1, v5); swap(v3, v5); swap(v4, v5); // facbdegh
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v5); swap(v3, v5); swap(v4, v5); // afcbdegh
																		}
																		else {
																			swap(v1, v2); swap(v2, v5); swap(v3, v5); swap(v4, v5); // acfbdegh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // fabeghcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // fabegchd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // fabegcdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // fabecghd
																			}
																			else {
																				swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // fabecgdh
																			}
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // fabecdgh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // fabceghd
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // fabcegdh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); // fabcedgh
																		}
																		else {
																			swap(v0, v5); swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // fabcdegh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // afbeghcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // afbegchd
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v6); // afbegcdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // afbecghd
																			}
																			else {
																				swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // afbecgdh
																			}
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v4); swap(v4, v5); // afbecdgh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // afbceghd
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v5, v6); // afbcegdh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); // afbcedgh
																		}
																		else {
																			swap(v1, v5); swap(v2, v5); swap(v3, v5); swap(v4, v5); // afbcdegh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v2)) {
														if (compare(v4, v2)) {
															if (compare(v6, v4)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v2)) {
																		swap(v2, v6); swap(v3, v5); swap(v5, v7); // abgfehcd
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // abgfechd
																		}
																		else {
																			swap(v2, v6); swap(v3, v5); swap(v5, v6); // abgfecdh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // abfgehcd
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v5); swap(v3, v6); swap(v6, v7); // abfgechd
																		}
																		else {
																			swap(v2, v5); swap(v3, v6); // abfgecdh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abfeghcd
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // abfegchd
																		}
																		else {
																			swap(v2, v5); swap(v3, v4); swap(v4, v6); // abfegcdh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abfecghd
																		}
																		else {
																			swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // abfecgdh
																		}
																	}
																	else {
																		swap(v2, v5); swap(v3, v4); swap(v4, v5); // abfecdgh
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abgfcehd
																		}
																		else {
																			swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // abgfcedh
																		}
																	}
																	else {
																		swap(v2, v6); swap(v3, v5); swap(v4, v6); // abgfcdeh
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // abfgcehd
																		}
																		else {
																			swap(v2, v5); swap(v3, v6); swap(v4, v5); // abfgcedh
																		}
																	}
																	else {
																		swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // abfgcdeh
																	}
																}
															}
															else {
																if (compare(v6, v4)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abfcgehd
																		}
																		else {
																			swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // abfcgedh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v5); swap(v3, v5); swap(v4, v6); // abfcgdeh
																		}
																		else {
																			swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // abfcdgeh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // abfceghd
																		}
																		else {
																			swap(v2, v5); swap(v3, v5); swap(v5, v6); // abfcegdh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v2, v5); swap(v3, v5); // abfcedgh
																		}
																		else {
																			swap(v2, v5); swap(v3, v5); swap(v4, v5); // abfcdegh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v6, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v6, v7); // acbgfehd
																		}
																		else {
																			swap(v1, v2); swap(v3, v6); swap(v4, v5); // acbgfedh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v3, v6); swap(v4, v5); swap(v5, v6); // acbgfdeh
																		}
																		else {
																			swap(v1, v2); swap(v3, v6); swap(v4, v6); // acbgdfeh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v4)) {
																			swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acbfgehd
																		}
																		else {
																			swap(v1, v2); swap(v3, v5); swap(v5, v6); swap(v6, v7); // acbfeghd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v1, v2); swap(v3, v5); swap(v4, v6); swap(v5, v6); // acbfgedh
																			}
																			else {
																				swap(v1, v2); swap(v3, v5); swap(v5, v6); // acbfegdh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v3, v5); swap(v4, v6); // acbfgdeh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v4)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v5)) {
																			swap(v1, v2); swap(v2, v3); swap(v4, v6); // acdbgfeh
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v4, v5); swap(v5, v6); // acdbfgeh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v3, v5); swap(v4, v5); swap(v5, v6); // acbfdgeh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v2); swap(v4, v6); // acbdgfeh
																			}
																			else {
																				swap(v1, v2); swap(v4, v5); swap(v5, v6); // acbdfgeh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v3, v5); // acbfedgh
																		}
																		else {
																			swap(v1, v2); swap(v3, v5); swap(v4, v5); // acbfdegh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v4, v5); // acdbfegh
																		}
																		else {
																			swap(v1, v2); swap(v4, v5); // acbdfegh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v7, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // abgcfehd
																		}
																		else {
																			swap(v3, v6); swap(v4, v5); swap(v6, v7); // abcgfehd
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abcfgehd
																		}
																		else {
																			swap(v3, v5); swap(v5, v6); swap(v6, v7); // abcfeghd
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v2, v6); swap(v3, v6); swap(v4, v5); // abgcfedh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v3, v6); swap(v4, v5); // abcgfedh
																			}
																			else {
																				swap(v3, v5); swap(v4, v6); swap(v5, v6); // abcfgedh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v3, v5); swap(v5, v6); // abcfegdh
																		}
																		else {
																			swap(v3, v5); // abcfedgh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // abgcfdeh
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v3, v6); swap(v4, v5); swap(v5, v6); // abcgfdeh
																			}
																			else {
																				swap(v3, v5); swap(v4, v6); // abcfgdeh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v3, v5); swap(v4, v5); swap(v5, v6); // abcfdgeh
																		}
																		else {
																			swap(v3, v5); swap(v4, v5); // abcfdegh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v2, v6); swap(v3, v6); swap(v4, v6); // abgcdfeh
																		}
																		else {
																			swap(v3, v6); swap(v4, v6); // abcgdfeh
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			if (compare(v6, v5)) {
																				swap(v4, v6); // abcdgfeh
																			}
																			else {
																				swap(v4, v5); swap(v5, v6); // abcdfgeh
																			}
																		}
																		else {
																			swap(v4, v5); // abcdfegh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
							else {
								if (compare(v7, v6)) {
									if (compare(v7, v4)) {
										if (compare(v7, v0)) {
											if (compare(v6, v2)) {
												if (compare(v6, v4)) {
													if (compare(v5, v3)) {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgefacdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hgefacbd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgefabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hgeafcdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); // hgeafcbd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hgeafbcd
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hgaefcdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hgaefcbd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hgaefbcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hagefcdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hagefcbd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hagefbcd
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v6, v7); // hgeacfdb
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hgaecfdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v6, v7); // hgacefdb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); // hagecfdb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); // hagcefdb
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); // hgeacfbd
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); // hgaecfbd
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); // hgacefbd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hagecfbd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v6, v7); // hagcefbd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); // hgeacbfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hgaecbfd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); // hgacebfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hgacbefd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); // hgeabfcd
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hgeabcfd
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hgaebfcd
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hgaebcfd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v6, v7); // hgabefcd
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hgabecfd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgabcefd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hagecbfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hagcebfd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); // hagebfcd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hagebcfd
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); // hagbefcd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); // hagbecfd
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hagcbefd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hagbcefd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v6, v7); // habgefcd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // habgecfd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // habgcefd
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v0)) {
															if (compare(v3, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hgeacdfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgaecdfb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hgacedfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hgacdefb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); // hgeacdbf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hgaecdbf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); // hgacedbf
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hgacdebf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgacdbef
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hgeacbdf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgaecbdf
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hgacebdf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgacbedf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hgacbdef
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hgeabcdf
																		}
																		else {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hgaebcdf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); // hgabecdf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hgabcedf
																			}
																			else {
																				swap(v0, v7); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hgabcdef
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hagecdfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hagecdbf
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hagecbdf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hagebcdf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); // hagcedfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hagcdefb
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hagcedbf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hagcdebf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); // hagcebdf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hagcbedf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hagbecdf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hagbcedf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); // habgecdf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // habgcedf
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hagcdbef
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hagcbdef
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hagbcdef
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // habgcdef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v3, v1)) {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // hegfacdb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hegafcdb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // heagfcdb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); // haegfcdb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); // hefgacdb
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hefagcdb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // heafgcdb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // haefgcdb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // hegacfdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hegacdfb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hegacdbf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); // heagcfdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // heagcdfb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // heagcdbf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); // haegcfdb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); // haegcdfb
																		}
																	}
																	else {
																		swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // haegcdbf
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hegfacbd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hegafcbd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // heagfcbd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // haegfcbd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // hefgacbd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); // hefagcbd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // heafgcbd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // haefgcbd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); // hegacfbd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hegacbfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hegacbdf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // heagcfbd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // heagcbfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // heagcbdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v6, v7); // haegcfbd
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // haegcbfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); // haegcbdf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v5, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // hegfabcd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); // hegafbcd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // heagfbcd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // haegfbcd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hefgabcd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hefagbcd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // heafgbcd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // haefgbcd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); // hegabfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); // hegabcfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hegabcdf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v6, v7); // heagbfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // heagbcfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); // heagbcdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // haegbfcd
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // haegbcfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // haegbcdf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v5, v2)) {
																		swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // heabgfcd
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // heabgcfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // heabgcdf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); // hefabgcd
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // heafbgcd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // heabfgcd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // haefbgcd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); // haebgfcd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // haebfgcd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // haebgcfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // haebgcdf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // habegfcd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // habefgcd
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // habegcfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // habegcdf
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v6, v3)) {
													if (compare(v6, v1)) {
														if (compare(v3, v1)) {
															if (compare(v4, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // heacgfdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // heacgdfb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // heacgdbf
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hefacgdb
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // heafcgdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // heacfgdb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); // haefcgdb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); // haecgfdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // haecfgdb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // haecgdfb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // haecgdbf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); // hacgefdb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v6); // hacegfdb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // hacefgdb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v5, v6); // hacgedfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // hacegdfb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // hacgdefb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // hacgedbf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hacegdbf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hacgdebf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // hacgdbef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // hefacgbd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // heafcgbd
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); // heacgfbd
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // heacfgbd
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // haefcgbd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // haecgfbd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // haecfgbd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v6, v7); // hacgefbd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // hacegfbd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // hacefgbd
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // heacgbfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // haecgbfd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // hacgebfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // hacegbfd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hacgbefd
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v5, v7); // hacgebdf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // hacgbedf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hacgbdef
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // heacgbdf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // haecgbdf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // hacegbdf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v4, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); // hefacbgd
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // heafcbgd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); // heacfbgd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); // heacbgfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); // heacbfgd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // heacbgdf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); // haefcbgd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // haecfbgd
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // haecbgfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); // haecbfgd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // haecbgdf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // hacebgfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hacebgdf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // hacefbgd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v7); // hacebfgd
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // hacbgefd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v5, v6); // hacbegfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v7); // hacbefgd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hacbgedf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // hacbegdf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // hacbgdef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); // hefabcgd
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); // heafbcgd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // heabfcgd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // heabcgfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); // heabcfgd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // heabcgdf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // haefbcgd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); // haebfcgd
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); // haebcgfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); // haebcfgd
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // haebcgdf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // habecgfd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // habecgdf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // habefcgd
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); // habecfgd
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // habcgefd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); // habcegfd
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); // habcefgd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // habcgedf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // habcegdf
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // habcgdef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v6, v1)) {
															if (compare(v4, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // hefacdgb
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); // heafcdgb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // heacfdgb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // heacdgfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); // heacdfgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // heacdgbf
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); // haefcdgb
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); // haecfdgb
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); // haecdgfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); // haecdfgb
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // haecdgbf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v5); // hacefdgb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v5, v6); // hacedgfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v4); // hacedfgb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v4, v6); swap(v5, v6); // hacdgefb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v5, v6); // hacdegfb
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); // hacdefgb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // hacdgebf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v4, v6); swap(v5, v7); // hacdgbef
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // hacedgbf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v5, v6); swap(v6, v7); // hacdegbf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hefacdbg
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // heafcdbg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // heacfdbg
																		}
																		else {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // heacdfbg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v6, v7); // haefcdbg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // haecfdbg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v6, v7); // haecdfbg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // hacefdbg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v6, v7); // hacedfbg
																			}
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v6, v7); // hacdefbg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // heacdbgf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); // haecdbgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v5, v7); // hacedbgf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // heacdbfg
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // haecdbfg
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // hacedbfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v7); swap(v5, v7); // hacdebgf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v5, v7); swap(v6, v7); // hacdebfg
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hacdbgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v4, v7); swap(v5, v7); // hacdbegf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hacdbefg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v4, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // hefacbdg
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // heafcbdg
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // heacfbdg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v6, v7); // heacbfdg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); // heacbdgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // heacbdfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // haefcbdg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // haecfbdg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // haecbfdg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // haecbdgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // haecbdfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // hacefbdg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // hacebfdg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // hacebdgf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hacebdfg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v6, v7); // hacbefdg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v5, v7); // hacbedgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // hacbedfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // hacbdgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // hacbdegf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // hacbdefg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v6, v7); // hefabcdg
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // heafbcdg
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // heabfcdg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // heabcfdg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // heabcdgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // heabcdfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // haefbcdg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // haebfcdg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v6, v7); // haebcfdg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); // haebcdgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // haebcdfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // habefcdg
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // habecfdg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // habecdgf
																		}
																		else {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // habecdfg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v6, v7); // habcefdg
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); // habcedgf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // habcedfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // habcdgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // habcdegf
																			}
																			else {
																				swap(v0, v7); swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // habcdefg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v1)) {
												if (compare(v5, v1)) {
													if (compare(v6, v5)) {
														if (compare(v6, v4)) {
															if (compare(v6, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ahgefcdb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); // ahgecfdb
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ahgecdfb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v6); // ahgcefdb
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); // ahgcedfb
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ahgcdefb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahgefcbd
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ahgefbcd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ahgecfbd
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v6, v7); // ahgcefbd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v3, v6); // ahcgefdb
																			}
																			else {
																				swap(v1, v7); swap(v3, v6); swap(v5, v6); // ahcgedfb
																			}
																		}
																		else {
																			swap(v1, v7); swap(v3, v6); swap(v6, v7); // ahcgefbd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ahcgdefb
																		}
																		else {
																			swap(v1, v7); swap(v4, v6); swap(v5, v6); // ahcdgefb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v6); // achgefdb
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); // achgedfb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v6, v7); // achgefbd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); // achgdefb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); // achdgefb
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); // acdhgefb
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); // ahegfcdb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v6); // ahegcfdb
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); // ahegcdfb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); // ahecgfdb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ahecgdfb
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); // ahecdgfb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ahegfcbd
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahegfbcd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v6, v7); // ahegcfbd
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ahecgfbd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v7); swap(v3, v4); swap(v4, v6); // ahcegfdb
																		}
																		else {
																			swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ahcegfbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ahcegdfb
																			}
																			else {
																				swap(v1, v7); swap(v3, v4); swap(v5, v6); // ahcedgfb
																			}
																		}
																		else {
																			swap(v1, v7); swap(v5, v6); // ahcdegfb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); // achegfdb
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); // achegdfb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // achegfbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); // achedgfb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v5, v6); // achdegfb
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); // acdhegfb
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v1)) {
															if (compare(v4, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ahefgcdb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); // ahefcgdb
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v5); // ahefcdgb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ahecfgdb
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); // ahecfdgb
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); // ahecdfgb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahefgcbd
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ahefgbcd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ahefcgbd
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahecfgbd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ahcefgdb
																			}
																			else {
																				swap(v1, v7); swap(v3, v4); swap(v4, v5); // ahcefdgb
																			}
																		}
																		else {
																			swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahcefgbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v3, v4); // ahcedfgb
																		}
																		else {
																			swap(v1, v7); // ahcdefgb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); // achefgdb
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); // achefdgb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // achefgbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v4); // achedfgb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v7); // achdefgb
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v7); // acdhefgb
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v6, v7); // ahefcdbg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ahecfdbg
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v6, v7); // ahecdfbg
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ahcefdbg
																			}
																			else {
																				swap(v1, v7); swap(v3, v4); swap(v6, v7); // ahcedfbg
																			}
																		}
																		else {
																			swap(v1, v7); swap(v6, v7); // ahcdefbg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v6, v7); // achefdbg
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v6, v7); // achedfbg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v6, v7); // achdefbg
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v6, v7); // acdhefbg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); // ahefcbgd
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ahefcbdg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahefbgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ahefbcgd
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahefbcdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ahecfbgd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ahcefbgd
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // achefbgd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahecfbdg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahcefbdg
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // achefbdg
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v1)) {
														if (compare(v6, v4)) {
															if (compare(v3, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahgecdbf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ahgcedbf
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahgcdebf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ahcgedbf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahcgdebf
																			}
																			else {
																				swap(v1, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahcdgebf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v6); swap(v6, v7); // achgedbf
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // achgdebf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // achdgebf
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acdhgebf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahgecbfd
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ahgecbdf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); // ahgebfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ahgebcfd
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahgebcdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ahgcebfd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ahcgebfd
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // achgebfd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); // ahgcebdf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v7); swap(v3, v6); swap(v5, v7); // ahcgebdf
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v5, v7); // achgebdf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ahegcdbf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahecgdbf
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ahecdgbf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahcegdbf
																			}
																			else {
																				swap(v1, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ahcedgbf
																			}
																		}
																		else {
																			swap(v1, v7); swap(v5, v6); swap(v6, v7); // ahcdegbf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // achegdbf
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v6); swap(v6, v7); // achedgbf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v5, v6); swap(v6, v7); // achdegbf
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v6); swap(v6, v7); // acdhegbf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ahegcbfd
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); // ahegcbdf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ahegbfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahegbcfd
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ahegbcdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahecgbfd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahcegbfd
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // achegbfd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ahecgbdf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v1, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ahcegbdf
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // achegbdf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ahecbgfd
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahecbgdf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); // ahecdbgf
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ahecbdgf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); // ahebgfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ahebgcfd
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahebgcdf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); // ahebcgfd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ahebcgdf
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); // ahebcdgf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v3, v4); swap(v5, v7); // ahcedbgf
																		}
																		else {
																			swap(v1, v7); swap(v5, v7); // ahcdebgf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); // achedbgf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v5, v7); // achdebgf
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); // acdhebgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ahcebgfd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahcebgdf
																			}
																			else {
																				swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ahcebdgf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // achebgfd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // achebgdf
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // achebdgf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); // ahecbfgd
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ahecbfdg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ahecdbfg
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahecbdfg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ahebfgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); // ahebfcgd
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ahebfcdg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v2, v4); swap(v3, v7); // ahebcfgd
																			}
																			else {
																				swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v6, v7); // ahebcfdg
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ahebcdfg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ahcedbfg
																		}
																		else {
																			swap(v1, v7); swap(v5, v7); swap(v6, v7); // ahcdebfg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v5, v7); swap(v6, v7); // achedbfg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v5, v7); swap(v6, v7); // achdebfg
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v5, v7); swap(v6, v7); // acdhebfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v3, v4); swap(v4, v7); // ahcebfgd
																			}
																			else {
																				swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ahcebfdg
																			}
																		}
																		else {
																			swap(v1, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahcebdfg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); // achebfgd
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // achebfdg
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // achebdfg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v6, v4)) {
														if (compare(v6, v1)) {
															if (compare(v3, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ahgcdbef
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ahcgdbef
																		}
																		else {
																			swap(v1, v7); swap(v4, v6); swap(v5, v7); // ahcdgbef
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v6); swap(v5, v7); // achgdbef
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v4, v6); swap(v5, v7); // achdgbef
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v6); swap(v5, v7); // acdhgbef
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahgcbefd
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ahgcbedf
																		}
																	}
																	else {
																		swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahgcbdef
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahcgbefd
																			}
																			else {
																				swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ahcgbedf
																			}
																		}
																		else {
																			swap(v1, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahcgbdef
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // achgbefd
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // achgbedf
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // achgbdef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ahcbgefd
																			}
																			else {
																				swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahcbgedf
																			}
																		}
																		else {
																			swap(v1, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ahcbgdef
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // achbgefd
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // achbgedf
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // achbgdef
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); // acbhgefd
																		}
																		else {
																			swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acbhgedf
																		}
																	}
																	else {
																		swap(v1, v2); swap(v3, v7); swap(v4, v6); swap(v5, v7); // acbhgdef
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v1, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahcdbgef
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // achdbgef
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acdhbgef
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acdbhgef
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahcbdgef
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // achbdgef
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acbhdgef
																		}
																		else {
																			swap(v1, v2); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acbdhgef
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v4, v3)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v7); swap(v3, v7); swap(v5, v6); // ahcbegfd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ahcbegdf
																			}
																			else {
																				swap(v1, v7); swap(v3, v7); swap(v5, v7); // ahcbedgf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); // achbegfd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // achbegdf
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); // achbedgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		swap(v1, v2); swap(v3, v7); swap(v5, v6); // acbhegfd
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v3, v7); swap(v5, v6); swap(v6, v7); // acbhegdf
																		}
																		else {
																			swap(v1, v2); swap(v3, v7); swap(v5, v7); // acbhedgf
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v1, v7); swap(v4, v7); swap(v5, v7); // ahcdbegf
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); // achdbegf
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); // acdhbegf
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); // acdbhegf
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ahcbdegf
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // achbdegf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); // acbhdegf
																		}
																		else {
																			swap(v1, v2); swap(v4, v7); swap(v5, v7); // acbdhegf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v7, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v7); swap(v3, v7); // ahcbefgd
																			}
																			else {
																				swap(v1, v7); swap(v3, v7); swap(v6, v7); // ahcbefdg
																			}
																		}
																		else {
																			swap(v1, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ahcbedfg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v7); swap(v3, v7); // achbefgd
																			}
																			else {
																				swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v6, v7); // achbefdg
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // achbedfg
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v3, v7); // acbhefgd
																		}
																		else {
																			swap(v1, v2); swap(v3, v7); swap(v6, v7); // acbhefdg
																		}
																	}
																	else {
																		swap(v1, v2); swap(v3, v7); swap(v5, v7); swap(v6, v7); // acbhedfg
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v1, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahcdbefg
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // achdbefg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acdhbefg
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acdbhefg
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v1, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahcbdefg
																		}
																		else {
																			swap(v1, v2); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // achbdefg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acbhdefg
																		}
																		else {
																			swap(v1, v2); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acbdhefg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v7, v1)) {
															if (compare(v6, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v1)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); // ahgbefcd
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v6, v7); // ahbgefcd
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ahbegfcd
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ahbefgcd
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); // ahgbecfd
																		}
																		else {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ahgbcefd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v4)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ahbgecfd
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ahbegcfd
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahbgcefd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ahbefcgd
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ahbefcdg
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ahbecgfd
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); // ahbecfgd
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ahbecfdg
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ahbcgefd
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); // ahbcegfd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); // ahbcefgd
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v6, v7); // ahbcefdg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			swap(v2, v7); swap(v3, v6); swap(v6, v7); // abhgefcd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // abhegfcd
																			}
																			else {
																				swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abhefgcd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // abhefcgd
																		}
																		else {
																			swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abhefcdg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v4)) {
																			swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // abhgecfd
																		}
																		else {
																			swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abhegcfd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // abhecgfd
																			}
																			else {
																				swap(v2, v7); swap(v3, v4); swap(v4, v7); // abhecfgd
																			}
																		}
																		else {
																			swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // abhecfdg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v2)) {
																			swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abhgcefd
																		}
																		else {
																			swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // abhcgefd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v2, v7); swap(v3, v7); swap(v5, v6); // abhcegfd
																			}
																			else {
																				swap(v2, v7); swap(v3, v7); // abhcefgd
																			}
																		}
																		else {
																			swap(v2, v7); swap(v3, v7); swap(v6, v7); // abhcefdg
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v4)) {
																			swap(v3, v7); swap(v4, v6); swap(v5, v6); // abchgefd
																		}
																		else {
																			swap(v3, v7); swap(v5, v6); // abchegfd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v3, v7); // abchefgd
																		}
																		else {
																			swap(v3, v7); swap(v6, v7); // abchefdg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v3)) {
															if (compare(v7, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v4)) {
																		if (compare(v6, v1)) {
																			swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ahgbecdf
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v5, v7); // ahbgecdf
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ahbegcdf
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahbecgdf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			if (compare(v4, v3)) {
																				swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahgbcedf
																			}
																			else {
																				swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ahgbcdef
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ahbgcedf
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahbgcdef
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ahbcgedf
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ahbcegdf
																			}
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ahbcgdef
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v4)) {
																			swap(v2, v7); swap(v3, v6); swap(v5, v7); // abhgecdf
																		}
																		else {
																			swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // abhegcdf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // abhgcedf
																		}
																		else {
																			swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abhgcdef
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v4)) {
																			if (compare(v4, v3)) {
																				swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abhcgedf
																			}
																			else {
																				swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // abhcgdef
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abhecgdf
																			}
																			else {
																				swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // abhcegdf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v4)) {
																				swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abchgedf
																			}
																			else {
																				swap(v3, v7); swap(v5, v6); swap(v6, v7); // abchegdf
																			}
																		}
																		else {
																			swap(v3, v7); swap(v4, v6); swap(v5, v7); // abchgdef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ahbecdgf
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); // ahbcedgf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // abhecdgf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v2, v7); swap(v3, v7); swap(v5, v7); // abhcedgf
																			}
																			else {
																				swap(v3, v7); swap(v5, v7); // abchedgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahbecdfg
																		}
																		else {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ahbcedfg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abhecdfg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // abhcedfg
																			}
																			else {
																				swap(v3, v7); swap(v5, v7); swap(v6, v7); // abchedfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v7, v1)) {
																		if (compare(v6, v4)) {
																			swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ahbcdgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // ahbcdegf
																			}
																			else {
																				swap(v1, v7); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ahbcdefg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abhcdgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // abhcdegf
																			}
																			else {
																				swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abhcdefg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v4)) {
																			swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abchdgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v3, v7); swap(v4, v7); swap(v5, v7); // abchdegf
																			}
																			else {
																				swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abchdefg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v4)) {
																			swap(v4, v7); swap(v5, v6); swap(v6, v7); // abcdhgef
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v4, v7); swap(v5, v7); // abcdhegf
																			}
																			else {
																				swap(v4, v7); swap(v5, v7); swap(v6, v7); // abcdhefg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v0)) {
											if (compare(v5, v2)) {
												if (compare(v7, v5)) {
													if (compare(v2, v1)) {
														if (compare(v3, v1)) {
															if (compare(v6, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); // ehgfacdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ehgafcdb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); // ehagfcdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // eahgfcdb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); // ehfgacdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); // ehfagcdb
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ehafgcdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eahfgcdb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ehfacgdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); // ehfacdgb
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ehfacdbg
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); // ehafcgdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); // ehafcdgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v6, v7); // ehafcdbg
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // eahfcgdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); // eahfcdgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // eahfcdbg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ehgfacbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehgafcbd
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // ehagfcbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eahgfcbd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v6, v7); // ehfgacbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ehfagcbd
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehafgcbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // eahfgcbd
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v6, v1)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehfacgbd
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ehfacbgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehfacbdg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // ehafcgbd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); // ehafcbgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // ehafcbdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eahfcgbd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // eahfcbgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eahfcbdg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v6, v1)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v5); swap(v5, v7); // ehgfabcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ehgafbcd
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehagfbcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // eahgfbcd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ehfgabcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehfagbcd
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ehafgbcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eahfgbcd
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v6, v2)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ehfabgcd
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); // ehfabcgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ehfabcdg
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehafbgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // ehafbcgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehafbcdg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // eahfbgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); // eahfbcgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // eahfbcdg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v6, v5)) {
																	if (compare(v6, v0)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); // ehgabfcd
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // ehagbfcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); // ehabgfcd
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // ehabfgcd
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); // ehabfcgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // ehabfcdg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v6, v7); // eahgbfcd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // eahbgfcd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eahbfgcd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // eahbfcgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eahbfcdg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); // eabhgfcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // eabhfgcd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); // eabhfcgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // eabhfcdg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v3)) {
														if (compare(v6, v1)) {
															if (compare(v6, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); // efhgacdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // efhgacbd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v6); swap(v6, v7); // efhgabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // efhagcdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efhagcbd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // efhagbcd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efahgcdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // efahgcbd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v6); // efahgbcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // eafhgcdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); // eafhgcbd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // eafhgbcd
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efhacgdb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // efahcgdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efachgdb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eafhcgdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // eafchgdb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // efhacgbd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v6); // efahcgbd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // efachgbd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // eafhcgbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eafchgbd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); // efhacbgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); // efahcbgd
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); // efachbgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // efacbhgd
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efhabgcd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // efhabcgd
																			}
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // efahbgcd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); // efahbcgd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efabhgcd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // efabhcgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); // efabchgd
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); // eafhcbgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // eafchbgd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); // eafhbgcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); // eafhbcgd
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // eafbhgcd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); // eafbhcgd
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); // eafcbhgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // eafbchgd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eabfhgcd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // eabfhcgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); // eabfchgd
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v3, v1)) {
																if (compare(v6, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // efhacdgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); // efahcdgb
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // efachdgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); // efacdhgb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efhacdbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // efahcdbg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efachdbg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // efacdhbg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v6, v7); // efacdbhg
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // efhacbdg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v6, v7); // efahcbdg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // efachbdg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efacbhdg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // efacbdhg
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efhabcdg
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // efahbcdg
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efabhcdg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // efabchdg
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v6, v7); // efabcdhg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v7, v2)) {
																	if (compare(v3, v1)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // eafhcdgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eafhcdbg
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // eafhcbdg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v6, v7); // eafhbcdg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); // eafchdgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // eafcdhgb
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // eafchdbg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eafcdhbg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eafchbdg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // eafcbhdg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // eafbhcdg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eafbchdg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eabfhcdg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // eabfchdg
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // eafcdbhg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v6, v7); // eafcbdhg
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // eafbcdhg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v6, v7); // eabfcdhg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v5, v1)) {
														if (compare(v3, v1)) {
															if (compare(v7, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); // ehgacfdb
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); // ehagcfdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); // ehacgfdb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ehacfgdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); // ehacfdgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // ehacfdbg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); // eahgcfdb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // eahcgfdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eahcfgdb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // eahcfdgb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eahcfdbg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // eachgfdb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eachfgdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // eacfhgdb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // eachfdgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v7); // eacfhdgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // eacfdhgb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eachfdbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // eacfhdbg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eacfdhbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // eacfdbhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ehgacfbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v6, v7); // ehagcfbd
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ehacgfbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ehacfgbd
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // eahgcfbd
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); // eahcgfbd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // eahcfgbd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v6); // eachgfbd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // eachfgbd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eacfhgbd
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ehacfbgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); // eahcfbgd
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v5); // eachfbgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // eacfhbgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v5, v7); // eacfbhgd
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // ehacfbdg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // eahcfbdg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // eachfbdg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eacfhbdg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // eacfbhdg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v6, v7); // eacfbdhg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v7, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehgacbfd
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ehagcbfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ehacgbfd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ehacbgfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); // ehacbfgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // ehacbfdg
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eahgcbfd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // eahcgbfd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); // eahcbgfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); // eahcbfgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v6, v7); // eahcbfdg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // eachgbfd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v5, v6); // eachbgfd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v7); // eachbfgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v6, v7); // eachbfdg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // eacbhgfd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v7); // eacbhfgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // eacbfhgd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // eacbhfdg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eacbfhdg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // eacbfdhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ehgabcfd
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehagbcfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ehabgcfd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); // ehabcgfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); // ehabcfgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v6, v7); // ehabcfdg
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // eahgbcfd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eahbgcfd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // eahbcgfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); // eahbcfgd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // eahbcfdg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // eabhgcfd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); // eabhcgfd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); // eabhcfgd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v6, v7); // eabhcfdg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // eabchgfd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); // eabchfgd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // eabcfhgd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // eabchfdg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eabcfhdg
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // eabcfdhg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v5, v1)) {
															if (compare(v7, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ehgacdfb
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); // ehagcdfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ehacgdfb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); // ehacdgfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); // ehacdfgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v6, v7); // ehacdfbg
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // eahgcdfb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eahcgdfb
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // eahcdgfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); // eahcdfgb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // eahcdfbg
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eachgdfb
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // eachdgfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v7); // eachdfgb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v4, v7); swap(v5, v6); // eacdhgfb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v4, v7); // eacdhfgb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v4, v5); swap(v5, v7); // eacdfhgb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // eachdfbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v4, v7); swap(v6, v7); // eacdhfbg
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eacdfhbg
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v4, v5); swap(v6, v7); // eacdfbhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehgacdbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ehagcdbf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehacgdbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // ehacdgbf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eahgcdbf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // eahcgdbf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eahcdgbf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // eachgdbf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eachdgbf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eacdhgbf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); // ehacdbgf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // eahcdbgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // eachdbgf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // ehacdbfg
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eahcdbfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eachdbfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v4, v7); swap(v5, v7); // eacdhbgf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eacdhbfg
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v5, v7); // eacdbhgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v5, v7); swap(v6, v7); // eacdbhfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v6, v7); // eacdbfhg
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v7, v2)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ehgacbdf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); // ehagcbdf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ehacgbdf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehacbgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // ehacbdgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ehacbdfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // eahgcbdf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eahcgbdf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // eahcbgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); // eahcbdgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // eahcbdfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eachgbdf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // eachbgdf
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v5, v7); // eachbdgf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // eachbdfg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eacbhgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // eacbhdgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eacbhdfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v5, v7); // eacbdhgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // eacbdhfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v6, v7); // eacbdfhg
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v7, v0)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ehgabcdf
																		}
																		else {
																			if (compare(v6, v1)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ehagbcdf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ehabgcdf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ehabcgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); // ehabcdgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v7); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // ehabcdfg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); // eahgbcdf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // eahbgcdf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eahbcgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // eahbcdgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eahbcdfg
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eabhgcdf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // eabhcgdf
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); // eabhcdgf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // eabhcdfg
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eabchgdf
																		}
																		else {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // eabchdgf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eabchdfg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); // eabcdhgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // eabcdhfg
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v6, v7); // eabcdfhg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v7, v1)) {
												if (compare(v6, v1)) {
													if (compare(v6, v5)) {
														if (compare(v5, v1)) {
															if (compare(v7, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // aehgfcdb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); // aehgcfdb
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // aehgcdfb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v6, v7); // aehcgfdb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aehcgdfb
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); // aehcdgfb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aehgfcbd
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // aehgfbcd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // aehgcfbd
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); // aehcgfbd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // aechgfdb
																		}
																		else {
																			swap(v1, v4); swap(v3, v7); swap(v4, v6); // aechgfbd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aechgdfb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // aechdgfb
																			}
																			else {
																				swap(v1, v4); swap(v4, v7); swap(v5, v6); // aecdhgfb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v6, v7); // acehgfdb
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acehgdfb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); // acehgfbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); // acehdgfb
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); // acedhgfb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); // acdehgfb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aehgcdbf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // aehcgdbf
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aehcdgbf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // aechgdbf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aechdgbf
																			}
																			else {
																				swap(v1, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aecdhgbf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v7); // acehgdbf
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acehdgbf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acedhgbf
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acdehgbf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aehgcbfd
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // aehgcbdf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v6, v7); // aehgbfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // aehgbcfd
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); // aehgbcdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // aehcgbfd
																			}
																			else {
																				swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // aechgbfd
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // acehgbfd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aehcgbdf
																			}
																			else {
																				swap(v1, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aechgbdf
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acehgbdf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v7, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aehfgcdb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); // aehfcgdb
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); // aehfcdgb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aehcfgdb
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); // aehcfdgb
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); // aehcdfgb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // aehfgcbd
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aehfgbcd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aehfcgbd
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v6); // aehcfgbd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aechfgdb
																			}
																			else {
																				swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // aechfdgb
																			}
																		}
																		else {
																			swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // aechfgbd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v3, v7); swap(v4, v7); // aechdfgb
																		}
																		else {
																			swap(v1, v4); swap(v4, v7); // aecdhfgb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acehfgdb
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); // acehfdgb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // acehfgbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); // acehdfgb
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v4, v7); // acedhfgb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); // acdehfgb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // aefhgcdb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aefhcgdb
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); // aefhcdgb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); // aefchgdb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); // aefchdgb
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); // aefcdhgb
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); // aefhgcbd
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // aefhgbcd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v6); // aefhcgbd
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aefchgbd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // aecfhgdb
																		}
																		else {
																			swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aecfhgbd
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v5); swap(v4, v7); // aecfhdgb
																			}
																			else {
																				swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // aecfdhgb
																			}
																		}
																		else {
																			swap(v1, v4); swap(v4, v5); swap(v5, v7); // aecdfhgb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); // acefhgdb
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); // acefhdgb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acefhgbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); // acefdhgb
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); // acedfhgb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); // acdefhgb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v7, v5)) {
															if (compare(v3, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v6, v7); // aehfcdbg
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aehcfdbg
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v6, v7); // aehcdfbg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aechfdbg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // aechdfbg
																			}
																			else {
																				swap(v1, v4); swap(v4, v7); swap(v6, v7); // aecdhfbg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acehfdbg
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v6, v7); // acehdfbg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v6, v7); // acedhfbg
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v6, v7); // acdehfbg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); // aehfcbgd
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aehfcbdg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // aehfbgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); // aehfbcgd
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // aehfbcdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); // aehcfbgd
																			}
																			else {
																				swap(v1, v4); swap(v3, v7); swap(v4, v5); // aechfbgd
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); // acehfbgd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v5); swap(v6, v7); // aehcfbdg
																			}
																			else {
																				swap(v1, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // aechfbdg
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // acehfbdg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aefhcdbg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v6, v7); // aefchdbg
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aefcdhbg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // aecfhdbg
																			}
																			else {
																				swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aecfdhbg
																			}
																		}
																		else {
																			swap(v1, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aecdfhbg
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v6, v7); // acefhdbg
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acefdhbg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acedfhbg
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acdefhbg
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); // aefhcbgd
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v5); swap(v6, v7); // aefhcbdg
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v5, v6); // aefhbgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v7); // aefhbcgd
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v6, v7); // aefhbcdg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); // aefchbgd
																			}
																			else {
																				swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // aecfhbgd
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // acefhbgd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aefchbdg
																			}
																			else {
																				swap(v1, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aecfhbdg
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acefhbdg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); // aehcbgfd
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // aehcbgdf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // aehcdbgf
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); // aehcbdgf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // aehbgfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aehbgcfd
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // aehbgcdf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // aehbcgfd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aehbcgdf
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // aehbcdgf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // aechdbgf
																		}
																		else {
																			swap(v1, v4); swap(v4, v7); swap(v5, v7); // aecdhbgf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); // acehdbgf
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); // acedhbgf
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); // acdehbgf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v3, v7); swap(v5, v6); // aechbgfd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // aechbgdf
																			}
																			else {
																				swap(v1, v4); swap(v3, v7); swap(v5, v7); // aechbdgf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); // acehbgfd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // acehbgdf
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); // acehbdgf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v7); // aehcbfgd
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v6, v7); // aehcbfdg
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aehcdbfg
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // aehcbdfg
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aehbfgcd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // aehbfcgd
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aehbfcdg
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); // aehbcfgd
																			}
																			else {
																				swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // aehbcfdg
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aehbcdfg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aechdbfg
																		}
																		else {
																			swap(v1, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aecdhbfg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acehdbfg
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acedhbfg
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acdehbfg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v3, v7); // aechbfgd
																			}
																			else {
																				swap(v1, v4); swap(v3, v7); swap(v6, v7); // aechbfdg
																			}
																		}
																		else {
																			swap(v1, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // aechbdfg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v7); // acehbfgd
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v6, v7); // acehbfdg
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // acehbdfg
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v7, v5)) {
														if (compare(v6, v5)) {
															if (compare(v7, v3)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // aecbhgfd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aecbhgdf
																			}
																			else {
																				swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // aecbhdgf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // acebhgfd
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acebhgdf
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // acebhdgf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); // acbehgfd
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acbehgdf
																		}
																		else {
																			swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); // acbehdgf
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v4); swap(v5, v7); // aecdbhgf
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v5, v7); // acedbhgf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); // acdebhgf
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v5, v7); // acdbehgf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v1, v4); swap(v3, v4); swap(v5, v7); // aecbdhgf
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); // acebdhgf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v3, v4); swap(v5, v7); // acbedhgf
																		}
																		else {
																			swap(v1, v2); swap(v5, v7); // acbdehgf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v3, v4); swap(v4, v7); // aecbhfgd
																			}
																			else {
																				swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // aecbhfdg
																			}
																		}
																		else {
																			swap(v1, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aecbhdfg
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); // acebhfgd
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // acebhfdg
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acebhdfg
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v3, v4); swap(v4, v7); // acbehfgd
																		}
																		else {
																			swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v6, v7); // acbehfdg
																		}
																	}
																	else {
																		swap(v1, v2); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acbehdfg
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v4); swap(v5, v7); swap(v6, v7); // aecdbhfg
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v5, v7); swap(v6, v7); // acedbhfg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v7); swap(v6, v7); // acdebhfg
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v5, v7); swap(v6, v7); // acdbehfg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v1, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // aecbdhfg
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // acebdhfg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v3, v4); swap(v5, v7); swap(v6, v7); // acbedhfg
																		}
																		else {
																			swap(v1, v2); swap(v5, v7); swap(v6, v7); // acbdehfg
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v3, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v6, v7); // aefcdbhg
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // aecfdbhg
																		}
																		else {
																			swap(v1, v4); swap(v4, v5); swap(v6, v7); // aecdfbhg
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v6, v7); // acefdbhg
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v6, v7); // acedfbhg
																		}
																	}
																	else {
																		swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v6, v7); // acdefbhg
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); // aefcbhgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v7); swap(v6, v7); // aefcbhdg
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v6, v7); // aefcbdhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v3, v5); swap(v5, v7); // aecfbhgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // aecfbhdg
																			}
																			else {
																				swap(v1, v4); swap(v3, v5); swap(v6, v7); // aecfbdhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); // acefbhgd
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // acefbhdg
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v6, v7); // acefbdhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // aecbfhgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aecbfhdg
																			}
																			else {
																				swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // aecbfdhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // acebfhgd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acebfhdg
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // acebfdhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); // acbefhgd
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acbefhdg
																		}
																		else {
																			swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v6, v7); // acbefdhg
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v4); swap(v6, v7); // aecdbfhg
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v6, v7); // acedbfhg
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v6, v7); // acdebfhg
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v6, v7); // acdbefhg
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v1, v4); swap(v3, v4); swap(v6, v7); // aecbdfhg
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v6, v7); // acebdfhg
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v3, v4); swap(v6, v7); // acbedfhg
																		}
																		else {
																			swap(v1, v2); swap(v6, v7); // acbdefhg
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v6, v3)) {
														if (compare(v4, v1)) {
															if (compare(v5, v2)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v5)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); // aebhgfcd
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // aebhfgcd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // aefbhgcd
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aebfhgcd
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); // aefbhcgd
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // aefbchgd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); // aebhfcgd
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // aebfhcgd
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); // aebfchgd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // aebhgcfd
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aebhgcdf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); // aebhcgfd
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v7); // aebhcfgd
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // aebhcgdf
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // aebchgfd
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aebchgdf
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); // aebchfgd
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // aebcfhgd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v2, v4); swap(v3, v7); swap(v4, v6); // abehgfcd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // abehfgcd
																			}
																			else {
																				swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abefhgcd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v2, v4); swap(v3, v7); swap(v4, v5); // abehfcgd
																		}
																		else {
																			swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // abefhcgd
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // abehgcfd
																		}
																		else {
																			swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abehgcdf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v5)) {
																				swap(v2, v4); swap(v3, v7); swap(v5, v6); // abehcgfd
																			}
																			else {
																				swap(v2, v4); swap(v3, v7); // abehcfgd
																			}
																		}
																		else {
																			swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // abehcgdf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // abechgfd
																		}
																		else {
																			swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abechgdf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v2, v4); swap(v3, v5); swap(v5, v7); // abefchgd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v4); swap(v3, v4); swap(v4, v7); // abechfgd
																			}
																			else {
																				swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // abecfhgd
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			swap(v3, v4); swap(v4, v7); swap(v5, v6); // abcehgfd
																		}
																		else {
																			swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abcehgdf
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v3, v4); swap(v4, v7); // abcehfgd
																		}
																		else {
																			swap(v3, v4); swap(v4, v5); swap(v5, v7); // abcefhgd
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // aefbhcdg
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aefbchdg
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // aefbcdhg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			if (compare(v7, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // aebhfcdg
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aebfhcdg
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // aebfchdg
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v6, v7); // aebfcdhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v6, v7); // aebhcfdg
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // aebchfdg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aebcfhdg
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // aebcfdhg
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // abehfcdg
																		}
																		else {
																			swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abefhcdg
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // abefchdg
																		}
																		else {
																			swap(v2, v4); swap(v3, v5); swap(v6, v7); // abefcdhg
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v7, v5)) {
																			if (compare(v7, v2)) {
																				swap(v2, v4); swap(v3, v7); swap(v6, v7); // abehcfdg
																			}
																			else {
																				swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // abechfdg
																			}
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abecfhdg
																			}
																			else {
																				swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // abecfdhg
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v3, v4); swap(v4, v7); swap(v6, v7); // abcehfdg
																			}
																			else {
																				swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abcefhdg
																			}
																		}
																		else {
																			swap(v3, v4); swap(v4, v5); swap(v6, v7); // abcefdhg
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v6, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); // aebhcdgf
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // aebchdgf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v2, v4); swap(v3, v7); swap(v5, v7); // abehcdgf
																			}
																			else {
																				swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // abechdgf
																			}
																		}
																		else {
																			swap(v3, v4); swap(v4, v7); swap(v5, v7); // abcehdgf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // aebhcdfg
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aebchdfg
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v2)) {
																				swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // abehcdfg
																			}
																			else {
																				swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abechdfg
																			}
																		}
																		else {
																			swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abcehdfg
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); // aebcdhgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // aebcdhfg
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v6, v7); // aebcdfhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v2, v4); swap(v3, v4); swap(v5, v7); // abecdhgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // abecdhfg
																			}
																			else {
																				swap(v2, v4); swap(v3, v4); swap(v6, v7); // abecdfhg
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			swap(v3, v4); swap(v5, v7); // abcedhgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v3, v4); swap(v5, v7); swap(v6, v7); // abcedhfg
																			}
																			else {
																				swap(v3, v4); swap(v6, v7); // abcedfhg
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v5, v7); // abcdehgf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v5, v7); swap(v6, v7); // abcdehfg
																			}
																			else {
																				swap(v6, v7); // abcdefhg
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
								else {
									if (compare(v6, v4)) {
										if (compare(v6, v0)) {
											if (compare(v7, v2)) {
												if (compare(v7, v4)) {
													if (compare(v5, v3)) {
														if (compare(v5, v1)) {
															if (compare(v5, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // ghefacdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghefacbd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // ghefabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); // gheafcdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gheafcbd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gheafbcd
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // ghaefcdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // ghaefcbd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // ghaefbcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gahefcdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gahefcbd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gahefbcd
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); // gheacfdb
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); // ghaecfdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); // ghacefdb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gahecfdb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v6, v7); // gahcefdb
																		}
																	}
																}
																else {
																	if (compare(v7, v0)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v6, v7); // gheacfbd
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // ghaecfbd
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v6, v7); // ghacefbd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); // gahecfbd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); // gahcefbd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gheacbfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // ghaecbfd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // ghacebfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // ghacbefd
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gheabfcd
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gheabcfd
																			}
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); // ghaebfcd
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // ghaebcfd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); // ghabefcd
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); // ghabecfd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // ghabcefd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gahecbfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); // gahcebfd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gahebfcd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gahebcfd
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v6, v7); // gahbefcd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gahbecfd
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gahcbefd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gahbcefd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); // gabhefcd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); // gabhecfd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gabhcefd
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v0)) {
															if (compare(v3, v1)) {
																if (compare(v5, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); // gheacdfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // ghaecdfb
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); // ghacedfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // ghacdefb
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gheacdbf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghaecdbf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // ghacedbf
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghacdebf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // ghacdbef
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v5, v7); // gheacbdf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // ghaecbdf
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v5, v7); // ghacebdf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // ghacbedf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghacbdef
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gheabcdf
																		}
																		else {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // ghaebcdf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // ghabecdf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // ghabcedf
																			}
																			else {
																				swap(v0, v6); swap(v1, v7); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // ghabcdef
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v3, v1)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gahecdfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gahecdbf
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gahecbdf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gahebcdf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gahcedfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gahcdefb
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); // gahcedbf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gahcdebf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gahcebdf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gahcbedf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); // gahbecdf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gahbcedf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gabhecdf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gabhcedf
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gahcdbef
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gahcbdef
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gahbcdef
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gabhcdef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v2, v1)) {
														if (compare(v3, v1)) {
															if (compare(v5, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gehfacdb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gehafcdb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // geahfcdb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // gaehfcdb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v6, v7); // gefhacdb
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); // gefahcdb
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // geafhcdb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gaefhcdb
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); // gehacfdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); // gehacdfb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gehacdbf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // geahcfdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // geahcdfb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // geahcdbf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v6, v7); // gaehcfdb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gaehcdfb
																		}
																	}
																	else {
																		swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); // gaehcdbf
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gehfacbd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gehafcbd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // geahfcbd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); // gaehfcbd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); // gefhacbd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gefahcbd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // geafhcbd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // gaefhcbd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gehacfbd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gehacbfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gehacbdf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); // geahcfbd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // geahcbfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // geahcbdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); // gaehcfbd
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); // gaehcbfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gaehcbdf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v5, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gehfabcd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gehafbcd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); // geahfbcd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // gaehfbcd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gefhabcd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gefahbcd
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // geafhbcd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gaefhbcd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v6, v7); // gehabfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gehabcfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v7); swap(v3, v6); swap(v5, v7); // gehabcdf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); // geahbfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); // geahbcfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // geahbcdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); // gaehbfcd
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gaehbcfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gaehbcdf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v5, v2)) {
																		swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); // geabhfcd
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // geabhcfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // geabhcdf
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gefabhcd
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); // geafbhcd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // geabfhcd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // gaefbhcd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // gaebhfcd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gaebfhcd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gaebhcfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gaebhcdf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); // gabehfcd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gabefhcd
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gabehcfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gabehcdf
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v7, v3)) {
													if (compare(v7, v1)) {
														if (compare(v3, v1)) {
															if (compare(v4, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v1)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); // geachfdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // geachdfb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // geachdbf
																		}
																	}
																	else {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // gefachdb
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // geafchdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // geacfhdb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // gaefchdb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gaechfdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gaecfhdb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gaechdfb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gaechdbf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v6, v7); // gachefdb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // gacehfdb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gacefhdb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // gachedfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gacehdfb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gachdefb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v5, v7); // gachedbf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // gacehdbf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // gachdebf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gachdbef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // gefachbd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // geafchbd
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // geachfbd
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // geacfhbd
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); // gaefchbd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); // gaechfbd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gaecfhbd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); // gachefbd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v7); // gacehfbd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // gacefhbd
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v5, v3)) {
																	if (compare(v4, v2)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // geachbfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gaechbfd
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v5, v6); // gachebfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // gacehbfd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // gachbefd
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // gachebdf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gachbedf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // gachbdef
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // geachbdf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gaechbdf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gacehbdf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v4, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gefacbhd
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // geafcbhd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // geacfbhd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // geacbhfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v6, v7); // geacbfhd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); // geacbhdf
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // gaefcbhd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gaecfbhd
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gaecbhfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gaecbfhd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gaecbhdf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gacebhfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gacebhdf
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gacefbhd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gacebfhd
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gacbhefd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gacbehfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v6, v7); // gacbefhd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gacbhedf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v5, v7); // gacbehdf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gacbhdef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v6, v7); // gefabchd
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // geafbchd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // geabfchd
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // geabchfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // geabcfhd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // geabchdf
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gaefbchd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // gaebfchd
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gaebchfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v6, v7); // gaebcfhd
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); // gaebchdf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // gabechfd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // gabechdf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gabefchd
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // gabecfhd
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gabchefd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // gabcehfd
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v6, v7); // gabcefhd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // gabchedf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); // gabcehdf
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gabchdef
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v7, v1)) {
															if (compare(v4, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // gefacdhb
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // geafcdhb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // geacfdhb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // geacdhfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // geacdfhb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // geacdhbf
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v6, v7); // gaefcdhb
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gaecfdhb
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gaecdhfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v6, v7); // gaecdfhb
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); // gaecdhbf
																		}
																	}
																}
															}
															else {
																if (compare(v5, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // gacefdhb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // gacedhfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v6, v7); // gacedfhb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // gacdhefb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v5, v7); swap(v6, v7); // gacdehfb
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v6, v7); // gacdefhb
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v4)) {
																		if (compare(v4, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v4, v7); swap(v5, v7); // gacdhebf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // gacdhbef
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v5, v7); // gacedhbf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v5, v7); // gacdehbf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // gefacdbh
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // geafcdbh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // geacfdbh
																		}
																		else {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // geacdfbh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); // gaefcdbh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // gaecfdbh
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); // gaecdfbh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v5); // gacefdbh
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v4); // gacedfbh
																			}
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); // gacdefbh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // geacdbhf
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gaecdbhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // gacedbhf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // geacdbfh
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // gaecdbfh
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v5, v6); // gacedbfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v6); swap(v5, v6); swap(v6, v7); // gacdebhf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v5, v6); // gacdebfh
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v4, v6); swap(v5, v7); // gacdbhef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gacdbehf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v4, v6); swap(v5, v6); // gacdbefh
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v4, v2)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v1)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // gefacbdh
																		}
																		else {
																			if (compare(v5, v2)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // geafcbdh
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // geacfbdh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); // geacbfdh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // geacbdhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // geacbdfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // gaefcbdh
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gaecfbdh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // gaecbfdh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gaecbdhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gaecbdfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gacefbdh
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v6); // gacebfdh
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gacebdhf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gacebdfh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v6); // gacbefdh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gacbedhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v5, v6); // gacbedfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gacbdhef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gacbdehf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gacbdefh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v1)) {
																if (compare(v4, v0)) {
																	if (compare(v5, v2)) {
																		if (compare(v5, v0)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v5); swap(v3, v6); // gefabcdh
																		}
																		else {
																			if (compare(v5, v1)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // geafbcdh
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // geabfcdh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // geabcfdh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // geabcdhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // geabcdfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v5, v1)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // gaefbcdh
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // gaebfcdh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); // gaebcfdh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gaebcdhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // gaebcdfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v5, v2)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // gabefcdh
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // gabecfdh
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gabecdhf
																		}
																		else {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // gabecdfh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); // gabcefdh
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // gabcedhf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // gabcedfh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // gabcdhef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // gabcdehf
																			}
																			else {
																				swap(v0, v6); swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // gabcdefh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v4, v1)) {
												if (compare(v5, v1)) {
													if (compare(v7, v5)) {
														if (compare(v7, v4)) {
															if (compare(v7, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aghefcdb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // aghecfdb
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aghecdfb
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v6, v7); // aghcefdb
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); swap(v6, v7); // aghcedfb
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aghcdefb
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // aghefcbd
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aghefbcd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); // aghecfbd
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v7); // aghcefbd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v3, v7); swap(v6, v7); // agchefdb
																			}
																			else {
																				swap(v1, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // agchedfb
																			}
																		}
																		else {
																			swap(v1, v6); swap(v3, v7); // agchefbd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agchdefb
																		}
																		else {
																			swap(v1, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agcdhefb
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v6, v7); // acghefdb
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); swap(v6, v7); // acghedfb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v7); // acghefbd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acghdefb
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acgdhefb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acdghefb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // agehfcdb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v6, v7); // agehcfdb
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // agehcdfb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // agechfdb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agechdfb
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // agecdhfb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); // agehfcbd
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // agehfbcd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v7); // agehcfbd
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); // agechfbd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // agcehfdb
																		}
																		else {
																			swap(v1, v6); swap(v3, v4); swap(v4, v7); // agcehfbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agcehdfb
																			}
																			else {
																				swap(v1, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // agcedhfb
																			}
																		}
																		else {
																			swap(v1, v6); swap(v5, v7); swap(v6, v7); // agcdehfb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v6, v7); // acgehfdb
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acgehdfb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); // acgehfbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); swap(v6, v7); // acgedhfb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v5, v7); swap(v6, v7); // acgdehfb
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); swap(v6, v7); // acdgehfb
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v1)) {
															if (compare(v4, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agefhcdb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // agefchdb
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v6, v7); // agefcdhb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agecfhdb
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // agecfdhb
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v6, v7); // agecdfhb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // agefhcbd
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agefhbcd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); // agefchbd
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // agecfhbd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agcefhdb
																			}
																			else {
																				swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // agcefdhb
																			}
																		}
																		else {
																			swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // agcefhbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v3, v4); swap(v6, v7); // agcedfhb
																		}
																		else {
																			swap(v1, v6); swap(v6, v7); // agcdefhb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acgefhdb
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v6, v7); // acgefdhb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // acgefhbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v6, v7); // acgedfhb
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v6, v7); // acgdefhb
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v6, v7); // acdgefhb
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v5); // agefcdbh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // agecfdbh
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v4); // agecdfbh
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v3, v4); swap(v4, v5); // agcefdbh
																			}
																			else {
																				swap(v1, v6); swap(v3, v4); // agcedfbh
																			}
																		}
																		else {
																			swap(v1, v6); // agcdefbh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); // acgefdbh
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v4); // acgedfbh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v6); // acgdefbh
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); // acdgefbh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // agefcbhd
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // agefcbdh
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // agefbhcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agefbchd
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // agefbcdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agecfbhd
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agcefbhd
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acgefbhd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // agecfbdh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // agcefbdh
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // acgefbdh
																			}
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v1)) {
														if (compare(v7, v4)) {
															if (compare(v3, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // aghecdbf
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v7); // aghcedbf
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v7); // aghcdebf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v3, v7); swap(v5, v7); // agchedbf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // agchdebf
																			}
																			else {
																				swap(v1, v6); swap(v4, v7); swap(v5, v7); // agcdhebf
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v7); // acghedbf
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v7); // acghdebf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v7); // acgdhebf
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v7); // acdghebf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // aghecbfd
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aghecbdf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // aghebfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aghebcfd
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // aghebcdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); // aghcebfd
																			}
																			else {
																				swap(v1, v6); swap(v3, v7); swap(v5, v6); // agchebfd
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); // acghebfd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v5, v6); swap(v6, v7); // aghcebdf
																			}
																			else {
																				swap(v1, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // agchebdf
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // acghebdf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); // agehcdbf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // agechdbf
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); // agecdhbf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // agcehdbf
																			}
																			else {
																				swap(v1, v6); swap(v3, v4); swap(v5, v7); // agcedhbf
																			}
																		}
																		else {
																			swap(v1, v6); swap(v5, v7); // agcdehbf
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v7); // acgehdbf
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v7); // acgedhbf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v5, v7); // acgdehbf
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v7); // acdgehbf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); // agehcbfd
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // agehcbdf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); // agehbfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // agehbcfd
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agehbcdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // agechbfd
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // agcehbfd
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // acgehbfd
																			}
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agechbdf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v1, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agcehbdf
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acgehbdf
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agecbhfd
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // agecbhdf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // agecdbhf
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agecbdhf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // agebhfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agebhcfd
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // agebhcdf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // agebchfd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); // agebchdf
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // agebcdhf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // agcedbhf
																		}
																		else {
																			swap(v1, v6); swap(v5, v6); swap(v6, v7); // agcdebhf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); swap(v6, v7); // acgedbhf
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v5, v6); swap(v6, v7); // acgdebhf
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); swap(v6, v7); // acdgebhf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agcebhfd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // agcebhdf
																			}
																			else {
																				swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agcebdhf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acgebhfd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // acgebhdf
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acgebdhf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // agecbfhd
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // agecbfdh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // agecdbfh
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // agecbdfh
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // agebfhcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // agebfchd
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // agebfcdh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v6, v7); // agebcfhd
																			}
																			else {
																				swap(v1, v6); swap(v2, v4); swap(v3, v6); // agebcfdh
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // agebcdfh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			swap(v1, v6); swap(v3, v4); swap(v5, v6); // agcedbfh
																		}
																		else {
																			swap(v1, v6); swap(v5, v6); // agcdebfh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v5, v6); // acgedbfh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v5, v6); // acgdebfh
																			}
																			else {
																				swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v5, v6); // acdgebfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // agcebfhd
																			}
																			else {
																				swap(v1, v6); swap(v3, v4); swap(v4, v6); // agcebfdh
																			}
																		}
																		else {
																			swap(v1, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // agcebdfh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // acgebfhd
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); // acgebfdh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // acgebdfh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v7, v4)) {
														if (compare(v7, v1)) {
															if (compare(v3, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aghcdbef
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agchdbef
																		}
																		else {
																			swap(v1, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agcdhbef
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acghdbef
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acgdhbef
																		}
																	}
																	else {
																		swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acdghbef
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); // aghcbefd
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aghcbedf
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v7); swap(v4, v6); swap(v5, v7); // aghcbdef
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // agchbefd
																			}
																			else {
																				swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agchbedf
																			}
																		}
																		else {
																			swap(v1, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // agchbdef
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // acghbefd
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acghbedf
																		}
																	}
																	else {
																		swap(v1, v2); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // acghbdef
																	}
																}
															}
														}
														else {
															if (compare(v7, v3)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agcbhefd
																			}
																			else {
																				swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // agcbhedf
																			}
																		}
																		else {
																			swap(v1, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agcbhdef
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acgbhefd
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // acgbhedf
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acgbhdef
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // acbghefd
																		}
																		else {
																			swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v7); // acbghedf
																		}
																	}
																	else {
																		swap(v1, v2); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // acbghdef
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v1, v6); swap(v4, v6); swap(v5, v7); // agcdbhef
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v7); // acgdbhef
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v7); // acdgbhef
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v7); // acdbghef
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // agcbdhef
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // acgbdhef
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v7); // acbgdhef
																		}
																		else {
																			swap(v1, v2); swap(v4, v6); swap(v5, v7); // acbdghef
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v4, v3)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // agcbehfd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v3, v6); swap(v5, v7); // agcbehdf
																			}
																			else {
																				swap(v1, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // agcbedhf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // acgbehfd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v7); // acgbehdf
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // acgbedhf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		swap(v1, v2); swap(v3, v6); swap(v5, v7); swap(v6, v7); // acbgehfd
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v3, v6); swap(v5, v7); // acbgehdf
																		}
																		else {
																			swap(v1, v2); swap(v3, v6); swap(v5, v6); swap(v6, v7); // acbgedhf
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v1, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agcdbehf
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acgdbehf
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acdgbehf
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acdbgehf
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agcbdehf
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acgbdehf
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acbgdehf
																		}
																		else {
																			swap(v1, v2); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acbdgehf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v6, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v6); swap(v3, v6); swap(v6, v7); // agcbefhd
																			}
																			else {
																				swap(v1, v6); swap(v3, v6); // agcbefdh
																			}
																		}
																		else {
																			swap(v1, v6); swap(v3, v6); swap(v5, v6); // agcbedfh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v6, v7); // acgbefhd
																			}
																			else {
																				swap(v1, v2); swap(v2, v6); swap(v3, v6); // acgbefdh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v5, v6); // acgbedfh
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v3, v6); swap(v6, v7); // acbgefhd
																		}
																		else {
																			swap(v1, v2); swap(v3, v6); // acbgefdh
																		}
																	}
																	else {
																		swap(v1, v2); swap(v3, v6); swap(v5, v6); // acbgedfh
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v6, v2)) {
																			swap(v1, v6); swap(v4, v6); swap(v5, v6); // agcdbefh
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v4, v6); swap(v5, v6); // acgdbefh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v6); swap(v4, v6); swap(v5, v6); // acdgbefh
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v4, v6); swap(v5, v6); // acdbgefh
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v2)) {
																			swap(v1, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // agcbdefh
																		}
																		else {
																			swap(v1, v2); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // acgbdefh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v3, v6); swap(v4, v6); swap(v5, v6); // acbgdefh
																		}
																		else {
																			swap(v1, v2); swap(v4, v6); swap(v5, v6); // acbdgefh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v3)) {
														if (compare(v6, v1)) {
															if (compare(v7, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v1)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v6, v7); // aghbefcd
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v7); // agbhefcd
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); // agbehfcd
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // agbefhcd
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); swap(v6, v7); // aghbecfd
																		}
																		else {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aghbcefd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v7, v4)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); // agbhecfd
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // agbehcfd
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // agbhcefd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // agbefchd
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // agbefcdh
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // agbechfd
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // agbecfhd
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); // agbecfdh
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // agbchefd
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // agbcehfd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v6, v7); // agbcefhd
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); // agbcefdh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			swap(v2, v6); swap(v3, v7); // abghefcd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v6); swap(v3, v4); swap(v4, v7); // abgehfcd
																			}
																			else {
																				swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // abgefhcd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abgefchd
																		}
																		else {
																			swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // abgefcdh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v4)) {
																			swap(v2, v6); swap(v3, v7); swap(v5, v6); // abghecfd
																		}
																		else {
																			swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // abgehcfd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abgechfd
																			}
																			else {
																				swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // abgecfhd
																			}
																		}
																		else {
																			swap(v2, v6); swap(v3, v4); swap(v4, v6); // abgecfdh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v2)) {
																			swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // abghcefd
																		}
																		else {
																			swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abgchefd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // abgcehfd
																			}
																			else {
																				swap(v2, v6); swap(v3, v6); swap(v6, v7); // abgcefhd
																			}
																		}
																		else {
																			swap(v2, v6); swap(v3, v6); // abgcefdh
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v7, v4)) {
																			swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abcghefd
																		}
																		else {
																			swap(v3, v6); swap(v5, v7); swap(v6, v7); // abcgehfd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v3, v6); swap(v6, v7); // abcgefhd
																		}
																		else {
																			swap(v3, v6); // abcgefdh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v3)) {
															if (compare(v6, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v7, v4)) {
																		if (compare(v7, v1)) {
																			swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v5, v7); // aghbecdf
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // agbhecdf
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agbehcdf
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // agbechdf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			if (compare(v4, v3)) {
																				swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v7); // aghbcedf
																			}
																			else {
																				swap(v1, v6); swap(v2, v7); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aghbcdef
																			}
																		}
																		else {
																			if (compare(v4, v3)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agbhcedf
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // agbhcdef
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // agbchedf
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v7); // agbcehdf
																			}
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // agbchdef
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v4, v2)) {
																		if (compare(v7, v4)) {
																			swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // abghecdf
																		}
																		else {
																			swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abgehcdf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abghcedf
																		}
																		else {
																			swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // abghcdef
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v4)) {
																			if (compare(v4, v3)) {
																				swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // abgchedf
																			}
																			else {
																				swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abgchdef
																			}
																		}
																		else {
																			if (compare(v4, v2)) {
																				swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // abgechdf
																			}
																			else {
																				swap(v2, v6); swap(v3, v6); swap(v5, v7); // abgcehdf
																			}
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v7, v4)) {
																				swap(v3, v6); swap(v4, v7); swap(v5, v7); // abcghedf
																			}
																			else {
																				swap(v3, v6); swap(v5, v7); // abcgehdf
																			}
																		}
																		else {
																			swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // abcghdef
																		}
																	}
																}
															}
														}
														else {
															if (compare(v4, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agbecdhf
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // agbcedhf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abgecdhf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // abgcedhf
																			}
																			else {
																				swap(v3, v6); swap(v5, v6); swap(v6, v7); // abcgedhf
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v4, v2)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // agbecdfh
																		}
																		else {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v5, v6); // agbcedfh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // abgecdfh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v2, v6); swap(v3, v6); swap(v5, v6); // abgcedfh
																			}
																			else {
																				swap(v3, v6); swap(v5, v6); // abcgedfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v6, v1)) {
																		if (compare(v7, v4)) {
																			swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // agbcdhef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // agbcdehf
																			}
																			else {
																				swap(v1, v6); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // agbcdefh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // abgcdhef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abgcdehf
																			}
																			else {
																				swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // abgcdefh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v4)) {
																			swap(v3, v6); swap(v4, v6); swap(v5, v7); // abcgdhef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abcgdehf
																			}
																			else {
																				swap(v3, v6); swap(v4, v6); swap(v5, v6); // abcgdefh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v4)) {
																			swap(v4, v6); swap(v5, v7); // abcdghef
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v4, v6); swap(v5, v6); swap(v6, v7); // abcdgehf
																			}
																			else {
																				swap(v4, v6); swap(v5, v6); // abcdgefh
																			}
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
									else {
										if (compare(v4, v0)) {
											if (compare(v5, v2)) {
												if (compare(v6, v5)) {
													if (compare(v2, v1)) {
														if (compare(v3, v1)) {
															if (compare(v7, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); swap(v6, v7); // eghfacdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eghafcdb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v6, v7); // egahfcdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eaghfcdb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v6, v7); // egfhacdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v6, v7); // egfahcdb
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egafhcdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // eagfhcdb
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egfachdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v6, v7); // egfacdhb
																		}
																	}
																	else {
																		swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); // egfacdbh
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); swap(v6, v7); // egafchdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v6, v7); // egafcdhb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); // egafcdbh
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eagfchdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // eagfcdhb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // eagfcdbh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v2)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v7); // eghfacbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v7); // eghafcbd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); // egahfcbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // eaghfcbd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); // egfhacbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); // egfahcbd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v7); // egafhcbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eagfhcbd
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v7, v1)) {
																		swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v7); // egfachbd
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egfacbhd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // egfacbdh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v7); // egafchbd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // egafcbhd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v5, v6); // egafcbdh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // eagfchbd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eagfcbhd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // eagfcbdh
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v7, v1)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		if (compare(v5, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v5); swap(v5, v6); swap(v6, v7); // eghfabcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eghafbcd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v5); swap(v5, v6); // egahfbcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); // eaghfbcd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v7); swap(v5, v6); // egfhabcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v7); swap(v5, v6); // egfahbcd
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egafhbcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // eagfhbcd
																		}
																	}
																}
															}
															else {
																if (compare(v5, v0)) {
																	if (compare(v7, v2)) {
																		swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egfabhcd
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // egfabchd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v5); swap(v3, v4); swap(v4, v6); // egfabcdh
																		}
																	}
																}
																else {
																	if (compare(v6, v0)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // egafbhcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egafbchd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // egafbcdh
																			}
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); // eagfbhcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // eagfbchd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // eagfbcdh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v0)) {
																if (compare(v7, v5)) {
																	if (compare(v7, v0)) {
																		swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v6, v7); // eghabfcd
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); // egahbfcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // egabhfcd
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egabfhcd
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // egabfchd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v5); // egabfcdh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); // eaghbfcd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); // eagbhfcd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // eagbfhcd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eagbfchd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // eagbfcdh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // eabghfcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eabgfhcd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // eabgfchd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // eabgfcdh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v3)) {
														if (compare(v7, v1)) {
															if (compare(v7, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // efghacdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); swap(v5, v6); // efghacbd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v7); // efghabcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // efgahcdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // efgahcbd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v7); // efgahbcd
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // efaghcdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // efaghcbd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // efaghbcd
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			if (compare(v3, v1)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); // eafghcdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // eafghcbd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eafghbcd
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // efgachdb
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); // efagchdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // efacghdb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // eafgchdb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eafcghdb
																		}
																	}
																}
																else {
																	if (compare(v5, v0)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efgachbd
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // efagchbd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // efacghbd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eafgchbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // eafcghbd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v0)) {
																if (compare(v2, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // efgacbhd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v6, v7); // efagcbhd
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // efacgbhd
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efacbghd
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // efgabhcd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // efgabchd
																			}
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // efagbhcd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // efagbchd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // efabghcd
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efabgchd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // efabcghd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v2, v1)) {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // eafgcbhd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eafcgbhd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // eafgbhcd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v6, v7); // eafgbchd
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eafbghcd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // eafbgchd
																			}
																		}
																		else {
																			if (compare(v2, v1)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // eafcbghd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eafbcghd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // eabfghcd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eabfgchd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // eabfcghd
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v0)) {
															if (compare(v3, v1)) {
																if (compare(v7, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efgacdhb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // efagcdhb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // efacgdhb
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // efacdghb
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // efgacdbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v5, v6); // efagcdbh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // efacgdbh
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v5, v6); // efacdgbh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); // efacdbgh
																		}
																	}
																}
															}
															else {
																if (compare(v2, v1)) {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v6); // efgacbdh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); // efagcbdh
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v6); // efacgbdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // efacbgdh
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v4); swap(v4, v5); // efacbdgh
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // efgabcdh
																		}
																		else {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v6); swap(v4, v5); // efagbcdh
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // efabgcdh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); swap(v5, v6); // efabcgdh
																			}
																			else {
																				swap(v0, v4); swap(v1, v5); swap(v2, v4); swap(v3, v5); // efabcdgh
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v6, v2)) {
																	if (compare(v3, v1)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eafgcdhb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // eafgcdbh
																		}
																	}
																	else {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // eafgcbdh
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v6); // eafgbcdh
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // eafcgdhb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eafcdghb
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // eafcgdbh
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // eafcdgbh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // eafcgbdh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v5, v1)) {
																		if (compare(v2, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // eafcbgdh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // eafbgcdh
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // eafbcgdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // eabfgcdh
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // eabfcgdh
																		}
																	}
																}
																else {
																	if (compare(v2, v1)) {
																		if (compare(v3, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // eafcdbgh
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v5); // eafcbdgh
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // eafbcdgh
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v5); // eabfcdgh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v5, v3)) {
													if (compare(v5, v1)) {
														if (compare(v3, v1)) {
															if (compare(v6, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v6, v7); // eghacfdb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v6, v7); // egahcfdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v6, v7); // egachfdb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); swap(v6, v7); // egacfhdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v6, v7); // egacfdhb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); // egacfdbh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // eaghcfdb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); // eagchfdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // eagcfhdb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eagcfdhb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // eagcfdbh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v7); // eacghfdb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // eacgfhdb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eacfghdb
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eacgfdhb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // eacfgdhb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eacfdghb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v6, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // eacgfdbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v6); // eacfgdbh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // eacfdgbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v5); // eacfdbgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); // eghacfbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); // egahcfbd
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); // egachfbd
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v7); // egacfhbd
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); // eaghcfbd
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // eagchfbd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eagcfhbd
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // eacghfbd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // eacgfhbd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // eacfghbd
																		}
																	}
																}
															}
															else {
																if (compare(v7, v3)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // egacfbhd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // eagcfbhd
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // eacgfbhd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eacfgbhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // eacfbghd
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // egacfbdh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // eagcfbdh
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v5); // eacgfbdh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // eacfgbdh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v5); swap(v5, v6); // eacfbgdh
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v5); // eacfbdgh
																			}
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v6, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); // eghacbfd
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); // egahcbfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); // egachbfd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // egacbhfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // egacbfhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); // egacbfdh
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // eaghcbfd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eagchbfd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // eagcbhfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v6, v7); // eagcbfhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); // eagcbfdh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eacghbfd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // eacgbhfd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v6, v7); // eacgbfhd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v6); // eacgbfdh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eacbghfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // eacbgfhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eacbfghd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v6); // eacbgfdh
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // eacbfgdh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v5); // eacbfdgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eghabcfd
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); // egahbcfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egabhcfd
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // egabchfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v6, v7); // egabcfhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); // egabcfdh
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); // eaghbcfd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // eagbhcfd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eagbchfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // eagbcfhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // eagbcfdh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v5)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eabghcfd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // eabgchfd
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v6, v7); // eabgcfhd
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); // eabgcfdh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eabcghfd
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // eabcgfhd
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eabcfghd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // eabcgfdh
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // eabcfgdh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // eabcfdgh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v3, v1)) {
														if (compare(v5, v1)) {
															if (compare(v6, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // eghacdfb
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); swap(v6, v7); // egahcdfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); swap(v6, v7); // egachdfb
																			}
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); swap(v6, v7); // egacdhfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v6, v7); // egacdfhb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); // egacdfbh
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eaghcdfb
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // eagchdfb
																		}
																	}
																	else {
																		if (compare(v7, v1)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eagcdhfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // eagcdfhb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // eagcdfbh
																		}
																	}
																}
															}
															else {
																if (compare(v7, v1)) {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // eacghdfb
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eacgdhfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // eacgdfhb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // eacdghfb
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v4, v6); swap(v6, v7); // eacdgfhb
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // eacdfghb
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v6); // eacgdfbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v4, v6); // eacdgfbh
																		}
																	}
																	else {
																		if (compare(v6, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v4, v5); swap(v5, v6); // eacdfgbh
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v4, v5); // eacdfbgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v7, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v7); // eghacdbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v7); // egahcdbf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v7); // egachdbf
																		}
																		else {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v7); // egacdhbf
																		}
																	}
																}
																else {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // eaghcdbf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eagchdbf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // eagcdhbf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eacghdbf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // eacgdhbf
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v4, v6); swap(v5, v7); // eacdghbf
																		}
																	}
																}
															}
															else {
																if (compare(v6, v3)) {
																	if (compare(v7, v5)) {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // egacdbhf
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eagcdbhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eacgdbhf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v5, v6); // egacdbfh
																		}
																		else {
																			if (compare(v6, v2)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // eagcdbfh
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // eacgdbfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v1)) {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eacdgbhf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v4, v6); swap(v5, v6); // eacdgbfh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v5, v6); swap(v6, v7); // eacdbghf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v5, v6); // eacdbgfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); // eacdbfgh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v2, v1)) {
															if (compare(v6, v2)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v1)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eghacbdf
																		}
																		else {
																			if (compare(v7, v2)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v5, v6); swap(v6, v7); // egahcbdf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // egachbdf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // egacbhdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egacbdhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // egacbdfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v1)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eaghcbdf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // eagchbdf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); // eagcbhdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // eagcbdhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // eagcbdfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // eacghbdf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v5, v7); // eacgbhdf
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // eacgbdhf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v6); swap(v5, v6); // eacgbdfh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // eacbghdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eacbgdhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // eacbgdfh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // eacbdghf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v3, v4); swap(v5, v6); // eacbdgfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v3, v4); // eacbdfgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v1)) {
																if (compare(v6, v0)) {
																	if (compare(v7, v2)) {
																		if (compare(v7, v0)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v7); swap(v3, v4); swap(v4, v6); swap(v5, v7); // eghabcdf
																		}
																		else {
																			if (compare(v7, v1)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // egahbcdf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // egabhcdf
																			}
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v7); // egabchdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // egabcdhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v6); swap(v2, v4); swap(v3, v6); swap(v5, v6); // egabcdfh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v7, v1)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // eaghbcdf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // eagbhcdf
																		}
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // eagbchdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eagbcdhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // eagbcdfh
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v3)) {
																		if (compare(v7, v2)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // eabghcdf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); // eabgchdf
																		}
																	}
																	else {
																		if (compare(v7, v5)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // eabgcdhf
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // eabgcdfh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // eabcghdf
																		}
																		else {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // eabcgdhf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // eabcgdfh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // eabcdghf
																			}
																			else {
																				swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // eabcdgfh
																			}
																		}
																		else {
																			swap(v0, v4); swap(v1, v4); swap(v2, v4); swap(v3, v4); // eabcdfgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
										else {
											if (compare(v6, v1)) {
												if (compare(v7, v1)) {
													if (compare(v7, v5)) {
														if (compare(v5, v1)) {
															if (compare(v6, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aeghfcdb
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v6, v7); // aeghcfdb
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aeghcdfb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); // aegchfdb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); // aegchdfb
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aegcdhfb
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); swap(v5, v6); // aeghfcbd
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v5); // aeghfbcd
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); // aeghcfbd
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v6, v7); // aegchfbd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v3, v6); swap(v4, v7); // aecghfdb
																		}
																		else {
																			swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // aecghfbd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // aecghdfb
																			}
																			else {
																				swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aecgdhfb
																			}
																		}
																		else {
																			swap(v1, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aecdghfb
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); // aceghfdb
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); // aceghdfb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // aceghfbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acegdhfb
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acedghfb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acdeghfb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v7); // aeghcdbf
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aegchdbf
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v7); // aegcdhbf
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aecghdbf
																			}
																			else {
																				swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // aecgdhbf
																			}
																		}
																		else {
																			swap(v1, v4); swap(v4, v6); swap(v5, v7); // aecdghbf
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aceghdbf
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v7); // acegdhbf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v7); // acedghbf
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v7); // acdeghbf
																		}
																	}
																}
															}
															else {
																if (compare(v7, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); // aeghcbfd
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aeghcbdf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v7); // aeghbfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); // aeghbcfd
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v7); swap(v5, v6); swap(v6, v7); // aeghbcdf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aegchbfd
																			}
																			else {
																				swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aecghbfd
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aceghbfd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v7); swap(v5, v7); // aegchbdf
																			}
																			else {
																				swap(v1, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // aecghbdf
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // aceghbdf
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v6, v5)) {
															if (compare(v6, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); // aegfhcdb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aegfchdb
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v6, v7); // aegfcdhb
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); // aegcfhdb
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aegcfdhb
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v6, v7); // aegcdfhb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aegfhcbd
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v7); swap(v5, v7); // aegfhbcd
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v7); // aegfchbd
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aegcfhbd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // aecgfhdb
																			}
																			else {
																				swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aecgfdhb
																			}
																		}
																		else {
																			swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aecgfhbd
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // aecgdfhb
																		}
																		else {
																			swap(v1, v4); swap(v4, v6); swap(v6, v7); // aecdgfhb
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); // acegfhdb
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acegfdhb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // acegfhbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v6, v7); // acegdfhb
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v6, v7); // acedgfhb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v6, v7); // acdegfhb
																		}
																	}
																}
															}
														}
														else {
															if (compare(v5, v2)) {
																if (compare(v3, v1)) {
																	if (compare(v6, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); // aefghcdb
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); // aefgchdb
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aefgcdhb
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aefcghdb
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v6, v7); // aefcgdhb
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aefcdghb
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v2, v1)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v6, v7); // aefghcbd
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aefghbcd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aefgchbd
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v7); // aefcghbd
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aecfghdb
																			}
																			else {
																				swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // aecfgdhb
																			}
																		}
																		else {
																			swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // aecfghbd
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aecfdghb
																		}
																		else {
																			swap(v1, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aecdfghb
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v3, v1)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acefghdb
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v6, v7); // acefgdhb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // acefghbd
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v5, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acefdghb
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acedfghb
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acdefghb
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v5, v1)) {
														if (compare(v6, v5)) {
															if (compare(v3, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); // aegfcdbh
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v5, v6); // aegcfdbh
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); // aegcdfbh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // aecgfdbh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v3, v6); swap(v4, v6); // aecgdfbh
																			}
																			else {
																				swap(v1, v4); swap(v4, v6); // aecdgfbh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v6); // acegfdbh
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); // acegdfbh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v4, v6); // acedgfbh
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); // acdegfbh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aegfcbhd
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v4, v6); swap(v5, v6); // aegfcbdh
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v7); // aegfbhcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); swap(v6, v7); // aegfbchd
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v5); swap(v5, v6); // aegfbcdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); swap(v6, v7); // aegcfbhd
																			}
																			else {
																				swap(v1, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // aecgfbhd
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // acegfbhd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v5); // aegcfbdh
																			}
																			else {
																				swap(v1, v4); swap(v3, v6); swap(v4, v5); // aecgfbdh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v5); // acegfbdh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v3, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v5, v6); // aefgcdbh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); // aefcgdbh
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); swap(v5, v6); // aefcdgbh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v3, v5); swap(v4, v6); // aecfgdbh
																			}
																			else {
																				swap(v1, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // aecfdgbh
																			}
																		}
																		else {
																			swap(v1, v4); swap(v4, v5); swap(v5, v6); // aecdfgbh
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v6, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); // acefgdbh
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); swap(v5, v6); // acefdgbh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v4, v5); swap(v5, v6); // acedfgbh
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); swap(v5, v6); // acdefgbh
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v2, v1)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); swap(v6, v7); // aefgcbhd
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v4, v5); // aefgcbdh
																		}
																	}
																	else {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v5, v7); swap(v6, v7); // aefgbhcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v6); swap(v6, v7); // aefgbchd
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v6); // aefgbcdh
																			}
																		}
																	}
																}
																else {
																	if (compare(v7, v3)) {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aefcgbhd
																			}
																			else {
																				swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aecfgbhd
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acefgbhd
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v5, v2)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v6); swap(v5, v6); // aefcgbdh
																			}
																			else {
																				swap(v1, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // aecfgbdh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // acefgbdh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v7, v5)) {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v7, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); swap(v6, v7); // aegcbhfd
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v7); // aegcbhdf
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aegcdbhf
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); swap(v6, v7); // aegcbdhf
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v2)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); // aegbhfcd
																		}
																		else {
																			if (compare(v5, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); // aegbhcfd
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v7); swap(v5, v6); swap(v6, v7); // aegbhcdf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aegbchfd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v7); // aegbchdf
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aegbcdhf
																			}
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aecgdbhf
																		}
																		else {
																			swap(v1, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aecdgbhf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acegdbhf
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acedgbhf
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acdegbhf
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // aecgbhfd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v6); swap(v5, v7); // aecgbhdf
																			}
																			else {
																				swap(v1, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // aecgbdhf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // acegbhfd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v7); // acegbhdf
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // acegbdhf
																			}
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v2, v1)) {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v6, v7); // aegcbfhd
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v6); // aegcbfdh
																		}
																	}
																	else {
																		if (compare(v3, v1)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v4, v6); swap(v5, v6); // aegcdbfh
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v6); swap(v5, v6); // aegcbdfh
																		}
																	}
																}
																else {
																	if (compare(v5, v2)) {
																		if (compare(v7, v2)) {
																			swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v7); // aegbfhcd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aegbfchd
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v5); swap(v5, v6); // aegbfcdh
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v6, v7); // aegbcfhd
																			}
																			else {
																				swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); // aegbcfdh
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v6); swap(v3, v4); swap(v4, v6); swap(v5, v6); // aegbcdfh
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // aecgdbfh
																		}
																		else {
																			swap(v1, v4); swap(v4, v6); swap(v5, v6); // aecdgbfh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			if (compare(v6, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v4, v6); swap(v5, v6); // acegdbfh
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v4, v6); swap(v5, v6); // acedgbfh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v6); swap(v5, v6); // acdegbfh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v6); swap(v6, v7); // aecgbfhd
																			}
																			else {
																				swap(v1, v4); swap(v3, v6); // aecgbfdh
																			}
																		}
																		else {
																			swap(v1, v4); swap(v3, v6); swap(v5, v6); // aecgbdfh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v6, v7); // acegbfhd
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v6); // acegbfdh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v6); swap(v5, v6); // acegbdfh
																		}
																	}
																}
															}
														}
													}
												}
											}
											else {
												if (compare(v2, v1)) {
													if (compare(v6, v5)) {
														if (compare(v7, v5)) {
															if (compare(v6, v3)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aecbghfd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // aecbghdf
																			}
																			else {
																				swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aecbgdhf
																			}
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acebghfd
																		}
																		else {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // acebghdf
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acebgdhf
																			}
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // acbeghfd
																	}
																	else {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v7); // acbeghdf
																		}
																		else {
																			swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // acbegdhf
																		}
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v4); swap(v5, v6); swap(v6, v7); // aecdbghf
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v5, v6); swap(v6, v7); // acedbghf
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); swap(v6, v7); // acdebghf
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v5, v6); swap(v6, v7); // acdbeghf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v1, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // aecbdghf
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // acebdghf
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v3, v4); swap(v5, v6); swap(v6, v7); // acbedghf
																		}
																		else {
																			swap(v1, v2); swap(v5, v6); swap(v6, v7); // acbdeghf
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // aecbgfhd
																			}
																			else {
																				swap(v1, v4); swap(v3, v4); swap(v4, v6); // aecbgfdh
																			}
																		}
																		else {
																			swap(v1, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // aecbgdfh
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // acebgfhd
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); // acebgfdh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // acebgdfh
																		}
																	}
																}
																else {
																	if (compare(v5, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v6, v7); // acbegfhd
																		}
																		else {
																			swap(v1, v2); swap(v3, v4); swap(v4, v6); // acbegfdh
																		}
																	}
																	else {
																		swap(v1, v2); swap(v3, v4); swap(v4, v6); swap(v5, v6); // acbegdfh
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v4); swap(v5, v6); // aecdbgfh
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v5, v6); // acedbgfh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v5, v6); // acdebgfh
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); swap(v5, v6); // acdbegfh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v1, v4); swap(v3, v4); swap(v5, v6); // aecbdgfh
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v5, v6); // acebdgfh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v3, v4); swap(v5, v6); // acbedgfh
																		}
																		else {
																			swap(v1, v2); swap(v5, v6); // acbdegfh
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v1)) {
															if (compare(v3, v1)) {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v4, v5); // aefcdbgh
																	}
																	else {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v3, v5); swap(v4, v5); // aecfdbgh
																		}
																		else {
																			swap(v1, v4); swap(v4, v5); // aecdfbgh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v5, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v4, v5); // acefdbgh
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v4, v5); // acedfbgh
																		}
																	}
																	else {
																		swap(v1, v2); swap(v2, v3); swap(v3, v4); swap(v4, v5); // acdefbgh
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v5, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); swap(v6, v7); // aefcbghd
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v5); swap(v5, v6); // aefcbgdh
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v5); // aefcbdgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // aecfbghd
																			}
																			else {
																				swap(v1, v4); swap(v3, v5); swap(v5, v6); // aecfbgdh
																			}
																		}
																		else {
																			swap(v1, v4); swap(v3, v5); // aecfbdgh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // acefbghd
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v5); swap(v5, v6); // acefbgdh
																		}
																	}
																	else {
																		swap(v1, v2); swap(v2, v4); swap(v3, v5); // acefbdgh
																	}
																}
															}
														}
														else {
															if (compare(v5, v3)) {
																if (compare(v4, v1)) {
																	if (compare(v4, v2)) {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aecbfghd
																			}
																			else {
																				swap(v1, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // aecbfgdh
																			}
																		}
																		else {
																			swap(v1, v4); swap(v3, v4); swap(v4, v5); // aecbfdgh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v7, v3)) {
																				swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acebfghd
																			}
																			else {
																				swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // acebfgdh
																			}
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v4); swap(v4, v5); // acebfdgh
																		}
																	}
																}
																else {
																	if (compare(v6, v3)) {
																		if (compare(v7, v3)) {
																			swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // acbefghd
																		}
																		else {
																			swap(v1, v2); swap(v3, v4); swap(v4, v5); swap(v5, v6); // acbefgdh
																		}
																	}
																	else {
																		swap(v1, v2); swap(v3, v4); swap(v4, v5); // acbefdgh
																	}
																}
															}
															else {
																if (compare(v3, v1)) {
																	if (compare(v4, v3)) {
																		if (compare(v4, v2)) {
																			swap(v1, v4); // aecdbfgh
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); // acedbfgh
																		}
																	}
																	else {
																		if (compare(v4, v1)) {
																			swap(v1, v2); swap(v2, v3); swap(v3, v4); // acdebfgh
																		}
																		else {
																			swap(v1, v2); swap(v2, v3); // acdbefgh
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v4, v2)) {
																			swap(v1, v4); swap(v3, v4); // aecbdfgh
																		}
																		else {
																			swap(v1, v2); swap(v2, v4); swap(v3, v4); // acebdfgh
																		}
																	}
																	else {
																		if (compare(v4, v3)) {
																			swap(v1, v2); swap(v3, v4); // acbedfgh
																		}
																		else {
																			swap(v1, v2); // acbdefgh
																		}
																	}
																}
															}
														}
													}
												}
												else {
													if (compare(v7, v3)) {
														if (compare(v4, v1)) {
															if (compare(v5, v2)) {
																if (compare(v7, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v7, v5)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // aebghfcd
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // aebgfhcd
																		}
																	}
																	else {
																		if (compare(v5, v1)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aefbghcd
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // aebfghcd
																		}
																	}
																}
																else {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); swap(v6, v7); // aefbgchd
																		}
																		else {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aefbcghd
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // aebgfchd
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aebfgchd
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // aebfcghd
																		}
																	}
																}
															}
															else {
																if (compare(v6, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // aebghcfd
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // aebghcdf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // aebgchfd
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v6, v7); // aebgcfhd
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v7); // aebgchdf
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // aebcghfd
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // aebcghdf
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // aebcgfhd
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // aebcfghd
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v2)) {
																if (compare(v5, v2)) {
																	if (compare(v7, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v6, v7); // abeghfcd
																			}
																			else {
																				swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v5, v7); swap(v6, v7); // abegfhcd
																			}
																		}
																		else {
																			swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v7); // abefghcd
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v2, v4); swap(v3, v6); swap(v4, v5); swap(v6, v7); // abegfchd
																		}
																		else {
																			swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abefgchd
																		}
																	}
																}
																else {
																	if (compare(v7, v2)) {
																		if (compare(v5, v3)) {
																			swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); swap(v6, v7); // abeghcfd
																		}
																		else {
																			swap(v2, v4); swap(v3, v6); swap(v4, v7); swap(v5, v7); // abeghcdf
																		}
																	}
																	else {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v2, v4); swap(v3, v6); swap(v5, v7); swap(v6, v7); // abegchfd
																			}
																			else {
																				swap(v2, v4); swap(v3, v6); swap(v6, v7); // abegcfhd
																			}
																		}
																		else {
																			swap(v2, v4); swap(v3, v6); swap(v5, v7); // abegchdf
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v6, v5)) {
																		if (compare(v5, v3)) {
																			if (compare(v7, v5)) {
																				swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abecghfd
																			}
																			else {
																				swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v6, v7); // abecgfhd
																			}
																		}
																		else {
																			swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v7); // abecghdf
																		}
																	}
																	else {
																		if (compare(v5, v2)) {
																			swap(v2, v4); swap(v3, v5); swap(v5, v6); swap(v6, v7); // abefcghd
																		}
																		else {
																			swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abecfghd
																		}
																	}
																}
																else {
																	if (compare(v7, v5)) {
																		if (compare(v5, v3)) {
																			swap(v3, v4); swap(v4, v6); swap(v5, v7); swap(v6, v7); // abceghfd
																		}
																		else {
																			swap(v3, v4); swap(v4, v6); swap(v5, v7); // abceghdf
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			swap(v3, v4); swap(v4, v6); swap(v6, v7); // abcegfhd
																		}
																		else {
																			swap(v3, v4); swap(v4, v5); swap(v5, v6); swap(v6, v7); // abcefghd
																		}
																	}
																}
															}
														}
													}
													else {
														if (compare(v5, v3)) {
															if (compare(v4, v1)) {
																if (compare(v5, v2)) {
																	if (compare(v5, v1)) {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v6); // aefbgcdh
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); swap(v5, v6); // aefbcgdh
																			}
																			else {
																				swap(v1, v4); swap(v2, v5); swap(v3, v4); swap(v4, v5); // aefbcdgh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v2)) {
																			if (compare(v6, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v4, v5); // aebgfcdh
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // aebfgcdh
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); swap(v5, v6); // aebfcgdh
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v5); // aebfcdgh
																			}
																		}
																	}
																}
																else {
																	if (compare(v6, v5)) {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); // aebgcfdh
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); // aebcgfdh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // aebcfgdh
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v5); // aebcfdgh
																		}
																	}
																}
															}
															else {
																if (compare(v5, v2)) {
																	if (compare(v6, v2)) {
																		if (compare(v6, v5)) {
																			swap(v2, v4); swap(v3, v6); swap(v4, v5); // abegfcdh
																		}
																		else {
																			swap(v2, v4); swap(v3, v5); swap(v4, v6); swap(v5, v6); // abefgcdh
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			swap(v2, v4); swap(v3, v5); swap(v5, v6); // abefcgdh
																		}
																		else {
																			swap(v2, v4); swap(v3, v5); // abefcdgh
																		}
																	}
																}
																else {
																	if (compare(v4, v2)) {
																		if (compare(v6, v5)) {
																			if (compare(v6, v2)) {
																				swap(v2, v4); swap(v3, v6); // abegcfdh
																			}
																			else {
																				swap(v2, v4); swap(v3, v4); swap(v4, v6); // abecgfdh
																			}
																		}
																		else {
																			if (compare(v6, v3)) {
																				swap(v2, v4); swap(v3, v4); swap(v4, v5); swap(v5, v6); // abecfgdh
																			}
																			else {
																				swap(v2, v4); swap(v3, v4); swap(v4, v5); // abecfdgh
																			}
																		}
																	}
																	else {
																		if (compare(v6, v3)) {
																			if (compare(v6, v5)) {
																				swap(v3, v4); swap(v4, v6); // abcegfdh
																			}
																			else {
																				swap(v3, v4); swap(v4, v5); swap(v5, v6); // abcefgdh
																			}
																		}
																		else {
																			swap(v3, v4); swap(v4, v5); // abcefdgh
																		}
																	}
																}
															}
														}
														else {
															if (compare(v6, v3)) {
																if (compare(v7, v5)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // aebgcdhf
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // aebcgdhf
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v2, v4); swap(v3, v6); swap(v5, v6); swap(v6, v7); // abegcdhf
																			}
																			else {
																				swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abecgdhf
																			}
																		}
																		else {
																			swap(v3, v4); swap(v4, v6); swap(v5, v6); swap(v6, v7); // abcegdhf
																		}
																	}
																}
																else {
																	if (compare(v4, v1)) {
																		if (compare(v6, v2)) {
																			swap(v1, v4); swap(v2, v4); swap(v3, v6); swap(v5, v6); // aebgcdfh
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // aebcgdfh
																		}
																	}
																	else {
																		if (compare(v4, v2)) {
																			if (compare(v6, v2)) {
																				swap(v2, v4); swap(v3, v6); swap(v5, v6); // abegcdfh
																			}
																			else {
																				swap(v2, v4); swap(v3, v4); swap(v4, v6); swap(v5, v6); // abecgdfh
																			}
																		}
																		else {
																			swap(v3, v4); swap(v4, v6); swap(v5, v6); // abcegdfh
																		}
																	}
																}
															}
															else {
																if (compare(v4, v2)) {
																	if (compare(v4, v1)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // aebcdghf
																			}
																			else {
																				swap(v1, v4); swap(v2, v4); swap(v3, v4); swap(v5, v6); // aebcdgfh
																			}
																		}
																		else {
																			swap(v1, v4); swap(v2, v4); swap(v3, v4); // aebcdfgh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v2, v4); swap(v3, v4); swap(v5, v6); swap(v6, v7); // abecdghf
																			}
																			else {
																				swap(v2, v4); swap(v3, v4); swap(v5, v6); // abecdgfh
																			}
																		}
																		else {
																			swap(v2, v4); swap(v3, v4); // abecdfgh
																		}
																	}
																}
																else {
																	if (compare(v4, v3)) {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v3, v4); swap(v5, v6); swap(v6, v7); // abcedghf
																			}
																			else {
																				swap(v3, v4); swap(v5, v6); // abcedgfh
																			}
																		}
																		else {
																			swap(v3, v4); // abcedfgh
																		}
																	}
																	else {
																		if (compare(v6, v5)) {
																			if (compare(v7, v5)) {
																				swap(v5, v6); swap(v6, v7); // abcdeghf
																			}
																			else {
																				swap(v5, v6); // abcdegfh
																			}
																		}
																		else {
																			// abcdefgh
																		}
																	}
																}
															}
														}
													}
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		};

	} // detail

	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	// fastest_sort_w8<T, Compare>
	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	template<typename T, typename Compare = std::less<>>
	class fastest_sort_w8
	{
	public:

		static constexpr size_t max_size() noexcept
		{
			return 8;
		}

		template<typename Iterator>
		static void sort(const Iterator &first, const Iterator &last, const Compare &compare = Compare{})
		{
			using Impl = detail::FastestSortW8Impl<T, Compare>;
			Iterator i = first;
			if (i == last) {
				return;
			}
			T &v0 = *i; ++i;
			if (i == last) {
				return;
			}
			T &v1 = *i; ++i;
			if (i == last) {
				Impl::stable_sort2(v0, v1, compare);
				return;
			}
			T &v2 = *i; ++i;
			if (i == last) {
				Impl::stable_sort3(v0, v1, v2, compare);
				return;
			}
			T &v3 = *i; ++i;
			if (i == last) {
				Impl::stable_sort4(v0, v1, v2, v3, compare);
				return;
			}
			T &v4 = *i; ++i;
			if (i == last) {
				Impl::stable_sort5(v0, v1, v2, v3, v4, compare);
				return;
			}
			T &v5 = *i; ++i;
			if (i == last) {
				Impl::stable_sort6(v0, v1, v2, v3, v4, v5, compare);
				return;
			}
			T &v6 = *i; ++i;
			if (i == last) {
				Impl::stable_sort7(v0, v1, v2, v3, v4, v5, v6, compare);
				return;
			}
			T &v7 = *i; ++i;
			if (i == last) {
				Impl::stable_sort8(v0, v1, v2, v3, v4, v5, v6, v7, compare);
				return;
			}
			assert(false && "Too many input data. Please add more source code.");
		}
	};

	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	// sort2(v0, v1, compare)
	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	template<typename T, typename Compare = std::less<>>
	inline void sort2(T &v0, T &v1, const Compare &compare = Compare{})
	{
		using Impl = detail::FastestSortW8Impl<T, Compare>;
		Impl::stable_sort2(v0, v1, compare);
	}

	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	// sort3(v0, v1, v2, compare)
	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	template<typename T, typename Compare = std::less<>>
	inline void sort3(T &v0, T &v1, T &v2, const Compare &compare = Compare{})
	{
		using Impl = detail::FastestSortW8Impl<T, Compare>;
		Impl::stable_sort3(v0, v1, v2, compare);
	}

	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	// sort4(v0, v1, v2, v3, compare)
	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	template<typename T, typename Compare = std::less<>>
	inline void sort4(T &v0, T &v1, T &v2, T &v3, const Compare &compare = Compare{})
	{
		using Impl = detail::FastestSortW8Impl<T, Compare>;
		Impl::stable_sort4(v0, v1, v2, v3, compare);
	}

	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	// sort5(v0, v1, v2, v3, v4, compare)
	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	template<typename T, typename Compare = std::less<>>
	inline void sort5(T &v0, T &v1, T &v2, T &v3, T &v4, const Compare &compare = Compare{})
	{
		using Impl = detail::FastestSortW8Impl<T, Compare>;
		Impl::stable_sort5(v0, v1, v2, v3, v4, compare);
	}

	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	// sort6(v0, v1, v2, v3, v4, v5, compare)
	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	template<typename T, typename Compare = std::less<>>
	inline void sort6(T &v0, T &v1, T &v2, T &v3, T &v4, T &v5, const Compare &compare = Compare{})
	{
		using Impl = detail::FastestSortW8Impl<T, Compare>;
		Impl::stable_sort6(v0, v1, v2, v3, v4, v5, compare);
	}

	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	// sort7(v0, v1, v2, v3, v4, v5, v6, compare)
	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	template<typename T, typename Compare = std::less<>>
	inline void sort7(T &v0, T &v1, T &v2, T &v3, T &v4, T &v5, T &v6, const Compare &compare = Compare{})
	{
		using Impl = detail::FastestSortW8Impl<T, Compare>;
		Impl::stable_sort7(v0, v1, v2, v3, v4, v5, v6, compare);
	}

	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	// sort8(v0, v1, v2, v3, v4, v5, v6, v7, compare)
	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	template<typename T, typename Compare = std::less<>>
	inline void sort8(T &v0, T &v1, T &v2, T &v3, T &v4, T &v5, T &v6, T &v7, const Compare &compare = Compare{})
	{
		using Impl = detail::FastestSortW8Impl<T, Compare>;
		Impl::stable_sort8(v0, v1, v2, v3, v4, v5, v6, v7, compare);
	}

	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	// sort(first, last, compare)
	//---+---------+---------+---------+---------+---------+---------+---------+---------+---------+
	template<typename Iterator, typename Compare = std::less<>>
	inline void sort(const Iterator &first, const Iterator &last, const Compare &compare = Compare{})
	{
		using Value = typename std::iterator_traits<Iterator>::value_type;
		fastest_sort_w8<Value, Compare>::sort(first, last, compare);
	}

} // fsw8
